/*****************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_baslccmerge.sas                                             
 *  TITLE:   Merges the input dataset with the Cross Classification and 
 *           Basel Mapping Datasets.  This is called for Scenario datasets
 *           since they can contain STANDARD CCV instead of Internal.
 *  SUPPORT: stchri
 *
 * PARAMETERS:
 *
 *  numCCVs       = number of Cross Classification variables, defaults to 2.
 *  data          =  the name of the input dataset with Standard Columns.
 *  mapCols       = space-separated list of the CCV Columns. 
 *                  ex: BASEL_BL_UID BASEL_RC_UID in &data.
 *  baselUids     = space-separated list of the CCV_VALUE_UID from 'ccvValuesData'
 *                  for the Mapping Columns. Pass -1 for non-Standard CCVs.
 *                  Ex: 1119 1121 -1
 *  ccData        = the name of the Cross Classification UID dataset with the following variables:
 *                  CCUID, CCVUID_<n> (one for each Cross Classification)
 *  ccvValuesData = the name of the CCV_VALUES dataset (usually VARCOMP.ccv_values)
 *  varcomp       = libref for the comp folder (e.g. vc1)
 *  dataOut       = the name of the output dataset.
 *  ostat         = name of the out status dataset.  Will contain the following:
 *                  status = 0 on success, 1 otherwise.
 *                  reason = printable message.
 *      
 *  Example: %orva_baslccmerge(numCCVs=2,
 *                         data=libname.rawlossi,
 *                         mapCols=BASEL_BL_UID BASEL_RC_UID,
 *                         baselUids=1119 1121
 *                         ccData=VARPROJ.CC,
 *                         ccValuesData=VARCOMP.ccv_values
 *                         dataOut=libname.lossi,
 *                         ostat=status);
 * 
 *****************************************************************************************/
%macro orva_baslccmerge(
                        numCCVs      = 2,
                        data         = ,
                        mapCols      = ,
                        baselUids    = ,
                        ccData       = ,
                        ccValuesData = ,
                        varcomp      = ,
                        dataOut      = ,
                        bdDataOut    = ,
                        oStat        = status
                        );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_baslccmerge(;
      %put .                  numCCVs      = &numCCVs,;
      %put .                  data         = &data,;
      %put .                  mapCols      = &mapCols,;
      %put .                  baselUids    = &baselUids,;
      %put .                  ccData       = &ccData,;
      %put .                  ccValuesData = &ccValuesData,;
      %put .                  dataOut      = &dataOut,;
      %put .                  bdDataOut    = &bdDataOut,;
      %put .                  oStat        = &oStat;
      %put .                  );
   %end;

   /*- create the status dataset up front with Unknown status -*/
   %let reason = %sysfunc(sasmsg(&orvads, bac_unknown_error, noquote));
   %orva_status_ds( statusds = &ostat, status = 1, reason = &reason );
   
   %let status = 0;
   %let reason = %sysfunc(sasmsg(&orvads, bac_successful_note, noquote));
   %let abort  = 0;

   /*- check input parameters -*/
   %if "&data" = "" or "&ccData"="" or "&mapCols"="" or "&ccValuesData"="" %then %do;
     %put &orvar_error %sysfunc(sasmsg(&orvads, bac_fields_to_specify_title, noquote));
     %let abort  = 1;
     %let reason = %sysfunc(sasmsg(&orvads, bac_missing_params_error, noquote));
     %goto DONE;
   %end;

   /*- get 'baselUids' into macro variables: CCV_UID_i -*/
   %let STANDARD_FOUND = false;

   %if %quote(&baselUids.) ne %str() %then %do;
      data _NULL_;      
         %let nmaps = 1;
         %let map = %scan(&baselUids, &nmaps, ' ');

         %do %while (&map ne );
             %if &map ne -1 %then %do;
               call symput("STANDARD_FOUND", "true");
             %end;
             %let i = %eval(&nmaps-1);
             call symput("CCV_UID_&i", &map);
             %let nmaps = %eval(&nmaps+1);
             %let map = %scan(&baselUids, &nmaps, ' ');
         %end;     
      run;     
   %end;
   %if &STANDARD_FOUND eq false %then %do;
      /*- let ccmerge do the work since no BASEL columns chosen -*/
      %let ccvProp =;
       %checkUseCCVProperty() ;
      %if &ccvProp eq true %then %do;
          /*in case of scenario uses ccv properties as dimension*/
         %orva_ccmerge(
                        numCCVs       = &numCCVs,
                        data          = &data,
                        mapCols       = &mapCols, 
                        ccData        = &ccData,
                        mapDataPrefix = &mapDataPrefix,
                        dataOut       = &dataOut, 
                        ostat         = &oStat
                        );
      %end;
      %else %do;
         /*In case of scenario uses ccv as dimension*/

         /* Need to get the right mapping columns.  mapCols has the CCV
            property - we want the CCV */
         %let CCVCol_uids =;
         %let CCVCol_names =;
         %let model_ccv_mapping_ds = ;
         %orvu_find_dataset(
                            nameIn  = model_ccv_mapping, 
                            Libs    = varprj varprj_m, 
                            nameOut = model_ccv_mapping_ds
                            );

         data _ccv_names_lookup;
            set &modeL_ccv_mapping_ds.;
            keep ccv_uid;
         run;

         /* Get CCV names - the order of the CCVs needs to be preserved */
         data _ccv_names_lookup;
           set _ccv_names_lookup;
           length CCV_FIELD_REF $40.;
           if _N_ eq 1 then do;
               declare hash ccvuidHash(dataset: "&varcomp..CCV");
               rc = ccvuidHash.defineKey("ccv_uid");
               rc = ccvuidHash.defineData('CCV_FIELD_REF');
               rc = ccvuidHash.defineDone();    
           end;
           rc  = ccvuidHash.find();
         run;

         proc sql noprint;
            select 
                     CCV_FIELD_REF
            into
                     :CCVCol_names separated by " "
            from
                     _ccv_names_lookup
            ;
         quit;

         %orva_map_scenario_ccv_dim(
                                    numCCVs       = &numCCVs.,
                                    data          = &data.,
                                    mapCols       = &CCVCol_names., 
                                    ccData        = &ccData., 
                                    mapDataPrefix = &mapDataPrefix.,
                                    dataOut       = &dataOut., 
                                    ostat         = &oStat.
                                    );
      %end;

      %goto DONE;
   %end;
   
   /*- get 'mapCols' into macro variables: CCV_NAME_i -*/
   data _NULL_;      
         %let nmaps = 1;
         %let map = %scan(&mapCols, &nmaps, ' ');

         %do %while (&map ne );
           %let i = %eval(&nmaps - 1);
           call symput("CCV_NAME_&i", "&map");
           %let nmaps = %eval(&nmaps + 1);
           %let map = %scan(&mapCols, &nmaps, ' ');
       %end;        
   run;
   
   data _work; 
      set &data; 
   run;
   
   /*- create Basel Mapping datasets -*/
   %do i = 0 %to &numCCVs - 1;
      %if &&CCV_UID_&i ne -1 %then %do;
         data _baselmap_&i; 
             set &ccValuesData (rename= CCV_VALUE_UID=CCVUID_&i);
             where CCV_UID = &&CCV_UID_&i;
             &&CCV_NAME_&i = input(CCV_DB_VALUE, best.);
             drop CCV_DB_VALUE CCV_NAME CCV_UID VERSION_UID;
         run;
      %end;
      %else %do;
         data _work; set _work;       
           CCMAP_&i = &&CCV_NAME_&i;
         run;
      %end;
   %end;
       
   /*- merge each Basel Mapping Datasets into the Input Dataset -*/
   %do i = 0 %to &numCCVs - 1; 
      %if &&CCV_UID_&i ne - 1 %then %do;
         proc sql noprint; 
            CREATE TABLE 
                           _work&i
            AS SELECT 
                           * 
            FROM
                           _work L, 
                           _baselmap_&i M
            WHERE 
                           L.&&CCV_NAME_&i = M.&&CCV_NAME_&i
                           ;   
      quit;

      data _work; 
         set _work&i; 
      run;
   %end;
   %else %do;
      proc sql noprint; 
         CREATE TABLE 
                        _work&i
         AS SELECT 
                        * 
         FROM 
                        _work L 
         LEFT JOIN 
                        &mapDataPrefix.&i M
               ON 
                        L.ccmap_&i = M.ccmap_&i
                        ;   
      quit;

      data _work; 
         set _work&i; 
         drop ccmap_&i; 
         if ccvuid_&i eq . then ccvuid_&i=0;
      run;
      %end; 
   %end;

   /*- merge the input Dataset with the Cross Classification Dataset -*/
   proc sql noprint; 
      CREATE TABLE 
                     &dataOut.
      AS SELECT 
                     * 
      FROM 
                     _work L, 
                     &ccData C
      WHERE 
                     L.ccvuid_0 =  C.ccvuid_0
                     %do i = 1 %to &numCCVs - 1; 
                        AND 
                           L.ccvuid_&i = C.ccvuid_&i
                     %end;
                     ;   
   quit;
 
%DONE:
   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, bac_abnormal_error, noquote));
      %let status = 1;    
   %end;   
   %orva_status_ds( statusds = &ostat, status = &status, reason = &reason );
     
   /*- delete unwanted data sets -*/
   %orvu_delete_data(data =
                           _work: 
                           _baselmap_: 
                           _ccv_names_lookup
                           );
%mend orva_baslccmerge;

/* checkUseCCVProperty macro is to check if the scenario uses the ccv property (vs CCV) as dimension.
 * this is done by check the signs on UID's on scenario_details where TYPE_CD="C". For CCV 
 * properties, the value is negative. The number(1,2 or 3) indicates which dimension it is 
 * on. It sets ccvProp to true if ccv property is used. Otherwise false. 
 */
%macro checkUseCCVProperty();
   data _null_;
      set &varorgLibref..SCENARIO_DETAILS (keep = 
                                                   &mapCols 
                                                   parent_uid 
                                                   type_cd
                                                   );
      where parent_uid = &scenarioUid and type_cd = 'C';

      %let nmaps = 1;
      %let map = %scan(&mapCols, &nmaps, %str( ));

      if &map lt 0 then do;
         call symputX("ccvProp", "true");
      end;
      else do;
        call symputX("ccvProp", "false");
      end;
   run;
%mend checkUseCCVProperty;
