/*****************************************************************************************
 *
 * Copyright (C), 2011
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orva_extra_loss                                      
 *       
 * PURPOSE: Approximates the effect of an additional large loss on VaR.  Meant to be called in 
 *          orva_simulate AFTER orva_get_model_info has been called. Results are stored in
 *          varstat and varstat_ins datasets.
 *
 *
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *     ccuid       - the CCUID this is being run for
 *     spriority   - simpriority_sev dataset
 *     fpriority   - simpriority_freq dataset
 *     varstat     - varstat dataset to write results to
 *     varstat_ins - varstat_ins dataset to write results to
 *
 * NOTES:  None
 *
 *****************************************************************************************/
%macro orva_extra_loss( 
                        ccUid         = ,
                        spriority     = ,
                        fpriority     = ,
                        varstat       = ,
                        varstat_ins   = 
                       );

   %if %upcase(&ORV_DEBUG.) EQ TRUE %then %do;
      %put #orva_extra_loss(; 
      %put .                 ccUid         = &ccUid.,;
      %put .                 spriority     = &spriority.,;
      %put .                 fpriority     = &fpriority.,;
      %put .                 varstat       = &varstat.,;
      %put .                 varstat_ins   = &varstat_ins.;
      %put .                 );
   %end;

   %local 
      found_error
      sev_scales_mod
      smean_mod
      MAX_SCALE_mod
      min_pi_mod
      var_value
      extreme_loss
      temp_MAX_SCALE
      temp_MIN_SCALE
      temp_scales
      ;

   /*****************************************************************/
   /*        Initialize the state to unknown error                  */
   /*****************************************************************/
   %let ORV_REASON  = unknown_error;
   %let ORV_ABORT   = &ORV_FALSE;
   %let found_error = &ORV_TRUE;
   
   data _ex_results;
      P_1_delta = .;
      P_2_delta = .;
      P_3_delta = .;
   run;

   %if &CAN_PROCEED. eq &ORV_FALSE. %then %do;
      %let ORV_ABORT   = &ORV_TRUE.;
      %let found_error = &ORV_FALSE.;
      %goto DONE;
   %end;

   /*****************************************************************/
   /*           Get the severity mean and parameters                */
   /*****************************************************************/
   data _ex_sev_parms;
      set &Sparm_dataset_name;
      where CCUID = &CCUID;

      array scales{*} scales1 - scales&nsev_mix. (&sev_scales.);
      array weights{*} weights1 - weights&nsev_mix. (&SPARAM1_WEIGHTS.);
      array parms param_:;
      smean = 0;
      true_param_1 = param_1;

      /* Get the mean for the mixture and param_1 for each distribution
         in the mixture */
      do i = 1 to &nsev_mix.;
         %if %sysevalf(&SCALE_TRANSFORM eq ID, boolean) %then %do; 
            param_1 = scales[i];
         %end;
         %else %do;
            param_1 = log(scales[i]);
         %end;
         
         %Moments&SDIST;
         smean = smean + weights[i]*mean;
      end;

      param_1 = true_param_1;

      if count_kri = . then do;
         count_kri = 0;
      end;

      N_S = obsCount;

      /* Get the severity parameters as macro variables (needed for quant&dist macro )*/
      do i = 1 to dim(parms);
         call symputX(vname(parms[i]), parms[i]);
      end;

      keep 
          smean 
          param_:
          count_KRI
          paramcount
          N_S
          ;
   run;

   /*****************************************************************/
   /*         Get the Frequency Distribution Information            */
   /*****************************************************************/
   data _ex_freq_parms;
      set &Fparm_dataset_name;
      where CCUID = &CCUID;

      FParam_1 = &FParam_1;
      %Moments&Fdist.();

      fmean     = mean;
      fvariance = variance;
      N_F       = Freq_ObsCount;
      %VCDF&fdist.(0, mass_at_0);

      /* If the frequency mean is less than 1, then we have
         to worry about ecdf computed below becoming negative. */
      if fmean < 1 then do;
         /* Get original non-central moments to then compute the
            new mean and variance */
         m1 = fmean;
         m2 = fvariance + m1**2;

         m1_new = m1/(1 - mass_at_0);
         m2_new = m2/(1 - mass_at_0);

         fmean     = m1_new;
         fvariance = m2_new - m1_new**2;
      end;

      keep 
          FParam:
          fmean 
          fvariance 
          N_F
          mass_at_0
          ;
   run;

   /*****************************************************************/
   /*           Grab the VaR values from simulation                 */
   /*****************************************************************/
   %let var_value = none;

   data _ex_var_sim; 
      set &varstat;
      where CCUID=&CCUID;

      var_value = P_1;
      pctl = pctl1;
      call symputx("var_value",put(var_value,BEST20.));
      output;

      var_value = P_2;
      pctl = pctl2;
      output;

      var_value = P_3;
      pctl = pctl3;
      output;

      keep 
          var_value 
          pctl
          ;
   run;

   /*--- Must have VaR Values -------------*/
   %if &var_value eq none %then %do;
      %let ORV_REASON = exl_no_var_values;
       %goto DONE;
   %end;

   /*****************************************************************/
   /*                    MERGE PARAMETER DATASETS                   */
   /*****************************************************************/ 
   data _ex_parms;
      merge 
            _ex_sev_parms 
            _ex_freq_parms
            ;

      do i = 1 to 3;
         output;
      end;

      drop i;
   run;

   /*****************************************************************/
   /*      GET THE VaR VALUE IMPLIED BY THE ORIGINAL PARAMTERS      */
   /*****************************************************************/
   data _temp_;
      merge _ex_var_sim _ex_parms;
      threshold = 0;

      pctl = (pctl/100);

      /* If the frequency mean is less than 1, then we have
         to worry about ecdf computed below becoming negative. */
      if fmean < 1 then do;
         pctl_adj  = (pctl - mass_at_0)/(1 - mass_at_0 );
      end;
      else do;
         pctl_adj = pctl;
      end;

      ecdf = 1 - ( (1 - pctl) / fmean );

   run;

   data _ex_parms;
      set _temp_;
      drop
         var_value
         pctl_adj
         ecdf
         ;
   run;

   %orv_inverse_severity_CDF(out = _ex_VaR_apprx);

   data _ex_VaR_apprx;
      set _ex_VaR_apprx;
      set _ex_parms;

      if mass_at_0 >= pctl then do;
         VaR_apprx = 0;
      end;
      else do;
         VaR_apprx = quant + smean*(fmean - 1 + fvariance/fmean );
      end;

      keep VaR_apprx;
   run;   

   /****************************************************************/
   /*                    COMPUTE THE EXTREME LOSS                  */
   /****************************************************************/
   %let extreme_loss = not_computed;

   data _temp_;
      ecdf = &ORV_EXTRA_LOSS_QUANTILE.;
      threshold = 0;
   run;

   %orv_inverse_severity_CDF(out = _ex_extreme_loss);

   data _NULL_;
      set _ex_extreme_loss;
      call symputx("extreme_loss", put(quant,BEST20.));
   run;

   %if &extreme_loss eq not_computed %then %do;
      %let ORV_ABORT   = &ORV_TRUE;
      %let ORV_REASON  = exl_unable_to_copute_extreme_loss;
      %let found_error = &ORV_TRUE.;
      %goto DONE;
   %end;

   /****************************************************************/
   /*                 UPDATE THE SEVERITY PARAMETERS               */
   /****************************************************************/
   /* Update the severity moments and use them to find a scale_effect that best captures the
      effect on the severity */
   %let smean_mod    = not_computed;

   data _ex_sev_update;

      set _ex_parms;
      if _N_ = 1;

      array scales{*} scales1 - scales&nsev_mix. (&sev_scales.);
      array scales_mod{*} scales_mod1 - scales_mod&nsev_mix.;
      array weights{*} weights1 - weights&nsev_mix. (&SPARAM1_WEIGHTS.);
      array moms{4} mom1-mom4;
      array m{4} m1-m4;
      array m_mod{4} m_mod1-m_mod4;
      array r{4} r1-r4;

      do i=1 to 4;
         moms[i] = 0;
         r[i] = 0;
      end;

      param_1_true = param_1;

      /* For mixtures, need the moments of the mixture */
      do i = 1 to &nsev_mix.;
         if upcase("&SCALE_TRANSFORM.") eq "ID" then do;
            param_1 = scales[i];
         end;
         else do;
            param_1 = log(scales[i]);
         end;

         %moments&sdist.;

         do j = 1 to 4;   
            if not missing(m[j]) then do;
               moms[j] = moms[j] + weights[i]*m[j];
            end;
         end;
      end;

      /* Compute the scale factor for adjusting the severity scale */
      if moms[1] = 0 then do;
         scale_factor = 1;
      end;
      else do;
         numerator = 0;
         denominator = 0;

         do i = 1 to 4;
            if moms[i] ne 0 then do;
               m_mod[i] = (N_S*moms[i] + &extreme_loss.**i)/(N_S + 1);
               r[i] = moms[i]/m_mod[i];
               numerator + r[i]**(1/i);
               denominator + r[i]**(2/i);
            end;
         end;

         scale_factor = numerator/denominator;
      end;

      /* Update the severity scales */
      do i = 1 to &nsev_mix.;
         scales_mod[i] = scale_factor * scales[i];
         call symputX(compress("sev_scales_mod"||put(i,10.)), scales_mod[i]);
     end;

     /* Update the largest and smallest scale parameter */
     call symputX("MAX_SCALE_mod", scale_factor * &MAX_SCALE.);
     call symputX("MIN_SCALE_mod", scale_factor * &MIN_SCALE. );

      /* Get the new severity mean.  */
      call symputX("smean_mod",m_mod[1]);
   run;
 
   %if &smean_mod eq not_computed %then %do;
      %let ORV_ABORT   = &ORV_TRUE;
      %let ORV_REASON  = exl_unable_to_copute_modified_sev;
      %let found_error = &ORV_TRUE.;
      %goto DONE;
   %end;

   /****************************************************************/
   /*                UPDATE THE FREQUENCY MOMENTS                  */
   /****************************************************************/
   data _ex_parms;
      set _ex_parms;

      /* Adjusted frequency moments */
      fmean_mod = fmean + 1 / N_F;
      fvariance_mod = fvariance + 1 / N_F - (1 / N_F)**2;

      /* Adjust the mass at 0 */
      number_of_0s  = round(mass_at_0 * N_F);
      mass_at_0_mod = mass_at_0 + (number_of_0s / N_F / N_F);

      /* If the frequency mean is less than 1, then we have
         to worry about ecdf computed below becoming negative. */
      if fmean_mod < 1 then do;
         /* Get original non-central moments to then compute the
            new mean and variance */
         fm1 = fmean_mod;
         fm2 = fvariance_mod + m1**2;

         fm1_new = fm1/(1 - mass_at_0_mod);
         fm2_new = fm2/(1 - mass_at_0_mod);

         fmean_mod     = fm1_new;
         fvariance_mod = fm2_new - fm1_new**2;
      end;
   run;

   /**************************************************************/
   /*   GET EFFECT OF THE EXTRA LOSS BY USING VAR APPROXIMATION  */
   /**************************************************************/
   /* Inverse CDF used to approximate VaR with extra loss*/
   data _temp_;
      set _ex_parms;

      /* If the frequency mean is less than 1, then we have
         to worry about ecdf computed below becoming negative. */
      if fmean_mod < 1 then do;
         pctl_adj  = (pctl - mass_at_0_mod)/(1 - mass_at_0_mod );
      end;
      else do;
         pctl_adj  = pctl;
      end;

      ecdf = 1 - ( (1 - pctl_adj) / fmean_mod );

   run;

   %let temp_MAX_SCALE = &MAX_SCALE.;
   %let MAX_SCALE = &MAX_SCALE_mod.;

   %let temp_MIN_SCALE = &MIN_SCALE.;
   %let MIN_SCALE = &MIN_SCALE_mod.;

   %let temp_scales = &sev_scales.;
   %let sev_scales = &sev_scales_mod1;
   %do i = 2 %to &nsev_mix.;
      %let sev_scales = &sev_scales. &&sev_scales_mod&i.;
   %end;
   %orv_inverse_severity_CDF(out = _ex_VaR_apprx_mod);

   %let MAX_SCALE     = &temp_MAX_SCALE.;
   %let MIN_SCALE     = &temp_MIN_SCALE.;
   %let sev_scales = &temp_scales.;

   data _ex_VaR_apprx_mod;
      set _ex_VaR_apprx_mod;
      set _ex_parms;

      if mass_at_0_mod >= pctl then do;
         VaR_apprx_mod = 0;
      end;
      else do;
         VaR_apprx_mod = quant + &smean_mod. * (fmean_mod - 1 + fvariance_mod / fmean_mod );
      end;

      keep VaR_apprx_mod; 
   run;

   /* Get change in VaR due to extra loss and apply to the original VaR */
   data _ex_results(keep = 
                           P_1_delta
                           P_2_delta
                           P_3_delta 
                           ); 
      set _ex_VaR_apprx;
      set _ex_VaR_apprx_mod;

      retain 
         P_1_delta 0 
         P_2_delta 0 
         P_3_delta 0
         ;
      Delta_VaR = VaR_apprx_mod - VaR_apprx;

      /* Need Delta_VaR >= 0 */
      if Delta_VaR < 0 then do;
         call symputX("ORV_ABORT", &ORV_TRUE);
         call symputX("ORV_REASON", "exl_effect_of_extreme_loss_negative");
         stop;
      end;

      if _N_ = 1 then do;
         P_1_delta = Delta_VaR;
      end;
      else if _N_ = 2 then do;
         P_2_delta = Delta_VaR;
      end;
      else if _N_ = 3 then do;
         P_3_delta = Delta_VaR;
         output;
      end;
   run;

   %if &ORV_ABORT. eq &ORV_TRUE. %then %do;
      %let found_error = &ORV_TRUE.;
      %goto DONE;
   %end;

   /*********************************************************************/
   /*                              FINISHED                             */
   /*********************************************************************/

   /* ---- Check for unknown errors ---- */
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %goto DONE;
   %end;   

   %let ORV_REASON  = success_note;
   %let found_error = &ORV_FALSE; 

   %DONE:

   %if &ORV_ABORT. %then %do;
      %if &found_error %then %do;
         %put &orvar_error.: %sysfunc(sasmsg(&orvads, exl_not_computed_for_this_cell, noquote, &CCUID));
         %put &orvar_error.: %sysfunc(sasmsg(&orvads, &ORV_REASON, noquote));
      %end;
   %end; 
   %else %do;
      data &varstat.;
         merge &varstat. _ex_results;
         P_1_extra_loss = P_1 + P_1_delta;
         P_2_extra_loss = P_2 + P_2_delta;
         P_3_extra_loss = P_3 + P_3_delta;
      run;

      data &varstat_ins.;
         merge &varstat_ins. _ex_results;
         P_1_extra_loss = P_1 + P_1_delta;
         P_2_extra_loss = P_2 + P_2_delta;
         P_3_extra_loss = P_3 + P_3_delta;
      run;
   %end;

   %orvu_delete_data(data = _ex_:)

%mend orva_extra_loss;

/* -------------------------------------------------------------------- 
   -------------------------------------------------------------------- */

/* Compute the inverse CDF of a mixture of severity models that only
   differ by their scale */
%macro orv_inverse_severity_CDF(
                                 out     = , 
                                 epsilon = 0.0001
                                 );
   %if %upcase(&ORV_DEBUG.) EQ TRUE %then %do;
      %put @orv_inverse_severity_CDF(;
      %put .                          out     = &out.,; 
      %put .                          epsilon = &epsilon.;
      %put .                          );
   %end;

   /* Get upper bounds on the inverse cdf */
   %if %upcase("&SCALE_TRANSFORM.") eq "ID" %then %do;
      %let param_1 = %sysevalf(&MAX_SCALE.);
   %end;
   %else %do;
      %let param_1 = %sysevalf(%sysfunc(log(&MAX_SCALE.)));
   %end;

   %Quant&SDIST.();

   /* Get lower bounds on the inverse cdf */
   %if %upcase("&SCALE_TRANSFORM.") eq "ID" %then %do;
      %let param_1 = %sysevalf(&MIN_SCALE.);
   %end;
   %else %do;
      %let param_1 = %sysevalf(%sysfunc(log(&MIN_SCALE.)));
   %end;

   data _ex_UB;
      set _temp_;
      UB = quant;
      keep UB;
   run;

   %Quant&SDIST.();

   data _temp_;
      set _temp_;
      set _ex_UB;
      LB = quant;
      keep LB UB cdf;
   run;

   proc fcmp 
            data = _temp_ 
            out  = &out.
            ;
      eps = abs(&epsilon.); 

      array weights{*} &SPARAM1_WEIGHTS.;
      array scales{*} &SEV_SCALES.;

      guess = (UB + LB)/2;
      /* If we only have one scale, then delta starts at 0 and we are done */
      delta = abs(UB - LB);

      /* Find inverse CDF using bisection */
      do while(delta > eps);
         val = 0;

         do i = 1 to dim(scales);
            if upcase("&SCALE_TRANSFORM.") eq "ID" then do;
               param_1 = scales[i];
            end;
            else do;
               param_1 = log(scales[i]);
            end;

            %Vcdf&SDist.(guess, temp_cdf);
            val = val + weights[i] * temp_cdf;
         end;

         if val > cdf then do;
            UB = guess;
         end;
         else do;
            LB = guess;
         end;

         newGuess =(UB + LB) / 2;
         delta = abs(guess - newGuess) / guess;
         guess = newGuess;
      end;

      quant = guess;

      keep quant;
   run;
%mend orv_inverse_severity_CDF;
