/*****************************************************************************************
 * Copyright (C), 2012
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orva_compute_net_exposure                                      
 *       
 * PURPOSE: Given a count dataset, for each count get pobs and fscale.  Then compute net
 *          exposure
 *         
 * CREATES: adds pobs, fscale, and net_exposure tothe counts data.
 *          Also stores avg count weighted pobs as avgCountWtdPobs in the
 *          dataset specified by the out_pobs argument
 *                  
 * SUPPORT: stmajo
 *
 * PARAMETERS:
 *     in_counts    - datset with the counts that net_exposure is computed for
 *
 *     in_losses    - datset with the individual losses
 *
 *     cell_where   - the usual subsetting where clause
 *
 *     dataType     - I for internal, etc.
 *
 *     ccv_property - ccv property model is stratified on.  Empty for cell level models
 *
 *     stratified   - true or false : is the frequency model stratified
 *
 *     sev_model    - dataset with sparm info for the associated severity model - top get pobs.
 *
 *     scaleData    - dataset with any scaling info (e.g. VARPRJ.scalei)
 *
 *     out_counts   - output dataset with original counts plus pobs, fscale, net_exposure
 *
 *     out_pobs     - dataset with average count weighted pobs (stored as avgCountWtdPobs)
 *
 * NOTES: 
 *
 *****************************************************************************************/
%macro orva_freq_compute_net_exposure( 
                                      in_counts    = ,
                                      in_losses    = ,
                                      cell_where   = ,
                                      dataType     = ,
                                      ccv_property = ,
                                      stratified   =  ,
                                      sev_model    = ,
                                      scaleData    = ,
                                      out_counts   = ,
                                      out_pobs     = 
                                      );
   %if %upcase(&ORV_DEBUG.) EQ TRUE %then %do; 
      %put #orva_freq_compute_net_exposure(; 
      %put .                                in_counts    = &in_counts.,;
      %put .                                in_losses    = &in_losses.,;
      %put .                                cell_where   = &cell_where.,;
      %put .                                dataType     = &dataType.,;
      %put .                                ccv_property = &ccv_property.,;
      %put .                                stratified   = &stratified.,;
      %put .                                sev_model    = &sev_model.,;
      %put .                                scaleData    = &scaleData.,;
      %put .                                out_counts   = &out_counts.,;
      %put .                                out_pobs     = &out_pobs.;
      %put .                                );
   %end;
 
   %local stratified_model;

   /* First get the losses for this cell, will need them when computing pobs (need to 
      evaluate the severity cdf at the threshold). */
   data _cne_cell_losses (compress = binary);
      set &in_losses.;
      where &cell_where.;
   run;

   /* Get POBS  */

   /* What is POBSERV_METHOD ? */
   %let pobsFlag = 0;

   data _NULL_;
      set &sev_model.;

      if strip(upcase(POBSERV_METHOD)) = 'DO_NOT_USE' then do;
         call symputX("pobsFlag", 0);
      end;
      else do;
         call symputX("pobsFlag", 1);
      end;
   run;

   %let avgCountWtdPobs = .;
   %orva_freq_getpobs( 
                      countdata     = &in_counts., 
                      sevinfo       = &sev_model.,
                      ccv_property  = &ccv_property.,
                      stratified    = &stratified.,
                      pobsFlag      = &pobsFlag.,
                      outcountdata  = &out_counts., 
                      out_wtd_pobs  = avgCountWtdPobs
                      )
   %if &ORV_ABORT. %then %do;
      %return;
   %end;

   /* Add in FSCALE info here (extracted from get_offset_and... macro in ffit) */
   %orv_getFSCALE( 
                  indata     = &out_counts.,
                  stratified = &stratified.,
                  datatype   = &datatype.,
                  scaleData  = &scaleData.,
                  outdata    = &out_counts.
                  )

   /* Now get the net exposure.  This is exposure*pobs/fscale.  The impied count for each
      substratum is count/net_exposure.  When summing over substratum, will need to use the 
      weighted sum of the net_exposures. */
   /* If Model is not stratified, aggregate losses to the cell level.  Need to get weighted
      sums of exposure, POBS, and FSCALE.  */
   data &out_counts.;
      set &out_counts.;
      net_exposure = exposure * pobs / FSCALE;

      if missing(net_exposure) then do;
         call symputX("ORV_ABORT", &ORV_TRUE.);
         call symputX("ORV_REASON", "%sysfunc(sasmsg(&orvads, freqf_cant_compute_net_expsr, noquote))");
         stop;
      end;
   run;

   data &out_pobs.;
      avgCountWtdPobs = &avgCountWtdPobs.;
      pobsFlag        = &pobsFlag.;
   run;

   %orvu_delete_data(data = _cne:)

%mend orva_freq_compute_net_exposure;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_getFSCALE( 
                     indata     = ,
                     stratified = ,
                     datatype   = ,
                     scaleData  = ,
                     outdata    = 
                     );
   %local nobs_scale;
   %let nobs_scale = 0;

   %if &scaleData. NE %str() %then %do;
      %orvu_getObsCount(&scaleData., nobs_scale);
   %end; 

   %if &nobs_scale. NE 0 %then %do;
      /* INTERNAL  and EXTERNAL: for EXTERNAL, FREQ_MODEL_IS_STRATIFIED is false */
      %if %upcase(&dataType.) eq I or %upcase(&dataType.) eq E %then %do;
         /*- merge count data with Scaling data by INTERNAL_BL_UID and YEAR -*/
         proc sql noprint; 
            CREATE TABLE 
                              &outdata.
               AS SELECT
                              * 
               FROM
                              &indata. AS L 
               LEFT JOIN
                              &scaleData.( keep = 
                                                  BLUID_SC 
                                                  YEAR 
                                                  FSCALE
                                                  ) AS S
               ON 

                              year(L.interval_start) = S.YEAR
                              AND
                              L.INTERNAL_BL_UID = S.BLUID_SC
               ;   
         quit;       
      %end; 

      data &outdata.;
         set &outdata.(drop = BLUID_SC YEAR);
         if missing(FSCALE) then do;
            FSCALE = 1;
         end;
      run;

   %end;
   %else %do;
      data &outdata.;
         set &indata.;
         FSCALE = 1;
      run;
   %end;

%mend orv_getFSCALE;
