/*****************************************************************************************
 *
 * Copyright (C), 2011
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    ORVA_freq_fit                                      
 *       
 * PURPOSE: Wrapper for orva_freq_fit_dist and other frequency fitting macros.
 *             1) gets all necessary fitting options out of the fitoptsData or initData
 *             2) Gets the counts for the current cell
 *             3) Using threshold options, computes pobs for each count
 *             4) Computes net exposure for each count (using pobs, fscale, exposure)
 *             5) Aggregates counts if necessary (total cells, cell level models)
 *             6) Performs KRI selection if necessary
 *             7) Passes KRIs and count data to freq_fit_dist to do actual fitting
 *                for each distribution (so 1 freq_fit_dist call for each requested
 *                distribution)
 *             8) Appends all fit outputs with fitting options to the fparm dataset
 *
 * CREATES: fparmX     - fitted parameters and fitting options
 *          fparm_covX - parameter covariance matrix
 *          fbluidestX - business line fixed effects
 *          fparm_mdlX - frequency model (by IBL if stratified model)
 *
 * SUPPORT: stmajo
 *
 * PARAMETERS:
 *     countdata        - count data from orva_freq_counts
 *     fitoptsData      - fitting options (NA if 'initData' is passed)
 *     initData         - fparm dataset to use as starting point
 *                        for fitting during incremental load
 *     iteration        - iteration to refit (only used with initData)
 *     instrataprops    - dataset containing stratprops used in orva_freq_make_counts (I think)
 *     useCellOverrides - use cell overrides (vs project options)
 *     cTool            - true/false: are we running Consortium Tool?
 *     pobservInfo      - sparm dataset containing severity fit for probability of observability
 *     sevIteration     - iteration id for sev fit this goes with
 *                        (NOT related to PObservInfo)
 *     sevIterSelected  - True/Fale: is sevIteration selected?
 *     datatype         - datatype I(nternal), (C)onsortium, (E)xternal
 *     ccUid            - ccUid for cell to fit
 *     ccData           - Cross Classification dataset
 *     allibls          - the losses/org/internal_businesslines dataset
 *     CCV_NAMES        - namse of current cross class variables
 *     parallel         - 1/0: parallel processing?
 *     outparm          - output fparm dataset to write results to
 *     outBluidEst      - output Dummy Estimates dataset (one dimension of dummies per newprop
 *     outCov           - output Covariance Matrix dataset
 *     outModels        - output Frequency Models dataset. One obs per stratum.  The sum of these
 *                        is the total mean for the cell.
 *     scaleData        - scaling dataset (currently only for internal data)
 *
 * NOTES:  Basic outline of how this macro works:
 *
 *         1) Get the fitting options and assign them to macros
 *
 *         2) Using fitting options, make any adjustments to the count data (what threhsold to use, etc.)
 *
 *         3) If there are KRIs, standardize them
 *
 *         4) Loop over distributions and create (one for each distribution):
 *            _ffit_countout_est_&ndist - these are combined to form fparm
 *            _ffit_covmat_&ndist       - these are combined to form fcov
 *            _ffit_fbluidest_&ndist    - these are combined to form fbluidest
 *            _ffit_freqmodels_&ndist   - these are combined to form fparm_mdl
 *
 *         5) Combine output from each distribution and appeand to tables in varprj.
 *          
 *
 *****************************************************************************************/
%macro orva_freq_fit( 
                     countdata        = ,
                     lossdata         = ,
                     fitoptsData      = ,  
                     initData         = ,
                     iteration        = , 
                     interval         = YEAR,
                     instrataprops    = ,
                     useCellOverrides = true,  
                     cTool            = true,
                     pobservInfo      = ,
                     sevIteration     = ,
                     sevIterSelected  = ,
                     datatype         = , 
                     ccUid            = , 
                     ccData           = ,
                     allibls          = VO1.INTERNAL_BUSINESSLINE,
                     CCV_NAMES        = , 
                     parallel         = 0,
                     outParm          = ,
                     outBluidEst      = ,
                     outCov           = ,
                     outModels        = ,
                     scaleData        = 
                     );   
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_freq_fit(; 
      %put .               countdata        = &countdata.,;
      %put .               lossdata         = &lossdata.,;
      %put .               fitoptsData      = &fitoptsData.,;  
      %put .               initData         = &initData.,;
      %put .               iteration        = &iteration.,; 
      %put .               interval         = &interval.,;
      %put .               instrataprops    = &instrataprops.,;
      %put .               useCellOverrides = &useCellOverrides.,;  
      %put .               cTool            = &cTool.,;
      %put .               pobservInfo      = &pobservInfo.,;
      %put .               sevIteration     = &sevIteration.,;
      %put .               sevIterSelected  = &sevIterSelected.,;
      %put .               datatype         = &datatype.,; 
      %put .               ccUid            = &ccUid.,; 
      %put .               ccData           = &ccData.,;
      %put .               allibls          = &allibls.,;
      %put .               CCV_NAMES        = &CCV_NAMES.,; 
      %put .               parallel         = &parallel.,;
      %put .               outParm          = &outParm.,;
      %put .               outBluidEst      = &outBluidEst.,;
      %put .               outCov           = &outCov.,;
      %put .               outModels        = &outModels.,;
      %put .               scaleData        = &scaleData.;
      %put .               );   
   %end;
         
   /*------------------------------------------------------------*/
   /*            Initialize macros and output datasets           */
   /*------------------------------------------------------------*/
   /* Initialize the status variables */
   %let ORV_REASON          = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %let ORV_ABORT           = &ORV_FALSE; 
   %let freq_fit_successful = &ORV_FALSE;

   /* Create the subsetting macro 'cell_where'  and find number of CCVS */
   %let CELL_WHERE = ;
   %let NUMCCVS    = ;
   %orvu_cell_where(
                     ccuid = &ccUid, 
                     ccData = &ccData
                     ); 
   
   /* Clear out the output datasets */
   %orvu_delete_data(
                     data = 
                           _ffit_parm 
                           _ffit_cov 
                           _ffit_blest 
                           _ffit_Freqmodels
                           );

   /* Some initial error checking */
   %orv_checkFreqFitInputs();
   %if &ORV_ABORT %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   /* Initialize fit options here, some with default values: */
   /* Stratum Properties */
   %let Freq_StrataProps             = ;
   %let Freq_Model_Basis_Props       = INTERNAL_BL_UID;
   %let Freq_Model_Is_Stratified     = false;

   /* KRI Options */
   %let Freq_Use_Dummies             = false;
   %let Freq_Use_Kris                = false;
   %let Freq_Kris_Suggested          = ;
   %let Freq_Kris_Forced             = ;
   %let Freq_Selection_Method_Kris   = Stepwise;
   %let FREQ_NAMESOFKRIS             = ;
   %let FREQ_NAMESOFKRIS_comma       = ;

   /* Fit measure options */
   %let Freq_FavFitMeasure           = LOGL;
   %let Freq_FitMeasures             = LOGL . SBC . AIC .;

   /*Distribution options */
   %let fdists                       = ;

   /* Threshold options */
   %let Freq_Use_Data_Threshold      = true;
   %let Freq_Missing_Default         = 0;
   %let Freq_Cut_Off                 = 0;
   %let Freq_Keep_Larger_Thresholds  = false;
   %let Freq_UseStratumMean_Default  = false;
   %let Freq_UseStratumMean_UserSpec = false;
   %let Freq_Zero_Count_Default      = 0;

   /*Probability of observability options */
   %let Freq_Pobserv_Source          = ;

   /*------------------------------------------------------------*/
   /*               Get the fitting options                      */
   /*------------------------------------------------------------*/
   %orv_freqFitOptionsRead( );
   %if &ORV_ABORT %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   %if &Freq_Model_Is_Stratified. eq false %then %do;
      %let Freq_Use_Dummies = false;
   %end;

   /* Initialize the output datasets to what they would be in case of error (need fdists for this,
   which isn't available until call to orv_freqFitOptionsRead ).  Puts fitting options in _ffit_parm */
   %orv_initializeOutputDatasets();

   /*-------------------------------------------------------------------------*/
   /*      Get the count data for the fit with offset and log adjusted count. */ 
   /*      Add zero-count substrata if model is stratified.                   */
   /*-------------------------------------------------------------------------*/
   /* Variables that will get assigned values here to be used later
   in the fitting */
   %let avgCOUNTwtdpobs   = .;
   %let Freq_ObsCount     = 0;  /* <= Number of frequency observations */
   %let Freq_Strata_Count = 1;

   /*------------------------------------------------------------------------- */
   /*     Get the counts for this cell.                                        */
   /*------------------------------------------------------------------------- */
   %orv_getCellCounts( 
                       in_counts  = &countdata.,
                       cell_where = &cell_where.,
                       dataType   = &dataType.,
                       out_data   = _ffit_counts 
                       );
   %if &ORV_ABORT %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   /* Determine the number of strata for the frequency model */
   %if %lowcase(&Freq_Model_Is_Stratified.) eq true %then %do;
      proc sql noprint;
         SELECT
                  count(distinct &Freq_Model_Basis_Props.)
         INTO
                  :Freq_Strata_Count
         FROM
                  _ffit_counts
         ;
      quit;
      %if &Freq_Strata_Count. eq 1 %then %do;
         %let Freq_Model_Is_Stratified = false;
      %end;
   %end;

   /*------------------------------------------------------------------------- */
   /*     Add in the threshold options to the cell counts                      */
   /*------------------------------------------------------------------------- */
   %orv_addThresholdOptions( 
                             in_counts  = _ffit_counts, 
                             out_counts = _ffit_counts 
                             ); 
   %if &ORV_ABORT %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   /*------------------------------------------------------------------------- */
   /*  Now apply the threshold options to get the threshold used to colc. pobs */
   /*------------------------------------------------------------------------- */
   %orva_freq_thresholds(  
                           countdata     = _ffit_counts, 
                           instrataprops = &freq_model_basis_props.,
                           stratified    = &Freq_Model_Is_Stratified.,
                           outcountdata  = _ffit_counts
                           );

   /*------------------------------------------------------------------------- */
   /*     For the cell, compute net expore (pobs, exposure, FScale)            */
   /*------------------------------------------------------------------------- */
   %orva_freq_compute_net_exposure( 
                                    in_counts    = _ffit_counts,
                                    in_losses    = &lossdata.,
                                    cell_where   = &CELL_WHERE.,
                                    dataType     = &dataType.,
                                    ccv_property = &freq_model_basis_props.,
                                    stratified   = &Freq_Model_Is_Stratified.,
                                    sev_model    = &pobservInfo.,
                                    scaleData    = &scaleData.,
                                    out_counts   = _ffit_counts,
                                    out_pobs     = avgCOUNTwtdpobs_data 
                                    );
   %if &ORV_ABORT. %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   data _null_;
      set avgCOUNTwtdpobs_data;
      call symputX("avgCOUNTwtdpobs",avgCOUNTwtdpobs);
      call symputX("pobsFlag",pobsFlag);
   run;

   /* check Mean Probability of Observability */
   %if  %sysevalf(&avgCountWtdPobs. < 0.1, BOOLEAN)  AND (&pobsFlag. EQ 0) %then %do;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, fgp_mean_cw_pobs_less_error, noquote, &avgCOUNTwtdpobs.));
      data _ffit_parm;
         set _ffit_parm;
         reason = "&ORV_REASON.";
      run;
      %let ORV_REASON = &ORV_REASON. %sysfunc(sasmsg(&orvads, fgp_specify_avgcount_pobs_error,noquote, &avgCOUNTwtdpobs.));
      %goto DONE_FREQ_FIT;
   %end; 

  /* Throw out any KRIs that have missing values from the forced / suggested list */
   %if (%lowcase(&freq_use_KRIs.) eq true) %then %do;
      %let Freq_numSuggestedKRIs = ;
      %let Freq_KRIs_suggested_mod = ;
      %orv_fremove_missing_variable(
                                    dataset   = _ffit_counts,
                                    variables = &Freq_KRIs_suggested.,
                                    out_list  = Freq_KRIs_suggested_mod,
                                    out_count = Freq_numSuggestedKRIs
                                    );
      %if %lowcase(&Freq_KRIs_suggested_mod.) ne %lowcase(&Freq_KRIs_suggested.) %then %do;
         %put &orvar_warning. %sysfunc(sasmsg(&orvads., freqf_missing_kri_values, noquote));
         %let Freq_KRIs_suggested   = &Freq_KRIs_suggested_mod.;
      %end;
      
      %orv_fremove_novar_variable(
                                    dataset   = _ffit_counts,
                                    variables = &Freq_KRIs_suggested.,
                                    out_list  = Freq_KRIs_suggested_mod,
                                    out_count = Freq_numSuggestedKRIs
                                    );

      %if %lowcase(&Freq_KRIs_suggested_mod.) ne %lowcase(&Freq_KRIs_suggested.) %then %do;
         %put &orvar_warning. %sysfunc(sasmsg(&orvads., freqf_novar_kri_values, noquote));
         %let Freq_KRIs_suggested = &Freq_KRIs_suggested_mod.;
      %end;

      %let Freq_numForcedKRIs = ;
      %let Freq_Kris_Forced_mod = ;
      %orv_fremove_missing_variable(
                                    dataset   = _ffit_counts,
                                    variables = &Freq_Kris_Forced.,
                                    out_list  = Freq_Kris_Forced_mod,
                                    out_count = Freq_numForcedKRIs
                                    );

      %if %lowcase(&Freq_KRIs_forced_mod.) ne %lowcase(&Freq_KRIs_forced.) %then %do;
         %put &orvar_warning. %sysfunc(sasmsg(&orvads., freqf_missing_kri_values, noquote));
         %let Freq_KRIs_forced = &Freq_KRIs_forced_mod.;
      %end;

      %orv_fremove_novar_variable(
                                    dataset   = _ffit_counts,
                                    variables = &Freq_Kris_Forced.,
                                    out_list  = Freq_Kris_Forced_mod,
                                    out_count = Freq_numForcedKRIs
                                    );

      %if %lowcase(&Freq_KRIs_forced_mod.) ne %lowcase(&Freq_KRIs_forced.) %then %do;
         %put &orvar_warning. %sysfunc(sasmsg(&orvads., freqf_novar_kri_values, noquote));
         %let Freq_KRIs_forced = &Freq_KRIs_forced_mod.;
      %end;

      %if &Freq_numSuggestedKRIs. eq 0 and &Freq_numForcedKRIs. eq 0 %then %do;
         %let freq_use_KRIs = false;
      %end;
         /* Get lists (space and comma delimeted) of the selected KRIs (if KRIs used) */
      %if &Freq_Use_Kris = true %then %do;
         %orv_getKriNames(
                           space_name = FREQ_NAMESOFKRIS, 
                           comma_name = FREQ_NAMESOFKRIS_comma
                           );
      %end;
   %end;

   /*------------------------------------------------------------------------- */
   /*     Perform any aggregation of counts.  For total cells, counts          */
   /*     will be sum of counts in component CCUIDs.  If model is cell         */
   /*     level, then counts will also be summed over the slicing dimension    */
   /*     In either case, net eposure will be a weighted sum of net exposures  */
   /*------------------------------------------------------------------------- */
   %orva_freq_aggregate_counts( 
                                in_counts                = _ffit_counts, 
                                freq_model_is_stratified = &freq_model_is_stratified.,
                                model_ccv_property       = &freq_model_basis_props.,
                                keep_KRIs                = &Freq_Use_Kris., 
                                KRI_names                = %quote(&FREQ_NAMESOFKRIS_comma.), 
                                out_counts               = _ffit_counts
                                );

   %if &ORV_ABORT. %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   /* Drop observations with missing counts */
   data _ffit_counts;
      set _ffit_counts;

      if missing(the_count) then do;
         delete;
      end;
   run;

   /* Store the total number of frequency observations */
   proc sql noprint;
      SELECT 
               count(INTERVAL_START)
      INTO
               :Freq_ObsCount
      FROM
               _ffit_counts
      ;
   quit;

   /* --------------------------------------------------------------------------- */
   /*           From net exposure, compute offset and log(adjusted count)         */
   /* --------------------------------------------------------------------------- */
   %orv_getOffsetAndLogAdjCount( indata    = _ffit_counts,
                                 datatype  = &datatype.,
                                 scaleData = &scaleData.,
                                 outdata   = _ffit_counts
                                 );

   %if &ORV_ABORT. %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   /*-----------------------------------------------------------------------*/
   /*      KRI selection- standardize KRIs and then variable selection      */
   /*-----------------------------------------------------------------------*/
   %let selectedKRIs          =  ;
   %let selectedKRIs_comma    =  ;
   %let selectedDummies       =  ;
   %let selectedDummies_comma =  ;
   %let numSelectedKRIs       = 0;
   %let numSelectedDummies    = 0;
   %if (%lowcase(&freq_use_KRIs.) eq true) or (%lowcase(&freq_model_is_stratified.) eq true) %then %do;

      %orv_kriSelection;

      %if &ORV_ABORT %then %do;
         %goto DONE_FREQ_FIT;
      %end;
   %end;
   %if %lowcase(&freq_use_KRIs) eq false %then %do;
      /* Create _ffit_counts_standard - just _ffit_counts when there are no KRIs */
      data _ffit_counts_standard;
          set _ffit_counts;
      run;
   %end;

   /*------------------------------------------------------------*/
   /*   IF Parallel processing, then distribute distributions    */
   /*                     over the remote hosts                  */
   /*------------------------------------------------------------*/
   %let ORVP_STATUS = 1;
   %let ndist       = 1;

   %if &parallel eq 1 %then %do;
      %orvp_distribute_freq_fit_dist(dists = &fdists);

      %if (&ORVP_STATUS ne 0) %then %do;
         %put &orvar_warning %sysfunc(sasmsg(&orvads, freqf_no_proces_paralfits_seq_warn, noquote));
      %end;
   %end;
   
   /*------------------------------------------------------------*/
   /*    If not in parallel, then Loop over the distributions    */
   /*                       and perform a fit                    */
   /*------------------------------------------------------------*/
   %if (&ORVP_STATUS NE 0) %then %do;
      %let dist = %scan(&fdists, &ndist, ' ');

      %do %while (&dist NE %str() );
         %orva_freq_fit_dist(
                              fdist             = &dist.,
                              ndist             = &ndist., 
                              CountData         = _ffit_counts,
                              countDataStandard = _ffit_counts_standard, 
                              ccuid             = &ccuid.,
                              avg_COUNTwtdpobs  = &avgCOUNTwtdpobs.,
                              strata_count      = &Freq_Strata_Count.,
                              allibls           = &allibls.
                              );

         %let ndist = %eval(&ndist. + 1);
         %let dist  = %scan(&fdists., &ndist., ' ');
      %end;
   %end;
   %let ndist = %eval(&ndist. - 1);

   /*------------------------------------------------------------*/
   /*    Combine the output datasets from each distribution      */
   /*    creates: _ffit_parm, _ffit_cov, _ffit_blest,            */
   /*             _ffit_freqModels                               */
   /*------------------------------------------------------------*/
   %orv_combineFreqFitOutput();

   /*------------------------------------------------------------*/
   /*            Format and save the fit options                 */
   /*------------------------------------------------------------*/
   data _ffit_parm;
      %orv_formatFreqFitOptions;
      set _ffit_parm;
      %orv_saveFreqFitOptions;
   run;

   /* Check for any errors from the fitting process */
   %if &ORV_ABORT. %then %do;
      %goto DONE_FREQ_FIT;
   %end;

   /*------------------------------------------------------------*/
   /*         If we get here, then no errors were found          */
   /*------------------------------------------------------------*/
   %let freq_fit_successful = &ORV_TRUE.;

   /*------------------------------------------------------------*/
   %DONE_FREQ_FIT: 
   /*------------------------------------------------------------*/

   /* Display any error messages */
   %if &freq_fit_successful  eq &ORV_FALSE. %then %do; 
      %put &orvar_error. &ORV_REASON.;

      %let sev_dist_name = ;
      data _NULL_;
         set &pobservInfo.;
         call symputX("sev_dist_name",distribution);
      run;
      %put &orvar_error. %sysfunc(sasmsg(&orvads, ffit_no_fits_performed, noquote, &sev_dist_name., &CCUID.));
   %end;  
 
   /* Append to the output to the parm dataset. */
   %orva_append_parm_dataset( 
                              ccUid           = &ccuid,
                              sevIteration    = &sevIteration,
                              sevIterSelected = &sevIterSelected,
                              measCol         = &Freq_FitMeasures,
                              selectedMeas    = &Freq_FavFitMeasure,
                              tmpParm         = _ffit_parm,
                              finalParm       = &outparm,
                              tmpCov          = _ffit_cov,
                              finalCov        = &outcov,
                              tmpBlEst        = _ffit_blest, 
                              finalBlEst      = &outBluidEst,
                              tmpFModels      = _ffit_freqModels, 
                              finalFModels    = &outModels
                              %if &iteration NE %str() %then %do;
                                 ,iteration   = %eval(&iteration - 1)
                              %end;
                              );

   /* Delete all of the temprorary datasets */
   %orvu_delete_data( data = _ffit_: )
%mend orva_freq_fit;

/*****************************************************************************/
/*                          END OF ORVA_FREQFIT                              */
/*****************************************************************************/
%macro orv_checkFreqFitInputs();
   %local nobs;
   /* Check countdata */
   %let nobs = 0;

   %if &countdata NE %str() %then %do;
      %orvu_getObsCount(&countdata, nobs);
   %end; 
   %if &nobs <= 0 OR &countdata eq %str() %then %do;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, ffit_no_count_data_error, noquote));
   %end;

   /* Check instrataprops */
   %let nobs = 0;

   %if &instrataprops NE %str() %then %do;
      %orvu_getObsCount(&instrataprops,nobs);
   %end; 

   %if &nobs LE 0 %then %do;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, ft_isp_ds_empty_warning, noquote));
   %end;  
%mend orv_checkFreqFitInputs;
/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_freqFitOptionsRead( );
   /* Get the stratum properties from the instrataprops dataset.  
   This is a list of all the possible stratification dimensions (CCVS)*/
   proc sql noprint;
      select distinct
                        ClassificationVariable
      into
                        :Freq_StrataProps separated by " "
      from
                        &instrataprops.
      ;
   quit;

   /* For incremental load, all options are in initData dataset */
   %if &initData NE %str() %then %do;
      %orv_readIncrementalOptions();
   %end; 
   /* Otherwise, get options from fitopts dataset */
   %else %if &fitOptsData NE %str() %then %do;
      %orv_readOptionsFromFitopts();
   %end;

   %let freq_model_basis_props = INTERNAL_BL_UID;

   /* Get lists (space and comma delimeted) of the selected KRIs (if KRIs used) */
   %if &Freq_Use_Kris = true %then %do;
      %orv_getKriNames(
                           space_name = FREQ_NAMESOFKRIS, 
                           comma_name = FREQ_NAMESOFKRIS_comma
                           );
   %end;
%mend orv_freqFitOptionsRead;
/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_readIncrementalOptions();
   data _NULL_; 
      set &initData; 
      where CCUID = &CCUID and ITERATION = &iteration;
      /*- get KRI and Indicator Info -*/
      %if &datatype EQ I %then %do;
         if ^missing(Freq_Use_Kris) then do;
            call symputX("Freq_Use_Kris", strip(Freq_Use_Kris));
         end;

         if ^missing(Freq_Kris_Suggested) then do;
            call symputX("Freq_Kris_Suggested", strip(Freq_Kris_Suggested));
         end;

         if ^missing(Freq_Kris_Forced)then do;
            call symputX("Freq_Kris_Forced", strip(Freq_Kris_Forced));
         end;

         if ^missing(Freq_Selection_Method_Kris) then do;
            call symputX("Freq_Selection_Method_Kris",strip(Freq_Selection_Method_Kris)); 
         end;

         if ^missing(Freq_Use_Dummies) then do;
            call symputX("Freq_Use_Dummies",strip(Freq_Use_Dummies)); 
         end;

         if ^missing(Freq_Model_Is_Stratified) then do;
            call symputX("Freq_Model_Is_Stratified",strip(Freq_Model_Is_Stratified)); 
         end;

         if ^missing(Freq_Model_Basis_Props) then do;
            call symputX("Freq_Model_Basis_Props", strip(Freq_Model_Basis_Props));
         end;
         else do;
            call symputX("Freq_Model_Basis_Props", "INTERNAL_BL_UID");
         end;
      %end;
      %else %do;
         call symputX("Freq_Model_Basis_Props", "");
      %end;  

      /* Pobserv options */
      if ^missing(Freq_Pobserv_Source) then do;
         call symputX("Freq_Pobserv_Source", Freq_Pobserv_Source);
      end;

      /*- get Threshold options -*/
      if ^missing(Freq_Use_Data_Threshold) then do;
         call symputX("Freq_Use_Data_Threshold", Freq_Use_Data_Threshold);
      end;

      if ^missing(Freq_Missing_Default) then do;
         call symputX("FreqOpt_Missing_Default", Freq_Missing_Default);
      end;

      if ^missing(Freq_Cut_Off) then do;
         call symputX("Freq_Cut_Off", Freq_Cut_Off);
      end;

      if ^missing(Freq_Keep_Larger_Thresholds) then do;
         call symputX("Freq_Keep_Larger_Thresholds", Freq_Keep_Larger_Thresholds);
      end;

      if ^missing(Freq_UseStratumMean_Default) then do;
         call symputX("&Freq_UseStratumMean_Default", Freq_UseStratumMean_Default);
      end;

      if ^missing(Freq_UseStratumMean_UserSpec) then do;
         call symputX("Freq_UseStratumMean_UserSpec", Freq_UseStratumMean_UserSpec);
      end;

      if ^missing(Freq_Zero_Count_Default) then do;
         call symputX("Freq_Zero_Count_Default", Freq_Zero_Count_Default);
      end;

      /*- Distribution options -*/      
      call symputX("fdists", distribution); 
   run;
%mend orv_readIncrementalOptions;
/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_readOptionsFromFitopts();
   %if &useCellOverrides EQ false %then %do;
      %let ccuidForOptions = .;
   %end;
   %else %do;
      %let ccuidForOptions  = &CCUID;
   %end;

   /*- get Distribution options -*/
   %orva_readfitopts( 
                     fitoptsdata = &fitoptsData,
                     type        = DIST,
                     fittype     = FREQ,
                     ccUid       = &ccuidForOptions,
                     datatype    = &datatype
                     );
   %if &fdists EQ %str() %then %do;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, ffit_no_dist_sel_for_fit_warning, noquote));
      %let ORV_ABORT  = 1;
      %return;
  %end;  

   /*- get KRI Control Options -*/
   %if &datatype EQ I %then %do;
      %orva_readfitopts( 
                        fitoptsdata = &fitoptsData,
                        type        = KRICONTROL,
                        fittype     = FREQ,
                        ccUid       = &ccuidForOptions,
                        datatype    = &datatype
                        );
   %end;

   /*- get Indicator Control Options -*/
   %if &datatype EQ I %then %do;
      %orva_readfitopts( 
                        fitoptsdata = &fitoptsData,
                        type        = INDICATOR,
                        fittype     = FREQ,
                        ccUid       = &ccuidForOptions,
                        datatype    = &datatype
                        );
   %end;

   /*- get POBSERV options -*/
   %orva_readfitopts( 
                     fitoptsData = &fitoptsData,
                     fittype     = FREQ,
                     type        = POBSERV,
                     datatype    = &datatype,
                     ccUid       = &ccuidForOptions
                     );

   /*- get Threshold options -*/
   %orva_readfitopts( 
                     fitoptsData = &fitoptsData, 
                     fittype     = FREQ, 
                     type        = THRESHOLD, 
                     datatype    = &datatype, 
                     ccuid       = &ccuidForOptions
                     );

   /*- get fit measure options -*/
   %orva_readfitopts(
                     fitoptsdata = &fitoptsData,
                     type        = FITMEASURE,
                     ccUid       = &ccuidForOptions,
                     fittype     = FREQ,
                     datatype    = &datatype
                     );
   /*- get KRIs -*/
      
   %if &Freq_Use_Kris EQ true %then %do;    
      %orva_readfitopts( 
                        fitoptsdata = &fitoptsData,
                        type        = KRI,
                        fittype     = FREQ,
                        ccUid       = &ccuidForOptions,
                        datatype    = &datatype
                        );
      /* If someone checks the  use kris box but doesn't actually select any...*/
      %if &Freq_Kris_Suggested EQ %str() AND &Freq_Kris_Forced EQ %str() %then %do;
         %let Freq_Use_Kris = false;
      %end;       
   %end; 
%mend orv_readOptionsFromFitopts;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_getKriNames(
                           space_name = , 
                           comma_name =
                           );
   /* What it does: produces a space-delimeted and comma-delimeted list of the selected
   (forced and suggested) KRIs.  These are stored in the macro variables speccified by
   space_name and comma_name */
   %let &space_name. = %str(); 
   %let &comma_name. = %str(); 

   %if %lowcase(&FREQ_USE_KRIs.) eq true %then %do;
      /* Add in names of suggested and forced KRIs */
      data _ffit_KRI_list;
         length KRI_name $ 100;

         /* Forced KRIs */
         %let forced_count   = 1;   
         %let current_forced = %scan(&FREQ_KRIs_FORCED.,&forced_count.," ");

         %do %while(&current_forced ne %str());
            KRI_name = "&current_forced";
            output;
            %let forced_count = %eval(&forced_count. + 1);
            %let current_forced = %scan(&FREQ_KRIs_FORCED.,&forced_count.," ");
         %end;

         %let forced_count = %eval(&forced_count. - 1);

         /* Suggested KRIs */
         %let SUGGESTED_count   = 1;   
         %let current_SUGGESTED = %scan(&FREQ_KRIs_SUGGESTED.,&SUGGESTED_count.," ");

         %do %while(&current_SUGGESTED ne %str());
            KRI_name = "&current_SUGGESTED";
            output;
            %let SUGGESTED_count   = %eval(&SUGGESTED_count. + 1);
            %let current_SUGGESTED = %scan(&FREQ_KRIs_SUGGESTED.,&SUGGESTED_count., " ");
         %end;

         %let SUGGESTED_count = %eval(&SUGGESTED_count. - 1);
      run;

      /* If no KRIs were selected, then set the use KRIs flag to false */
      %if &forced_count. + &SUGGESTED_count. = 0 %then %do;
         %let FREQ_USE_KRIs = false;
      %end;
      /* Otherwise, get a comma delimeted and space delimeted list of the KRIs */
      %else %do;
         PROC SQL noprint;
            SELECT
                     KRI_name,
                     KRI_name
            INTO
                     :&space_name. separated by " ",
                     :&comma_name. separated by ","
            FROM
                     _ffit_KRI_list
            ;
         quit;
      %end;
   %end;
%mend orv_getKriNames;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_initializeOutputDatasets();
   %local ndist dist;

   data _ffit_parm;
      format reason &ORVAR_REASON_FMT distribution $12.;
      %orv_formatFreqFitOptions;

      %let ndist = 1;
      %let dist  = %scan(&fdists, &ndist, ' ');

      %do %while (&dist NE %str() );
         distribution = "&dist";   
         reason = "&ORV_REASON"; 
         status = 1;
         LOGL = SBC = AIC = .;
         %orv_saveFreqFitOptions;
         output;
         %let ndist = %eval(&ndist + 1);
         %let dist  = %scan(&fdists, &ndist, ' ');
      %end;
   run;

   %orvu_nullify_data(data =  
                              _ffit_cov 
                              _ffit_blest
                              _ffit_FreqModels 
                              );
%mend orv_initializeOutputDatasets;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_getCellCounts( in_counts  = ,
                          cell_where = ,
                          dataType   = ,
                          out_data   =  );

   %local 
      deleted_rows 
      nobs_counts;

   %let deleted_rows = &ORV_FALSE.;

   /* Get the counts in this cell */
   data &out_data.(compress = binary);
      set &in_counts.;
      where &cell_where.;

      /* Set missing exposure values to 0 */
      exposure = max(exposure,0);

      /* Drop observations with missing data or no exposure */
      if exposure eq 0 or missing(the_count) then do;
         call symputX("deleted_rows", &orv_true.);
         delete;
      end;
   run;

   /* Make sure there are some observations here */
   %if &deleted_rows. eq &orv_true. %then %do;
      %put &orvar_warning. %sysfunc(sasmsg(&orvads., freqf_some_rows_were_dropped, noquote));
   %end;

   %let nobs_counts = 0;
   %orvu_getObsCount(&out_data., nobs_counts);

   %if &nobs_counts. = 0 %then %do;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads., freqf_all_rows_dropped, noquote));
      %let ORV_ABORT  = &orv_true.;
      %return;
   %end;
%mend orv_getCellCounts;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

/* NOTE: This macro assumes that the frequency fitting options have already been pulled
         and are stored in macro variables */
%macro orv_addThresholdOptions( 
                               in_counts  = , 
                               out_counts = 
                               ); 
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_addThresholdOptions(; 
      %put .                         in_counts  = &in_counts.,; 
      %put .                         out_counts = &out_counts.;
      %put .                         ); 
   %end;

   data &out_counts.;
      set &in_counts.;
      format 
            Freq_Missing_Default   
            Freq_Zero_Count_Default 
            Freq_Cut_Off            10.5
            ;
      length 
            Freq_Use_Data_Threshold 
            Freq_UseStratumMean_Default 
            Freq_Keep_Larger_Thresholds 
            $10
            ;

      Freq_Use_Data_Threshold     = "&Freq_Use_Data_Threshold";
      Freq_Missing_Default        = &Freq_Missing_Default;
      Freq_Zero_Count_Default     = &Freq_Zero_Count_Default;
      Freq_UseStratumMean_Default = "&Freq_UseStratumMean_Default";
      Freq_Cut_Off                = &Freq_Cut_Off;
      Freq_Keep_Larger_Thresholds = "&Freq_Keep_Larger_Thresholds"; 

      array prop &Freq_STRATAPROPS;

      do i = 1 to dim(prop);
         if missing(prop[i]) then do;
            prop[i] = -1;
         end;
      end;
   run;
%mend orv_addThresholdOptions;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_getOffsetAndLogAdjCount( 
                                   indata    = ,
                                   datatype  = ,
                                   scaleData = ,
                                   outdata   = 
                                   );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_getOffsetAndLogAdjCount(; 
      %put .                             indata    = indata.,;
      %put .                             datatype  = datatype.,;
      %put .                             scaleData = scaleData.,;
      %put .                             outdata   = outdata.;
      %put .                             );
   %end;

   /* Drop obs with zero net exposure; 
   create log_adjcount for use in Proc Reg variable selection.  */
   data &outdata(compress = binary);
      set &indata; 

      if net_exposure LE 0 then do;
         offset = 1;  
         log_adjcount = .;
         the_count = .;
      end;

      else do;
         offset = safe_log(net_exposure); 
         log_adjcount = safe_log(the_count) - offset;
      end;
   run;

   /* Check to see if anything went wrong in here */
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, ffit_err_computing_offset, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
   %end;
%mend orv_getOffsetAndLogAdjCount;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_kriSelection;
   /* Standardize the KRIs */
   %if %lowcase(&freq_use_KRIs.) eq true %then %do;
      %orv_standardizeKRIs( 
                           indata   = _ffit_counts,
                           outdata  = _ffit_counts_standard,
                           outstats = _ffit_standard_stats
                           );
   %end;
   %if &ORV_ABORT %then %do;
      %return;
   %end;
   
   /* Clear datatsets that will get used */
   %orvu_nullify_data( data = 
                              _ffit_fvs_outparms 
                              _ffit_selected_KRIs 
                              _ffit_fvs_status                            
                              _ffit_dummyMap 
                              _ffit_selected_Dummies 
                              );

   /* Perform variable selection */
   %if %lowcase(&freq_use_KRIs.) eq true %then %do;
      %orva_freq_variable_select ( 
                                  data                      = _ffit_counts_standard,
                                  selection                 = &Freq_Selection_Method_Kris.,
                                  outCompressedData         = _ffit_counts_standard,
                                  outParms                  = _ffit_fvs_outparms,
                                  outSelKRIs                = _ffit_selected_KRIs,
                                  outNotSelected            = _ffit_KRI_not_selected,
                                  outDroppedDummies         = _ffit_droppedDummies,
                                  outDummyMap               = _ffit_dummyMap,
                                  outSelDummies             = _ffit_selected_Dummies
                                  );
   %end;
   %else %do;
      %orva_freq_variable_select ( 
                                  data                      = _ffit_counts,
                                  selection                 = &Freq_Selection_Method_Kris.,
                                  outCompressedData         = _ffit_counts,
                                  outParms                  = _ffit_fvs_outparms,
                                  outSelKRIs                = _ffit_selected_KRIs,
                                  outNotSelected            = _ffit_KRI_not_selected,
                                  outDroppedDummies         = _ffit_droppedDummies,
                                  outDummyMap               = _ffit_dummyMap,
                                  outSelDummies             = _ffit_selected_Dummies
                                  );
   %end;

   %if &ORV_ABORT %then %do;
      %return;
   %end;

   /* Get KRI information */
   data _NULL_;
      set _ffit_fvs_outparms;
      call symputX("numSelectedKRIs", numSelectedKRIs);
   run;
   
   %if &numSelectedKRIs. > 0 %then %do;
      proc sql noprint;
         SELECT
                  _NAME_,
                  compress('"'||_NAME_||'"')
         INTO
                  :selectedKRIs       SEPARATED BY " ",
                  :selectedKRIs_comma SEPARATED BY ","
         FROM
                  _ffit_selected_KRIs
                  ;
      quit;
   %end;

   /* Get Dummy information (if appropriate) */
   data _NULL_;
      set _ffit_fvs_outparms;
      call symputX("numSelectedDummies", numSelectedDummies);
   run;
   %if &numSelectedDummies. > 0 %then %do;
      proc sql noprint;
         SELECT
                  upcase(_NAME_),
                  upcase(compress('"'||_NAME_||'"'))
         INTO
                  :selectedDummies       SEPARATED BY " ",
                  :selectedDummies_comma SEPARATED BY ","
         FROM
                  _ffit_selected_Dummies
         ;
      quit;
   %end;
%mend orv_kriSelection;

/*------------------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------------*/
%macro orv_standardizeKRIs( 
                           indata   = ,
                           outdata  = ,
                           outstats = 
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_standardizeKRIs(; 
      %put .                     indata   = indata.,;
      %put .                     outdata  = outdata.,;
      %put .                     outstats = outstats.;
      %put .                     );
   %end;

   /* Save original KRI values, w/ obs in same order as current _ffit_counts */
   data _ffit_only_KRIs ; 
      set &indata(keep = KRI_:) ;
   run;

   /* Standardize the KRIs into dataset _ffit_counts2 */    
   ods listing close;
   ods output Standard =  &outstats ; 
   proc standard 
               data = &indata 
               mean = 0 
               std  = 1 
               print 
               out  = &outdata;
               var &FREQ_NAMESOFKRIS 
               ;
   run;
   ods listing; 

   proc datasets nolist nodetails; /* Modify names of unstandardized KRIs by adding prefix "un_" */
      MODIFY _ffit_only_KRIs;  /* Contains original KRI values */ 
      RENAME 
            %let nname = 1;
            %let thename = %scan(&FREQ_NAMESOFKRIS, &nname, ' ');

            %do %while (&thename NE %str() );
               &thename = un_&thename  
               %let nname   = %eval(&nname + 1);
               %let thename = %scan(&FREQ_NAMESOFKRIS, &nname, ' ');
            %end;
            ;  
   quit;

   /* outdata will have the standarized KRIs (KRI_:) 
   and the unstandardized KRIs (un_KRI_:) */
   data &outdata(compress = binary);
      set &outdata; 
      set _ffit_only_KRIs;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %let ORV_ABORT  = &ORV_TRUE.;
      %let ORV_REASON = %sysfunc(sasmsg(&orvuds, freqf_unknown_standardizing_error, noquote));
   %end;
%mend orv_standardizeKRIs;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_combineFreqFitOutput();
   data _ffit_cov; 
      set
         %do i = 1 %to &ndist.;
            _ffit_covmat_&i. 
         %end;
         ;
   run;

   data _ffit_parm;
      set
         %do i = 1 %to &ndist.;
            _ffit_countout_est_&i. 
         %end;
         ;
   run;

   data _ffit_blest;
      set
         %do i = 1 %to &ndist.;
            _ffit_fbluidest_&i.
         %end;
         ;
   run;

   data _ffit_freqModels;
      set
         %do i = 1 %to &ndist.;
            _ffit_freqmodels_&i. 
         %end;
         ;
   run;

   %if &SYSERR. ne 0 and &SYSERR. ne 4 %then %do;
      %let ORV_ABORT = &ORV_TRUE.;
   %end;
%mend orv_combineFreqFitOutput;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_formatFreqFitOptions;
   length Freq_Use_KRIs              $ 8;
   length Freq_Use_Dummies           $ 8;
   length Freq_Model_Is_Stratified   $ 8;
   length Freq_Kris_Suggested        $ 10000;
   length Freq_Kris_Forced           $ 10000;
   length Freq_Pobserv_Source        $ 12;
   length Freq_Model_Basis_Props     $ 256;
   length Freq_Selection_Method_Kris $ 32;
%mend orv_formatFreqFitOptions;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_saveFreqFitOptions;
   Freq_Pobserv_Source          = "&Freq_Pobserv_Source";
   Freq_Use_Kris                = "&Freq_Use_Kris";
   Freq_Use_Dummies             = "&Freq_Use_Dummies";
   Freq_Kris_Suggested          = "&Freq_Kris_Suggested";
   Freq_Kris_Forced             = "&Freq_Kris_Forced";
   Freq_Selection_Method_Kris   = "&Freq_Selection_Method_Kris";
   Freq_Use_Data_Threshold      = "&Freq_Use_Data_Threshold";
   Freq_Keep_Larger_Thresholds  = "&Freq_Keep_Larger_Thresholds";
   Freq_UseStratumMean_Default  = "&Freq_UseStratumMean_Default";
   Freq_UseStratumMean_UserSpec = "&Freq_UseStratumMean_UserSpec";
   Freq_Fit_Interval            = "&interval.";
   Freq_Model_Is_Stratified     = "&Freq_Model_Is_Stratified";
   Freq_Missing_Default         = &Freq_Missing_Default;
   Freq_Cut_Off                 = &Freq_Cut_Off;
   Freq_Zero_Count_Default      = &Freq_Zero_Count_Default;
   Freq_Model_Basis_Props       = "&Freq_Model_Basis_Props";
%mend orv_saveFreqFitOptions;

%macro orv_fremove_missing_variable(
                                   dataset   = ,
                                   variables = ,
                                   out_list  = ,
                                   out_count = 
                                   );

   %if &variables. eq %str() %then %do;
      %let &out_list. = %str();
      %let &out_count. = 0;
   %end;
   %else %do;
      %local number_of_variables;
      %let number_of_variables = %sysfunc(countw(&variables., " "));
      data _rmv_temp_dataset_;
         array list[*] &variables.;    
         array flags[&number_of_variables.] Flag1-Flag&number_of_variables.;      
         retain flags;
         set &dataset. end = last;
         do i = 1 to dim(list);
            if list(i) = . then do;
               flags(i) = 1;
            end;
         end;
         if last then do;
            output;
         end;
      run;

      data _rmv_temp_dataset_;
         set _rmv_temp_dataset_;
         length _name_ $500;
         array list[*] &variables.; 
         array flags[&number_of_variables.] Flag1-Flag&number_of_variables.; 

         count = 0;
         do i = 1 to &number_of_variables.;
            if flags(i) ne 1 then do;
               _name_ = vname(list(i));
               count = count + 1;
               output;
            end;
         end;
         call symputX("&out_count.", count);
         
         keep _name_;
      run;

      proc sql noprint;
         SELECT
                  _name_
         INTO
                  :&out_list. separated by " "
         FROM
                  _rmv_temp_dataset_
         ;
      quit;

      %orvu_delete_data( data = _rmv_temp_dataset_ );
   %end;
%mend orv_fremove_missing_variable;

/* Remove variables that don't have any variance ----------------*/
%macro orv_fremove_novar_variable(
                                  dataset   = ,
                                  variables = ,
                                  out_list  = ,
                                  out_count = 
                                  );

   %if &variables. eq %str() %then %do;
      %let &out_list. = %str();
      %let &out_count. = 0;
   %end;
   %else %do;
      data _ffit_kris_temp;
         set &dataset.;
         keep &variables.;
      run;

      ods listing close;
      ods output Standard = _ffit_standard_stats2; 
      proc standard 
                  data = _ffit_kris_temp 
                  mean = 0 
                  std  = 1 
                  print 
                  out  = _ffit_temp;
                  var &variables. 
                  ;
      run;
      ods listing; 

      proc sql noprint;
         select
                  name
         into
                  :&out_list. separated by " "
         from
                  _ffit_standard_stats2
         where
                  stdDev ne 0
         ;
      quit;
 
      %local number_of_variables;
      %let &out_count. = %sysfunc(countw(&&&out_list, " "));
   %end;
%mend orv_fremove_novar_variable;
