/*************************************************************************************
 * Copyright (C), 2011
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orva_freq_fit_dist.sas                                             
 * PURPOSE: Submacro of %orva_ffit() called once for each  distribution.
 *                    
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *    fdist             - Current distribution name.
 *    ndist             - Current distribution number.
 *    CountData         - counts dataset (conpressed by orva_ffit).
 *    countDataStandard - Counts with standardized KRIs
 *    ccuid             - current cell key.
 *    avg_COUNTwtdpobs  - average count weighted pobs.
 *                        This is computed in freq_fit and put into the fit datatset for use 
 *                        later on.
 *    strata_count      - Number of strata, aka cell slices, for the frequency model.  Cell
 *                        level models have strata_count = 1, >1 for 'stratified' models.
 *                        This is computed in freq_fit and put into the fit datatset for use 
 *                        later on.
 *    
 *    allibls           - The losses/org/internal_businesslines dataset
 *
 * NOTES: Please see the accompanying doc for %orva_freq_fit().
 *        The output of this macro is the datasets:
 *            _ffit_countout_est_&ndist - parameter estimates for the compressed(counts) data
 *            _ffit_covmat_&ndist       - covariance matrix
 *            _ffit_fbluidest_&ndist    - The fixed effect for each business line
 *            _ffit_freqmodels_&ndist   - The frequency model - by business line if stratified
 *
 *       If there are parameter changes, please make sure %orvp_distribute_freq_fit_dist
 *       has the same update.
 **************************************************************************************/
%macro orva_freq_fit_dist( 
                           fdist             = ,
                           ndist             = , 
                           CountData         = ,
                           countDataStandard = ,
                           ccuid             = ,
                           avg_COUNTwtdpobs  = ,
                           strata_count      = ,
                           allibls           = 
                         );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_freq_fit_dist(; 
      %put .                    fdist             = &fdist.,;
      %put .                    ndist             = &ndist.,; 
      %put .                    CountData         = &CountData.,;
      %put .                    countDataStandard = &countDataStandard.,;
      %put .                    ccuid             = &ccuid.,;
      %put .                    avg_COUNTwtdpobs  = &avg_COUNTwtdpobs.,;
      %put .                    strata_count      = &strata_count.,;
      %put .                    allibls           = &allibls.;
      %put .                    );
   %end;

   /* Error handling macros */
   %let dist_fit_successful = &ORV_FALSE.;

   /* load macro for this frequency distribution */
   %let PARAM_COUNT = 0;
   %orvd_&fdist;
       
   /* intialize output datasets */
   %orv_InitializeFitData;
  
   /* Call the relevant fitting macros */
   %InitFit&fdist(countdata = &CountDataStandard);

   /* The Fit should produce a covariance matrix and a dataset with all of the parameter estimates.
      The fbluidest, freqmodels, and fparm datasets are constructed afterwards. */
   %Fit&fdist(
              CountData = &CountDataStandard, 
              outparm   = _ffit_countout_est_&ndist, 
              outcov    = _ffit_covmat_&ndist 
              );

   /* If we are using KRIs, check that all were kep by countreg that proc reg selected.
      If any were dropped, add them back in with a coefficient of 0 */
   %if %lowcase(&freq_use_KRIs.) eq true %then %do;
      %orv_checkForDroppedKRIs;
   %end;

   /* Check for successful fit. If not, abort. */
   data _NULL_;
      set _ffit_countout_est_&ndist.;
      call symputX("fit_status", status);
   run;

   %if &fit_status. eq 1 %then %do;
      /* Abort - set status and leave */
      data _ffit_countout_est_&ndist;
         length Distribution $12;
         set _ffit_countout_est_&ndist.;
         Distribution = upcase("&fdist.");
         CCUID        = &ccUid.;
         paramcount   = &PARAM_COUNT.;
         status       = 1;
      run;

      %return;
   %end;
   %else %do;
      /* drop any extraneous info from the fit */
      data _ffit_countout_est_&ndist.;
         set _ffit_countout_est_&ndist.;

         /* If intercept does not exist, add it */
         if missing(intercept) then do;
            intercept = log(FParam_1);
         end;

         keep
            reason
            status
            logl
            AIC
            SBC
            intercept
            FParam:
            &SelectedKRIs.
            &SelectedDummies. 
            ;
      run;
   %end;

   /* IF KRIs, adjust the estimates (because we standardized the KRIs) */
   %if &numSelectedKRIs gt 0 %then %do;
      /* Unstandardize the parameter estimates and the covariance matrix */
      %orv_unstandardize;

      %if &ORV_ABORT. %then %do;
         %goto DONE_FREQ_FIT_DIST;
      %end;
   %end;

   /* Do any post-fitting work here. Right now this is just covariance matrix
      adjustments for the fact that countreg estimates an intercept.  Need to
      make this adjustment AFTER the unstandardization process. */

   %PostFit&fdist(
                  parm   = _ffit_countout_est_&ndist, 
                  cov    = _ffit_covmat_&ndist 
                  );

   /* Create the output fbluid dataset */
   %orv_createFbluidDatasets;
   
   /* Add in distribution name, moments, and other info to the output datasets */
   %orv_FinalizeFreqDatasets();

   /* Create the Freq models dataset */
   %orv_CreateFreqModels;
   %if &ORV_ABORT. %then %do;
      %goto DONE_FREQ_FIT_DIST;
   %end;

   /* If we get here, then no errors were found */
   %let dist_fit_successful = &ORV_TRUE.;

   /* --------------------------------------------------------- */
   %DONE_FREQ_FIT_DIST:
   /* --------------------------------------------------------- */
   
   %if &dist_fit_successful. eq &ORV_FALSE. %then %do;
      /* May have produced parameter estiamtes or covariance matrix that are meaningless.  Set everything
         back to blank with the appropriate message */
      %orv_InitializeFitData;
      %put &orvar_error &ORV_REASON;
      %put &orvar_error %sysfunc(sasmsg(&orvads, ffdist_fit_failure, noquote, &FDIST.));
   %end;
%mend orva_freq_fit_dist;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_InitializeFitData;
   %orvu_nullify_data(data = 
                             _ffit_countout_est_&ndist. 
                             _ffit_covmat_&ndist.
                             _ffit_fbluidest_&ndist.   
                             _ffit_freqmodels_&ndist. 
                             );
   data _ffit_countout_est_&ndist.;
      length Distribution $12;
      Distribution = upcase("&fdist");
      CCUID = &ccUid;
      paramcount = &PARAM_COUNT;
      status = 1;
      reason = "&ORV_REASON.";

      %do i=1 %to &PARAM_COUNT;
         Param_&i. = .;
      %end;

      mean = .;
      variance = .;
   run;

   data _ffit_fbluidest_&ndist;
      CCUID = &ccUid;
      length 
             Distribution        $12 
             Level_Variable_Name $32
             ;
      Distribution = upcase("&fdist");
      Estimate = .;
      _NAME_ = "";
      Level_Variable_Name = "";
      VALUE = .;
   run;

%mend orv_InitializeFitData;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_checkForDroppedKRIs;

   /* First, get a list of the KRIs from the fitting */
   %local 
      fit_KRIs
      _i_
      kri_name
      ;
   data _ffit_KRIs_;
      length NAME $ 1000;
      set _ffit_countout_est_&ndist;
      array nums[*] _NUMERIC_;
      do i = 1 to dim(nums);
         if lowcase(substr(vname(nums[i]),1,3)) eq "kri" then do;
            NAME = vname(nums[i]);
            output;
         end;
      end;
      keep NAME;
   run;

   proc sql noprint;
      select 
               NAME
      into:
               fit_KRIs separated by " "
      from
               _ffit_KRIs_
      ;
   quit;

   /* Now, go through the list of KRIs chosen by proc reg. If any are not in the list
      produced from the fit, add them back in */
   %if %sysfunc(countw(&fit_KRIs.," ")) lt %sysfunc(countw(&selectedKRIs.," ")) %then %do;
      %do _i_ = 1 %to %sysfunc(countw(&selectedKRIs.," "));
         %let kri_name = %scan(&selectedKRIs., &_i_., " ");

         %if %index(&fit_KRIs., &kri_name.) eq 0 %then %do;
            data _ffit_countout_est_&ndist;
               set _ffit_countout_est_&ndist;
               &kri_name. = 0;
            run;

            data _ffit_covmat_&ndist;
               set _ffit_covmat_&ndist. end = last;
               array nums[*] _NUMERIC_;
               &kri_name. = 0;
               output;
               if last then do;
                  _NAME_ = "&kri_name.";
                  do i = 1 to dim(nums);
                     nums[i] = 0;
                     &kri_name. = 0;
                  end;
                  output;
               end;
               drop i;
            run;

         %end;
      %end;
   %end;

%mend orv_checkForDroppedKRIs;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_createFbluidDatasets;
   /*------------------------------------------------------------------*/
   /*                      CREATE _ffit_FBLUIDEST                      */
   /*------------------------------------------------------------------*/
   %if %lowcase(&Freq_model_is_stratified.) eq true %then %do;
      %if %eval(&numSelectedDummies) GT 0 %then %do;
         data _ffit_fbluidesta;
            set _ffit_countout_est_&ndist.(keep = Dummy: );
         run;

         proc transpose 
            data = _ffit_fbluidesta 
            out  = _ffit_fbluidest_&ndist.( rename = ( COL1 = Estimate ) )
            ;
            var Dummy: ;
         run;

         proc sort data = _ffit_dummymap;
            by _NAME_;
         run;

         data _ffit_dummymap;
            set _ffit_dummymap;
            _NAME_ = upcase(_NAME_);
         run;

         proc sort data = _ffit_fbluidest_&ndist.;
            by _NAME_;
         run;

         data _ffit_fbluidest_&ndist.;
            set _ffit_fbluidest_&ndist.;
            _NAME_ = upcase(_NAME_);
         run;

         data _ffit_fbluidest_&ndist.;
            merge 
               _ffit_dummymap 
               _ffit_fbluidest_&ndist.
            ;
            by 
               _NAME_
            ;
         run;
      %end; 
      %else %do;
         data _ffit_fbluidest_&ndist;
            set _ffit_dummymap;
            Estimate = . ;
            _NAME_ = upcase(_NAME_);
         run;
      %end;

      data _ffit_fbluidest_&ndist.;
         CCUID = &ccUid.;
         length Distribution $12;
         Distribution = upcase("&fdist.");
         set _ffit_fbluidest_&ndist.;

         if missing(Estimate) then do;
            Estimate = 0;
         end; 
      run;
   %end;
%mend orv_createFbluidDatasets;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_FinalizeFreqDatasets();
   /* finalize parameter estimates dataset */
   data _ffit_countout_est_&ndist.;
      length Distribution $12;
      length SelectedKRIs selectedDummies $ 10000;

      set _ffit_countout_est_&ndist;
      %Moments&fdist.;

      Distribution = upcase("&fdist");
      CCUID                   = &ccUid;
      paramcount              = &PARAM_COUNT;
      COUNT_KRI               = &numSelectedKRIs;
      COUNT_DUMMIES           = &numSelectedDummies; 
      Avg_count_weighted_pobs = &avg_COUNTwtdpobs.;
      Freq_ObsCount           = &Freq_ObsCount.;
      Freq_DegreesOfFreedom   = Freq_ObsCount - paramcount;
      Freq_Strata_count       = &Strata_count.;
      SelectedKRIs            = %if &numSelectedKRIs. gt 0 %then %do;
                                   "&SelectedKris"; 
                                %end;
                                %else %do;
                                   "";
                                %end;
      SelectedDummies     = ""; /* <= Filled in later when renaming indicators. */

   run;
      
   /* finalize covariance matrix dataset */
   data _ffit_covmat_&ndist.; 
      length Distribution $32 _NAME_ $32;
      set _ffit_covmat_&ndist.;
      Distribution = upcase("&fdist");
      CCUID = &ccUid;
   run;

   /* Rename indicators (dummies) */
   %orv_RenameIndicators;
            
%mend orv_FinalizeFreqDatasets;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_RenameIndicators;
   %if &numSelectedDummies. > 0 %then %do;
      /* Create the rename statement */
      proc sql noprint;
         SELECT
                  trim(_NAME_)||"="||trim(_DISPLAY_NAME_),
                  trim(_DISPLAY_NAME_)
         INTO
                  :rename_stmt SEPARATED BY " ",
                  :new_KRI_names SEPARATED BY " "
         FROM
                  _ffit_DummyMap
         WHERE
                  upcase(trim(_NAME_)) in (&selectedDummies_comma.)
                  ;
      quit;

      /* Rename variables and change the values of '_NAME_' column in covmat */
      data _ffit_dummymap;
         set _ffit_dummymap;
         _NAME_ = upcase(_NAME_);
      run;

      data _ffit_covmat_&ndist(rename = (&rename_stmt.));
         set _ffit_covmat_&ndist;
         length 
                _DISPLAY_NAME_ $100 
                _NAME_         $32
                ;

         if _N_ = 1 then do;
            declare hash dispNameLookup(dataset: '_ffit_DummyMap');
            dispNameLookup.defineKey('_NAME_');
            dispNameLookup.defineData('_DISPLAY_NAME_');
            dispNameLookup.defineDone();         
         end;
         if lowcase(substr(_NAME_,1,5)) eq "dummy" then do;
            _NAME_ = upcase(_NAME_);
            dispNameLookup.find();
            _NAME_ = trim(_DISPLAY_NAME_);
         end;
         drop _DISPLAY_NAME_;
      run;

      /* Rename variables in paramter estimate dataset */
      proc datasets nolist nodetails;
         modify _ffit_countout_est_&ndist;
         rename &rename_stmt.;
      quit;
      data _ffit_countout_est_&ndist;
         set _ffit_countout_est_&ndist;
         selectedDummies = "&new_KRI_names.";
      run;
   %end;

%mend orv_RenameIndicators;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/

%macro orv_unstandardize;
   %local 
         KRI_NAMES
         KRI_STDs 
         KRI_Means 
         Intercept_Adj 
         FP1_old 
         Intercept_Adj_Val
         ;

   /* From standardized KRI data, get info for adjusting parameters */
   proc sql noprint;
      SELECT
               Name,
               put(stdDev,best10.),
               put(Mean,best10.),
               put(Mean,best10.)||"/"||put(stdDev,best10.)||"*"||Name
      INTO
               :KRI_NAMES     SEPARATED BY " ",
               :KRI_STDs      SEPARATED BY " ",
               :KRI_Means     SEPARATED BY " ",
               :Intercept_Adj SEPARATED BY "+"
      FROM
               _ffit_standard_stats
      WHERE
               Name in (&selectedKRIs_comma.)
               ;
   quit;

   /* Adjust the parameter estimates */
   %orv_unstandardize_parameters;

   %if &ORV_ABORT. eq &ORV_TRUE. %then %do;
      %return;
   %end;

   /* Adjust the covariance matrix of the parameter estimates */
   %orv_unstandardize_covariance;

%mend orv_unstandardize;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_unstandardize_parameters;
   %local 
         nm 
         curr_std 
         counter
         ;

   /* Adjust the parameters - see stmajo's doc on covariance adjustments for details (last example) */
   data _ffit_countout_est_&ndist.;
      set _ffit_countout_est_&ndist.;

      call symputX("intercept_old",intercept);
      call symputX("Intercept_Adj_Val",put((&Intercept_Adj.),BEST20.));

      /* Adjust everything in the exponent before evaluating it */
      intercept = intercept - (&Intercept_Adj.);

      %do counter=1 %to &numSelectedKRIs.;
         %let nm = %scan(&selectedKRIs.,&counter.," ");
         %let curr_std = %scan(&KRI_STDs.,&counter.," ");
         &nm. = &nm./&curr_std.;
      %end;

   run;

%mend orv_unstandardize_parameters;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_unstandardize_covariance;
   /* Adjust Covariance Matrix */

   %local 
         varnames
         varnames_quote
         row_name
         row_ID
         col_name
         col_ID
         nm 
         mean 
         std
         ind
         j
         ;

   /* Create the Hessian matrix (see stmajo's doc on changing covariance matrices- last example) */

   %let varnames = ;
   %let varnames_quote = ;
   data _null_;
      set _ffit_covmat_&ndist.;
      if _N_ = 1;
      length 
         varnames       $ 32000
         varnames_quote $ 32000
         ;
      array vars _NUMERIC_;
      do i = 1 to dim(vars);
         varnames = strip(varnames) || " " || vname(vars[i]);
         varnames_quote = strip(varnames_quote) || ' "' || vname(vars[i]) || '"';
      end;
      call symput("varnames",varnames);
      call symput("varnames_quote",varnames_quote);
   run;

   data _ffit_Hessian&ndist.;

      if _N_ = 0 then do;
         set _ffit_covmat_&ndist.(drop = _NAME_);
      end;

      /* Lopp over the variable names in the covariance dataset */
      %do row_ID = 1 %to %sysfunc(countw(&varnames.," "));
         %let row_name = %scan(&varnames., &row_ID.);

         /* KRI rows */
         %if %lowcase(%substr(&row_name.,1,3)) eq kri %then %do;
            %do col_ID = 1 %to %sysfunc(countw(&varnames.," "));
               %let col_name = %scan(&varnames., &col_ID.);
               %if &col_name. eq &row_name. %then %do;
                  %let ind   = %sysfunc(findw(&KRI_NAMES.,&col_name.," ",E));
                  %let std   = %scan(&KRI_STDs., &ind., " ");
                  &col_name. = 1 / &std.;
               %end;
               %else %do;
                  &col_name. = 0;
               %end;
            %end;
            output;
         %end;

         /* Dummy rows */
         %if %lowcase(%substr(&row_name.,1,3)) eq dum %then %do;
            %do col_ID = 1 %to %sysfunc(countw(&varnames.," "));
               %let col_name = %scan(&varnames., &col_ID.);
               %if &col_name. eq &row_name. %then %do;
                  &col_name. = 1;
               %end;
               %else %do;
                  &col_name. = 0;
               %end;
            %end;
            output;
         %end;

         /* Intercept Row */
         %if %lowcase(&row_name.) eq intercept %then %do;
            %do col_ID = 1 %to %sysfunc(countw(&varnames.," "));
               %let col_name = %scan(&varnames., &col_ID.);

               /* KRI columns */
               %if %lowcase(%substr(&col_name.,1,3)) eq kri %then %do;
                  %let ind   = %sysfunc(findw(&KRI_NAMES.,&col_name.," ",E));
                  %let mean  = %scan(&KRI_Means., &ind., " ");
                  %let std   = %scan(&KRI_STDs., &ind., " ");
                  %let nm    = %scan(&selectedKRIs., &ind., " ");
                  &col_name. = -&mean. / &std.;
               %end;

               /* Dummy Columns */
               %if %lowcase(%substr(&col_name.,1,3)) eq dum %then %do;
                  &col_name. = 0;
               %end;

               /* Intercept Column */
               %if %lowcase(&col_name.) eq intercept %then %do;
                  Intercept = 1;
               %end;

               /* Other Model parameters */
               %do j = 2 %to &Param_Count;
                  FParam_&j. = 0;
               %end;
            %end;

            output;
         %end;

         /* FParam_* Rows */
         %if %lowcase(%substr(&row_name.,1,3)) eq fpa %then %do;
            %do col_ID = 1 %to %sysfunc(countw(&varnames.," "));
               %let col_name = %scan(&varnames., &col_ID.);
               %if &col_name. eq &row_name. %then %do;
                  &col_name. = 1;
               %end;
               %else %do;
                  &col_name. = 0;
               %end;
            %end;

            output;
         %end;
      %end;
   run; 

   /* Now, compute:  fcov_&ndist = Hessian * cov * Hessian_T */
   proc iml;
      _NAME_={&varnames_quote.};
      use _ffit_covmat_&ndist;

      read all var _num_ into C; 
      close _ffit_covmat_&ndist;

      use _ffit_Hessian&ndist.;
      read all var _num_ into M;

      close _ffit_Hessian&ndist.;
      X = M * C * (M`);

      create fcov_&ndist from X[
                                colname = _NAME_ 
                                rowname = _NAME_
                                ];
      append from X[rowname=_NAME_];
      close fcov_&ndist;
   quit;

   data _ffit_covmat_&ndist;
      length _NAME_ $ 32;
      set fcov_&ndist;
   run;

%mend orv_unstandardize_covariance;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_CreateFreqModels;
   %local 
         i
         modString 
         mTableHeader 
         ;

   %do i = 2 %to &Param_Count.;
      %local Displayname&i.;
   %end;

   /* Get the frequency model and predicted moments (used to be in orva_freq_calc_mean) */
   /* The predicted moments will overwrite the moments determined from %mom&fdist. */
   %orv_CalcMomsAndModelStrings;
   %if &ORV_ABORT %then %do; 
      %return;
   %end;

   /*- put distribution into the freqmodels datasets -*/
   %DisplayModel&dist( 
                      fparmnames, 
                      _ffit_freqmodels_&ndist, 
                      _ffit_countout_est_&ndist, 
                      modString, 
                      mTableHeader
                      );

   data _ffit_freqmodels_&ndist;  
      if _N_ = 1 then do;
         length 
               display         $ 1000 
               meanTableHeader $ 1000
               model_level     $ 1000
               level_display   $ 1000
               ; 
         set _ffit_countout_est_&ndist(keep =  
                                             Distribution 
                                             FParam_1
                                             %if &numSelectedKRIs gt 0 %then %do;
                                                &SelectedKRIs 
                                             %end; 
                                             %if &numSelectedDummies gt 0 %then %do;
                                                Dummy_: 
                                             %end;
                                             );
      end;
      set _ffit_freqmodels_&ndist;

      display         = "&modString";
      meanTableHeader = "&mTableHeader";  

      %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
         model_level = "%sysfunc(sasmsg(&orvads, ffdist_internal_business_line, noquote))";
      %end;
      %else %do;
         model_level = "%sysfunc(sasmsg(&orvads, ffdist_cell, noquote))";
         internal_bl_uid = .;
      %end;
 
      %do i = 2 %to &Param_Count.;
         &&Displayname&i. = FParam_&i.;
      %end;
   run;

   /* Get the text that will displayed in the left column of the table.
      For cell level models, it is the names of the CCVs in that cell.
      For IBL models, it is the name of the IBLs */
   /* For display, get the name of the CCV's for this CCUID */
   %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
      data _ffit_freqmodels_&ndist;
         set _ffit_freqmodels_&ndist;
         length Internal_BL_Name $ 1000;

         if _N_ = 1 then do;
            /* Look-up table for the IBL names */
            declare Hash IBLLookup(dataset: "&allibls.");
            IBLLookup.defineKey("Internal_BL_UID");
            IBLLookup.defineData("Internal_BL_Name");
            IBLLookup.defineDone();
         end;

         IBLLookup.find();
         level_display = Internal_BL_Name;
         drop Internal_BL_Name;
      run;
   %end;
   %else %do;
      %let ccds   = ;
      %orvu_find_dataset(
                         nameIn  = cc, 
                         Libs    = VARPRJ VARPRJ_M, 
                         nameOut = ccds
                         );
      %let ccmap0 = ;
      %orvu_find_dataset(
                         nameIn  = ccmapval_0, 
                         Libs    = VARPRJ VARPRJ_M, 
                         nameOut = ccmap0
                         );
      %let ccmap1 = ;
      %orvu_find_dataset(
                         nameIn  = ccmapval_1, 
                         Libs    = VARPRJ VARPRJ_M, 
                         nameOut = ccmap1
                         );
      %if &NUMCCVS. = 3 %then %do;
         %let ccmap2 = ;
         %orvu_find_dataset(
                            nameIn  = ccmapval_2, 
                            Libs    = VARPRJ VARPRJ_M, 
                            nameOut = ccmap2
                            );
      %end;
      
      proc sql;
         CREATE TABLE 
                           _ffit_work AS
         SELECT DISTINCT 
                           * 
         FROM 
                           &ccds.   cc,
                           &ccmap0. ccv0,
                           &ccmap1. ccv1
                           %if &NUMCCVS. = 3 %then %do; 
                              ,&ccmap2. ccv2 
                           %end;
         WHERE 
                           cc.CCUID = &CCUID.
                  AND
                           cc.CCVUID_0 = ccv0.CCVUID_0 
                  AND 
                           cc.CCVUID_1 = ccv1.CCVUID_1  
                          %if &NUMCCVS. = 3 %then %do; 
                              AND 
                              cc.CCVUID_2 = ccv2.CCVUID_2 
                           %end;
                           ;  
      quit;

      data _null_;
         set _ffit_work;
         txt_and = sasmsg("&orvrds", " / ", "noquote");
         level_display = strip(CCVVALUE_0) || " " || strip(txt_and) || " " || strip(CCVVALUE_1)
         %if &NUMCCVS. = 3 %then %do;
            || " " || strip(txt_and) || " " || strip(CCVVALUE_2) 
         %end;
         ;
         call symput("level_display",trim(level_display));
      run;
    
      data _ffit_freqmodels_&ndist;
         set _ffit_freqmodels_&ndist;
         level_display = "&level_display.";
      run;
   %end;
%mend orv_CreateFreqModels;

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_CalcMomsAndModelStrings;           
   %local 
         i 
         TheFreqKriModel 
         TheFreqKriModel_Full
         ;

   /* Store the KRI models in macros */
   %let TheFreqKriModel = 0;
   %let TheFreqKriModel_Full = 0;
   %orv_CreateFreqKriModelString;

   %if &ORV_ABORT eq &orv_true %then %do; 
      %return;
   %end;

   /* Sort the counts by the strata */
   %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
      proc sort data = &countdata.; 
         by &Freq_Model_basis_props.; 
      run;
   %end;

   /* _ffit_data1 is used to compute the historical mean.  For each row in count data, 
       it has the model-predicted mean */
   /* _ffit_freqmodels_&ndist will have the entry for this distribution in the fparm_mdli dataset */
   data _ffit_data1 
        _ffit_freqmodels_&ndist.( keep = 
                                        %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
                                           &Freq_model_basis_props.
                                        %end;
                                        StratumModel
                                        parameter_model
                                        FParam:
                                        );

      set &CountData. end = last;

      %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
         by &Freq_model_basis_props.;
      %end;

      format 
            StratumModel $1000. 
            parameter_model $1000.
            ;

      if _N_ = 1 then do;
         /* Look-up table for the dummy variables */
         %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
            declare Hash EstLookup(dataset: "_ffit_fbluidest_&ndist.");
            EstLookup.defineKey("Value");
            EstLookup.defineData("Estimate");
            EstLookup.defineDone();
         %end;

         /* Look-Up table for the model parameters */
         declare Hash ParmLookup(dataset: "_ffit_countout_est_&ndist.");
         ParmLookup.defineKey("CCUID");

         %do i = 1 %to &PARAM_COUNT.;
            ParmLookup.defineData("FParam_&i.");
         %end;

         ParmLookup.defineDone();
      end;

      /* Compute the Fparam_1 modified by KRIs for this row, then get 
         the mean and variance*/
      Estimate = 0;
      CCUID    = &CCUID.;

      %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
         Value = &Freq_Model_Basis_Props.;
         EstLookup.find(); 
      %end;

      if missing(Estimate) then Estimate = 0;
      ParmLookup.find();
      FParam_1 = exp(log(FParam_1) + &TheFreqKriModel. + Estimate);

      if missing(FParam_1) then do;
         FParam_1 = 0;
      end;

      %Moments&fdist.;

      /* Get the original parameters again */
      ParmLookup.find();

      /* Create the stratum model string for the fparm_mdli dataset */
      if (&numSelectedKRIs. > 0) or (&numSelectedDummies. > 0 and Estimate ne 0) then do;
         stratumModel = " * exp(";
         if (&numSelectedKRIs. > 0) then do;
            stratumModel =  strip(stratumModel) || "&TheFreqKriModel";
         end;
         if (&numSelectedDummies. > 0) and (Estimate ne 0) then do;
            if (&numSelectedKRIs. > 0) then do;
               stratumModel =  strip(stratumModel) || " + ";
            end;
            stratumModel =  strip(stratumModel) || Estimate;
         end;
         stratumModel =  strip(stratumModel) || ")";
      end;
      
      parameter_model = put(FParam_1, BEST10.) || " " || strip(stratumModel);
      stratumModel = "FParam_1 " || strip(stratumModel);

      output _ffit_data1;
      
      %if %lowcase(&freq_model_is_stratified.) eq true %then %do;
         if last.&Freq_model_basis_props. then do;
            output _ffit_freqmodels_&ndist.;
         end;
      %end;
      %else %do;
         if last then do;
            output _ffit_freqmodels_&ndist.;
         end;
      %end;
   run;

   /* Get summary statistics */
   proc summary data = _ffit_data1;
      var 
         mean 
         variance
         ;

      %if %lowcase(&freq_model_is_stratified) eq true %then %do;
         by &Freq_Model_Basis_Props; 
      %end;

      weight net_exposure; 
      output
            out  = _ffit_summary(drop = _TYPE_ _FREQ_) 
            mean = avgMean avgVariance
            ;
   run;

   /* Include summary statistics in the Freqmodels dataset */
   data _ffit_freqmodels_&ndist.(rename = (avgMean = HistoricalStratumMean));
      set _ffit_summary;
      set _ffit_freqmodels_&ndist.;

      %if %lowcase(&freq_model_is_stratified) eq true %then %do;
         by &Freq_Model_Basis_Props;
      %end;
   run;

   /* Get predicted mean and variance */
   data _NULL_;
      set _ffit_summary end=last;
      retain MTot 0 VTot 0;
      MTot= Mtot + avgMean;
      VTot = VTot + avgVariance;

      if last then do;
         call symputX("ThePredMean", put(MTot,BEST20.));
         call symputX("ThePredVar",  put(VTot,BEST20.));
      end;  
   run;

   /* Put predicted moments into the parameter estimate dataset, as well as the full KRI model */
   data _ffit_countout_est_&ndist.;
      set _ffit_countout_est_&ndist.;
      length Freq_KriModel_Full $ 2000;

      Mean     = &ThePredMean.;
      Variance = &ThePredVar.;
      Freq_KriModel_Full = "exp(&TheFreqKriModel_Full)"; 

      if missing(Mean) OR missing(Variance) then do;
         call symput("ORV_REASON", "%sysfunc(sasmsg(&orvads, fcm_mean_or_variance_missing_error, noquote))");
         reason = "%sysfunc(sasmsg(&orvads, fcm_mean_or_variance_missing_error, noquote))";
      end;
   run;
%mend orv_CalcMomsAndModelStrings;  

/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
%macro orv_CreateFreqKriModelString;
   /*- Create the Frequency KRI Model as a macro- only has KRI effects -*/
   %if &numSelectedKRIs >0 or &numSelectedDummies > 0 %then %do;
      data _null_;
         set _ffit_countout_est_&ndist.;
         retain something_there 0;

         length TheModel TheModelFull $ 10000 ID $ 12 plus $ 3;
         TheModel = "";
         TheModelFull = "";

         /* If there are KRIs, add them to the model */
         %if &numSelectedKRIs > 0 %then %do;  
            array MyKRIs[&numSelectedKRIs] &SelectedKRIs.;
            TheModel = MyKRIs[1] || " * " || vname(MyKRIs[1]);
            something_there = something_there + 1;

            do i=2 to &numSelectedKRIs.;
               TheModel = trim(TheModel) || " + " || MyKRIs[i] || " * " || vname(MyKRIs[i]);   
            end;
            TheModelFull = TheModel;
         %end;

         /* If there are Dummies, add them to the full model */
         %if &numSelectedDummies > 0 %then %do;  
            array MyDummies Dummy_:;

            do i = 1 to dim(MyDummies);
               /*Get the numeric identifier at the end of the Dummy name */
               if MyDummies[i] ne . then do;
                  ID = compress(vname(MyDummies[i]),'A','F');
                  if something_there then do;
                     plus = " + ";
                  end;
                  else do;
                     plus = "";
                  end;
                  TheModelFull = trim(TheModelFull) || plus 
                     || MyDummies[i] || " * " || "(&Freq_Model_Basis_Props. =" || strip(ID) || ")";
                  something_there = something_there + 1;
               end;
            end;
         %end;

         call symput("TheFreqKriModel", trim(TheModel)); 
         call symput("TheFreqKriModel_Full", trim(TheModelFull)); 
      run;    
      %let TheFreqKriModel = %trim(&TheFreqKriModel); 
   %end;

   %put %sysfunc(sasmsg(&orvads, fcm_freq_kri_model_vlabel, noquote, &TheFreqKriModel_Full));

   /* ERROR checking */
   %if &SYSERR NE 0 & &SYSERR NE 4 %then %do;
      %let ORV_ABORT = &ORV_TRUE.;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, fcm_cant_construct_kri_model_string_error, noquote));
      data &countout_est; 
         set &countout_est;
         status=1;
         reason="&ORV_REASON.";
      run;
   %end;
%mend orv_CreateFreqKriModelString;
