/*****************************************************************************************
 *
 * NAME: ORVA_freq_kri_effect.sas                                             
 * DESCRIPTION: Compute the effect of the kris on the frequency distribution
 * SUPPORT: jariou       
 *
 * PARAMETERS: 
 *
 *    freqPriorityDataset     = Name of the frequency priority dataset. 
 *                              The data set has one row for each cell containing:
 *                              SOURCE_TYPE   I,C,T,S,B  
 *                                                        I = Internal
 *                                                        C = Consortium
 *                                                        T = Consortium Tool
 *                                                        S = Scenario
 *                                                        B = BucketScenario
 *                                                        M = Mixed
 *                              SOURCE_NAME   for scenarios
 *                              DATASET_NAME  freq fits dataset
 *                              CCUID         cell key
 *                              FMODELS_DS    dataset from %orva_ffit
 *                              BLESTS_DS     dataset from %orva_ffit
 *    freqKriDatasetBl        = Dataset with the simulation values for the Businessline
 *                              frequency KRI variables. Must have:
 *                              BLUID_KRI and <risk_factors> and RISK_FACTOR_VARIABLE_DATE
 *    freqFriDatasetFirm      = Dataset with the simulation values for the Firm
 *                              frequency KRI variables. Must have:
 *                              BLUID_KRI and <risk_factors> and RISK_FACTOR_VARIABLE_DATE
 *    ccData                  = Name of the Cross Classification dataset containing:
 *                              CCUID, CCVUID_0, CCVUID_1
 *    outFreqStatus           = Success or failure of the frequency effect calculation 
 *    outFParmEffect          = Frequency parameters output dataset 
 *                                             
 **************************************************************************************/
%macro ORVA_freq_kri_effect(  
                              freqPriorityDataset     = ,
                              freqKriDatasetBl        = ,   /* Only needed for data w/ KRIs */
                              freqKriDatasetFirm      = ,   /* Only needed for data w/ KRIs */
                              ccData                  = ,
                              outFreqStatus           = outFreqStatus,
                              outFParmEffect          = outfParmEffect,
                              interval                = YEAR,
                              SimulationDate          = ,
                              DataVersion             = 
                              );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_freq_kri_effect(;  
      %put .                      freqPriorityDataset     = &freqPriorityDataset,;
      %put .                      freqKriDatasetBl        = &freqKriDatasetBl,;
      %put .                      freqKriDatasetFirm      = &freqKriDatasetFirm,;
      %put .                      ccData                  = &ccData,;
      %put .                      outFreqStatus           = &outFreqStatus,;
      %put .                      outFParmEffect          = &outFParmEffect,;
      %put .                      interval                = &interval,;
      %put .                      SimulationDate          = &SimulationDate,;
      %put .                      DataVersion             = &DataVersion;
      %put .                      );
   %end;

   /*- create the status data set upfront -*/
   %let ORV_REASON = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( 
                     statusds = &outFreqStatus, 
                     status   = 1, 
                     reason   = &ORV_REASON 
                     );
   
   %let status = 0;
   %let ORV_REASON = %sysfunc(sasmsg(&orvads, successful_note, noquote)) ;
   %let ORV_ABORT  = &ORV_FALSE;
   %let CELL_WHERE = 1;
   %let NUMCCVS    = 0;

   /*- check &freqPriorityDataset input parameter -*/
   %if "&freqPriorityDataset" = "" %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, fke_specify_freqPriority_ds_error, noquote));
        %let ORV_ABORT  = &ORV_TRUE;
        %let ORV_REASON = %sysfunc(sasmsg(&orvads, fke_freqPriority_missing_error, noquote));
        %goto DONE;
   %end;

   %let ds_nobX = 0;
   %let dsid = %sysfunc(open(&freqPriorityDataset,I));
   %if &dsid NE 0 %then %do;
        %let ds_nobX = %sysfunc(attrn(&dsid,nlobs));
        %let rc = %sysfunc(close(&dsid));
   %end;
   %else %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, fke_cant_open_freqPriority_ds_error, noquote));
        %let ORV_ABORT  = &ORV_TRUE;
        %let ORV_REASON = %sysfunc(sasmsg(&orvads, fke_cant_open_freqPriority_ds_error, noquote));
        %goto DONE;     
   %end;
   %if &ds_nobX LE 0 %then %do; /* should have at least one obs */
        %put &orvar_error %sysfunc(sasmsg(&orvads, fke_freqPriority_ds_empty_warning, noquote, &freqPriorityDataset));
        %let ORV_ABORT  = &ORV_TRUE;
        %let ORV_REASON = %sysfunc(sasmsg(&orvads, fke_freqPriority_ds_empty_warning, noquote, &freqPriorityDataset));
        %goto DONE;
   %end;

   /*- Find the number of rows in freqPriorityDataset, excluding rare events -*/
   /*- Get the list of CCUID's, datasets, etc. for which calculations need to be made -*/
   proc SQL noprint;
      SELECT 
            put(count(SOURCE_TYPE), 8.-L)
      INTO
            :NROWS
      FROM
            &freqPriorityDataset
      WHERE
            SOURCE_TYPE NE "R"
            ;
      SELECT 
            dataset_name, 
            CCUID, 
            SOURCE_TYPE, 
            SCENARIO_TYPE,
            FMODELS_DS,
            BLESTS_DS
      INTO 
            :DSNAME1 - :DSNAME&NROWS , 
            :CCUID1 - :CCUID&NROWS , 
            :SOURCE_TYPE1 - :SOURCE_TYPE&NROWS, 
            :SCENARIO_TYPE1 - :SCENARIO_TYPE&NROWS,
            :FREQMODEL1 - :FREQMODEL&NROWS,
            :BLESTS1 - :BLESTS&NROWS
      FROM 
            &freqPriorityDataset 
      WHERE
            SOURCE_TYPE NE "R"
            ;
   quit;  
   %if &SYSERR NE 0 & &SYSERR NE 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, fke_cant_make_macros_with_these_names, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, fke_cant_make_macros_with_these_names, noquote));
      %goto DONE;
   %end;

   /************************************************************************/
   /* Loop through the rows to calculate the KRI & dummy effects           */
   /************************************************************************/
   /*  Take care of source types that would not have output datasets       */
   /*  from calling orva_ffit().                                           */
   /*  These would include mixed (Relative Relations) cells and scenarios  */
   /************************************************************************/
   %do _fke_index = 1 %to &NROWS;
      /* These cases amount to all the possibilities except for scenarios */
      %if 
         "&&SOURCE_TYPE&_fke_index" NE "I" 
         AND 
         "&&SOURCE_TYPE&_fke_index" NE "C" 
         AND 
         "&&SOURCE_TYPE&_fke_index" NE "T" 
         AND 
         "&&SOURCE_TYPE&_fke_index" NE "F" 
         AND 
         "&&SOURCE_TYPE&_fke_index" NE "M" 
      %then %do;      
         %orv_scenariosGetFreqEffect(
                                       dsName       = &&DSNAME&_fke_index, 
                                       CCUID        = &&CCUID&_fke_index,
                                       sourceType   = &&SOURCE_TYPE&_fke_index,
                                       scenarioType = &&SCENARIO_TYPE&_fke_index
                                       );
         %goto NEXTCCUID; 
      %end; 
      %else %if "&&SOURCE_TYPE&_fke_index" EQ "M" %then %do;
         data _fke_single_cell_results; 
            set  &&DSNAME&_fke_index; 
            where CCUID = &&CCUID&_fke_index;
            Source = "&&SOURCE_TYPE&_fke_index";
            keep ccuid distribution source FParam_:;
         run;

         %orvu_append(
                      base         = _fke_all_cells, 
                      new          = _fke_single_cell_results, 
                      canUseAppend = false
                      );

         %goto NEXTCCUID; 
      %end;
 
      %local breakReason;

      %orv_fittedModelsGetEffect(
                                 dsName          = &&DSNAME&_fke_index, 
                                 CCUID           = &&CCUID&_fke_index,
                                 sourceType      = &&SOURCE_TYPE&_fke_index
                                 );

      %if &breakReason = DONE %then %do;
         %goto DONE;
      %end;
   
%NEXTCCUID:
   %end; 

   /*********** adjust fparams for CTool models (Source=T) **********/
   data &outFParmEffect(keep = 
                               Distribution 
                               Fparam_: 
                               CCUID 
                               d_: 
                               Source 
                               );
      set _fke_all_cells;

      if Source = "T" then do;
         width_fit = intnx( Freq_Fit_Interval, '03Apr2001'd, 1 ) - intnx( Freq_Fit_Interval, '03Apr2001'd, 0 );
         width_sim = intnx( "&interval", '03Apr2001'd, 1 ) - intnx( "&interval", '03Apr2001'd, 0 );
         ratio     = width_sim / width_fit ;
         call ChangeInterval( 
                              Distribution, 
                              ratio, 
                              &theFreqSimCallParams 
                              );
      end;
   run;
%DONE:
  
    %if &ORV_ABORT %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, fke_abnormal_error, noquote));
        %orva_status_ds( 
                        statusds = &outFreqStatus, 
                        status   = 1, 
                        reason   = &ORV_REASON 
                        );
    %end;  
    %else %do;
        %orva_status_ds( 
                        statusds = &outFreqStatus, 
                        status   = 0, 
                        reason   = &ORV_REASON 
                        );
    %end;
    
    %orvu_delete_data(data = _fke: );

%mend ORVA_freq_kri_effect;

%macro orv_scenariosGetFreqEffect(
                                    dsName       = , 
                                    CCUID        = , 
                                    sourceType   = , 
                                    scenarioType =
                                    );

   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_scenariosGetFreqEffect(;
      %put .                            dsName       = &dsName,; 
      %put .                            CCUID        = &CCUID,; 
      %put .                            sourceType   = &sourceType,;
      %put .                            scenarioType = &scenarioType;
      %put .                            );
   %end;

   data _NULL_;
      length 
            DATASET_NAME $&ORV_DATASET_NAME_LENGTH.
            ;
      DATASET_NAME     = "&dsName";
      dotPosition      = index(DATASET_NAME,".");
      scenarioPosition = index(DATASET_NAME,"Scenario");

      if scenarioPosition > 0 then do;
         DATASET_NAME = substr(DATASET_NAME,1, dotPosition ) || 
                        substr(DATASET_NAME,scenarioPosition,8) || 
                        substr(DATASET_NAME,scenarioPosition + 8)
                        ;
         call symput("dsName", DATASET_NAME);
      end;
   run;

   /*****************************************************************/
   /*   If this is a bucket scenario, I need to look at the dataset */
   /*   that is named VARPRJ??.scenario_plotbrkdwn_XXXX             */
   /*****************************************************************/
   %if &scenarioType eq B %then %do;
      %let plotLoc = %index(&dsName, _plot_);

      %if &plotLoc > 0 %then %do;
         %let dsName = %substr(&dsName, 1, &plotLoc + 4)brkdwn%substr(&dsName, &plotLoc + 5);
      %end;

      /********************************************/
      /*   Get the number of rows in the scenario */
      /*   put it  in RowCount                    */
      /********************************************/
      proc SQL noprint;
         SELECT 
                  count(CCUID)
         INTO   
                  :rowCount
         FROM   
                  &dsName
         WHERE  
                  CCUID  = &CCUID
                  ;
      quit;

      data _fke_scenario; 
         length 
                  Distribution      $12 
                  Freq_Fit_Interval $12 
                  source            $2
                  ;
         set &dsName;
         where CCUID = &CCUID;

         Source            = "&sourceType";
         Freq_Fit_Interval = "";
         Distribution      = "";
         FParam_1          = . ;
         FParam_2          = . ;

         /********************************************************/
         /*  Obtain the name and parameters of the frequency     */
         /*  distribution based on the Mean, Min, and Max inputs */
         /********************************************************/
         call BucketFreqSimDef(  
                                 Frequency_Mean,
                                 Frequency_Min,
                                 Frequency_Max,
                                 Distribution,
                                 FParam_1,
                                 FParam_2
                                 );
         keep 
               CCUID 
               Distribution 
               Fparam_: 
               Frequency_: 
               Scenario_: 
               Source 
               Freq_Fit_Interval
               ;
      run;
   %end;
   %else %do;/* Model scenario */
      proc SQL noprint;
         SELECT 
                  count(CCUID)
         INTO   
                  :rowCount
         FROM   
                  &dsName
         WHERE  
                  CCUID  = &CCUID
                  ;
      quit;

      data _fke_scenario; 
         length 
                  Distribution      $12 
                  Freq_Fit_Interval $12 
                  source            $2
                  ;

         set &dsName;
         where CCUID       = &CCUID;
         Source            = "&sourceType";
         Freq_Fit_Interval = "";
         Distribution      = Frequency_Dist;

         keep 
               CCUID 
               Distribution 
               Fparam_: 
               Frequency_: 
               Scenario_: 
               Source 
               Freq_Fit_Interval;
      run;
   %end;

   %if &rowCount GT 1 %then %do;
      %orva_freq_sum_model(
                           inData       = _fke_scenario, 
                           distModelVar = Distribution, 
                           outData      = _fke_single_cell_results
                           );
   %end;
   %else %do;
      data _fke_single_cell_results;
         set _fke_scenario; 
      run; 
   %end;

   %orvu_append(
                  base         = _fke_all_cells, 
                  new          = _fke_single_cell_results, 
                  canUseAppend = false
                  );
%mend orv_scenariosGetFreqEffect;

/***** get models from FREQMODEL dataset *****/
%macro orv_fittedModelsGetEffect(
                                 dsName          = , 
                                 CCUID           = , 
                                 sourceType      = 
                                 );

   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_fittedModelsGetEffect(;
      %put .                           dsName          = &dsName,; 
      %put .                           CCUID           = &CCUID,; 
      %put .                           sourceType      = &sourceType;
      %put .                           );
   %end;

   /* ------------------------------------------------------------------- */
   /*                        FITTED MODEL SCENARIO                        */
   /* ------------------------------------------------------------------- */
   /***********************************************************************/
   /* Start by handling the case where one has a fitted model scenario,   */
   /* i.e. the case where a model is fitted and then copied as scenario.  */
   /* Presently the UI does not allow KRI fits to be copied such.         */
   /***********************************************************************/
   %if %upcase(&sourceType) = F %then %do;
      data _fke_single_cell_results; 
         set  &&DSNAME&_fke_index; 
         where CCUID = &ccuid;
         Source = "&sourceType";

         if missing(intercept) then do;
            intercept = safe_log(FParam_1);
         end;

         keep 
              ccuid 
              distribution 
              source 
              FParam_:
              Freq_Fit_Interval
              intercept
              ;
     run;

     %goto COMPLETE_Handle_Fitted_Models;
   %end;

   /* ------------------------------------------------------------------- */
   /*                             FITTED MODEL                            */
   /* ------------------------------------------------------------------- */
   /* STRATEGY: FParam_1_mod will have the final fparam_1 value.  For cell level
      models without KRIs, this will be just the original fpram_1. The cases to 
      consider are: 
         (1) stratified or not.  If stratified, we need to get the Fparam_1 at each
             IBL and then sum them up.  Considering adding some dist macro functionality
             to genericise this
         (2) KRIs or not.  If there are KRIs, firm, IBL, both?  
             
   */
   %local 
         breakReason
         nobs
         Fparam_1_mod
         FREQ_ITERATION
         FITINTERVALUSED
         Freq_Model_is_stratified
         Freq_KRI_model
         Freq_Use_KRIs
         FParam_1_orig
         IBL_list
         IBL_count
         have_IBL_KRIs
         have_firm_KRIs
         _i_
         intercept
         ;

   %let breakReason = NONE;

   /* Get basic information about the fitted model */  
   data _null_;
      set &dsname.;
      where ccuid = &CCUID.;

      array parms FParam_:;

      if missing(intercept) then do;
         intercept = safe_log(FParam_1);
      end;

      call symputX( "FREQ_ITERATION",           Iteration                    );
      call symputX( "FITINTERVALUSED",          Freq_Fit_Interval            );
      call symputX( "Freq_Model_is_stratified", Freq_Model_is_stratified     );
      call symputX( "Freq_KRI_model",           substr(Freq_KRIModel_full,4) );
      call symputX( "Freq_Use_KRIs",            Freq_Use_KRIs                );
      call symputX( "FParam_1_orig",            FParam_1                     );
      call symputX( "intercept",                intercept                    );
      call symputX( "Freq_Distribution_Name",   Distribution                 );
      call symputX( "paramcount",               paramcount                   );

      do i = 1 to paramcount;
         if lowcase(vname(parms[i])) ne "fparam_1" then do;
            call symputX(vname(parms[i]), parms[i]);
         end;
      end;

   run;

   /* First handle the simplest case - the model is not stratified */
   %if %lowcase(&Freq_Model_is_stratified.) eq false %then %do;
      /* If we use KRIs, then use the firm KRI data to get the adjusted Fparam_1,
         otherwise, there is nothing to do */
      %if %lowcase(&Freq_Use_KRIs.) eq true %then %do;
         /* Check if there are firm KRIs, if there aren't then there is a problem */
         %let nobs = 0;
         %orvu_getObsCount(
                           &freqKriDatasetFirm, 
                           nobs
                           );
         %if nobs eq 0 %then %do;
            %let ORV_ABORT   = &ORV_TRUE;
            %let ORV_REASON  = %sysfunc(sasmsg(&orvads, fke_firm_KRI_data_missing , noquote));
            %let breakReason = DONE;
            %put &orvar_error &ORV_REASON.;
            %goto DONE_Handle_Fitted_Models;
         %end;

         data _null_;
            set &freqKriDatasetFirm.;
            Fparam_1_mod = exp(&intercept. + &Freq_KRI_model.);
            call symputX("Fparam_1_mod", Fparam_1_mod);
         run;
      %end; 
      %else %do;
         %let Fparam_1_mod = &Fparam_1_orig.;
      %end;
   %end;

   /* Now handle models that are stratified */
   %if %lowcase(&Freq_Model_is_stratified.) eq true %then %do;
      data _fke_models1;
         set &&FREQMODEL&_fke_index.;
         WHERE
               CCUID     = &CCUID.
           AND
               Iteration = &FREQ_ITERATION.
               ;
      run;

      /* check for empty dataset */
      %let nobs = 0;
      %orvu_getObsCount(
                        _fke_models1, 
                        nobs
                        );

      %if &nobs LE 0 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, fke_FModels_empty_warning, noquote));
         %let ORV_ABORT   = &ORV_TRUE;
         %let ORV_REASON  = %sysfunc(sasmsg(&orvads, fke_FModels_empty_warning, noquote));
         %let breakReason = DONE;
         %goto DONE_Handle_Fitted_Models;
      %end;    

      /* Get a list of all the IBLs */
      proc sql noprint;
         SELECT 
               internal_bl_uid
         INTO 
               :IBL_list separated by " "
         FROM 
               _fke_models1
               ;
      quit;

      %let IBL_count = %sysfunc(countw("&IBL_list."));

      /* Do we have KRIs?  */
      %let have_firm_KRIs = false;
      %let have_IBL_KRIs  = false;

      %if %lowcase(&Freq_Use_KRIs.) eq true %then %do;
         /* If we have IBL KRIs, Get a list of IBLs in the model and keep those KRIs */
         %let nobs = 0;

         %if &freqKriDatasetBL. ne %str() %then %do;
            %orvu_getObsCount(
                              &freqKriDatasetBL., 
                              nobs
                              );
         %end;

         %if &nobs. > 0 %then %do;
            %let have_IBL_KRIs  = true;

            data _fke_IBL_KRIs;
               set &freqKriDatasetBL.;
               where bluid_kri in (&IBL_list.);
               rename bluid_kri = internal_bl_uid;
            run;

            proc sort data = _fke_IBL_KRIs;
               by internal_bl_uid;
            run;
         %end;

         /* If we have firm KRIs, get those.  If we also have IBL KRIs, 
            repeat each observation for each IBL so that we can merge 
            the 2 datasets together */
         %let nobs = 0;

         %if &freqKriDatasetFirm. ne %str() %then %do;
            %orvu_getObsCount(
                              &freqKriDatasetFirm., 
                              nobs
                              );
         %end;

         %if &nobs. > 0 %then %do;
            %let have_firm_KRIs  = true;

            data _fke_firm_KRIs;
               set 
                  /* Repeat each row for each IBL so we can merge later. */
                  %do _i_ = 1 %to &IBL_count.;
                     &freqKriDatasetFirm.
                  %end;
                  ;
               %if &have_IBL_KRIs. eq false %then %do;
                  /* Add in internal_bl_uid in case there are dummies or IBL KRIs */
                  %do _i_ = 1 %to &IBL_count.;
                     if _N_ = &_i_. then do;
                        internal_bl_uid = %scan(&IBL_list., &_i_., " ");
                     end;
                  %end;
               %end;
            run;
         %end;

         /* Now create the combined KRI dataset and apply the frequency model to each row */
         data _fke_strat_model;

            %if &have_IBL_KRIs. eq true %then %do;
               set _fke_IBL_KRIs;
            %end;
            %if &have_firm_KRIs. eq true %then %do;
               set _fke_firm_KRIs;
            %end;

            Fparam_1 = exp(&intercept. + &Freq_KRI_model.);
         run;
      %end;

      /* Now handle the case with no KRIs */
      %else %do;
         data _fke_strat_model;
            set _fke_models1(keep = 
                                    FParam_:
                                    internal_bl_uid
                                    );
            Fparam_1 = exp(&intercept. + &Freq_KRI_model.);
         run;
      %end;

      data _fke_strat_model;
         set _fke_strat_model;
         length distribution $100;
         distribution = strip("&Freq_Distribution_Name.");

         %do i = 2 %to &paramcount.;
            FParam_&i. = &&FParam_&i.;
         %end;
      run;

      /* Now, each row of _fke_strat_model has the modfied fparam_1 for that IBL.  Need to
         get the FParam_1 for the cell model and we're done */

      %orva_freq_sum_model(
                           inData       = _fke_strat_model, 
                           distModelVar = Distribution, 
                           outData      = _fke_summed_model
                           );

      data _null_;
         set _fke_summed_model;
         call symputX("FParam_1_mod", FParam_1);
      run;
   %end;

   data _fke_single_cell_results; 
      length Distribution $12;
      length Freq_Fit_Interval $12;

      CCUID        = &CCUID;
      Distribution = upcase("&Freq_Distribution_Name");


      Source = "&sourceType";

      if Source = "T" then do; 
         Freq_Fit_Interval = "&FITINTERVALUSED";
      end;
      else do;
         Freq_Fit_Interval = "";
      end;

      Fparam_1 = &FParam_1_mod.;

      %do i = 2 %to &paramcount.;
         FParam_&i. = &&FParam_&i;
      %end;
   run;

%COMPLETE_Handle_Fitted_Models:

   %orvu_append(
                  base         = _fke_all_cells , 
                  new          = _fke_single_cell_results , 
                  canUseAppend = false
                  );

%DONE_Handle_Fitted_Models:

%mend orv_fittedModelsGetEffect;
