/*******************************************************************************
 * orva_freq_make_counts
 *
 * SUPPORT: stmajo
 *
 *
 * NOTE1: Keep in mind that each firm has its own set of bluids.  Therefore for
 *        a given bluid, own will have one value.
 *
 * NOTE2: When the word 'partition' is used in the comments, the strict 
 *        mathematical defninition is implied: a partition of a set A is a 
 *        collection of subsets of A whose pairwise intersections are empty and 
 *        whose union is A itself.
 *
 *
 * WHAT IT DOES: This macro creates a dataset with the loss counts for each cell 
 *               and time interval together with the corresponding KRI 
 *               information.  There are 5 parts to this:
 *
 *     1) Determine all of the time intervals that the counts will be broken into.  
 *        This is determined by the fit interval, the intervals for the firm KRIs, 
 *        and the intervals for the IBL KRIs. Let G be the set of dates defined by 
 *        the global range. Let P_I be the partition of G defined by the fit 
 *        interval.  Let P_B and P_F be the partitions defined by the intervals in 
 *        the business line KRI and firm KRI datasets respectively. What we want 
 *        is a partition of G, P_G such that if pg is in P_G, then pg is a subset 
 *        of exactly one element of P_I, P_B, and P_F. In other words, every 
 *        interval in P_I, P_B, and P_F can be expressed as the union of 1 or more 
 *        intervals in P_G.
 *
 *     2) Determine the list of classifications that will be used to bin 
 *        (slice-up, stratify) the data. These are:
 *          (a) CCVUID_0 ... CCVUID_N  where N is &NUMCCVS. - 1
 *          (b) INTERNAL_BL_UID
 *          (c) Any Classifications that have missing ranges defined
 *
 *      NOTE ON (2c): The reason these are included is so that we can properly 
 *                    adjust the interval lengths for missing ranges. Suppose 
 *                    there are 2 classifications, classA and classB.  The 
 *                    project is defined on classA.  Now suppose a missing 
 *                    range is defined for one item in classB for one time 
 *                    interval.  This range doesn't occupy any one cell, 
 *                    instead, it occupies a portion of all the cells.  The 
 *                    only way to know how much to adjust the interval by is to 
 *                    break the counts down by classA and classB.
 *
 *     3) Create all of the bins implied by the classifications selected in step 2.
 *
 *     4) Combine the time interval bins with the classification variable bins 
 *        (steps 1 and 3). Separate the counts into these combined bins.  
 *        Merge in the KRI data.
 *
 *     5) Adjust the interval lengths of the counts according to the missing 
 *        range table.
 *
 *
 * ARGUMENTS:
 *        lossdata           = the losses (e.g. VARPRJ.lossi)
 *
 *        dataType           =   I = internal, C=consortium, P=public
 *
 *        ForTheseCCUIDs     = list of ccuids to make the counts on.  use '_all_'
 *                             to make counts on all ccuids (dangerous - can get a
 *                             lot of empty cells), '_losses_' to only make counts on
 *                             cells with losses, or just past a space separated list.
 *
 *        cTool              = true or false, are we in the Consortium Tool?
 *
 *        MapDataPrefix      = the ccvuid mapping dataset library and dataset name 
 *                             beginning.  e.g. varprj.ccmapi_
 *
 *        kribl_ds           = dataset with IBL level KRIs with filters applied
 *
 *        krifirm_ds         = dataset with firm level KRIs with filters applied
 *                             EACH OF THESE DATASETS has every possible interval, 
 *                             from global start to BEYOND global end, and maybe 
 *                             some extra intervals in between, if the user 
 *                             recorded KRIs at a date other than a fit interval 
 *                             boundary.
 *
 *        interval           = user-selected fit interval
 *
 *        global_range       = the global range table
 *
 *        missing_range      = the MODIFIED missing range table.
 *                             This was the output dataset "outmissrange" from 
 *                             %orva_getfreqmomts macro.
 *
 *        CCVs               = List of the 'mapped from' classifications for the 
 *                             project.
 *
 *        numccvs            = number of cross-classification defined for the 
 *                             project.
 *
 *        mapping_uids       = The map UIDs for each map in the project, passed
 *                             as a space-separated list.
 *
 *        allibls            = INTERNAL_BUSINESSLINE dataset.  Has all of the 
 *                             internal business lines. There may be some for 
 *                             which KRIs are not recorded.
 *
 *        outClassifications = the classifications that were used to make the 
 *                             counts.
 *
 *        outCounts          = output Counts dataset.
 *
 *        outStatus          = name of the out status dataset.  Will contain 
 *                             the following: 
 *                                STATUS = 0 on success, 1 otherwise.
 *                                REASON = printable message
 *
 *        loss_mapping_date  = first_date_of_event. Eventually this will be 
 *                             variable. It is the date that gets used in the 
 *                             fitting.
 ******************************************************************************/
%macro orva_freq_make_counts(
                               lossdata           = ,
                               dataType           = ,
                               ForTheseCCUIDs     = _losses_,
                               cTool              = ,
                               MapDataPrefix      = ,
                               kribl_ds           = ,
                               krifirm_ds         = ,
                               fit_interval       = ,
                               global_range       = ,
                               missing_range      = varprj.missing_Range,
                               CCVs               = ,
                               numccvs            = ,
                               mapping_uids       = ,
                               allibls            = ,
                               outClassifications = ,
                               outCounts          = ,
                               outStatus          = ,
                               loss_mapping_date  = financialdate,
                               debug              = 
                            );
 
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put #orva_freq_make_counts( ;
      %put .                      lossdata           = &lossdata, ;
      %put .                      dataType           = &dataType, ;
      %put .                      ForTheseCCUIDs     = &ForTheseCCUIDs, ;
      %put .                      cTool              = &cTool, ;
      %put .                      MapDataPrefix      = &MapDataPrefix, ;
      %put .                      kribl_ds           = &kribl_ds, ;
      %put .                      krifirm_ds         = &krifirm_ds, ;
      %put .                      fit_interval       = &fit_interval, ;
      %put .                      global_range       = &global_range, ;
      %put .                      missing_range      = &missing_range, ;
      %put .                      CCVs               = &CCVs, ;
      %put .                      numccvs            = &numccvs, ;
      %put .                      mapping_uids       = &mapping_uids, ;
      %put .                      allibls            = &allibls, ;
      %put .                      outClassifications = &outClassifications, ;
      %put .                      outCounts          = &outCounts, ;
      %put .                      outStatus          = &outStatus, ;
      %put .                      loss_mapping_date  = &loss_mapping_date ;
      %put .                      ) ;
   %end;
 
   /******************************************************************************/
   /*                Macro variables that are local this macro                   */
   /******************************************************************************/
   %local
          i                               /* generic looping index */
          reason                          /* key for the message lookup table - why we aborted (if we did) or 'success' */
          status                          /* set to 0 when the end is reached with no error found */
          abort                           /* set to &orv_true if an error is found */
          firm_kris_present               /* set to &orv_true if there are firm kris recorded */
          firm_kri_list                   /* comma - separated list of KRIs with '_F' added */
          bl_kris_present                 /* set to &orv_true if there are bl kris recorded */
          bl_kri_list                     /* A comma separated list of KRIs */
          ibl_list                        /* space separated list of business lines */
          interval_date_formatting        /* A string with the date format for the interval_start and interval_end */
          missing_bls                     /* list of bls not found in the bl kri data */
          current_bls                     /* list of distinct bl's found in kribl dataset */
          nmiss                           /* number of business lines in &allibls. not found in the bl kri data */
          rename_statement                /* string that adds _F to end of names of the Firm KRIs */
          label_statement                 /* string that adds (F) to end of labels of the Firm KRIs */
          misssing_classifications        /* list of missing classifications */
          missing_classifications_count   /* number of classifications with missing range observations */
          all_classifications             /* list of all these classifications */
          all_classifications_count       /* number of total classifications the counts are broken down by */
          missing_class_name              /* used when looping through names of classifications with missing ranges */
          CCV_num                         /* The position in the list of CCVs corresponding to Internal_BL_UID */
          missing_range_classes           /* The list of all possible missingrange classifications */
          missing_range_classes_count     /* The number of possible missing range  classifications */
          missing_range_table_rows        /* The number of rows in the missing range table */
          extra_classes                   /* Classes in the missing range table that are not used to define the cells of the project */
          extra_classes_count             /* The number of extra classes */
          name                            /* Macro variable to hold the results of a %scan */
          classID_list                    /* List of classification uids from missing range table */
          obs_count                       /* number of observations in a dataset */
          model_on_IBL                    /* Will we enable modeling on IBL? */
          org_path                        /* Path to the losses/org folder */
          comp_path                       /* Path to this versions components folder */
          select_list                     /* A string like: r.amount,r.interval_start,r.interval_end ... */
          start_list                      /* List of interval_starts */
          end_list                        /* List of interval_ends */
          ccv_temp_list                   /* used to remove __totals_only__ from the list of ccvs */
          ;
 
   /* --------------------------------------------------------------- */
   /*     STEP 0: Error Checking and initialization                   */
   /* --------------------------------------------------------------- */
 
   /*- create the status data set up front -*/
   %let reason = %sysfunc(sasmsg(&orvads, fmc_ffting_abnormal_err, noquote));
   %let status = 1;
   %let abort  = &orv_false.;
   %let model_on_IBL = &orv_false.;
   %orva_status_ds( 
                   statusds = &outStatus.,
                   status   = &status.,
                   reason   = &reason. 
                   );
 
   /* The interval start and end formats need ot be consistent throughout.
                  Decide what to use here
      and then use this macro for formatting throughout.  Changes to format now only need to be done once. */
   %let interval_date_formatting = %str(interval_start date9. interval_end date9.);
 
   /* There will be a lot of checks against &ctool and dataType, make it a little cleaner by applying %locase() now */
   %let ctool    = %lowcase(&ctool.);
   %let dataType = %lowcase(&dataType.);
 
   /* Check that IBL is mapped to one of the CCVs.  If it is, set
      model_on_IBL to true, otherwise we skip IBL KRIs. */
   %let CCVs = %lowcase(&CCVs.);
   %let CCV_num = %sysfunc(findw(&CCVs., internal_bl_uid ));
   %if &CCV_num. > 0 %then %do;
      %let model_on_IBL = &orv_true.;
   %end;
 
   /* If one of the CCVs is __TOTALS__ONLY__, then we ignore it */
   %let CCV_num = %sysfunc(findw(&CCVs., __totals__only__ ));
   %if &CCV_num. > 0 %then %do;
      /* remove __TOTALS__ONLY__ from the list of CCVs */
      %let name = ;
      %let ccv_temp_list = ;
      %do i = 1 %to &numccvs.;
         %let name = %scan(&CCVS.,&i.);
         %if %lowcase(&name.) ne __totals__only__ %then %do;
            %let ccv_temp_list = &ccv_temp_list. &name.;
         %end;
      %end;
      %let numccvs = %eval(&numccvs. - 1);
      %let CCVs = &ccv_temp_list.;
      %let name = ;
   %end;

   /* Does the loss data exist? */
   %if &lossData. = %str() %then %do;
      %let abort  = &orv_true.;
      %let reason = %sysfunc(sasmsg(&orvads., fmc_lossData_missing_err, noquote));
      %goto DONE;
   %end;
 
   /* Does the global range data exist?*/
   %if &global_range. = %str() %then %do;
      %let abort  = &orv_true.;
      %let reason = %sysfunc(sasmsg(&orvads, fmc_specify_global_range_ds_err, noquote));
      %goto DONE;
   %end;
 
   /* Get the global start and global end date, make sure the former comes before the latter */
   data _null_;
      set &global_range.;
      call symputX( "range", end_date-start_date+1);
      call symputX( "global_end", end_date);
      call symputX( "global_start", start_date);
   run;
   %if %eval(&range.) LE 0 %then %do;
      %let abort  = &ORV_TRUE.;
      %let reason = %sysfunc(sasmsg(&orvads., fmc_start_not_before_end_err, noquote));
      %goto DONE;
   %end;
 
   /* Figure out where the cc and ccmap_* datasets are (varprj or varprj_m) */
   %let ccds = ;
   %orvu_find_dataset(
                      nameIn  = cc, 
                      Libs    = varprj varprj_m, 
                      nameOut = ccds
                      );

   /* -------------------------------------------------------------------- */
   /*     STEP 1: Create all of the time intervals that will be used       */
   /* -------------------------------------------------------------------- */
   %orv_CreateTimeIntervals( 
                            out_intervals   = _fmc_all_intervals,
                            out_KRIfirm_ds  = _fmc_firm_KRIs,
                            out_KRIbl_ds    = _fmc_bl_KRIs,
                            firmKRI_list    = firm_KRI_list,
                            blKRI_list      = bl_KRI_list,
                            cTool           = &cTool.,
                            global_start    = &global_start.,
                            global_end      = &global_end.,
                            fit_interval    = &fit_interval.,
                            KRIfirm_ds      = &KRIfirm_ds.,
                            KRIbl_ds        = &KRIbl_ds.,
                            allibls         = &allibls.,
                            interval_format = &interval_date_formatting.
                            );
 
   /* --------------------------------------------------------------- */
   /*     STEP 2: Determine the classifications that will be used     */
   /* --------------------------------------------------------------- */
   %orv_GetListofClassifications( 
                                 out_class_data       = &outClassifications.,
                                 out_class_list       = all_classifications,
                                 out_class_count      = all_classifications_count,
                                 out_miss_class       = missing_range_classes,
                                 out_miss_class_count = missing_range_classes_count,
                                 out_xtra_class       = extra_classes,
                                 out_xtra_class_count = extra_classes_count,
                                 numccvs              = &numccvs.,
                                 bl_kris_present      = &bl_kris_present.,
                                 firm_kris_present    = &firm_kris_present.,
                                 missing_range        = &missing_range.,
                                 cTool                = &cTool. 
                                 );

   /* --------------------------------------------------------------- */
   /*     STEP 3: Create the bins defined by classifications above    */
   /* --------------------------------------------------------------- */
   %orv_CreateCountBins( 
                        out_bins            = _fmc_countBins,
                        lossData            = &lossdata.,
                        ForTheseCCUIDs      = &ForTheseCCUIDs.,
                        all_classifications = &all_classifications. ,
                        cTool               = &cTool.,
                        missing_range       = &missing_range.,
                        xtra_class          = extra_classes,
                        xtra_class_count    = extra_classes_count,
                        bl_kris_present     = &bl_kris_present.,
                        allibls             = &allibls.,
                        CCVs                = &CCVs.
                        );


    /* ---------------------------------------------------------------------------------- */
   /*     STEP 4: merge the time interval and classification variable bins.  Add in the  */
   /*             counts and the KRIs                                                    */
   /* ---------------------------------------------------------------------------------- */
   %orv_MergeBinsCountsAndKRIs( 
                               out_counts                = &outCounts.,
                               interval_data             = _fmc_all_intervals,
                               bin_data                  = _fmc_countBins,
                               lossData                  = &lossdata.,
                               cTool                     = &cTool.,
                               all_classifications       = &all_classifications.,
                               all_classifications_count = &all_classifications_count.,
                               dataType                  = &dataType.,
                               bl_kris_present           = &bl_kris_present.,
                               firm_kris_present         = &firm_kris_present.,
                               loss_mapping_date         = &loss_mapping_date.,
                               firm_KRI_list             = %quote(&firm_KRI_list.),
                               firm_KRI_data             = _fmc_firm_KRIs,
                               bl_KRI_list               = %quote(&bl_KRI_list.),
                               bl_KRI_data               = _fmc_bl_KRIs 
                               );
   %if &abort. eq &orv_true. %then %do;
      %goto DONE;
   %end;

   /* -------------------------------------------------------------------- */
   /*     STEP 5: Adjust the interval lengths according to missing ranges  */
   /* -------------------------------------------------------------------- */
 
   /* If there were any missing ranges defined, adjust the interval_length appropriately */
   %if &missing_range_classes_count. > 0 %then %do;
      %orv_adjustForMissingRanges( 
                                  missing_range               = &missing_range.,
                                  missing_range_classes       = &missing_range_classes.,
                                  missing_range_classes_count = &missing_range_classes_count.,
                                  inCounts                    = &outCounts.,
                                  outCounts                   = &outCounts.,
                                  cTool                       = &cTool.,
                                  all_classifications         = &all_classifications.,
                                  all_classifications_count   = &all_classifications_count. 
                                  );
   %end;
 
   /* ----------------------------------------------------------------------------------------- */
   /*  STEP 6: A Little formatting of the output.  Add some varables that freq_fit will use.    */
   /* ----------------------------------------------------------------------------------------- */
   data &outCounts.;
      set &outCounts.;
 
      /* The interval is likely only a portion of the fit interval, so record the actual fit interval
         that it falls in.  */
      start_of_curr_fitinterval = intnx("&fit_interval.", interval_start, 0);
      end_of_curr_fitinterval   = intnx("&fit_interval.", interval_start, 1);
      fitinterval_length        = (end_of_curr_fitinterval - start_of_curr_fitinterval);
 
      /* If there were no missing ranges, then adj_length was not created */
      if missing(adj_length) then do;
         adj_length = interval_length;
      end;
 
      /* Compute exposure */
      exposure = adj_length / fitinterval_length;
 
      /* Map the CCVUID_#'s to the corresponding CCUID (makes life easier in backtesting) */
      ccuid = .;
 
      if _N_ eq 1 then do;
        declare hash ccuidLookup(dataset: "&ccds.");
        %do i = 0 %to %eval(&numccvs. - 1);
           rc = ccuidLookup.defineKey("ccvuid_&i.");
        %end;

        rc = ccuidLookup.defineData("ccuid");
        rc = ccuidLookup.defineDone();
      end;
 
      rc = ccuidLookup.find();
 
      /* Frequency fitting uses 'the_count'*/
      rename count = the_count;
      drop 
         rc
         interval_id
      ;
   run;
 
   /*---------------------------------------- DONE ---------------------------------------*/
   %let status = 0;
 
%DONE:
   %if &abort %then %do;
      %put &orvar_error. &reason.;
   %end;
 
   %orva_status_ds( 
                   statusds = &outStatus.,
                   status   = &status.,
                   reason   = &reason.
                   );

   data &outStatus.;
      set &outStatus.;
      model_on_IBL = ("&model_on_IBL." = "&orv_true.");
   run;

   %orvu_delete_data(data= _fmc_:);

%mend orva_freq_make_counts;
 
/*----------------------------------------------------------------------------------------------------------------*
 * Macro Name: orv_CreateTimeIntervals                                                                            *
 *                                                                                                                *
 * What it Does: Creates a dataset with all of the time intervals that the losses will be grouped into.           *
 *               In the process, it also cleans the BL and Firm KRI lists - add '_F' to firm KRI names,           *
 *               add anymissing BLUIDs in the BL KRI data, edit the intervals to fall within the global           *
 *               range.                                                                                           *
 *                                                                                                                *
 * Subordinate Macros:                                                                                            *
 *                                                                                                                *
 *     orv_alterFirmKRINames                                                                                      *
 *                                                                                                                *
 *     orv_addMissingIBLs                                                                                         *
 *                                                                                                                *
 *     orv_SortAndTrimKriDatasets                                                                                 *
 *                                                                                                                *
 *     orv_duplicateIntervalsByBl                                                                                 *
 *                                                                                                                *
 * Arguments:                                                                                                     *
 *                                                                                                                *
 *    out_intervals    - The dataset to create with the intervals.  2 or 3 rows: Interval_start, interval_end,    *
 *                       and possibly BLUID (if bl KRIs are present).                                             *
 *                                                                                                                *
 *    out_KRIfirm_ds   - Name of the Firm KRI dataset that will be produced (altered names and trimmed intervals  *
 *                                                                                                                *
 *    out_KRIbl_ds     - Name of the BL KRI dataset that will be produced (add missing BLs and trimmed intervals) *
 *                                                                                                                *
 *    firmKRI_list     - Name of macro variable to hold a comma-separated list of the firm KRIs                   *
 *                                                                                                                *
 *    blKRI_list       - Name of macro variable to hold a comma-separated list of the BL KRIs                     *
 *                                                                                                                *
 *    cTool            - True or false.  Whether or not we are in CTool                                           *
 *                                                                                                                *
 *    global_start     - The global start date for the project                                                    *
 *                                                                                                                *
 *    global_end       - The global end date for the project                                                      *
 *                                                                                                                *
 *    fit_interval     - The fit interval for the project                                                         *
 *                                                                                                                *
 *    KRIfirm_ds       - The dataset with Firm KRI values that are passed in (unaltered)                          *
 *                                                                                                                *
 *    KRIbl_ds         - The dataset with bl KRI values that are passed in (unaltered)                            *
 *                                                                                                                *
 *    allibls          - A datset with a list of all of the internal_bl_uids                                      *
 *                                                                                                                *
 *    interval_format - The date format for the interval_start and interval_end values.                           *
 *                      e.g. %str(interval_start date9. interval_end date9.)                                      *
 *                                                                                                                *
 *----------------------------------------------------------------------------------------------------------------*/
%macro orv_CreateTimeIntervals(
                               out_intervals   = ,
                               out_KRIfirm_ds  = ,
                               out_KRIbl_ds    = ,
                               firmKRI_list    = ,
                               blKRI_list      = ,
                               cTool           = ,
                               global_start    = ,
                               global_end      = ,
                               fit_interval    = ,
                               KRIfirm_ds      = ,
                               KRIbl_ds        = ,
                               allibls         = ,
                               interval_format =
                               );
 
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_CreateTimeIntervals( ;
      %put .                         out_intervals   = &out_intervals.,;
      %put .                         out_KRIfirm_ds  = &out_KRIfirm_ds.,;
      %put .                         out_KRIbl_ds    = &out_KRIbl_ds.,;
      %put .                         firmKRI_list    = &firmKRI_list.,;
      %put .                         blKRI_list      = &blKRI_list.,;
      %put .                         cTool           = &cTool.,;
      %put .                         global_start    = &global_start.,;
      %put .                         global_end      = &global_end.,;
      %put .                         fit_interval    = &fit_interval.,;
      %put .                         KRIfirm_ds      = &KRIfirm_ds.,;
      %put .                         KRIbl_ds        = &KRIbl_ds.,;
      %put .                         allibls         = &allibls.,;
      %put .                         interval_format = &interval_format.;
      %put .                         );
   %end;
 
   %let cTool = %lowcase(&cTool.);
   /* --------- Get the intervals from global range, firm KRI data, and BL KRI data ---------- */
 
    /* First create the intervals based on the fit interval, global start, and global end (the global range) */
   data _fmc_global_intervals;
      format &interval_format.;
      retain n 0 done 0;
 
      do while(done ne 1);
         interval_start = intnx("&fit_interval.",&global_start.,n,"beginning");

         if interval_start < &global_start. then do;
            interval_start = &global_start;
         end;

         interval_end =  intnx("&fit_interval.",interval_start,0,"end");

         if  interval_end >= &global_end then do;
            done = 1;
            interval_end = &global_end;
         end;

         interval_length = interval_end - interval_start + 1;
         output;
         n = n + 1;
      end;
 
      keep interval_:;
   run;
 
   /* Check for firm KRIs.  If they are found, get the firm KRI intervals,
      sort them, and adjust them so that they conform to the global range.
      Rename firm KRIs so that they have an '_F' at the end */
   /* Creates _fmc_firm_kris_sorted and _fmc_firm_intervals. */
   %let firm_kris_present  = &orv_false.;
   %let obs_count = 0;
   %if %upcase(&dataType.) eq I %then %do;
      %orvu_ds_exists( 
                      Dataset   = &krifirm_ds.,
                      OutNOBS   = obs_count,
                      OutStatus = _fmc_ds_exists
                      );
   %end;

   %if &obs_count. > 0 %then %do;
      /* The dataset exists, but this does not guarantee that there are KRIs in there, so check */
      proc contents
                   data = &krifirm_ds.
                   out = _fmc_temp
                   noprint
                   ;
      run;
 
      data _null_;
         set _fmc_temp end=last;
         retain KRI_found 0;

         if lowcase(substr(NAME,1,3)) eq "kri" then do;
            KRI_found = 1;
         end;

         if last and KRI_found = 1 then do;
            call symputX("firm_kris_present","&orv_true.");
         end;
      run;

      %if &firm_kris_present. eq &orv_true. %then %do;
          /* Dont want to work with the original data  - make a copy */
         data &out_KRIfirm_ds.;
            set &krifirm_ds.;
         run;
 
         /* Add '_F' to the end of the firm KRI names, to distinguish them from the bl KRIs */
         %orv_alterFirmKRINames(
                                in_KRIfirm_ds  = &out_KRIfirm_ds.,
                                out_KRIfirm_ds = &out_KRIfirm_ds.
                                );
 
         /* Sort by interval_start and adjust intervals to comply with global range and the fit interval */
         %orv_SortAndTrimKriDatasets(
                                     KRIdata_in      = &out_KRIfirm_ds. ,
                                     KRIdata_out     = &out_KRIfirm_ds.,
                                     KRI_list        = %quote(&firmKRI_list.),
                                     type            = firm,
                                     global_start    = &global_start.,
                                     global_end      = &global_end.,
                                     cTool           = &cTool.,
                                     interval_format = &interval_format. 
                                     );

         /* Grab just the intervals. */
         data _fmc_firm_intervals;
            set &out_KRIfirm_ds.(keep =
                                      interval_:
                                      %if (&cTool. EQ true) %then %do;
                                         own
                                      %end;
                                      );
         run;
      %end;
   %end;

   /* Do the same for the BL KRIs.  Also, if there are any bluids in the 
      &allibls. dataset that are not in the KRI dataset, add them in with 
      interval_start = global_start and interval_end = global_end and set all 
      KRI values to 0. */

   /* Creates _fmc_bl_kris_sorted and _fmc_bl_intervals. */
   %let obs_count       = 0;
   %let bl_kris_present = &orv_false.;

   %if %upcase(&dataType.) eq I and &model_on_IBL. eq &orv_true. %then %do;
      %orvu_ds_exists( 
                      Dataset   = &kribl_ds.,
                      OutNOBS   = obs_count,
                      OutStatus = _fmc_ds_exists
                      );
   %end;

   /* If we have IBL KRIs AND IBL is a 'mapped from' property, then we can 
      add in IBL KRI stuff */
   %if &obs_count. > 0  %then %do;
      /* The dataset exists, but this does not guarantee that there are KRIs in 
         there, so check */
      proc contents
                   data = &kribl_ds.
                   out = _fmc_temp
                   noprint
                   ;
      run;

      data _null_;
         set _fmc_temp end=last;
         retain KRI_found 0;

         if lowcase(substr(NAME,1,3)) eq "kri" then do;
            KRI_found = 1;
         end;

         if last and KRI_found = 1 then do;
            call symputX("bl_kris_present","&orv_true.");
         end;
      run;
 
      %if &bl_kris_present. eq &orv_true. %then %do;
         /* Dont want to work with the original data  - make a copy */
         data &out_KRIbl_ds.;
            set &kribl_ds.;
         run;
 
         %orv_addMissingIBLs(
                             in_KRIbl_ds     = &out_KRIbl_ds.,
                             allibls         = &allibls.,
                             out_KRIbl_ds    = &out_KRIbl_ds.,
                             global_start    = &global_start.,
                             global_end      = &global_end.,
                             interval_format = &interval_format.
                             );
 
         /* Sort by interval_start and bluid. Adjust intervals to comply with
           global range and the fit interval */
         %orv_SortAndTrimKriDatasets(
                                     KRIdata_in      = &out_KRIbl_ds.,
                                     KRIdata_out     = &out_KRIbl_ds.,
                                     KRI_list        = %quote(&blKRI_list.),
                                     type            = bl,
                                     global_start    = &global_start.,
                                     global_end      = &global_end.,
                                     cTool           = &cTool.,
                                     interval_format = &interval_format. 
                                     );
 
         /* Grab just the intervals. */
         data _fmc_bl_intervals;
            set &out_KRIbl_ds.(keep = interval_:);
         run;

       %end;
    %end;
 
   /* ------- Combine the intervals.  ------- */
   data _fmc_all_intervals;
      set
         _fmc_global_intervals
         %if &bl_kris_present. eq &orv_true. %then %do;
            _fmc_bl_intervals
         %end;
         %if &firm_kris_present. eq &ORV_TRUE. %then %do;
            _fmc_firm_intervals
         %end;
         ;
      keep 
         interval_start
         interval_end
         ;
   run;
 
   /* ------- Cut these intervals down to the smallest set we need ------- */
   /* sort by interval_start and interval_end; Only want each interval_start 
     and end to appear once */

   /* Need to run the sort twice.  This is because we want to get rid of all 
      duplicate interval_starts and all duplicate interval_ends, but don't 
      require that both values be repeated for an observation to get deleted. 
      So get all the distinct interval_starts (by bluid and own) and then all 
      the interval_ends, and then put them together. */

   /* This may seem a bit magical, but because the KRI dates and the global 
      range dates partition the interval from global_start to global_end, we 
      know that for every duplicate interval_start there will be a duplicate 
      interval_end and vice-versa.  Therefore, for a given bluid, the list of 
      distinct interval_starts and interval_ends will also partition the 
      interval and each interval in the global_range, firm KRI and bl KRI 
      datasets will be a combination of these intervals because its start date 
      and end date will be represented. */
 
   proc sort
            data = _fmc_all_intervals(keep = interval_start)
            out  = _fmc_all_interval_starts
            nodupkey
            ;
      by
         interval_start
         ;
   run;
 
   proc sort
            data  = _fmc_all_intervals(keep = interval_end)
            out   = _fmc_all_interval_ends
            nodupkey
            ;
      by
         interval_end;
         ;
   run;
 
   /* This is the final interval dataset */
   data &out_intervals.;
      set _fmc_all_interval_starts;
      set _fmc_all_interval_ends;

      interval_length = interval_end - interval_start + 1;
   run;
%mend orv_CreateTimeIntervals;
 
/* Add '_F' to the end of the firm KRI names. */
%macro orv_alterFirmKRINames(
                             in_KRIfirm_ds  = ,
                             out_KRIfirm_ds =
                             );
 
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_alterFirmKRINames( ;
      %put .                       in_KRIfirm_ds  = &in_KRIfirm_ds, ;
      %put .                       out_KRIfirm_ds = &out_KRIfirm_ds ;
      %put .                       ) ;
   %end;
 
   %let rename_statement = ;
   %let label_statement  = ;
   proc contents
      data =  &in_KRIfirm_ds.(keep=KRI_:)
      out  = _fmc_firm_kri_names(keep=NAME LABEL)
      noprint;
   run;
 
   proc sql noprint;
      SELECT
               strip(name) || " = " || strip(name) || "_F",
               strip(name) || "_F = " || strip(label) || "(F)"
      INTO
               :rename_statement separated by " ",
               :label_statement  separated by " "
      FROM
               _fmc_firm_kri_names
               ;
   quit;
 
   data &out_KRIfirm_ds.;
      set &in_KRIfirm_ds.(rename = ( &rename_statement. ));
      label &label_statement.;
   run;
%mend orv_alterFirmKRINames;
 
/* Add bluids not in the bl KRI data to the list, with KRI values for those bluids set to 0 */
%macro orv_addMissingIBLs(
                          in_KRIbl_ds     = ,
                          allibls         = ,
                          out_KRIbl_ds    = ,
                          global_start    = ,
                          global_end      = ,
                          interval_format =
                          );
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_addMissingIBLs( ;
      %put .                    in_KRIbl_ds     = &in_KRIbl_ds, ;
      %put .                    allibls         = &allibls, ;
      %put .                    out_KRIbl_ds    = &out_KRIbl_ds, ;
      %put .                    global_start    = &global_start, ;
      %put .                    global_end      = &global_end, ;
      %put .                    interval_format = &interval_format ;
      %put .                    ) ;
   %end;
 
   /* Check for bluids in &allibls that are not in the KRI data.  If found, add them in. */
   proc sql noprint;
      SELECT
             distinct bluid_kri
      INTO
             :current_bls SEPARATED BY " "
      FROM
             &in_KRIbl_ds.
             ;
   quit;
 
   proc sql noprint;
      SELECT
             distinct internal_bl_uid
      INTO
             :missing_bls SEPARATED BY " "
      FROM
             &allibls.
      WHERE
             internal_bl_uid NOT IN (&current_bls.)
             ;
   quit;
 
   %let nmiss = %sysfunc(countw(%str(&missing_bls.)));
 
   %if &nmiss. > 0 %then %do;
      data &out_KRIbl_ds.;
         format &interval_format.;
         set &in_KRIbl_ds. end=last;
         array kris kri:;
         output;

         if last then do;
            do i = 1 to &nmiss.;
               bluid_kri      = scan("&missing_bls.", i, ' ');
               interval_start = &global_start.;
               interval_end   = &global_end.;

               do j = 1 to dim(kris);
                  kris[j] = 0;
               end;

               output;
            end;
         end;

         drop i j;
      run;
   %end;
%mend orv_addMissingIBLs;
 
/* Sort the KRI dataset by interval_start (and BLUID if appropriate).  Make sure inervals conform to the global range */
/* type is either 'firm' or 'bl' */
/*
   Produces the dataset:
      _fmc_&type._kris_sorted   % The above dataset sorted by interval_start and possibly bluid
   Modifies:
      _fmc_&type._KRIs          % Sorted KRI list iwth the intervals adjusted according to the global range
 
   Assigns values to the macros:
      &type._KRIs_present       % set to &orv_true. if KRI data of the specified type is found
      &type._KRI_list           % Comma-separated list of the KRIs of the given type (with '_F' for firm)
*/
%macro orv_SortAndTrimKriDatasets(
                                  KRIdata_in      = ,
                                  KRIdata_out     = ,
                                  KRI_list        = ,
                                  type            = ,
                                  global_start    = ,
                                  global_end      = ,
                                  cTool           = ,
                                  interval_format =
                                  );
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_SortAndTrimKriDatasets( ;
      %put .                            KRIdata_in      = &KRIdata_in, ;
      %put .                            KRIdata_out     = &KRIdata_out, ;
      %put .                            KRI_list        = &KRI_list, ;
      %put .                            type            = &type, ;
      %put .                            global_start    = &global_start, ;
      %put .                            global_end      = &global_end, ;
      %put .                            cTool           = &cTool, ;
      %put .                            interval_format = &interval_format ;
      %put .                            ) ;
   %end;
 
   %let type = %lowcase(&type.);
 
   /* If there are KRIs, get the intervals from the KRI dataset */
   proc sort
            data = &KRIdata_in.
            out = _fmc_&type._kris_sorted
            ;
      by
         %if (&cTool EQ true) %then %do;
            own
         %end;
         %if %lowcase(&type.) eq bl %then %do;
            bluid_kri
         %end;
         interval_start
         ;
   run;
 
   /* Trim intervals so that they fall in the global range */
   data &KRIdata_out.;
      format &interval_format.;
      set _fmc_&type._kris_sorted;
 
      /* Remove intervals before the global start date. */
      if interval_end <= &global_start. then do;
         delete;
      end;

      /* Make global_start the start of the first interval */
      if interval_end >= &global_start. and interval_start <  &global_start. then do;
         interval_start = &global_start.;
      end;
 
      /* Remove intervals after the global end. */
      if interval_start > &global_end. then do;
         delete;
      end;
 
      /* Make global_start the start of the first interval */
      if interval_start <= &global_end. and interval_end > &global_end. then do;
         interval_end = &global_end.;
      end;
 
      interval_length = interval_end - interval_start + 1;

      %if (&cTool EQ true) %then %do;
         if missing(own) then do;
            own 0;
         end;
      %end;
   run;
 
   /* Finally, create the comma-separated list of KRI names */
   proc contents
      data =  &KRIdata_in.(keep=KRI_:)
      out  = _fmc_&type._kri_names(keep=NAME LABEL)
      noprint;
   run;
 
   proc sql noprint;
      SELECT
            name
      INTO
            :&KRI_list. separated by ","
      FROM
            _fmc_&type._kri_names
            ;
   quit;
%mend orv_SortAndTrimKriDatasets;
 
/*----------------------------------------------------------------------------------------------------------------*
 * Macro Name: orv_GetListofClassifications                                                                       *
 *                                                                                                                *
 * What it Does: Determines which classifications will be used when breakign the data up into bins.  These        *
 *               Are determined by the ccvuid's, the missing range table, and internal_bl_uid.  These are         *
 *               written to a datset as well as a macro variable (as a space-separated list).                     *
 *                                                                                                                *
 * Subordinate Macros:                                                                                            *
 *                                                                                                                *
 *    orv_getNumRec                                                                                               *
 *                                                                                                                *
 * Arguments:                                                                                                     *
 *                                                                                                                *
 *    out_class_data       - Dataset with a list of all the classifications to bin data by                        *
 *                                                                                                                *
 *    out_class_list       - Macro variable with list of all classifications used in bins                         *
 *                                                                                                                *
 *    out_class_count      - Macro variable with the count of the classifications used in bins                    *
 *                                                                                                                *
 *    out_miss_class       - Macro variable to store the list of classifications with aty least one               *
 *                           non-missing value in the missing range table                                         *
 *                                                                                                                *
 *    out_miss_class_count - Macro variable to store the number of classifications with aty least one             *
 *                           non-missing value in the missing range table                                         *
 *                                                                                                                *
 *    out_xtra_class       - Macro variable with the list of classifications in out_miss_class but not in         *
 *                           out_class_list.                                                                      *
 *                                                                                                                *
 *    out_xtra_class_count - Macro variable with the number of classifications in out_xtra_class                  *
 *                                                                                                                *
 *                                                                                                                *
 *    numccvs              - The number of CCVs the projrct is defined on (1,2 or 3)                              *
 *                                                                                                                *
 *    bl_kris_present      - Are there bl KRIs?  Set to orv_true or orv_false                                     *
 *                                                                                                                *
 *    firm_kris_present    - Are there firm KRIs?  Set to orv_true or orv_false                                   *
 *                                                                                                                *
 *    missing_range        - Dataset with all of the missing ranges                                               *
 *                                                                                                                *
 *    cTool                - Are we in cTool? true or false                                                       *
 *                                                                                                                *
 *----------------------------------------------------------------------------------------------------------------*/
%macro orv_GetListofClassifications(
                                    out_class_data       = ,
                                    out_class_list       = ,
                                    out_class_count      = ,
                                    out_miss_class       = ,
                                    out_miss_class_count = ,
                                    out_xtra_class       = ,
                                    out_xtra_class_count = ,
                                    numccvs              = ,
                                    bl_kris_present      = ,
                                    firm_kris_present    = ,
                                    missing_range        = ,
                                    cTool                =
                                    );
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_GetListofClassifications( ;
      %put .                              out_class_data       = &out_class_data, ;
      %put .                              out_class_list       = &out_class_list, ;
      %put .                              out_class_count      = &out_class_count, ;
      %put .                              out_miss_class       = &out_miss_class, ;
      %put .                              out_miss_class_count = &out_miss_class_count, ;
      %put .                              out_xtra_class       = &out_xtra_class, ;
      %put .                              out_xtra_class_count = &out_xtra_class_count, ;
      %put .                              numccvs              = &numccvs, ;
      %put .                              bl_kris_present      = &bl_kris_present, ;
      %put .                              firm_kris_present    = &firm_kris_present, ;
      %put .                              missing_range        = &missing_range, ;
      %put .                              cTool                = &cTool ;
      %put .                              ) ;
   %end;
 
   %let cTool = %lowcase(&cTool.);
 
   /* Right now this is fairly straightforward.  Classifications are:
       (1) ccvuid_0 - ccvuid_(numccvs - 1)
       (2) internal_bl_uid if IBL is mapped to one the CCVs
       (3) those for which a missing range is defined
   */
   /* NOTe: If we add classX because there is a missing range for classX, then there will be a redundency
      if classX is mapped to something used to define one of the ccvuids.  This isn't a problem. In fact,
      if we don't include the ccvuid that classX is mapped to when making the bins, we will need to add it
      in later and the result is the same. */
 
   /* -------- (1) The CCVUID_'s ------------- */
   %let &out_class_list.  = ;
   %let &out_class_count. = &numccvs.;
   %do i = 0 %to %eval(&numccvs. - 1);
      %let &out_class_list. = &&&out_class_list. ccvuid_&i.;
   %end;
 
   /* -------- (2) Internal Business Line ------------ */
   %if &model_on_IBL. eq &orv_true. %then %do;
      %let &out_class_list = &&&out_class_list. internal_bl_uid;
      %let &out_class_count = %eval(&&&out_class_count. + 1);
   %end;
 
   /* -------- (3) Classifications with missing ranges ------------ */
   %let &out_miss_class_count. = 0;
   %let &out_miss_class.       = ;
 
   %let &out_xtra_class. = ;
   %let &out_xtra_class_count. = 0;
 
   %let missing_range_table_rows = 0;
   %orvu_ds_exists( 
                   Dataset   = &missing_range.,
                   OutNOBS   = missing_range_table_rows,
                   OutStatus = _fmc_ds_exists
                   );
 
   %if &missing_range_table_rows. > 0 %then %do;
      /* First need to go through and get a list of the classifications with missing ranges.  This will
         be used again later when adjusting the count intervals for missing ranges. */
      %let &out_miss_class. =  
                              internal_bl_uid 
                              internal_rc_uid 
                              legal_entity_uid 
                              process_uid
                              class_1 
                              class_2 
                              class_3 
                              class_4 
                              class_5
                              class_6 
                              class_7 
                              class_8 
                              class_9 
                              class_10
                              ;
 
      /* If we're in consortium tool, 'own' is also a possibility */
      %if &cTool. eq &orv_true. %then %do;
         %let &out_miss_class. = &out_miss_class. own;
      %end;
 
      /* From the list of possible classificaitons with missing ranges, select the ones for which a missing
         range is actually defined */
      %let &out_miss_class_count. = %sysfunc(countw("&&&out_miss_class."," "));
 
      /* Right now the list is all possible classifications.  Go through the dataset and only select the classifications
         that have at least one non-missing variable.  */
      data _null_;
         length missing_range_classes $ 1000;
         set &missing_range.(keep = &&&out_miss_class.) end = last;
 
         /* This array will index all the classification variables */
         array classifications _NUMERIC_;
 
         /* c[i] will indicate (1 = true) if a non-missing value exists for classification[i] */
         array c c1 - c&&&out_miss_class_count.;
         retain c;
 
         do i=1 to &&&out_miss_class_count.;
            if _N_ = 1 then do;
               c[i] = 0;
            end;
            if not missing(classifications[i]) then do;
               c[i] = 1;
            end;
         end;
 
         if last then do;
            count = 0;

            do i = 1 to &&&out_miss_class_count.;
               if c[i] = 1 then do;
                  missing_range_classes = strip(missing_range_classes) || " " || strip(vname(classifications[i]));
                  count = count + 1;
               end;
            end;

            call symputX("&out_miss_class_count.",count);
            call symputX("&out_miss_class.", strip(missing_range_classes));
         end;
      run;
 
      %let &out_miss_class.  = %lowcase(&&&out_miss_class.);
 
      /* Go through the missing range classifications.  If they are not already in the list, add them. */
      %do i = 1 %to &&&out_miss_class_count.;
         %let missing_class_name = %scan(&&&out_miss_class.,&i.," ");

         %if %sysfunc(findw(&&&out_class_list., &missing_class_name. )) = 0 %then %do;
            %let &out_class_list.       = &&&out_class_list. &missing_class_name.;
            %let &out_class_count.      = %eval(&&&out_class_count. + 1);
            %let &out_xtra_class.       = &&&out_xtra_class. &missing_class_name.;
            %let &out_xtra_class_count. = %eval(&&&out_xtra_class_count. + 1);
         %end;
      %end;
   %end; /*missing_range_table_rows > 0 */

   /* Write the classifications out to a dataset */
   data &out_class_data.;
      length ClassificationVariable $ 100;
 
      %do i = 1 %to &&&out_class_count.;
         ClassificationVariable = "%scan(&&&out_class_list.,&i.," ")";
         output;
      %end;
   run;
%mend orv_GetListofClassifications;
 
/*----------------------------------------------------------------------------------------------------------------*
 * Macro Name: orv_CreateCountBins                                                                                *
 *                                                                                                                *
 * What it Does: Based on a list of classifations, creates a list of all combinations of classes. Possible nodes  *
 *               for each classification are those that appear in the loss data. Internal_bl_uids that do not     *
 *               appear in the loss data are added in also.                                                       *
 *                                                                                                                *                                                                                        *
 *                                                                                                                *
 * Arguments:                                                                                                     *
 *                                                                                                                *
 *    out_bins            - The dataset to produce wityh all of the bins                                          *
 *                                                                                                                *
 *    lossData            - Dataset with all the losses                                                           *
 *                                                                                                                *
 *    ForTheseCCUIDs      - See description in freq_make_counts_header                                            *
 *                                                                                                                *
 *    all_classifications - A list of all the classififaction variables to use when making bins                   *
 *                                                                                                                *
 *    cTool               - Are we in cTool? true / false.                                                        *
 *                                                                                                                *
 *    missing_range       - The missing range table                                                               *
 *                                                                                                                *
 *    xtra_class          - List of classifications in the missing range table that are not used to define        *
 *                          the project.                                                                          *
 *                                                                                                                *
 *    xtra_class_count    - The number of extra classes.                                                          *                                       *
 *                                                                                                                *
 *    allibls             - Dataset with a list of all the internal_bl_uid nodes                                  *
 *                                                                                                                *
 *    bl_kris_present     - are there any IBL KRIs to consider?                                                   *
 *                                                                                                                *
 *    CCVs                - List of CCVs that are mapped from when defining the ccvuid's (each ccvuid is          *
 *                          the result of a mapping from one CCV to another).                                     *
 *                                                                                                                * 
 *----------------------------------------------------------------------------------------------------------------*/
%macro orv_CreateCountBins(
                           out_bins            = ,
                           lossData            = ,
                           ForTheseCCUIDs      = ,
                           all_classifications = ,
                           cTool               = ,
                           missing_range       = ,
                           xtra_class          = ,
                           xtra_class_count    = ,
                           allibls             = ,
                           bl_kris_present     = ,
                           CCVs                = 
                           );
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_CreateCountBins( ;
      %put .                     out_bins            = &out_bins, ;
      %put .                     lossData            = &lossData, ;
      %put .                     all_classifications = &all_classifications, ;
      %put .                     cTool               = &cTool, ;
      %put .                     missing_range       = &missing_range, ;
      %put .                     xtra_class          = &xtra_class, ;
      %put .                     xtra_class_count    = &xtra_class_count, ;
      %put .                     allibls             = &allibls, ;
      %put .                     bl_kris_present     = &bl_kris_present, ;
      %put .                     CCVs                = &CCVs ;
      %put .                     ) ;
   %end;
 
   %let cTool = %lowcase(&cTool.);
 
   /* Make a bin for each of the classification combinations for which there are counts */
 
   /* Make a temp dataset for input to &orva_designmatrix:  keep only the classifications in &all_classifications. in lossdata */
   data _fmc_designinput;
     set &lossData.(keep = 
                           &all_classifications.
                           %if &ctool. eq true %then %do;
                              own
                           %end;
                           );
      %if &ctool. eq true %then %do;
         if own = . then do;
            own = 0;
         end;
      %end;
   run;
 
   %if %lowcase(&dataType.) eq i %then %do;
      %let missing_range_table_rows = 0;
      %orvu_ds_exists( 
                      Dataset   = &missing_range.,
                      OutNOBS   = missing_range_table_rows,
                      OutStatus = _fmc_ds_exists
                      );
    
      %if &missing_range_table_rows. > 0 %then %do; 
         data _fmc_designinput_miss;
           set &missing_range.(keep = 
                                 &all_classifications.
                                 );
         run;

         data _fmc_designinput;
            set _fmc_designinput
                _fmc_designinput_miss
                ;
         run;
      %end;
   %end;

    /* SPARSE = false only creates those combos of variable levels that actually 
    occur in "designinput". Otherwise we would have waaaaay too many zero counts.  
    And we might be estimating for cells that don't exist in the customer's organization. */
    %orva_designmatrix(
                       A             = _fmc_designinput,
                       design        = _fmc_designoutput,
                       sparse        = false,
                       combosonly    = true,
                       drop          = false,
                       status_design = _fmc_design_status
                       );
   data _null_;
      set _fmc_design_status;

      if status ne 0 then do;
         call symputX("abort","&orv_true.");
         call symputX("reason",reason);
      end;
   run;
 
   /* Add in any zero count cell slices if IBL is mapped to a CCV. */
   /* First get a list of all the cells / cell slices.  */
   /* Need path to components and org folders */
   data _NULL_;
      org_path = pathname(scan("&allibls.",1,"."));
      /* org_path is "C:\ .... \loss\org" */
      comp_path = substr(org_path,1,length(strip(org_path)) - 8) || "components";
      call symputX("org_path",org_path);
      call symputX("comp_path",comp_path);
   run;
 
   libname orglib  "&org_path.";
   libname complib "&comp_path.";

   %orva_create_matrix(
                       cc_data              = &ccds.,
                       numCCVs              = &numccvs.,
                       mapping_uids         = &mapping_uids.,
                       datatype             = &datatype.,
                       split_cells_by_ibl   = &model_on_IBL.,
                       org_lib              = orglib,
                       comp_lib             = complib,
                       outdata              = _fmc_matrix
                       );

   /* Now merge in the bins from design_matrix */
   /* Get a list of all the variables in _fmc_designoutput other than ccvui_#
      and internal_bluid.  The 'matrix' dataset has only these variables, so 
      we need to know which variables from the designoutput to add in. */
   %let select_list = ;

   %let dsid = %sysfunc(open(_fmc_designoutput,i));
   %let varlist = ;
   %let varcount = %sysfunc(attrn(&dsid.,nvars));

   %do i = 1 %to &varcount.;
      %let varlist = &varlist. %sysfunc(varname(&dsid,&i));
   %end;

   %let rc = %sysfunc(close(&dsid.));

   data _fmc_varnames;
      length name $ 100;

      %do i = 1 %to &varcount.;
         name = "%scan(&varlist.,&i.)";
         output;
      %end;
   run;
 
   /* These are the variables not in the _fmc_matrix */
   data _fmc_varnames;
      set _fmc_varnames;
 
      where strip(lowcase(name)) not in (
                                         "ccvuid_0"
                                         "ccvuid_1"
                                         "ccvuid_2"
                                         %if &model_on_IBL. eq &orv_true. %then %do;
                                            "internal_bl_uid"
                                         %end;
                                         );
   run;

   /* Does _fmc_varnames have anything in it? */
   %let N_extra_vars = 0;
   %let dsid = %sysfunc(open(_fmc_varnames, I));

   %if &dsid. ne 0 %then %do;
      %let N_extra_vars = %sysfunc(attrn(&dsid., nlobs));
   %end;

   %let rc = %sysfunc(close(&dsid.));
 
   /* If it does, make a list */
   %if &N_extra_vars. > 0 %then %do;
      proc sql noprint;
         SELECT
               strip(name)
         INTO
               :select_list SEPARATED BY ", "
         FROM
               _fmc_varnames
               ;
      quit;
   %end;

   /* Which cells do we actually want counts for?  Adjust the matrix
      accordingly */
   %if %lowcase(%substr(&ForTheseCCUIDs.,1,1)) ne _ %then %do;
      /* ForTheseCCUIDs is not _all_ or _losses_, 
         so only select cells from the specified list */
      data _fmc_matrix_trimmed;
         set _fmc_matrix;
         where ccuid in (&ForTheseCCUIDs.);
      run;

   %end;
   %else %if %lowcase(&ForTheseCCUIDs.) eq _losses_ %then %do;
      /* Trim out any rows in the matrix that correspond with cells that have no losses */
      proc sql noprint;
         CREATE TABLE
                     _fmc_cells_with_losses
         AS SELECT distinct
                     %do i = 1 %to %eval(&numccvs. - 1);
                        ccvuid_&i.,
                     %end;
                     ccvuid_0
         FROM
                     _fmc_designoutput 
         ;
      quit;

      proc sql noprint;
         CREATE TABLE
                     _fmc_matrix_trimmed
         AS SELECT
                     %if &model_on_IBL. eq &orv_true. %then %do;
                        l.internal_bl_uid,
                     %end;
                     %do i = 1 %to %eval(&numccvs. - 1);
                        l.ccvuid_&i.,
                     %end;
                     l.ccvuid_0

         FROM
                     _fmc_matrix as l
         INNER JOIN
                     _fmc_cells_with_losses as r
         ON
                     l.ccvuid_0 eq r.ccvuid_0
                     %do i = 1 %to %eval(&numccvs. - 1);
                        AND l.ccvuid_&i. eq r.ccvuid_&i.
                     %end;
                     ;
      quit;
 
   %end;
   %else %if %lowcase(&ForTheseCCUIDs.) eq _all_ %then %do;
      data _fmc_matrix_trimmed;
         set _fmc_matrix;
      run;
   %end;

   /* Use the matrix and designmatrix output to get a list of all the bins */
   proc sql noprint;
      CREATE TABLE
                  &out_bins.
      AS SELECT
                  %if &model_on_IBL. eq &orv_true. %then %do;
                     l.internal_bl_uid,
                  %end;
                  %do i = 1 %to %eval(&numccvs. - 1);
                     l.ccvuid_&i.,
                  %end;
                  l.ccvuid_0
                  %if &N_extra_vars. > 0 %then %do;
                     , &select_list.
                  %end;
      FROM
                  _fmc_matrix_trimmed as l
      LEFT JOIN
                  _fmc_designoutput as r
             ON
                  l.ccvuid_0 eq r.ccvuid_0
                  %do i = 1 %to %eval(&numccvs. - 1);
                     AND l.ccvuid_&i. eq r.ccvuid_&i.
                  %end;
                  %if &model_on_IBL. eq &orv_true. %then %do;
                     AND l.internal_bl_uid eq r.internal_bl_uid
                  %end;
                  ;
   quit;

   %if &abort = &orv_true. %then %do;
      %return;
   %end;
%mend orv_CreateCountBins;
 
/*----------------------------------------------------------------------------------------------------------------*
 * Macro Name: orv_MergeBinsCountsAndKRIs                                                                         *
 *                                                                                                                *
 * What it Does: Takes the time intervals and the bins defined by the classifications and uses them to put the    *
 *               counts into bins.  The KRI data is then added in also.                                           *
 *                                                                                                                *
 * Subordinate Macros:                                                                                            *
 *                                                                                                                *
 *    _orv_mergeCountsAndKRIs                                                                                     *
 *                                                                                                                *
 * Arguments:                                                                                                     *
 *                                                                                                                *
 *    out_counts                - The dataset to produce with the counts for each bin                             *
 *                                                                                                                *
 *    interval_data             - Dataset with all of the time intervals that counts are divided into             *
 *                                                                                                                *
 *    bin_data                  - Dataset with all of the bins defined by classification variables                *
 *                                                                                                                *
 *    cTool                     - Are we in cTool? true / false.                                                  *
 *                                                                                                                *
 *    lossData                  - Dataset will all of the losses                                                  *
 *                                                                                                                *
 *    all_classifications       - List of classifications used to make the bin_data                               *
 *                                                                                                                *
 *    all_classifications_count - The number of classifications used to make bin_data                             *
 *                                                                                                                *
 *    bl_kris_present           - Are there any BL KRIs present? orv_true or orv_false                            *
 *                                                                                                                *
 *    firm_kris_present         - Are there any firm KRIs present? orv_true or orv_false                          *
 *                                                                                                                *
 *    loss_mapping_date         - Date to use when mapping losses to bins                                         *
 *                                                                                                                *
 *    firm_KRI_list             - List (seaparated by comma) of all the firm KRIs                                 *
 *                                                                                                                *
 *    firm_KRI_data             - Dataset with all the firm KRI values                                            *
 *                                                                                                                *
 *    bl_KRI_list               - List (separated by comma) of all the bl KRIs                                    *
 *                                                                                                                *
 *    bl_KRI_data               - Dataset with all the bl KRI values                                              *
 *                                                                                                                *
 *----------------------------------------------------------------------------------------------------------------*/
%macro orv_MergeBinsCountsAndKRIs(
                                   out_counts                = ,
                                   interval_data             = ,
                                   bin_data                  = ,
                                   lossData                  = ,
                                   cTool                     = ,
                                   all_classifications       = ,
                                   all_classifications_count = ,
                                   dataType                  = ,
                                   bl_kris_present           = ,
                                   firm_kris_present         = ,
                                   loss_mapping_date         = ,
                                   firm_KRI_list             = ,
                                   firm_KRI_data             = ,
                                   bl_KRI_list               = ,
                                   bl_KRI_data               =
                                   );
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_MergeBinsCountsAndKRIs( ;
      %put .                            out_counts                = &out_counts, ;
      %put .                            interval_data             = &interval_data, ;
      %put .                            bin_data                  = &bin_data, ;
      %put .                            lossData                  = &lossData, ;
      %put .                            cTool                     = &cTool, ;
      %put .                            all_classifications       = &all_classifications, ;
      %put .                            all_classifications_count = &all_classifications_count, ;
      %put .                            dataType                  = &dataType, ;
      %put .                            bl_kris_present           = &bl_kris_present, ;
      %put .                            firm_kris_present         = &firm_kris_present, ;
      %put .                            loss_mapping_date         = &loss_mapping_date, ;
      %put .                            firm_KRI_list             = &firm_KRI_list, ;
      %put .                            firm_KRI_data             = &firm_KRI_data, ;
      %put .                            bl_KRI_list               = &bl_KRI_list, ;
      %put .                            bl_KRI_data               = &bl_KRI_data ;
      %put .                            ) ;
   %end;
 
   %let cTool = %lowcase(&cTool.);
 
   /* Index the interval dataset for faster merging */
   %let interval_count = 0;
   data &interval_data.;
      set &interval_data. end=last;
      interval_id = _N_;
      if last then do;
         call symputX("interval_count",_N_);
      end;
   run;

   /* Now transpose the interval dataset (just the end dates) - this will
      be used later to quickly add all of the interval_end dates as variables
      in the loss data (temporarily) to make finding the interval associated with 
      each loss quicker */
   data _fmc_ends;
      set &interval_data.;
      keep interval_end;
   run;

   proc transpose
      data = _fmc_ends
      out  = _fmc_trans_ends
      ;
   run;

   /* ----------------- Merge this with the interval dataset.   ------------- */
   /* Will get a row for every interval / classification1 / ... /classificationN combination  */
   proc sql noprint;
      CREATE TABLE
                        _fmc_all_count_bins
      AS SELECT DISTINCT
                        /* We select each of the classifications */
                        %do i = 1 %to &all_classifications_count.;
                           %let name = %scan(&all_classifications., &i., " ");
                            l.&name.,
                        %end;
                        /* and ctool if we are in consortium tool */
                        %if (&cTool eq true) %then %do;
                           l.own,
                        %end;

                        /* and the time intervals. */
                        r.interval_start,
                        r.interval_end,
                        r.interval_length,
                        r.interval_id
      FROM
                        &bin_data. as l
      LEFT JOIN
                        &interval_data. as r
 
      ON
                        1 = 1

      ORDER BY
                        %if (&cTool EQ true) %then %do;
                           l.own,
                        %end;
                        %do i = 1 %to &all_classifications_count.;
                           %let name = %scan(&all_classifications., &i., " ");
                            l.&name.,
                        %end;
                        r.interval_id
                        ;
   quit;

   /* Go through the losses and identify which time interval each loss occurred in.  This makes
      the count creation much faster, since we can select based on equality now (using interval_id)
      rather than using a >= and <= check */
   proc sort 
      data = &lossdata.( keep = 
                               &loss_mapping_date. 
                               threshold 
                               &all_classifications.
                               )
      out = _fmc_tmp_loss
      ;
      by &loss_mapping_date.;
   run;

   data _fmc_tmp_loss;
      set _fmc_tmp_loss;
      retain 
         current_interval 1
         col1 - col&interval_count.
         abort 0
         ;
      array ends col1 - col&interval_count.;

      if _N_ = 0 then do;
         set &interval_data.;
      end;

      /* Populate the ens[] array with the interval end dates.  Because
         of the retain statement, these values will then be available in
         each row.  Doing it this way is MUCH faster than repeatedly 
         setting the interval end dataset */
      if _N_ = 1 then do;
         _NAME_ = "interval_end";
         declare hash endHash(dataset: "_fmc_trans_ends");
         rc = endHash.defineKey("_NAME_");
         %do i = 1 %to &interval_count.;
            rc = endHash.defineData("col&i.");
         %end;
         rc = endHash.defineDone();
         rc = endHash.find();
      end;

      if &loss_mapping_date. <= ends[current_interval] then do;
         interval_id = current_interval;
      end;
      else do;
         do while (&loss_mapping_date. > ends[current_interval] and abort ne 1 );
            current_interval = current_interval + 1;
            if current_interval > dim(ends) then do;
               abort = 1;
               /* Avoid array index out of bounds */
               current_interval = dim(ends);
               call symputX("abort","&orv_true.");
            end;
         end;
         interval_id = current_interval;
      end;

      drop
         _NAME_
         %do i = 1 %to &interval_count.;
            col&i.
         %end;
         rc
         current_interval
         abort
         ;
   run;

   %if &abort. eq &orv_true. %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, fmc_loss_outside_global_range, noquote));
      %goto DONE_MBCK;
   %end;

   /* Combine the time intervals and the bins to get a row for each
      bucket that the losses will be placed in */
   proc sql noprint;
      create table
                           _fmc_bins_all
      as select
                           l.*,
                           r.*
      from
                           &bin_data. as l
      left join
                           &interval_data. as r
      on
                           1=1
      ;
   quit;

   /* Index the bins for faster grouping of the counts */
   data _fmc_bins_all; 
      set _fmc_bins_all; 

      %do i=1 %to &all_classifications_count.;
         %let name = %scan(&all_classifications., &i., " ");
         if missing(&name.) then &name. = -1; 
      %end;

      bin = _n_;
   run;

   /* Merge in the losses to get counts for each bin */
   proc sql noprint;
      CREATE TABLE 
                           &outcounts. 
      AS SELECT DISTINCT
                           l.*, 
                           r.threshold,
                           sum((r.interval_id > 0)) as the_count 
      FROM 
                   
                  _fmc_bins_all as l

      LEFT JOIN

                  _fmc_tmp_loss as r
                  
      ON 
                  /* If name is missing, then this is a total cell for that CCV. */
                  %do i=1 %to &all_classifications_count.;
                     %let name = %scan(&all_classifications., &i., " ");
                  ( l.&name. = -1 OR l.&name. = r.&name.) 
                  AND
                  %end;
                  l.interval_id = r.interval_id
      GROUP BY 
                  l.bin,
                  r.threshold
      ;
   quit;

   /* ----------------- Add in the KRI data.  ------------- */
   %if &firm_kris_present. = &orv_true. %then %do;
      %_orv_mergeCountsAndKRIs(  
                               inCounts                  = &out_counts.,
                               outCounts                 = &out_counts.,
                               type                      = firm,
                               KRI_list                  = %quote(&firm_KRI_list.),
                               KRI_data                  = &firm_KRI_data.,
                               cTool                     = &cTool.,
                               all_classifications       = &all_classifications.,
                               all_classifications_count = &all_classifications_count. 
                               );
   %end;
   %if &bl_kris_present. = &orv_true. %then %do;
      %_orv_mergeCountsAndKRIs(  
                               inCounts                  = &out_counts.,
                               outCounts                 = &out_counts.,
                               type                      = bl,
                               KRI_list                  = %quote(&bl_KRI_list.),
                               KRI_data                  = &bl_KRI_data.,
                               cTool                     = &cTool.,
                               all_classifications       = &all_classifications.,
                               all_classifications_count = &all_classifications_count. 
                               );
   %end;

   %DONE_MBCK:

%mend orv_MergeBinsCountsAndKRIs;
 
/* Merge KRI data with the count data by matching intervals and possibly bluids */
%macro _orv_mergeCountsAndKRIs(
                               inCounts                  = ,
                               outCounts                 = ,
                               type                      = ,
                               KRI_list                  = ,
                               KRI_data                  = ,
                               cTool                     = ,
                               all_classifications       = ,
                               all_classifications_count =
                               );
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @_orv_mergeCountsAndKRIs( ;
      %put .                         inCounts                  = &inCounts, ;
      %put .                         outCounts                 = &outCounts, ;
      %put .                         type                      = &type, ;
      %put .                         KRI_list                  = &KRI_list, ;
      %put .                         KRI_data                  = &KRI_data, ;
      %put .                         cTool                     = &cTool, ;
      %put .                         all_classifications       = &all_classifications, ;
      %put .                         all_classifications_count = &all_classifications_count ;
      %put .                         ) ;
   %end;
 
   proc sql noprint;
      CREATE TABLE
                        &outCounts.
      AS SELECT DISTINCT
                        &KRI_list. ,
                        l.*
      FROM
                       &inCounts. as l
      LEFT JOIN
                  &KRI_data. as r
             ON
                  l.interval_start <= r.interval_start
                  AND
                  l.interval_end   >= r.interval_end
                  %if &type. = bl %then %do;
                     AND l.internal_bl_uid = r.bluid_kri
                  %end;
                  %if &cTool. = &orv_true. %then %do;
                     AND l.own = r.own
                  %end;
      GROUP BY
                  %do i=1 %to &all_classifications_count.;
                     %let name = %scan(&all_classifications., &i., " ");
                     l.&name.,
                  %end;
                  %if (&cTool eq true) %then %do;
                     l.own,
                  %end;
                  l.interval_start,
                  l.interval_end,
                  l.threshold
                  ;
   quit;
%mend _orv_mergeCountsAndKRIs;
 
/*----------------------------------------------------------------------------------------------------------------*
 * Macro Name: orv_adjustIntervalForMissingRanges                                                                 *
 *                                                                                                                *
 * What it Does: Creates a variable 'adj_length' in the count data which is the interval length adjusted by the   *
 *               length of any missing ranges relevant to that interval.                                          *
 *                                                                                                                *
 * Subordinate Macros:  NONE                                                                                      *
 *                                                                                                                *
 * Arguments:                                                                                                     *
 *                                                                                                                *
 *    missing_range               - The missing range table                                                       *
 *                                                                                                                *
 *    missing_range_classes       - The classifications that have missing ranges defined                          *
 *                                                                                                                *
 *    missing_range_classes_count - The number of classes that have missing ranges defined                        *
 *                                                                                                                *
 *    inCounts                    - The count data passed in - interval_length must be present                    *
 *                                                                                                                *
 *    outCounts                   - The dataset to produce with the counts for each bin                           *
 *                                                                                                                *
 *    cTool                       - Are we in cTool? true / false.                                                *
 *                                                                                                                *
 *    all_classifications         - The list of classifications used in making the bins                           *
 *                                                                                                                *
 *    all_classifications_count   - The number of classifications used to make bin_data                           *
 *                                                                                                                *
 *----------------------------------------------------------------------------------------------------------------*/
%macro orv_adjustForMissingRanges(
                                   missing_range               = ,
                                   missing_range_classes       = ,
                                   missing_range_classes_count = ,
                                   inCounts                    = ,
                                   outCounts                   = ,
                                   cTool                       = ,
                                   all_classifications         = ,
                                   all_classifications_count   =
                                   );
    %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put @orv_adjustForMissingRanges( ;
      %put .                            missing_range               = &missing_range, ;
      %put .                            missing_range_classes       = &missing_range_classes, ;
      %put .                            missing_range_classes_count = &missing_range_classes_count, ;
      %put .                            inCounts                    = &inCounts, ;
      %put .                            outCounts                   = &outCounts, ;
      %put .                            cTool                       = &cTool, ;
      %put .                            all_classifications         = &all_classifications, ;
      %put .                            all_classifications_count   = &all_classifications_count ;
      %put .                            ) ;
   %end;
 
   %let cTool = %lowcase(&cTool.);
 
   data _fmc_missing_range_data;
      set &missing_range.(keep = 
                                 &missing_range_classes.
                                 start_missing_date
                                 end_missing_date 
                                 );
   run;
 
   proc sql noprint;
      CREATE TABLE
                        &outCounts.
      AS SELECT DISTINCT
                        l.*,
                        (min(l.interval_end,r.end_missing_date)  - max(l.interval_start,r.start_missing_date) + 1)*(not missing(r.start_missing_date)) as the_missing_length,
                        l.interval_length - (min(l.interval_end,r.end_missing_date) - max(l.interval_start,r.start_missing_date) + 1)*(not missing(r.start_missing_date)) as adj_length
      FROM
                        &inCounts. AS l
      LEFT JOIN
                        _fmc_missing_range_data as r
 
             ON
                        /* Some part of the missing range interval lies inside the interval in the count data */
                        NOT ( (r.end_missing_date < l.interval_start) OR (r.start_missing_date > l.interval_end) )
 
                        /* For each classification in the missing range table, either the classification is missing,
                        the classification is missing in the count data, or the classifications match */
                        %do i = 1 %to &missing_range_classes_count.;
                        %let missing_class_name = %scan(&missing_range_classes., &i., " ");
                        AND (   missing(r.&missing_class_name.)
                             OR missing(l.&missing_class_name.)
                             OR l.&missing_class_name. = r.&missing_class_name.
                            )
                        %end;
      ORDER BY
                     %do i = 1 %to &all_classifications_count.;
                        %let name = %scan(&all_classifications., &i., " ");
                        l.&name.,
                     %end;
                     %if (&cTool eq true) %then %do;
                        l.own,
                     %end;
                     l.interval_start,
                     l.interval_end
                     ;
   quit;
%mend orv_adjustForMissingRanges;
