/*****************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    %orva_freq_variable_select.sas                                             
 *  PURPOSE: Selects significant regressors to be used later in frequency fitting
 *           
 *  SUPPORT: jariou
 *
 *  
 *
 *  PARAMETERS:
 *
 *     data  =  Count dataset outputted by orva_freq_make_counts, with the following variables: OWN, each CCV,
 *              THRESHOLD, pobs, interval_start, interval_end, adj_length, the_count, CCVUID_:, CCUID
 *              possibly KRI_:, possibly? internal_bl_uid (if there were any bl-level KRIs).
 *
 *     selection = the model-selection option to be used to choose the best set of KRIs (regressors).  
 *                 This should be a string, provided by the user, and it may take on 
 *                 one of the following 5 values: Forward, Backward, Stepwise.  
 *                 Note that this is analogous to, but separate from, the parameter that 
 *                 orva_sev_fit_kri gets for selecting the best severity KRIs.  
 *                 Defaults to FORWARD.
 *
 *     outCompressedData = Output dataset containing (compressed) frequency counts, plus 
 *             the firm & businessline level KRIs (called KRI_:) and the panel dummy 
 *             variables for CCVUID and possibly OWN (called CCVUID_x_:, where x=0 to NUMCCVs-1).  
 *
 *     outParms          = Parameter dataset output by macro.
 *
 *     outSelKRIs        = dataset with a list of selected KRIs. 
 *
 *     outStatus         =  the status dataset for the macro- contains status (1 for failure) 
 *                          and reason (string)
 *
 *     outNotSelected    = dataset with a list of KRIs that were not selected
 * 
 *     outDroppedDummies = Dataset with the dummy variables that were not selected.  This is
 *                         used in the construction of the fbluid dataset.
 *
 *     outDummymap       = Dataset with the mapping from the dummy variable names to the strata
 *                         they represent.
 *
 *     outSelDummies     = Dataset with the names of the selected dummies
 *
 *  NOTE: If the model is at a finer-than-cell level, but no dummies are to be used, treat the model like
 *        one with dummies, but with none selected.
 *****************************************************************************************/
%macro orva_freq_variable_select ( 
                                  data                      = ,
                                  selection                 = ,
                                  outCompressedData         = ,
                                  outParms                  = ,
                                  outSelKRIs                = ,
                                  outNotSelected            = ,
                                  outDroppedDummies         = ,
                                  outDummyMap               = ,
                                  outSelDummies             =
                                  );                     
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_freq_variable_select (; 
      %put .                            data                      = &data,;
      %put .                            selection                 = &selection,;
      %put .                            outCompressedData         = &outCompressedData,;
      %put .                            outParms                  = &outParms,;
      %put .                            outSelKRIs                = &outSelKRIs,;
      %put .                            outNotSelected            = &outNotSelected,;
      %put .                            outDroppedDummies         = &outDroppedDummies,;
      %put .                            outDummyMap               = &outDummyMap,;
      %put .                            outSelDummies             = &outSelDummies;
      %put .                            );                     
   %end;

/*  Macro variables that were already set in orva_freq_fit.sas:
 *
 *  Freq_Kris_Suggested - space-delimited list of the KRIs (bl and firm) suggested for inclusion 
 *              in the model.  The selection heuristic (in proc reg) will choose the significant
 *              KRIs from this list.
 *                  
 *  Freq_Kris_Forced - space-delimited list of the KRIs (bl and firm) that MUST be included 
 *              in the model.  These enter the the "INCLUDE=" option on proc reg's 
 *              model statement.
 *              
 *  FREQ_USE_KRIS - true or false
 *
 *  FREQ_USE_DUMMIES - true or false
 *
 *  FREQ_MODEL_BASIS_PROPS
 */
   %local FOUND_ERROR;
   /* create the status data set up front */
   %let ORV_REASON         = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %let ORV_ABORT          = &ORV_TRUE;
   %let FOUND_ERROR        = &ORV_FALSE.;
   %let numDummies         = 0;
   %let dummyNames         = ;
   %let numSelectedKRIs    = 0;    
   %let numSelectedDummies = 0;
   %let SelectedKRIs       = ;
   %let SelectedDummies    = ;

   /* clear out out temporary output datasets */
   %orvu_nullify_data(data = &outParms);

   /* Trim the KRIs to remove any that we don't have values for */
   %let counts_data = ;
   %orvu_find_dataset( 
                      namein  = freqcounts_i, 
                      libs    = VARPRJ VARPRJ_M, 
                      nameout = counts_data
                      );
   data _fsel_kris;
      set &counts_data.;
	  keep KRI_:;
   run;
   %let usable_KRIs = %orv_varList(_fsel_kris);
   %let Freq_Kris_Forced    = %orv_listIntersection(&Freq_Kris_Forced., &usable_KRIs.);
   %let Freq_Kris_Suggested = %orv_listIntersection(&Freq_Kris_Suggested., &usable_KRIs.);

   /* Determine the number of Freq_Kris_Forced and Freq_Kris_Suggested */
   %let numForcedKRIs    = %orvu_str_getwordcount(strIn=&Freq_Kris_Forced, delimiters=%str( ));
   %let numSuggestedKRIs = %orvu_str_getwordcount(strIn=&Freq_Kris_Suggested, delimiters=%str( ));

   /* Set up dummy variables per level of stratification.  If no dummies but stratified,
      still call this to create the 'dummymap'.  */
   %if %lowcase(&freq_model_is_stratified) eq true %then %do;
      %_orva_createDummyVariables( 
                                  indata                  = &data., 
                                  outdata                 = &data.,
                                  makeDummiesOnM          = &Freq_model_basis_props,
                                  outDroppedDummiesM      = &outDroppedDummies, 
                                  outDummyMapM            = &outDummyMap,
                                  numDummiesM             = numDummies, 
                                  dummyNamesM             = dummyNames
                                  ); 
   %end;
   
   /* Check whether there are more possible parameters than count observations */
   %_orva_check_freq_paramcount;
   %if &FOUND_ERROR %then %do;
      %goto DONE;
   %end;

   /* If there are suggested variables, perform selection (all dummies are suggested by default) */
   %if %eval(&numDummies + &numSuggestedKRIs) GT 0 %then %do;
      %_orva_selectFreqVariables( indata = &data );
   %end;
   %else %do;  /* no need to do selection */
      %let numSelectedKRIs    = &numForcedKRIs;
      %let numSelectedDummies = 0;

      %if &numForcedKRIs GT 0 %then %do;
         data &outSelKRIs;
            length _NAME_ $ 32;
            %let nkri = 1;
            %let kri = %scan(&Freq_Kris_Forced, &nkri, ' ');

            %do %while (&kri NE %str() );
               _NAME_ = "&kri";
               output;
               %let nkri = %eval(&nkri + 1);
               %let kri  = %scan(&Freq_Kris_Forced, &nkri, ' ');
            %end;
         run; 
      %end;
   %end;

   data &outCompressedData(compress = binary);
      set &data;
      drop log_adjcount;
   run;

   data &outParms;
      length Freq_Kris_Suggested Freq_Kris_Forced $ 1000;
      numSelectedKRIs     = &numSelectedKris;
      numSelectedDummies  = &numSelectedDummies;
      Freq_Kris_Suggested = "&Freq_Kris_Suggested";
      Freq_Kris_Forced    = "&Freq_Kris_Forced";
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %goto DONE;
   %end;

   /* Only set abort to false if we get here without any errors */
   %let ORV_ABORT = &ORV_FALSE.;
%DONE:
   /*- delete unwanted data sets -*/
   %orvu_delete_data(data = _fvs_:);
%mend orva_freq_variable_select;

%macro _orva_check_freq_paramcount;
   %let countobs = 0;  
   %orvu_getObsCount(&data, countobs);

   %if %upcase(&Freq_Use_Kris) EQ TRUE %then %do;
      /* TODO: '+2' is not a good way to go here.  For NEGBIN it should be '+3' and other distributions could
         be defined with more paramaters... */
      %let param_limit = %eval(&numForcedKris + &numDummies + 2);
      /*  If the selection method is backwards, include suggested KRIs in the check - 
      PROC REG can produce a garbage fit in this case */
      %if %upcase(&selection) EQ BACKWARD %then %do;
         %let param_limit = %eval(&param_limit + &numSuggestedKris);
      %end;
   %end;
   %else %do;
      /* TODO: same as 'TODO' above */
      %let param_limit = %eval(&numDummies + 2);
   %end;
   %if %sysevalf(&countobs LE &param_limit, BOOLEAN) %then %do;
      %if %upcase(&selection) EQ BACKWARD %then %do;
         %let ORV_REASON = %sysfunc(sasmsg(&orvads, fs_too_many_suggested_error, noquote, &countobs));
      %end;
      %else %do;
         %let ORV_REASON = %sysfunc(sasmsg(&orvads, fs_too_many_params_error, noquote, &countobs));
      %end;
      %let FOUND_ERROR = &ORV_TRUE.;
   %end;
%mend _orva_check_freq_paramcount;

%macro _orva_createDummyVariables( 
                                  indata             = ,
                                  outdata            = ,
                                  makeDummiesOnM     = ,
                                  outDroppedDummiesM = , 
                                  outDummyMapM       = ,
                                  numDummiesM        = , 
                                  dummyNamesM        =
                                  );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;   
      %put @_orva_createDummyVariables(; 
      %put .                            indata             = &indata,;
      %put .                            outdata            = &outdata,;
      %put .                            makeDummiesOnM     = &makeDummiesOnM,;
      %put .                            outDroppedDummiesM = &outDroppedDummiesM,; 
      %put .                            outDummyMapM       = &outDummyMapM,;
      %put .                            numDummiesM        = &numDummiesM,; 
      %put .                            dummyNamesM        = &dummyNamesM;
      %put .                            );
   %end;
   
   %let &numDummiesM = 0;
   %let &dummyNamesM = ;    

   %if &makeDummiesOnM NE %str() %then %do;
      data _fsel_designinput;
         set &indata(keep = &makeDummiesOnM );
      run;

      %orva_designmatrix(  
                         A                  = _fsel_designinput, 
                         design             = _fsel_designoutput, 
                         sparse             = true, 
                         combosonly         = false, 
                         drop               = true, 
                         outDroppedDums     = &outDroppedDummiesM,
                         keepallowndummies  = false, 
                         status_design      = _fsel_statusdesign
                         );

      /* Any subtrata which were not used for dummies (always drop at least 1) are included here.  While there
         is technically no dummy for these, they are part of the fbluid datasets, which is created from
         the dummy map, so including them now makes life easier down the road. */
      %_orva_MakeDummyMap( 
                           props   = &makeDummiesOnM, 
                           thedata = _fsel_designoutput, 
                           themap  = &outDummyMapM
                           );
      /* in design output, rename the 'props' with a 'the_' prefix so that the sql 'on' clause
         to folow does not throw warnings to the log */
      data _fsel_designoutput;
         set _fsel_designoutput;
         rename &Freq_Model_Basis_Props. = Z&Freq_Model_Basis_Props.;
      run;

      /***** add dummy var columns to &indata *****/
      proc sql noprint;
         CREATE TABLE 
                           _fsel_work4 
         AS SELECT 
                           *
         FROM 
                           &indata l 
         LEFT OUTER JOIN
                           _fsel_designoutput r
         ON 
                           l.&Freq_Model_Basis_Props. EQ r.Z&Freq_Model_Basis_Props.
         ;
      quit;

      /* After this data step, Dummy_&FReq_model_basis_pros._: will be dummies.  Drop the 'the_' prefixed
         variables that were created earlier. */
      data &outdata; 
         set _fsel_work4(drop = Z&Freq_Model_Basis_Props.);
      run;

      /***** Determine the names of and the number of dummy variables. *****/
      proc contents 
                  data = &outdata 
                  out  = _fsel_content(keep = NAME) 
                  noprint
                  ;
      run;

      data _fsel_content2;
         set _fsel_content;
         if compare(upcase(scan(NAME, 1, '_')), 'DUMMY') EQ 0 then output;
      run;

      %if %lowcase(&freq_use_Dummies) eq true %then %do;
         %orvu_getObsCount(_fsel_content2, &numDummiesM);
         %if &&&numDummiesM GT 0 %then %do;
            proc sql noprint;
               select 
                        NAME
               into 
                        :&dummyNamesM separated by " "
               from 
                        _fsel_content2;
            quit;
         %end;
      %end;
   %end;
   %else %do;
      data &outdata;
         set &indata;
      run;  
   %end;
%mend _orva_createDummyVariables;

%macro _orva_selectFreqVariables( indata = );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;   
      %put @_orva_selectFreqVariables(;
      %put .                           indata = &indata;
      %put .                           );
   %end;

   /* Select the variables with significant impact on the mean count */
   proc reg 
            data   = &indata 
            outest = _fvs_reg_a 
            noprint  /* tableout */
            ;
   /* Don't use KRI_: or you'll use the severity KRIs too, which is wrong */
      model 
         log_adjcount = &Freq_Kris_Forced 
                        &Freq_Kris_Suggested 
                        /* Only include dummies if requested */
                        %if %lowcase(&Freq_use_dummies) eq true %then %do;
                           &dummyNames.
                        %end;
         / selection  = &selection 
           SLENTRY    = 0.05 
           SLSTAY     = 0.05
           %if &numForcedKris GT 0 %then %do; 
              include = &numForcedKris  
           %end;   
           ;
   run;
   
   proc transpose 
      data = _fvs_reg_a(
                        drop = 
                               _RMSE_ 
                               Intercept 
                               log_adjcount
                               ) 
            out = _fvs_reg
            ;
      id _TYPE_; 
   run;

   data _regResults_ 
      &outSelDummies(keep = _NAME_)
      &outSelKris(keep = _NAME_)
      &outNotSelected(rename = (PARMS = Estimate));

      length _NAME_ $ 32;         
      set _fvs_reg end = last;

      if PARMS NE . then do;
         if upcase(scan(_NAME_, 1, "_")) EQ "KRI" then do;
            numSelectedKRIs + 1;
            output &outSelKris;
         end;
         else do;
            numSelectedDummies + 1;
            output &outSelDummies;
         end;
         output _regResults_;
      end;
      else do;
         output &outNotSelected;
      end;

      if last then do;
         call symputX("numSelectedKris", numSelectedKRIs);
         call symputX("numSelectedDummies", numSelectedDummies);
      end;

      drop 
          numSelectedKRIs 
          numSelectedDummies 
          ;
   run;

%mend _orva_selectFreqVariables;

/*
 * Purpose: To make a map that reflects the corresponding relations of the 
 * dummy name, the level variable name and the unique value. The relation is
 * represented differently in the input dataset that usually was generated
 * by %orva_designmatrix. For example: The input
 * ------------------------------------------
 * |a |b  | dummy_a_1 |dummy_a_2 |dummy_b_1 |
 * +------+-----------+----------+----------+
 * |1 |10 |   1       |   0      | 1        |
 * |1 |11 |   1       |   0      | 0        |
 * |2 |10 |   0       |   1      | 1        |
 * |2 |11 |   0       |   1      | 0        |
 * |3 |10 |   0       |   0      | 1        |
 * |3 |11 |   0       |   0      | 0        |
 * ------------------------------------------
 * is mapped to:
 * ----------------------------------------
 * |Value |Level_variable_name | _name_   |
 * +------+--------------------+----------+
 * | 1    |a                   |dummy_a_1 |
 * | 2    |a                   |dummy_a_2 |
 * | 10   |b                   |dummy_b_1 |
 * ----------------------------------------
 * 
 * Notice that the value 3(of a) and 11(of b) are dropped as model required.
 * a and b here represent two levels of distinct value seperation.  --xiaoma 
 */

%macro _orva_MakeDummyMap( 
                          props   = ,
                          thedata = , 
                          themap  = 
                          );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;   
      %put @_orva_MakeDummyMap(; 
      %put .                    props   = &props,;
      %put .                    thedata = &thedata,; 
      %put .                    themap  = &themap;
      %put .                    );
   %end;

   data &themap;
      set _NULL_;
   run;

   %let nprops = 1;
   %let prop_curr = %upcase(%scan(&props, &nprops, ' '));

   %do %while (&prop_curr ne );
      %let dsid   = 0; 
      %let dumnum = 0;
      %let nobs   = 0;
      %let dsid   = %sysfunc(open(&thedata, I));

      %if &dsid NE 0 %then %do;
         %let nobs   = %sysfunc(attrn(&dsid, nlobs));
         %let dumnum = %sysfunc(varnum(&dsid, Dummy_&prop_curr._1));
         %let rc     = %sysfunc(close(&dsid));
      %end;

      %if &dumnum GT 0 AND &nobs GT 0 %then %do;
         data _dm1_&nprops;
            set &thedata;

            keep &prop_curr 
                 Dummy_&prop_curr._:
            ;
         run;

         proc sort 
               data = _dm1_&nprops 
               out  = _dm2_&nprops 
               nodupkey
               ;
            by &prop_curr;
         run;

         %let numLevels = 0;
         %orvu_getObsCount(
                            _dm2_&nprops, 
                            numLevels
                            );

         %if &numLevels GT 0 %then %do;
            data _dm3_&nprops(
                              keep = 
                                    _NAME_ 
                                    VALUE 
                                    Level_Variable_Name
                                    _DISPLAY_NAME_
                                    );
               set _dm2_&nprops(
                                 rename = (&prop_curr = VALUE)
                                 );

               length level_variable_name _NAME_ $ 32;
               array Dum{&numLevels} Dummy_&prop_curr._1 - Dummy_&prop_curr._&numLevels;
               Level_Variable_Name = "&prop_curr";
               named = 0;

               do j = 1 to &numLevels;
                  if not named then do;
                     if Dum{j} = 1 then do;
                        _NAME_ = "Dummy_&prop_curr._" || trim(left(j));
                        named = 1;
                     end;
                     /* This is the last stratum, on which there is no dummy (n strata get n-1 dummies) */
                     else if j = &numLevels then do;
                        _NAME_ = "Dummy_&prop_curr._" || trim(left(j));
                     end;
                  end;
               end;

               if missing(_NAME_) then DO;
                  /* TODO: This is an error.  Put an error message here */
               end;

               _DISPLAY_NAME_ = compress("Dummy_"||level_variable_name||"_"||value);
            run;

            data &themap;
               set 
                  &themap 
                  _dm3_&nprops
                  ;
            run;
         %end;
      %end;

      %let nprops = %eval(&nprops + 1);
      %let prop_curr = %upcase(%scan(&props, &nprops, ' '));
   %end;
    
   %orvu_delete_data( data = _dm: );
%mend _orva_MakeDummyMap;

/*     Macro that returns the list of variables from a specified dataset.

   PARAMETERS:
       &ds - the dataset

   EXAMPLES:
       %orv_varList(sashelp.class);
       %orv_varList(doesnotexist);
*/

%macro orv_varList(ds);
   %local
		nvars 
		nobs 
		varlist 
		i
		;
   %let dsid = %SYSFUNC(open(&ds));

   %if &dsid ne 0 %then %do;
      %let nvars  = %sysfunc(attrn(&dsid, nvars));

      %if &nvars LT 1 %then %do;
         %put WARNING: &ds. has no variables.;
      %end;
      %else %do;
         %do i = 1 %to &nvars;
            %let varlist = &varList. %sysfunc(varname(&dsid., &i.));
         %end;

         %let rc = %sysfunc(close(&dsid.));
         &varlist
         %return;
      %end;
   %end;
   %else %do;
      %put ERROR: Open for data set &dset failed - %sysfunc(sysmsg());
   %end;
%mend orv_varList;

/*
   Macro that simulates the IN function used in data steps.
   That is, it checks to see if &string is an element of &list.

   NOTE:
       It is not enough that &string be contained in an element of
   &list.  Only if &string IS AN ELEMENT of &list, then true
   is returned.

   EXAMPLE:
       %orv_listIn(%STR(el1 el2 el3), el2);      - will return 1 (true)
       %orv_listIn(%STR(el1 el2 el3), el4);      - will return 0 (false)
*/
%macro orv_listIn(list, string);
   %if %length(&list.) <= 0 %then %do;
      %let list = %str( );
   %end;

   %if (%sysfunc(findw(%str(&list), &string))) %then 1;
   %else 0;
%mend orv_listIn;

/*
   Macro that returns the intersection of two lists.

   EXAMPLE:
      %orv_listIntersection(%STR(element1 element2 element3), %STR(element2 element4)); returns element2
      %orv_listIntersection(%STR(element1 element2 element3), %STR(element4 element5)); returns empty string
*/
%macro orv_listIntersection(list1, list2);
   %local return i;
   %let return = %str( );
   %let i = 1;

   %do %while (%length(%qscan(&list1, &i, %str( ))) GT 0);
       %if (%orv_listIn(%str(&list2), %qscan(&list1, &i, %str( )))) %then %do;
          %let return = &return %sysfunc(strip(%qscan(&list1, &i, %str( ))));
       %end;
       %let i = %eval(&i + 1);
   %end;
   &return.
%mend orv_listIntersection;
