/*****************************************************************************************
 *   Copyright (c) 2004 by SAS Institute Inc.  Cary NC 27513  USA
 *
 *   Name: ORVA_freqcorr.sas                                             
 *   Description: Compute the correlation of the frequency of the losses.
 *          The correlations are computed at a smaller interval, interval,
 *          then the interval of the simulation, final_interval,
 *          so that more points can be used to create the covariance matrix. 
 *          A normal copula is used for the joint distribution and smoothing is
 *          used to create the bins.
 *          
 *
 * PARAMETERS
 *
 * intData     = the name of the internal loss data with KRIS.  Must have the following variables:
 *               AMOUNT and CCUID.
 * intSParm    = the name of the internal Severity fitting data set.
 *
 * intFParm    = the name of the internal Frequency fitting data set. 
 *               The data set has one row for each cell containing the FPARM data + CCUID.
 *
 * rrFParm     = the name of the RR Frequency fitting data set. 
 *               The data set has one row for each cell containing the FPARM data + CCUID.
 *
 * ccLibref    =  the name of the libref with the cc data information
 *
 * useWght     = Flag to use a weighting scheme:          
 *                       1 to turn it on 
 * 
 * lambda      = weight to use if usewght = 1             
 *                       0.0001 <= lambda <= 0.9999   
 *
 * GenTotals   = 0 or 1 to indicate whether covariances for total cells are needed.
 *      
 * interval    = SAS interval, defaults to QTR
 *
 * outStatus   = Success or failure - Default is outstatus 
 *
 * outCorr     = Correlation matrix result               
 *                                                      
 ****************************************************************/
 %macro ORVA_freqcorr(
                      intdata   = ,
                      intSParm  = ,
                      intFParm  = ,  
                      rrFParm   = ,
                      useWght   = 0,
                      lambda    = .5,
                      interval  = QTR,
                      CCLibref  = varprj., 
                      GenTotals = 1,
                      outStatus = outstatus,
                      outCorr   = outcorr 
                      );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #ORVA_freqcorr(; 
      %put .               intdata    = &intdata.,;
      %put .               intSParm   = &intSParm.,;
      %put .               intFParm   = &intFParm.,; 
      %put .               rrFParm    = &rrFParm.,;
      %put .               useWght    = &useWght.,;
      %put .               lambda     = &lambda.,;
      %put .               interval   = &interval.,;
      %put .               CCLibref   = &CCLibref.,; 
      %put .               GenTotals  = &GenTotals.,;
      %put .               outStatus  = &outStatus.,;
      %put .               outCorr    = &outCorr.;
      %put .               );                     
   %end;

   /*- create the status data set upfront -*/
   %let reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( 
                   statusds = &outstatus, 
                   status   = 1, 
                   reason   = &reason 
                   );
   %let status = 0;
   %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));
   %let abort  = 0;
   %let width  = 5;

   /*- error checking -*/
   %if &intdata = %str() %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_specify_intdata_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, fcorr_IntData_missing_error, noquote));
      %goto DONE;
   %end;

   %if  &usewght = 1 %then %do;
      %if %sysevalf(&lambda < 0.0001, BOOLEAN) 
         or 
          %sysevalf(&lambda > 0.9999, BOOLEAN) %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_lambda_out_of_range_error, noquote));
         %let abort  = 1;
         %let reason = &orvar_error %sysfunc(sasmsg(&orvads, fcorr_lambda_out_of_range_error, noquote));
         %goto DONE;
      %end;
   %end;

   /* -------------------------- */
   /* STEP 1: Get the count data */
   /* -------------------------- */
   %let count_data = ;
   %orvu_find_dataset(
                      nameIn  = freqcounts_i, 
                      Libs    = varprj varprj_m, 
                      nameOut = count_data
                      );

   data _fcor_counts;
      set &count_data.;
      array ccvuids ccvuid_:;
      do i = 1 to dim(ccvuids);
         if missing(ccvuids[i]) then do;
            delete;
         end;
      end;  
      drop i;
   run;

   proc sort data = _fcor_counts;
      by CCUID;
   run;
 
   /* ------------------------------------------------------------------- */
   /* STEP 2: Match up each observation with the correct model options    */
   /* ------------------------------------------------------------------- */

   /* Drop the KRI coefficients from the fparm dataset.  They
      share the same names as the KRIs themseslves and so will
      cause roblems for the merge */
   data _fcor_freq_parms;
      set &intFParm.;
      keep freq: ccuid;
   run;

   proc sort data = _fcor_freq_parms;
      by CCUID;
   run;

   data _fcor_counts;
      merge 
            _fcor_counts (in = in_freqcounts)
            _fcor_freq_parms (in = in_selfparm)
            ;
      by CCUID;

      if in_freqcounts and in_selfparm then do;
         output;
      end;
   run;

   /* ------------------------------------------------------------------- */
   /* Step 3: add in net exposure and aggregate the counts to the model   */
   /*         level - same thing we do in freq_fit. Do this cell by cell. */
   /* ------------------------------------------------------------------- */

   %let ccuids = ;
   proc sql noprint;
      select distinct
                  CCUID
      into
                  :CCUIDs separated by " "
      from
                  _fcor_counts
      ;
   quit;

   /* Get the cc dataset */
   %let ccdata = ;
   %orvu_find_dataset(
                      nameIn  = cc, 
                      Libs    = varprj varprj_m, 
                      nameOut = ccdata
                      );

   /* get the scaling data */
   %let scaledata = ;
   %orvu_find_dataset(
                      nameIn  = scalei, 
                      Libs    = varprj varprj_m, 
                      nameOut = scaledata
                      );

   /* Will need a list of all the KRIs separated by commas to pass in here */
   %let FREQ_NAMESOFKRIS_comma = ;
   data _null_;
      set _fcor_counts;
      if _N_ = 1;
      length FREQ_NAMESOFKRIS_comma $ 1000;
      array KRIs kri:;

      if dim(KRIs) > 0 then do;
         FREQ_NAMESOFKRIS_comma = strip(vname(KRIs[1]));

         do i=2 to dim(KRIs);
            FREQ_NAMESOFKRIS_comma = strip(FREQ_NAMESOFKRIS_comma) || "," || strip(vname(KRIs[i]));
         end;
      end;

      call symput("FREQ_NAMESOFKRIS_comma",FREQ_NAMESOFKRIS_comma);
   run;

   %let ccuid_count = %sysfunc(countw("&CCUIDs."));
   %do i=1 %to &ccuid_count.;
      %let current_ccuid = %scan(&ccuids.,&i.,' ');

      /* Create the cell_where string */
      %let CELL_WHERE = ;
      %orvu_cell_where(
                        ccuid = &current_ccUid., 
                        ccData = &ccdata.
                        );

      data _fcor_counts&current_ccuid.;
         set _fcor_counts;
         where ccuid eq &current_ccUid.;
         call symput("freq_model_basis_props",freq_model_basis_props);
         call symput("freq_model_is_stratified",freq_model_is_stratified);
         call symput("Freq_Use_Kris",Freq_Use_Kris);
      run;

      /* Now apply the threshold options to get the threshold used to calc. pobs */
      %orva_freq_thresholds(  
                            countdata     = _fcor_counts&current_ccuid., 
                            instrataprops = &freq_model_basis_props.,
                            stratified    = &freq_model_is_stratified.,
                            outcountdata  = _fcor_counts&current_ccuid.
                            );

      /* For the cell, compute net expore (pobs, exposure, FScale) */
      data _fcor_temp_sparms;
         set &intSParm.;
         where ccuid eq &current_ccUid.;
      run;

      %orva_freq_compute_net_exposure( 
                                       in_counts    = _fcor_counts&current_ccuid.,
                                       dataType     = I,
                                       ccv_property = &freq_model_basis_props.,
                                       stratified   = &freq_model_is_stratified.,
                                       sev_model    = _fcor_temp_sparms,
                                       scaleData    = &scaledata.,
                                       out_counts   = _fcor_counts&current_ccuid.,
                                       out_pobs     = _fcor_junk 
                                       );

      /* Perform any aggregation of counts. If model is cell level, then 
         counts will be summed over the slicing dimension. Net exposure 
         will be a weighted sum of net exposures */
      %orva_freq_aggregate_counts( 
                                  in_counts                = _fcor_counts&current_ccuid., 
                                  freq_model_is_stratified = false,
                                  model_ccv_property       = ,
                                  keep_KRIs                = false, 
                                  KRI_names                = , 
                                  out_counts               = _fcor_counts&current_ccuid.
                                  );

      /* Abort if any errors were found */
      %if &ORV_ABORT. eq &ORV_TRUE. %then %do;
         %goto DONE;
      %end;

      /* freq_agrgegate_counts only keeps the interval, net_exposure, and count
         info, so add CCUID back in */
      data _fcor_counts&current_ccuid.;
         set _fcor_counts&current_ccuid.;
         ccuid = &current_ccuid.;
      run;

   %end;

   /* Merge the counts for each cell back together */
   data _fcor_counts;
      set 
      %do i=1 %to &ccuid_count.;
         %let current_ccuid = %scan(&ccuids.,&i.,' ');
         _fcor_counts&current_ccuid.
      %end;;
      adjCount = the_Count / net_exposure;
   run;

   /* Make sure we have a alpha char for the key */
   data _fcor_counts; 
      set _fcor_counts;
      length _newkey $ 32;

      _newkey = compress('Cell_'|| put(CCUID, &ORVAR_CELLFMT));
      n = adjCount;
      keep 
          _newkey 
          n 
          interval_start
          ;
   run;

   proc sort data = _fcor_counts; 
      by interval_start;
   run;

   /* make the counts into columns */
   proc transpose 
                  data = _fcor_counts 
                  out  = _transpose(drop = _NAME_);
      by interval_start;
      id _newkey;
      var n;
   run;

   data _transpose; 
      set _transpose;
      drop 
          interval_start 
          i;
   run;

   %if ( &usewght = 1 ) %then %do;
      data _NULL_;
         set _transpose;
         call symputX( 'nobs', _N_);
      run;

      data _transpose;
         set _transpose;
         omega = (( &lambda) ** (&nobs - _N_ ) ) * (1 - &lambda);
      run ;

      proc corr 
               data   = _transpose 
               outk   = &outcorr 
               vardef = weight
               noprint 
               ;
         weight omega;
      run;
   %end;
   %else %do;
      proc corr 
               data = _transpose 
               outk = &outcorr 
               noprint
               ;
      run;
   %end;

   data &outcorr; 
      set &outcorr;
      where _type_ = "CORR";
      /* Convert back to a pearson */
      array y[*] _NUMERIC_;

      do i = 1 to dim(y);
         y[i] = sin(constant('pi')*y[i]/2); 
      end;

      output;
   run;

   /* get an identity matrix of the system to make the covariance matrix the same */
   /* size as all of the other generated covariance matrices ---------------------*/
   %ORVA_COMBINE_COV(    
                     ccLibref  = &cclibref,  
                     outCov    = _Identity_,
                     GenTotals = 0,
                     ostat     = _status_ 
                     );
   /* check the status */
   data _NULL_; 
      set _status_;
      call symputX('i_status', status);
   run;
   %if &i_status = 1 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_id_mat_create_failed_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, fcorr_id_mat_create_failed_error, noquote));
      %goto DONE;
   %end;

   /* combine the two matrices -------------------------------*/
   proc sort 
            data = _Identity_; 
      by _NAME_; 
   run;

   proc sort 
            data = &outcorr; 
      by _NAME_; 
   run;

   data &outcorr; 
      merge 
           _Identity_ 
           &outcorr
           ; 
      by _NAME_; 
      length _TYPE_ $ 8;
      array miss[*] _numeric_;

      do i = 1 to dim( miss );
         if miss[i] = . then do;
            miss[i] = 0;
         end;

         /* We do not want negative correlation */
         if miss[i] < 0 then do;
            miss[i] = 0;
         end;
      end;

      _TYPE_='CORR';
   run;

   %if &GenTotals = 1 %then %do;
      %orva_rollup_cov(
                       cov      = &OutCorr, 
                       ccLibref = &ccLibref
                       );
   %end;

%DONE:

   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_abnormal_error, noquote));
      %let status = 1;
   %end;  
   %orva_status_ds( 
                   statusds = &outstatus, 
                   status   = &status, 
                   reason   = &reason 
                   );

   %orvu_delete_data( data =
                            _transpose 
                            _freqcnts 
                            _fcor:
                            );
%mend;
