/*-------------------------------------------------------------------
 * Copyright (C), 2005
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 *
 * %orva_get_hill_estimate(data=, UseKRI=, tailpnts=, hillplot=)
 *
 * Computes the tail index for a data set based on Hill's algorithm.

 * Input:
 *    data = loss data with loss value=AMOUNT
 *    tailpnt = subset of the tail numbers use in hill's method
 * NOTE: This version is based on IML code.
 *-------------------------------------------------------------------*/
%macro orva_get_hill_estimate(
                              data        = ,
                              UseKri      = NO,
                              tailpnts    = 2000,
                              hillplot    = _hillplot
                              );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put #orva_get_hill_estimate(;
      %put .                        data        = &data,;
      %put .                        UseKri      = &UseKri,;
      %put .                        tailpnts    = &tailpnts,;
      %put .                        hillplot    = &hillplot;
      %put .                        );
   %end;

   /* We do not make distinction between fitting with or without KRI's */
   %let AMOUNT_VAR = amount;
    
   /*    Set default 2000 points in hill estimate */
   /*    User can overwrite this option by specify tailpnts=*/
   /*    In most cases, all the data points will be used*/
   /*    Otherwise use all the data points to estimate tail*/
   /*    Selected K depend on the number of points used in this estimation*/
   /*    The bootstrap sample size n1 and n2 are adjusted according to */
   /*                    the number of points used*/
   /*      Consider increasing the bootstrap replicates number from 10 to 100*/
   /*      Empirical did not show a lot improvment, but might take longer to run*/
   proc sort 
            data = &data 
            out  = _tmpData
            ;
     by &amount_var;
   run;

   proc sql noprint;
      SELECT 
            count(*) 
      INTO 
            :NOBS
      FROM 
            _tmpData
            ;
   quit;

   /*If # of obs in the data set is less than 6, no hill estimation will be done*/
   %if  %sysevalf( &nobs < 6, boolean ) %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, ghe_not_enough_pts_for_tailest_error, noquote));
      %let status = 1;
      %put "&reason";
   %end;
   %else %do;
      /* If # of obs in the data set is less than the specified tailpnts used for estimation*/
      /* then we use all the data points and reset tailpnts to the number of obs*/
      %if  %sysevalf( &tailpnts > &nobs, boolean ) %then %do;
         %put "%sysfunc(sasmsg(&orvads, ghe_all_data_for_tailest_note, noquote))";
         %let tailpnts = &NOBS;

         data _work_;
            set _tmpData;
         run;
      %end;

   /*If # of obs in the data set is more than the specifed tailpnts, */
   /*only the largest &tailpnts order statistics will be used for estimation*/
   /*NOTE: we set default to 2000, this should be large enough for most data*/
   /*      unless user rewrite our setting*/
    %else %do;
        %put "%sysfunc(sasmsg(&orvads, ghe_largest_data_for_tailest_note, noquote, &tailpnts))";

        data _work_;
            set _tmpData;
            if _n_ > (&nobs - &tailpnts) then output;
        run;
    %end;

   /*The tuning parameter, n1, is selected w.r.t */
   /*the number of points used for estimation*/
   /*We only pick 10 to 90 quantiles as k1*/
   /*We only do 10 bootstrap samples for each n1 or n2*/
   /*This can be increased to 100 for better accuracy */
   /*but might need more computing time*/
   %let n1_S = %sysfunc(int(%sysevalf(&tailpnts / 10)));

   %if  %sysevalf( &n1_S < 1, boolean ) %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, ghe_not_enough_pts_for_tailest_error, noquote));
      %let status = 1;
      %put "&reason";
      %goto GETOUT;
   %end;

   %let n1_L = %sysfunc(int(%sysevalf(&tailpnts/2)));
   %let n1_U = %eval(&n1_L + &n1_S * (%sysfunc(int(%sysevalf((&tailpnts - &n1_L) / &n1_S))) - 1));
   %let k_L  = 10;
   %let k_U  = 90;
   %let k_S  = 10;
   %let boot = 10;

    /*    Call hill plot here*/
    %get_hill_plot(
                   inputds = _work_, 
                   outds   = &hillplot,
                   var     = &amount_var,
                   size    = 100
                   )

   /*Find bootstrap sample and then calculate Q for each n1 and n2*/
   %do n_1 = &n1_L %to &n1_U %by &n1_S;
      %let n_2 = %eval(&n_1 * &n_1 / &tailpnts);

      %get_bootsamp(
                    _work_,
                    boot_&n_1,
                    &amount_var,
                    &tailpnts,
                    &n_1,
                    &boot
                    )
      %compute_q(
                 boot_&n_1,
                 &n_1,
                 Q_&n_1,
                 &boot,
                 &K_L,
                 &K_U,
                 &K_S
                 )

      %get_bootsamp(
                    _work_,
                    boot_&n_2,
                    &amount_var,
                    &tailpnts,
                    &n_2,
                    &boot
                    )
      %compute_q(
                 boot_&n_2,
                 &n_2,
                 Q_&n_2,
                 &boot,
                 &K_L,
                 &K_U,
                 &K_S
                 )

      proc sort data = Q_&n_1(rename = (col1 = q1 col2 = n1 col3 = k1));
         by n1 k1;
      run;

      proc sort data = Q_&n_2(rename = (col1 = q2 col2 = n2 col3 = k2));
         by n2 k2;
      run;

      proc summary data = Q_&n_1;
         var q1;
         by n1 k1;
         output out = mq1 mean = Q1;
      run;

      proc summary data = Q_&n_2;
         var q2;
         by n2 k2;
         output out = mq2 mean = Q2;
      run;

      proc append 
                  base = bootsampq1 
                  data = mq1
                  ;
      run;

      proc append 
                  base = bootsampq2 
                  data = mq2
                  ;
      run;
   %end;
 
    /*    Find the k1 and k2 for each n1 and n2 that minize Q*/
    proc sort data = bootsampq1;
      by n1 q1;
    run;
    
    data bootsampq1;
        set bootsampq1;
        if missing(q1) then delete;
    run;

    data  bootsampq1stat;
        set bootsampq1;
        by n1;
        if first.n1 then output;
        drop _type_ _freq_;
    run;

    proc sort data = bootsampq2;
       by n2 q2;
    run;

    data bootsampq2;
        set bootsampq2;
        if missing(q2) then delete;
    run;

    data  bootsampq2stat;
        set bootsampq2;
        by n2;
        if first.n2 then output;
        drop _type_ _freq_;
    run;

    data bootsampqstat;
        merge  
             bootsampq1stat  
             bootsampq2stat
             ;
    run;

    /*Find combination of n1 and n2 that minimize R*/
    data mqstat;
        set bootsampqstat;
        R = Q1 * Q1 / Q2;
        if missing(R) then delete;
    run;

    proc sort data = mqstat;
        by R;
    run;

    data mqfinal;
        set mqstat;
        if _n_ = 1 then output;
    run;

    /*Find K  */
    data mqfinal;
        set mqfinal;
        k_1 = k1 * k1 / k2;
        k_2 = (log(k1) * log(k1)) / (2 * log(n1) - log(k1))**2;
        k_3 = (log(n1) - log(k1)) / log(n1);
        k   = int(k_1 * k_2**k_3);
        call symputX("K",k);
    run;
    
    proc sort data = _work_;
        by  
          descending 
          &amount_var
          ;
    run;

   /* If hill estimation failed (K=.) or K is out of reasonable range,*/
   /* pick the 5th largest order statistics for the user*/
   %if   %sysevalf( &k > &tailpnts, boolean ) OR 
         %sysevalf( &k <= 0, boolean ) %then %do;
      %put %sysfunc(sasmsg(&orvads, ghe_hillest_failed_error, noquote));

      data _null_;
         set _work_ end = last;

         if _n_ = 5 then do;
            call symput("PARAM_xLimit", put(&amount_var, BEST20.));
            call symput("TailBoundaryValue", put(&amount_var, BEST20.));
         end;

         if last then do;
            call symput("PARAM_pnLimit", put((_N_ - 5) / _N_, BEST20.));
         end;
      run;    
   %end;

   /*If hill estimation suceeded, */
   /*then pick the largest (k+1)th order statistics as the cut off point*/
   %else %do;
      %put %sysfunc(sasmsg(&orvads, ghe_hillest_success_note, noquote, %eval(&k + 1)));

      data _null_;
         set _work_ end = last;

         if _n_ = %eval(&k + 1) then do;
            call symput("PARAM_xLimit", put(&amount_var, BEST20.));
            call symput("TailBoundaryValue", put(&amount_var, BEST20.));
         end;

         if last then do;
            call symput("PARAM_pnLimit", put((_N_-&k-1)/_N_, BEST20.));
         end;
      run;    
   %end;

%GETOUT:
      %orvu_delete_data( data =
                                _tmpData 
                                boot_: 
                                Q_: 
                                mq1 
                                mq2 
                                bootsampq1 
                                bootsampq2 
                                bootsampq1stat 
                                bootsampq2stat
                                bootsampqstat 
                                mqstat 
                                mqfinal 
                                _work_
                                );
 
   %end;
%mend orva_get_hill_estimate;

/*Generate specified number of bootstrp samples (nboot=10) */
/*with sample size equal to n1 or n2 */
%macro get_bootsamp(
                     inputds , /*input data set: contain all points used for tail estimation*/
                     outds   ,     /* output bootstrap sample data set for n1 or n2*/
                     var     ,        /* variable of interest */
                     nobs    ,       /* number of observations in input data set */
                     n       ,        /*bootstrap sample size*/
                     nboot       /* number of bootstrap samples */
                     );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put @get_bootsamp(;
      %put .              inputds = &inputds,;
      %put .              outds   = &outds,;
      %put .              var     = &var,;
      %put .              nobs    = &nobs,;
      %put .              n       = &n,;
      %put .              nboot   = &nboot; 
      %put .              );
   %end;

   proc iml;
      use &inputds;
      read all var{&var} into amount;
      close &inputds;

      tailpnts = &nobs;
      n = &n;
      boot = &nboot;

      bootsamp = j(n * boot, 2, .);
      c = j(n * boot, 1, n * 12345);
      index = uniform(c);

      do l = 1 to boot;
         do m = 1 to n;
         p = index[(l - 1) * n + m, ];
         tmp = floor(p * tailpnts) + 1;
         bootsamp[(l - 1) * n + m, 1] = amount[tmp, ];
         bootsamp[(l - 1) * n + m, 2] = n;
         end;
      end;

      create &outds from bootsamp[colname = "&amount_var"];
      append from bootsamp;
   quit; 
%mend get_bootsamp;

/*Find Q for each bootstrap sample and selected K*/
%macro compute_q(
                  inputds  ,  /* input data set for n1 or n2  */
                  index    ,  /* n1 or n2*/
                  outputds ,  /* output data set containing  q  */
                  nboot    ,  /*boot strap sample times*/
                  K_L      ,  /*lowest quantile for bootstrap sample*/
                  K_U      ,  /*highest quantile for bootstrap sample*/
                  K_S         /*step size*/
                  );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put @compute_q(;
      %put .           inputds  = &inputds,;
      %put .           index    = &index,;
      %put .           outputds = &outputds,;
      %put .           nboot    = &nboot,;
      %put .           K_L      = &K_L,;
      %put .           K_U      = &K_U,;
      %put .           K_S      = &K_S;
      %put .           );
   %end;

   proc iml;
      use &inputds;
      read all var{&amount_var} into bootsamp;
      close &inputds;

      K_L  = &K_L;
      K_U  = &K_U;
      K_S  = &K_S;
      n    = &index;
      boot = &nboot;

      nk = (k_U-k_L) / k_S + 1;
      Q  = j(nk * boot, 3, .);

      do i = 1 to boot;
         work = bootsamp[(i - 1) * n + 1:(i - 1) * n + n];
         call sort(work, 1, 1);

         do k_loop = k_L to k_U by k_S;
            p = (k_loop - k_L) / k_S + 1;
            obsindex = int(n * k_loop / 100);

            if obsindex = 0 then do;
               Q[(i - 1) * nk + p, 1] = .;
               Q[(i - 1) * nk + p, 2] = n;
               Q[(i - 1) * nk + p, 3] = obsindex;
            end;
            else do;
               logdiff   = 0;
               logdiffsq = 0;

               do j = 1 to obsindex;
                  logdiff = logdiff + (log(work[j, 1]) - log(work[obsindex + 1, 1]));
                  logdiffsq = logdiffsq + (log(work[j, 1]) - log(work[obsindex + 1, 1]))**2;
               end;

               gamma = logdiff / obsindex;
               M = logdiffsq / obsindex;
               /*  Now put the number in the right row???*/
               Q[(i - 1) * nk + p, 1] = (M - 2 * gamma**2)**2;
               Q[(i - 1) * nk + p, 2] = n;
               Q[(i - 1) * nk + p, 3] = obsindex;
            end;
         end;
      end;

      create &outputds from Q;
      append from Q;
   quit;
%mend compute_q;

/*Get the Hill plot of the largest size (100) order statistics
for the input data set*/
%macro get_hill_plot(
                     inputds, /*input data set used to gerenate hill plot*/
                     outds,  /*output data set contains K and alpha*/
                     var,    /* variable name for analysis*/
                     size,   /*number of obs used the plot*/
                     );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put @get_hill_plot(;
      %put .               inputds = &inputds,;
      %put .               outds   = &outds,;
      %put .               var     = &var,;
      %put .               size    = &size ;  
      %put .               );
   %end;

   proc sort 
            data = &inputds 
            out  = _hilldata
            ; 
      by descending &var;
   run;
    
   proc sql noprint;
      SELECT 
               count(*) 
      INTO 
               :INPUTOBS
      FROM 
               _hilldata
               ;
   quit;

   %if  %sysevalf( &size > &INPUTOBS, boolean ) %then %do;
      %let size = &INPUTOBS;

      data _hilldata;
         set _hilldata;
      run;
   %end;
   %else %do;
      data _hilldata;
         set _hilldata;
         if _n_ <= &size then output;
      run;
   %end;

   proc iml;
      use _hilldata;
      read all var{&var} into amount;
      close _hilldata;
      H = j(&size - 1, 2, .);

      do i = 1 to &size - 1;
         logdiff = 0;
         tmp = log(amount[i + 1]);

         do j = 1 to i;
            logdiff = logdiff + (log(amount[j]) - tmp);
         end;

         H[i, 1] = (logdiff / i)**(-1);
         H[i, 2] = i + 1;
      end;

      create &outds from H;
      append from H;
   quit;
%mend get_hill_plot;
