/******************************************************************************
 * Copyright (C), 2011
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 *
 * NAME:     orva_get_model_info.sas
 * PURPOSE:  Produces macro variables that define the complete severity and frequency
 *           model.  Called in orva_simulate.  The macros defined here are used in 
 *           orva_conf_interval, orva_var_gradient, and orva_extra_loss.
 *
 * SUPPORT:  stmajo
 *
 * Parameters:
 *
 *   SimPriority_Sev  = Dataset with severity model info.
 *   SimPriority_Freq = Dataset with frequency model info.
 *   loss*            = the losses for the model used in this cell with KRI data appended
 *
 * NOTE: The macros defined in here need to be declared in orva_simulate so that their
 *       scope extends to orva_conf_interval and orva_extra_loss.
 *
 ***********************************************************************************/
%macro orva_get_model_info(
                           SimPriority_Sev   = ,
                           SimPriority_Freq  = ,
                           lossi             = ,
                           lossc             = 
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_get_model_info(;
      %put .                     SimPriority_Sev   = &SimPriority_Sev,;
      %put .                     SimPriority_Freq  = &SimPriority_Freq,;
      %put .                     lossi             = &lossi,;
      %put .                     lossc             = &lossc;
      %put .                     );
   %end;

   /* Define "infinity" to check against later */
   %let LOG_INFINITY = 30;  /* <= EXP(LOG_INFINITY) is larger than 10 trillion */

   /* Set up the CELL_WHERE macro */
   %let CELL_WHERE = ;
   %orvu_cell_where(
                     ccuid  = &ccuid,
                     ccData = &ccData
                     );

   /* Both severity and frequency models may need these datasets, so intialize here */
   %let firm_kri_sims  = %str();
   %let strat_kri_sims = %str();

   /*****************************************************************/
   /*           Get the Severity Distribution Information           */
   /*****************************************************************/
   data _NULL_;
      set &SimPriority_Sev.;
      where CCUID = &ccuid AND upcase(SOURCE_TYPE) in ( "I", "C", "P", "T");
      call symput("Sev_Source_Type", SOURCE_TYPE);
      call symput("Sparm_dataset_name", dataset_name);
   run;

   /* Source type of T is Consortium tool, which uses 'C' datasets like
      sparmc, selfparmc, etc. */
   %if %upcase(&Sev_Source_Type.) eq T %then %do;
      %let Sev_Source_Type = C;
   %end;

   /* Error check: */
   %orvl_ds_exists(&Sparm_dataset_name, "orva_conf_interval", false);
   %if &ds_nobs eq 0 %then %do;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON =  gmi_no_sev_info_error;
      %return;
   %end;

   /* Pull out the distribution and KRI information */
   data _NULL_;
      set &Sparm_dataset_name;
      where CCUID = &CCUID;

      length firm_kri_list $32000;
      length strat_kri_list $32000;

      n_f_kris       = 0;
      n_strat_kris   = 0;
      firm_kri_list  = "";
      strat_kri_list = "";

      call symputx("sev_iteration",iteration);
      sdist = strip(distribution);
      call symputx("sdist", sdist);

      if count_kri = . then do;
         count_kri = 0;
      end;

      call symputx("sev_KRI_count", count_kri);
      call symputx("sparamcount", paramcount);
      call symputx("sev_kri_model", KRI_MODEL);
      call symputX("Sev_ObsCount", obsCount);
      call symputx("_POBS_", put(Avg_count_weighted_pobs, BEST20.));

      /* get parameters as macros */
      array params param_:;

      do i = 1 to dim(params);
         call symputX(compress("S"||vname(params[i])), put(params[i], BEST20.));
      end;
      if not missing(param_pnlimit) then do;
         call symputX("param_xlimit", put(param_xlimit, BEST20.));
         call symputX("param_pnlimit", put(param_pnlimit, BEST20.));
      end;

      /* Get separated counts / lists of firm-level stratum level KRIs */
      %if count_kri > 0 %then %do;
         array kri[*] Param_1_KRI_:;

         do i = 1 to dim(kri);
            if kri[i] ne . then do;
               len = length(trim(vname(kri[i])));

               if len eq 16 then do;
                  strat_kri_list = trim(strat_kri_list) || "  " || trim(substr(vname(KRI(i)), 9, 8));
                  n_strat_kris = n_strat_kris + 1;
               end;
               else do;
                  firm_kri_list = trim(firm_kri_list) || "  " || trim(substr(vname(KRI(i)), 9, 10));
                  n_f_kris = n_f_kris + 1;
               end;
            end;
         end;
      %end;

      call symputx( "SEV_N_F_KRIs", n_f_kris);
      call symputx( "SEV_FIRM_KRI_LIST", trim(firm_kri_list));
      call symputx( "SEV_N_STRAT_KRIs", n_strat_kris);
      call symputx( "SEV_STRAT_KRI_LIST", trim(strat_KRI_LIST));
   run;

   /* If body/tail, get the body disttribution.  If not, set body_dist to the name of
      the distribution.  */
   %let BODY_DIST = ;
   %orva_set_sev_dist(
                        distIn  = &sdist.,
                        distOut = sdist
                        );
   %if &BODY_DIST eq . %then %do;
      %let BODY_DIST = &sdist.;
   %end;

   /* load the severity distribution macro */
   %orvd_&sdist;

   /* If there are severity stratum-level KRIs, then the severity distribution is a 
   mixture over strata. To avoid continually conditioning on KRIs, 
   treat the case with no KRIs like there are KRIs, but only 1 stratum. */
   %if %lowcase(&scale_transform.) eq id %then %do;
      %let SEV_SCALES = &Sparam_1;
      %let MAX_SCALE  = &Sparam_1;
      %let MIN_SCALE  = &Sparam_1;
   %end;
   %else %do;
      %let SEV_SCALES = %sysfunc(exp(&Sparam_1.));
      %let MAX_SCALE  = &SEV_SCALES;
      %let MIN_SCALE  = &SEV_SCALES;
   %end;
   %let nsev_mix        = 1;
   %let SPARAM1_WEIGHTS = 1;
   %let Sev_Strat_IDs   = 0;
   /* This will be pulled from a dataset eventually */
   %let SEV_STRAT_TYPE = INTERNAL_BL_UID;

   /* Get the severity parameters used in simulation- if KRIs are present, get the scale parameter
      implied by the KRI model.  For cell slice models, get weights for each slice also */
   %if &SEV_KRI_COUNT > 0 %then %do;
      /* If cell slice model, get weights for each cell and grab the cell-slice level KRIs */
      %if &SEV_N_STRAT_KRIs gt 0 %then %do;
         /* Weights associated with each cell slice and the list of the cell slice IDs */
         proc sort 
                  data = &&loss&Sev_Source_Type 
                  out  = _sim_sev_kridata
                  ;
            where &CELL_WHERE;
            by &SEV_STRAT_TYPE; 
         run;

         data _gmi_weights;
            set  _sim_sev_KRIdata;
            by &SEV_STRAT_TYPE;
            retain weight;
   
            if first.&SEV_STRAT_TYPE then do;
               weight = 0;
            end;

            weight = weight + 1;

            if last.&SEV_STRAT_TYPE then do;
               output;
            end;
         run;

         proc sql noprint;
            SELECT
                  weight / sum(weight),
                  &SEV_STRAT_TYPE,
                  count(weight)
            INTO
                  :Sparam1_weights SEPARATED BY " ",
                  :Sev_Strat_IDs SEPARATED BY " ",
                  :nsev_mix
            FROM
                  _gmi_weights
                  ;
         quit;

         /* Cell slice level KRIs */
         %ORVU_find_dataset( 
                              namein  = kribl_sim, 
                              libs    = VARPRJ VARPRJ_M, 
                              nameout = strat_kri_sims
                              );

         data _sim_sev_strat_kris;
            set &strat_kri_sims;
            where BLUID_KRI in ( &Sev_Strat_IDs );
         run;

         proc sort data = _sim_sev_strat_kris;
            by RISK_FACTOR_VARIABLE_DATE;
         run;
      %end;

      /* If firm KRIs, get the simulated KRI values */
      %if &SEV_N_F_KRIs gt 0 %then %do;
         %orvu_find_dataset( 
                              namein  = krifirm_sim, 
                              libs    = VARPRJ VARPRJ_M, 
                              nameout = firm_kri_sims
                              );
         data _sim_sev_firm_kris;
            set
            /* When (if) we merge the firm and cell slice kri datasets, they need to be the same size */
            %do i=1 %to &nsev_mix.;
               &firm_kri_sims
            %end; ;
         run;

         proc sort data = _sim_sev_firm_kris;
            by RISK_FACTOR_VARIABLE_DATE;
         run;
      %end;

      /* Get the severity scale parameters for each loss group.  Do this by 
         combining kribl_sim and krifirm_sim and then constructing the parameters
         'from scratch' */
      data _sim_sev_kris_combined;
         MERGE  
               %if &SEV_N_STRAT_KRIs gt 0 %then %do;
                  _sim_sev_strat_kris 
               %end;
               %if &SEV_N_F_KRIs gt 0 %then %do;
                  _sim_sev_firm_kris
               %end; 
               ;
         BY RISK_FACTOR_VARIABLE_DATE; 
      run;

      /* Now go through and compute the scale for each business line */
      %let scale_error = false;
      data _sim_sev_kris_combined;
         set _sim_sev_kris_combined end = last;
         retain 
               max_theta 0 
               min_theta 0
               ;

         /* get model intercept */
         %if %lowcase(&SCALE_TRANSFORM.) eq id %then %do; 
            intercept = log(&SParam_1);
         %end;
         %else %do;
            intercept = &SParam_1.;
         %end;

         /* compute the scale for this slice.  Check that it isn't too large */
         log_theta = intercept + &sev_kri_model.;

         if log_theta gt &LOG_INFINITY then do;
            call symput("scale_error", "true");
         end;
         else do;
            theta = exp(log_theta);
         end;

         max_theta = max(theta, max_theta);

         if _N_ = 1 then do;
            min_theta = theta;
         end;
         else do;
            min_theta = min(theta, min_theta);
         end;

         /* Store the largest and smallest scales */
         if last then do;
            call symputx("MAX_SCALE", put(max_theta,BEST20.));
            call symputx("MIN_SCALE", put(min_theta,BEST20.));
         end;
      run;
      %if &scale_error eq true %then %do;
         %let ORV_ABORT   = &ORV_TRUE;
         %let found_error = &ORV_FALSE;
         %let ORV_REASON  = gmi_scale_too_large_error;
         %return;
      %end;

      /* Put the scales into a macro */
      proc sql noprint; 
         SELECT 
                  theta
         INTO 
                  :SEV_SCALES SEPARATED BY " "
         FROM 
                  _sim_sev_kris_combined
      quit;

      %let nsev_mix = %sysfunc(strip(&nsev_mix));
   %end;

   /*****************************************************************/
   /*         Get the Frequency Distribution Information            */
   /*****************************************************************/
   /* Get the frequency distribution dataset */
   %let freq_strat_dsname = none;

   data _NULL_; 
      set &SimPriority_Freq;
      where &CCUID = ccuid AND upcase(SOURCE_TYPE) in ( "I", "C", "P", "T") ;
      call symput("Fparm_dataset_name", dataset_name);
      call symput("Freq_Source_Type", SOURCE_TYPE);
      /* Dataset with loss grouping specific KRI coeficients ( fparm_mdli ) */
      call symput("freq_strat_dsname", fmodels_ds);
   run;

   /* Source type of T is Consortium tool, which uses 'C' datasets like
      sparmc, selfparmc, etc. */
   %if %upcase(&Freq_Source_Type.) eq T %then %do;
      %let Freq_Source_Type = C;
   %end;

   /* Error checks: */
   %orvl_ds_exists(
                     &Fparm_dataset_name, 
                     "orva_get_model_info", 
                     false
                     );

   %if &ds_nobs eq 0 %then %do;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON =  gmi_no_freq_info_error;
      %return;
   %end;

   %if &freq_strat_dsname eq none %then %do;  
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = gmi_no_freq_kri_info_error;
      %return;
   %end;

   %orvl_ds_exists(
                     &freq_strat_dsname, 
                     "orva_get_model_info", 
                     false
                     );

   %if &ds_nobs eq 0 %then %do;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = gmi_no_freq_kri_info_error;
      %return;
   %end; 

   /* Pull out the distribution and KRI information */
   data _NULL_;
      set &Fparm_dataset_name;
      where CCUID = &CCUID;
      call symputx("fiteration", iteration);
      call symput("fdist", strip(distribution));
      call symputX("Freq_ObsCount", Freq_ObsCount);

      if missing(count_kri) then do;
         count_kri = 0;
      end;

      call symputx("freq_KRI_count", count_kri);
      call symput("freq_KRIs", strip(SelectedKRIs));
      call symput("FREQ_MODEL_BASIS_PROPS", strip(Freq_Model_Basis_Props));

      if missing( COUNT_DUMMIES ) then do;
         COUNT_DUMMIES = 0;
      end;

      if missing(intercept) then do;
         intercept = safe_log(FParam_1);
      end;

      call symputx("freq_dummy_count", COUNT_DUMMIES);
      call symputx("fparamcount", paramcount);
      call symput("Freq_Fit_Interval", strip(freq_fit_interval));
      call symput("Freq_Model_Is_Stratified", Freq_Model_Is_Stratified);
      call symput("Freq_Strata_Count", Freq_Strata_Count);
      call symput("Freq_KriModel_Full", Freq_KriModel_Full);
      call symput("Freq_dummy_list", selectedDummies);
      call symput("Freq_intercept", intercept);
   run;

   /* Load the distribution macros */
   %orvd_&fdist;

   %let fit_sim_ratio = 1;

   /* Get the parameters used in simulation. */
   data _NULL_;
      set &Fparm_dataset_name;
      where ( CCUID = &CCUID. );

      %do i = 1 %to &fparamcount.;
         call symputx("FPARAM_&i.", put(FPARAM_&i.,BEST20.));
      %end;

      call symputX("FParam_1_0", FParam_1);
   run;

   /* If there are KRIs, get the effective param_1 for each stratum */
   %if ( (&FREQ_KRI_COUNT gt 0) or (&freq_dummy_count gt 0) ) %then %do;
      /* Need to distinguish between firm level and risk category level KRIs */
      data _null_;
         set &Fparm_dataset_name;
         where ( CCUID = &CCUID );
         length firm_kri_list $2000;
         length strat_kri_list $2000;

         %if &freq_dummy_count gt 0 %then %do;
            if _N_ = 1 then do;
               array  DUMMY DUMMY_:;
               length DUMMY_LIST $2000;
               length DUMMY_VALS $2000;
               length DUMMY_STRATS $2000;
      
               do i = 1 to dim(DUMMY);
                  if DUMMY[i] NE . then do;
                     DUMMY_LIST = trim(DUMMY_LIST) || " " || trim(vname(DUMMY(i)));
                     DUMMY_VALS = trim(DUMMY_VALS) || " " || DUMMY(i);
                     DUMMY_STRATS  = trim(DUMMY_STRATS)  || " " || trim( compress(lowcase(vname(DUMMY(i))), 'abcdefghijklmnopqrstuvwxyz.-_ '));
                  end;
               end;

               call symputx( "FREQ_DUMMY_LIST", DUMMY_LIST );
               call symputx( "FREQ_DUMMY_VALS", DUMMY_VALS );
               call symputx( "FREQ_DUMMY_STRATS", DUMMY_STRATS );
            end;
         %end;

         n_f_kris = 0;
         n_strat_kris = 0;

         %if &freq_KRI_count gt 0 %then %do;
            array kri[*] KRI_:;

            do i = 1 to dim(kri);
               if kri[i] ne . then do;
                  y = trim(vname(kri[i]));
                  z = length(y);

                  if z eq 8 then do;
                     strat_kri_list = trim(strat_kri_list) || "  " || trim(vname(KRI(i)));
                     n_strat_kris = n_strat_kris + 1;
                  end;
                  else do;
                     firm_kri_list = trim(firm_kri_list) || "  " || trim(vname(KRI(i)));
                     n_f_kris = n_f_kris + 1;
                  end;
               end;
            end;
         %end;

         call symputx( "FREQ_N_F_KRIs", n_f_kris);
         call symputx( "FREQ_FIRM_KRI_LIST", firm_kri_list);
         call symputx( "FREQ_N_STRAT_KRIs", n_strat_kris);
         call symputx( "FREQ_STRAT_KRI_LIST", STRAT_KRI_LIST);
      run;

      /* Get the KRI datasets and the cell slice IDs */
      %if &FREQ_N_STRAT_KRIs gt 0 %then %do;

         proc sql noprint;

            SELECT DISTINCT 
                  &Freq_Model_Basis_Props.
            INTO
                  :Freq_Strat_IDs separated by " "
            FROM
                  &freq_strat_dsname
            WHERE
                  iteration = &fiteration.
            ;
         quit;

         /* If we didn;t already pull this for severity, pull it here */
         %if &strat_kri_sims. eq %str() %then %do;
            %ORVU_find_dataset( 
                                 namein  = kribl_sim, 
                                 libs    = VARPRJ VARPRJ_M, 
                                 nameout = strat_kri_sims
                                 );
         %end;
      %end;

      %if &FREQ_N_F_KRIs gt 0 %then %do;
         %if &firm_kri_sims eq %str() %then %do;
            %ORVU_find_dataset( 
                                 namein  = krifirm_sim, 
                                 libs    = VARPRJ VARPRJ_M, 
                                 nameout = firm_kri_sims
                                 );
         %end;
      %end;

      data _sim_freq_kris_combined;
         retain FParam_1_sum 0;
         /* For cell slice models, need the cell slice IDS- one row per slice */
         %if %lowcase(&Freq_Model_Is_Stratified.) eq true %then %do;
           set &freq_strat_dsname.( keep = 
                                          &FREQ_MODEL_BASIS_PROPS. iteration CCUID 
                                          where = ( 
                                                      (CCUID = &CCUID.) 
                                                   and 
                                                      (iteration = &fiteration.) 
                                                   ) 
                                    ) 
                                    end  = last
                                    ;
         %end;

         /* If cell slice level KRIs, add them in */
         %if &FREQ_N_STRAT_KRIs. gt 0 %then %do;
            set &strat_kri_sims.( where = (BLUID_KRI in ( &Freq_strat_IDS. ) )
                                  drop  = RISK_FACTOR_VARIABLE_DATE
                                  );
         %end;

         /* If firm level KRIs, add them in also */
         %if &FREQ_N_F_KRIs. gt 0 %then %do;
            set
               %do i = 1 %to &Freq_Strata_Count.;
                  &firm_kri_sims.(drop = RISK_FACTOR_VARIABLE_DATE)
            %end; ;
         %end;

         Fparam_1 = exp( &freq_intercept. + %substr(&Freq_KriModel_Full.,4) );

         %do i = 2 %to &FParamCount;
            FParam_&i = &&FParam_&i;
         %end;

         FParam_1_sum = FParam_1_sum + FParam_1;

         call symputX("FParam_1", FParam_1_sum);

         drop 
               iteration 
               bluid_kri
               FParam_1_sum
               ;
      run;

   %end;

   /* Adjust for fit interval / simualtion interval differences */
   data _NULL_;
      fit_sim_ratio = intck("&Freq_Fit_Interval", 0, 366) / intck("&SimInterval", 0, 366);
      call symput('fit_sim_ratio', put(fit_sim_ratio, BEST20.));
   run;

   %if &fit_sim_ratio NE 1 %then %do;
      data _null_;
         %do i = 1 %to &FParamCount.;
            FParam_&i. = &&FParam_&i.;
         %end;
         distribution = "&fdist.";
         /* adjust macro variable values for fit / simulation interval */
         call ChangeInterval(distribution, &fit_sim_ratio., &theFreqSimCallParams.);
         %do i = 1 %to &FParamCount.;
            call symputX("FParam_&i.",FParam_&i.);
         %end;
      run;

      /* If there are KRIs, change the values in this dataset as well */
      %if ( (&FREQ_KRI_COUNT gt 0) or (&freq_dummy_count gt 0) ) %then %do;
         data _sim_freq_kris_combined;
            set _sim_freq_kris_combined;
            distribution = "&fdist.";
            call ChangeInterval(distribution, &fit_sim_ratio., &theFreqSimCallParams.);
         run;
      %end;
   %end;

   %orvu_delete_data(data= _gmi_:)

%mend orva_get_model_info;
