/*****************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    ORVA_GetMomts.sas                                             
 * PURPOSE: Calculates frequency and severity moments and aggregations for all cells.           
 *          Calculates the aggregates, mean, and count for the Loss Effects Breakdown.
 * SUPPORT: stchri
 *
 * PARAMETERS:
 *
 *   lossData      =  the name of the loss data with the following variables:
 *   dataType      =  (I)nternal, (C)onsortium, (P)ublic.
 *   global_range  =  global range dataset for this project.
 *   missing_range =  name of the missing ranges dataset with the following variables:
 *                    &CCV_NAMES, START_MISSING_DATE, END_MISSING_DATE.
 *   interval      =  DATE interval (YEAR, MONTH ...). Default is YEAR.
 *   ccData        =  Needed for input to orvu_CELL_WHERE, to subset losses for this cell.
 *   numCCvs       =  Number of cross-classification variables in project.
 *   CCV_NAMES     =  space-delimited list of cross-classification variables in project.
 *   outSevMomts   =  MEAN, VARIANCE, STD, SKEWNESS, KURTOSIS, COUNT, MINLOSS, MAXLOSS, CCV_n, SUM (for AMOUNT)
 *                    LEGAL_LIABILITY (aggregate), LEGAL_LIABILITY_MEAN, LEGAL_LIABILITY_N (count)
 *                    REGULATORY_ACTION (aggregate), REGULATORY_ACTION_MEAN, REGULATORY_ACTION_N (count)
 *                    RESTITUTION (aggregate), RESTITUTION_MEAN, RESTITUTION_N (count)
 *                    LOSS_TO_ASSETS (aggregate), LOSS_TO_ASSETS_MEAN, LOSS_TO_ASSETS_N (count)
 *                    LOSS_OF_RECOURSE (aggregate), LOSS_OF_RECOURSE_MEAN, LOSS_OF_RECOURSE_N (count)
 *                    WRITE_DOWNS (aggregate), WRITE_DOWNS_MEAN, WRITE_DOWNS_N (count)
 *   outFreqMomts  =  CCUID, MEAN, VARIANCE, CCV_n
 *   outStatus     =  Success or failure - default is status
 *   sharedRollup  =  preform shared loss rollup of data first.
 *   singleEventMult = column that specifies shared loss key.
 *   outprops      =  Output dataset for panel properties.  This only needs to be created 
 *                    once per project per datatype, so it is ignored if sharedRollup=false
 *   outmissprops  =  Output dataset for panel props with missing ranges.  
 *                    This only needs to be created once per project per
 *                    datatype, so it is ignored if sharedRollup=false
 *   outmissrange  =  Updated missing range table.  This only needs to be created 
 *                    once per project per datatype, so it is ignored if sharedRollup=false
 *   scaleData     =  scaling dataset (currently only for internal data)
 *   
 *****************************************************************************************/
 %macro ORVA_GetMomts( 
                        lossData        = ,
                        dataType        = I,
                        global_range    = ,
                        missing_range   = ,
                        interval        = YEAR,
                        ccData          = ,
                        numCCVs         = 2,
                        CCV_NAMES       = INTERNAL_BL_UID INTERNAL_RC_UID , 
                        outSevMomts     = osstat,              
                        outFreqMomts    = ofstat,              
                        outSevMomtsRU   = ,              
                        outFreqMomtsRU  = ,              
                        outStatus       = statusgetmomts,
                        singleEventMult = ,
                        outprops        = ,
                        outmissprops    = ,
                        outmissrange    = , 
                        outCountData    = theOutCountData, 
                        ctool           = false,
                        scaleData       =
                        );  
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #ORVA_GetMomts(; 
      %put .               lossData        = &lossData.,;
      %put .               dataType        = &dataType.,;
      %put .               global_range    = &global_range.,;
      %put .               missing_range   = &missing_range.,;
      %put .               interval        = &interval.,;
      %put .               ccData          = &ccData.,;
      %put .               numCCVs         = &numCCVs.,;
      %put .               CCV_NAMES       = &CCV_NAMES.,; 
      %put .               outSevMomts     = &outSevMomts.,;              
      %put .               outFreqMomts    = &outFreqMomts.,;              
      %put .               outSevMomtsRU   = &outSevMomtsRU.,;              
      %put .               outFreqMomtsRU  = &outFreqMomtsRU.,;
      %put .               outStatus       = &outStatus.,;
      %put .               singleEventMult = &singleEventMult.,;
      %put .               outprops        = &outprops.,;
      %put .               outmissprops    = &outmissprops.,;
      %put .               outmissrange    = &outmissrange.,; 
      %put .               outCountData    = &outCountData.,; 
      %put .               ctool           = &ctool.,;
      %put .               scaleData       = &scaleData.;
      %put .               );  
   %end;

   /*- create the status data set up front -*/
   %let reason = %sysfunc( sasmsg(&orvads, unknown_error, noquote) );
   %orva_status_ds( 
                   statusds = &outstatus, 
                   status   = 1, 
                   reason   = &reason 
                   );
    
   %let status = 0;
   %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));
   %let abort  = 0;

   /*- empty out results -*/
   data &outSevMomts; 
      set _NULL_; 
   run;

   data &outFreqMomts; 
      set _NULL_; 
   run;

   data &outSevMomtsRU; 
      set _NULL_; 
   run;

   data &outFreqMomtsRU; 
      set _NULL_; 
   run;

   /*- error checking -*/
   %if &lossData = %str() %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, gmom_specify_global_range_ds_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, gmom_missing_lossData_error, noquote));
      %goto DONE;
   %end;

   %if &global_range = %str() %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, gmom_specify_global_range_ds_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, gmom_specify_global_range_ds_error, noquote));
      %goto DONE;
   %end;

   /***** get BYVARS macro variable *****/
   data _NULL_;
      length byvars $100;

      %do i = 0 %to &NUMCCVS - 1; 
         byvars = trim(left(byvars)) || " CCVUID_&i";
      %end;    

      call symput("BYCCVS", byvars);
   run;

   /* first handle normal moments */   
   %orv_getMomentsNoRollup(
                           lossData       = &lossData, 
                           CCV_NAMES      = &CCV_NAMES,       
                           byCCVs         = &BYCCVS, 
                           numCCVs        = &numCCvs,
                           ccData         = &ccData, 
                           interval       = &interval,
                           missing_range  = &missing_range,
                           outFreqMomts   = &outFreqMomts,
                           outSevMomts    = &outSevMomts,
                           global_range   = &global_range,
                           dataType       = &dataType,
                           outCounts      = &outCountData
                           );
    
   /* second handle shared loss rollup */
   %if &abort eq 0 %then %do;
      %let sharedLossCount = 0;

      %if &singleEventMult ne %str() %then %do;
         PROC SQL noprint;
            SELECT 
                     count(*) 
            INTO 
                     :sharedLossCount
            FROM 
                     &lossData
            WHERE 
                     &singleEventMult notin (0, .)
                     ;
         quit;
      %end;

      %if &sharedLossCount eq 0 %then %do;
         data &outSevMomtsRU; 
            set &outSevMomts; 
         run;

         data &outFreqMomtsRU; 
            set &outFreqMomts; 
         run;
      %end;
      %else %do;
         %orv_getMomentsRollup(
                               lossData        = &lossData, 
                               byCCVs          = &BYCCVS, 
                               numCCVs         = &numCCvs,
                               ccData          = &ccData, 
                               singleEventMult = &singleEventMult, 
                               dataType        = &dataType, 
                               outSevMomts     = &outSevMomtsRU, 
                               outFreqMomts    = &outFreqMomtsRU,
                               global_range    = &global_range
                               ); 
      %end;
   %end;
%DONE:

   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, gmom_abnormal_error, noquote));
      %let status = 1;
   %end;  

   %orva_status_ds( 
                   statusds = &outstatus, 
                   status   = &status, 
                   reason   = &reason 
                   );
%mend ORVA_GetMomts;

/*********************************************************************************
 * Shared Loss Rollup Moments
 * 1. Use the %orva_rollup() macro to break up the losses into 2 datasets:
 *    'singles' and 'shares'.
 *
 *    The 'shares' contain shared losses that are 'rolled up' for all possible
 *    permutations of CCVs and thus may have duplicates since entries are
 *    created (rolled up) in aggregate cells.
 *
 *    The 'singles' contain non-shared losses and should never have duplicates.
 *
 * 2. The %orva_cell_slices macro creates a dataset of 'slices' - combinations
 *    of cells that constitute permutations of the CCVs.
 *
 * 3. call normal moments macros for each slice.
 *********************************************************************************/
%macro orv_getMomentsRollup(
                            lossData        = , 
                            byCCVs          = , 
                            numCCVs         = , 
                            ccData          = ,
                            singleEventMult = , 
                            dataType        = , 
                            outSevMomts     = , 
                            outFreqMomts    = ,
                            global_range    =
                            );
   
   /***** rollup shared losses in _rollup dataset *****/
   %orva_rollup(
                lossData        = &lossData, 
                byCCVs          = &BYCCVS, 
                numCCVs         = &numCCvs,
                ccData          = &ccData, 
                singleEventMult = &singleEventMult, 
                outDataShared   = _rollup, 
                outDataSingle   = _norollup, 
                dataType        = &dataType
                );     

   /**** create slices for each level of aggregation *****/
   %let first      = true;    
   %let NUM_SLICES = 0;
   %let NUM_CCVARS = 0;

   %orva_cell_slices(
                     ccData         = &ccData,
                     cellKey        = CCUID,
                     outData        = _slices,
                     guardForTotals = true
                     );

   /***** loop over slices *****/
   %do slice = 1 %to &NUM_SLICES;
      /****** get slice of losses ******/
      %orv_getDataBySlice( 
                          sharedData  = _rollup,
                          singleData  = _norollup,
                          ccData      = &ccData,
                          sliceData   = _slices,
                          sliceNum    = &slice,
                          outData     = _slice
                          );

      /***************************************************************************
      * SEVERITY MOMENTS
      ****************************************************************************/
      %orv_getSeverityMoments(
                              lossData    = _slice, 
                              outSevMomts = _sliceMomts, 
                              byCCVs      = CCUID, 
                              numCCVs     = &numCCVs, 
                              ccData      = &ccData, 
                              byclass     = false
                              );

      %orvu_append(
                   base         = &outSevMomts,
                   new          = _sliceMomts,
                   canUseAppend = true
                   );     

      /***************************************************************************
      * FREQUENCY MOMENTS
      ***************************************************************************/
      %orva_getfreqmomts( 
                         lossData       = _slice,
                         dataType       = &dataType,
                         global_range   = &global_range,
                         interval       = &interval,
                         cTool          = &cTool,
                         missing_range  = &missing_range,
                         CCV_NAMES      = &CCV_NAMES, 
                         NUMCCVS        = &NUMCCVS,
                         outFreqMomts   = _sliceMomts,
                         outStatus      = _ofreqStatus,
                         rollup         = true
                         %if &scaleData NE %str() %then %do;
                            ,scaleData     = &scaleData
                         %end;
                         );    

      /***** stack up slice *****/
      %if &slice eq 1 %then %do;
         data &outFreqMomts; 
            set _sliceMomts; 
         run;
      %end;
      %else %do;
         proc append 
                     base = &outfreqMomts 
                     data = _sliceMomts
                     ; 
         run;
      %end;
   %end; /* looping over aggregation slices */
    
   /***** fix up CCVUID_<num> in moments datasets, 
   while the CCUID is correct the individual
   CCCUIDs are arbitrary at this point *****/
   data _NULL_; 
      set _slices( where = (whereClauseLength = 0));
      call symput('hashData', trim(left(keyHash)));
   run;

   data &outSevMomts; 
      set &outSevMomts(drop = 
                              %do i = 0 %to &numCCVs - 1; 
                                 CCVUID_&i 
                              %end; 
                              );
      if _N_ eq 1 then do;
         declare hash ccHash(dataset: "&ccData");
         rc = ccHash.defineKey("CCUID");
         rc = ccHash.defineData(&hashData);
         rc = ccHash.defineDone();

         %do i = 0 %to &numCCVs - 1;
            call missing( CCVUID_&i);
         %end;
      end;

      rc = ccHash.find();
      drop rc;
   run;

   data &outFreqMomts; 
      set &outFreqMomts;

      if _N_ eq 1 then do;
         declare hash ccHash(dataset: "&ccData");
         rc = ccHash.defineKey("CCUID");
         rc = ccHash.defineData(&hashData);
         rc = ccHash.defineDone();

         %do i = 0 %to &numCCVs - 1;
            call missing( CCVUID_&i);
         %end;
      end;

      rc = ccHash.find();
      drop rc;
   run;
   /******************************************************************************

   /*- clean up -*/
   %orvu_delete_data( data =
                             _slice: 
                             _data 
                             _rollup 
                             _norollup 
                             _ofreqStatus
                             );
%mend orv_getMomentsRollup;

%macro orv_getMomentsNoRollup(
                              lossData       = , 
                              byCCVs         = , 
                              numCCVs        = , 
                              ccData         = ,
                              CCV_NAMES      = , 
                              global_range   = ,
                              interval       = ,
                              missing_range  = ,
                              outSevMomts    = ,      
                              outFreqMomts   = ,
                              dataType       = ,
                              outCounts      =
                              );

   /******************************************************************************
   * SEVERITY MOMENTS
   * handled with one call to %orv_getSeverityMoments() using class variables
   *******************************************************************************/
   %orv_getSeverityMoments(
                           lossData    = &lossData, 
                           outSevMomts = _sliceMomts, 
                           byCCVs      = &BYCCVS, 
                           numCCVs     = &numCCVs, 
                           ccData      = &ccData
                           );

   /*- get CCUIDs back from ccData -*/
   proc sql noprint nowarn;
      CREATE TABLE 
                     &outSevMomts 
      AS SELECT 
                     * 
      FROM
                     _sliceMomts M 
         LEFT JOIN 
                     &ccData C 
         ON 
                     1
                     %do i = 0 %to &numCCVs - 1;    
                        and M.CCVUID_&i eq C.CCVUID_&i    
                     %end;
                     ;
   quit;

   /******************************************************************************
   * FREQUENCY MOMENTS
   * handled with one call to %orva_getfreqmomts() for interior cells and
   * one call to %orva_freqmomtsaggr() for aggregate cells.
   *******************************************************************************/ 
   %orva_getfreqmomts( 
                      lossData       = &lossData,
                      dataType       = &dataType,
                      global_range   = &global_range,
                      interval       = &interval,
                      missing_range  = &missing_range,
                      CCV_NAMES      = &CCV_NAMES,               
                      NUMCCVS        = &NUMCCVS,
                      outFreqMomts   = _sliceMomts,
                      outStatus      = &outStatus,
                      rollup         = false,
                      ccdata         = &ccData.,
                      outprops       = &outprops,
                      outmissprops   = &outmissprops,
                      outmissrange   = &outmissrange,
                      outCountData   = _outCounts1,
                      ctool          = &ctool
                      );    
                   
   /*- get CCUIDs back from ccData -*/
   proc sql noprint nowarn;
      CREATE TABLE 
                     _sliceMomts2 
      AS SELECT 
                     * 
      FROM
                     _sliceMomts M 
         LEFT JOIN 
                     &ccData C 
         ON
                     M.CCUID eq C.CCUID
                     ;
   quit;
        
   proc sql noprint nowarn;
      CREATE TABLE 
                     &outCounts 
      AS SELECT 
                     * 
      FROM
                     _outcounts1 M 
         LEFT JOIN 
                     &ccData C 
         ON
                     M.CCUID eq C.CCUID
                     ;
   quit;
      
  
   %orva_cell_slices(
                     ccData         = &ccData,
                     cellKey        = CCUID,
                     outData        = _slices,
                     guardForTotals = true
                     );  

   %orva_freqmomtsaggr(
                       slicesDataset = _slices, 
                       interiorCells = _sliceMomts2,
                       numCCVs       = &NUMCCVS,
                       outDataset    = &outFreqMomts,
                       ccData        = &ccData
                       );
      
  
   /*- clean up -*/
   %orvu_delete_data( data =
                            _sliceMomts 
                            _sliceMomts2 
                            _slices
                            );
%mend orv_getMomentsNoRollup;

/**********************************************************************************
  orv_getSeverityMoments: calculate SEVERITY moments using proc summary
 **********************************************************************************/
%macro orv_getSeverityMoments(
                              lossData    = , 
                              outSevMomts = , 
                              byCCVs      = , 
                              numCCVs     = , 
                              ccData      = , 
                              byclass     = true 
                              );
   /*- make sure loss dataset is sorted -*/
   proc sort 
            data = &lossData 
            out  = _sortedData( keep = 
                                       &byCCVs 
                                       AMOUNT 
                                       LEGAL_LIABILITY 
                                       RESTITUTION 
                                       LOSS_TO_ASSETS
                                       REGULATORY_ACTION
                                       LOSS_OF_RECOURSE 
                                       WRITE_DOWNS 
                                       NON_INSURANCE_RECOVERIES
                                       );
      by &byCCVs; 
   run;

   /*- handle moments for AMOUNT -*/
   proc summary 
               data   = _sortedData   
               vardef = n 
               noprint
               ;
      output 
            out  = _SevMoments1(drop = _type_ _freq_) 
            mean = MEAN 
            var  = VARIANCE 
            skew = SKEWNESS 
            kurt = KURTOSIS 
            n    = COUNT 
            min  = MINLOSS 
            max  = MAXLOSS 
            sum  = SUM
            ;
      var AMOUNT; 

      %if &byclass eq true %then %do;
         class &byCCVs;
      %end;
      %else %do;
         by &byCCVs;
      %end;
   run;

   /*for loss categories: legal_liability, regulatory_action, etc, set the zero-valued loss to
   missing, before doing PROC SUMMARY, to get the no zero 
   count of the losses
   */
   data _sortedData;
      set _sortedData;

      if LEGAL_LIABILITY = 0 then do;
         LEGAL_LIABILITY = .;
      end;

      if REGULATORY_ACTION = 0 then do;
         REGULATORY_ACTION = .;
      end;

      if RESTITUTION = 0 then do;
         RESTITUTION = .;
      end;

      if LOSS_TO_ASSETS = 0 then do;
         LOSS_TO_ASSETS = .;
      end;

      if LOSS_OF_RECOURSE =0 then do;
         LOSS_OF_RECOURSE = .;
      end;

      if WRITE_DOWNS = 0 then do;
         WRITE_DOWNS = .;
      end;

      if NON_INSURANCE_RECOVERIES = 0 then do;
         NON_INSURANCE_RECOVERIES = .;
      end;
   run;

   /*- do this separately because different stats are needed for this group -*/
   proc summary data = _sortedData;
      output 
            mean =   
                     LEGAL_LIABILITY_MEAN 
                     REGULATORY_ACTION_MEAN 
                     RESTITUTION_MEAN 
                     LOSS_TO_ASSETS_MEAN 
                     LOSS_OF_RECOURSE_MEAN 
                     WRITE_DOWNS_MEAN 
                     NON_INSURANCE_RECOVERIES_MEAN
            sum = 
                     LEGAL_LIABILITY 
                     REGULATORY_ACTION 
                     RESTITUTION 
                     LOSS_TO_ASSETS 
                     LOSS_OF_RECOURSE 
                     WRITE_DOWNS 
                     NON_INSURANCE_RECOVERIES
            n   = 
                     LEGAL_LIABILITY_N 
                     REGULATORY_ACTION_N 
                     RESTITUTION_N 
                     LOSS_TO_ASSETS_N 
                     LOSS_OF_RECOURSE_N 
                     WRITE_DOWNS_N 
                     NON_INSURANCE_RECOVERIES_N
            out =
                     _SevMoments2(drop = _type_ _freq_)
                     ;

      var 
            LEGAL_LIABILITY 
            REGULATORY_ACTION 
            RESTITUTION LOSS_TO_ASSETS 
            LOSS_OF_RECOURSE 
            WRITE_DOWNS 
            NON_INSURANCE_RECOVERIES
            ;

      %if &byclass eq true %then %do;
         class &byCCVs;
      %end;
      %else %do;
         by &byCCVs;
      %end;
   run;

   /*- merge moments datasets together -*/
   proc sort data = _SevMoments1; 
      by &byCCVs; 
   run; 

   proc sort data = _SevMoments2; 
      by &byCCVs; 
   run; 

   data &outSevMomts; 
      merge 
            _sevMoments1 
            _sevMoments2
            ;
      by &byCCVs;
      STD = sqrt(VARIANCE);
      Kurtosis = Kurtosis + 3;

      %do i = 0 %to &numCCVs - 1;    
         if CCVUID_&i eq . then do;
            CCVUID_&i = 0;    
         end;
      %end;
   run;

   /*- cleanup our temporary datasets -*/
   %orvu_delete_data( data =
                            _sorteddata 
                            _SevMoments1 
                            _SevMoments2
                            );
%mend orv_getSeverityMoments;

%macro orv_getDataBySlice(
                          sharedData =,
                          singleData =,
                          ccData     =,
                          sliceData  =,
                          sliceNum   =,
                          outData    =
                          );
   data _NULL_; 
      set &sliceData; 
      where index eq &sliceNum;
      call symputX("ccWhereNum", whereClauseLength);
      call symput("ccWhere", trim(left(whereClause)));
      call symput("guardedWhere", trim(left(guardedWhereClause)));
      call symput("ccKeep", trim(left(keepClause)));
      call symputX("ccKeepNum", keepClauseLength);
      call symput("keyHash", keyHash);
   run;
    
   /*** non-Aggregate Slice:
   need all non aggregate entries from both datasets ***/  
   %if &ccWhereNum eq 0 %then %do;
      data _DAT&sliceNum.; 
         set 
            &sharedData(where = (&guardedWhere)) 
            &singleData(where = (&ccWhere))
            ;
      run;
   %end;
   /*** GRAND-Aggregate slice:
   Need all entries from &singleData
   GuardedWhereClause works on &sharedData
   need to set all CCUIDs to grand aggregate cell ***/  
   %else %if &ccKeepNum eq 0 %then %do;
      /*- get CCUID for the grand aggregate cell -*/
      data _NULL_; 
         set &ccData; 
         where 1 
               %do i = 0 %to &numCCVs - 1;    
                  and CCVUID_&i eq 0    
               %end;
               ;
         call symputX('CCUID',CCUID); 
      run;   

      data _DAT&sliceNum.; 
         set &sharedData(where = (&guardedWhere)) 
         &singleData;
         CCUID = &CCUID;
      run;
   %end;
   /*** simple-Aggregate slice (along 1 dimension):
   Need all entries from &singleData 
   GuardedWhereClause works on &sharedData
   use hashmap to get corrected CCUID ***/ 
   %else %do;
      data _cc; 
         set &ccData; 
         where &ccWhere; 
         keep CCUID &ccKeep.; 
      run;

      data _DAT&sliceNum.; 
         set &singleData;

         if _N_ eq 1 then do;
            declare hash ccHash(dataset: '_cc', hashexp: 6);
            rc = ccHash.defineKey(&keyHash.);
            rc = ccHash.defineData("CCUID");
            rc = ccHash.defineDone();
         end;

         rc = ccHash.find();
      run;

      data _DAT&sliceNum.; 
         set &sharedData(where = (&guardedWhere)) 
         _DAT&sliceNum.;          
      run;
   %end;
   
   data &outData; 
      set _DAT&sliceNum.; 
   run;

   /*- cleanup our temporary datasets -*/
   %orvu_delete_data( data = 
                            _DAT&sliceNum. 
                            _cc
                            );
%mend orv_getDataBySlice;
