/***************************************************************************************** 
 * Copyright (C), 2008
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_init_global_functions.sas                                             
 *  SUPPORT: sasdje
 *  PURPOSE: Setup various FCMP function libraries needed by the OpRiskVaR.
 *           Currently creates default distortion functions and fast Fourier routines.
 *           This is called by the %orva_init macro at startup time.
 *  PARAMS:
 *          cmpLib            = Library where to store the FCMP functions
 *                              It is up to the caller to make sure that this is in the 
 *                              cmp search path.
 *
 *
 *****************************************************************************************/
%macro orva_init_global_functions( cmplib = work.orvar );
   proc fcmp outlib = &cmplib..utils;
      /* Function to keep Param_1 to escape to infinity */
      /* when fitting with KRI's                        */
      function Safe_Id(x);
         if x LE 25 then do;
            return(x);
         end;
         else do;
            return(25 + 5 * (exp((x - 25) / (x + 25)) - 1));
         end;
      endsub;

      /* Safe Log funtion to avoid problems when taking */
      /* log of zero (0) in KRI severity fitting        */
      function Safe_Log(x);
         return(log(x + 1E-30));
      endsub;

      function run_macro_ds( macro $, dataset_name $,ccuid, parmindex, _sparm $);
         ok = run_macro( "orva_append_data", dataset_name ,ccuid, parmindex, _sparm );
         put macro =;
         return(ok);
      endsub;

      /*- Shell sort ------------------*/
      subroutine orva_sort_mat( mat[*,*], col);
         outargs mat;
         inc = 1;
         n = dim1(mat);
                   /*Determine the starting increment */
         do while ( inc <= n);
            inc = inc * 3;
            inc +=1;
         end;
         do while (inc > 1);      /* begin the main loop */
            inc = floor(inc / 3);

            do i = inc + 1 to n;
               v = mat[i, col];
               j = i;
               t =  mat[j-inc , col];

               do while (t > v);
                  mat[j, col] = mat[j - inc, col];
                  j = j - inc;
                  if j <= inc then goto next;
                  t =  mat[j - inc , col];
               end;
         next:      
               mat[j, col] = v;
            end;
         end;
      endsub;
   
     function SVRTUTIL_EDF(y, dim, x[*], F[*], Ftype);
        if (missing(y)) then
            return(.);

        eps = 2.220446E-16; /* constant('MACEPS') */
        if (Ftype = 1 or Ftype = 2) then do;
            /* individual data */
            if (y <= x[1]+eps) then do; /* y <= x[1] */
                if (y < x[1]-eps) then /* y < x[1] */
                    return (0); /* assume F is 0 in [0,x[1]) */
                else /* y = x[1] */
                    return (F[1]);
            end;

            if (y > x[dim]+eps) then /* y > x[dim] */
                return (F[dim]); /* assume that F is flat after last point */

            /* at this point, dim > 1 and y is in (x[1], x[dim]] interval */
            edf = .;
            ileft = 2;
            iright = dim;
            i = int((dim+2)/2);
            do while (i >= 2 and i <= dim);
                if (y < x[i]-eps) then do;
                    if (y >= x[i-1]+eps) then do; /* x[i-1] <= y < x[i] */
                        edf = F[i-1]; /* F is F[i-1] in [x[i-1],x[i]) */
                        go to doneEDF;
                    end;
                    /* y < x[i-1]; move search to left half of current interval
*/
                    iright = i-1;
                end;
                else do;
                    if (y <= x[i]+eps) then do; /* y = x[i] */
                        edf = F[i]; /* F jumps to F[i] at x[i] */
                        go to doneEDF;
                    end;
                    /* y > x[i]; move search to right half of current interval
*/
                    ileft = i+1;
                end;
                i = int((ileft+iright)/2);
            end;
        end;
        else do;
            /* grouped data */
            if (y <= x[1]+eps) then do;
                /* y <= x[1] */
                if (y < x[1]-eps) then /* y < x[1] */
                    return (0); /* assume F is 0 in [0,x[1]) */
                else do;
                    /* y = x[1] */
                    if (1 < dim) then do;
                        if (x[2] <= x[1]+eps) then
                            edf = F[2];
                        else
                            edf = F[1];
                    end;
                    else
                        edf = F[1];
                    go to doneEDF;
                end;
            end;

            /* at this point, dim > 1 and y is in (x[1], x[dim]] interval */
            if (y > x[dim]+eps) then /* y > x[dim] */
                return (F[dim]); /* assume that F is flat after last point */

            edf = .;
            ileft = 2;
            iright = dim;
            i = int((dim+2)/2);

            do while (i >= 2 and i <= dim);
                if (y < x[i]-eps) then do;
                    if (y > x[i-1]+eps) then do; /* x[i-1] < y < x[i] */
                        edf = F[i-1] + (y-x[i-1])*(F[i]-F[i-1])/(x[i]-x[i-1]);
                        go to doneEDF;
                    end;
                    /* y <= x[i-1]; move search to left half of current interval
*/
                    iright = i-1;
                end;
                else do;
                    if (y <= x[i]+eps) then do; /* y = x[i] */
                        if (i < dim) then do;
                            if (x[i+1] <= x[i]+eps) then
                                edf = F[i+1];
                            else
                                edf = F[i];
                        end;
                        else
                            edf = F[i];
                        go to doneEDF;
                    end;
                    /* y > x[i]; move search to right half of current interval
*/
                    ileft = i+1;
                end;
                i = int((ileft+iright)/2);
            end;
        end;
        doneEDF:
            return(edf);
     endsub;
   run;

   proc proto package = &cmplib..FFT;
      %include "&UCM_PATH.orva_proto.sas";
   run;
%mend orva_init_global_functions;
