/*****************************************************************************************
 * Copyright (C), 2011
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   NAME:    ORVA_insurance_set_effect.sas                                            
 *   PURPOSE: Determines how the insurance will be applied before any 
 *            simulation actually gets done.
 *
 *   SUPPORT: jariou            
 *
 *   PARAMETERS:
 *
 *    COVERAGEDATA1 = The name of the dataset containing primary
 *                    insurance coverage probabilities.
 *
 *    COVERAGEDATA2 = The name of the dataset containing secondary
 *                    insurance coverage probabilities.
 *
 *    INSURANCEDATA = The name of the dataset containing the insurance accounts.
 *
 *    SIMDATE       = The simulation date.
 *
 *    VERSION       = current data version
 *
 *    CCUID         = The cell number for the simulation.
 *
/*---------------------------------------------------------------------------------*/
%macro orva_insurance_set_effect(
                                 CoverageData1  = ,
                                 CoverageData2  = ,
                                 InsuranceData  = ,
                                 simDate        = ,
                                 Version        = ,
                                 CCUID          = 
                                 );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_insurance_set_effect(;
      %put .                           CoverageData1  = &CoverageData1.,;
      %put .                           CoverageData2  = &CoverageData2.,;
      %put .                           InsuranceData  = &InsuranceData.,;
      %put .                           simDate        = &simDate.,;
      %put .                           Version        = &Version.,;
      %put .                           CCUID          = &CCUID.;
      %put .                           );
   %end;

   %Let CoverageCount          = 0;
   %let pInsuranceCount        = 0;
   %Let InsuranceContractCount = 0;
   %Let pInsurance             = %str();
   %Let InsuranceArray         = %str();

   %Let Insurance_UID_1        = %str();
   %Let Insurance_Index_1      = %str();
   %Let Deductible_1           = %str();
   %Let The_Limit_1            = %str();
   %Let Eff_Agg_Limit_1        = %str();
   %Let Payout_Haircut_1       = %str();
   %Let PolicyExpiryDate_1     = %str();

   %Let Insurance_UID_2        = %str();
   %Let Insurance_Index_2      = %str();
   %Let Deductible_2           = %str();
   %Let The_Limit_2            = %str();
   %Let Eff_Agg_Limit_2        = %str();
   %Let Payout_Haircut_2       = %str();
   %Let PolicyExpiryDate_2     = %str();

   /******************************************************/
   /* Put simulation date into a numerical format: BEGIN */
   /******************************************************/
   %if "&simDate" eq "" %then %do;
      %let simDate = %sysfunc(today());
   %end;

   data _NULL_;
      format SimDate BEST12.;
      SimDate = &simDate;
      call symputX("simDate", SimDate);
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to put simulation date into a numerical format.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to put simulation date into a numerical format.;
   %end;

   /******************************************************/
   /* Put simulation date into a numerical format: END   */
   /******************************************************/
   %if "&coverageData1" ne "" AND "&coverageData2" ne "" %then %do;
      %orvl_ds_exists(
                      &CoverageData1, 
                      "Ins", 
                      false
                      );
      %let cov1 = &ds_nobs;
      %orvl_ds_exists(
                      &CoverageData2, 
                      "InsCov", 
                      false
                      );

      %if (&cov1 ne 0 and &ds_nobs ne 0) %then %do;
         %_orva_insuranceSetEffect(
                                    CoverageData1  = &coverageData1,
                                    CoverageData2  = &coverageData2,
                                    InsuranceData  = &InsuranceData,
                                    simDate        = &simDate,
                                    Version        = &Version,
                                    CCUID          = &CCUID 
                                    );
         %end;
         %else %do;
            %let coverageData1 = ;
            %let coverageData2 = ;
            %let CoverageCount = 0;
      %end;
   %end;
%mend orva_insurance_set_effect;

%macro _orva_insuranceSetEffect(
                                 CoverageData1  = ,
                                 CoverageData2  = ,
                                 InsuranceData  = ,
                                 simDate        = ,
                                 Version        = ,
                                 CCUID          = 
                                 );

   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orva_insuranceSetEffect(;
      %put .                          CoverageData1  = &CoverageData1,;
      %put .                          CoverageData2  = &CoverageData2,;
      %put .                          InsuranceData  = &InsuranceData,;
      %put .                          simDate        = &simDate,;
      %put .                          Version        = &Version,;
      %put .                          CCUID          = &CCUID;
      %put .                          );
   %end;

   %let InsuranceContractCount = 0;

   data _tmp_InsuranceData_ ;
      length UID $8;
      set &InsuranceData end = last;
      where Version_UID = &Version;
      UID = Insurance_UID;

      %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
         if _N_ = 1 then do;
            put "           ------------------------------------------------------------------";
         end;

         put "           | Insurance Name : "  Insurance_Name;
         put "           | Company Name   : "  Insurance_Company_Name;
         put "           | UID            : "  Insurance_UID;
         put "           -----------------------------------------------------------------";
      %end;

      call symputX("InsuranceContractCount", _N_);
      Insurance_Index = _N_;
   run;
   %if &InsuranceContractCount eq 0 %then %do;
      %let CoverageCount = 0;
      %goto DONE_INS_EFFECT;
   %end;

   proc sort data = _tmp_InsuranceData_;
      by insurance_uid;
   run;

   proc SQL noprint;
      SELECT
               "Insurance_" || compress(UID)
      INTO
               :InsuranceArray SEPARATED BY " "
      FROM
               _tmp_InsuranceData_;
               ;
      CREATE TABLE 
                     VARPRJ.Ins_Contracts
      AS SELECT
                     "Insurance_" || compress(UID) AS Insurance
      FROM
                     _tmp_InsuranceData_
                     ;
   quit;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to create the InsuranceArray macro variable.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to create the InsuranceArray macro variable.;
      %goto DONE_INS_EFFECT;    
   %end;

   /*************************************************************/ 
   /* Compute aggregate probability of coverage                 */
   /* i.e. product of coverage by insurance 1 and insurance 2   */
   /*************************************************************/ 
   %let CoverageCount = 0;

   /*----- come up with CoverageProp1 and CoverageProp2 -----*/
   proc sort 
            data = &coverageData1 
            out  = _ins1
            ; 
      where CCUID = &CCUID; 

      by 
         CCUID 
         Insurance_1
         ; 
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to sort &coverageData1.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to sort &coverageData1.;
      %goto DONE_INS_EFFECT;    
   %end;

   proc sort 
            data = &coverageData2 
            out  = _ins2
            ; 
      where CCUID = &CCUID; 
      by 
         CCUID 
         Insurance_1
         ; 
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to sort &coverageData2.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to sort &coverageData2.;
      %goto DONE_INS_EFFECT;    
   %end;
   
   data _AllcoverageDatar;
      merge _ins1 _ins2;
      by CCUID Insurance_1;

      if missing(PInsCurrent2) then do;
         Insurance_2  = 0;
         PInsCurrent2 = 1;
      end; 

      PInsurance = PInsCurrent1 * PInsCurrent2;

      if missing(PInsurance) then PInsurance = 0;

      keep 
            Insurance_1 
            Insurance_2 
            PInsurance 
            CoverageProp
            ;
      call symputx("CoverageCount",_N_);
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_compute_agg_pvcovg_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_compute_agg_pvcovg_error, noquote));
      %goto DONE_INS_EFFECT;    
   %end;

   data _AllcoverageDatar;
      if _N_ eq 1 then do;
         declare hash insHash(dataset: '_AllcoverageDatar', hashexp: 5);
         rc = insHash.defineKey('Insurance_1');
         rc = insHash.defineData('CoverageProp');
         rc = insHash.defineDone();
      end;

      set _AllcoverageDatar;

      CoverageProp1    = CoverageProp;
      save_Insurance_1 = Insurance_1;
      Insurance_1      = Insurance_2;
      CoverageProp2    = insHash.find();
      Insurance_1      = save_Insurance_1;

      drop 
            CoverageProp 
            save_Insurance_1
            ;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_compute_covg_prop_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_compute_covg_prop_error, noquote));
      %goto DONE_INS_EFFECT;    
   %end;

   /*----- Add the insurance details to the coverage dataset -----*/
   data _AllcoverageDatar(
                           keep =   
                                 Insurance_1 
                                 Insurance_2 
                                 PInsurance 
                                 Insurance_Index_1
                                 Deductible_1 
                                 The_Limit_1 
                                 Eff_Agg_Limit_1 
                                 Agg_Limit_1 
                                 Payout_Haircut_1
                                 PolicyExpiryDate_1
                                 CoverageProp1
                                 CoverageProp2
                                 );

      merge 
            _AllcoverageDatar(in = ini) 
            _tmp_InsuranceData_(rename=
                                       (
                                          insurance_uid     =  insurance_1 
                                          Insurance_index   =  insurance_index_1
                                          Deductible        =  Deductible_1
                                          The_Limit         =  The_Limit_1
                                          Agg_Limit         =  Agg_Limit_1
                                          Payout_Haircut    =  Payout_Haircut_1
                                          Effective_To_Date =  PolicyExpiryDate_1
                                       )
                                    );

      format PolicyExpiryDate_1 BEST12.;
      BY 
         insurance_1
         ;

      if ini;
      Eff_Agg_Limit_1 = CoverageProp1 * Agg_Limit_1;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_merge_pinfo_covginfo_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_merge_pinfo_covginfo_error, noquote));
      %goto DONE_INS_EFFECT;    
   %end;

   /*----- Sort insurance data by the second level of insurance coverage -----*/
   proc sort data = _AllCoverageDatar;
      by insurance_2;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_sort_insdat_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_sort_insdat_error, noquote));
      %goto DONE_INS_EFFECT;    
   %end;

   data _AllcoverageDatar(keep =   
                                 Insurance_1 
                                 Insurance_2 
                                 PInsurance 
                                 Deductible_1 
                                 The_Limit_1 
                                 Eff_Agg_Limit_1 
                                 Agg_Limit_1  
                                 Payout_Haircut_1 
                                 Insurance_index_1
                                 PolicyExpiryDate_1
                                 Deductible_2 
                                 The_Limit_2 
                                 Eff_Agg_Limit_2 
                                 Agg_Limit_2 
                                 Payout_Haircut_2
                                 Insurance_index_2
                                 PolicyExpiryDate_2   
                                 );
      merge 
            _AllcoverageDatar (in = ini) 
            _tmp_InsuranceData_(rename =
                                        (
                                          insurance_uid     =  insurance_2 
                                          Insurance_Index   =  Insurance_index_2
                                          Deductible        =  Deductible_2
                                          The_Limit         =  The_Limit_2
                                          Agg_Limit         =  Agg_Limit_2
                                          Payout_Haircut    =  Payout_Haircut_2
                                          Effective_to_Date =  PolicyExpiryDate_2
                                          )
                                         );
      by insurance_2;
      format 
            PolicyExpiryDate_2 BEST12.;

      if ini;
      Eff_Agg_Limit_2 = CoverageProp2 * Agg_Limit_2;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to merge coverage information with insurance data.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to merge coverage information with insurance data.;
      %goto DONE_INS_EFFECT;    
   %end;

   /*----- Count the number of insurance policies -----*/
   data _NULL_;
      set _tmp_InsuranceData_ end = last;

      if last then do;
         call symputx("InsurancePolicyCount",_N_);
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to compute the number of insurance policies.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to compute the number of insurance policies.;
      %goto DONE_INS_EFFECT;    
   %end;

   %let pInsuranceCount = 0;
   /*---- Count the number of insurance probabilities ----*/
   data _NULL_;
      set _AllcoverageDatar end = last;

      if last then do;
         call symputx("pInsuranceCount", _N_);
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to compute the number of insurance probabilities.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to compute the number of insurance probabilities.;
      %goto DONE_INS_EFFECT;    
   %end;

   proc sql noprint;
      SELECT 
               put(pInsurance,       BEST20.),
               Insurance_1,
               Insurance_Index_1,
               put(Deductible_1,     BEST20.),
               put(The_Limit_1,      BEST20.),
               put(Eff_Agg_Limit_1,  BEST20.),
               put(Payout_Haircut_1, BEST20.),
               PolicyExpiryDate_1,
               Insurance_2,
               Insurance_Index_2,
               put(Deductible_2,     BEST20.),
               put(The_Limit_2,      BEST20.),
               put(Eff_Agg_Limit_2,  BEST20.),
               put(Payout_Haircut_2, BEST20.),
               PolicyExpiryDate_2
      INTO 
               :pInsurance          SEPARATED BY ", " ,
               :Insurance_UID_1     SEPARATED BY " "  ,
               :Insurance_Index_1   SEPARATED BY " "  ,
               :Deductible_1        SEPARATED BY " "  , 
               :The_Limit_1         SEPARATED BY " "  ,
               :Eff_Agg_Limit_1     SEPARATED BY " "  ,
               :Payout_Haircut_1    SEPARATED BY " "  ,
               :PolicyExpiryDate_1  SEPARATED BY " "  ,
               :Insurance_UID_2     SEPARATED BY " "  ,
               :Insurance_Index_2   SEPARATED BY " "  ,
               :Deductible_2        SEPARATED BY " "  ,
               :The_Limit_2         SEPARATED BY " "  ,
               :Eff_Agg_Limit_2     SEPARATED BY " "  ,
               :Payout_Haircut_2    SEPARATED BY " "  ,
               :PolicyExpiryDate_2  SEPARATED BY " "  
      FROM 
               _AllcoverageDatar
               ;
   quit;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error Unable to populate the insurance coverage macros variables for this cell.;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = Unable to populate the insurance coverage macros variables for this cell.;
      %goto DONE_INS_EFFECT;    
   %end;

%DONE_INS_EFFECT:
   %orvu_delete_data(data = 
                           _ins: 
                           _AllcoverageDatar
                           );
%mend _orva_insuranceSetEffect; 
