/*****************************************************************************************
 * Copyright (C), 2012
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   NAME:    ORVA_Loss_Sim.sas                                            
 *   PURPOSE: Simulate aggregate loss from frequency and severity distributions 
 *            This IS THE FASTER MC SIMULATION VERSION (b)!!!! 
 *   SUPPORT: jariou            
 *
 *
 * PARAMETERS:
 *
 * SPARMPRIORITYDATA = the name of the priority parameter data set with the following variables:
 *                     dataset_name...
 *
 * LOSSKRIDATA   =  Dataset containing losses with KRI info
 *
 * KRIWEIGHTS =  Dataset containg ccUID, BL_UID, and the proportions
 *                  of losses of each BL_UID by CCUID
 *
 * freqSimData   = The data set containing the frequency draws from orva_freq_sim.sas (required in 3.2)
 *
 * CCUID         = The cell number for the simulation.
 * 
 * ccData        = Project's Cross Classification dataset.
 * 
 * seedData      = Dataset of seeds for each cell CCUID.
 *
 * confIntData   = The output file from the orva_cisetup macro.  Contains the indices into the 
 *                 order stats for the percentile point estimates and conf intervals.
 *
 * rareScenSim   = output from orva_rare_scen_sim().  Contains all the rare events scenarios
 *                 that are part of the project.
 *                 
 * selectedScenUids = subset of the project's rare event scenarios to include in the simulation
 *
 * outSimNaked   =  the name of the replication output data set with the following variables:
 *                    _REP_ and LOSS. (W/O insurance)
 *
 * outsiminsured  =  the name of the replication output data set. (W/ insurance)
 *
 * OUTSTATNaked  =  the name of the output statistics data set with the following variables:
 *                  STATUS, REASON, FDIST, SDIST, MEAN, NPCTL, PCTL1-PCTLn, P_1-P_n,
 *                  LCL1-LCLn, UCL1-UCLn, and COVERAGE1-COVERAGn. (W/O insurance)
 *
 * OUTSTATInsured   =  the name of the output statistics data set. (W/ insurance)
 *
 *  parallel        = 1: use parallel processing
 *                    0: don't use parallel processing
 *
 ***********************************************************************************/
/*                  Other helper macros in this file                               */
/*                                                                                 */
/*    %macro orv_handleBucketScenario(baseData=, counter=);                            */
/*    %macro orv_handleModelScenario(baseData=,counter=);                              */
/*    %macro orv_handleModelCase(DistNameVar=, ScaleEffect=);                          */
/*    %macro orv_handleKRIs();                                                         */
/*    %macro orv_createMainLookupTable;                                                */
/*                                                                                 */
/***********************************************************************************/
/*         Main list of macro variables created and/or used in this file           */ 
/*---------------------------------------------------------------------------------*/
/* Source Types:  I   Internal, i.e. model parameters and KRIs                    */
/*                                                                                 */
/*                F   Fit Scenario, i.e. model parameters                          */
/*                M   Mixed, i.e. model parameters                                 */
/*                N   Mixed Scenario, i.e model parameters                         */
/*                P   Public, i.e. model parameters                                */
/*                C   Consortium Fit, i.e. model parameters                        */
/*                T   Consortium Tool Model, i.e. parameters and KRIs              */
/*                                                                                 */
/*                S   Scenario proxy                                               */
/*                R   Rare Event Scenario                                          */
/*                                                                                 */
/*---------------------------------------------------------------------------------*/
/*   Scenario Type:  M   Model Based  Senario                                      */
/*                   B   Buckect Based Scenario                                    */
/*---------------------------------------------------------------------------------*/
/*                      List of macro variables available                          */
/*---------------------------------------------------------------------------------*/
/*                                                                                 */
/*    sParmPriorityData = Dataset with list of priorities for severity simulation  */
/*      SOURCE_TYPE1-nMixIntervals               =  Source_Type                    */
/*      SOURCE_NAME1-nMixIntervals               =  Source_Name                    */
/*      DATASET_NAME1-nMixIntervals              =  Dataset_name                   */
/*      SECONDARY_DATASET_NAME1-nMixIntervals    =  Secondary Dataset Name         */
/*      SCENARIO_TYPE1-nMixIntervals             =  Scenario Type                  */
/*      LOWER_BOUND1-nMixIntervals               =  Lower Bound                    */
/*      UPPER_BOUND1-nMixIntervals               =  Upper Bound                    */
/*---------------------------------------------------------------------------------*/
/*                                                                                 */
/*    nReps             = Number of simulations to run                             */
/*    seedData          = Name of seed dataset                                     */
/*    seed              = The seed value for this run                              */
/*    outStatInsured    = Dataset name of simulation with insurance applied        */
/*    sevParmData       = Dataset name of severity parameters                      */
/*    MAXLOSS           = Maximum simulated loss amount for this cell              */
/*    nMixIntervals     = Number of mixing intervals in the simulation             */
/*    nRareEventScen    = Number of rare event scenarios                           */
/*---------------------------------------------------------------------------------*/
%macro ORVA_Loss_Sim(   
                        sParmPriorityData  =  ,  
                        insuranceMethod    =  ,
                        lossKriData        =  ,
                        lossKriDataC       =  ,
                        freqSimData        =  ,
                        ccUid              =  ,
                        ccData             =  ,
                        seedData           =  ,
                        lookupSize         =  1000,
                        outsimNaked        =  outsimNaked,     
                        outsimInsured      =  outsimInsured,
                        outstatNaked       =  outstatNaked,   
                        outstatInsured     =  outstatInsured,
                        ConfIntData        =  ,
                        parallel           =  0,
                        nstreams           =  1,
                        cpiAdjustment      =  1.0,
                        rareScenSim        =  ,
                        selectedScenUids   = ,
                        combineMethod      = MAXIMUM
                        );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #ORVA_Loss_Sim(;   
      %put .               sParmPriorityData  = &sParmPriorityData.,;  
      %put .               insuranceMethod    = &insuranceMethod.,;
      %put .               lossKriData        = &lossKriData.,;
      %put .               lossKriDataC       = &lossKriDataC.,;
      %put .               freqSimData        = &freqSimData.,;
      %put .               ccUid              = &ccUid.,;
      %put .               ccData             = &ccData.,;
      %put .               seedData           = &seedData.,;
      %put .               lookupSize         = &lookupSize.,;
      %put .               outsimNaked        = &outsimNaked.,;     
      %put .               outsimInsured      = &outsimInsured.,;
      %put .               outstatNaked       = &outstatNaked.,;   
      %put .               outstatInsured     = &outstatInsured.,;
      %put .               ConfIntData        = &ConfIntData.,;
      %put .               parallel           = &parallel.,;
      %put .               nstreams           = &nstreams.,;
      %put .               cpiAdjustment      = &cpiAdjustment.,;
      %put .               rareScenSim        = &rareScenSim.,;
      %put .               selectedScenUids   = &selectedScenUids.,;
      %put .               combineMethod      = &combineMethod.;
      %put .               );
   %end;

   /*- create the status data set upfront -*/
   %let ORV_REASON = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( 
                     statusds = &outstatNaked,   
                     status   = 1, 
                     reason   = &ORV_REASON 
                     );
   %orva_status_ds( 
                     statusds = &outstatInsured, 
                     status   = 1, 
                     reason   = &ORV_REASON 
                     );
   
   %let ORV_ABORT  = &ORV_FALSE;
   %let status     = 0;
   %let ORV_REASON = %sysfunc(sasmsg(&orvads, successful_note, noquote));
  
   %global PARAM_COUNT;
   %global TOTAL_SCEN_MEAN;

   %let KRI_ADJUST         = 0;
   %let fdist              = ;
   %let sdist              = ;
   %let PROXY_RUN_DONE     = No;  
   %let AVERAGE_KRI_EFFECT = 1;
   %let UPPER_BOUNDS       = %str();
   %let HandledKRIs        = No;
 
   /*- error checking -*/
   /* sParmPriorityData must be specified */
   %if ("&sParmPrioritydata" EQ "") %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_specify_sparmprioritydata_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_missing_SPARMPRIORITYDATA_error, noquote));
      %goto DONE;
   %end;
   
   /*****************************************/
   /* Get the number of replications: BEGIN */ 
   /*****************************************/
   /* Set the macro variable nreps to know how many simulations to run */
   data _null_;
        set &ConfIntData;
        call symputX("nreps", nreps);
   run;
 
   /* The number of simulations must be positive */
   %if &nreps <= 0 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_nreps_be_positive, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_nreps_be_positive, noquote));
      %goto DONE;
   %end;

   /* nreps must be an integer */
   %if %sysevalf(&nreps, integer) ^= &nreps %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_nreps_be_positive, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_nreps_be_positive, noquote));
      %goto DONE;
   %end;
   /*****************************************/
   /* Get the number of replications: END */
   /*****************************************/

   /*****************************************/
   /* Get the seed information: BEGIN */
   /*****************************************/
   /* seedData must be specified */
   %if "&seedData" EQ "" %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_provide_seedData_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_missing_seedData_error , noquote));
      %goto DONE;
   %end;
   /*****************************************/
   /* Get the seed information: END */
   /*****************************************/
   
   %orv_getFrequencySim(
                           theCCUID = &ccuid, 
                           freqData = &freqSimData
                           );    
   %if &ORV_ABORT %then %do;
      %GOTO DONE;
   %end;

   proc sql noprint;
      SELECT
               count(Source_Type) 
      INTO
               :The_Source_Type_Count
      FROM
               &sParmPriorityData
      WHERE
               CCUID = &ccuid
            AND
               Source_Type NE "R"
      ;
      SELECT
               Source_Type
      INTO
               :The_Source_Type SEPARATED BY " "
      FROM
               &sParmPriorityData
      WHERE
               CCUID = &ccuid
      AND
               Source_Type NE "R"
               ;
   quit;

   %if &The_Source_Type_Count = 1 %then %do;
      %if &The_Source_Type EQ S %then %do;
         %if %upcase(&Use_42_Scen_Calc) = YES %then %do;
            %orv_checkWhatKindOfSim();
            %GOTO DONE_WITH_BASE_SIM;
         %end;
      %end;
   %end;

   /* Severity simulation preparation: START */
   /* It creates _SevParmsProxy_ */
   %orv_createMainLookupTable(
                              ccuid          = &CCUID, 
                              ccData         = &ccData, 
                              _combineMethod = &combineMethod
                              );
   %if &ORV_ABORT %then %do;
      %GOTO DONE;
   %end;

   %local this_is_a_total_cell;
   %orvu_is_this_a_total_cell(
                                 CCUID          = &CCUID,
                                 ccData         = &ccData,
                                 om_isTotalCell = this_is_a_total_cell
                                 );

   /* Severity simulation preparation: END */

   /*- Monte Carlo simulation -*/
   /************************ Parallelized loop ***************************/
   data _NULL_;
      now = datetime();
      call symputX('_TimeStart_',now);
   run;

   %let AllReps = %str();
   %global ORVP_STATUS;
   %let ORVP_STATUS = 1;

   %if ("&parallel" eq "1") %then %do;
      %orvp_distribute_loss_sim(
                                 nreps        = &nreps,
                                 nstreams     = &nstreams,
                                 seedData     = &seedData,
                                 isTotalCell  = &this_is_a_total_cell
                                 );

      %if (&ORVP_STATUS eq 0) %then %do;
         data &outsimNaked;
            set _osimNaked_;
         run;

         data &outsimInsured;
            set _osimInsured_;
         run;
      %end;
      %else %do;
         %put &orvar_warning %sysfunc(sasmsg(&orvads, losim_no_procs_for_paralstim_warning, noquote));
      %end;
   %end;

   %if (&ORVP_STATUS ^= 0) %then %do;
      data _NULL_;
         reps = ceil(&nreps/&nstreams);
         call symputX('REPS', reps);
      run;

      %let AllReps   = &reps;
      %let CumReps   = &reps;
      %let NITER     = 1;
      %let STARTREP  = 1;
      %let NREPSLEFT = &nreps;

      %do %while (&NREPSLEFT gt 0);
         %let SEED = 0;

         data _NULL_; 
            set &seedData; 
            WHERE 
                  CCUID eq &ccUID 
            AND
                  HOST eq &NITER
                  ;
            length tseed $ 32;
            tseed = put(seed,32.);
            call symputX("SEED", strip(tseed));
         run;

         %orva_loss_sim_stream(
                                 startRep        = &STARTREP,
                                 nreps           = &REPS,
                                 streamNum       = &NITER,
                                 seed            = &SEED, 
                                 isTotalCell     = &this_is_a_total_cell,
                                 insuranceMethod = &insuranceMethod
                                 );

         %orvu_append(
                        base         = &outsimNaked,
                        new          = _osimNaked_&NITER,
                        canUseAppend = true
                        );     

         %orvu_append(
                        base         = &outsimInsured,
                        new          = _osimInsured_&NITER,
                        canUseAppend = true
                        );     

            
         %orvu_delete_data( data = 
                                    _osimNaked_&NITER 
                                    _osimInsured_&NITER
                                    );
            
         %let STARTREP  = %eval(&STARTREP + &REPS);
         %let NREPSLEFT = %eval(&NREPSLEFT - &REPS);

         %if &NREPSLEFT lt &REPS %then %do;
            %let REPS = &NREPSLEFT;
         %end;

         %let CumReps = %eval(&CumReps + &reps);
         %let AllReps = &AllReps &CumReps;
         %let NITER   = %eval(&NITER + 1);
      %end;
   %end;

   %orv_getRawSummary(
                        simData    = &outsimNaked, 
                        outSummary = raw_summary
                        );
%DONE_WITH_BASE_SIM:
   /* Handle the addition of rare event scenarios to the original */
   /* simulation datasets                                        */
   %if &rareScenSim NE %str() and %trim(%left("&selectedScenUids")) ne "" %then %do;
      /* Keep only the data generated for this CCUID */
      /* ANd then only for the selected Scenarios */

      data _RareScenSimThisCCUID_;
         set &rareScenSim;
         WHERE 
               (CCUID = &CCUID) 
         AND 
               (SCENARIO_UID in (&selectedScenUids))
               ;
      run;

      %let nRareSims = 0;
      %orvu_getObsCount(_RareScenSimThisCCUID_, nRareSims);

      %if &nRareSims GT 0 %then %do;
         proc sort 
                  data = _RareScenSimThisCCUID_ 
                  out  = _rareScenSimThisCCUID_sorted_
                  ;
            by _rep_;
         run;
         
         /* Take the sum (by _rep_) of the rare event scenarios selected */
         proc summary 
                     data = _rareScenSimThisCCUID_sorted_;
            var loss_amount;
            by _rep_;

            output
                  sum = rare_loss
                  out = _rareSumsThisCCUID_(keep = _rep_ rare_loss )
                  ;
         run;

         /* Add rare event scenarios to the Naked simulation */
         data &outsimNaked;
            merge 
                  &outsimNaked  
                  _rareSumsThisCCUID_
                  ;
            by _rep_;
            loss = loss + rare_loss;
            keep _rep_ loss;
         run;
      
         /* Add rare event scenarios to the Insured simulation */
         data &outsimInsured;
            merge 
                  &outsimInsured  
                  _rareSumsThisCCUID_
                  ;
            by _rep_;
            loss = loss + rare_loss;
            keep _rep_ loss Insurance_:;
         run;
      %end; /* There is something in the rare event scenario dataset */
   %end; /* A rare event scenario dataset is passed in */

   data _NULL_;
      now     = datetime();
      elapsed = now - symget('_TimeStart_');
      call symput('_TimeElapsed_', elapsed);
   run;
   %put %sysfunc(sasmsg(&orvads, losim_loss_sim_time_note, noquote, &_TimeElapsed_));

   /************************ Parallelized loop ***************************/
   %orva_get_var_stats(
                        simData          = &outsimNaked, 
                        simCount         = &nreps,
                        outStat          = &outStatNaked, 
                        lossVar          = loss,
                        confIntData      = &confIntData,
                        includeInsurance = NO
                        );   

   %orva_compute_expected_shortfall(
                                    simData = &outsimNaked, 
                                    outstat = &outstatNaked, 
                                    nreps   = &nreps 
                                    );

   data &outStatNaked;
      set &outStatNaked;
      set raw_summary;
   run; 

   %orva_get_var_stats(
                        simData          = &outsimInsured, 
                        simCount         = &nreps,
                        outStat          = &outstatInsured, 
                        lossVar          = loss,
                        confIntData      = &confIntData,
                        includeInsurance = YES                          
                        );  

   %orva_compute_expected_shortfall(
                                    simData = &outsimInsured, 
                                    outstat = &outstatInsured, 
                                    nreps   = &nreps 
                                    );
%DONE:
   %if &ORV_ABORT %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_abnormal_error, noquote));
      %put &ORV_REASON;
      %orva_status_ds( 
                        statusds = &outstatNaked,   
                        status   = 1, 
                        reason   = &ORV_REASON 
                        );
      %orva_status_ds( 
                        statusds = &outstatInsured, 
                        status   = 1, 
                        reason   = &ORV_REASON 
                        );
   %end;    
    
   %let HandledKRIs = No;
   %orvu_delete_data(data = 
                           _temp0_
                           _temp_
                           _loss_ 
                           _TParm_ 
                           _tmp_kri_summ_
                           _RareScenSimThisCCUID_
                           _RareScenSimThisCCUID_sorted_
                           _rareSumsThisCCUID_
                           _transposed_data_
                           );
%mend ORVA_Loss_Sim;

%macro orv_handleBucketScenario(
                                 baseData = , 
                                 counter  = , 
                                 outData  =
                                 );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_handleBucketScenario(;
      %put .                          baseData = &baseData.,; 
      %put .                          counter  = &counter.,; 
      %put .                          outData  = &outData.;  
      %put .                          );
   %end;

   data _null_;
      set &DSNAME;
      where CCUID = &CCUID;
      call symputX("NBUCKETS", NBUCKETS);
   run;

   /* All bucket scenarios are merged together at this point? */
   data _DS_CCUID_;
      set &DSNAME;
      WHERE 
            CCUID = &CCUID;
      Threshold = 0;
   run;

   data _temp_;
      set &baseData;
      call symputX("effLookupSize", _N_);
   run;

   data _lookup&counter.TMP;
      merge _temp_ _DS_CCUID_; 
      by threshold; 
   run;

   data &outData(KEEP = 
                        ecdf 
                        quant&counter. 
                        index
                        );
      set _lookup&counter.TMP;
      retain  
            quant&counter 0  
            BucketIndex   1 
            lastP         0 
            bucketProb    0 
            lastCdf       0 
            lastQ         0 
            ;

      array final_bucketwgt[&nbuckets];
      array final_bucket[&nbuckets];

      ecdf  = 1 - (1 - _N_ / (&effLookupSize + 1) )**2.5;
      index = _N_;

      if _N_ = 1 then do; 
         bucketProb = final_bucketwgt[1];
         lastQ = final_bucket0;
      end;

      do while (ecdf > bucketProb );
         bucketIndex = BucketIndex + 1;
         lastQ       = final_bucket[BucketIndex];
         bucketProb  = bucketProb + final_bucketwgt[BucketIndex];
         lastCdf     = bucketProb;
      end;

      if BucketIndex = 1 then do;
         quant&counter = (final_bucket[bucketIndex] - final_bucket0) * (ecdf - lastCdf) / final_bucketwgt[bucketIndex] + lastQ;  
      end;
      else do;
         quant&counter = (final_bucket[BucketIndex] - final_bucket[BucketIndex - 1]) * (ecdf - lastCdf) / final_bucketwgt[bucketIndex] + lastQ;  
      end;

      lastQ   = quant&counter;
      lastCdf = ecdf;
   run;
%mend orv_handleBucketScenario;

/* 
    Original method to create a lookup table for simulation. 
    Since the quantile of the mixture of the various model 
    scenarios would be a mess to obtain, I simply simulate 
    the mixture itself and order the results as approximate 
    appropriate quantiles
*/
%macro orv_handleModelScenario(
                                 baseData = , 
                                 counter  = , 
                                 outData  =
                                 );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_handleModelScenario(;
      %put .                         baseData = &baseData.,; 
      %put .                         counter  = &counter.,; 
      %put .                         outData  = &outData.;
      %put .                         );
   %end;

   data _NULL_;
      set &baseData;
      call symputX("effLookupSize", _N_);
   run;

   /* obtain the parameter dataset for that cell */
   data _DS_CCUID_;
      set &DSNAME;
      where CCUID = &CCUID;
      call symputX("modelCount",_N_);
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_nreps_be_positive, noquote, &CCUID));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_nreps_be_positive, noquote, &CCUID));
      %goto DONE;    
   %end;
    
   /* If only one line, then treat this as an internal parametric fit and get out of here*/
   %if &modelCount = 1 %then %do;
      %orv_handleModelCase( 
                              DistNameVar = Severity_Dist, 
                              ScaleEffect = 1, 
                              Counter     = &Counter,
                              outData     = &outData
                              );
      %goto DONE_MODEL_SCEN;
   %end;
    
   /* If we are here, we know that there are more than one line. */
   /* Get the frequency mean for each model scenario, as well as */
   /* the aggregated mean frequency                              */
   data _DS_CCUID_; 
      set _DS_CCUID_ end = last;
      retain TotalFreqMean 0;
      call FreqMoments(
                        Frequency_Dist, 
                        &theFreqSimCallParams, 
                        FreqMean, 
                        fm2
                        );
      TotalFreqMean = TotalFreqMean + FreqMean;

      if last then do;
         call symputX("TotalFreqMean", put(TotalFreqMean, BEST20.));
      end;

      drop fm2;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_compute_freq_mean_for_modelscen_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_compute_freq_mean_for_modelscen_error, noquote));
      %goto DONE;    
   %end;

   /* Order by value of frequency mean */
   proc sort data = _DS_CCUID_;
      by FreqMean;
   run;
    
   /* Find out how many from each scenario needs to be simulated                          */
   /* I floor the proportion and then make sure everything adds up to the desired count   */
   /* If does not add up, I add the missing count to the one with the largest proportion */
   data _DS_CCUID_;
      set _DS_CCUID_ end = Last;
      retain TotalCount 0;
      subCount   = floor(FreqMean * 20 * &lookupSize / &TotalFreqMean);
      TotalCOunt = TotalCOunt + subCOunt;

      if last then do;
         SubCount   = SubCount + (20 * &lookupSize - TotalCount);
         TotalCount = 20 * &lookupSize;
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_compute_reps_scen_num_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_compute_reps_scen_num_error, noquote));
      %goto DONE;    
   %end;

   %do modelIndex = 1 %to &modelCount;
      data tmpSParm;
         set _DS_CCUID_;

         if _N_= &modelIndex then do;
            output;
            call symput("Distribution",Severity_Dist);
         end;
      run;

      %orva_set_sev_dist(
                        distIn = &Distribution, 
                        distOut = Distribution
                        );
      %orvd_&Distribution;

      %if &modelIndex = 1 %then %do;    
         data _lookupModelScen(KEEP = quant&counter);
            set tmpSParm;
            call streaminit(12345 * &modelIndex);

            do i = 1 to SubCount;
               %sim&Distribution(quant&counter);
               output;
            end;
         run;
      %end;
      %else %do;
         data _tempQuant_(KEEP = quant&counter);
            set tmpSParm;
            call streaminit(12345 * &modelIndex);

            do i = 1 to SubCount;
               %sim&Distribution(quant&counter);
               output;
            end;
         run;
         proc append 
                     base = _lookupModelScen
                     data = _tempQuant_
                     ;
         run;
      %end;
   %end; /* of loop over modelIndex */

   proc sort data = _lookupModelScen;
      by quant&counter;
   run;

   data _lookupModelScen;
      set _lookupModelScen nobs = obscount;
      ecdfVar = _N_ / (Obscount + 1);
   run;

   %orva_compactify_lookup_array(
                                 inData     = _lookupModelScen,
                                 quantVar   = quant&counter,
                                 ecdfVar    = ecdfVar,
                                 lookUpSize = &lookUpSize,
                                 outData    = &outData
                                 );
%DONE_MODEL_SCEN:
%mend orv_handleModelScenario;

/* This macro handles creating a lookup table for a mixing interval where the  */
/* severity is described as a distribution model (either fit or scenario)      */
/* The only parameter is the name of the variable holding theseverity          */
/* distribution name.                                                          */
%macro orv_handleModelCase(
                           DistNameVar = , 
                           ScaleEffect = , 
                           Counter     = ,  
                           outData     =
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_handleModelCase(;
      %put .                     DistNameVar = &DistNameVar.,; 
      %put .                     ScaleEffect = &ScaleEffect.,;
      %put .                     Counter     = &Counter.,;  
      %put .                     outData     = &outData.;
      %put .                     );
   %end;

   /* Set the distribution name in the SDist macro */
   data _null_;
      set &DSNAME;
      where CCUID = &CCUID;
      call symput("SDIST", &DistNameVar);
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_set_sdist_macro_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_set_sdist_macro_error, noquote));
      %goto DONE_HANDLE_MODEL;    
   %end;

   %orva_set_sev_dist(
                        distIn = &sdist, 
                        distOut = sdist
                        );
    /* Initialize the severity distribution */
    %orvd_&SDIST;

   /* Create the macros Param_1, Param_2, .... needed for the */
   /* quantile calculations below                             */
   proc transpose 
                 data = &DSNAME 
                 out  = _transposed_data_
                 ;
      where CCUID = &ccuid;
   run;

   data _NULL_;
      set _transposed_data_;
      where substr(upcase(_NAME_), 1, 5 ) = "PARAM";
      call symputX(_NAME_, COL1);
   run;

   proc sql noprint;
      SELECT
            _NAME_ || put(COL1, BEST20.) 
      INTO
            :TheParams SEPARATED BY " "
      FROM
            _transposed_data_
      WHERE
            substr(upcase(_NAME_), 1, 5 ) = "PARAM"
            ;
   quit;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_get_set_these_macros_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_get_set_these_macros_error, noquote));
      %goto DONE_HANDLE_MODEL;    
   %end;

   /* Check to see  if a sample proxy dataset has been created yet */
   /* If no, then do one                                           */
   %if "&PROXY_RUN_DONE" = "No" %then %do;
      data _temp0_;
         retain &TheParams;
         call streaminit(12345);

         do i = 1 to 20 * &lookupSize;
            threshold = 0;
            %Sim&sdist(Quantile)
            output;
         end;
      run;

      proc sort data=_temp0_;
         by Quantile;
      run;

      data _temp0_;
         set _temp0_ nobs = ObsCount;
         ecdfVar = _N_ / (ObsCount + 1);
      run;

      %let PROXY_RUN_DONE = Yes;
      %orva_compactify_lookup_array( 
                                    inData     = _temp0_,
                                    quantVar   = quantile,
                                    ecdfVar    = ecdfVar,
                                    lookUpSize = &lookUpSize,
                                    outData    = _temp0_
                                    );
   %end;

   /* Get a clean copy of the _temp0_ dataset */
   data _temp_;
      set _temp0_;
      retain threshold 0;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_create_clean_copy_temp_ds_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_create_clean_copy_temp_ds_error, noquote));
      %goto DONE_HANDLE_MODEL;    
   %end;

   /* Run the quantile macro for the distribution */
   /* It expect a _temp_ dataset and writes to it */
   %Quant&SDist;

   /* Get rid of extranuous variables in _temp_ */
   data &outData(keep = 
                        ecdf 
                        quant&counter
                        );
      set _temp_(rename = (quant = quant&counter));
      quant&nInt = Quant&nInt * &ScaleEffect;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_modify_lu_ds_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_modify_lu_ds_error, noquote));
      %goto DONE_HANDLE_MODEL;    
   %end;
%DONE_HANDLE_MODEL:
%mend orv_handleModelCase;

%macro orv_handleKRIs(
                        ccuid  = , 
                        ccData =
                        );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_handleKRIs(;
      %put .                ccuid  = &ccuid.,; 
      %put .                ccData = &ccData.;
      %put .                );
   %end;

   %global KRI_CHOICES;
   %let kridata = ;

   %if "&SOURCE_TYPE" EQ "I" %then %do;
      %let kridata = &lossKriData;
   %end;
   %else %if "&SOURCE_TYPE" EQ "T" %then %do;
      %let kridata = &lossKriDataC;
   %end;
    
   /* If KRI's then compute the KRI values for this cell */
   %if ("&kridata" ne "") %then %do;   
      /* Get a list of the kri variables */
      /* See if there are KRIs */

      /* Subset the severity parm data to the CCUID of the cell */
      data _SEVPARMTMP;
         set &SevParmData;
         where CCUID = &ccUID;
         call symputX("KRI_ADJUST", COUNT_KRI);
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_subset_sev_paramds_for_cell_error, noquote, &CCUID));
         %let ORV_ABORT  = &ORV_TRUE;
         %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_subset_sev_paramds_for_cell_error, noquote, &CCUID));
         %goto DONE_HANDLE_KRIs;    
      %end;
    
      /* Check if KRI's were used in fitting this cell */
      %if %sysevalf(&KRI_ADJUST > 0, boolean) %then %do;
         %orva_get_kri_log(data = _SEVPARMTMP);

         /* Calculate the KRI_SCALE_EFFECT for each loss in the dataset */
         %let CELL_WHERE = ;
         %orvu_cell_where(
                           ccuid  = &CCUID, 
                           ccData = &ccData
                           );
         data _tmp_kri_(keep = Kri_Hash_Index KRI_SCALE_EFFECT BLUID);
            set &KriData end = last; 
            where &CELL_WHERE;
            array KRIS[*] &KRI_LIST;
            retain averageEffect 0;

            do i = 1 to dim(KRIS); 
               if KRIS[i] eq . then do;
                  KRIS[i] = 0; 
               end;
            end;

            KRI_SCALE_EFFECT = exp(&KRI_LOG);
            averageEffect = averageEffect + KRI_SCALE_EFFECT;
            Kri_Hash_Index   = _N_;

            if last then do;
               call symputX("AVERAGE_KRI_EFFECT", put(averageEffect / _N_, BEST20.));
            end;

            drop i;
         run;
         %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
            %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_calc_kri_scale_effect_error, noquote));
            %let ORV_ABORT  = &ORV_TRUE;
            %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_calc_kri_scale_effect_error, noquote));
            %goto DONE_HANDLE_KRIs;    
         %end;
                
         %let dsid = %sysfunc(open(_tmp_kri_));

         %if &dsid %then  %do;
            %let KRI_CHOICES = %sysfunc(attrn(&dsid,nobs));
            %let rc = %sysfunc(close(&dsid));
         %end;
         %else %do;
            %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_get_kri_choices_tot_num_error, noquote));
            %let ORV_ABORT  = &ORV_TRUE;
            %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_get_kri_choices_tot_num_error, noquote));
            %goto DONE_HANDLE_KRIs;    
         %end;
      %end;
      %else %do;
         /* No KRI's are used */
         %let KRI_LOG            = 0;
         %let KRI_SCALE_ARRAY    = 1;
         %let KRI_PROB_ARRAY     = 1;
         %let AVERAGE_KRI_EFFECT = 1;
      %end;
   %end;
   %else %do;
      %let KRI_ADJUST         = 0;
      %let AVERAGE_KRI_EFFECT = 1;
   %end; /* END OF "If there are KRI's */
%DONE_HANDLE_KRIs:
%mend orv_handleKRIs; 

%macro orv_createMainLookupTable(
                                 ccuid          = , 
                                 ccData         = , 
                                 _combineMethod = 
                                 );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_createMainLookupTable(;
      %put .                           ccuid          = &ccuid.,; 
      %put .                           ccData         = &ccData.,; 
      %put .                           _combineMethod = &_combineMethod.;
      %put .                           );
   %end;
 
   %global LOOKUP_QUANTILES;

   /* Parse the correct macros to combine simulation from more than one source */
   %if &_combineMethod. NE %str() %then %do;
      %orva_sim_combine_&_combineMethod ();
   %end;

   /* Set the lookup table up */
   data _temp0_;
      do i = 1 to &lookupSize;
         threshold = 0;
         output;
      end;
   run;

   /* Extract the following from sParmPriorityData: maxloss  */
   %let NOBS    = 0;
   %let MAXLOSS = 10000000000;
   
   data _SevParmsProxy_;
      set &sParmPriorityData end = LAST;
      WHERE 
            CCUID = &ccUID 
            AND 
            SOURCE_TYPE NE "R"
            ;
      retain MaximumLoss 0;
      MaximumLoss = max(MaximumLoss, max_loss);

      if last then do;
         /* If all max_loss are missing  */
         if MaximumLoss = 0 then do;
            MaximumLoss = 10000000000;
         end;
   
         call symputX("MAXLOSS",put(MaximumLoss, BEST20.));
         call symputX('NOBS',_N_);
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_get_max_likely_loss_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_get_max_likely_loss_error, noquote));
      %goto DONE_LOOKUP_MAIN;    
   %end;

   /* Compactify _SevParmsProxy_ */
   %if &nobs GT 1 %then %do;
      %orv_compactifyRanges(
                            data  = _SevParmsProxy_, 
                            out   = _SevParmsProxy_, 
                            ccuid = &CCUID
                            );
   %end;
    
   /* Check how many mixing intervals there are */ 
   %let dsid = %sysfunc(open(_SevParmsProxy_));

   %if &dsid %then  %do;
      %let nMixIntervals = %sysfunc(attrn(&dsid,nobs));
      %let rc = %sysfunc(close(&dsid));
   %end;
   %else %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_no_data_for_cell_error, noquote, &sParmPriorityData));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_no_data_for_cell_error, noquote, &sParmPriorityData));
      %goto DONE_LOOKUP_MAIN;
   %end;

   /* Check to see whether KRI's are used at all in any range */
   %let KRI_USED = No;

   data _NULL_;
      set _SevParmsProxy_;
      where SOURCE_TYPE = "I" or SOURCE_TYPE = "T";
      call symput("KRI_USED", "Yes");
      call symput("SevParmData", DATASET_NAME);
      call symput("SOURCE_TYPE", SOURCE_TYPE);       
   run;

   %if "&KRI_USED" = "Yes" %then %do;
      %orv_handleKRIs( 
                      ccuid  = &CCUID, 
                      ccData = &ccData
                      );

      %if &ORV_ABORT %then %do;
         %GOTO DONE_LOOKUP_MAIN;
      %end;
   %end;

   /*  Loop through each simulation interval    */
   /*  and set the appropriate macros            */
   %do nInt = 1 %to &nMixIntervals;
      /* Set the Severity Dataset information line per line */
      data _null_;
         Index = &nInt;
         set _SevParmsProxy_ point = Index;
         call symput( "DSNAME"        ,   DATASET_NAME  );
         call symput( "SOURCE_TYPE"   ,   SOURCE_TYPE   );
         call symput( "SCENARIO_TYPE" ,   SCENARIO_TYPE );
         call symputX( "LOWER_BOUND"   ,  put(LOWER_BOUND, BEST20.) );
         call symputX( "UPPER_BOUND"   ,  put(UPPER_BOUND, BEST20.) );
         stop;
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_get_detail_info_mixintv_error, noquote, &nInt));
         %let ORV_ABORT  = &ORV_TRUE;
         %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_get_detail_info_mixintv_error, noquote, &nInt));
         %goto DONE_LOOKUP_MAIN;    
      %end;
        
      /* Create a string of upper bounds values */
      %let UPPER_BOUNDS = &UPPER_BOUNDS &UPPER_BOUND;
      %let mixProb = %sysevalf(&UPPER_BOUND - &LOWER_BOUND); 

      /* Case where  Scenario */
      %if  "&SOURCE_TYPE" EQ "S" %then %do;
         /*Case where the scenario is of Model type */
         %if  "&SCENARIO_TYPE" EQ "M"  %then %do;
            %orv_handleModelScenario(  
                                       baseData         = _temp0_, 
                                       counter          = &nInt,
                                       outData          = _lookup&&nInt
                                       ); 
            %if &ORV_ABORT %then %do;
               %GOTO DONE_LOOKUP_MAIN;
            %end;
         %end;
         /* Case where the scenario is of type Bucketed */
         %else %do;
            %orv_handleBucketScenario( 
                                          baseData         = _temp0_, 
                                          counter          = &nInt, 
                                          outData          = _lookup&&nInt
                                          );
            %if &ORV_ABORT %then %do;
               %goto DONE_LOOKUP_MAIN;    
            %end;
         %end;
      %end; /* Case where source is scenario */  
      /* normal data case */
      %else %do;
         %if "&SOURCE_TYPE" EQ "I"  or  "&SOURCE_TYPE" EQ "T"  %then %do; 
            %let SevParmData = &DSNAME;

            /* I want to call the HandleKRIs macro only once */ 
            %if"&HandledKRIs" EQ "No" %then %do;
               %orv_handleKRIs( 
                                 ccuid  = &CCUID, 
                                 ccData = &ccData
                                 );
               %if &ORV_ABORT %then %do;
                  %goto DONE_LOOKUP_MAIN;    
               %end;

               %let HandledKRIs = Yes;
            %end;
         %end;

         %orv_handleModelCase( 
                              DistNameVar = Distribution, 
                              ScaleEffect = &AVERAGE_KRI_EFFECT,
                              counter     = &nInt,
                              outData     = _lookup&nInt
                              );

         %if &ORV_ABORT %then %do;
           %goto DONE_LOOKUP_MAIN;    
         %end;
      %end;

      %if &_combineMethod NE %str() %then %do;
         /* Create the lookup table for this mixing interval  */
         /* and merge with other mixing intervals             */
         %Stack_Data_&_combineMethod.(
                                      partialLookup      = _lookup&nInt.,
                                      partialLookupIndex = &nInt.,
                                      totalLookup        = _lookups_
                                      );
         %if &ORV_ABORT %then %do;
            %goto DONE_LOOKUP_MAIN;    
         %end;
      %end;
      %else %do;
         data _lookups_;
            set _lookup&nInt.;
            index = _N_;
         run;
      %end;
   %end; /* End of the interval looping */

   %if &_combineMethod NE %str() %then %do;
      %Final_Lookup_Table_&_combineMethod.(
                                           totalLookup    = _lookups_, 
                                           combinedLookup = _LookUp_Table_
                                           );
      %if &ORV_ABORT %then %do;
         %goto DONE_LOOKUP_MAIN;    
      %end;
   %end;
   %else %do;
      data _LookUp_Table_(keep = 
                                 ecdf 
                                 quantile
                                 );
         format quantile BEST12.;
         set _lookups_ ;
         quantile = quant1;
      run;
   %end;

   /* Try this SQL to create the lookup table macro */
   proc SQL noprint;
      SELECT 
            put(quantile, BEST20.)
      INTO
            :LOOKUP_QUANTILES SEPARATED by " "
      FROM
            _LookUp_Table_
      ;
   quit;
%DONE_LOOKUP_MAIN:
%mend orv_createMainLookupTable;

/* This macro compactifies the severity priority ranges */
/* i.e. it combines two or more consecutuve ranges if   */
/* they have the same DATASET_NAME.                     */
%macro orv_compactifyRanges(
                              data  = , 
                              out   = , 
                              ccuid =
                              );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_compactifyRanges(;
      %put .                      data  = &data.,; 
      %put .                      out   = &out.,; 
      %put .                      ccuid = &ccuid.;
      %put .                      );
   %end;

   data &out; 
      set &data; 
      where CCUID = &ccuid; 
   run;

   proc sort data = &out; 
      by LOWER_BOUND; 
   run;

   data &out; 
      set &out;
      length dsname $&ORV_DATASET_NAME_LENGTH.;
      retain dsname "";

      if dsname ne DATASET_NAME then do;
         output;
      end;

      dsname = DATASET_NAME;
      drop dsname;
    run;

    proc sort data = &out; 
        by descending LOWER_BOUND; 
    run;

    data &out; 
      set &out;
      retain lbound -1;

      if lbound ne -1 then do;
         UPPER_BOUND = lbound;
      end;
      else do;
         UPPER_BOUND = 1;
      end;

      lbound = LOWER_BOUND;
      drop lbound;
    run;

    proc sort data = &out; 
        by LOWER_BOUND; 
    run;
%mend orv_compactifyRanges;

%macro orv_getRawSummary(
                           simData    = , 
                           outSummary = 
                           );

   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_getRawSummary(;
      %put .                   simData    = &simData.,; 
      %put .                   outSummary = &outSummary.;
      %put .                   );
   %end;

   data &outSummary ;
      set &simData  end = last;
      where _rep_ in (&AllReps);
      retain 
               GrandTotalLoss   0 
               GrandTotalLoss2  0 
               GrandTotalCount  0 
               GrandTotalCount2 0
               ;
      GrandTotalLoss = GrandTotalLoss + TotalLoss; 
      GrandTotalLoss2 = GrandTotalLoss2 + TotalLoss2;
      GrandTotalCount = GrandTotalCount + TotalCount;
      GrandTotalCount2 = GrandTotalCount2 + TotalCount2;

      if last then do;
         SampleSize = _rep_;
         RawLossMean  = GrandTotalLoss  / GrandTotalCount;
         RawLossStd   = sqrt(GrandTotalLoss2 / GrandTotalCount - RawLossMean * RawLossMean);
         RawCountMean = GrandTotalCount / _rep_;
         RawCountVar  = GrandTotalCount2/ _rep_ - RawCountMean * RawCountMean;
         output;
      end;

      keep 
            RawLossMean 
            RawLossStd 
            RawCountMean 
            RawCountVar
            ;
   run;

   data &simData;
      set &simData;
      keep _rep_ loss;
   run;
%mend orv_getRawSummary;

%macro orv_getFrequencySim(
                           theCCUID = , 
                           freqData =
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_getFrequencySim(;
      %put .                     theCCUID = &theCCUID.,; 
      %put .                     freqData = &freqData.;
      %put .                     );
   %end;

   /*-     Obtain the name of the column of frequencies          -*/
   %local _CELL;
   data _NULL_;
      CELL = compress('Cell_'|| put(&theCCUid, &ORVAR_CELLFMT));
      call symputX('_CELL', CELL, L);
   run;

   %local cell_found;
   %let cell_found = false;

   data _loss_(keep = Freq_Hash_Index nLoss);
      set &freqData end = last;   
      /* There should be a column in &freqData for each cell in the VaR matrix */
      nloss = &_CELL;
      Freq_Hash_Index = _N_;

      if missing(nloss) then do;
         call symputX("ORV_ABORT", &ORV_TRUE );
         call symput("ORV_REASON", "%sysfunc(sasmsg(&orvads, losim_freqsim_missing_error, noquote, &_CELL))" );
         stop;
      end;

      if last then do;
         call symput("cell_found ", "true");
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_cant_get_simulated_freqs_error, noquote));
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_cant_get_simulated_freqs_error, noquote));
   %end;
   %if &ORV_ABORT %then %do;
      %GOTO GFS_DONE;
   %end;
    
   /*- error when cell isn't even in correlation matrix -*/
   %if &cell_found eq false %then %do;
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON = %sysfunc(sasmsg(&orvads, losim_freqsim_missing_error, noquote, &_CELL));
   %end;
%GFS_DONE:
%mend orv_getFrequencySim;

%macro orv_getScenarioSim(
                              theCCUID =, 
                              scenData =, 
                              outData  = 
                              );    
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_getScenarioSim(;
      %put .                    theCCUID = &theCCUID.,; 
      %put .                    scenData = &scenData.,; 
      %put .                    outData  = &outData.; 
      %put .                    );    
   %end;

   data &outData;
      set &scenData;
      where CCUID = &theCCUID;
      rename Loss_Amount = Loss;
   run;
%mend orv_getScenarioSim;

/*****************************************/
/* Get the proxy scenario simulation     */
/* result if appropriate:         BEGIN  */
/*****************************************/
%macro orv_checkWhatKindOfSim();
   %let dsid = %sysfunc(open(VARPRJ.Proxy_Sim, I));

   %if &dsid ^= 0 %then %do;
      %let rc   = %sysfunc(close(&dsid));
      %orv_getScenarioSim(
                              theCCUID = &ccuid, 
                              scenData = VARPRJ.Proxy_Sim,
                              outData  = &outSimNaked
                              );    

      /* In this case, skipping raw summary calculation and putting missing values */
      data raw_summary ;
         RawLossMean  = .;
         RawLossStd   = .;
         RawCountMean = .;
         RawCountVar  = .;
      run;

      /* This also is ususally handled by orv_getRawSummary */
      data &outsimNaked;
         set &outsimNaked;
         keep _rep_ loss;
      run;

      data &outsimInsured;
        set &outsimNaked;
      run;
   %end;
%mend orv_checkWhatKindOfSim;
