/*****************************************************************************************
 *  Copyright (C), 2010
 *  SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_map_scenario_ccv_dim.sas                                             
 *  Purpose: If the scenario effect is applied to CCV properties, Merges the Dataset with the 
 *           Cross Classification Variables with the project Mapping Datasets. If the scenario
 *           effect is applied to CCV, use the uid of dimentions(such as uids of internal business 
 *           and risk event to find the ccvuid.  
 *
 *  SUPPORT: xiaoma
 *
 *  PARAMETERS:
 *
 *  numCCVs = number of Cross Classification variables, defaults to 2.
 *
 *  data  =  the name of the "raw" loss dataset. It should contain the
 *           original Cross Classification columns from the Relational Database. 
 *
 *  mapCols  = space-separated list of Cross Classification columns (left-hand side), 
 *             ex: INTERNAL_BL_UID INTERNAL_RC_UID.
 *             The special case of __TOTALS__ONLY__ signifies that the mapping can be
 *             ignored for that dimension.  In that case we map all values to the total
 *             for that dimension.
 *
 *  ccData = the name of the Cross Classification UID dataset with the following variables:
 *           CCUID, CCVUID_<n> (one for each Cross Classification)
 *
 *  mapDataPrefix = prefix for the name of the mapping datasets. Should be one
 *                  for each mapping called &mapDataPrefix<n>.
 * 
 *  dataOut = the name of the output dataset (VaR Loss Dataset).
 *
 *  ostat = name of the out status dataset.  Will contain the following:
 *           status = 0 on success, 1 otherwise.
 *           reason = printable message.
 *  CTool - Are we in tthe Consortium Tool?
 *****************************************************************************************/
%macro orva_map_scenario_ccv_dim(
                                  numCCVs       = 2,
                                  data          = ,
                                  mapCols       = ,
                                  ccData        = ,
                                  mapDataPrefix = ,
                                  dataOut       = ,
                                  oStat         = status,
                                  missingRanges = false,
                                  CTool         = false
                                  );

    /*- create the status dataset up front with Unknown status -*/
    %let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
    %orva_status_ds( statusds=&oStat, status=1, reason=&reason );

    %let status=0;
    %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote)) ;
    %let abort=0;

    /*- check input parameters -*/
    %if &data=%str() or &ccData=%str() or &mapCols=%str() or &mapDataPrefix=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, ccm_specify_these_fields_error, noquote)) ;
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, ccm_missing_parameters_error, noquote)) ;
        %goto DONE;
    %end;      
   
    %let mapCols=%upcase(&mapCols);    

    /*- setup mapping columns in dataset: CCMAP_&i -*/    
    data _work(compress=binary); set &data;      
        %let nmaps=1;
        %let map=%scan(&mapCols, &nmaps, ' ');
        %do i=0 %to &numCCVs-1;      
            %if &map eq __TOTALS__ONLY__ %then %do;
                CCMAP_&i = .;
            %end;
            %else %do;  
                CCMAP_&i = &map;
            %end;
            %let nmaps=%eval(&nmaps+1);
            %let map=%scan(&mapCols, &nmaps, ' ');
        %end;        
    run;

    /* in scenario_details, if the scenario effect is on the CCV, then the values of mapCols of type_cd=V   
     * are the CCV_UID values. Then from the ccv_values dataset, we can find the ccvuid_0 or 1. Now we add
     * the ccv_uid_# column to the details 
     */
   %do i = 0 %to &numCCVs-1;
      %let col = %scan(&mapCols, &i+1, ' ');
      proc sql noprint;
           create table _work_tmp as 
                select w.*, d.&col as ccv_uid_&i from _work as w, &varorgLibref..SCENARIO_DETAILS as d
                where d.type_cd='V' and  w.scenario_uid = d.parent_uid;
      quit;
      data _work; 
         set _work_tmp;
      run;
   %end;
   
   %do i = 0 %to &numCCVs-1;
      proc sql noprint;
          create table _work_tmp as
             select w.*, cv.ccv_value_uid as ccvuid_&i. from _work as w, &varcompLibref..ccv_values as cv
             where w.ccmap_&i = input(cv.ccv_db_value, best20.) and cv.ccv_uid = w.ccv_uid_&i;
      quit;
      data _work; 
         set _work_tmp;
      run;

   %end;

    /*- Finally, merge the Loss Dataset with the CC Dataset -*/
    data &dataOut(compress=binary); set _work;
        if _N_ eq 1 then do;
            declare hash ccHash(dataset: "&ccData");
            %do i = 0 %to &numCCVs-1; 
                rc = ccHash.defineKey("ccvuid_&i");
            %end;
            rc = ccHash.defineData('CCUID');
            rc = ccHash.defineDone(); 
            call missing(CCUID);      
        end;
        rc  = ccHash.find();
        drop rc ccmap_: ccv_uid_:;
    run; 
%DONE:
    %if &abort %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, ccm_abnormal_error, noquote)) ;
        %let status=1;    
    %end;   
    %orva_status_ds( statusds=&ostat, status=&status, reason=&reason );
  
    /*- delete unwanted datasets -*/
    proc datasets library=work memtype=data nolist nodetails;
        delete _work: ;
    quit;
%mend orva_map_scenario_ccv_dim;
