/*************************************************************************************
 * Copyright (C), 2007
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME   : orva_readfitopts                                             
 * PURPOSE: Read fitting options from dataset           
 * SUPPORT: stchri
 *
 * PARAMETERS :
 *
 *   fitoptsData = fitting options dataset.
 *   fittype     = SEV,FREQ
 *   type        = THRESHOLD, ADVANCED, DIST, FITMEASURE, SEVUSERINPUT, KRI
 *   datatype    = I,C,P
 *   ccUid       = ccuid for cell we are fitting. If ccUid is missing, only project-level 
 *                 options are used to set the requisite macro variables.
 *                 Otherwise, this macro attempts to find cell-level options for the value of ccUid.  
 *                 (if none are found, then the project-level options are used.)
 *
 * OUTPUT (macro variables): 
 *   type = THRESHOLD:
 *           for fittype = SEV the following macro variables are set:
 *           threshold
 *           gofThreshold
 *           defaultThreshold
 *
 *           for fittype = FREQ and type = THRESHOLD the following macro variables are set:
 *           Freq_Use_Data_Threshold 
 *           Freq_Missing_Default 
 *           Freq_Zero_Count_Default  
 *           Freq_UseStratumMean_Default 
 *           Freq_UseStratumMean_UserSpec
 *           Freq_Cut_Off 
 *           Freq_Keep_Larger_Thresholds
 * 
 *  type = DIST:
 *          for fittype=SEV the following macro variable is set:
 *            dists
 *          for fittype=FREQ the following macro variable is set:
 *            fdists
 *
 *  type = KRI:
 *          for fittype=SEV the following macro variable is set:
 *             Sev_Kris_Suggested
 *             Sev_Kris_Forced
 *          for fittype=FREQ the following macro variable is set:
 *             Freq_Kris_Suggested
 *             Freq_Kris_Forced
 *
 *  type = POBSERV: (both fittypes)
 *          Probability_Observed
 *          Source_Method
 *          Source_Type
 *          Source_Uid           
 *
 *  type = FITMEASURE:
 *         for fittype=SEV the following macro variable is set:
 *           Sev_FitMeasures=;
 *           Sev_FavFitMeasure=;
 *         for fittype=FREQ the following macro variable is set:
 *           Freq_FitMeasures=;
 *           Freq_FavFitMeasure=;
 **************************************************************************************/
%macro orva_readfitopts( 
                        fitoptsData = , 
                        fittype     = SEV, 
                        type        = THRESHOLD, 
                        datatype    = I, 
                        ccUid       = 
                        );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;                             
      %put #orva_readfitopts(; 
      %put .                  fitoptsData = &fitoptsData.,; 
      %put .                  fittype     = &fittype.,; 
      %put .                  type        = &type.,; 
      %put .                  datatype    = &datatype.,; 
      %put .                  ccUid       = &ccUid.;
      %put .                  );
   %end;

   /* Set CCUID to MISSING for project level options */
   %if &ccuid EQ %str() %then %do;
      %let ccuid = .;
   %end;

   /* 
   * Note that THRESHOLD and POBSERV options are datatype (I,C,P) dependent
   * when we are dealing with project level options, but the rest are not.
   * In this second case set datatype to EMPTY.
   */
   %if &ccuid EQ . %then %do;
      %if &type ne THRESHOLD and &type ne POBSERV %then %do;
         %let datatype = %str();
      %end;
   %end;
    
   /* Get our optiont prefix based on fitting TYPE */
   %if &fittype EQ SEV %then %do;
      %let prefix  = Sev_;
      %let prefixL = 4;
   %end;
   %else %do;
      %let prefix  = Freq_;
      %let prefixL = 5;
   %end;   

   /* Finally, handle option type requested */
   %if &type EQ THRESHOLD %then %do;
      %orv_read_threshold_options;
   %end;
   %else %if &type EQ POBSERV %then %do;
      %orv_read_pobserv_options;
   %end;
   %else %if &type EQ DIST %then %do;
      %orv_read_dist_options;
   %end;
   %else %if &type EQ SEVUSERINPUT %then %do;
      %orv_read_sui_options;
   %end;
   %else %if &type EQ KRI %then %do;
      %orv_read_kri_options;
   %end;
   %else %if &type EQ INDICATOR %then %do;
      %orv_read_indicator_options;
   %end;
   %else %if &type EQ KRICONTROL %then %do;
      %orv_read_kricontrol_options;
   %end;
   %else %if &type EQ FITMEASURE %then %do;
      %orv_read_fitmeasure_options;
   %end;
   %else %if &type EQ ADVANCED %then %do;
      %orv_read_advanced_options;
   %end;
%mend orva_readfitopts;

/**
 * Generic macro to retrieve fit options by
 *   type, ccuid, datatype, and prefix
 * datatype is ignored if ccuid = .
 */
%macro orv_get_options(
                       flag      = options_exist,
                       type      = ,
                       datatype  = ,
                       ccuid     = ,
                       usePrefix = true
                       );
   data _NULL_; 
      set &fitoptsData; 
      where 
            CCUID = &ccuid 
        AND 
            type EQ "&type" 
            %if &datatype ne %str() %then %do;
               AND datatype EQ "&datatype"
            %end;
            %if &usePrefix EQ true %then %do;
               AND substr(name, 1, &prefixL) = "&prefix"
            %end;
            ;

      if cvalue NE "" then do;
         call symput(name, trim(left(cvalue)));
      end;
      else do;
         call symput(name, trim(left(put(nvalue, BEST.))));
      end;

      call symput("&flag", "true");
   run;
%mend orv_get_options;

%macro orv_read_kricontrol_options;
   %if &fittype EQ SEV %then %do;
      %let Sev_Use_Kris              = false;
      %let Sev_Selection_Method_Kris = Stepwise;
   %end;
   %else %do;
      %let Freq_Use_Kris              = false;
      %let Freq_Selection_Method_Kris = Stepwise;
   %end;

   /*** read options as specified by ccUid ***/
   %let options_exist = false;

   %orv_get_options(
                    flag      = options_exist,
                    type      = &type,
                    datatype  = &datatype,
                    ccuid     = &ccuid,
                    usePrefix = true
                    );
   /*** if that failed and ccUid was specified, try project-level instead ***/ 
   %if &options_exist EQ false AND &ccuid NE . %then %do;
      %orv_get_options(
                       flag      = options_exist,
                       type      = &type,
                       ccuid     = .,
                       usePrefix = true
                       );
   %end;
%mend orv_read_kricontrol_options;

%macro orv_read_sui_options;
   %let SevUserInput        = true;
   %let TailBoundaryPercent = 95;
   %let TailBoundaryValue   = 1000000;
   %let TailBoundaryChoice  = TailBoundaryPercent;

   /*** read options as specified by ccUid ***/
   %let options_exist = false;

   %orv_get_options(
                    flag      = options_exist,
                    type      = &type,
                    datatype  = &datatype,
                    ccuid     = &ccuid,
                    usePrefix = false
                    );
   /*** if that failed and ccUid was specified, try project-level instead ***/ 
   %if &options_exist EQ false AND &ccuid NE . %then %do;
      %orv_get_options(
                       flag      = options_exist,
                       type      = &type,
                       ccuid     = .,
                       usePrefix = false
                       );
   %end;
%mend orv_read_sui_options;

/*************************************************************
 * Macro to read the POBSERV options 
 ************************************************************/
%macro orv_read_pobserv_options;    
   /*** default values we expect to find ****/
   %let Probability_Observed = .;
   %let Source_Method        = .;
   %let Source_Type          = .;
   %let Source_Uid           = 0;           

   /*** get probability of observability using ccuid specified ***/
   %let options_exist = false;

   %orv_get_pobserv(
                    flag     = options_exist,
                    datatype = &datatype,
                    ccuid    = &ccuid
                    );

   /*** if not found and cell specified try project-level ***/
   %if &options_exist EQ false AND &ccuid NE . %then %do;
      %orv_get_pobserv(
                       flag     = options_exist,
                       datatype = &datatype,
                       ccuid    = .
                       );
   %end;
%mend orv_read_pobserv_options;

%macro orv_get_pobserv(
                       flag     = options_exist,
                       datatype = ,
                       ccuid    =
                       );
   data _NULL_; 
      set &fitoptsData; 
      where 
            CCUID = &ccuid 
        AND 
            type EQ "&type" 
        AND 
            datatype EQ "&datatype"  
        AND 
            substr(name, 1, &prefixL) = "&prefix"
            ;

      if name EQ "Sev_PObserv_Source" then do;
         call symput('Source_Method', 'Infer');
         call symput('Source_Type', trim(left(cvalue)));
         call symput('Source_Uid', trim(left(nvalue)));
      end;
      else if name EQ "Freq_PObserv_Source" then do;
         call symput('Freq_PObserv_Source', trim(left(cvalue)));
      end;
      else if name EQ "Sev_Probability_Observed" then do;
         call symputX('Probability_Observed', put(nvalue, BEST20.));
      end;
      else if name EQ "Sev_PObserv_Method" then do;
         call symput('Source_Method', trim(left(cvalue)));
      end;  

      call symput("&flag", "true");
    run;
%mend orv_get_pobserv;

/*************************************************************
 * Macro to read the THRESHOLD options 
 ************************************************************
 *   Frequency Options
 *
 *       Freq_Cut_Off                 : NUMERIC
 *       Freq_Keep_Larger_Thresholds  : TRUE or FALSE
 *       Freq_Missing_Default         : NUMERIC
 *       Freq_UseStratumMean_Default  : TRUE or FALSE
 *       Freq_Use_Data_Threshold      : TRUE or FALSE
 *       Freq_Zero_Count_Default      : NUMERIC
 *
 *   Severity Options
 *
 *       Sev_Cut_Off                  : NUMERIC
 *       Sev_GOF_Threshold            : NUMERIC
 *       Sev_Keep_Larger_Thresholds   : TRUE or FALSE
 *       Sev_Missing_Default          : NUMERIC
 *       Sev_Use_Data_Threshold       : TRUE or FALSE
 *       Sev_Use_GOF_Threshold        : TRUE or FALSE
 ************************************************************/
%macro orv_read_threshold_options;
   /*** default values we expect to find ****/
   %if &fittype EQ SEV %then %do;
      %let threshold                  = .;
      %let gofThreshold               = .;
      %let defaultThreshold           = .;
      %let Sev_Use_Data_Threshold     = true;
      %let Sev_Use_GOF_Threshold      = false;
      %let Sev_Missing_Default        = 0;
      %let Sev_Cut_Off                = 10000;
      %let Sev_GOF_Threshold          = .;
      %let Sev_Keep_Larger_Thresholds = .;
   %end;
   %else %do;
      %let Freq_Use_Data_Threshold      = true;
      %let Freq_Missing_Default         = 0;
      %let Freq_Zero_Count_Default      = 0;
      %let Freq_UseStratumMean_Default  = false;
      %let Freq_UseStratumMean_UserSpec = false;
      %let Freq_Cut_Off                 = . ;
      %let Freq_Keep_Larger_Thresholds  = true;
   %end;           
     
   /*** read options as specified by ccUid ***/
   %let options_exist = false;

   %orv_get_options(
                    flag     = options_exist,
                    type     = &type,
                    datatype = &datatype,
                    ccuid    = &ccuid
                    );

   /*** if that failed and ccUid was specified, try project-level instead ***/ 
   %if &options_exist EQ false AND &ccuid NE . %then %do;
      %orv_get_options(
                       flag     = options_exist,
                       type     = &type,
                       datatype = &datatype,
                       ccuid    = .
                       );
   %end;
     
   %if &fittype EQ SEV %then %do;
      %if %upcase(&Sev_Use_Data_Threshold) NE TRUE %then %do;
         %let threshold = &Sev_Cut_Off.;
      %end;

      %if %upcase(&Sev_Use_GOF_Threshold) EQ TRUE %then %do;
         %let gofThreshold = &Sev_GOF_Threshold.;
      %end;

      %let defaultThreshold = &Sev_Missing_Default.;
   %end;     
%mend orv_read_threshold_options;

/*************************************************************
 * Macro to read the distribution options into a
 * space-delimited string, placed in either dists or fdists
 ************************************************************/
%macro orv_read_dist_options;
   /*** initialize our macro variables ***/
   %if &fittype EQ SEV %then %do;
      %let dists = ;
   %end;
   %else %do;
      %let fdists = ;
   %end;

   /*** look for distributions as requested, and concatenate ***/
   %let options_exist = false;

   %orv_get_dists(
                  flag     = options_exist,
                  ccuid    = &ccUid,
                  datatype = &datatype
                  );
   /*** if we attempted cell-level and it failed, try project-level ***/
   %if &options_exist eq false AND &ccUid NE . %then %do;
      %orv_get_dists(
                     flag     = options_exist,
                     ccuid    = .,
                     datatype = &datatype
                     );   
   %end;
%mend orv_read_dist_options;

%macro orv_get_dists(
                     flag    = options_exist,
                     ccuid    = ,
                     datatype =
                     );
   data _NULL_; 
      set &fitoptsData end = LAST; 
      where 
            CCUID = &ccuid 
        AND 
            type EQ "DIST" 
        AND 
            substr(name, 1, &prefixL) = "&prefix"
            %if &ccuid ne . %then %do; 
               AND datatype = "&datatype" 
            %end;
            ;
      length dists $512;
      retain dists;
      dists = trim(left(dists)) || " " || trim(left(cvalue)) ;

      if LAST then do;
            %if &fittype EQ SEV %then %do;
               call symput('dists', trim(left(dists)));
            %end;
            %else %do;
               call symput('fdists', trim(left(dists)));
            %end;
   
         call symput("&flag", "true");
      end;
   run;
%mend orv_get_dists;

/*************************************************************
 * Macro to read the FITMEASURE options 
 ************************************************************/
%macro orv_read_fitmeasure_options;
   /*** initialize our macro variables ***/
   %if &fittype EQ SEV %then %do;
      %let Sev_FitMeasures   = ;
      %let Sev_FavFitMeasure = ;
   %end;
   %else %do;
      %let Freq_FitMeasures   = ;
      %let Freq_FavFitMeasure = ;
   %end;

   /*** look for distributions as requested, and concatenate ***/
   %let options_exist = false;
  
   %orv_get_options(
                    flag      = options_exist,
                    type      = &type,
                    ccuid     = &ccUid,
                    usePrefix = true,
                    datatype  = &datatype
                    );
   /*** if we attempted cell-level and it failed, try project-level ***/
   %if &options_exist eq false AND &ccUid NE . %then %do;
      %orv_get_options(
                       flag      = options_exist,
                       type      = &type,
                       ccuid     = .,
                       usePrefix = true
                       );   
   %end;
%mend orv_read_fitmeasure_options;

/**************************************************************
 * Macro to read the KRI options into a
 * space-delimited string, placed in the following macro vars:
 *   Sev_Kris_Suggested, Sev_Kris_Forced
 *     or
 *   Freq_Kris_Suggested, Freq_Kris_Forced
 *************************************************************/
%macro orv_read_kri_options;
   /*** initialize our macro variables ***/
   %if &fittype EQ SEV %then %do;
      %let Sev_Kris_Suggested = ;
      %let Sev_Kris_Forced    = ;
   %end;
   %else %do;
      %let Freq_Kris_Suggested = ;
      %let Freq_Kris_Forced    = ;
   %end;     

   /*** look for kris as requested, and concatenate ***/
   %let options_exist = false;
   %orv_get_kris(
                 flag     = options_exist,
                 ccuid    = &ccUid,
                 datatype = &datatype
                 );

   /* Don't attempt if cell level KRICONTROL is set to true ------*/
   %local 
         Sev_Use_Kris 
         Freq_Use_Kris 
         Sev_Selection_Method_Kris 
         Freq_Selection_Method_Kris
         ;
   %let Sev_Use_Kris  = false;
   %let Freq_Use_Kris = false;

   data _NULL_; 
      set &fitoptsData; 
      where CCUID = &ccuid AND type EQ "KRICONTROL"  ;
      call symput(name, compress(cvalue));
   run;

   /*** if we attempted cell-level and it failed, try project-level ***/
   %if 
      (&Sev_Use_Kris = false and &fittype = SEV) 
     or
      (&Freq_Use_Kris = false and &fittype = FREQ) 
      %then %do;
      /*** if we attempted cell-level and it failed, try project-level ***/
      %if &options_exist eq false AND &ccUid NE . %then %do;
         %orv_get_kris(
                       flag     = options_exist,
                       ccuid    = .,
                       datatype = &datatype
                       );   
      %end;
   %end;
%mend orv_read_kri_options;

/**************************************************************
 * Macro to read the INDICATOR options into a
 *  placed in the following macro vars:
 *************************************************************/
%macro orv_read_indicator_options;
   /*** initialize our macro variables ***/
   %if &fittype EQ FREQ %then %do;
      %let Freq_Model_Is_Stratified = ;
      %let Freq_Model_Basis_Props   = ;
      %let Freq_Use_Dummies         = ;
   %end;     

   /*** look for  ***/
   %let options_exist = false;

   %orv_get_options(
                    flag      = options_exist,
                    type      = INDICATOR,
                    ccuid     = &ccUid,
                    datatype  = &datatype, 
                    usePrefix = false
                    );
   /*** if we attempted cell-level and it failed, try project-level ***/
   %if &options_exist eq false AND &ccUid NE . %then %do;
      %orv_get_options(
                       flag      = options_exist,
                       type      = INDICATOR,
                       ccuid     = .,
                       usePrefix = false
                       );
   %end;
%mend orv_read_indicator_options;

/**************************************************************
 * Macro to read the ADVANCED options into a
 *  placed in the following macro vars:
 *************************************************************/
%macro orv_read_advanced_options;
   /*** initialize our macro variables ***/
   %if &fittype EQ SEV %then %do;
      %let Min_Fitting_Count = ;
   %end;     

   /*** look for  ***/
   %let options_exist = false;

   %orv_get_options(
                    flag      = options_exist,
                    type      = ADVANCED,
                    ccuid     = &ccUid,
                    datatype  = &datatype, 
                    usePrefix = false
                    );
   /*** if we attempted cell-level and it failed, try project-level ***/
   %if &options_exist eq false AND &ccUid NE . %then %do;
      %orv_get_options(
                       flag      = options_exist,
                       type      = ADVANCED,
                       ccuid     = .,
                       usePrefix = false
                       );
   %end;
%mend orv_read_advanced_options;

%macro orv_get_kris(
                  flag     = options_exist,
                  ccuid    = ,
                  datatype =
                  );

   proc sql noprint;
      /* Suggested */
      SELECT distinct
                        name
      INTO
                        :&fittype._Kris_Suggested separated by " "
      FROM
                        &fitoptsData.
      WHERE
                        CCUID = &ccuid 
                        AND 
                        type EQ "KRI" 
                        AND 
                        cvalue EQ "&fittype"
                        %if &ccuid NE . %then %do; 
                           AND 
                           datatype = "&datatype" 
                        %end;
                        AND
                        nvalue = 1
      ;
      /* Forced */
      SELECT distinct
                        name
      INTO
                        :&fittype._Kris_Forced separated by " "
      FROM
                        &fitoptsData.
      WHERE
                        CCUID = &ccuid 
                        AND 
                        type EQ "KRI" 
                        AND 
                        cvalue EQ "&fittype"
                        %if &ccuid NE . %then %do; 
                           AND 
                           datatype = "&datatype" 
                        %end;
                        AND
                        nvalue = 2
      ;
   quit;

   /* Check that there are any KRI rows for this CCUID and this fit type.  */
   data _NULL_; 
      set &fitoptsData end = LAST; 
      where 
            CCUID = &ccuid 
        AND 
            type EQ "KRI" 
        AND 
            cvalue EQ "&fittype"
            %if &ccuid NE . %then %do; 
               AND datatype = "&datatype" 
            %end;
            ;
      if last then do;
         /* Only runs if the dataset has at least one row */
         call symput("&flag", "true");
      end;
   run;

%mend orv_get_kris;
