/*****************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved. 
 *
 * NAME    : orva_rollup.sas  
 * PURPOSE : Combines shared losses.           
 * SUPPORT : stchri         
 *
 * PARAMETERS :
 *
 * lossData        -  the name of the loss data with the following variables.
 * byCCVs          -  space-delimited list of cross classification variables.
 * numCCVs         -  number of CC variables in byCCVs.
 * ccData          -  Cross Classification dataset.
 * singleEventMult -  Single event multiple loss key (Internal and Consortium only).
 * outDataShared   -  dataset with shared losses.  This dataset may contain duplicates
 *                    since shared losses can span across cells and are contained in
 *                    aggregate cells.
 * outDataSingle   -  dataset of non-shared losses.  No duplicates are ever contained
 *                    in this dataset because it does not conatin aggregates.
 * dataType        -  I or C only.
 *
 *****************************************************************************************/
 %macro orva_rollup(
                     lossData        =, 
                     byCCVs          =, 
                     numCCVs         =, 
                     ccData          =,
                     singleEventMult =, 
                     outDataShared   =, 
                     outDataSingle   =, 
                     dataType        =
                     );

      /***** get subset of losses that are SHARED *****/
      proc sort data=&lossData(where = (&singleEventMult notin (0,.))) 
          out=_sorted;
          by &BYCCVS &singleEventMult;
      run;  
      
      /***** handle INTERNAL losses *****/
      %if &dataType eq I %then %do;
           
           /***** Get a list of variables to add to id statement *****/
         proc contents 
            data = _sorted(drop = AMOUNT LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                             LOSS_OF_RECOURSE  WRITE_DOWNS INSURANCE_RECOVERY_1 INSURANCE_RECOVERY_2 
                             NON_INSURANCE_RECOVERIES &singleEventMult CCUID
                             DATE_OF_DISCOVERY DATE_OF_PROVISION DATE_OF_SETTLEMENT 
                             FIRST_DATE_OF_EVENT FINANCIALDATE CCVUID_:
                             %do i = 1 %to &numCCVs;
                                %scan(&byCCVs, &i, ' ')
                             %end;
                             )
            out  = _c
            noprint
            ; 
           run;

           proc sql noprint;
              SELECT distinct
                               name
              into
                               :idvars separated by " "
              from
                               _c
              ;
           quit;

           /***** 
            *  Use proc summary to roll up the shared losses by CCVUIDs and &SingleEventMult      
            ****/
           proc summary data=_sorted;
              class &BYCCVS &SingleEventMult;
              var AMOUNT LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                  LOSS_OF_RECOURSE WRITE_DOWNS INSURANCE_RECOVERY_1 INSURANCE_RECOVERY_2 
                  NON_INSURANCE_RECOVERIES
                  DATE_OF_DISCOVERY DATE_OF_PROVISION DATE_OF_SETTLEMENT 
                  FIRST_DATE_OF_EVENT FINANCIALDATE;
              id &idvars;
              output sum ( AMOUNT LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                           LOSS_OF_RECOURSE WRITE_DOWNS INSURANCE_RECOVERY_1 INSURANCE_RECOVERY_2 
                           NON_INSURANCE_RECOVERIES ) 
                         = AMOUNT LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                           LOSS_OF_RECOURSE WRITE_DOWNS INSURANCE_RECOVERY_1 INSURANCE_RECOVERY_2 
                           NON_INSURANCE_RECOVERIES
                     min ( DATE_OF_DISCOVERY DATE_OF_PROVISION DATE_OF_SETTLEMENT 
                           FIRST_DATE_OF_EVENT FINANCIALDATE )
                         = DATE_OF_DISCOVERY DATE_OF_PROVISION DATE_OF_SETTLEMENT 
                           FIRST_DATE_OF_EVENT FINANCIALDATE
                     out=_shared(drop= _type_ _freq_);
           run;

      %end; 
      %else %if &dataType eq C %then %do;
          
         /***** Get a list of variables to add to id statement *****/
         proc contents data=_sorted out=_c noprint; run;
         data _NULL_; set _c;
           retain idvars; length idvars $ 11024;
           if upcase(name) notin ("AMOUNT" "LEGAL_LIABILITY" "REGULATORY_ACTION" 
                                  "RESTITUTION" "LOSS_TO_ASSETS" "LOSS_OF_RECOURSE" 
                                  "WRITE_DOWNS" "INSURANCE_RECOVERY" "NON_INSURANCE_RECOVERIES"
                                  "&SingleEventMult" "CCUID" "FIRST_DATE_OF_EVENT" "FINANCIALDATE")  
                           and substr(name,1,7) ne "CCVUID_" 
                           %do i = 1 %to &numCCVs;
                              and upcase(name) ne "%scan(&byCCVs, &i, ' ')" 
                            %end;
              then do;
              idvars = trim(idvars) || ' ' || trim(name);
              call symput( 'idvars', trim(left(idvars)) );
           end;
         run;

         /***** 
          *  Use proc summary to roll up the shared losses by CCVUIDs and &SingleEventMult
          ****/
         proc summary data=_sorted;
            class &BYCCVS &SingleEventMult;
            var AMOUNT LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                LOSS_OF_RECOURSE WRITE_DOWNS INSURANCE_RECOVERY NON_INSURANCE_RECOVERIES
                FIRST_DATE_OF_EVENT FINANCIALDATE;
            id &idvars;
            output sum ( AMOUNT LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                         LOSS_OF_RECOURSE WRITE_DOWNS INSURANCE_RECOVERY NON_INSURANCE_RECOVERIES )
                       = AMOUNT LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                         LOSS_OF_RECOURSE WRITE_DOWNS INSURANCE_RECOVERY NON_INSURANCE_RECOVERIES
                   min ( FIRST_DATE_OF_EVENT FINANCIALDATE )
                       = FIRST_DATE_OF_EVENT FINANCIALDATE
                   out=_shared(drop=_type_ _freq_);
         run;
      %end;
     
      /***** create SINGLE output dataset *****/     
      data _single (compress=binary); 
         set &lossData(where=(&SingleEventMult in (0,.)));
         %do i = 0 %to &numCCVs-1;  
              if CCVUID_&i eq . then CCVUID_&i = 0;
         %end;
      run;   
      proc sql noprint nowarn; create table _join as select * from
         _single S left join &ccData C on 1
         %do i = 0 %to &numCCVs-1;  
              and S.CCVUID_&i eq C.CCVUID_&i 
         %end;
         ;
      quit;
      data &outDataSingle (compress=binary); set _join; run;

      /***** create SHARED output dataset *****/     
      data _shared (compress=binary); 
        set _shared; where &SingleEventMult ne .;
        %do i = 0 %to &numCCVs-1;   
              if CCVUID_&i eq . then CCVUID_&i = 0;
        %end;
      run;
      proc sql noprint nowarn; create table _join as select * from
        _shared S left join &ccData C on 1
        %do i = 0 %to &numCCVs-1;   
              and S.CCVUID_&i eq C.CCVUID_&i 
        %end;
        ;
      quit;
      data &outDataShared (compress=binary); set _join; run;
   
      /***** cleanup our temporary datasets *****/
      proc datasets library=work memtype=data nolist;
         delete _c _shared _sorted _join _single;
       quit;
     
%mend orva_rollup;
