/*****************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    ORVA_scenarioData.sas                                             
 *  PURPOSE: Flattens out (denormalizes) the Scenario data and merges with the 
 *           Cross Classification and Mapping Datasets.
 *
 *  SUPPORT: stchri
 *
 *  PARAMETERS:
 *
 *    scenarioUID   = Scenario UID.
 *    versionUID    = NOT USED
 *    numCCVs       = number of Cross Classification variables, defaults to 2.
 *    mapCols       = space-separated list of mapping columns, 
 *                    ex: INTERNAL_BL_UID INTERNAL_RC_UID in &data.
 *    ccData        = the name of the Cross Classification dataset with the following variables:
 *                    CCUID, CCVUID_<n> (one for each Cross Classification)
 *    mapDataPrefix = prefix for the name of the mapping datasets. Should be one
 *                    for each mapping called &mapDataPrefix<n>.   
 *    scenarioOut   = the name of the output dataset (Scenario Data). This will be
 *                    denormalied and merges with the CCV data.
 *    scenGroupOut  = the name of the output dataset for Group Frequencies.
 *    bucketOut     = the name of the output dataset for Buckets.
 *    outStatus     = name of the out status dataset.  Will contain the following:
 *                    status = 0 on success, 1 otherwise.
 *                    reason = printable message.
 *      
 *  Example: %ORVA_scenarioData(
 *                               scenarioUid   = 1002,
 *                               numCCVs       = 2,
 *                               mapCols       = INTERNAL_BL_UID INTERNAL_RC_UID,
 *                               ccData        = libname.CC,
 *                               mapDataPrefix = libname.CCMAPI_,
 *                               scenarioOut   = libname.Scen_331,
 *                               scenGroupOut  = libname.ScenGroup_331,
 *                               bucketOut     = libname.Buck_331,
 *                               outStatus     = status,
 *                               varorgLibref  = VO1
 *                               );
 * 
 *****************************************************************************************/
%macro ORVA_scenarioData(
                         scenarioUID   = ,
                         versionUID    = ,
                         numCCVs       = ,
                         mapCols       = ,
                         baselUids     = ,
                         ccData        = ,
                         mapDataPrefix = ,
                         scenarioOut   = ,
                         scenGroupOut  = ,
                         bucketOut     = ,
                         outStatus     = ,
                         varorgLibref  = ,
                         varcompLibref =
                         );
   %if %upcase(&ORV_DEBUG.) EQ TRUE %then %do;
      %put #ORVA_scenarioData(;
      %put .                   scenarioUID   = &scenarioUID.,;
      %put .                   versionUID    = &versionUID.,;
      %put .                   numCCVs       = &numCCVs.,;
      %put .                   mapCols       = &mapCols.,;
      %put .                   baselUids     = &baselUids.,;
      %put .                   ccData        = &ccData.,;
      %put .                   mapDataPrefix = &mapDataPrefix.,;
      %put .                   scenarioOut   = &scenarioOut.,;
      %put .                   scenGroupOut  = &scenGroupOut.,;
      %put .                   bucketOut     = &bucketOut.,;
      %put .                   outStatus     = &outStatus.,;
      %put .                   varorgLibref  = &varorgLibref.,;
      %put .                   varcompLibref = &varcompLibref.;
      %put .                   );
   %end;

   /** create the status dataset up front **/
   %let reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( 
                   statusds = &outStatus, 
                   status   = 1, 
                   reason   = &reason 
                   ); 
   %let status = 0;
   %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));

   /** subset _EVENTS by VERSION AND SCENARIO_UID **/
   data _events (KEEP = 
                       SCENARIO_EVENT_UID
                       SCENARIO_UID
                       EVENT_NAME
                       EVENT_DESC
                       EVENT_TYPE
                       FREQUENCY_DIST
                       FREQUENCY_PARM0
                       FREQUENCY_PARM1
                       FREQUENCY_PARM2
                       VERSION_UID
                       );
      set &varorgLibref..SCENARIO_EVENT;
      where SCENARIO_UID = &scenarioUID;
   run;

   /** subset SCENARIO_EFFECT Joined to SCENARIO_EVENT by SCENARIO_UID **/
   proc sql noprint;
      CREATE TABLE 
                  _detailsStepOneEff 
      AS 
      SELECT 
                  * 
      FROM 
                  _events a,
                  &varorgLibref..SCENARIO_EFFECT b
      WHERE 
                  a.SCENARIO_EVENT_UID = b.SCENARIO_EVENT_UID 
         AND 
                  a.SCENARIO_UID = &scenarioUID
                  ;
   quit;

   /** subset SCENARIO_EFFECT_GROUP Joined to SCENARIO_EVENT by SCENARIO_UID **/
   proc sql noprint;
      CREATE TABLE 
                  _detailsStepOneGrp 
      AS 
      SELECT 
                  * 
      FROM 
                  _events a,
                  &varorgLibref..SCENARIO_EFFECT_GROUP b
      WHERE 
                  a.SCENARIO_EVENT_UID = b.SCENARIO_EVENT_UID 
         AND 
                  a.SCENARIO_UID = &scenarioUID
                  ;
   quit;

   /** add the details **/
   proc sql noprint;
      CREATE TABLE 
                  _detailsEff 
      AS 
      SELECT 
                  * 
      FROM 
                  _detailsStepOneEff a,
                  &varorgLibref..SCENARIO_DETAILS b
      WHERE 
                  a.SCENARIO_EFFECT_UID = b.PARENT_UID 
         AND 
                  b.TYPE_CD = 'E'
                  ;
   quit;

   /** add the details **/
   proc sql noprint;
      CREATE TABLE 
                  _detailsGrp 
      AS 
      SELECT 
                  * 
      FROM 
                  _detailsStepOneGrp a,
                  &varorgLibref..SCENARIO_DETAILS b
      WHERE 
                  a.SCENARIO_EFFECT_GROUP_UID = b.PARENT_UID 
         AND 
                  b.TYPE_CD = 'G'
                  ;
   quit;

   %orvu_delete_data( data = 
                             _detailsStepOneEff 
                             _detailsStepOneGrp 
                             _events
                             );

   /** make sure that the CCUID are actually columns in the Scenario Data **/
   %if "&baselUids" eq "" %then %do; 
      proc contents 
                  data = _detailsEff 
                  out  = _cat 
                  noprint
                  ; 
      run;

      data _NULL_; 
         set _cat;
         retain found;

         if missing(found) then do;
            found = 0;
         end;

         %let nCOL = 1;
         %let colname = %scan(&mapCols, &nCOL, ' ');

         %do %while(&colname ne ); 
            if "&colname" eq upcase(trim(left(name))) then do;
               found = found + 1;
            end;

            %let nCOL = %eval(&nCOL + 1);
            %let colname = %scan(&mapCols, &nCOL, ' ');
         %end;

         call symput("FOUND",found);
      run;

      %orvu_delete_data( data = _cat);

      %if &found ne &numCCVs %then %do;
         data &outStatus;
            format reason &ORVAR_REASON_FMT;
            status = 2;
            reason = "ERROR: Scenario &scenarioUID is not valid for project CCVs: &mapCols";
            put reason;
            dropped = 0;
            bucket_max = 0;
            sparm_max = 0;
         run;

         %goto DONE;
      %end;
   %end;

   %let BUCKET_MAX = 0;
   %let SPARM_MAX  = 0;

   /***************************************************************/
   /****************** BUCKETED Scenario **************************/
   /***************************************************************/
   %if &bucketOut ne %str() %then %do;
      /***** join UPPER_BOUNDS with BUCKETS *****/
      /** collapse bounds **/
      data _bounds0; 
         set &varorgLibref..BUCKET_UPPER_BOUNDS;
      run;
      
      proc sql noprint; 
         SELECT 
                  MAX(BUCKET_NUM) 
         INTO 
                  :BUCKET_MAX 
         FROM 
                  _bounds0
                  ; 
      quit;
      
      proc sort data = _bounds0; 
         by 
            BUCKET_UID 
            BUCKET_NUM
            ;
      run;
      
      data _bounds1; 
         set _bounds0; 
         by BUCKET_UID;

         retain 
               count 
               %do i = 0 %to &BUCKET_MAX; 
                  bucket&i 
               %end; 
               ;

         if missing(count) then do;
            count = 0;
         end;
         else do;
            count = count + 1;  
         end;

         %do i = 0 %to &BUCKET_MAX; 
            if count = &i then do;
               bucket&i = UPPER_BOUND_VALUE; 
            end;
         %end;

         if LAST.BUCKET_UID then do;
            count = .;
            output;

            %do i = 0 %to &BUCKET_MAX; 
               bucket&i = .; 
            %end; 
            ;
         end;

         drop 
               count 
               UPPER_BOUND_VALUE 
               BUCKET_BOUND_UID 
               BUCKET_NUM
               ;
      run;

      /** join collapsed bounds with buckets **/
      data _buckets0; 
         set &varorgLibref..BUCKET;
      run;

      proc sql noprint;
         CREATE TABLE 
                        &bucketOut 
         AS SELECT 
                        * 
         FROM 
                        _buckets0 b,
                        _bounds1 AS d
         WHERE 
                        b.BUCKET_UID eq d.BUCKET_UID
                        ;
      quit;
     
      /**** join SCENARIOS WEIGHTS *****/
      /** collapse weights **/
      data _weights0; 
         set &varorgLibref..BUCKET_WEIGHTS;
      run;
      
      proc sql noprint; 
         SELECT 
                  MAX(BUCKET_NUM) 
         INTO 
                  :BUCKET_MAX 
         FROM 
                  _weights0
                  ; 
      quit;

      proc sort data = _weights0; 
         by 
            SCENARIO_EFFECT_UID 
            BUCKET_NUM
            ; 
      run;

      data _weights1; 
         set _weights0; 
         by SCENARIO_EFFECT_UID;

         retain 
               count 
               %do i = 1 %to %eval(&BUCKET_MAX + 1); 
                  bucketwgt&i 
               %end; 
               ;

         if count eq . then DO;
            count = 1;
         end;
         else do;
            count = count + 1;  
         end;

         %do i = 1 %to %eval(&BUCKET_MAX + 1); 
            if count = &i then do;
               bucketwgt&i = WEIGHT_VALUE; 
            end;
         %end;

         if LAST.SCENARIO_EFFECT_UID then do;
            count = 0;
            output;
            %do i = 1 %to %eval(&BUCKET_MAX + 1); 
               bucketwgt&i = .; 
            %end; 
            ;
         end;

         drop 
               count 
               WEIGHT_VALUE 
               BUCKET_NUM 
               BUCKET_WEIGHT_UID
               ;
      run;

      /** join collapsed weights with details **/
      proc sql noprint;
         CREATE TABLE 
                        _scenarios 
         AS 
         SELECT 
                        * 
         FROM 
                        _detailsEff d,
                        _weights1 AS w
         WHERE 
                        d.SCENARIO_EFFECT_UID eq w.SCENARIO_EFFECT_UID
                        ;
      quit;

      data _scenarios; 
         set _scenarios (rename =(
                                    FREQUENCY_PARM0 = FREQUENCY_MEAN
                                    FREQUENCY_PARM1 = FREQUENCY_MIN
                                    FREQUENCY_PARM2 = FREQUENCY_MAX
                                    ));
         if (EVENT_TYPE = 0) then do;
            FREQUENCY_DIST = '';
            FREQUENCY_MEAN = SIMPLE_FREQUENCY_PARM0;
            FREQUENCY_MIN  = SIMPLE_FREQUENCY_PARM1;
            FREQUENCY_MAX  = SIMPLE_FREQUENCY_PARM2;
         end;

         drop 
               SIMPLE_FREQUENCY_DIST
               SIMPLE_FREQUENCY_PARM0
               SIMPLE_FREQUENCY_PARM1
               SIMPLE_FREQUENCY_PARM2
               ;
      run;

      /** cleanup for bucket scenarios **/
      %orvu_delete_data( data = 
                                _bounds: 
                                _weights: 
                                _buckets: 
                                ) ;
   %end;
   /***************************************************************/
   /****************** MODEL Scenario *****************************/
   /***************************************************************/
   %else %do;
      /** collapse sparms **/
      data _sevs0; 
         set &varorgLibref..MODEL_SCENARIO_SEV_PARMS;
      run;

      proc sql noprint; 
         SELECT 
                  MAX(PARM_NUM) 
         INTO 
                  :SPARM_MAX 
         FROM 
                  _sevs0
                  ; 
      quit;

      proc sort data = _sevs0; 
         by 
            SCENARIO_EFFECT_UID 
            PARM_NUM
            ; 
      run;
      
      data _sevs1; 
         set _sevs0; 
         by SCENARIO_EFFECT_UID;

         retain 
               count 
               %do i = 1 %to &SPARM_MAX; 
                  Param_&i 
               %end; 
               ;

         if missing(count) then do;
            count = 1;
         end;
         else do;
            count = count + 1;  
         end;

         %do i = 1 %to &SPARM_MAX; 
            if count = &i then Param_&i = PARM_VALUE; 
         %end;

         if LAST.SCENARIO_EFFECT_UID then do;
            count = . ;
            output;

            %do i = 1 %to &SPARM_MAX; 
               Param_&i = . ; 
            %end; 
            ;
         end;

         drop 
               count 
               PARM_VALUE 
               PARM_NUM
               ;
      run;

      /** join severities with scenarios **/
      proc sql noprint;
         CREATE TABLE 
                        _scenarios 
         AS 
         SELECT 
                        * 
         FROM 
                        _detailsEff d,
                        _sevs1   s
         WHERE 
                        d.SCENARIO_EFFECT_UID = s.SCENARIO_EFFECT_UID
                        ;
      quit;

      data _scenarios; 
         set _scenarios (rename = (
                                    Frequency_Parm0 = FParam_1 
                                    Frequency_Parm1 = FParam_2
                                    ));

         if (EVENT_TYPE = 0) then do;
            FREQUENCY_DIST = SIMPLE_FREQUENCY_DIST;
            FParam_1       = SIMPLE_FREQUENCY_PARM0;
            FParam_2       = SIMPLE_FREQUENCY_PARM1;
         end;

         drop 
               SIMPLE_FREQUENCY_DIST
               SIMPLE_FREQUENCY_PARM0
               SIMPLE_FREQUENCY_PARM1
               SIMPLE_FREQUENCY_PARM2
               FREQUENCY_PARM2
               ;
      run;

      /** cleanup for models scenarios **/
      %orvu_delete_data( data = _sevs: );
   %end;
  
   /***************************************************************/
   /*********************** merge with CCUIDs *********************/  
   /***************************************************************/
   %let dsid         = %sysfunc(open( _scenarios, I ));
   %let INITIAL_NOBS = %sysfunc(attrn( &dsid, nlobs));
   %let rc           = %sysfunc(close( &dsid ));

   /***** Use %orva_baslccmerge() to map and merge *****/
   %orva_baslccmerge(
                     numCCVs      = &numCCvs, 
                     data         = _scenarios,
                     mapCols      = &mapCols,
                     baselUids    = &baselUids, 
                     ccData       = &ccData, 
                     ccValuesData = &varcompLibref..CCV_VALUES,
                     varcomp      = &varcompLibref.,
                     dataOut      = &scenarioOut, 
                     ostat        = &outStatus
                     );
  
   %let dsid       = %sysfunc( open( &scenarioOut, I ));
   %let FINAL_NOBS = %sysfunc( attrn( &dsid, nlobs ));
   %let rc         = %sysfunc( close( &dsid ));
 
   /*- make sure dataset is sorted -*/
   proc sort data = &scenarioOut; 
      by CCUID; 
   run;
 
   data &outStatus; 
      set &outStatus;
      sparm_max  = &SPARM_MAX;
      bucket_max = %eval(&BUCKET_MAX + 1);
      dropped    = &INITIAL_NOBS - &FINAL_NOBS;
      total      = &INITIAL_NOBS;
   run;

   /** Create Scenario Group dataset **/
   %orvl_ds_exists(_detailsGrp, "Scenario Group Processing", false);

   %if &ds_nobs > 0 %then %do;
      %orva_baslccmerge(
                        numCCVs      = &numCCvs, 
                        data         = _detailsGrp,
                        mapCols      = &mapCols,
                        baselUids    = &baselUids, 
                        ccData       = &ccData, 
                        ccValuesData = &varcompLibref..CCV_VALUES,
                        varcomp      = &varcompLibref.,
                        dataOut      = &scenGroupOut, 
                        ostat        = groupOutStatus
                        );

      /*- Reduce dataset -*/
      data &scenGroupOut(keep = 
                                Group_ID 
                                Group_Prob 
                                CCUID 
                                Event_ID 
                                CCVUID_: 
                                Frequency_Dist
                                FParam_1
                                FParam_2
                                ); 
         set &scenGroupOut (rename = (
                                      SCENARIO_EFFECT_GROUP_UID = Group_ID
                                      Frequency_Parm0           = FParam_1
                                      Frequency_Parm1           = FParam_2
                                      SCENARIO_EVENT_UID        = Event_ID
                                      ));
      run;
      
      /* Make sure that no duplicate records remain */
      proc sort 
               data = &scenGroupOut 
               out  = &scenGroupOut 
               noduprecs
               ;
         by Event_ID ;
      run;

      /*- make sure dataset is sorted -*/
      proc sort data = &scenarioOut; 
         by CCUID; 
      run;
   %end;

   /*** cleanup ***/   
   %orvu_delete_data( data = 
                            _scenarios: 
                            _details:
                            );
%DONE:
%mend ORVA_scenarioData;                                                                                          
