/*************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME   : ORVA_SFit_KRI.sas                                             
 * PURPOSE: fit operational risk severity distributions with KRIs.
 *          Currently this macro should only be called by %orva_sfit.            
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *    DATA         = Specifies the name of the input data set to be analyzed.
 *    CCUID        = Specifies cell to fit.
 *    CCDATA       = Specifies the name of the CC Dataset.
 *    OUTPARM      = Name of the output data set containing parameter estimates.
 *    PARALLEL     = 1: use parallel processing
 *                   0: don't use parallel processing
 *
 * NOTES: This macro assumes that numerous macro variables are set by the caller.
 *    See orva_sfit() for a list. 
 *
 *    Please see the accompanying doc for pdfs and cdfs of the distributions
 *
 **************************************************************************************/
%macro ORVA_SFit_kri(
                     data             = ,
                     whereClause      = ,
                     useQLIM          = false,
                     outparm          = SPARM,
                     outcov           = ,
                     parallel         = 0
                     );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put #ORVA_SFit_kri(;
      %put .               data             = &data.,;
      %put .               whereClause      = &whereClause.,;
      %put .               useQLIM          = &useQLIM.,;
      %put .               outparm          = &outparm.,;
      %put .               outcov           = &outcov.,;
      %put .               parallel         = &parallel.;
      %put .               );
   %end;            

   %let dists       = %upcase(&dists);
   %let abort       = 0; /* This flag indicates abortion due to error */
   %let no_kri_stop = 0; /* This flag indicates intentional abortion of the macro because no
                           Distributions are KRI fitable- no error is displayed */
   %let status = 0;
   %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));

   /* save thresholds for parm dataset */
   %let orig_thresh     = &threshold;
   %let orig_gofthresh  = &gofthreshold;
   %let orig_defthresh  = &defaultthreshold;
   %let orig_percthresh = &thresholdPercent;
   
   /* First go through dusts and check for any that are not fitable with KRIs.
     Remove any that are found.  Display a message in the log for each one. 
     If the list becomes empty, abort */
   data _NULL_;
      length 
            dist_to_check   $15 
            dist_string_out $1000 
            tmp             $500 
            skipped_dists   $1000
            ;
      i = 1;
      dist_string_out = " ";
      dist_to_check = scan("&dists.", i, ' ');
      skipped_dists = " ";
      retain n_skipped 0;

      do while(dist_to_check NE "");
         /* If this dist is KRI fitable, keep it*/
         if IsSevKriFitable(dist_to_check) then do;
            dist_string_out = trim(dist_string_out)||' '||trim(dist_to_check);
         end;
         /* otherwise, put it in a list of skipped distributions */
         else do;
            skipped_dists = trim(skipped_dists)|| ' '||trim(dist_to_check);
            n_skipped = n_skipped + 1;
         end;

         i = i + 1;
         dist_to_check = scan("&dists.",i,' ');
      end;

      call symput("dists", dist_string_out);
      call symput("skipped_dists", skipped_dists);
      call symput("n_skipped", n_skipped);
   run;

   /* Now check for distributions that are skipped */
   %if &n_skipped %then %do;
      %do i = 1 %to &n_skipped.;
            %let tmp = "%sysfunc(sasmsg(&orvads, sfkri_skip_fit_because_not_kri_fitable, noquote,%scan(&skipped_dists,&i)))";
            %put &orvar_warning &tmp;
      %end;
   %end;
   /* If there are no KRI fitable distributions.  We will abort, but need to do some other stuff first. */

   /*-------------------------------------------------
   * prepare output data sets
   * Note:
   *  AIC: Akaike Information Criterion
   *  SBC: Schwartz Bayesian Criterion
   *------------------------------------------------*/
   data &outcov; 
      set _NULL_; 
   run;

   data &outparm;
      format  
               distribution $12. 
               status 8. 
               reason &ORVAR_REASON_FMT
               KRI_Select_Method $12.
               Sev_Keep_Larger_Thresholds $5.
               ;
      distribution = "";
      status       = 1;
      reason       = "%sysfunc(sasmsg(&orvads, unknown_error, noquote))";
      mean         = .;
      std          = .;
      skewness     = .;
      kurtosis     = .;
      logl         = .;
      aic          = .;
      sbc          = .;
      ad_v         = .;
      ad_p         = .;
      cvm_v        = .;
      cvm_p        = .;
      ks_v         = .;
      ks_p         = .;
   run;
   %if &data = %str() %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_specify_datat_error, noquote));
      %let abort  = 1;
      %let status = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_missing_DATA_error, noquote));
      %goto DONE;
   %end;

   /* Now check if there are any distributions left to fit (were none of them KRI fittable?)*/
   %if %trim(&dists) eq %str() %then %do;
      %let no_kri_stop = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_no_kri_fitable_dists, noquote));
      %goto DONE;
   %end;

   /*- If defaultTHRESHOLD is passed an empty string,
   set it equal to "essentially" 0 instead -*/
   %if &defaultTHRESHOLD = %str() OR &defaultTHRESHOLD = . %then %do;
      %let defaultTHRESHOLD = 0;
   %end;
   %else %if %sysevalf(&defaultTHRESHOLD LT 0.001, boolean) %then %do;
      %let defaultTHRESHOLD = 0;
   %end;

   /*- If THRESHOLD is passed as parameter,
   then put it in the dataset as a repeated value  -*/
   %if &threshold ^= %str() && &threshold ^= . %then %do;

      %if %sysevalf(&threshold < 0, boolean)  %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_be_neg_error, noquote));
         %let abort  = 1;
         %let status = 1;      
         %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_be_neg_error, noquote));
         %goto DONE; 
      %end;       

      %if %sysevalf(&THRESHOLD LT 0.001, boolean) %then %do;
         %let THRESHOLD = 0;
      %end;

      /* Add logs of losses and logs of threshold to the dataset     */
      /* Note: This is the reason why I said "essentiallly" 0 above  */
      data _LogWorkdata_(compress = binary);
         set &data end = LAST;
         where &whereClause;

         /* Keep original value when original threshold is ;arger than the cutoff */
         /* In all other cases, override threshold.                               */
         %if  %upcase(&Sev_Keep_Larger_Thresholds) = TRUE %then %do;
            if Threshold <= &threshold then do;
               THRESHOLD = &threshold;
            end;
         %end;
         %else %do;
            THRESHOLD = &threshold;
         %end;

         LogAMOUNT    = log(AMOUNT);
         LogTHRESHOLD = Safe_log(THRESHOLD);

         if last then do;
            call symputX("obscount", _N_);
         end;
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_set_thresh_error, noquote));
         %let abort  = 1;
         %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_set_thresh_error, noquote));
         %goto DONE;
      %end;

      %let minThreshold = &threshold;
   %end; /* THRESHOLD is passed as parameter */
   %else %do;   /*  Threshold is not passed as a parameter  */
      data _LogWorkdata_(compress = binary);
         set &data end = LAST; 
         where &whereClause;

         if THRESHOLD = . then do;
            THRESHOLD = &defaultThreshold;
         end;

         LogAMOUNT    = log(AMOUNT);
         LogTHRESHOLD = Safe_log(THRESHOLD);

         if last then do;
            call symputX("obscount", _N_);
         end;
      run;

      /* Because fitting may be sensitive to the order of data, sort from largest to smallest, 
         which should have smallest pdf values come first to help avoid catastrophic cancellation */
      proc sort data = _LogWorkdata_; 
         by descending AMOUNT; /* amount and logamount will be sorted in descending order */ 
      run;

      /* Get the minimum threshold value  */
      proc summary data = _LogWorkdata_ noprint;
         var threshold;
         output
               out = _junk_
               min = min
               ;
      run;

      data _null_;
         set _junk_;

         if _N_ = 1 then do;
            call symputX("minThreshold", put(min, BEST20.));
         end;
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_find_thresh_error, noquote));
         %let abort  = 1;
         %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_find_thresh_error, noquote));
         %goto DONE;
      %end;
   %end;
   /*- end of threshold testing -*/

   /* Make sure there is something in the dataset  */
   %let dsid     = %sysfunc(open(_LogWorkdata_, I));
   %let nparmobs = %sysfunc(attrn(&dsid, nlobs));
   %let rc       = %sysfunc(close(&dsid));

   %if %sysevalf(&nparmobs = 0, boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_no_obs_indat_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_no_obs_indat_error, noquote));
      %goto DONE;
   %end;

   /*- data sanity check AND drop when amount < threshold -*/
   data _LogWorkdata_(compress = binary);
      set _LogWorkData_ nobs = nObserv;
      retain NMiss 0 NNeg 0 NExcess 0;

      if missing(AMOUNT)then do;
         NMiss = NMiss + 1;
      end;

      if AMOUNT < 0 then do;
         NNeg = NNeg + 1;
      end;

      if AMOUNT > threshold then do;
         NExcess = NExcess + 1;
         output;
      end;

      if _N_ = nObserv then do;
         call symputX("NMiss",   NMiss);
         call symputX("NNeg",    NNeg);
         call symputX("NExcess", NExcess);   
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_check_use_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_check_use_error, noquote));
      %goto DONE;
   %end;

   %if %sysevalf(&Min_Fitting_Count > &nExcess, boolean) %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_not_enough_points, noquote));
      %let abort = 0;
      %let orva_sfit_skip = 1;
      %goto DONE;
   %end;

   %if %sysevalf(&NMiss >= 1, boolean) %then %do;
      %put &orvar_warning %sysfunc(sasmsg(&orvads, sfkri_missing_losses_error, noquote, &NNeg));
   %end;

   %if %sysevalf(&NNeg >= 1, boolean) %then %do;
      %put &orvar_warning %sysfunc(sasmsg(&orvads, sfkri_neg_losses_errror, noquote, &NNeg));
   %end;

   %if %sysevalf(&NExcess < 1, boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_not_above_thresh_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_invalid_obs_error, noquote));
      %goto DONE;
   %end;

   /* Obtain the list of KRI's that do not have missing values */
   %let numSuggestedKRIs = ;
   %orv_remove_missing_variable(
                                 dataset   = _LogWorkdata_,
                                 variables = &Sev_KRIs_suggested.,
                                 out_list  = Sev_KRIs_suggested,
                                 out_count = numSuggestedKRIs
                                 );
   /* Further prune that list to the variables that have non zero variance */
   %orv_remove_novar_variable( 
                              dataset   = _LogWorkdata_,
                              variables = &Sev_KRIs_suggested.,
                              out_list  = Sev_KRIs_suggested,
                              out_count = numSuggestedKRIs
                              );
   %let KRIs = &Sev_KRIs_suggested.;

   %let numForcedKRIs = ;
   %orv_remove_missing_variable(
                                 dataset   = _LogWorkdata_,
                                 variables = &Sev_Kris_Forced.,
                                 out_list  = Sev_Kris_Forced,
                                 out_count = numForcedKRIs
                                 );
   %orv_remove_novar_variable( 
                              dataset   = _LogWorkdata_,
                              variables = &Sev_Kris_Forced.,
                              out_list  = Sev_Kris_Forced,
                              out_count = numForcedKRIs
                              );

   %LEt  KRI_MODEL_NO_INTERCEPT = ;
   %LEt  KRI_MODEL              = ;
   %LEt  KRI_PARAMS             = ;
   %LEt  KRI_COMMA              = ;
   %let  COUNT_KRI              = 0;

   /* Check the kri counts against the number of observations */
   %let kri_count_status = ;
   %orv_adjustSeverityKriList( 
                              OBS_COUNT      = &nExcess,
                              N_SUGGESTED    = numSuggestedKRIs, 
                              SUGGESTED_LIST = Sev_KRIs_suggested, 
                              N_FORCED       = numForcedKRIs, 
                              FORCED_LIST    = Sev_Kris_Forced
                              );
   /* Check the status  */
   %if ( (&kri_count_status = STOP) or (&kri_count_status = UNKNOWN) ) %then %do;
      %let abort = 1;
      %goto DONE;
   %end;
   %else %do;
      /* Run the selection macro only if there are some suggested KRI's*/
      %if &kri_count_status NE NO_SUGG %then %do;
         /* kri_count_status is either 'OK' or 'SWITCH' */
         %let select_status = 0;
         %orva_sev_kri_select_&Sev_Selection_Method_Kris.();

         %if &select_status EQ 1 %then %do;
            %let abort = 1;
            %goto DONE;
         %end;  
      %end;
      %else %do;
      /* There are only forced KRIs */
         %let COUNT_KRI = &numForcedKRIs;
         %let KRIS = &Sev_Kris_Forced;
      %end;
   %end;

   %if &COUNT_KRI GT 0 %then %do;
      %let USEQLIM = true;
      %if &USEQLIM = true  %then %do;
         proc qlim 
                  data   = _LogWorkdata_ 
                  outest = _QLimParams_ 
                  covout 
                  noprint
                  ;
            model LogAMOUNT = &KRIs / truncated(LB = logthreshold);
         quit;

         data _linParams_;
            set _QLimParams_;
            where _TYPE_ = "PARM";

            length KRI_MODEL_NOI $10240. ;
            length KRI_MODEL $10240.;
            length KRI_INIT_PARM $20000.;

            array KRI_VARS[&COUNT_KRI] &KRIs; ;

            CountKri  = &COUNT_KRI;
            KRI_MODEL = "Intercept"; 
            
            KRI_MODEL_NOI = " ";
            KRI_INIT_PARM = " "; 

            do i = 1 to &COUNT_KRI;
               if missing(KRI_VARS[i]) then do;
                  CountKri = CountKri - 1;
               end;
               else do;
                  KRI_MODEL_NOI = trim(KRI_MODEL_NOI)  ||    " + "    || vname(KRI_VARS[i]) || " * Param_1_" || vname(KRI_VARS[i]);

                  if i = 1 then do;
                     KRI_INIT_PARM = trim(KRI_INIT_PARM)  || "Param_1_" ||trim(vname(KRI_VARS[i])) || "=" || trim( KRI_VARS[i]);
                  end;
                  else do;
                     KRI_INIT_PARM = trim(KRI_INIT_PARM) ||"," || "Param_1_" ||trim(vname(KRI_VARS[i])) || "=" || trim( KRI_VARS[i]);
                  end;
               end;
            end;

            if count_kri > 0 then do;
               plus = "+";
            end;
            else do;
               plus = " ";
            end;

            KRI_MODEL      = "Intercept" || plus || KRI_MODEL_NOI; 
            KRI_MODEL_YESI = put(Intercept, BEST20.) || plus || KRI_MODEL_NOI; 
            KRI_MODEL_NOI  = "(" || trim(KRI_MODEL_NOI) || ")";

            call symputX("COUNT_KRI", CountKri);
            call symput("The_Intercept", put(Intercept, BEST20.));
            call symput("TheKRI_MODEL", trim(KRI_MODEL));
            call symput("KRI_MODEL_NO_INTERCEPT", trim(KRI_MODEL_NOI));
            call symput("KRI_MODEL_YES_INTERCEPT", trim(KRI_MODEL_YESI));
            call symput("KRI_MODEL", trim(KRI_MODEL));
            call symput("KRI_INIT_PARM", trim(KRI_INIT_PARM));
         run;
      %end;
   %end;

   %put %sysfunc(sasmsg(&orvads, sfkri_note_header, noquote));
   %put KRI_MODEL_NO_INTERCEPT  = &KRI_MODEL_NO_INTERCEPT;
   %put KRI_MODEL               = &KRI_MODEL;
   %put KRI_PARAMS              = &KRI_PARAMS;
   %put COUNT_KRI               = &COUNT_KRI;
   %put KRIs                    = &KRIs;
   %put KRI_COMMA               = &KRI_COMMA;
   %put -><<>><<>><<>><<>><<>><<>><<>><<>><<>><<>><<>><<>><<>><<><-;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_calc_reg_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_calc_reg_error, noquote));
      %goto DONE;
   %end;

   %if &COUNT_KRI eq 0 %then %do; 
      /* No KRI's were significant so just call sfit_all and get out of here*/
      %ORVA_SFitALL( 
                    data             = &data,
                    whereClause      = &whereClause,
                    outparm          = &outparm, 
                    outcov           = &outcov,
                    parallel         = &parallel
                    );
      %return;
   %end; /* No KRI's were significant */


   /*- Build the linear fit parameters dataset-*/
   /* Rename the regression parameters to Param_1_.... */
   data _LinParams_(
                     keep = 
                           Intercept 
                           Param_: 
                           KRI_Select_Method
                           );
      set _LinParams_;
      length tmp $1024. ;
      KRI_Select_Method = "%upcase(&Sev_Selection_Method_Kris)";

      %if %sysevalf(&COUNT_KRI > 0, boolean) %then %do;
         /*- loop over KRI's -*/
         %let nkri = 1;
         %let kri  = %scan(&KRIs, &nkri, ' ');
         tmp = resolve(&kri);
         Param_1_&kri = input(tmp,Best20.);

         %do %while (&kri ne );  
            tmp          =  resolve(&kri);
            Param_1_&kri =  Input(tmp,Best20.);
            %let nkri    =  %eval(&nkri + 1);
            %let kri     =  %scan(&KRIs, &nkri, ' ');
         %end;
      %end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_store_params_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_store_params_error, noquote));
      %goto DONE;
   %end;

      /* Creation of the covariance matrix */
      %let AVG_ADJ_GOF = .;

      %if %sysevalf(&COUNT_KRI > 0, boolean) %then %do; /* COUNT_KRI > 0 */
         /* Joining the values of the estimated parameters to the rest of the dataset*/
         /* _LogWorkdata_Kri_ contains the original data, the log of the original data 
         the kri variables and the kri parameters */
         data _LogWorkdata_kri_(
                                 drop = 
                                       KRI_LOG_SUM 
                                       KRI_LOG2_SUM 
                                       KRI_LOG3_SUM 
                                       KRI_LOG4_SUM 
                                       KRI_SCALE_SUM 
                                       KRI_SCALE2_SUM 
                                       KRI_SCALE3_SUM 
                                       KRI_SCALE4_SUM
                                 compress = binary
                                 );
         
         set _LogWorkdata_ nobs = _nobs_;

         if _N_ eq 1 then do;
            set _LinParams_;
         end;
         
         retain   
               KRI_LOG_SUM     0 
               KRI_LOG2_SUM    0 
               KRI_LOG3_SUM    0 
               KRI_LOG4_SUM    0 
               KRI_SCALE_SUM   0 
               KRI_SCALE2_SUM  0 
               KRI_SCALE3_SUM  0 
               KRI_SCALE4_SUM  0
               %if &orig_gofthresh NE . %then %do; 
                  AVG_ADJ_GOF_T   0
               %end;
               ;

         KRI_LOG               =  Intercept + &KRI_MODEL_NO_INTERCEPT;
         KRI_LOG_SUM           =  KRI_LOG_SUM + KRI_LOG;
         KRI_LOG2_SUM          =  KRI_LOG2_SUM + KRI_LOG**2;
         KRI_LOG3_SUM          =  KRI_LOG3_SUM + KRI_LOG**3;
         KRI_LOG4_SUM          =  KRI_LOG4_SUM + KRI_LOG**4;

         KRI_SCALE             =  exp(KRI_LOG);
         KRI_SCALE_SUM         =  KRI_SCALE_SUM + KRI_SCALE;
         KRI_SCALE2_SUM        =  KRI_SCALE2_SUM + KRI_SCALE**2;
         KRI_SCALE3_SUM        =  KRI_SCALE3_SUM + KRI_SCALE**3;
         KRI_SCALE4_SUM        =  KRI_SCALE4_SUM + KRI_SCALE**4;

         KRI_ADJ_AMOUNT        =  AMOUNT / KRI_SCALE;
         KRI_ADJ_THRESHOLD     =  THRESHOLD / KRI_SCALE;
         KRI_ADJ_LogAMOUNT     =  LogAMOUNT - KRI_LOG;
         KRI_ADJ_LOGTHRESHOLD  =  LOGTHRESHOLD - KRI_LOG;

         %if "&orig_gofthresh" NE "." %then %do; 
            AVG_ADJ_GOF_T         = AVG_ADJ_GOF_T + &gofThreshold./KRI_SCALE;
         %end;

         if _N_ = _nobs_ then do;
            call symputX("KRI_LOG_AVG"    , put(KRI_LOG_SUM/_nobs_, BEST20.));
            call symputX("KRI_LOG2_AVG"   , put(KRI_LOG2_SUM/_nobs_, BEST20.));
            call symputX("KRI_LOG3_AVG"   , put(KRI_LOG3_SUM/_nobs_, BEST20.));
            call symputX("KRI_LOG4_AVG"   , put(KRI_LOG4_SUM/_nobs_, BEST20.));

            call symputX("KRI_SCALE_AVG"  , put(KRI_SCALE_SUM/_nobs_, BEST20.));
            call symputX("KRI_SCALE2_AVG" , put(KRI_SCALE2_SUM/_nobs_, BEST20.));
            call symputX("KRI_SCALE3_AVG" , put(KRI_SCALE3_SUM/_nobs_, BEST20.));
            call symputX("KRI_SCALE4_AVG" , put(KRI_SCALE4_SUM/_nobs_, BEST20.));


            %if "&orig_gofthresh" NE "." %then %do; 
               call symputX("AVG_ADJ_GOF_T"  , put(AVG_ADJ_GOF_T/_nobs_, BEST20.));
            %end;
            %else %do;
               call symput("AVG_ADJ_GOF_T"  , .);
            %end;
         end;
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_compute_kri_error, noquote));
         %let abort  = 1;
         %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_compute_kri_error, noquote));
         %goto DONE;
      %end;
   %end;      /*  COUNT_KRI > 0  */
   %else %do; /*  COUNT_KRI = 0  */
      data _LogWorkdata_kri_(compress = binary);
         set _LogWorkdata_ nobs = _nobs_;
         KRI_LOG              = 0;
         KRI_SCALE            = 1;
         KRI_ADJ_AMOUNT       = AMOUNT;
         KRI_ADJ_THRESHOLD    = THRESHOLD;
         KRI_ADJ_LogAMOUNT    = LogAMOUNT ;
         KRI_ADJ_LogTHRESHOLD = LogTHRESHOLD ;

         if _N_ = _nobs_ then do;
            call symputX("KRI_LOG_AVG"    , 0);
            call symputX("KRI_LOG2_AVG"   , 0);
            call symputX("KRI_LOG3_AVG"   , 0);
            call symputX("KRI_LOG4_AVG"   , 0);

            call symputX("KRI_SCALE_AVG"  , 1);
            call symputX("KRI_SCALE2_AVG" , 1);
            call symputX("KRI_SCALE3_AVG" , 1);
            call symputX("KRI_SCALE4_AVG" , 1);
         end;
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_add_todat_error, noquote));
         %let abort  = 1;
         %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_add_todat_error, noquote));
         %goto DONE;
      %end;
   %end; /*  COUNT_KRI = 0  */

   %let orig_TailBoundaryValue   = &TailBoundaryValue;
   %let orig_TailBoundaryPercent = &TailBoundaryPercent;

   /* If the tail boundary point is supplied, figure out      */
   /* what tail percentile it represents                      */      
   %if %upcase(&TailBoundaryChoice) EQ GET_HILL_ESTIMATE %then %do;
      %if %index(&dists, _GPD_) > 0 OR %index(&dists, LOGNGPD) > 0 %then %do;
         %orva_get_hill_estimate(
                                 data     = _LogWorkdata_KRI_,
                                 UseKri   = YES,
                                 tailpnts = 2000,
                                 hillplot = _hillplot
                                 );
      %end;
   %end;

   /* Get summary statistics for the loss variable*/
   proc summary 
               data   = _LogWorkdata_KRI_ 
               vardef = n 
               noprint
               ;
      var KRI_ADJ_AMOUNT;
      output 
            out = _stat_(keep = 
                                 obsCount 
                                 sample_mean 
                                 sample_variance 
                                 sample_skewness 
                                 sample_kurtosis 
                                 sample_uss 
                                 sample_q1 
                                 sample_q3
                                 )
            n        =  obsCount 
            mean     =  sample_mean 
            var      =  sample_variance 
            skewness =  sample_skewness 
            kurtosis =  sample_kurtosis 
            uss      =  sample_uss 
            q1       =  sample_q1 
            q3       =  sample_q3
            ;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_create_sum_stats_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_create_sum_stats_error, noquote));
      %goto DONE;
   %end;

   /* Get summary statistics for the logs of the loss variable*/
   proc summary 
               data   = _LogWorkdata_KRI_ 
               vardef = n 
               noprint
               ;
      var KRI_ADJ_LogAMOUNT;
      output 
            out = _log_stat_(keep = 
                                    sample_log_mean 
                                    sample_log_variance 
                                    sample_log_skewness 
                                    sample_log_kurtosis                          
                                    sample_log_uss 
                                    sample_log_q1 
                                    sample_log_q3
                                    )
            mean     = sample_log_mean 
            var      = sample_log_variance 
            skewness = sample_log_skewness
            kurtosis = sample_log_kurtosis
            uss      = sample_log_uss
            q1       = sample_log_q1
            q3       = sample_log_q3
            ; 
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_cant_trans_sumstats_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_cant_trans_sumstats_error, noquote));
      %goto DONE;
   %end;

   data _stat_ ;
      set _stat_;
      set _log_stat_;
   run;
     
   data _NULL_;
      now = datetime();
      call symput('_TimeStart_',now);
   run;
    
   %global ORVP_STATUS;
   %let ORVP_STATUS = 1;

   /********************** do dists in parallel *****************************/
   %if ("&parallel" eq "1") %then %do;
      %orvp_distribute_sfitkri(
                               cellData    = _LogWorkdata_KRI_, 
                               summaryData = _stat_, 
                               dists       = &dists
                               );

      %if(&ORVP_STATUS eq 0) %then %do;
         %orvu_append(
                      base = &outparm, 
                      new  = _parmnlp_
                      );
         %orvu_append(
                      base = &outcov, 
                      new  = _outcov_
                      );
      %end;
      %else %do;
         %put &orvar_warning %sysfunc(sasmsg(&orvads, sfkri_no_procs_error, noquote));
      %end;
   %end;
   
   /********************** loop over distributions *****************************/
   %if (&ORVP_STATUS ^= 0) %then %do;
      %let ndist = 1;
      %let dist  = %scan(&dists, &ndist, ' ');
      %let TailBoundaryChoice_temp  = &TailBoundaryChoice;
      %let TailBoundaryPercent_temp = &TailBoundaryPercent;
      %let TailBoundaryValue_temp   = &TailBoundaryValue;

      %do %while (&dist ne );
         %orva_sfitkri_dist( 
                            cellData    = _LogWorkdata_KRI_, 
                            dist        = &dist, 
                            ndist       = &ndist, 
                            summaryData = _stat_
                            );

         %orvu_append(
                      base = &outparm, 
                      new  = _parmnlp_&ndist
                      );       
         %orvu_append(
                      base = &outcov,  
                      new  = _covmat_&ndist
                      );
         %let ndist = %eval(&ndist + 1);
         %let dist  = %scan(&dists, &ndist, ' ');
         /* If any of these variables get changed during fitting for one distribution,
            we want them reset to their original values for the next */
         %let TailBoundaryChoice  = &TailBoundaryChoice_temp;
         %let TailBoundaryPercent = &TailBoundaryPercent_temp;
         %let TailBoundaryValue   = &TailBoundaryValue_temp;
      %end;
   %end;
   /************** End of the distribution loop ***************************************/

   /* Set missing numerical values in outcov to zero */
   data &outcov;
      length RowName $20;
      set &outcov;
      array nums _NUMERIC_;

      do i = 1 to dim(nums);
         if nums(i) = . then nums(i) = 0;
      end;

      if RowName = '' then do;
         RowName = _NAME_;
      end;

      drop i _NAME_;
   run;

   data _NULL_;
      now = datetime();
      elapsed = now - symget('_TimeStart_');
      call symput('_TimeElapsed_',elapsed);
   run;
   %put ORVA_SFit_Kri Elapsed Time: &_TimeElapsed_;
   
   /*- clean up SParm dataset -*/
   data &outparm;
      length TailBoundaryChoice $32;
      set &outparm; 

      if _N_ = 1 then do;
         delete; /* fake row in case of error */
      end;

      %if &orig_thresh = %str() %then %do; 
         threshold = .; 
      %end;
      %else %do; 
         threshold = &orig_thresh; 
      %end;

      gofThreshold     = &orig_gofthresh; 
      defaultThreshold = &orig_defthresh;
      percentThreshold = &orig_percthresh;
      Sev_Keep_Larger_Thresholds = "&Sev_Keep_Larger_Thresholds";

      TailBoundaryChoice  = "&TailBoundaryChoice  ";
      TailBoundaryValue   = &orig_TailBoundaryValue ;
      TailBoundaryPercent = &orig_TailBoundaryPercent ;
   run;
%DONE:
   
   %let gofThreshold     = &orig_gofthresh; 
   %let defaultThreshold = &orig_defthresh;
   %let percentThreshold = &orig_percthresh;
   %let threshold        = &orig_thresh;

   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_abnormal_error, noquote));

      data &outparm;
         set &outparm;
         status = 1;
         reason = "&reason";
      run;
   %end;

   %if &no_kri_stop %then %do;
      %put &orvar_warning &reason.;
      data &outparm;
         set &outparm;
         status = 1;
         reason = "&reason";
      run;
   %end;

   %if &orva_sfit_skip %then %do;
      data &outparm;
         set &outparm;
         status = 1;
         reason = "&reason";
      run;
   %end;

   %orvu_delete_data( data =
                            _stat_ 
                            _log_stat_ 
                            _logl_ 
                            _cs_ 
                            _parmnlp_: 
                            _covmat_ 
                            _SelectionParameters_ 
                            _linParams_ 
                            _initParams_ 
                            _tmp_ 
                            _estSummStat_     
                            _paramest_ 
                            _initParams 
                            _kri_params_
                            _junk_
                            _logworkdata_
                            _logworkdata_kri_
                            _covmat_KRI_tmp_
                            &outcov._KRI
                            _regResults_
                             );
%mend ORVA_SFit_KRI;

/*************************************************************************************
* PURPOSE: Take a list of suggested and forced KRIs, compare to number of observations
*          and decide what to do with the list.  This procedure should be common to all 
*          of the sev_kri selection methods.
*
* OUTPUT:  The macro variable kri_count_status should be defined outside of this 
*          macro.  It will be assigned one of these values:
*              UNKNOWN - Unable to determine what to do. Stop fitting.
*              STOP    - Insufficient observations to continue fitting - abort the fit
*              OK      - Everything is fine - continue to kri selection
*              SWITCH  - Forced KRIs are being treated like suggested KRIs, suggested
*                        KRIs are throewn out.
*              NO_SUGG - Ok to continue fitting, but there are only forced KRIs
*
*************************************************************************************/
%macro orv_adjustSeverityKriList( 
                                 OBS_COUNT      = , /* Pass the actual number */
                                 N_SUGGESTED    = , /* Pass the name of the variable */
                                 SUGGESTED_LIST = , /* Pass the name of the variable */
                                 N_FORCED       = , /* Pass the name of the variable */
                                 FORCED_LIST    =  /* Pass the name of the variable */
                                 );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;               
      %put @orv_adjustSeverityKriList(;
      %put .                           OBS_COUNT      = &OBS_COUNT.,;
      %put .                           N_SUGGESTED    = &N_SUGGESTED.,;
      %put .                           SUGGESTED_LIST = &SUGGESTED_LIST.,;
      %put .                           N_FORCED       = &N_FORCED.,;
      %put .                           FORCED_LIST    = &FORCED_LIST.;
      %put .                           );
   %end;

    /* Initialize the status macro */
   %let kri_count_status = UNKNOWN;
   
   /* Need at least 3 observations to consider fitting with KRIs.  Otherwise, stop */
   %if &OBS_COUNT <= 3 %then %do;
      let kri_count_status = STOP;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_insufficient_obs_error, noquote));
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfkri_insufficient_obs_error, noquote));
   %end;
   /* Next, compare the total KRI count to the number of observations.  If there
      are more observations than Forced KRIs plus 2 for parameters (plus 2 if there
      are suggested kris) then we're good to go */
   %else %if %eval(&&&N_FORCED + 2 + 2 * %eval(&&&N_SUGGESTED > 0) <=  &OBS_COUNT) %then %do;
      %if %eval(&&&N_SUGGESTED > 0) %then %do;
         %let kri_count_status = OK;
      %end;
      %else %do;
         %let kri_count_status = NO_SUGG;
      %end;
   %end;
   %else %if %eval(&&&N_FORCED > 0) %then %do; 
   /* If there are forced KRIs, but not enough observations to handle the forced and
      suggested KRIs, then throw out the suggested KRIs and treat the forced as 
      suggested.  */
      %let kri_count_status = SWITCH;
      %let reason = %sysfunc(sasmsg(&orvads, sfkri_too_many_forced_warning, noquote));
      %put &orvar_warning %sysfunc(sasmsg(&orvads, sfkri_too_many_forced_warning, noquote));
      %let &SUGGESTED_LIST   = &&&FORCED_LIST;
      %let &N_SUGGESTED      = &&&N_FORCED;
      %let &N_FORCED         = 0;
      %let &FORCED_LIST      = %str();
   %end;
%mend orv_adjustSeverityKriList;

%macro orv_remove_missing_variable(
                                   dataset   = ,
                                   variables = ,
                                   out_list  = ,
                                   out_count = 
                                   );

   %if &variables. eq %str() %then %do;
      %let &out_list. = %str();
      %let &out_count. = 0;
   %end;
   %else %do;
      %local number_of_variables;
      %let number_of_variables = %sysfunc(countw(&variables., " "));
      data _rmv_temp_dataset_;
         array list[*] &variables.;    
         array flags[&number_of_variables.] Flag1-Flag&number_of_variables.;      
         retain flags;
         set &dataset. end = last;
         do i = 1 to dim(list);
            if list(i) = . then do;
               flags(i) = 1;
            end;
         end;
         if last then do;
            output;
         end;
      run;

      data _rmv_temp_dataset_;
         set _rmv_temp_dataset_;
         length _name_ $500;
         array list[*] &variables.; 
         array flags[&number_of_variables.] Flag1-Flag&number_of_variables.; 

         count = 0;
         do i = 1 to &number_of_variables.;
            if flags(i) ne 1 then do;
               _name_ = vname(list(i));
               count = count + 1;
               output;
            end;
         end;
         call symputX("&out_count.", count);
         
         keep _name_;
      run;

      proc sql noprint;
         SELECT
                  _name_
         INTO
                  :&out_list. separated by " "
         FROM
                  _rmv_temp_dataset_
         ;
      quit;

      %orvu_delete_data( data = _rmv_temp_dataset_ );
   %end;
%mend orv_remove_missing_variable;

/* Remove variables that don't have any variance ----------------*/
%macro orv_remove_novar_variable(
                                 dataset   = ,
                                 variables = ,
                                 out_list  = ,
                                 out_count = 
                                 );

   %if &variables. eq %str() %then %do;
      %let &out_list. = %str();
      %let &out_count. = 0;
   %end;
   %else %do;
      %local number_of_variables;
      %let number_of_variables = %sysfunc(countw(&variables., " "));
      data _rmv_temp_dataset_;
         array list[*] &variables.; 
         length _name_ $32;   
         array flags[&number_of_variables.] Flag1-Flag&number_of_variables.;      
         retain flags;
         set &dataset. end = last;

          /* Make sure lags are setup -----------------*/
         do i = 1 to dim(list);
            lagx = lag(list[i]);
         end;

         if _n_ ne 1 then do;
            do i = 1 to dim(list);
                         /* if the value changes at any point note it */
               if list[i] ne lag(list[i] ) then flags[i] = 1;
               
            end;
         end;

         if last then do;
            count = 0;
            do i = 1 to &number_of_variables.;
               if flags(i) = 1 then do;
                  _name_ = vname(list(i));
                  count = count + 1;
                  output;
               end;
            end;
            call symputX("&out_count.", count);   
           
         end;

          keep _name_;
      run;

      proc sql noprint;
         SELECT
                  _name_
         INTO
                  :&out_list. separated by " "
         FROM
                  _rmv_temp_dataset_
         ;
      quit;

      %orvu_delete_data( data = _rmv_temp_dataset_ );
   %end;
%mend orv_remove_novar_variable;
