/*************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME   : ORVA_SFitAll.sas                                             
 * PURPOSE: fit operational risk severity distributions             
 *          Currently this macro should only be called by %orva_sfit.            
 * SUPPORT: jariou
 *
 * Please see the accompanying doc for pdfs and cdfs of the distributions
 *
 * PARAMETERS :
 *
 *   DATA             = Specifies the name of the input data set to be analyzed.
 *   WHERECLAUSE      = Specifies which rows in Data are used in the fitting.
 *   OUTPARM          = Name of the output data set containing parameter estimates.
 *   PARALLEL         = 1: use parallel processing
 *                      0: don't use parallel processing
 *   OUTCOV           = name of the output covarinace dataset.
 *   
 * NOTES: Please see the accompanying doc for pdfs and cdfs of the distributions
 *
 **************************************************************************************/
%macro ORVA_SFitAll( 
                    data         = ,
                    whereClause  = ,
                    outparm      = ,
                    outcov       = ,
                    parallel     = 0
                    );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put #ORVA_SFitAll(; 
      %put .              data         = &data.,;
      %put .              whereClause  = &whereClause.,;
      %put .              outparm      = &outparm.,;
      %put .              outcov       = &outcov.,;
      %put .              parallel     = &parallel.;
      %put .              );
   %end;

   %let dists = %upcase(&dists);

   /* save thresholds for parm dataset */
   %let orig_thresh     = &threshold;
   %let orig_gofthresh  = &gofthreshold;
   %let orig_defthresh  = &defaultthreshold;
   %let orig_percthresh = &thresholdPercent;

   /*-------------------------------------------------
   * Prepare output data sets
   * Note:
   *   AIC: Akaike Information Criterion
   *   SBC: Schwartz Bayesian Criterion
   *------------------------------------------------*/
   data &outcov; 
      set _NULL_; 
   run;

   data &outparm;
      format 
            distribution $12. 
            reason &ORVAR_REASON_FMT  
            Sev_Keep_Larger_Thresholds $5.
            ;
      distribution = "";
      status       = 1;
      reason       = "%sysfunc(sasmsg(&orvads, unknown_error, noquote))";
      mean         = .;
      std          = .;
      skewness     = .;
      kurtosis     = .;
      logl         = .;
      aic          = .;
      sbc          = .;
      ad_v         = .;
      ad_p         = .;
      cvm_v        = .;
      cvm_p        = .;
      ks_v         = .;
      ks_p         = .;
   run;
   %if &data = %str()  %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_specify_data_error, noquote));
      %let abort  = 1;
      %let status = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_missing_DATA_error, noquote));
      %goto DONE;
   %end;

   %if &defaultThreshold = %str() %then %do;
      %let defaultThreshold = 0.000000001;
   %end;

   data _workdata_;
      set &data; 
      where &whereClause;

      if missing(THRESHOLD) then do;
         THRESHOLD = &defaultThreshold;
      end;
   run;

   %let dsid     = %sysfunc(open(_workdata_, I));
   %let nparmobs = %sysfunc(attrn(&dsid, nlobs));
   %let rc       = %sysfunc(close(&dsid));

   %if &nparmobs = 0 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_no_obs_indat_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_no_obs_indat_error, noquote));
      %goto DONE;
   %end;

   /*- If threshold is passed as parameter,
   then put it in the dataset as a repeated value  -*/
   %if &threshold ^= . %then %do;
      %if %sysevalf(&threshold <0, boolean)  %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_cant_be_neg_error, noquote));
         %let abort  = 1;
         %let status = 1;      
         %let reason = %sysfunc(sasmsg(&orvads, sfa_THRESHOLD_neg_error, noquote));
         %goto DONE; 
      %end;       

      data _workdata_;
         set _workData_;

         /* Keep original value when original threshold is ;arger than the cutoff */
         /* In all other cases, override threshold.                               */
         %if  %upcase(&Sev_Keep_Larger_Thresholds) = TRUE %then %do;
            if Threshold <= &threshold then do;
               THRESHOLD = &threshold;
            end;
         %end;
         %else %do;
            THRESHOLD = &threshold;
         %end;
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_cant_set_thresh_error, noquote));
         %let abort  = 1;
         %let reason = %sysfunc(sasmsg(&orvads, sfa_cant_set_thresh_error, noquote));
         %goto DONE;
      %end;

      %let minThreshold = &threshold;
   %end;
   %else %do;
      proc summary data = _workData_;
         output 
               min = minthreshold 
               out = _junk_
               ;
         var THRESHOLD;
      run;    

      data _null_;
         set _junk_;
         call symput("minThreshold", put(minthreshold, BEST20.));
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_cant_find_thresh_error, noquote));
         %let abort  = 1;
         %let reason = %sysfunc(sasmsg(&orvads, sfa_cant_find_thresh_error, noquote));
         %goto DONE;
      %end;
   %end;
    /*- end of threshold testing -*/

   /*  Now lets look at the goodness of fit threshold*/
   %if &gofThreshold ^= . %then %do;
      %if %sysevalf(&gofThreshold < &minThreshold,boolean) %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_greater_error, noquote));
         %let abort  = 1;
         %let status = 1;      
         %let reason = %sysfunc(sasmsg(&orvads, sfa_GOFTHRESHOLD_low_error, noquote));
         %goto DONE; 
      %end;

      %if %sysevalf(&gofThreshold<0,boolean)  %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_pos_error, noquote));
         %let abort  = 1;
         %let status = 1;      
         %let reason = %sysfunc(sasmsg(&orvads, sfa_GOFTHRESHOLD_neg_error, noquote));
         %goto DONE; 
      %end;       
   %end;
   %else %do;
      %let gofThreshold = &minThreshold;
   %end;

   /*- Make sure only one variable name is passed -*/
   %if %length(%scan(AMOUNT,2))%then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_one_allowed_error, noquote));
      %let abort  = 1;
      %let status = 1;      
      %let reason = %sysfunc(sasmsg(&orvads, sfa_one_allowed_error, noquote));
      %goto DONE;
   %end;

   /*- data sanity check -*/
   data _null_;
      set _workData_ nobs = nObserv;
      retain 
            NMiss   0 
            NNeg    0 
            NExcess 0
            ;

      if missing(AMOUNT) then do;
         NMiss = NMiss + 1;
      end;

      if AMOUNT < 0 then do;
         NNeg = NNeg + 1;
      end;

      if AMOUNT > THRESHOLD then do;
         NExcess = NExcess + 1;
      end;

      if _N_ = nObserv then do;
         call symput("NMiss", NMiss);
         call symput("NNeg", NNeg);
         call symput("NExcess", NExcess);   
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do; 
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_cant_check_use_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_cant_check_use_error, noquote));
      %goto DONE;
   %end;

   %if %sysevalf(&Min_Fitting_Count > &nExcess, boolean) %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_not_enough_points, noquote));
      %let abort = 0;
      %let orva_sfit_skip = 1;
      %goto DONE;
   %end;

   %if %sysevalf(&NMiss >= 1, boolean) %then %do;
      %put &orvar_warning %sysfunc(sasmsg(&orvads, sfa_miss_loss_error, noquote));
      %let abort  = 0;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_miss_loss_error, noquote));
   %end;

   %if %sysevalf(&NNeg >= 1, boolean) %then %do;
      %put &orvar_warning %sysfunc(sasmsg(&orvads, sfa_neg_loss_error, noquote));
      %let abort  = 0;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_neg_loss_error, noquote));
   %end;

   %if %sysevalf(&NExcess < 1, boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_not_above_thresh_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_not_above_thresh_error, noquote));
      %goto DONE;
   %end;

   /* Because fitting may be sensitive to the order of data, sort from largest to smallest, 
      which should have smallest pdf values come first to help avoid catastrophic cancellation */
   proc sort data = _workData_; 
      by descending AMOUNT; 
   run;

   data _LogWorkdata_;
      set _workData_ end = last;
      where AMOUNT > THRESHOLD;
      LogAMOUNT = log(AMOUNT);

      if last then do;
         call symput("obscount", _N_);
      end;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_cant_subset_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_cant_subset_error, noquote));
      %goto DONE;
   %end;

   %let orig_TailBoundaryValue   = &TailBoundaryValue;
   %let orig_TailBoundaryPercent = &TailBoundaryPercent;

   %if %upcase(&TailBoundaryChoice) EQ GET_HILL_ESTIMATE %then %do;
      %if %index(&dists, _GPD_) > 0 OR %index(&dists, LOGNGPD) > 0 %then %do;
         %orva_get_hill_estimate(
                                 data     = _LogWorkdata_,
                                 UseKri   = NO,
                                 tailpnts = 2000,
                                 hillplot = _hillplot
                                 );
      %end;
   %end;

   /* For incremental load and  migration, skip this */
   proc summary 
               data   = _LogWorkdata_  
               vardef = n 
               noprint
               ;
      var AMOUNT;
      output 
      out = 
            _stat_(keep =   
                           obsCount 
                           sample_mean 
                           sample_variance 
                           sample_skewness 
                           sample_kurtosis 
                           sample_uss 
                           sample_q1 
                           sample_q3
                           )
            n        =  obsCount 
            mean     =  sample_mean 
            var      =  sample_variance 
            skewness =  sample_skewness 
            kurtosis =  sample_kurtosis 
            uss      =  sample_uss 
            q1       =  sample_q1 
            q3       =  sample_q3
            ;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_cant_create_sum_stats_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_cant_create_sum_stats_error, noquote));
      %goto DONE;
   %end;
       
   proc summary 
               data   = _LogWorkdata_  
               vardef = n 
               noprint
               ;
      var LogAMOUNT;
      output 
         out = 
               _log_stat_( keep =   
                                 sample_log_mean 
                                 sample_log_variance 
                                 sample_log_skewness 
                                 sample_log_kurtosis 
                                 sample_log_uss 
                                 sample_log_q1 
                                 sample_log_q3
                                 )
               mean     =  sample_log_mean 
               var      =  sample_log_variance 
               skewness =  sample_log_skewness 
               kurtosis =  sample_log_kurtosis 
               uss      =  sample_log_uss 
               q1       =  sample_log_q1 
               q3       =  sample_log_q3
               ;
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_cant_trans_stats_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, sfa_cant_trans_stats_error, noquote));
      %goto DONE;
   %end;
         
   data _stat_;
      set _stat_;
      set _log_stat_;
   run;
   /* End of migration/incremental load */

   /********************** loop over distributions *****************************/
   data _NULL_;
      now = datetime();
      call symput('_TimeStart_', now);
   run;

   %global ORVP_STATUS;
   %let ORVP_STATUS = 1;

   %if "&parallel" eq "1" %then %do;
      %orvp_distribute_sfit(
                            cellData    = _LogWorkdata_,
                            summaryData = _stat_,
                            initData    = &initData,
                            dists       = &dists
                            );

      %if (&ORVP_STATUS eq 0) %then %do;
        %orvu_append(
                     base = &outparm, 
                     new = _parmnlp_
                     );

        %if &outcov ne %str() %then %do;
           %orvu_append(
                        base = &outcov, 
                        new  = _outcov_
                        );
        %end;
      %end;
      %else %do;
         %put &orvar_warning %sysfunc(sasmsg(&orvads, sfa_no_procs_error, noquote));
      %end;
   %end;

   %if (&ORVP_STATUS ^= 0) %then %do;
      %let ndist = 1;
      %let dist  = %scan(&dists, &ndist, ' ');
      %let TailBoundaryChoice_temp  = &TailBoundaryChoice;
      %let TailBoundaryPercent_temp = &TailBoundaryPercent;
      %let TailBoundaryValue_temp   = &TailBoundaryValue;

      %do %while ("&dist" ne "");
         %orva_sfit_dist(
                         cellData    = _LogWorkdata_,
                         dist        = &dist,
                         ndist       = &ndist,
                         summaryData = _stat_
                         );
         %orvu_append(
                      base = &outparm, 
                      new  = _parmnlp_&ndist
                      );

         %if &outcov ne %str() %then %do;
            %orvu_append(
                         base = &outcov, 
                         new = _covmat_&ndist
                         );
         %end;
 
         %let ndist = %eval(&ndist + 1);
         %let dist  = %scan(&dists, &ndist, ' ');

         /* If any of these variables get changed during fitting for one distribution,
            we want them reset to their original values for the next */
         %let TailBoundaryChoice  = &TailBoundaryChoice_temp;
         %let TailBoundaryPercent = &TailBoundaryPercent_temp;
         %let TailBoundaryValue   = &TailBoundaryValue_temp;
      %end;
   %end;

   data _NULL_;
      now = datetime();
      elapsed = now - symget('_TimeStart_');
      call symput('_TimeElapsed_', elapsed);
   run;

   %put ORVA_SFitAll Elapsed Time: &_TimeElapsed_;
   /************** End of the distribution loop ***************************************/

   /*- clean up SParm dataset -*/
   data &outparm;
      length TailBoundaryChoice $32;
      set &outparm; 

      if _N_ = 1 then do;
         Delete; /* fake row in case of error */
      end;

      %if &orig_thresh = %str() %then %do; 
         threshold = .; 
      %end;
      %else %do; 
         threshold = &orig_thresh; 
      %end;
   
      gofThreshold     = &orig_gofthresh; 
      defaultThreshold = &orig_defthresh;
      percentThreshold = &orig_percthresh;
      Sev_Keep_Larger_Thresholds = "&Sev_Keep_Larger_Thresholds";
      COUNT_KRI = 0;
      TailBoundaryChoice  = "&TailBoundaryChoice  ";
      TailBoundaryValue   = &orig_TailBoundaryValue ;
      TailBoundaryPercent = &orig_TailBoundaryPercent ;
   run;

%DONE:
   %let gofThreshold     = &orig_gofthresh; 
   %let defaultThreshold = &orig_defthresh;
   %let percentThreshold = &orig_percthresh;
   %let threshold        = &orig_thresh;

   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfa_abnormal_error, noquote));
      %let dist = %scan(&dists, 1, ' ');

      data &outparm;
         set &outparm(obs = 1);
         status = 1;
         reason = "&reason"; 
         distribution = "&dist";
      run;
   %end;

   %if &orva_sfit_skip %then %do;
      data &outparm;
         set &outparm;
         status = 1;
         reason = "&reason";
      run;
   %end;

   %orvu_delete_data( data =
                             _workdata_ 
                             _stat_ 
                             _logl_ 
                             _cs_ 
                             _parmnlp_ 
                             _covmat_ 
                             _tmp_ 
                             _logworkdata_
                             );
%mend ORVA_SFitAll;
