/*****************************************************************************************
 * Copyright (C), 2010
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   NAME:    orva_sim_scenario_certain.sas                                            
 *
 *   PURPOSE: Computes scenario simulation or certain scenarios.
 *
 *   SUPPORT: jariou            
 *   
 *   PARAMETERS: 
 *              simCount = the number of values to be simulated
 *              dataIn   = dtaset passed  in
 *              scenSpec = scenario specification type, i.e. Model=M or Bucket=B
 *              seed     = seed
 *              dataOut  = output dataset 
 ***********************************************************************************/
%macro orva_sim_scenario_certain(
                                 simCount        = , 
                                 dataIn          = , 
                                 scenSpec        = , 
                                 seed            = , 
                                 scenKind        = , 
                                 insuranceEffect = VARPRJ.Ins_effect,
                                 dataOut         =
                                 ); 
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_sim_scenario_certain(;
      %put .                          simCount        = &simCount.,; 
      %put .                          dataIn          = &dataIn.,; 
      %put .                          scenSpec        = &scenSpec.,; 
      %put .                          seed            = &seed.,; 
      %put .                          scenKind        = &scenKind.,; 
      %put .                          insuranceEffect = &insuranceEffect.,;
      %put .                          dataOut         = &dataOut.;
      %put .                          ); 
   %end;

   %if %upcase( &scenSpec ) = M %then %do; 
      %orv_scenarioSimCertainModel(
                                    simCount        = &simCount, 
                                    dataIn          = &dataIn, 
                                    seed            = &seed, 
                                    insuranceEffect = &insuranceEffect,
                                    dataOut         = &dataOut
                                    ); 
   %end;
   %else %do;
      %orv_scenarioSimCertainBucket(
                                    simCount        = &simCount, 
                                    dataIn          = &dataIn, 
                                    seed            = &seed, 
                                    insuranceEffect = &insuranceEffect,
                                    dataOut         = &dataOut
                                    ); 
   %end;

   %if %upcase( &scenKind ) NE RARE %then %do;
      proc sort 
               data = &dataOut 
               out  = &dataOut
               ;
         by 
            CCUID 
            _rep_
            ;
      run;

      proc summary data = &dataOut ;
         by 
            ccuid 
            _rep_
            ;
         var 
            Event_Count 
            Loss_count  
            Loss_Amount
            ;
         output
               out = &dataOut( keep = 
                                       CCUID 
                                       _REP_ 
                                       Event_Count 
                                       Loss_count  
                                       Loss_Amount
                                       )
               sum = 
                     Event_Count 
                     Loss_count  
                     Loss_Amount
                     ;
      run;
   %end;
%mend orva_sim_scenario_certain; 

/********************************************************/
/* Case where the scenario specification is ModelBased  */
/********************************************************/
%macro orv_scenarioSimCertainModel(
                                    simCount        = , 
                                    dataIn          = , 
                                    seed            = , 
                                    insuranceEffect = ,
                                    dataOut         =
                                    ); 
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_scenarioSimCertainModel(;
      %put .                             simCount        = &simCount.,; 
      %put .                             dataIn          = &dataIn.,; 
      %put .                             seed            = &seed.,; 
      %put .                             insuranceEffect = &insuranceEffect.,;
      %put .                             dataOut         = &dataOut.;
      %put .                             ); 
   %end;

/*   proc sort data = &insuranceEffect;*/
/*      by CCUID;*/
/*   run;*/
   
   data _NULL_;
      call symput( "dataDataIn", scan(translate("&dataIn", '.', ' '), 2));
      call symput( "dataInsuranceEffect", scan(translate("&insuranceEffect", '.', ' '), 2));
   run;

   proc sql noprint;
      CREATE TABLE   
                     __tmp__
      AS SELECT
                     *
      FROM
                     &dataIn L 
      LEFT JOIN
                     &InsuranceEffect( WHERE = (ccuid ne 0)) R
             ON
                     L.CCUID = R.CCUID
                     ;
                           
   quit;

   data &dataOut;
      set __tmp__;
      call streaminit( &SEED. );

      do _rep_ = 1 to &simCount;
         Event_Count = FreqSim( Frequency_Dist, &theFreqSimCallParams );
         Loss_Count  = Event_Count;
         Loss_Amount = 0;  

         /* Accumulate the severity amounts for each iteration */
         do i = 1 to Loss_Count;
            Loss_Amount = SevSim(Severity_Dist, &theSevSimCallParams);
            output;
         end;
      end;

      keep  SCENARIO_UID
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT
            ;
   run;

%mend orv_scenarioSimCertainModel;

/********************************************************/
/* Case where the scenario specification is BucketBased */
/********************************************************/
%macro orv_scenarioSimCertainBucket(
                                    simCount        = , 
                                    dataIn          = , 
                                    seed            = , 
                                    insuranceEffect = ,
                                    dataOut         =
                                    );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orv_scenarioSimCertainBucket(;
      %put .                              simCount        = &simCount.,; 
      %put .                              dataIn          = &dataIn.,; 
      %put .                              seed            = &seed.,; 
      %put .                              insuranceEffect = &insuranceEffect.,;
      %put .                              dataOut         = &dataOut.;
      %put .                              );
   %end;

   /*****************************************************************/
   /*   If this is a bucket scenario, I need to look at the dataset */
   /*   that is named VARPRJ??.scenario_plotbrkdwn_XXXX             */
   /*****************************************************************/
   %let _tailNumber = %index(&dataIn,_);
   %let _tailString = %substr(&dataIn, %eval(&_tailNumber + 1));

   %do %while (&_tailNumber GT 0);
      %let _lastTailNumber = &_tailNumber;
      %let _lastTailString = &_tailString;
      %let _tailNumber = %index(&_lastTailString,_);
      %let _tailString = %substr(&_lastTailString, %eval(&_tailNumber + 1));
   %end;

   %let _scenLoc = %index(%upcase(&dataIn), SCENARIO_);

   %if &_scenLoc > 0 %then %do;
      %let dataIn = %substr(&dataIn,1,&_scenLoc + 8)plotbrkdwn_&_tailString;
   %end;

   /* I need to figure out the size fo the array*/
   data _NULL_;
      set &dataIn ;
      retain maxBucketCount 0;
      array buckWeights bucketwgt:;
      maxBucketCount = max( maxBucketCount , nBuckets);
      call symputX( "buckWeightsSize", maxBucketCount );
      call symputX( "buckSize", maxBucketCount + 1 );
   run;   
   
   proc sort data = &insuranceEffect;
      by CCUID;
   run;
            
   proc sql noprint;
      CREATE TABLE 
                  &dataOut(drop = ccuidX)
      AS SELECT
                  ScenEffect.*,
                  InsEffect.*
      FROM
                  &dataIn(rename=(ccuid = ccuidX)) ScenEffect
      LEFT JOIN
                  
                  &insuranceEffect InsEffect
            ON
                  ScenEffect.ccuidX = InsEffect.ccuid
            ;
   quit;
   
   data &dataOut;
      set &dataOut;
      call streaminit(&SEED.);
      
      array    BucketWeights[*] bucketwgt1-bucketwgt&buckWeightsSize;
      array tmpBucketWeights[&buckWeightsSize] _temporary_;

      array    Buckets[*] bucket0-bucket&buckWeightsSize;
      array tmpBuckets[&buckSize] _temporary_;

      do i = 1 to &buckSize;
         tmpBuckets[i] = Buckets[i];
      end;

      do i = 1 to &buckWeightsSize;
         tmpBucketWeights[i] = BucketWeights[i];
      end;

      call NormalizeWeights( 
                              tmpBucketWeights,
                              nBuckets
                              );
 
      call BucketFreqSimDef(  
                              Frequency_Mean,
                              Frequency_Min,
                              Frequency_Max,
                              Frequency_dist,
                              FParam_1,
                              FParam_2
                              );

      do _rep_ = 1 to &simCount;
         /* Simulate frequencies as in the model case now that the */
         /* call to "BucketFreqSimDef" created the correct model   */
         Event_Count = FreqSim(
                                 Frequency_Dist, 
                                 &theFreqSimCallParams
                                 );
         Loss_Count  = Event_Count;
         Loss_Amount = 0;

         /* Accumulate the severity amounts for each iteration */
         do i = 1 to Loss_Count;
            Loss_Amount  = BucketSevSim( 
                                        Seed,
                                        tmpBuckets,
                                        tmpBucketWeights,
                                        nBuckets
                                        );
            output;
         end;
      end;

      keep  
            SCENARIO_UID 
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT
            iLoss
            ;
   run;
%mend orv_scenarioSimCertainBucket;
