/******************************************************************************
 * Copyright (C), 2008
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 *
 * NAME:     orva_var_gradient.sas
 * PURPOSE:  Calculate the gradient of the VaR values with respect to the parameters. 
 *           Called by %orva_conf_interval().
 * SUPPORT:  stmajo
 *
 * Parameters:
 *
 *   ccUid              = key for this cell.
 *   Sparm_dataset_name = Name of the input data set containing severity parameter estimates.
 *   Fparm_dataset_name = Name of the input data set containing frequency parameter estimates.
 *   SimInterval        = Interval used for frequency simulation
 *   varstat            = Used to get the var numbers for the cell 
 *   freq_gradient      = The name of the output dataset (for frequency) for the ccUid.
 *   sev_gradient       = The name of the output dataset (for severity) for the ccUid.
 *
 * NOTE: This is, for intents and purposes, a submacro of orva_conf_interval.  It is not 
 *       designed to be used as a stand alone macro.  Any errors are passed through to 
 *       orva_conf_interval for handling.  Error messages are not written to the log 
 *       by this macro.
 ***********************************************************************************/
%macro orva_var_gradient( 
                           ccUid              = ,
                           SimInterval        = ,
                           varstat            = ,
                           maxloss            = ,
                           Sparm_dataset_name = ,
                           Fparm_dataset_name = , 
                           freq_gradient      = ,
                           sev_gradient       = 
                           );

    %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_var_gradient(; 
      %put .                   ccUid              = &ccUid.,;
      %put .                   SimInterval        = &SimInterval.,;
      %put .                   varstat            = &varstat.,;
      %put .                   maxloss            = &maxloss.,;
      %put .                   Sparm_dataset_name = &Sparm_dataset_name.,;
      %put .                   Fparm_dataset_name = &Fparm_dataset_name.,; 
      %put .                   freq_gradient      = &freq_gradient.,;
      %put .                   sev_gradient       = &sev_gradient.;
      %put .                   );
   %end;

   /* Get the list of evaluation points  */
   %let p_1 = none;

   data _vg_var_; 
      set &varstat;
      where CCUID = &CCUID;

      call symputX("P_1", put(p_1, BEST20.));
      call symputX("P_2", put(p_2, BEST20.));
      call symputX("P_3", put(p_3, BEST20.));

      var_value = p_1;
      quantile  = pctl1;
      output;

      var_value = p_2;
      quantile  = pctl2;
      output;

      var_value = p_3;
      quantile  = pctl3;
      output;

      keep 
            var_value 
            quantile
            ;
   run;

   /*Check that the VaR values were recovered */
   %if &p_1 eq none %then %do;  
      %let ORV_ABORT  = &ORV_TRUE;
      %let ORV_REASON =  vargrad_no_var_values_error;
      %goto DONE;
   %end;

   /**************************************************/
   /*  DETERMINE DISCRETIZATION INTERVAL INFORMATION */
   /**************************************************/

   /* Determine the largest VaR value */
   data _NULL_;
      V = max(&P_1, &P_2, &P_3);
      call symput("Var_Max", put(V,Best20.));
   run;
   /* If all VaR are 0 or missing, then skip to the end */
   %if %sysevalf(&Var_Max le 0, boolean) %then %do;
      %let ORV_ABORT = &ORV_TRUE;
      %let found_error = &ORV_FALSE;
      %let ORV_REASON =  vargrad_no_var_gt_zero_error;
      %goto DONE;
   %end;

   /* Check the largest VaR value against 99.99% quantile of the severity distribution.  Use the larger of the 2
      to determine the region over which to perform the discretization.  This becomes important when the severity
      scale is so large that the maxloss truncates most observations. For severity mixtures, use the model
      with the larget scale parameter. */
   %let sev_upper_bnd = NOT_COMPUTED;
   data _temp_;
      threshold = 0;
      ecdf = .9999;
      output;
   run;
   %if %lowcase(&scale_transform.) eq id %then %do;
      %let param_1 = &MAX_SCALE.;
   %end;
   %else %do;
      %let param_1 = %sysfunc(log(&max_scale.));
   %end;
   %do i=2 %to &SPARAMCOUNT.;
      %let PARAM_&i. = &&Sparam_&i.;
   %end;
   %QUANT&SDIST();
   data _NULL_;
      set _temp_;
      call symputx("sev_upper_bnd",put(quant,BEST20.));
   run;  

   %if &sev_upper_bnd. eq NOT_COMPUTED %then %do;
      %let ORV_ABORT = &ORV_TRUE;
      %let found_error = &ORV_TRUE;
      %let ORV_REASON =  vargrad_severity_upper_bound_not_computed;
      %goto DONE;
   %end;
   
   /* If maxloss is less than the 99.99% quantile, then use maxloss
      as the upper bound here.  All the mass to the right of maxloss will get
      set to 0 */
   %let sev_upper_bnd = %sysfunc(min(&sev_upper_bnd , &maxloss));

   /* Use the larger of VaR and sev_upper_bound as the upper bound */
   %let upper_bnd = %sysfunc(max(&sev_upper_bnd , &Var_Max));

   /* Will have 1 scale parameter for each stratum, so compute total # of severity parameters 
      in the mixture model */
   %let SPARM_TOT = %eval(&SPARAMCOUNT + &NSEV_MIX -1);

   /* total number of columns we will need in results */
   %let tot_cols = %eval(&SPARM_TOT + &FPARAMCOUNT + 2);

   /* Determine the number of intervals to use in discretization */
   /* use 2**n points. */
   %let pwr = 11;
   %let npoints = %eval(2**&pwr);
   %let h = %sysevalf((&upper_bnd/(&npoints/2)));

   /* Min_Tail_Boundary separates the body from the tail.  For mixdist, this is literally the tail boundary point.
      For others, it is the 95% quantile (somewhat arbitrarily chosen).  We want at least 10 points below this
      boundary to get meaningful results.  If none are below, then mixdist causes errors, since derivatives for
      the body are missing */
   %let Min_Tail_Boundary = -1;
   %if %trim(%upcase(&SDIST)) eq MIXDIST or %trim(%upcase(&SDIST)) eq LOGNGPD or %trim(%upcase(&SDIST)) eq LOGNGPD1 %then %do;
   /* Loop over the severity distributions (may have a mixture of several when there are KRIs) and find the 
      smallest body/tail boundary value. */
      %let scale_param = %scan(&SEV_SCALES, 1, ' ');
      %let Min_Tail_Boundary = %sysevalf(&Param_xLimit * &scale_param);
      %do w=2 %to &NSEV_MIX;
         %let scale_param = %scan(&SEV_SCALES, &w, ' ');
         %let Min_Tail_Boundary = %sysfunc(min(%sysevalf(&Param_xLimit * &scale_param),&Min_Tail_Boundary));
      %end;
   %end;
   %else %do;
      /* Use the 95% quantile as the body/tail boundary */
      data _temp_;
         threshold = 0;
         ecdf = 0.95;
         output;
      run;
      %if %lowcase(&scale_transform.) eq id %then %do;
         %let param_1 = &MIN_SCALE.;
      %end;
      %else %do;
         %let param_1 = %sysfunc(log(&min_scale.));
      %end;
      %do i=2 %to &SPARAMCOUNT.;
         %let PARAM_&i. = &&Sparam_&i.;
      %end;
      %QUANT&SDIST();
      data _NULL_;
         set _temp_;
         call symputx("min_tail_boundary",put(quant,BEST20.));
      run;
   %end;

   %if &min_tail_boundary. le 0 %then %do;
      %let ORV_ABORT = &ORV_TRUE;
      %let found_error = &ORV_TRUE;
      %let ORV_REASON =  vargrad_body_tail_boundary_not_computed;
      %goto DONE;
   %end;

   %if %sysevalf(10*&h > &Min_Tail_Boundary,BOOLEAN) %then %do;
      /* There are too few discretization points in the body of the distribution. See how many discretization 
         points will be needed to get at least 10 in the body.  If too many, use another method.  The
         '+1' in this calculation is because we want half the points to lie above the discretization interval
         to provide 'padding' for the FFT.  This reduces the aliasing error. */
      %let pwr = %sysfunc(ceil(%sysevalf(%sysfunc(log(10*&upper_bnd/&Min_Tail_Boundary))/%sysfunc(log(2)) + 1)));

      %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
         %Put INCREASED PWR TO: &pwr.;
      %end;
   %end;

   %if %sysevalf(&pwr gt 13, BOOLEAN) %then %do;
   /* In this case, can not get enough points in the body without using a LOT of points which
      is too computationally expensive (and sometimes impossible), so use the normal approximation method */
      %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
         %put RESORTING TO NORMAL APPROXIMATION OF AGGREGATE FOR PARAMETER VARIANCE CI CALCULATION;
      %end;
      %_orv_normal_param_grad( 
                              varpct    = _vg_var_, 
                              outgrad   = _vg_pgrad 
                              );

   %end;
   %else %do;
      /* Compute the number of discretization points and the interval width */
      %let npoints = %sysevalf(2**&pwr);
      %let h = %sysevalf((&upper_bnd/(&npoints/2)));
      /* DEBUG INFO */
      %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
         %put ---------------- CALLING FFT_PARAM_GRAD -----------------;
         %put DIST IS: &SDIST;
         %put POWER IS: &PWR (npoints is &npoints.);
         %put MIN TAIL BOUNDARY IS: &Min_Tail_Boundary;
         %put INTERVAL WIDTH IS: &h;
         %let rat = %sysevalf(&Min_Tail_Boundary/&h);
         %put RATIO is: &rat.;
      %end;
      %_orv_fft_param_grad( varpct         =_vg_var_, 
                            n_disc_points  = &npoints,
                            disc_int_width = &h,
                            outgrad        =_vg_pgrad );
   %end;

   /* Check for errors */
   %if &ORV_ABORT %then %do;
      %goto DONE; 
   %end;

   /* Add in the gradient effects of the KRIs */
   %if &SEV_KRI_COUNT >0 %then %do;
      %_orv_sev_kri_gradient( firm_kri_data  = _sim_sev_firm_kris, 
                              N_F_KRIs       = &SEV_N_F_KRIs,
                              firm_kri_list  = &SEV_FIRM_KRI_LIST,
                              strat_kri_data = _sim_sev_strat_kris, 
                              N_STRAT_KRIs   = &SEV_N_STRAT_KRIs,
                              STRAT_KRI_LIST = &SEV_STRAT_KRI_LIST,
                              prefix         = DS_, 
                              scales         = &sev_scales, 
                              ingrad         = _vg_pgrad, 
                              outgrad        = _vg_pgrad );
   %end;
   %if &FREQ_KRI_COUNT. > 0 or &freq_dummy_count > 0 %then %do;
      %_orv_freq_kri_gradient( N_F_KRIs          = &FREQ_N_F_KRIs.,
                               firm_kri_list     = &FREQ_FIRM_KRI_LIST.,
                               N_STRAT_KRIs      = &FREQ_N_STRAT_KRIs.,
                               STRAT_KRI_LIST    = &FREQ_STRAT_KRI_LIST.,
                               N_DUMMIES         = &FREQ_DUMMY_COUNT.,
                               Dummy_List        = &Freq_dummy_list.,
                               combined_kri_data = _sim_freq_kris_combined, 
                               ingrad            = _VG_pgrad, 
                               outgrad           = _vg_pgrad );
   %end;
   data &sev_gradient;
      set _vg_pgrad;
      drop 
      %if ( (&SEV_N_STRAT_KRIs eq 0) and (&sev_KRI_count gt 0) ) %then %do; 
        DSparam_1_0 
      %end; 
      df: ;
   run;
   data &freq_gradient;
      set _vg_pgrad;
      drop ds:;
   run;

   /* Store gradient data along with model data for use later (KRI Slider Bar calcualtion */
   /* First the severity info */
   /* Collect the KRIs names and their values, along with distribution info*/
   data _vg_sev_added_info;
      set &Sparm_dataset_name;
      where CCUID = &CCUID;
   run;

   data _vg_sev_added_info;
      length scale_transform $ 12;
      set _vg_sev_added_info;
      scale_transform = "&SCALE_TRANSFORM";
      keep
      %do i=1 %to &sparamcount;
         param_&i
      %end;
      %if &sev_KRI_count gt 0 %then %do;
         param_1_:
      %end;
      paramcount distribution ccuid scale_transform; 
   run;

   data &sev_gradient;
      set
      %do i=1 %to 3;
         _vg_sev_added_info
      %end; ;
      set &sev_gradient;
   run;

   /* Now store in VARPRJ */
   %let sev_grad_ds = %str();
   %orvu_find_dataset( namein=sev_gradient, libs = VARPRJ VARPRJ_M, nameout=sev_grad_ds);
   %if &sev_grad_ds NE %str() %then %do;
      data varprj.sev_gradient;
         set &sev_grad_ds;
      run;
      /* TODO: eventually, java will delete sev_gradient and freq_gradient prior to simulating, so this will
               be unnecessary */
      proc sql noprint;
         delete from VARPRJ.sev_gradient where ccuid = .;
         delete from VARPRJ.sev_gradient where ccuid = &ccUid;
      quit;
      data varprj.sev_gradient;
         set VARPRJ.sev_gradient &sev_gradient;
      run;
   %end;
   %else %do;
      data varprj.sev_gradient;
         set &sev_gradient;
      run;
   %end;
   proc sort data=VARPRJ.sev_gradient; by ccuid quantile; run;

   /* Now Frequency info */
   /* Collect the KRIs names and their values, along with distribution info*/
   data _NULL_;
      set &Fparm_dataset_name;
      where ( CCUID = &CCUID. );
      if &freq_KRI_count gt 0 then do;
         call symput("SelectedKRIs",SelectedKRIs);
      end;
   run;

/* First rename KRI coefficients to distinguish them from KRI values (insert Param_1_
   in the beginning) Frequency only- severity KRIs already labeled this way */
   data _vg_freq_added_info;
      set &Fparm_dataset_name;
      where ( CCUID = &CCUID. ); 
      %if &freq_KRI_count gt 0 %then %do;
         length new_kri_list $2000;
         array kri[*] &SelectedKRIs;
         do i=1 to dim(kri);
            new_kri_list = trim(new_kri_list) || " " || "Param_1_" || trim(vname(kri[i]));
         end;
         call symput("new_list",new_kri_list);
      %end;
   run;

   data _vg_freq_added_info;
      set _vg_freq_added_info;
      %if &freq_KRI_count gt 0 %then %do;
         array old_kri[*] &SelectedKRIs;
         array new_kri[*] &new_list.;
         do i=1 to dim(old_kri);
            new_kri[i] = old_kri[i];
         end;
      %end;
      keep  Fparam_:   
            paramcount 
            distribution 
            ccuid 
            %if &freq_KRI_count gt 0 %then %do;
            Param_1_:
            %end;
            %if &freq_dummy_count NE 0 %then %do;
               dummy: 
            %end;
            selectedKRIs 
            ccuid 
            seviter;
   run;

   data &freq_gradient;
      set
      %do i=1 %to 3;
         _vg_freq_added_info
      %end; ;
      set &freq_gradient;
   run;

   /* Now store in VARPRJ */
   %let freq_grad_ds = %str();
   %orvu_find_dataset( namein=freq_gradient, libs = VARPRJ VARPRJ_M, nameout=freq_grad_ds);
   %if &freq_grad_ds NE %str() %then %do;
      data VARPRJ.freq_gradient;
         set &freq_grad_ds;
      run;
      /* TODO: eventually, java will delete sev_gradient and freq_gradient prior to simulating, so this will
               be unnecessary */
      proc sql noprint;
         delete from VARPRJ.freq_gradient where ccuid = .;
         delete from VARPRJ.freq_gradient where ccuid = &ccUid;
      quit;
      data varprj.freq_gradient;
         set VARPRJ.freq_gradient &freq_gradient;
      run;
   %end;
   %else %do;
      data varprj.freq_gradient;
         set &freq_gradient;
      run;
   %end;
   proc sort data=VARPRJ.freq_gradient; by ccuid quantile; run;

%DONE:
   /* ALL OF THE ERROR HANDLING IS DONE BY ORVA_CONF_INTERVAL.  MESSAGES ARE DISPLAYED THERE */

   %orvu_delete_data(data= _vg_: );

%mend orva_var_gradient;


%macro _orv_freq_kri_gradient( 
                               N_F_KRIs          =, /* Number of firm level KRIs */
                               firm_kri_list     =, /* space delimited string of firm level KRIs (names) */
                               N_STRAT_KRIs      =, /* Number of risk category level KRIs */
                               STRAT_KRI_LIST    =, /* space delimited string of stratum level KRIs (names) */
                               N_DUMMIES         =, /* Number of Dummies */
                               Dummy_List        =, /* Space delimeted list of dummies */
                               combined_kri_data =, /* data set with firm and stratum level KRIs */
                               ingrad            =, 
                               outgrad           =
                              );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_freq_kri_gradient(; 
      %put .                        N_F_KRIs          = &N_F_KRIs,;
      %put .                        firm_kri_list     = &firm_kri_list,;
      %put .                        N_STRAT_KRIs      = &N_STRAT_KRIs,;
      %put .                        STRAT_KRI_LIST    = &STRAT_KRI_LIST,;
      %put .                        N_DUMMIES         = &N_DUMMIES,;
      %put .                        Dummy_List        = &Dummy_List,;
      %put .                        combined_kri_data = &combined_kri_data,;
      %put .                        ingrad            = &ingrad,; 
      %put .                        outgrad           = &outgrad;
      %put .                        );
   %end;

   %local name;
   %local i;

   /* All the derivatives have a common factor of DVaR/DParam_1.  
      This is added in when everything is combined with the &ingrad. 
      dataset */

   /* The Fparam_1 macro variable has the 'effective' FParam_1, the one
      implied by the full KRI model.  The dataset Fparam_1 is the estimated
      FParam_1 (exp(intercept) for KRI models) */
   
   /* Derivative wrt firm KRI parameters */
   %if &N_F_KRIs gt 0 %then %do;
      data _vg_firm_der;
         set &combined_kri_data;
         if _N_ = 1;

         array kri[*] &FIRM_KRI_LIST;
         array dkri[*] %do i=1 %to &N_F_KRIs; 
                          %let name = %scan(&FIRM_KRI_LIST,&i);
                          DF_&name.
                       %end; ;

            do i=1 to &N_F_KRIs;
            dkri[i] = &FParam_1.*kri[i];
            end;
            output;

         keep %do i=1 %to &N_F_KRIs;
                 %let name = %scan(&FIRM_KRI_LIST,&i);
                 DF_&name. 
              %end; ;
      run;
   %end;

   /* Derivative wrt cell slice (stratum) level KRIs */
   %if &N_STRAT_KRIs gt 0 %then %do;
      data _vg_strat_der;
         set &combined_kri_data. end=last;
         array kri[*] &STRAT_KRI_LIST;
         array dkri[*] %do i=1 %to &N_STRAT_KRIs; 
                          %let name =%scan(&STRAT_KRI_LIST,&i);
                          DF_&name. 
                       %end; ;

         retain %do i=1 %to &N_STRAT_KRIs; 
                   %let name = %scan(&STRAT_KRI_LIST,&i);
                   DF_&name. 0
                %end; ;

         do i=1 to &N_STRAT_KRIs;
            dkri[i] = dkri[i] + FParam_1*kri[i];
         end;
         if last then do;
            output;
         end;
         keep %do i=1 %to &N_STRAT_KRIs;
                 %let name = %scan(&STRAT_KRI_LIST,&i);
                 DF_&name.
              %end; ;
      run;
   %end;

   /* Derivative wrt indicator variables (dummies) */
   %if &N_DUMMIES gt 0 %then %do;
      data _vg_dum_der;
         set &combined_kri_data. end=last;
         /* TODO: IBL specific code here */
         where &Freq_Model_Basis_Props. in (&Freq_Dummy_Strats.);
         array ddum[*] %do i=1 %to &N_DUMMIES; 
                          %let name = %scan(&Freq_Dummy_Strats.,&i);
                          DF_DUMMY_&name. 
                       %end; ;
         retain %do i=1 %to &N_DUMMIES; 
                   %let name = %scan(&Freq_Dummy_Strats.,&i);
                   DF_Dummy_&name. 0
                %end; ;
         ddum[_N_] = FParam_1;
         if last then do;
            output;
         end;
         keep DF_DUMMY_:;
      run;
   %end;

   data &outgrad; 
      set &ingrad; 
      %if &N_F_KRIs gt 0 %then %do;
         set _vg_firm_der _vg_firm_der _vg_firm_der;
      %end;
      %if &N_STRAT_KRIs gt 0 %then %do;
         set _vg_strat_der _vg_strat_der _vg_strat_der;
      %end;
      %if &N_DUMMIES gt 0 %then %do;
         set _vg_dum_der _vg_dum_der _vg_dum_der;
      %end;

      array derivatives DF_:;
      do i=1 to dim(derivatives);
         derivatives[i] = derivatives[i]*DFParam_1;
      end;
   run;

%mend _orv_freq_kri_gradient;

/* for variable descriptions, see freq_kri_gradient macro in this file */
%macro _orv_sev_kri_gradient( 
                              firm_kri_data  = , 
                              N_F_KRIs       = ,
                              firm_kri_list  = ,
                              strat_kri_data = , 
                              N_STRAT_KRIs   = ,
                              STRAT_KRI_LIST = ,
                              prefix         = , 
                              scales         = , 
                              ingrad         = , 
                              outgrad        =
                              );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_sev_kri_gradient(; 
      %put .                       firm_kri_data  = &firm_kri_data.,; 
      %put .                       N_F_KRIs       = &N_F_KRIs.,;
      %put .                       firm_kri_list  = &firm_kri_list.,;
      %put .                       strat_kri_data = &strat_kri_data.,;
      %put .                       N_STRAT_KRIs   = &N_STRAT_KRIs.,;
      %put .                       STRAT_KRI_LIST = &STRAT_KRI_LIST.,;
      %put .                       prefix         = &prefix.,;
      %put .                       scales         = &scales.,;
      %put .                       ingrad         = &ingrad.,;
      %put .                       outgrad        = &outgrad.;
      %put .                       );
   %end;

   %if &N_F_KRIs gt 0 %then %do;
      /* Need the sum of DVaRDScale*Scale for each var value (for firm level KRIs derivs) */
         data _vg_temp;
            set &ingrad;
            array dscales[*] DSparam_1_:;
            sum=0;
            %do i=1 %to &nsev_mix;
               %let scl = %scan(&SCALES,&i,' ');
               sum = sum + dscales[&i]*&SCL;
            %end;
         run;
         proc sql noprint; 
            select sum
            into : STRAT_FACTOR separated by " "
            from _vg_temp
         run;
         quit;

      /* Get the derivatives for the firm level kris */
      proc fcmp data=&firm_kri_data out = _vg_firm_der;
         array kri[*] &FIRM_KRI_LIST;
         array strat_factors[*] &STRAT_FACTOR;
         array dkri[*] %do i=1 %to &N_F_KRIs; 
                          %let n=%scan(&FIRM_KRI_LIST,&i);
                          &Prefix&n 
                       %end; ;
         do var_points = 1 to 3;
            do i=1 to &N_F_KRIs;
               dkri[i] = strat_factors[var_points]*kri[i] ;
            end;
            output;
         end;

         keep %do i=1 %to &N_F_KRIs;
                 %let n=%scan(&FIRM_KRI_LIST,&i);
                 &Prefix&n 
              %end; ;
      run;
   %end;

   %if &N_STRAT_KRIs gt 0 %then %do;
      /* add the scales to the loss group kri dataset */
      data _vg_temp;
         %do i=1 %to &nsev_mix;
            %let stratid = %scan(&SEV_STRAT_IDS,&i,' ');
            %let scale = %scan(&scales,&i,' ');
            BLUID_KRI = &stratid;
            scale = &scale;
            output;
         %end;
      run;

      proc sort data = _vg_temp;
         by bluid_kri;
      run;

      proc sort data = &strat_kri_data;
         by bluid_kri;
      run;

      /* Get the derivatives for the loss group level kris */
      data _vg_temp;
        MERGE _vg_temp &strat_kri_data; 
        BY BLUID_KRI; 
      RUN: 

      data _vg_dscale;
         set &ingrad;
         keep DSParam_1_:;
      run;
      proc transpose data=_vg_dscale out=_vg_dscale; run;

      data _vg_temp;
         set _vg_temp;
         set _vg_dscale end=last;
         retain 
         %do i=1 %to &N_STRAT_KRIs.;
            %let suffix = %scan(&STRAT_KRI_LIST.,&i,' ');
            &prefix&suffix._1 0
            &prefix&suffix._2 0
            &prefix&suffix._3 0
         %end; ;
         %do i=1 %to &N_STRAT_KRIs.;
            %let suffix = %scan(&STRAT_KRI_LIST.,&i,' ');
            &prefix&suffix._1 = &prefix&suffix._1 + COL1*scale*&suffix.;
            &prefix&suffix._2 = &prefix&suffix._2 + COL2*scale*&suffix.;
            &prefix&suffix._3 = &prefix&suffix._3 + COL3*scale*&suffix.;
         %end; ;
         if last then do;
            output;
         end;
         keep &prefix.:;
      run;

      data _vg_strat_der; 
         set _vg_temp;

         %do i=1 %to 3;
            %do j=1 %to &N_STRAT_KRIs.;
               %let suffix = %scan(&STRAT_KRI_LIST.,&j,' ');
               &prefix.&suffix. = &prefix.&suffix._&i.;
            %end;
            output;
         %end;
         keep %do j=1 %to &N_STRAT_KRIs.;
                 %let suffix = %scan(&STRAT_KRI_LIST.,&j,' ');
                 &prefix.&suffix.
              %end; ;
      run;

   %end;
   data &outgrad; 
      set &ingrad; 
      %if &N_F_KRIs gt 0 %then %do;
         set _vg_firm_der;
      %end;
      %if &N_STRAT_KRIs gt 0 %then %do;
         set _vg_strat_der;
      %end;
   run;
%mend _orv_sev_kri_gradient;


/* Use FFT to get derivative of VaR wrt parameters */
%macro _orv_fft_param_grad( 
                            varpct         =, /* Dataset with VaR values */
                            n_disc_points  =, /* Number of discretization points */
                            disc_int_width =,
                            outgrad        =  /* output dataset */
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_fft_param_grad(; 
      %put .                     varpct         = &varpct.,;
      %put .                     n_disc_points  = &n_disc_points.,;
      %put .                     disc_int_width = &disc_int_width.,;
      %put .                     outgrad        = &outgrad.;
      %put .                     );
   %end;

   %let npoints = &n_disc_points;
   %let h = &disc_int_width;

   /* Get the index of the discretized severity corresponding to maxloss */
   /* Note: if max_loss_ind is larger than npoints, then it will be ignored 
            See code dealing with max_loss_ind below */
   %let max_loss_ind = %sysfunc(max(%sysfunc(round(&maxloss./&h.)),1));

   /* Generate points on which to compute the severity derivatives */
   data _vg_derivs_indata;
      do i = 1 to &npoints.;
         quant = (2*i - 1) * &h. / 2;
         output;
      end;
   run;

   /* Calculate the derivatives of severity cdf wrt severity parameters */
   /* If there are severity KRIs, may need to do this for each risk category */
   /* The severity macros look for Param_1, not SParam_1*/
   %do i = 2 %to &SPARAMCOUNT;
      %let PARAM_&i. = &&Sparam_&i.;
   %end;
   %if %sysevalf(&SPARAMCOUNT > 1, BOOLEAN) %then %do;
      %do i=1 %to &nsev_mix;
         %if %lowcase(&scale_transform.) eq id %then %do;
            %let Param_1 = %scan(&SEV_SCALES, &i, ' ');
         %end;
         %else %do;
            %let Param_1 = %sysfunc(log(%scan(&SEV_SCALES, &i, ' ')));
         %end;
         %CdfDer&SDIST.(
                        scalefactor = 1, 
                        indata      = _vg_derivs_indata,
                        outdata     = _vg_grad_dist&i
                        );
         /* Check that proc model returned something. */
         %orvl_ds_exists(_vg_grad_dist&i, "orva_var_gradient", false);
         %if %sysevalf(&ds_nobs eq 0 or &SYSINFO ne 0) %then %do;
            %let ORV_ABORT = &ORV_TRUE;
            %let &ORV_REASON =  vargrad_no_model_output_error;
            %goto DONE_FFT;
         %end;
      %end;
   %end;

   proc fcmp library=&ORVAR_CMPLIB; 

      /***** DECLARE ARRAYS *****/

      /* var quantiles, var values, output dataset */
      array qv[3] /nosymbols;
      array var_vals[3] /nosymbols;
      array results[3,&tot_cols.]/nosymbols;
      /* Array containing the weights used in the severity mixture */
      array weights[&NSEV_MIX] &SPARAM1_WEIGHTS;

      /* discretized severity- real and imaginary.  Second copy is for freq. derivs */
      array f[1] /NOSYMBOLS;
      array f2[1] /NOSYMBOLS;
      array im_f[1] /NOSYMBOLS;
      array im_f2[1] /NOSYMBOLS;

      /* derivative of aggregate distribution wrt freq. params- real and imaginary */
      %do y = 1 %to &FPARAMCOUNT;
         array dfdf&y[1] /NOSYMBOLS;
         array im_dfdf&y[1] /NOSYMBOLS;
      %end;

      /* discretized aggregate distribution- real and imaginary */
      array g[1] /NOSYMBOLS;
      array im_g[1] /NOSYMBOLS;

      /* derivative of aggregate cdf  at VaR wrt freq. parameters */
      %do y = 1 %to &FPARAMCOUNT;
         array dQdf&y[1] /NOSYMBOLS;
      %end;

      /* derivative of aggreagate cdf at VaR */
      array dQdV[3] /NOSYMBOLS;

      /* indices corresponding to the var values and the quantiles */
      array var_indices[3] /NOSYMBOLS;
      array quant_indices[3] /NOSYMBOLS;

      /* derivative of severity distribution wrt parameters- real and imaginary */
      /* Need 1 set for each scale parameter (loss grouping) */
      %DO i=1 %TO &NSEV_MIX;   
         %DO z=2 %TO &SPARAMCOUNT.;      
            array dfdS&z._&i.[1] /NOSYMBOLS;   
         %END;
      %END;
      /* These will hold the combined (mixed) values */
      %DO z=2 %TO &SPARAMCOUNT;
         array dfdS&z.[1] /NOSYMBOLS;
         array im_dfdS&z.[1] /NOSYMBOLS;

         /* derivative of aggregate severity distribution wrt parameters- real and imaginary */
         array dgdS&z.[1] /NOSYMBOLS;
         array im_dgdS&z.[1] /NOSYMBOLS;

         /* derivative of aggregate cdf wrt severity parameters */
         array dQdS&z.[1] /NOSYMBOLS;
      %end;

      /* dynamically allocate arrays to avoid stack overflow */
      call dynamic_array( f, &Npoints);
      call dynamic_array( f2, &Npoints);
      call dynamic_array( im_f, &Npoints);
      call dynamic_array( im_f2, &Npoints);

      %do y = 1 %to &FPARAMCOUNT;
         call dynamic_array( dfdf&y, &Npoints);
         call dynamic_array( im_dfdf&y, &Npoints);
         call dynamic_array( dQdf&y, &Npoints);
      %end;

      call dynamic_array( g, &npoints);
      call dynamic_array( im_g, &npoints);

      %DO i=1 %TO &NSEV_MIX;   
         %DO z=2 %TO &SPARAMCOUNT.;      
            call dynamic_array(dfdS&z._&i., &Npoints);   
         %END;
      %END;   
      %DO z=2 %TO &SPARAMCOUNT;
         call dynamic_array( dfds&z., &Npoints);
         call dynamic_array( im_dfds&z., &Npoints);
         call dynamic_array( dgds&z., &Npoints);
         call dynamic_array( im_dgds&z., &Npoints);
         call dynamic_array( dQdS&z., &Npoints);
      %end;

      /***** GET QUANTILES *****/
      /* Load the quantiles */
      rc = read_array("&varpct", qv, "quantile");

      var_vals[1] = &P_1;
      var_vals[2] = &P_2;
      var_vals[3] = &P_3;

      /* Figure out the order of the quantiles (they ar now provided in order) */
      quant_indices[1] = 3;
      quant_indices[2] = 2;
      quant_indices[3] = 1;

      /***** GET indices *****/
      /* Find the indices of g corresponding to the VaR values */
      do i=1 to 3;
         var_indices[i] = max(floor(var_vals[quant_indices[i]]/&h+1),1);
      end;

      /***** LOAD DERIVATIVES *****/
      /*Load in the derivatives of the f's wrt sev. params */
      %do i=1 %to &NSEV_MIX;
         %do z=2 %to &SPARAMCOUNT.;
            rc = read_array("_vg_grad_dist&i", dfdS&z._&i., "DParam_&z.");
         %end;
      %end;

      /* We actually loaded derivative of Fx wrt parameters.
         f[i] = Fx[i+1] - Fx[i], where Fx is cdf of sev. dist. */
      do i=0 to &Npoints.-2;
         %do i=1 %to &NSEV_MIX;
            %do z=2 %to &SPARAMCOUNT.;
               dfdS&z._&i.[&Npoints.-i] = dfdS&z._&i.[&Npoints.-i] - dfdS&z._&i.[&Npoints.-i-1];
            %end;
         %end;
      end;

      /* Now get the values for the full mixture */
      do i=1 to &Npoints.;
         %do z = 2 %to &SPARAMCOUNT;
            dfdS&z[i] = 0;
            %do i=1 %to &NSEV_MIX;
               dfdS&z[i] = dfdS&z[i] + weights[&i]*dfdS&z._&i.[i];
            %end;
         %end;
      end;

      /***** DISCRETIZE SEVERITY *****/
      %if %trim(%upcase(&SDIST)) eq MIXDIST or %trim(%upcase(&SDIST)) eq LOGNGPD or %trim(%upcase(&SDIST)) eq LOGNGPD1 %then %do;
          param_xlimit = &param_xlimit;
          param_pnlimit = &param_pnlimit;
      %end;
      DO i=1 to &Npoints.;
         f[i] = 0; 
      end;
      %DO i=2 %TO &SPARAMCOUNT;
         Param_&i = &&SPARAM_&i.;
      %END;

      %DO w=1 %TO &NSEV_MIX;
         %if %lowcase(&scale_transform.) eq id %then %do;
            Param_1 = %scan(&SEV_SCALES, &w, ' ');
         %end;
         %else %do;
            Param_1 = %sysfunc(log(%scan(&sev_scales., &w., ' ')));
         %end;
         %VCdf&SDIST.(&h/2, prior);
         f[1]=f[1]+weights[&w]*prior;
         total_mass = prior;
         DO i=2 to &Npoints.;
            %VCdf&SDIST.(&h*(2*i-1)/2, next);
            new_mass = weights[&w]*(next - prior);
            f[i] = f[i] + new_mass;
            prior = next;
            total_mass = total_mass + new_mass;
         end;
      %END;

      /* Especially if maxloss truncates the severity, we may have some
         unaccounted for mass.  Place it in the last bin */
      remaining_mass = max(0, 1 - total_mass);
      f[&Npoints.] = f[&Npoints.] + remaining_mass;

      /* Initialize imaginary component */
      do i=1 to &Npoints.;
         im_f[i] = 0;
      end;

      /* Correct for truncation due to maxloss.  All mass to the right of maxloss gets placed at maxloss.  
         Maxloss is treated as fixed and not a function of the parameters for convenience  */
      do i=(&max_loss_ind.+1) to &Npoints.;
         f[&max_loss_ind.] = f[&max_loss_ind.] + f[i];
         f[i] = 0;
         %do z=2 %to &SPARAMCOUNT;
            dfdS&z[&max_loss_ind.] = dfdS&z[&max_loss_ind.] + dfdS&z[i];
            dfdS&z[i] = 0;
         %end;
      end;

      /***** AGGREGATE DISTRIBUTION *****/

      /* Transform */
      rc = zfft(f,im_f,&Npoints.,&Npoints.,&Npoints.,-1);
      do i=1 to &Npoints.;
         g[i] = f[i];   im_g[i] = im_f[i];
         f2[i] = f[i];  im_f2[i] = im_f[i];
      end;

      /* Apply pgf */
      call pgf&FDIST(%do y=1 %to &FPARAMCOUNT; &&FPARAM_&y, %end; g,im_g,&Npoints.);
      
      /* Inverse Transform */
      rc = zfft(g,im_g,&Npoints.,&Npoints.,&Npoints.,1);

      /* Rescale */
      do i=1 to &Npoints.;
         g[i] = g[i]/&Npoints.;
         g[i] = max(g[i],0);
      end;

      /***** DQDVaR *****/
      do i=1 to 3;
         if g[var_indices[i]] eq 0 then do;
            var_indices[i] = 1;
         end;
         if var_indices[i] ne 1 then do;
            dQdV[quant_indices[i]] = g[var_indices[i]]/&h;
         end;
      end;

      /***** GET DER. OF QUANTILE WRT TO PARAMETERS *****/

      /***** FREQ DERIVS *****/
      /* Frequency parameter- we already computed transform of f */
      /* Apply der. of pgf wrt frequency params*/
      %do y = 1 %to &FPARAMCOUNT;
         call pgfder&FDIST(%do y=1 %to &FPARAMCOUNT; dfdf&y,im_dfdf&y,&&FPARAM_&y, %end; f,im_f,&Npoints.);
      %end;
      /* Inverse Transform (dfdl becomes dgdl, f is severity pmf and g is aggregate pmf) */
      %do y=1 %to &FPARAMCOUNT;
         rc = zfft(dfdf&y,im_dfdf&y,&Npoints.,&Npoints.,&Npoints.,1);
      %end;
      /* Rescale */
      do i=1 to &Npoints.;
         %do y=1 %to &FPARAMCOUNT;
            dfdf&y[i] = dfdf&y[i]/&Npoints.;
         %end;
      end;      

      /***** SEV DERIVS *****/
      /* Severity parameter- start with the derivative of f wrt parameters.  Do not 
         transform, as there is no good way to predict the tail behavior of the 
         derivatives (see Embrechts and Frei 2009 for description of how one can
         use a transformation to stabilize FFT) */
      do i=1 to &Npoints.;
         %do z=2 %to &SPARAMCOUNT;
            im_dfds&z.[i] = 0;   
         %end;
      end;

      %do z=2 %to &SPARAMCOUNT;
         rc = zfft(dfdS&z.,im_dfdS&z.,&Npoints.,&Npoints.,&Npoints.,-1);
      %end;

      call pgfCompDer&FDIST(%do y=1 %to &FPARAMCOUNT; &&FPARAM_&y, %end;f2,im_f2,&Npoints.);

      /* compute complex product */
      %do z=2 %to &SPARAMCOUNT;
         do i=1 to &Npoints.;
            temp1 = f2[i]*dfds&z.[i] - im_f2[i]*im_dfds&z.[i];
            temp2 = f2[i]*im_dfds&z.[i] + dfds&z.[i]*im_f2[i];
            dfds&z.[i] = temp1;
            im_dfds&z.[i] = temp2;
         end;
      %end;

      /* Inverse Transform */
      %do z=2 %to &SPARAMCOUNT;
         rc = zfft(dfdS&z.,im_dfdS&z.,&Npoints.,&Npoints.,&Npoints.,1);
      %end;

      /* Rescale */
      %do z=2 %to &SPARAMCOUNT;
         do i=1 to &Npoints.;
            dfdS&z.[i] = dfdS&z.[i]/&Npoints.;
         end;
      %end;

      /***** QUANTILE DERIVS *****/
      /* Now get partials of the quantile functions */
      %DO y=1 %TO &FPARAMCOUNT;
         dQdf&y[1] = dfdf&y[1];
      %end;

      %DO z=2 %TO &SPARAMCOUNT;
         dQdS&z.[1] = dfdS&z.[1];
      %end;
      do i=2 to &npoints.;
         %DO y=1 %TO &FPARAMCOUNT;
            dQdf&y[i] = dQdf&y[i-1] + dfdf&y[i];
         %end;
         %DO z=2 %TO &SPARAMCOUNT;
            dQdS&z.[i] = dQdS&z.[i-1] + dfdS&z.[i];
         %end;
      end;

      /***** RESULTS *****/
      do i=1 to 3;
         if var_indices[i] ne 1 then do;
            %do m = 1 %to &NSEV_MIX;
               %let weight = %scan(&SPARAM1_WEIGHTS,&m,' ');
               /* DVaR /DScales */
               results[quant_indices[i],&m.] = &weight.* var_vals[quant_indices[i]]/ %scan(&SEV_SCALES, &m, ' ');
            %end;
            %DO z=2 %to &SPARAMCOUNT;
               %let ind = %eval(&NSEV_MIX + &z - 1);
               results[quant_indices[i],&ind] = -1*dQdS&z[var_indices[i]]/dQdV[quant_indices[i]];
            %end;   
            %do y = 1 %to &FPARAMCOUNT;
               %let ind = %eval(&SPARM_TOT + &y);
               results[quant_indices[i],&ind] = -1*dQdf&y[var_indices[i]]/dQdV[quant_indices[i]];
            %end;
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT.+1] = qv[quant_indices[i]];
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT+2] = var_vals[quant_indices[i]];
         end;
         else do;
            %DO z=1 %to %eval(&SPARAMCOUNT + &NSEV_MIX -1);
               results[quant_indices[i],&z.] =0;
            %end;   
            %do y = 1 %to &FPARAMCOUNT;
               results[quant_indices[i],&SPARM_TOT+&y] = 0;
            %end;
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT.+1] = qv[quant_indices[i]];
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT.+2] = var_vals[quant_indices[i]];
         end;
      end;

      /* How the results are labeled depends on whether or not there are severity risk 
         category level KRIs */
      rc = write_array("&outgrad.",results, 
         %if %sysevalf(&SEV_KRI_COUNT > 0, boolean) %then %do;
            %do z=1 %to &NSEV_MIX;
               /* Get the name of the loss grouping to append to the appropriate scale */
               %let suffix = %scan(&SEV_STRAT_IDS,&z,' ');
               "DSParam_1_&suffix.",
            %end;
            %do z=2 %to &SPARAMCOUNT;
               /*Now do the rest of the parameters */
               "DSparam_&z.",
            %end;
         %end;
         %else %do;
            "DSParam_1",
            %do z=2 %to &SPARAMCOUNT; 
               "DSParam_&z.", 
            %end;
         %end;
         %do y=1 %to &FPARAMCOUNT; 
            "DFParam_&y.", 
         %end;
         "Quantile", "VaR");
   run;

   /* Add in the derivative with respect to severity param_1 */
   %if %sysevalf(&SEV_KRI_COUNT > 0, boolean) %then %do;
      data &outgrad.;
         set &outgrad;
         DSparam_1 = 0;
         %do z=1 %to &nsev_mix;
            %let suffix = %scan(&SEV_STRAT_IDS,&z,' ');
            %let scale = %scan(&SEV_SCALES,&z,' '); 
            %if &SCALE_TRANSFORM = ID %then %do;
               DSparam_1 = DSparam_1 + DSparam_1_&suffix.*&SCALE./&SParam_1;
            %end; 
            %else %do;
               DSparam_1 = DSparam_1 +  DSparam_1_&suffix.*&SCALE.;
            %end;
         %end;
      run;
   %end;
   %else %do;
      data &outgrad.;
         set &outgrad;
         %if &SCALE_TRANSFORM = ID %then %do;
/*            DSparam_1 = DSparam_1;*/
         %end;
         %else %do;
            DSparam_1 = DSparam_1 * %scan(&SEV_SCALES,1,' ');
         %end;
      run;
   %end;
  
   %DONE_FFT: 
   
%mend _orv_fft_param_grad;

%macro _orv_normal_param_grad( 
                               varpct  = , 
                               outgrad = 
                              );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orv_normal_param_grad(; 
      %put .                        varpct  = &varpct.,;
      %put .                        outgrad = &outgrad.;
      %put .                        );
   %end;

   /* This dataset will be fed into proc model to get derivatives with respect parameters */
   data _vg_norm;
      set &varpct;
      %do i=1 %to &FPARAMCOUNT;
         FPARAM_&i = &&FPARAM_&i.;
      %end;
      %if %trim(%upcase(&SDIST)) eq MIXDIST or %trim(%upcase(&SDIST)) eq LOGNGPD or %trim(%upcase(&SDIST)) eq LOGNGPD1 %then %do;
         Param_xLimit  = &Param_xLimit;
         Param_pnLimit = &Param_pnLimit;
      %end;
      %do i=2 %to &SPARAMCOUNT;
        PARAM_&i. = &&Sparam_&i.;
      %end;
      smean = 0;
      svar = 0;
      %do k=1 %to &nsev_mix;
         %if %lowcase(&scale_transform.) eq id %then %do;
            Param_1_&k. = %scan(&SEV_SCALES, &k, ' ');
         %end;
         %else %do;
            Param_1_&k. = %sysfunc(log(%scan(&SEV_SCALES, &k, ' ')));
         %end;
         Param_1 = Param_1_&k;
         %moments&SDIST;
         smean      = smean +  %scan(&SPARAM1_WEIGHTS, &k, ' ')*mean;
         svar       = svar + %scan(&SPARAM1_WEIGHTS, &k, ' ')*std**2;
      %end;
      
      %moments&FDIST;
      fmean      = mean;
      fvar       = variance;
      agg_mean   = fmean*smean; /* mean of aggregate distribution */
      agg_var    = fmean*svar + fvar*smean**2; /* variance of aggregate distribution */
      var1       = max(probit(quantile/100)*sqrt(agg_var) + agg_mean,0);
      %do i=2 %to %eval(&NSEV_MIX + &SPARAMCOUNT + &FPARAMCOUNT - 1);
         var&i = var1;
      %end;
      index = _N_;
      output;
      keep fparam: param_: var: quantile smean svar fmean fvar index;
   run;

   %do j=1 %to 3;
      data _vg_norm&j;
         set _vg_norm;
         where index = &j;
         call symputx("current_var_guess",var1);
         call symputx("current_var_actual",var_value);
      run;
      /* Handle the VaR <= 0 case: */
      %if %sysevalf( (&current_var_guess. <= 0) or (&current_var_actual. <= 0),BOOLEAN) %then %do;
         %let ORV_ABORT = &ORV_TRUE;
         %let found_error = &ORV_FALSE;
         %let ORV_REASON =  vargrad_normal_approx_method_failed;
         %return;
      %end;
      %else %do;
         proc model data=_vg_norm&j noprint;
            dependent 
               %do i=1 %to %eval(&NSEV_MIX + &SPARAMCOUNT + &FPARAMCOUNT - 1); 
                  var&i 
               %end;
            ;
            retain
               %do i=1 %to &nsev_mix;
                  %if %lowcase(&scale_transform.) eq id %then %do;
                     Param_1_&i  %scan(&SEV_SCALES, &i, ' ')
                  %end;
                  %else %do;
                     Param_1_&i  %sysfunc(log(%scan(&SEV_SCALES, &i, ' ')))
                  %end;
               %end;
               %do i=2 %to &SPARAMCOUNT;
                  Param_&i &&SParam_&i
               %end;
               %do i=1 %to &FPARAMCOUNT;
                  FParam_&i &&FParam_&i
               %end;
            ;
            smean = 0;
            svar = 0;
            %do k=1 %to &nsev_mix;
               Param_1 = Param_1_&k;
               %moments&SDIST;
               smean      = smean + %scan(&SPARAM1_WEIGHTS, &k, ' ')*mean;
               svar       = svar + %scan(&SPARAM1_WEIGHTS, &k, ' ')*std**2;
            %end;
            %moments&FDIST;
            fmean      = mean;
            fvar       = variance;
            agg_mean   = fmean*smean;
            agg_var    = fmean*svar + fvar*smean**2;      
            var1 =  probit(quantile/100)*sqrt(agg_var) + agg_mean;
            %do i=2 %to %eval(&NSEV_MIX + &SPARAMCOUNT + &FPARAMCOUNT - 1);
               var&i = pred.var1;
            %end;
            %do i=1 %to &nsev_mix;
               DSParam_1_&i = getder(pred.var&i,Param_1_&i);
            %end;
            %do i=2 %to &SPARAMCOUNT;
               %let ind = %eval(&nsev_mix+&i-1);
               DSParam_&i = getder(pred.var&ind,Param_&i);
            %end;
            %do i=1 %to &FPARAMCOUNT;
               %let ind = %eval(&SPARAMCOUNT + &nsev_mix -1 + &i);
               DFParam_&i = getder(pred.var&ind,FParam_&i);
            %end;
            outvars 
               %do i=1 %to &nsev_mix;
                  DSParam_1_&i
               %end;
               %do i=2 %to &SPARAMCOUNT;
                  DSParam_&i
               %end;
               %do i=1 %to &FPARAMCOUNT;
                  DFParam_&i 
               %end;
               quantile
               var_value;
            solve 
               %do i=1 %to &nsev_mix;
                  Param_1_&i
               %end;
               %do i=2 %to &SPARAMCOUNT;
                  Param_&i
               %end;
               %do i=1 %to &FPARAMCOUNT;
                  FParam_&i
               %end;
               /out=_vg_norm_out&j(keep=
                                   quantile
                                   var_value 
                                   %do i=1 %to &nsev_mix;
                                      DSParam_1_&i
                                   %end;
                                   %do i=2 %to &SPARAMCOUNT;
                                      DSParam_&i
                                   %end;
                                   %do i=1 %to &FPARAMCOUNT;
                                      DFParam_&i 
                                   %end;
                                  );
         run;
         quit;
         /* Check that proc model returned something. */
         %orvl_ds_exists(_vg_norm_out&j, "orva_var_gradient", false);
         %if %sysevalf(&ds_nobs eq 0 or &SYSINFO ne 0) %then %do;
            %let ORV_ABORT = &ORV_TRUE;
            %let &ORV_REASON =  vargrad_no_model_output_error;
            %return;
         %end; 
      %end;
   %end;

   data &outgrad;
      set _vg_norm_out1 _vg_norm_out2 _vg_norm_out3;
      VaR = var_value;
      drop var_value;
   run;

   /* Adjust the severity scale parameter results */
   %if %sysevalf(&SEV_KRI_COUNT > 0, boolean) %then %do;
      data &outgrad;
         set &outgrad;
         DSParam_1 = 0;
         %do i=1 %to &NSEV_MIX;
            %let suffix = %scan(&SEV_STRAT_IDS,&i,' ');
            %let scale = %scan(&SEV_SCALES,1,' ');
            DSParam_1_&suffix.  = DSParam_1_&i;
            %if &SCALE_TRANSFORM = ID %then %do;
               DSparam_1 = DSparam_1 + DSParam_1_&suffix.*&SCALE./&SParam_1;
            %end; 
            %else %do;
               DSparam_1 = DSparam_1 + DSParam_1_&suffix.;
            %end;
         %end;
         drop 
            %do i=1 %to &NSEV_MIX;
               DSParam_1_&i
            %end;
            ;
      run;
   %end;
   %else %do;
      data &outgrad;
         set &outgrad;
         DSParam_1 = DSParam_1_1;
         drop DSParam_1_1;
      run;
   %end;
%mend _orv_normal_param_grad;
