/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_BURR.sas
 *   
 *   -Provides support for the BURR distribution-
 *
 * This file contains macro fragments to implement calculation specific to the 
 * BURR distribution
 *
 *  The macro %ORVD_BURR is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_BURR macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ********************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable into which result is stored
 *****************************************************************************************/
%macro CdfBURR(x, returnVal);
      &returnVal = 1-(1+(&x/(&Param_1))**&Param_3)**(-&Param_2);   
%mend CdfBURR;

%macro VCdfBURR(x, returnVal);
      &returnVal = 1-(1+(&x/Param_1)**Param_3)**(-Param_2);   
%mend VCdfBURR;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable into which result is stored
 *****************************************************************************************/
%macro PdfBURR(x, returnVal);
      &returnVal =&Param_2 * &Param_3 * ((&x / (&Param_1))**&Param_3)/(&x * (1+(&x / (&Param_1))**&Param_3)**(&Param_2 + 1));
%mend PdfBURR;

%macro VPdfBURR(x, returnVal );
      &returnVal =   Param_2  * Param_3 * ((&x / Param_1)**Param_3)/(   &x * (1 + (&x / Param_1 )**Param_3)**( Param_2 + 1));
%mend VPdfBURR;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantBURR(scaleFactor=1);
   data _temp_;
      set _temp_ nobs=nobs end=eof;
      retain truncdf 0;

      if _N_=1 then do;
         if THRESHOLD = 0 then do;
            truncdf = 0;
         end;
         else do;
            %CDFBURR(THRESHOLD,truncdf)     
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf   = ecdf *(1-truncdf) + truncdf;

      if  cdf/(&scaleFactor) LE 1.0 then do; 
         quant =(&Param_1) * ((1-cdf/(&scaleFactor))**(-1/&Param_2) - 1 )**(1/&Param_3);
      end;
      else do;
         quant = .;
      end;
   run;
%mend QuantBURR;
   
/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerBURR(scaleFactor=1, indata=, outdata=);
   data &outdata;
      set &indata;
      %cdfBURR(quant,cdf)
      cdf = cdf*&ScaleFactor.;
      DParam_1 = -1*&ScaleFactor.*&Param_2.*&Param_3.*(1+(quant/&Param_1)**(&Param_3))**(-1-&Param_2)*(quant/&Param_1)**(&Param_3)/(&Param_1);
      DParam_2 = cdf*log(cdf)/&Param_2;
      DParam_3 = &ScaleFactor.*&Param_2*(1+(quant/&Param_1)**(&Param_3))**(-1-&Param_2)*(quant/&Param_1)**(&Param_3)*log(quant/&Param_1);
   run;

%mend CdfDerBURR;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimBURR(loss);
    cdf=rand("UNIFORM");
    &loss = (Param_1) * ( cdf**(-1/Param_2) - 1 )**(1/Param_3);    
%mend SimBURR;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitBURR(summaryData = , KRI_FIT = NO);
   %let nparms=3;

   /*- initial guess based on Pareto moments -*/
   data _null_;
      set &summaryData;

      if sample_variance > sample_mean**2 then do;
         alpha =  2 * sample_variance / (sample_variance - sample_mean**2);
      end;
      else do;
         alpha = 2;
      end;

      theta = sample_mean * (alpha-1);

      call symputx("PARAM_2_INIT", put(alpha, BEST20.));
      call symputx("PARAM_1_INIT", put(theta, BEST20.));
      call symputx("PARAM_3_INIT", 2);
      call symputx("obsCount",obsCount);
   run;   
   
   %if &dist NE MIXDIST %then %do;
      %put **********************************;
      %put Distribution  =  BURR;
      %put Observations  =  &obsCount;
      %put Param_1_init  =  &Param_1_init;
      %put Param_2_init  =  &Param_2_init;
      %put Param_3_init  =  &Param_3_init;
      %put **********************************;
   %end;
%mend InitFitBURR;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitBURR;
   ods listing close;
   ods output 
            IterStop           =  _logl_
            ConvergenceStatus  =  _cs_ 
            covmat             =  _covmat_ 
            ;  

   proc nlp 
            data   = _logworkdata_ 
            outest = _parmnlp_
            &PROC_NLP_OPTIONS.
            ;

      parms 
            Param_1 = &Param_1_init, 
            Param_2 = &Param_2_init, 
            Param_3 = &Param_3_init;

      bounds Param_1 > 0 %Param_BoundsBURR(); 

      max logf;

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
               log(1 -( 1 + (THRESHOLD/Param_1)**Param_3 )**(-Param_2) + 1E-30);
      end;
      else if THRESHOLD=0 then do;
         lsdf = 0;
      end;
      else do;
         lsdf = Param_2 * log( 1 + (THRESHOLD/Param_1)**Param_3 );
      end;

      logf = lsdf+
             log(Param_2) +
             log(Param_3) +
             Param_3 * ( log(AMOUNT) - log(Param_1) )-
             log(AMOUNT) -
             (Param_2 + 1) * 
             log( 1 + (AMOUNT/param_1)**Param_3 )
             ;
   quit;
   ods listing;
%mend FitBURR;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriBURR;
   ods listing close;
   ods output 
               IterStop           =  _logl_
               ConvergenceStatus  =  _cs_ 
               covmat             =  _covmat_ 
               ;  

   proc nlp 
            data   = _LogWorkdata_KRI_(drop= PARAM_1:) 
            outest = _parmnlp_
            &PROC_NLP_OPTIONS.
            ;

      parms 
            Param_0 = &Param_1_init, 
            Param_2 = &Param_2_init, 
            Param_3 = &Param_3_init,
            &KRI_INIT_PARM
            ;

      bounds Param_0 > 0 %Param_BoundsBURR(); 
      max logf;
      LogParam_1 = Safe_ID(log(Param_0) + &KRI_MODEL_YES_INTERCEPT);
      Param_1    = exp(LogParam_1);

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf =  &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
                  log(1 -( 1 + (THRESHOLD/Param_1)**Param_3 )**(-Param_2) + 1E-30);
      end; 
      else if THRESHOLD=0 then do;
         lsdf = 0;
      end;
      else do;
         lsdf = Param_2 * log( 1 + (THRESHOLD/Param_1)**Param_3 );
      end;

      logf =   lsdf +
               log(Param_2) +
               log(Param_3) +
               Param_3 * ( LogAMOUNT - LogParam_1) -
               Log(AMOUNT) -
               (Param_2 + 1) * 
               log( 1 + (AMOUNT/Param_1)**Param_3 )
               ;
   quit;
   ods listing;
%mend FitKriBURR;

/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsBURR;
   %do i=1 %to 4;
      if &i < Param_2 * Param_3 then do;
         m&i= Param_1**&i * exp( LGamma(1+&i/Param_3 ) + LGamma(Param_2-&i/Param_3) - LGamma(Param_2));
      end;
      else do;
         m&i=.;
      end;
   %end;  

   if not missing(m1) then do;
      mean = m1;
   end;
   else do;
      mean = .;
   end;
   if not missing(m2) then do;
      std = sqrt(m2-m1*m1);
   end;
   else do;
      std = .;
   end;
   if not missing(m3) then do;
      skewness = (m3-3*m1*m2+2*(m1**3))/(std**3);
   end;
   else do;
      skewness = .;
   end;
   if not missing(m4) then do;
      kurtosis = (m4-4*m1*m3+6*(m1**2)*m2-3*(m1**4))/(std**4);
   end;
   else do;
      kurtosis = .;
   end;
%mend MomentsBURR;

%macro LimMomentsBurr(i,x,returnVal);
   /*  Use numerical integration (Simpson's Rule) of the ith power of the inverse cdf */
   %local n;
   %let   n = 300;   /* Must be an even integer */
   &returnVal = 0;

   /* CDF evaluated at x divided by number of steps */   
   %VCdfBURR(&x,_lm__h__) 
   _lm__h__ = _lm__h__/&n.;

   do _lm__j__ = 0 to &n;
      if _lm__j__ = 0 or _lm__j__=&n then do;
         _lm__coeff__ = 1;
      end;
      else if _lm__j__/2 = int(_lm__j__/2) then do;
         _lm__coeff__ = 2;
      end;
      else do;
         _lm__coeff__ = 4;
      end;

      &returnVal = &returnVal + _lm__coeff__*(((1-_lm__h__*_lm__j__)**(-1/Param_2)-1)**(1/Param_3))**&i.;
   end;

   &returnVal = &returnVal * Param_1**&i.;
   &returnVal = (&returnVal*_lm__h__/3) + (1-_lm__h__*&n.)*(&x.**&i.);
%mend LimMomentsBurr;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the location, param_1, param_2, and param_3 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitBURR;
   data _temp_;
      mean     = &mean;
      variance = &variance;
      kurtosis = &kurtosis;
      auss     = variance + mean*mean;
      Param_2  = 2*(mean*mean-auss)/(2*mean*mean-auss);
      Param_1  = mean*(param_2-1);
      Param_3  = 2;  
      call symputx("converge", put(mean*1e-8, BEST20.));
   run;
   Ods listing close;
   Ods output obsused = _cs_;
   proc model data=_temp_;
      dependent param_1 param_2 param_3;
         /*- fisrt 4 moment -*/
         m1=(param_1)*gamma(1+1/param_3)*gamma(param_2-1/param_3)/gamma(param_2);
         logm1=log(param_1)-log(param_3)+lgamma(1/param_3)+lgamma(param_2-1/param_3)-lgamma(param_2);
         m2=(param_1**2)*gamma(1+2/param_3)*gamma(param_2-2/param_3)/gamma(param_2);
         logm2 = 2*log(param_1)+ log(2)-log(param_3)+lgamma(2/param_3)+lgamma(param_2-2/param_3)-lgamma(param_2);
         m3=(param_1**3)*gamma(1+3/param_3)*gamma(param_2-3/param_3)/gamma(param_2);
         m4=(param_1**4)*gamma(1+4/param_3)*gamma(param_2-4/param_3)/gamma(param_2);

      eq.one=log(mean)-logm1;
      eq.two=log(mean**2+variance)-logm2;
      eq.three=log(kurtosis*(variance**2))-log(m4-4*m1*m3+6*m1*m1*m2-3*m1**4);

      solve param_1 param_2 param_3/ maxiter=10000 out=_sol_(keep=param_1 param_2
      param_3) converge=&converge;
   run;
   quit;
   Ods listing;
   data _null_;
      set _cs_;
      if label1 = "Failed" then do;
         if cvalue1 = 1 then do;
               call symputx("status",1);
               call symput("reason",sasmsg("&orvdds", "burr_no_mom_estimate_error", "quote"));
         end;
      end;
   run;  
   
   data _null_;
      set _sol_;
      call symputx("param_1", put(param_1, BEST20.));
      call symputx("param_2", put(param_2, BEST20.));
      call symputx("param_3", put(param_3, BEST20.));
   run;     
%mend MomFitBURR;

%macro Param_BoundsBURR();
               , 
   Param_2 > 0 , 
   Param_3 > 0
%mend;

%macro ValidateParamsBURR(Valid, Reason);
   if (Param_1 > 0) AND (Param_2 > 0) AND (Param_3 > 0) then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "burr_one_param_not_pos_error", "quote");
   end;
%mend ValidateParamsBURR;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_BURR(minimalInit = true);
   %let NEEDED_INPUT    = ;
   %let SCALE_TRANSFORM = ID;
   %let DIST_NAME       = Burr;
   %let PARAM_COUNT     = 3;
   %let COV_COUNT       = 6;
   %let ProbMeasure    = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..BURR;
         function Cdf_BURR(x,Param_1 &BURR_Param_List_Minus_1);
            return( 1 -(1+(x/Param_1)**Param_3)**(-Param_2));   
         endsub;

         function Pdf_BURR(x,Param_1 &BURR_Param_List_Minus_1);
            return(Param_2 * Param_3 * ((x / Param_1)**Param_3)/(x * (1+(x / Param_1 )**Param_3)**(Param_2 + 1)));
         endsub;
      run;
   %end;
%DONE:
   data BURR_STATUS;
      status          = 0;
   run;  
%mend ORVD_BURR;
