/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_exp.sas
 *   
 *   -Provides support for the Exponential distribution-
 *
 * This file contains macro fragments to implement calculation specific to the 
 * Exponential distribution
 *
 *  The macro %ORVD_EXP is at the end of this file. It is an empty macro that takes
 *  no paramter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_EXP macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 *  The list of other macro files that rely on fragments in this file:
 *
 ********************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        valueHolder = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfEXP(x,returnVal);
      &returnVal=1-exp(-(&x)/(&Param_1));   
%mend CdfEXP;

%macro VCdfEXP(x,returnVal);
      &returnVal=1-exp(-(&x)/Param_1);   
%mend VCdfEXP;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfEXP(x,returnVal);
      &returnVal=exp(-(&x)/(&Param_1))/(&Param_1);   
%mend PdfEXP;

%macro VPdfEXP(x,returnVal);
      &returnVal=exp(-(&x)/Param_1)/Param_1;   
%mend VPdfEXP;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantEXP(scaleFactor=1); /* scaleFactor is typically pnLimit/pLimit for body and tail fitting */
   data _temp_;
      set _temp_ nobs=nobs end=eof;
      retain truncdf 0;

      if _N_=1 then do;
         if THRESHOLD = 0 then do;
            truncdf = 0;
         end;
         else do;
            %CDFEXP(THRESHOLD,truncdf)
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf = ecdf * (1-truncdf) + truncdf;

      if  cdf/(&scaleFactor) LE 1.0 then do; 
         quant = -(&param_1) * log(1-cdf/(&scaleFactor));
      end;
      else do;
         quant = .;
      end;
   run;
%mend QuantEXP;
   
/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerEXP(scaleFactor=1, indata=, outdata=);
   data &outdata;
      set &indata;
      %CdfEXP(quant,cdf);
      cdf = cdf*&scaleFactor.;
      DParam_1 = (cdf - &scaleFactor)*quant/&param_1./&param_1.;
      keep quant cdf DParam_:;
   run;
%mend CdfDerEXP;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimEXP(loss);
    &loss = rand("EXPONENTIAL") * Param_1;
%mend SimEXP;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitEXP(summaryData = , KRI_FIT = NO);
   %let nparms=1;

   /*- initial guess based on moments -*/
   data _NULL_;
      set &summaryData;
      theta = sample_mean;
      call symputx("Param_1_init", put(theta, BEST20.));
      call symputx("obsCount", obsCount);
   run;

   %if &dist NE MIXDIST %then %do;  
      %put **********************************;
      %put Distribution =  EXP;
      %put Observations =  &obsCount;
      %put Param_1_init =  &Param_1_init;
      %put ******************************;
   %end;
%mend InitFitEXP;
   
/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitEXP;
   ods listing close;
   ods output 
            IterStop           =  _logl_
            ConvergenceStatus  =  _cs_ 
            covmat             =  _covmat_ 
            ;  

   proc nlp 
         data   = _logworkdata_ 
         outest = _parmnlp_
         &PROC_NLP_OPTIONS.
         ;

      parms 
         Param_1 = &Param_1_init;

      bounds Param_1 > 0; 
      max logf;

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf =  &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
                     log( 1 - exp(-THRESHOLD / Param_1) + 1E-30);
      end; 
      else if THRESHOLD=0 then do;
         lsdf = 0;
      end;
      else do;
         lsdf = (THRESHOLD / Param_1);
      end;

      logf = lsdf - log(Param_1) - AMOUNT / Param_1;
   quit;

   ods listing;
%mend FitEXP;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC MODEL
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKRIEXP;
   data _covmat_;
      RowName="Param_1";
      Param_1=.;
   run;

   ods listing close;
   ods output 
            IterStop           =  _logl_
            ConvergenceStatus  =  _cs_ 
            covmat             =  _covmat_ 
            ;  

      proc nlp 
            data   = _LogWorkdata_KRI_ (drop= PARAM_1:)  
            outest = _parmnlp_
            &PROC_NLP_OPTIONS.
            ;

         parms  Param_0  =  &Param_1_init,
                            &KRI_INIT_PARM;

         bounds Param_0 >0 %Param_BoundsEXP(); 

         max logf;
         LogParam_1 = Safe_ID( log(Param_0) + &KRI_MODEL_YES_INTERCEPT );
         Param_1    = exp(LogParam_1);

         if &THRESHOLDPERCENT >= 0 then do;
            lsdf = &THRESHOLDPERCENT / (1-&THRESHOLDPERCENT)*
                  log( 1 - exp(-THRESHOLD / Param_1) + 1E-30);
         end; 
         else if threshold=0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf = (THRESHOLD / Param_1 );
         end;

         logf =   lsdf - 
                  LogParam_1 - 
                  AMOUNT / Param_1 
                  ;
      quit;
   ods listing;
%mend FitKRIEXP;

/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsEXP;
    %do i=1 %to 4;
        m&i=(Param_1**&i) * gamma(&i+1);
    %end;

    mean  =  Param_1;
    std      =  Param_1;
    skewness =  2;
    kurtosis =  9;
%mend MomentsEXP;

%macro LimMomentsEXP(i,x,returnVal);
   &returnVal = 1;

   do _lm_ii=1 to &i;
      &returnVal = &returnVal * Param_1 * _lm_ii;
   end;

   &returnVal = &returnVal * cdf("gamma",&x/Param_1,&i,1);
%mend LimMomentsEXP;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the location, param_1, shape1, and shape2 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitEXP;
    %let Param_1=&mean;
%mend MomFitEXP;

%macro Param_BoundsExp();

%mend;

%macro ValidateParamsEXP(Valid, Reason);
   if Param_1 > 0 then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "exp_param_not_pos_error", "quote") ;
   end;
%mend ValidateParamsEXP;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_EXP(minimalInit  = true);    
   %let NEEDED_INPUT         = ;
   %let SCALE_TRANSFORM      = ID;
   %let DIST_NAME            = Exponential;
   %let PARAM_COUNT          = 1;
   %let CDF_TRANSFORM_FACTOR = 1;
   %let ProbMeasure          = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..EXP;
         function Cdf_Exp(x,Param_1 &EXP_Param_List_Minus_1);
            return( 1 - exp(-x /Param_1));   
         endsub;

         function Pdf_Exp(x,Param_1 &EXP_Param_List_Minus_1);
            return(exp(-x/Param_1)/Param_1);   
         endsub;
      run;
   %end; 
%DONE:
   data EXP_STATUS;
      status = 0;
   run;
%mend ORVD_EXP;
