/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_GAMMA.sas
 *   
 *   -Provides support for the GAMMA distribution-
 *
 * This file contains macro fragments to implement calculation specific to the 
 * GAMMA distribution
 *
 *  The macro %ORVD_GAMMA is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_GAMMA macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ********************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfGAMMA(x, returnVal);
    &returnVal = cdf("GAMMA", &x/&Param_1, &Param_2);
%mend CdfGAMMA;

%macro VCdfGAMMA(x, returnVal);
    &returnVal = cdf("GAMMA",&x, Param_2, Param_1);
%mend VCdfGAMMA;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfGAMMA(x, returnVal);
    &returnVal = pdf("GAMMA",&x, &Param_2, &Param_1);
%mend PdfGAMMA;

%macro VPdfGAMMA(x, returnVal);
   &returnVal = pdf("GAMMA",&x, Param_2, Param_1);
%mend VPdfGAMMA;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantGAMMA(scaleFactor=1);
   data _temp_;
      set _temp_ nobs=nobs end=eof;
      retain truncdf 0;

      if _N_=1 then do;
         if THRESHOLD = 0 then do;
            truncdf = 0;
         end;
         else do;
            %CDFGAMMA(THRESHOLD,truncdf)
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf=ecdf*(1-truncdf)+truncdf;

      if  cdf/(&scaleFactor) LE 1.0 then do; 
         quant=gaminv(cdf/(&scaleFactor),&param_2)*&param_1;;
      end;
      else do;
         quant = .;
      end;
   run;
%mend QuantGAMMA;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerGAMMA(scaleFactor=1, indata=, outdata=);
   data _temp_0;
      set &indata ;

      cdf =  &scaleFactor.*cdf("GAMMA", quant/&param_1, &param_2);
 
   run;

   proc model data=_temp_0 noprint;
      dependent cdf ;
      retain Param_1 &Param_1 Param_2 &Param_2;

      cdf = &scaleFactor.*cdf("GAMMA", quant/param_1, param_2);

      DParam_2 = getder(pred.cdf,Param_2);

      outvars DParam_2 quant;
         solve Param_2/out=&outdata(keep  =quant cdf DParam_2 );
   run;
   quit;
%mend CdfDerGAMMA;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimGAMMA(loss);
    &loss = rand("GAMMA", Param_2) * Param_1;
%mend SimGAMMA;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitGAMMA(summaryData = , KRI_FIT = NO);
   %let nparms=2;

   /*- initial guess based on moments -*/
   data _null_;
      set &summaryData;
      s = log(sample_mean) - sample_log_mean;
      param_2 = (3-s+sqrt( (s-3)**2 + 24*s ))/12/s;
      param_1 = sample_mean/param_2;

      call symputx("Param_1_init", put(param_1, BEST20.));
      call symputx("Param_2_init", put(param_2, BEST20.));
      call symputx("obsCount", obsCount);
   run;

   %if &dist NE MIXDIST %then %do;  
      %put ******************************;
      %put Distribution   =  &dist;
      %put Observations   =  &obsCount;
      %put Param_1_init    =  &Param_1_init;
      %put Param_2_init    =  &Param_2_init;
      %put ******************************;
   %end;
%mend InitFitGAMMA;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro Fitgamma;
   data _null_;
      scale_factor = &param_1_init*&param_2_init;
      call symputx("scale_factor", put(scale_factor, BEST20.));
      P1_init = 1/&Param_2_init;
      call symputx("P1_init", put(P1_init, BEST20.));
   run;

   ods listing close;
    ods output 
      IterStop           =  _logl_
      ConvergenceStatus  =  _cs_ 
      covmat             =  _covmat_ 
      ;  

    proc nlp 
            data   =  _logworkdata_ 
            outest =  _parmnlp_ 
            &PROC_NLP_OPTIONS.
            ;

      parms 
            P1 = &P1_init, 
            Param_2 = &Param_2_init
            ;

      bounds P1 > 0 %Param_BoundsGamma(); 
      max logf;
      _DELTA_ = 1.0e-6;

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
                  log(cdf("GAMMA", THRESHOLD/(P1*&scale_factor), Param_2) + 1E-30);     
      end; 
      else if THRESHOLD=0 then do;
         lsdf=0;
      end;
      else do;
         lsdf = log(sdf("GAMMA",THRESHOLD, Param_2, (P1*&scale_factor)) + 1E-30);
      end;

      logf =   (Param_2 - 1)*log(AMOUNT)-AMOUNT/(P1*&scale_factor)
               - Param_2 * log((P1*&scale_factor)) 
               - LGamma(Param_2) 
               - lsdf
               ;
    quit;
    ods listing;
    
   data _parmnlp_;
      set _parmnlp_;
        Param_1 = &scale_factor*P1;
      drop P1;
   run;

   %if %sysfunc(exist(_covmat_)) %then %do;
      data _covmat_;
         set _covmat_;

         if RowName = "P1" then do;
            Param_1  = &scale_factor * &scale_factor * P1;
            Param_2  = &scale_factor * Param_2;
            RowName = "Param_1";
         end;
         else do;
            Param_1 = &scale_factor * P1;
         end;
         drop P1;
      run;
   %end;
%mend Fitgamma;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriGAMMA;
   ods listing close;
   ods output 
               IterStop           =  _logl_
               ConvergenceStatus  =  _cs_ 
               covmat             =  _covmat_ 
               ;  

    proc nlp 
            data   =  _LogWorkdata_KRI_ (drop= PARAM_1:)  
            outest =  _parmnlp_ 
            &PROC_NLP_OPTIONS.
            ;

      parms 
            Param_0 = &Param_1_init, 
            Param_2 = &Param_2_init,
            &KRI_INIT_PARM
            ;

      bounds Param_0 > 0 %Param_BoundsGAMMA(); 
      max logf;
      _DELTA_    = 1.0e-6;
      LogParam_1 = Safe_ID(log(Param_0) + &KRI_MODEL_YES_INTERCEPT );
      Param_1    = exp(LogParam_1);

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
                  log(cdf("GAMMA", THRESHOLD/Param_1, Param_2) + 1E-30);     
      end; 
      else if THRESHOLD=0 then do;
         lsdf = 0;
      end;
      else do;
         lsdf = log(sdf("GAMMA",THRESHOLD, Param_2, Param_1) + 1E-30);
      end;

      logf =   (Param_2 - 1) * Log(AMOUNT) 
               - AMOUNT / Param_1 
               - Param_2 * LogParam_1  
               - LGamma(Param_2)
               - lsdf
               ;
   quit;
   ods listing;
%mend FitKriGAMMA;

/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsGAMMA;
   %do i=1 %to 4;
      m&i = (Param_1**&i); 

      do _lm_j = 1 to &i;
         m&i = m&i * (Param_2 + &i -_lm_j);
      end;
   %end;

   mean     = Param_1 * Param_2;
   std      = Param_1 * sqrt(Param_2);
   skewness = 2/sqrt(Param_2);
   kurtosis = 3 + 6/Param_2;
%mend MomentsGAMMA;

%macro LimMomentsGAMMA(i,x,returnVal);
   &returnVal = Param_1**&i*gamma(Param_2 +&i)*cdf("GAMMA", &x/Param_1,Param_2 +&i)/gamma(Param_2)
                +
                &x**&i*(1 - cdf("GAMMA", &x/Param_1,Param_2));
%mend LimMomentsGAMMA;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the location, param_1, param_2, and shape2 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitGAMMA;
    data _null_;
      Param_2 = &mean*&mean/&variance;
      Param_1 = &variance/&mean;
      call symputx("Param_1", put(Param_1, BEST20.));
      call symputx("Param_2", put(Param_2, BEST20.));
   run;
%mend MomFitGAMMA;

%macro Param_BoundsGamma();
   , Param_2>0
%mend;

%macro ValidateParamsGAMMA(Valid, Reason);
   if (Param_1 > 0) AND (Param_2 > 0) then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "gamma_one_param_not_pos_error", "quote");
   end;
%mend ValidateParamsGAMMA;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_GAMMA(minimalInit = true);
   %let NEEDED_INPUT         = ;
   %let SCALE_TRANSFORM      = ID;
   %let DIST_NAME            = GAMMA;
   %let PARAM_COUNT          = 2;
   %let CDF_TRANSFORM_FACTOR = 1;
   %let ProbMeasure          = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..GAMMA;
         function Cdf_GAMMA(x,Param_1 &GAMMA_Param_List_Minus_1);
            return( cdf("gamma", x, Param_2, Param_1));
         endsub;

         function Pdf_GAMMA(x,Param_1 &GAMMA_Param_List_Minus_1);
            return( pdf("GAMMA", x, param_2, Param_1  ));
         endsub;
      run;
   %end;
%DONE:
   data GAMMA_STATUS;
      status          = 0;
   run;  
%mend ORVD_GAMMA;
