/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_IGAUSS.sas
 *   
 *   -Provides support for the Inverse Gaussian distribution-
 *
 *
 * This file contains macro fragments to implement calculation specific to the 
 * Inverse Gaussian distribution
 *
 *  The macro %ORVD_IGAUSS is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_IGAUSS macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 *  The list of other macro files that rely on fragments in this file:
 *
 ********************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        valueHolder = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfIGAUSS(x, returnVal);
   &returnVal = cdf("IGAUSS", &x/(&Param_1), &Param_2);
%mend CdfIGAUSS;

%macro VCdfIGAUSS(x, returnVal);
   &returnVal = cdf("IGAUSS", &x/Param_1, Param_2);
%mend VCdfIGAUSS;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfIGAUSS(x, returnVal);
   &returnVal = pdf("IGAUSS", &x/(&Param_1), &Param_2) / (&Param_1);
%mend PdfIGAUSS;

%macro VPdfIGAUSS(x, returnVal );
   &returnVal = pdf("IGAUSS", &x/Param_1, Param_2) / Param_1;
%mend VPdfIGAUSS;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantIGAUSS(scaleFactor = 1);
   data _temp_;
      set _temp_ nobs = nobs;
      retain truncdf 0;

      if _N_ = 1 then do;
         if THRESHOLD = 0 then do;
            truncdf = 0;
         end;
         else do;
            %CDFIGAUSS(THRESHOLD, truncdf)
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf = ecdf * (1 - truncdf) + truncdf;

      if  cdf / (&scaleFactor) LE 1.0 then do; 
            quant = quantile("IGAUSS", cdf / (&scaleFactor), &Param_2) * &Param_1;
      end;
      else do;
         quant = .;
      end;
   run;
%mend QuantIGAUSS;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerIGAUSS(scaleFactor = 1, indata =, outdata =);
   data _temp_0;
      set &indata;
   
      cdf =  &scaleFactor.*cdf("IGAUSS", quant / (&Param_1), &Param_2);

   run;

   proc model data = _temp_0 noprint;
      dependent cdf ;
      retain 
            Param_1 &Param_1 
            Param_2 &Param_2
            ;

      cdf =  &scaleFactor. * cdf("IGAUSS", quant / (Param_1), Param_2); 

      DParam_2 = getder(pred.cdf, Param_2);

      outvars DParam_2 quant;
      solve Param_2 /out = &outdata(keep = cdf quant DParam_2 );
   run;
   quit;
%mend CdfDerIGAUSS;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimIGAUSS(loss);
   _u_   = rand("UNIFORM");
   &loss = quantile("IGAUSS", _u_, Param_2) * Param_1;
%mend SimIGAUSS;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitIGAUSS(summaryData = , KRI_FIT = NO);
   %let nparms = 2;

   /*- initial guess based on moments -*/
   data _null_;
      set &summaryData;
      mean     = sample_mean;
      variance = sample_variance;

      call symputx("Param_1_init", put(mean, BEST20.));
      call symputx("Param_2_init", put(mean * mean / variance, BEST20.));
      call symputx("obsCount", obsCount);
   run;

   %if &dist EQ IGAUSS %then %do;  
      %put **********************************;
      %put Distribution    =  &dist;
      %put Observations    =  &obsCount;
      %put Param_1_init    =  &Param_1_init;
      %put Param_2_init    =  &Param_2_init;
      %put **********************************;
   %end;
%mend InitFitIGAUSS;
 
/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitIGAUSS;
   data _null_;
      scale_factor = &Param_1_init / &Param_2_init;
      call symputx("Scale_Factor", put(scale_factor, BEST20.));
      P1_init = &Param_2_init;
      call symputx("P1_init", put(P1_init, BEST20.));
   run;

   ods listing close;
   ods output 
               IterStop          =  _logl_
               ConvergenceStatus =  _cs_ 
               covmat            =  _covmat_ 
               ;  

   proc nlp 
            data   =  _logworkdata_
            outest =  _parmnlp_
            &PROC_NLP_OPTIONS.
            ;
      
      parms 
            P1       = &P1_init,
            Param_2  = &Param_2_init
            ;
      bounds P1 > 0 %Param_BoundsIGAUSS(); 
         
      max logf;

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT / (1 - &THRESHOLDPERCENT) *
                  log(cdf("IGAUSS", THRESHOLD / P1 / &scale_factor, Param_2) + 1E-30);
      end; 
      else do;
         if THRESHOLD = 0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf = log( sdf("IGAUSS", THRESHOLD / P1 / &scale_factor, Param_2) + 1E-30);
         end;
      end;

      logf = 0.5 * (log(P1) + log(&scale_factor) + log(Param_2) - log(2) - log(3.141592) - 3 * log(AMOUNT) )
             - Param_2 * Amount / P1 / &scale_factor / 2
             + Param_2
             - Param_2 * P1 * &scale_factor / Amount/ 2
             - lsdf;
   quit;
   ods listing;

   data _parmnlp_;
      set _parmnlp_;
      Param_1 = &scale_factor * P1;
      drop P1;
   run;

   %if %sysfunc(exist(_covmat_)) %then %do;
      data _covmat_;
         set _covmat_;

         if RowName = "P1" then do;
            Param_1  = &scale_factor * &scale_factor * P1;
            Param_2  = &scale_factor * Param_2;
            RowName = "Param_1";
         end;
         else do;
            Param_1 = &scale_factor * P1;
         end;

         drop P1;
      run;
   %end;
%mend FitIGAUSS;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriIGAUSS;
   ods listing close;
   ods output 
               IterStop          =  _logl_
               ConvergenceStatus =  _cs_ 
               covmat            =  _covmat_ 
               ;  

   proc nlp 
         data   =  _LogWorkdata_KRI_(drop = PARAM_1:)  
         outest =  _parmnlp_
         &PROC_NLP_OPTIONS.
         ;

      parms 
            Param_0  = &Param_1_init,
            Param_2  = &Param_2_init,
                       &KRI_INIT_PARM
                       ;

      bounds Param_0 > 0 %Param_BoundsIGAUSS();
 
      max logf;
      LogParam_1 = Safe_ID(log(Param_0) + &KRI_MODEL_YES_INTERCEPT );
      Param_1    = exp(LogParam_1);

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT / (1 - &THRESHOLDPERCENT)*
                  log(cdf("IGAUSS", THRESHOLD / Param_1, Param_2) + 1E-30);
      end; 
      else do;
         if THRESHOLD = 0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf = log( sdf("IGAUSS",THRESHOLD / Param_1, Param_2) + 1E-30);
         end;
      end;

      logf = 0.5 *( LogParam_1
                    + log(Param_2) 
                    - log(2)
                    - log(3.141592)
                    - 3 * log(AMOUNT) 
                    )
                    - Param_2 * Amount / Param_1 / 2
                    + Param_2 
                    - Param_2 * Param_1 / Amount / 2
                    - lsdf;
   quit;
   ods listing;
%mend FitKriIGAUSS;

/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsIGAUSS;
   Mean     = Param_1 ;
   Std      = Param_1 / sqrt(Param_2);
   Skewness = 3 / sqrt(Param_2);
   Kurtosis = 3 + 15 / Param_2;

   m1 = Mean;
   m2 = Std**2 + m1 * m1;
   m3 = Std**3 * Skewness - 2 * (m1**3) + 3 * m1 * m2;
   m4 = Std**4 * Kurtosis + 3 * (m1**4) - 6 * (m1**2) * m2 + 4 * m1 * m3;
%mend MomentsIGAUSS;

%macro LimMomentsIGAUSS(i, x, returnVal);
   /* In this case there is no closed form solution, so use numerical  */
   /* integration of the ith power of the inverse cdf (Simpson's Rule) */
   %local n;
   %let n = 300;   /* Must be an even integer */
   &returnVal = 0;

   /* CDF evaluated at x divided by number of steps */   
   %VCdfIGAUSS(&x, _lm__h__); 
   _lm__h__ = _lm__h__ / &n.;

   do _lm__j__ = 1 to &n;
      if _lm__j__ = 0 or _lm__j__ = &n then do;
         _lm__coeff__ = 1;
      end;
      else if _lm__j__/2 = int(_lm__j__/2) then do;
         _lm__coeff__ = 2;
      end;
      else do;
         _lm__coeff__ = 4;
      end;

      &returnVal = &returnVal + _lm__coeff__ * quantile('IGAUSS', _lm__h__ * _lm__j__, Param_2)**&i.;
   end;

   &returnVal = &returnVal * Param_1**&i.;
   &returnVal = (&returnVal * _lm__h__/3) + (1 - _lm__h__ * &n.) * (&x.**&i.);
%mend LimMomentsIGAUSS;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding
 *                    values to the location, param_1, param_2, and shape2 macros
 *
 *   No Parameters
 *****************************************************************************************/
%macro MomFitIGAUSS;
   data _null_;
      Param_2 = &mean**2 / &variance;
      Param_1 = &mean;
      call symputx("param_1", put(param_1, BEST20.));
      call symputx("param_2", put(param_2, BEST20.));
   run;
%mend MomFitIGAUSS;

%macro Param_BoundsIGAUSS();
   , Param_2 > 0
%mend;

%macro ParamNamesIGAUSS(nameList =);
   %let &nameList = Location Scale;
%mend;

%macro ValidateParamsIGAUSS(Valid, Reason);
   if (Param_1 > 0) AND (Param_2 > 0)then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "igauss_one_param_not_pos_error", "quote");
   end;
%mend ValidateParamsIGAUSS;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_IGAUSS(minimalInit = true);
   %let NEEDED_INPUT    = ;
   %let SCALE_TRANSFORM = ID;
   %let DIST_NAME       = Inverse Gaussian;
   %let PARAM_COUNT     = 2;
   %let ProbMeasure     = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..IGAUSS;
         function Cdf_IGAUSS(x,Param_1 &IGAUSS_Param_List_Minus_1);
            return( cdf("IGAUSS", x / (Param_1), Param_2));
         endsub;

         function Pdf_IGAUSS(x,Param_1 &IGAUSS_Param_List_Minus_1);
            return( pdf("IGAUSS", x / (Param_1 ), Param_2) / (Param_1));
         endsub;
      run;
   %end;
%DONE:
   data IGAUSS_STATUS;
      status = 0;
   run;
%mend ORVD_IGAUSS;
