/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_LNGROBUST.sas
 *   
 *   -Provides support for the Robust version of the LogNormal-Gamma distribution-
 *
 *   This file contains macro fragments to implement calculation specific to the 
 *   LogNormal-Gamma distribution
 *
 *  The macro %ORVD_LNGROBUST is at the end of this file. It is an empty macro that takes
 *  no parameters and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_LNGROBUST macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ******************************************************************************/
%macro LNGROBUST_rep;
   do _lng_rep_i = 1 to &LNGROBUST_NQUANT;
      if Param_3 < 0.00002 then do;
         q[_lng_rep_i] = sqrt(1 + probit(_lng_rep_i / (&LNGROBUST_NQUANT + 1)) * sqrt(Param_3));
      end;
      else do;        
         q[_lng_rep_i] = sqrt(gaminv(_lng_rep_i /(&LNGROBUST_NQUANT + 1), 1 / Param_3)  *Param_3);
      end;
   end;
%mend LNGROBUST_rep;   

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfLNGRobust(x, returnVal);
   if &Param_3 < 1e-8 then do;
      &returnVal = cdf("LOGNORMAL", &x, &Param_1, &Param_2);   
   end;
   else do;
      &returnVal = 0;
      retain q&returnVal.1-q&returnVal&LNGROBUST_NQUANT;
      array q&returnVal q&returnVal.1-q&returnVal&LNGROBUST_NQUANT; 

      if _N_ = 1 then do;
         do lng_i = 1 to &LNGROBUST_NQUANT;
            if &Param_3 < 0.00002 then do;
               q&returnVal[lng_i] = sqrt(1 + probit(lng_i / (&LNGROBUST_NQUANT + 1)) * sqrt(&Param_3));
            end;
            else do;
               q&returnVal[lng_i] = sqrt(gaminv(lng_i / (&LNGROBUST_NQUANT + 1), 1 / &Param_3) * &Param_3);
            end;
         end;
      end; 

      do lng_i = 1 to &LNGROBUST_NQUANT;
         if &Param_2 * q&returnVal[lng_i] < 1e-8 then do;
            &returnVal = &returnVal + 1;
         end;
         else do;
            &returnVal = &returnVal + cdf('LOGNORMAL', &x, &Param_1, &Param_2 * q&returnVal[lng_i]);
         end; 
      end;

      &returnVal = &returnVal / &LNGROBUST_NQUANT;
      drop q&returnVal.1-q&returnVal&LNGROBUST_NQUANT;
   end;
%mend CdfLNGRobust;

%macro VCdfLNGRobust(x, returnVal);
    if Param_3 < 1e-8 then do;
        &returnVal = cdf("LOGNORMAL", &x, Param_1, Param_2);   
    end;
    else do;
        &returnVal = 0;
        retain q&returnVal.1-q&returnVal&LNGROBUST_NQUANT;
        array q&returnVal q&returnVal.1-q&returnVal&LNGROBUST_NQUANT; 

        if ( _N_ = 1 ) or missing(_N_) then do;
            do lng_i = 1 to &LNGROBUST_NQUANT;
                if Param_3 < 0.00002 then do;
                    q&returnVal[lng_i] = sqrt(1 + probit(lng_i / (&LNGROBUST_NQUANT + 1)) * sqrt(Param_3));
                end;
                else do;
                    q&returnVal[lng_i] = sqrt(gaminv(lng_i / (&LNGROBUST_NQUANT + 1), 1 / Param_3) * Param_3);
                end;
            end;
         end;

       do lng_i = 1 to &LNGROBUST_NQUANT;
          if Param_2 * q&returnVal[lng_i] < 1e-8 then do;
             &returnVal = &returnVal + 1;
          end;
          else do;
              &returnVal = &returnVal + cdf('LOGNORMAL', &x, Param_1 , Param_2 * q&returnVal[lng_i]);
           end; 
       end;

        &returnVal = &returnVal / &LNGROBUST_NQUANT;
        drop q&returnVal.1-q&returnVal&LNGROBUST_NQUANT;
    end;
%mend VCdfLNGRobust;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfLNGRobust(x, returnVal);
   if &Param_3 < 1e-8 then do;
      &returnVal = pdf("LOGNORMAL", &x, &Param_1, &Param_2);   
   end;
   else do;
      &returnVal = 0;
      retain q1-q&LNGROBUST_NQUANT;
      array q q1-q&LNGROBUST_NQUANT; 

      if _N_ = 1 then do;
         do i = 1 to &LNGROBUST_NQUANT;
            if &Param_3 < 0.00002 then do;
               q[i] = sqrt(1 + probit(i/(&LNGROBUST_NQUANT + 1)) * sqrt(&Param_3));
            end;
            else do;
               q[i] = sqrt( gaminv(i / (&LNGROBUST_NQUANT + 1), 1 / &Param_3) * &Param_3);
            end;
         end;
      end; 

      do lng_i = 1 to &LNGROBUST_NQUANT;
         &returnVal = &returnVal + pdf('LOGNORMAL', &x, &Param_1, &Param_2 * q[lng_i]);
      end; 

      &returnVal=&returnVal / &LNGROBUST_NQUANT;
   end;
%mend PdfLNGRobust;

%macro VPdfLNGRobust(x, returnVal);
    if Param_3 < 1e-8 then do;
        &returnVal = pdf("LOGNORMAL",&x, Param_1, Param_2);   
    end;
    else do;
        &returnVal = 0;
        retain q1-q&LNGROBUST_NQUANT;
        array q q1-q&LNGROBUST_NQUANT; 

        if (_N_ = 1) or missing(_N_) then do;
            do lng_i = 1 to &LNGROBUST_NQUANT;
                if Param_3 < 0.00002 then do;
                    q[lng_i] = sqrt(1 + probit(lng_i / (&LNGROBUST_NQUANT + 1)) * sqrt(Param_3));
                end;
                else do;
                    q[lng_i] = sqrt(gaminv(lng_i / (&LNGROBUST_NQUANT + 1), 1 / Param_3) * Param_3);
                end;
            end;
        end; 

        do lng_i = 1 to &LNGROBUST_NQUANT;
            &returnVal = &returnVal + pdf('LOGNORMAL', &x, Param_1, Param_2 * q[lng_i]);
        end; 

        &returnVal = &returnVal / &LNGROBUST_NQUANT;
        drop q1-q&LNGROBUST_NQUANT;
    end;
%mend VPdfLNGRobust;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantLNGRobust(scaleFactor = 1);
    %orvd_LOGN();
    %QuantLOGN(scaleFactor = &scaleFactor);
    %orvd_LNGRobust();

    %if %sysevalf( &Param_3 > 1.0e-8, boolean) %then %do;
        data _parm_;
            mu    = &Param_1;
            sigma = &Param_2;
            beta  = &Param_3;
        run;

        data _temp_;
            set _temp_; 
            quant = log(quant);
        run;

        proc model 
                  data      = _temp_ 
                  parmsdata = _parm_ 
                  out       = _temp1_ 
                  noprint
                  ;
            parms 
                  mu 
                  sigma 
                  beta
                  ;
            F = 0;
            T = 0;

            do i = 1 to &LNGROBUST_NQUANT;
                q = sqrt( gaminv( i / (&LNGROBUST_NQUANT + 1), 1 / beta) * beta);
                F = F + probnorm( (quant - mu) / sigma / q) / &LNGROBUST_NQUANT;

                if THRESHOLD NE 0 then do;
                    T = T + probnorm( (log(THRESHOLD) - mu) / sigma / q )/&LNGROBUST_NQUANT;
                end;
            end;

            F = (F - T * (&scaleFactor)) / (1 - T * (&scaleFactor)) / (&scaleFactor);
            eq.one = F - ecdf;        
            solve quant/converge = 1e-6;
        quit;   

        data _temp_;
            merge 
                  _temp_ 
                  _temp1_(keep = ecdf quant)
                  ;
            quant = exp(quant);
        run;
    %end;
%mend QuantLNGRobust; 

%macro CdfDerLNGRobust(
                       scaleFactor = 1, 
                       indata      = , 
                       outdata     =
                       );
      %CdfDerLNG(scaleFactor = &scalefactor, indata = &indata, outdata = &outdata);
      %orvd_LNGRobust;
%mend CdfDerLNGRobust;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimLNGRobust(loss);
   /*   
   This fix is done here because when the "Param_3" parameter becomes too small,
   the simulation of the gamma variate fails. However this corresponds to a very
   peaky distribution around 1. Another way to look at this is that when the Param_3 
   parameter is so small, the distribution degenerates to a log-normal and that is
   the choice that should have been made if one had used the AIC or BSC instead of 
   simply the best likellihood.
   */
   if Param_3 < 0.000000001 then do;
      /*    Take absolut value here because we need to take a square
         root later. This introduces a minimal bias that can be
         ignored because the probability of this ever occuring 
         is extremely low */
      gamma = abs(rand("NORMAL") * sqrt(Param_3) + 1.0);
   end;
   else do;
      gamma = rand("GAMMA", 1 / Param_3) * Param_3;
   end;

   &loss = rand("NORMAL") * Param_2 * sqrt(gamma) + Param_1;
   &loss = exp(&loss);
%mend SimLNGRobust;

/*****************************************************************************************
 *   What it does  :  Actually does the initialization of the minimization 
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitLNGRobust(summaryData = , KRI_FIT = NO );
   %let nparms = 3;
   /* Using proc severity's initiallization routine ------------*/
%mend InitFitLNGRobust;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using Proc Severity
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitLNGRobust( inputdataset = _LogWorkdata_);
   proc severity 
                  data    = &inputdataset 
                  obj     = cvmobj 
                  outstat = _fitstat_ 
                  outest  = _outest_ 
                  EDF     = STANDARD 
                  covout 
                  noprint 
                  ;
      nloptions 
               maxiter = 1000 
               maxfunc =  10000 
               tech    = tr
               ;
      loss amount / lt = threshold 
                  %if &THRESHOLDPERCENT >= 0 %then %do; 
                     (pobs = %sysevalf(1 - &THRESHOLDPERCENT)) 
                  %end; ;
                  ;
      dist lngamma;
      cvmobj = _cdf_(amount);
      cvmobj = (cvmobj - (1 - (_N_ - 1)/&obscount.))**2;
   run;

   data _logl_; 
      set _fitstat_; 
      output;
      output;       
      label1  = "Objective Function";
      nvalue1 = -Neg2LogLike/2;
      output;
      keep  
          label1  
          nvalue1
          ;
   run;

   %if %sysfunc(exist(_outest_)) %then %do;
      data _covmat_; 
         set _outest_(rename = (
                                 _NAME_ = RowName 
                                 MU     = Param_1 
                                 sigma  = Param_2 
                                 shape  = Param_3
                                 ));
         where _type_ = "COV";

         if upcase( RowName ) = "MU" then do;
            RowName = "Param_1"; 
         end;
         else if upcase( RowName ) = "SIGMA" then do;
            RowName = "Param_2"; 
         end;
         else if upcase( RowName ) = "SHAPE" then do;
            RowName = "Param_3"; 
         end;
      run;
     
      data _parmnlp_; 
         set _outest_;
         where _type_ = "EST";
         rename MU    = Param_1;
         rename sigma = Param_2;
         rename shape = Param_3;
         _type_ = "PARMS";
      run;
   %end;
 
   data _cs_; 
      set _fitstat_;

      if _status_ > 310 then do; 
         status = 1;
         reason = "Convergence Failure";
      end;
      else do;
         status = 0;
         reason = "Convergence";
      end;

      keep status reason;
   run;

   %orvu_delete_data( data = 
                             _fitstat_ 
                             _outest_
                             );
%mend FitLNGRobust;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC MODEL
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriLNGRobust( inputdataset = _LogWorkdata_KRI_);
   data _cs_;
      reason = sasmsg("&orvdds", "rob_no_kri_fit_error", "quote");
   run;
%mend FitKriLNGRobust;

/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
/*- actually, the moments of logarithm of losses -*/
%macro MomentsLNGRobust;
    %do i = 1 %to 4;
        if Param_3 * Param_2 * Param_2 < 2/(&i * &i) then do;
            if Param_3 < 1.0e-8 then do;
                m&i = exp(&i * Param_1 + &i * &i * Param_2 * Param_2 / 2);
            end;
            else do;
            m&i = exp(&i * Param_1) * (1 - Param_3 * &i * &i * Param_2 * Param_2 / 2)**(-1 / Param_3);
            end;
        end;
        else do;
            m&i = .;
        end;
    %end;  

    MEAN     = m1;
    STD      = sqrt(m2 - m1 * m1);
    SKEWNESS = (m3 - 3 * m1 * m2 + 2 * (m1**3)) / (std**3);
    KURTOSIS = (m4 - 4 * m1 * m3 + 6 * (m1**2) * m2 - 3 * (m1**4)) / (std**4);
%mend MomentsLNGRobust;

%macro LimMomentsLNGRobust( 
                           i, 
                           x, 
                           returnVal
                           );
   array q _lm_q1-_lm_q&LNGROBUST_NQUANT;                
   %LNG_REP;
   &returnVal      = 0;
   _lm_tmp_Param_2 = Param_2;

   do _lm_quants = 1 to &LNGROBUST_NQUANT;
      Param_2 = _lm_tmp_Param_2 * q[_lm_quants];
      %LimMomentsLOGN(&i, &x, _lm_tmp_returnVal)
      &returnVal = &returnVal + _lm_tmp_returnVal;
   end;

   Param_2 = _lm_tmp_Param_2;
   &returnVal = &returnVal / &LNGROBUST_NQUANT;
%mend LimMomentsLNGRobust;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the Param_1, Param_2, Param_3, and shape2 macros
 *
 *   No Parameters 
 *****************************************************************************************/
 %macro MomFitLNGRobust;
   data _null_;
      call symputx("Param_1", put(&mean, BEST20.));
      p2 = sqrt(&variance);
      p3 = &kurtosis / 3 - 1;
      call symput("Param_2", put(p2, BEST20.));
      call symput("Param_3", put(p3, BEST20.));
   run;
%mend MomFitLNGRobust;

%macro LNGRobust_unboundedmoments;
   %do i=1 %to 4;
      if Param_3 * Param_2 * Param_2 < 2 / (&i * &i) then do;
         m&i = (&i * Param_1) + log(1 - Param_3 * &i * &i * Param_2 * Param_2 / 2)*(-1 / Param_3); 
         m&i = exp(m&i); 
      end;
      else do;
         m&i = .;
      end;
   %end;  

   MEAN     = m1;
   STD      = sqrt(m2 - m1 * m1);
   SKEWNESS = (m3 - 3 * m1 * m2 + 2 * (m1**3)) / (std**3);
   KURTOSIS = (m4 - 4 * m1 * m3 + 6 * (m1**2) * m2 - 3 * (m1**4)) / (std**4);
%mend LNGRobust_unboundedmoments;

%macro MomFit_real_LNGRobust;  
   %let Param_1_init = .;
   %let Param_2_init = .;
   %let Param_3_init = .;

   data _null_;
      c = &mean**2 + &variance;

      if c <= 0 then do; 
         p1 = .;
         p2 = .;
         call symputx("status", 1);
         call symput("reason", "%sysfunc(sasmsg(&orvdds, rob_cant_compute_MOMest_error, noquote))");
      end;
      else do;
         c  = sqrt(c);
         p1 = log( &mean**2 / c );
         p2 = sqrt( 2 * log(c / &mean) );
      end;

      Param_3 = .5 / ( 8 * p2**2);
      call symputx("Param_1_init", put(p1, BEST20.));
      call symputx("Param_2_init", put(p2, BEST20.));
      call symputx("Param_3_init", put(Param_3, BEST20.));
   run;

   ods listing;
   proc nlp  
            maxit  = 10000 
            maxfu  = 10000 
            tech   = tr 
            outest = _parmnlp_ 
            ;
      min obj;
      parms 
            Param_1 = &Param_1_init,
            Param_2 = &Param_2_init,
            Param_3 = &Param_3_init
            ;

      bounds %Param_BoundsLNGROBUST(); 
      %LNGRobust_unboundedmoments;       

      f1 = (mean - &mean);
      f2 = (std * std - &variance);
      f4 = (kurtosis - &kurtosis);
      obj = f1**2 + sqrt(f2**2) + (f4**2)**(.25);
   run;

   data _null_;
      set _parmnlp_;

      if _TYPE_ = "PARMS" then do;
         call symputx("Param_1", put(Param_1, BEST20.));
         call symputx("Param_2", put(Param_2, BEST20.));
         call symputx("Param_3", put(Param_3, BEST20.));
      end;
   run;  
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvdds, rob_cant_find_est_error, noquote));
      %let status = 1;
      %let reason = %sysfunc(sasmsg(&orvdds, rob_cant_find_est_error, noquote));
   %end;
%mend MomFit_real_LNGRobust;

%macro Param_BoundsLNGRobust();
   Param_2 > 0, 
   Param_3 > 1.0e-5
%mend;

%macro ParamNamesLNGRobust(nameList = );
    %let &nameList = Location Scale Shape;
%mend;

%macro ValidateParamsLNGROBUST(Valid, Reason);
   if Param_1 > 0 AND Param_2 > 0 AND Param_3 > 0 then do;
       &Valid  = "true";
   end;
   else do;
       &Valid  = "false";
       &Reason = sasmsg("&orvdds", "rob_one_param_not_pos_error", "quote");
   end;
%mend ValidateParamsLNGROBUST;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_LNGRobust(minimalInit = true);
   %LET LNGROBUST_KRI_USED  = NO;
   %let LNGROBUST_NQUANT    = 31;
   %let LNGROBUST_NQUANT_p1 = %eval(&LNGROBUST_NQUANT+1);
   %let NEEDED_INPUT        = ;
   %let SCALE_TRANSFORM     = LOG;
   %let DIST_NAME           = Log Normal Gamma;
   %let PARAM_COUNT         = 3;
   %let ProbMeasure         = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
       proc fcmp 
                  outlib  = &ORVAR_CMPLIB..LNGRobust 
                  library = sashelp.svrtdist
                  ;
         function Cdf_LNGRobust(x,Param_1 &LNGROBUST_Param_List_Minus_1);
            if Param_3 < 1e-8 then do;
               return( cdf("LOGNORMAL", x, Param_1, Param_2));   
            end;
            else do;
               tmp = 0;
               array q[&LNGROBUST_NQUANT_p1]/ nosymbols; 
               call dynamic_array(q, .); 

               if q[&LNGROBUST_NQUANT_p1] ne Param_3 then do;
                  q[&LNGROBUST_NQUANT_p1] = Param_3;
   
                  do i_lng_rep = 1 to &LNGROBUST_NQUANT;
                     if Param_3 < 0.00002 then do;
                        q[i_lng_rep] = sqrt(1 + probit(i_lng_rep/(&LNGROBUST_NQUANT + 1)) * sqrt(Param_3));
                     end;
                     else do;
                        q[i_lng_rep] = sqrt(gaminv(i_lng_rep/(&LNGROBUST_NQUANT + 1),1/(Param_3)) * (Param_3));
                     end;
                  end;
               end;

               do i_lng_rep = 1 to &LNGROBUST_NQUANT;
                  if Param_2 * q[i_lng_rep] < 1e-8 then do;
                     tmp = tmp + 1;
                  end;
                  else do;
                     tmp = tmp + cdf('LOGNORMAL', x, Param_1, Param_2 * q[i_lng_rep]);
                  end; 
               end;

               return(tmp/&LNGROBUST_NQUANT);
            end;
         endsub;
         
         function Pdf_LNGRobust(x,Param_1 &LNGROBUST_Param_List_Minus_1);
            if Param_3 < 1e-8 then do;
               return( pdf("LOGNORMAL", x, Param_1, Param_2));   
            end;
            else do;
               tmp = 0;

               array q[&LNGROBUST_NQUANT_p1]/ nosymbols; 
               call dynamic_array(q, .); 

               if q[&LNGROBUST_NQUANT_p1] ne Param_3 then do;
                  q[&LNGROBUST_NQUANT_p1] = Param_3;

                  do i_lng_rep = 1 to &LNGROBUST_NQUANT;
                     if Param_3 < 0.00002 then do;
                        q[i_lng_rep] = sqrt(1 + probit(i_lng_rep/(&LNGROBUST_NQUANT + 1)) * sqrt(Param_3));
                     end;
                     else do;
                        q[i_lng_rep] = sqrt(gaminv(i_lng_rep/(&LNGROBUST_NQUANT + 1), 1/Param_3) * Param_3);
                     end; 
                  end;
               end;

               do i_lng_rep = 1 to &LNGROBUST_NQUANT;
                  tmp = tmp + pdf('LOGNORMAL', x, Param_1, Param_2 * q[i_lng_rep]);
               end;

               return(tmp/&LNGROBUST_NQUANT);
            end;        
         endsub;

         /*--- Define Lognormal Gamma Distribution for proc severity -------*/
         function LNGamma_DESCRIPTION() $256;
            length desc $256;
            desc = "Lognormal Gamma Distribution.";
            return(desc);
         endsub;

         function LNGamma_SCALETRANSFORM() $3;
            length xform $3;
            xform = "LOG";
            return (xform);
         endsub;

         function LNGamma_PDF(x, Mu, Sigma, Shape);
            return( Pdf_LNGRobust(x,Mu, Sigma, Shape));
         endsub;

         function LNGamma_CDF(x, Mu,Sigma,Shape);
            return(Cdf_LNGRobust(x, Mu, Sigma, Shape));
         endsub;

         subroutine LNGamma_PARMINIT(dim, x[*], nx[*], F[*], Ftype, Mu, Sigma, Shape);
            outargs Mu,Sigma,Shape;
            array m[4] / nosymbols;
            array logx[1000] / nosymbols;

            do i = 1 to dim;
               logx[i] = log(x[i]);
            end;

            call svrtutil_rawmoments(dim, logx, nx, 4, m);
            mu       = m[1];  
            sigma    = sqrt(m[2] - m[1]**2);
            kurtosis = (m[4] - 4 * m[1] * m[3] + 6 * (m[1]**2) * m[2] - 3 * (m[1]**4)) / (sigma**4); 
            Shape    = min(kurtosis, 3);

            if shape < 0 then do;
               shape = 0.01;
            end;
         endsub;

         subroutine LNGamma_LOWERBOUNDS(Mu, Sigma, Shape);
            outargs Mu,Sigma,Shape;
            Mu    = .; /* Mu has no lower bound */
            Sigma = 0; /* Sigma > 0 */
            Shape = 0; /* Xi > 0 */
         endsub;
      quit;
   %end;     
%DONE:
   data LNGRobust_STATUS;
      status = 0;
   run;
%mend ORVD_LNGRobust;
