/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_LOGN.sas
 *   
 *   -Provides support for the LogNormal distribution-
 *
 * This file contains macro fragments to implement calculation specific to the 
 * LogNormal distribution
 *
 *  The macro %ORVD_LOGN is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_LOGN macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ********************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfLOGN(x,returnVal);
       &returnVal = cdf("LOGNORMAL", &x, &param_1, &param_2);    
%mend CdfLOGN;

%macro VCdfLOGN(x,returnVal);
       &returnVal = cdf("LOGNORMAL", &x, Param_1, Param_2);    
%mend VCdfLOGN;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfLOGN(x,returnVal);
       &returnVal = pdf("LOGNORMAL", &x, &param_1, &param_2);    
%mend PdfLOGN;
      
%macro VPdfLOGN(x,returnVal);
       &returnVal = pdf("LOGNORMAL", &x, Param_1, Param_2);    
%mend VPdfLOGN;
      
/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantLOGN(scaleFactor = 1);
   data _temp_;
      set _temp_ nobs=nobs end=eof;
      retain truncdf 0;

      if _N_ = 1 then do;
         if THRESHOLD = 0 then do;
            truncdf=0;
         end;
         else do;
            %CDFLOGN(THRESHOLD,truncdf);
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf   =  ecdf*(1-truncdf)+truncdf;

      if  cdf/(&scaleFactor) LE 1.0 then do; 
         quant = exp( &param_1 + &param_2 * probit(cdf/(&scaleFactor)) );
      end;
      else do;
         quant = .;
      end;
   run;
%mend QuantLOGN;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerLOGN(scaleFactor=1, indata=, outdata=);
   data &outdata;
      set &indata;
      %CdfLOGN(quant,cdf);
      cdf = &scaleFactor.*cdf;
      y = (log(quant) - &Param_1)/(&Param_2)/sqrt(2);
      DParam_1 = -1*&scaleFactor/(&Param_2)/sqrt(2*3.14159)*exp(-y**2);
      DParam_2 = -1*&scaleFactor/(&Param_2)/sqrt(3.14159)*exp(-y**2)*y;
      keep quant cdf DParam_:;
   run;
%mend CdfDerLOGN;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters      
 *****************************************************************************************/
%macro SimLOGN(loss);
    &loss=exp(rand("NORMAL") * Param_2 + Param_1);
%mend SimLOGN;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitLOGN(summaryData = , KRI_FIT = NO );
   %let nparms=2;    

   /*- initial guess based on moments -*/
   data _null_;
      set &summaryData;
      avg_uss  = sample_uss / obsCount;
      sigma_sq = log(avg_uss)-2*log(sample_mean);
      mu       = 0.5*log(avg_uss)-sigma_sq;  
      sigma    = sqrt(sigma_sq);

      call symputx("Param_1_init", put(mu, BEST20.));
      call symputx("Param_2_init", put(sigma, BEST20.));
      call symputx("ObsCount",obsCount);
   run;

   %if &dist NE MIXDIST %then %do;  
      %put **********************************;
      %put Distribution  =  LOGN;
      %put Observations  =  &obsCount;
      %put Param_1_init  =  &Param_1_init;
      %put Param_2_init  =  &Param_2_init;
      %put **********************************;
   %end; 
%mend InitFitLOGN;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitLOGN;
 
   ods listing close;
        ods output 
         IterStop           =  _logl_
         ConvergenceStatus  =  _cs_ 
         covmat             =  _covmat_ 
         ;  

   proc nlp 
         data   =  _logworkdata_ 
         outest =  _parmnlp_
         &PROC_NLP_OPTIONS.
         ;

         parms 
            Param_1 = &Param_1_init,
            Param_2 = &Param_2_init
            ;

         bounds %Param_BoundsLOGN(); 
         max logf;

         if &THRESHOLDPERCENT >= 0 then do;
           lsdf = - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT) * 
                  log(cdf("LOGNORMAL", THRESHOLD, Param_1, Param_2) + 1E-30);
         end; 
         else if THRESHOLD=0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf = log(sdf("LOGNORMAL", THRESHOLD, Param_1, Param_2) + 1E-30);
         end;

         logf =   - lsdf - 0.5 * log(2)
                  - lgamma(0.5)
                  - log(AMOUNT)
                  - log(Param_2)
                  - (log(AMOUNT) - Param_1) * (log(AMOUNT) - Param_1)
                  / (2 * Param_2 * Param_2)
                  ;
   quit;
   ods listing;
%mend FitLOGN;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriLOGN;
   ods listing close;
   ods output 
         IterStop           =  _logl_
         ConvergenceStatus  =  _cs_ 
         covmat             =  _covmat_ 
         ;  

   proc nlp 
         data   =  _LogWorkdata_KRI_(drop = PARAM_1: )
         outest =  _parmnlp_
         &PROC_NLP_OPTIONS.
         ;

         parms 
            Param_0 = &Param_1_init,
            Param_2 = &Param_2_init,
            &KRI_INIT_PARM
            ;

         bounds %Param_BoundsLOGN(); 
         max logf;
         Param_1 = Safe_ID(Param_0 + &KRI_MODEL_YES_INTERCEPT);
             
         if &THRESHOLDPERCENT >= 0 then do;
            lsdf = - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT) * 
                   log(cdf("LOGNORMAL", THRESHOLD, Param_1 ,Param_2) + 1E-30);
         end; 
         else if THRESHOLD=0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf = log( sdf("LOGNORMAL", THRESHOLD, Param_1, Param_2) + 1E-30);
         end;

         logf =   - lsdf - 0.5 * log(2)
                  - lgamma(0.5)
                  - LogAMOUNT
                  - log(Param_2)
                  - (LogAMOUNT - Param_1 ) * (LogAMOUNT - Param_1)
                  / (2 * Param_2 * Param_2); 
   quit;
   ods listing;
%mend FitKriLOGN;

/*- actually, the moments of logarithm of losses -*/
%macro MomentsLOGN;
    %do i=1 %to 4;
        m&i=exp(&i*param_1 + &i*&i * param_2*param_2 /2);
    %end;

    mean=m1;
    std=sqrt(m2-m1*m1);
    skewness=(m3-3*m1*m2+2*(m1**3))/(std**3);
    kurtosis=(m4-4*m1*m3+6*(m1**2)*m2-3*(m1**4))/(std**4);
%mend MomentsLOGN;

%macro LimMomentsLOGN(i, x, returnVal);
         &returnVal = exp(&i*Param_1+&i*&i*Param_2*Param_2/2);
         &returnVal = &returnVal * cdf("NORMAL", (log(&x) - Param_1 - &i * Param_2 * Param_2)/Param_2);
         &returnVal = &returnVal + &x**&i * (1- cdf("NORMAL", (log(&x)-Param_1)/Param_2));
%mend LimMomentsLOGN;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the param_1, param_2, shape1, and shape2 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitLOGN;
    data _null_;
      c = &mean**2+&variance;
      
      if c <= 0 then do;
         p1 = .;
         p2 = .;
         call symputx("status",1);
         call symput("reason", sasmsg("&orvdds", "lognorm_no_mom_estimate_error", "quote"));
      end;
         else do;
         c = sqrt(c);
      p1 = log(&mean**2/c);
      p2 = sqrt(2*log(c/&mean));
      end;
      
      call symputx("param_1",put(p1, BEST20.));
      call symputx("param_2", put(p2, BEST20.));
   run;
%mend MomFitLOGN;

%macro Param_BoundsLOGN();
   Param_2 > 0
%mend;

%macro ValidateParamsLOGN(Valid, Reason);
   if (Param_2 > 0) then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "lognorm_param_2_not_pos_error", "quote");
   end;
%mend ValidateParamsLOGN;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_LOGN(minimalInit = true);
   %let NEEDED_INPUT    = ;
   %let SCALE_TRANSFORM = LOG;
   %let DIST_NAME       = Log Normal;
   %let PARAM_COUNT     = 2;
   %let ProbMeasure     = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..LOGN;
         function Cdf_LOGN(x,Param_1 &LOGN_Param_List_Minus_1);
            return( cdf("LOGNORMAL", x, Param_1 , Param_2));    
         endsub;

         function Pdf_LOGN(x,Param_1 &LOGN_Param_List_Minus_1);
            return( pdf("LOGNORMAL", x, Param_1 , Param_2));    
         endsub;
      run;
   %end;
%DONELOGN:
   data LOGN_STATUS;
     status = 0;
   run;
%mend ORVD_LOGN;
