/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_LOGNGPD.sas
 *   
 *   -Provides support for the mixture of a LogNormal distribution
 *   -to represent the body and a Generalized Pareto to represent the tail
 *
 * This file contains macro fragments to implement calculation specific to the 
 * LogNormal distribution
 *
 *  The macro %ORVD_LOGNGPD is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_LOGN macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ******************************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        valueHolder = name of variable inton which result is stored
 *****************************************************************************************/

/*The Cumulative Distribution Function (using macro variables as parameters)*/
%macro CdfLOGNGPD(x,returnVal);
   Param_pLimit = cdf("LOGNORMAL", (&Param_xLimit * exp(&Param_1)),(&Param_1), &Param_2);

   if &x LE &Param_xLimit * exp(&Param_1) then do;
      &returnVal  =  cdf("LOGNORMAL",&x,(&Param_1), &Param_2) * 
                  &Param_pnLimit/Param_pLimit;    ;    
   end;
   else do;
      if &Param_4 LE 0.00000000001 then do;
         &returnVal = &Param_pnLimit + 
                   (1-&Param_pnLimit)*
                   ( 1 - exp( -(&x - &Param_xLimit * exp(&Param_1))/
                   &Param_3/exp(&Param_1)));
      end;
      else do;
         &returnVal = &Param_pnLimit + 
                   (1-&Param_pnLimit) * 
                   (1-(1 + &Param_4*(&x - &Param_xLimit * exp(&Param_1))/
                    &Param_3/exp(&Param_1))**(-1/&Param_4));
      end;
   end;
%mend CdfLOGNGPD;

/*The Cumulative Distribution Function (using dataset variables as parameters)*/
%macro VCdfLOGNGPD(x,returnVal);
   Param_pLimit = cdf("LOGNORMAL", Param_xLimit * exp(Param_1), Param_1, Param_2) ;    

   if &x LE Param_xLimit * exp(Param_1) then do;
      &returnVal  =  cdf("LOGNORMAL", &x, Param_1, Param_2) * 
                  Param_pnLimit / Param_pLimit;    
   end;
   else do;
      if Param_4 LE 0.00000000001 then do;
         &returnVal = Param_pnLimit + 
                   (1 - Param_pnLimit) * 
                   (1-exp(-(&x - Param_xLimit * exp(Param_1))/
                    Param_3/exp(Param_1)));
      end;
      else do;
         &returnVal = Param_pnLimit + 
                   (1 - Param_pnLimit) * 
                   (1-(1 + Param_4 * (&x - Param_xLimit * exp(Param_1))/
                   Param_3/exp(Param_1))**(-1/Param_4));
      end;
   end;
%mend VCdfLOGNGPD;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        valueHolder = name of variable inton which result is stored
 *****************************************************************************************/
/*The Probability Distribution Function (using macro variables as parameters)*/
%macro PdfLOGNGPd(x,returnVal);
   Param_pLimit = cdf("LOGNORMAL", (&Param_xLimit * exp(&Param_1)), (&Param_1), &Param_2);

   if &x LE &Param_xLimit * exp(&Param_1) then do;
      &returnVal =   pdf("LOGNORMAL", &x, (&Param_1), &Param_2) * 
                  &Param_pnLimit / Param_pLimit;    
   end;
   else do;
      if &Param_4 LE 0.00000000001 then do;
         &returnVal = (1-&Param_pnLimit) * 
                    exp(-(&x - &Param_xLimit * exp(&Param_1))/
                    &Param_3/exp(&Param_1))/
                    &Param_3/exp(&Param_1);
      end;
      else do;
         &returnVal = (1-&Param_pnLimit) * 
                   ((1 + (&Param_4)*(&x - &Param_xLimit * exp(&Param_1))/
                   &Param_3/exp(&Param_1))**(-1/&Param_4 - 1))/
                   &Param_3/exp(&Param_1);
      end;
   end;
%mend PdfLOGNGPD;

/*The Probability Distribution Function (using dataset variables as parameters)*/
%macro VPdfLOGNGPd(x,returnVal);
   Param_pLimit = cdf("LOGNORMAL", Param_xLimit * exp(Param_1), Param_1, Param_2);    

   if &x LE Param_xLimit * exp(Param_1) then do;
      &returnVal = pdf("LOGNORMAL",&x, (Param_1), Param_2) * 
                Param_pnLimit/Param_pLimit;    
   end;
   else do;
      if Param_4 LE 0.00000000001 then do;
         &returnVal = (1 - Param_pnLimit) * 
                   exp(-(&x - Param_xLimit * exp(Param_1))/
                   Param_3/exp(Param_1))/
                   Param_3/exp(Param_1);
      end;
      else do;
         &returnVal = (1 - Param_pnLimit) * 
                   ((1 + (Param_4)*(&x - Param_xLimit * exp(Param_1))/
                   Param_3/exp(Param_1))**(-1/Param_4 -1))/
                   Param_3/exp(Param_1);
      end;
   end;
%mend VPdfLOGNGPD;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantLOGNGPD(scaleFactor=1);
   data _NULL_;
      Eff_Param_xLimit = &Param_xLimit * exp(&Param_1);
      Eff_Param_3 = &Param_3 * exp(&Param_1);

      %CdfLOGN(Eff_Param_xLimit ,tmp); 

      call symput("Param_pLimit", put(tmp,BEST20.)); 
      call symput("Eff_Param_xLimit", put(Eff_Param_xLimit,BEST20.) );
      call symput("Eff_Param_3", put(Eff_Param_3,BEST20.));
   run;

   %QuantLOGN(scaleFactor = (&Param_pnLimit/&Param_pLimit))
  
    %if %sysevalf(&Param_4 > 0, boolean) %then %do;
        %let quantTail = &Eff_Param_3 *((1/(1-tmp_cdf))**(&Param_4)-1)/(&Param_4);
    %end;
    %else %do;
        %let quantTail = - &Eff_Param_3 * log(1-tmp_cdf);
    %end;

    data _temp_;
      set _temp_ ;
   
      if (cdf > &Param_pnLimit) then do;
         tmp_cdf = (cdf-&Param_pnLimit)/(1-&Param_pnLimit);
         quant = &Eff_Param_xLimit + (&quantTail);
      end;

      drop tmp_cdf;
    run;
%mend QuantLOGNGPD;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerLOGNGPD(scaleFactor=1, indata=, outdata=);

   data _temp_0;
      set &indata;
      retain Param_plimit;

      Param_pLimit = cdf("LOGNORMAL", (&Param_xLimit * exp(&Param_1)),(&Param_1), &Param_2);

      if quant LE &Param_xLimit * exp(&Param_1) then do;
         cdf  = cdf("LOGNORMAL",quant,(&Param_1), &Param_2) * &Param_pnLimit/Param_pLimit;    ;    
      end;
      else do;
         if &Param_4 LE 0.00000000001 then do;
            cdf = &Param_pnLimit + (1-&Param_pnLimit)*(1-exp(-(quant - &Param_xLimit * exp(&Param_1))/&Param_3/exp(&Param_1)));
         end;
         else do;
            cdf = &Param_pnLimit + (1-&Param_pnLimit)*(1-(1 + &Param_4*(quant - &Param_xLimit * exp(&Param_1))/&Param_3/exp(&Param_1))**(-1/&Param_4));
         end;
      end;

      cdf1=cdf;
      cdf2=cdf;
   run;

   proc model data=_temp_0 noprint;
      dependent cdf cdf1 cdf2 ;
      retain Param_1 &Param_1 Param_2 &Param_2 Param_3 &Param_3 Param_4 &Param_4;

      Param_pLimit = cdf("LOGNORMAL", (&Param_xLimit * exp(Param_1)),(Param_1), Param_2);

      if quant LE &Param_xLimit * exp(Param_1) then do;
         cdf_temp  = cdf("LOGNORMAL",quant,(Param_1), Param_2) * &Param_pnLimit/Param_pLimit;    ;    
      end;
      else do;
         if Param_4 LE 0.00000000001 then do;
            cdf_temp = &Param_pnLimit + (1-&Param_pnLimit)*(1-exp(-(quant - &Param_xLimit * exp(Param_1))/Param_3/exp(Param_1)));
         end;
         else do;
            cdf_temp = &Param_pnLimit + (1-&Param_pnLimit)*(1-(1 + Param_4*(quant - &Param_xLimit * exp(Param_1))/Param_3/exp(Param_1))**(-1/Param_4));
         end;
      end;

      cdf = cdf_temp;
      cdf1=pred.cdf;
      cdf2=pred.cdf;

      DParam_2 = getder(pred.cdf,Param_2);
      DParam_3 = getder(pred.cdf1,Param_3);
      DParam_4 = getder(pred.cdf2,Param_4);

      outvars DParam_2 DParam_3 DParam_4 quant;
      solve Param_2 Param_3 Param_4 /
      out=&outdata(keep=cdf quant DParam_2 DParam_3 DParam_4 );
   run;
   quit;
%mend CdfDerLOGNGPD;
/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimLOGNGPD(loss);
   p=rand("UNIFORM");

   if p LE Param_pnLimit then do;
      _body_loop_count = 0;
      do until(&loss LE Param_xLimit );
         _body_loop_count = _body_loop_count + 1;
         if _body_loop_count < 100 then do;
            &loss = exp( rand("NORMAL") * Param_2 );
         end;
         else do;
            &loss = rand("UNIFORM") * Param_xLimit;
         end;
      end;
      drop _body_loop_count;
   end;

   else do;
      if Param_4 > 0 then do;
            &loss = Param_xLimit  + ((( 1 / rand("UNIFORM") )**(Param_4) - 1  ) * Param_3 )/ Param_4 ;        
      end;
      else do;
            &loss = Param_xLimit  - Param_3 * log(rand("UNIFORM"));
      end;
   end;

   &loss = &loss * exp(Param_1);
%mend SimLOGNGPD;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitLOGNGPD(summaryData = , KRI_FIT = NO);
   %InitFitLogn(summaryData = &summaryData, KRI_Fit = &KRI_FIT);
   %let nparms = &PARAM_COUNT;
   %put **********************************;
   %put Distribution    =  LOGNGPD;
   %put Observations    =  &obsCount;
   %put Param_1_init    =  &Param_1_init;
   %put Param_2_init    =  &Param_2_init;
   %put **********************************;


   /* BEGIN Does the fit use KRI's?  */
   %if %upcase(&KRI_FIT) EQ YES %then %do;
      %let _Loss_Data_    = _LogWorkdata_KRI_;
      %let _Loss_Amount_  = KRI_ADJ_AMOUNT;
   %end; 
   %else %do; 
      %let _Loss_Data_    = _LogWorkdata_;
      %let _Loss_Amount_  = AMOUNT;
   %end;
   /* END Does the fit use KRI's?    */
   
   /* If the tail boundary point is supplied, figure out      */
   /* what tail percentile it represents                      */      
   %if "%upcase(&TailBoundaryChoice)" NE "TAILBOUNDARYPERCENT" %then %do;
      %let TailCount = 0;

      data _NULL_;
         set   &_Loss_Data_
               nobs = TotalCount 
               end  = TheEnd
               ;
         where AMOUNT <= &TailBoundaryValue;

         retain BodyCount 0;
         BodyCount = BodyCount + 1;

         if TheEnd then do;
            call symputX("TotalCount", TotalCount);
            call symputX("BodyCount", BodyCount);
            call symputX("TailCount", TotalCount - BodyCount);
            call symputX("Param_pnLimit",put(BodyCount/TotalCount, BEST20.));
         end;
      run;

      %let TailBoundaryPercent = %sysevalf(&Param_pnLimit * 100);

      %if %sysevalf(&TailCount < 2, boolean) %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
         %let status=1;
         %let reason=%sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
         %goto LOCAL_DONE;
      %end;
   %end;
   %else %do;
      %let Param_pnLimit = %sysevalf(&TailBoundaryPercent/100);
   %end;

   %let PercentLimit = %sysevalf(100 * &Param_pnLimit.);

   /* At this point, the Param_pnLimit macro is defined */
   /* Get the right Param_xLimit                        */
   proc univariate data = &_Loss_Data_ PCTLDEF = 4 noprint;
      output 
            out     = __tmp__ 
            pctlpts = &PercentLimit 
            pctlpre = PCTL
            ;
      var &_Loss_Amount_;
   quit;

   proc transpose data=__tmp__ out =__tmp__;
   run;

   data _null_;
      set __tmp__;
      call symputx("Param_xLimit",put(COL1, BEST20.));
   run;

   /* Now Param_xLimit is defined in the right space */
   /* i.e in the AMOUNT space if no KRI's are used   */
   /*     in the AMOUNT/KRI_SCALE if KRI's are used  */

   /* BEGIN - Get initial values for tail parameters */
   proc summary data = &_Loss_Data_ noprint;
      var &_Loss_Amount_;
      where &_Loss_Amount_ > &Param_xLimit; 

      output 
            out  = _tail_stat_(keep=sample_mean2 sample_variance2 sample_size2)
            n    = sample_size2
            mean = sample_mean2 
            var  = sample_variance2 
            ; 
   run;

   %let TailStatCount = %str();
   %orvu_getObsCount(_tail_stat_,TailStatCount)

   %if %sysevalf(&TailStatCount = 0, boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %let status=1;
      %let reason=%sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %goto LOCAL_DONE;
   %end;

   proc transpose 
      data = _tail_stat_ 
      out  = _tail_stat_;
   run;
        
   data _null_;
      set _tail_stat_;
      call symputX(_name_, put(col1, BEST20.));
   run;
   %if %sysevalf(&sample_size2 < 2, boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %let status=1;
      %let reason=%sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %goto LOCAL_DONE;
   %end;

   data _null_;
      if &sample_variance2 > &sample_mean2**2 then do;
         alpha = (&sample_variance2 - &sample_mean2**2)/2 / &sample_variance2 ;
      end;
      else do;
         alpha = .49;
      end;

      theta = &sample_mean2*(1 - alpha);
      call symputX("Param_3_init", put(theta/exp(&Param_1_init.), BEST20.));
      call symputX("Param_4_init", put(alpha, BEST20.));
   run;
   /* END - Get initial values for tail parameters */

%LOCAL_DONE:

%mend InitFitLOGNGPD;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitLOGNGPD;

   /* This piece of code increases Param_pnLimit if the probability of observability is provided     */
   /* New_pnLimit = Old_pnLimit + (1-Old_pnLimit)*p0 where p0 is the probability below the threshold */
   %if %sysevalf(&THRESHOLDPERCENT >= 0, BOOLEAN) %then %do;
      %let Param_pnLimit = %sysevalf(&Param_pnLimit + (1-&Param_pnLimit)* &THRESHOLDPERCENT); 
   %end;

    ods listing close;
    ods output 
               IterStop          =  _logl_
               ConvergenceStatus =  _cs_ 
               covmat            =  _covmat_
               ;  

   proc nlp 
            data   = _LogWorkdata_  
            outest = _parmnlp_ 
            &PROC_NLP_OPTIONS.
            ;

      /* The list of parameters */
      parms 
            Param_1 = &Param_1_init,
            Param_2 = &Param_2_init,
            Param_3 = &Param_3_init,
            Param_4 = &Param_4_init
            ;

      /* The bounds on parameters */
      bounds 
            Param_2 > 0,
            Param_3 > 0,
            Param_4 > 0
            ;

      /* The optimization part */
      max logf;

      Param_xLimit   = &PARAM_xLimit;
      Param_pnLimit  = &PARAM_pnLimit;

      if THRESHOLD = 0 then do;
         lsdf = 0;
      end;
      else do;
         CdfOfThreshold = Cdf_LOGNGPD(THRESHOLD, Param_1 &LOGNGPD_Param_List_Minus_1);

         /* Handle the probability of observability */
         if &THRESHOLDPERCENT >= 0 then do;
            lsdf =  &THRESHOLDPERCENT/(1 - &THRESHOLDPERCENT) * 
                    log(CdfOfThreshold + 1E-30);
         end; 
         else do;
            lsdf = -log(1 - CdfOfThreshold + 1E-30);
         end;
      end;

      PdfOfLoss = Pdf_LOGNGPD(amount, Param_1 &LOGNGPD_Param_List_Minus_1);
        
      /* Combine all these pieces together     */
      /* to form the complete log-likelihood   */
      logf = log(PdfOfLoss + 1E-30) + lsdf;

   quit;
   ods listing;

    %if %sysfunc(exist(_cs_)) %then %do;
       data _null_;
           set _cs_;
           call symputx("status",status);
       run;

      %if &Status = 0 %then %do;
         data _parmnlp_;
            length TailBoundaryChoice $32;
            set _parmnlp_;
            where (index(_type_,"PARMS")>0);
            Param_pnLimit = &Param_pnLimit;

            if Param_4 LE 0.0000000001 then do;
               Param_4 = 0; 
            end;


            Param_xLimit  = &Param_xLimit/exp(Param_1);
         run;
      %end;
      %else %do;
         %let status = 1;
         %let reason = %sysfunc(sasmsg(&orvdds, log_opt_problem_no_fit_error, noquote));
      %end;
    %end;
    %else %do;
        %let status = 1;
        %let reason = %sysfunc(sasmsg(&orvdds, log_body_fit_fail_error, noquote));
    %end;
%LOCAL_DONE:
%mend FitLOGNGPD;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriLOGNGPD;
   /* This piece of code increases Param_pnLimit if the probability of observability is provided   */
   /* New_pnLimit = Old_pnLimit + (1-Old_pnLimit)*p0 where p0 is the probability of being hidden   */
   %if %sysevalf(&THRESHOLDPERCENT >= 0, BOOLEAN) %then %do;
      %let Param_pnLimit = %sysevalf(&Param_pnLimit + (1-&Param_pnLimit)* &THRESHOLDPERCENT); 
   %end;

   ods listing close;
   ods output 
               IterStop           =  _logl_
               ConvergenceStatus  =  _cs_ 
               covmat             =  _covmat_
               ;   

   proc nlp 
            data   = _LogWorkdata_KRI_(drop= PARAM_1:) 
            outest = _parmnlp_ 
            &PROC_NLP_OPTIONS.
            ;

      /* The list of parameters */     
      parms 
            Param_0 = &Param_1_init,
            Param_2 = &Param_2_init,
            Param_3 = &Param_3_init,
            Param_4 = &Param_4_init,
            &KRI_INIT_PARM
            ;

      /* The bounds on parameters */      
      bounds 
            Param_2 > 0,
            Param_3 > 0,
            Param_4 > 0
            ;


      /* The optimization  */
      max logf;
      Param_1 = Safe_Id(Param_0 + &KRI_MODEL_YES_INTERCEPT); 
      Param_xLimit  = &PARAM_xLimit * Exp(Param_1);
      Param_pnLimit = &PARAM_pnLimit;

      if THRESHOLD = 0 then do;
         lsdf = 0;
      end;
      else do;
         CdfOfThreshold = Cdf_LOGNGPD(THRESHOLD, Param_1 &LOGNGPD_Param_List_Minus_1);
         /* Handle the probability of observability */
         if &THRESHOLDPERCENT >= 0 then do;
            lsdf =  &THRESHOLDPERCENT/(1 - &THRESHOLDPERCENT) * 
                    log(CdfOfThreshold + 1E-30);
         end; 
         else do;
            lsdf = -log(1 - CdfOfThreshold + 1E-30);
         end;
      end;
   
      PdfOfLoss = Pdf_LOGNGPD(amount, Param_1 &LOGNGPD_Param_List_Minus_1);
      logf = log(PdfOfLoss + 1E-30) + lsdf;
    quit;
    ods listing;

   %if %sysfunc(exist(_cs_)) %then %do;
      data _null_;
         set _cs_;
         call symputx("status",status);
      run;
      
      %if &status = 0 %then %do;
         data _parmnlp_;
            length TailBoundaryChoice $32;
            set _parmnlp_ ;
            where (index(_type_,"PARMS")>0);
            Param_xLimit  = &Param_xLimit;
            Param_pnLimit = &Param_pnLimit;
         run;
      %end;
      %else %do;
         %let status = 1;
         %let reason = %sysfunc(sasmsg(&orvdds, log_opt_problem_no_fit_error, noquote));
      %end;
   %end;
   %else %do;
      %let status = 1;
      %let reason = %sysfunc(sasmsg(&orvdds, mixd_fit_fail_error, noquote));
   %end;
%LOCAL_DONE:
%mend FitKriLOGNGPD;

/*- actually, the moments of logarithm of losses -*/
%macro MomentsLOGNGPD;

   _lm_t_Param_3 = Param_3 * exp(Param_1);
   _lm_tmp_xLimit = Param_xLimit * exp(Param_1);
   %VCDFLOGN( _lm_tmp_xLimit, Param_pLimit)
   
   
    /* Part of the moment due to the body distribution     */
    /* Need to use limited expectation of lognormal        */
    /* See formula in "Loss Models" by Klugman             */
    /* on page 583 at the top                              */         
   %do i=1 %to 4;
      m&i = exp(&i*Param_1+&i*&i*Param_2*Param_2/2);
      m&i = m&i * cdf("NORMAL", (log(_lm_tmp_xLimit ) - Param_1 - &i * Param_2 * Param_2)/Param_2);
   %end;

   /* This piece is so that the contribution of the body */
   /*  distribution has the correct weight,               */
    /*  i.e Param_pnLimit=empirical probability            */
    %do i=1 %to 4;
        m&i = m&i * Param_pnLimit/Param_pLimit;
    %end;


   /* To each of the moments, add the part due to the GPD tail */
   if Param_4 EQ 0 then do;
      _lm_mGPD1 = _lm_t_Param_3 ;
   end;
   else if 1 < 1/Param_4 then do;
        _lm_mGPD1 = (_lm_t_Param_3 / Param_4) / (1/Param_4 - 1);
   end;
   else do;
        m1 = .;
   end;

   m1    = m1 + (1 - Param_pnLimit)*(_lm_mGPD1 + _lm_tmp_xLimit);

   if Param_4 EQ 0 then do;
      _lm_mGPD2 = 2 * _lm_t_Param_3 * _lm_t_Param_3;
   end;
   else if 2 < 1/Param_4 then do;
        _lm_mGPD2 = ((_lm_t_Param_3 / Param_4)**2)*2 / ((1/Param_4 - 1) * (1/Param_4 - 2));
   end;
   else do;
        m2 = .;
   end;

   m2    = m2 + (1 - Param_pnLimit) * (_lm_mGPD2 + 2 * _lm_tmp_xLimit * _lm_mGPD1 + _lm_tmp_xLimit**2 );

   if Param_4 EQ 0 then do;
      _lm_mGPD3 = 6 * _lm_t_Param_3 * _lm_t_Param_3 * _lm_t_Param_3;
   end;
   else if 3 < 1/Param_4 then do;
        _lm_mGPD3 = ((_lm_t_Param_3 / Param_4)**3)*6/((1/Param_4 - 1) * (1/Param_4 - 2) * (1/Param_4 - 3));
   end;
   else do;
        m3 = .;
   end;

   m3    = m3 + (1 - Param_pnLimit)*(_lm_mGPD3 + 3 * _lm_tmp_xLimit * _lm_mGPD2 + 3 * _lm_tmp_xLimit**2 * _lm_mGPD1  + _lm_tmp_xLimit**3 );


   if Param_4 EQ 0 then do;
      _lm_mGPD4 = 24 * _lm_t_Param_3 * _lm_t_Param_3 * _lm_t_Param_3 * _lm_t_Param_3;
   end;
   else if 4 < 1/Param_4 then do;
        _lm_mGPD4 = ((_lm_t_Param_3/Param_4)**4)*24 / ((1/Param_4 - 1) * (1/Param_4 - 2) * (1/Param_4 - 3) * (1/Param_4 - 4));
   end;
   else do;
        m4 = .;
   end;

   m4    = m4 + (1 - Param_pnLimit) * (_lm_mGPD4 + 4 * _lm_tmp_xLimit * _lm_mGPD3 + 6 * _lm_tmp_xLimit**2 * _lm_mGPD2 + 4 * _lm_tmp_xLimit**3 * _lm_mGPD1  + _lm_tmp_xLimit**4 );


   mean     = m1;
   std      = sqrt(m2-m1*m1);
   skewness = (m3-3*m1*m2+2*(m1**3))/(std**3);
   kurtosis = (m4-4*m1*m3+6*(m1**2)*m2-3*(m1**4))/(std**4);
%mend MomentsLOGNGPD;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the Param_1, Param_2, Param_4, and Param_3 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitLOGNGPD;
   /*- expect  the mean, variance, and kurtosis of logarithm of losses -*/    
   data _null_;
      Param_2=sqrt(&variance);
      call symputx("Param_1", put(&mean, BEST20.));
      call symputx("Param_2", put(Param_2, BEST20.));
   run;
%mend MomFitLOGNGPD;

%macro Param_BoundsLOGNGPD();

%mend;


%macro ValidateParamsLOGNGPD(Valid, Reason);
   if (Param_2 LE 0) OR (Param_3 LE 0) then do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "log_param_2_or_3_not_pos_error", "quote") ;
   end;
   else if (Param_4 < 0) then do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "log_param_4_neg_error", "quote") ;
   end;
   else if (Param_xLimit LE 0) then do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds","log_xlimit_param_not_pos_error", "quote") ;
   end;
   else if  (Param_pnLimit LE 0) OR (Param_pnLimit GE 1)  then do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds","log_pnlimit_param_not_0_1_error", "quote") ;
   end;
   else do;
      &Valid  = "true";
   end;
%mend ValidateParamsLOGNGPD;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_LOGNGPD(minimalInit = true); 
   %orvd_logn(minimalInit = &minimalInit);
   proc sql noprint; 
      drop table LOGN_status; 
   quit;

   %let NEEDED_INPUT         = TailBoundaryChoice TailBoundaryValue TailBoundaryPercent;
   %let SCALE_TRANSFORM      = LOG;
   %let DIST_NAME            = Log Normal-GPD;
   %let PARAM_COUNT          = 4;
   %let ProbMeasure          = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp 
               outlib = &ORVAR_CMPLIB..LOGNGPD 
               inlib  = &ORVAR_CMPLIB.;
         function Cdf_LOGNGPD(x,Param_1 &LOGNGPD_Param_List_Minus_1);
            Param_pLimit = Cdf_LOGN(Param_xLimit, Param_1 &LOGN_Param_List_Minus_1) ;    

            if x LE Param_xLimit then do;
               return( Cdf_LOGN(x,Param_1 &LOGN_Param_List_Minus_1) * Param_pnLimit/Param_pLimit);    
            end;
            else do;
               if Param_4 LE 0.00000000001 then do;
                  return( Param_pnLimit + (1-Param_pnLimit)*(1-exp(-(x-Param_xLimit)/Param_3/exp(Param_1))));
               end;
               else do;
                  return( Param_pnLimit + (1-Param_pnLimit)*(1-(1 + Param_4*(x-Param_xLimit)/Param_3/exp(Param_1))**(-1/Param_4)));
               end;
            end;
         endsub;

         function Pdf_LOGNGPD(x,Param_1 &LOGNGPD_Param_List_Minus_1);
            Param_pLimit = Cdf_LOGN(Param_xLimit, Param_1 &LOGN_Param_List_Minus_1) ;    

            if x LE Param_xLimit then do;
               return( Pdf_LOGN(x, Param_1, Param_2) * Param_pnLimit / Param_pLimit);    
            end;
            else do;
               if Param_4 LE 0.00000000001 then do;
                  return((1-Param_pnLimit) * exp(-(x-Param_xLimit)/Param_3/exp(Param_1))/Param_3/exp(Param_1));
               end;
               else do;
                  return( (1 - Param_pnLimit) * ((1 + (Param_4) * (x-Param_xLimit)/Param_3/exp(Param_1))**(-1/Param_4 -1))/Param_3/exp(Param_1));
               end;
            end;
         endsub;
      run;
   %end;
%DONE:
   data LOGNGPD_STATUS;
      status = 0;
   run;
%mend ORVD_LOGNGPD;
