/********************************************************************
 * Copyright (C), 2011
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_LOGNMIX.sas
 *   
 *  This file contains macro fragments to implement calculation specific to the 
 *  orvd_LOGNMIX distribution
 *
 *  The macro %ORVD_orvd_LOGNMIX is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_orvd_LOGNMIX macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragments then become defined and available for other 
 *  macros to call.
 *
 ********************************************************************/


/***********************************************************************************//**
 *   CDFLOGNMIX
 *
 *   What it does  :  Evaluate the cumulative distribution function (CDF).  
 *                    Parameters are assumed to exist as macro variables when called.
 *
 *   Parameters    :
 *        x         = value at which the CDF is evaluated
 *        returnVal = name of dataset variable into which result is stored
 ***********************************************************************************/
%macro CDFLOGNMIX(x, returnval);
    &returnval = (1 - &Param_3) * cdf("LOGNORMAL", &x, &Param_1, &Param_2) 
                 + &Param_3 * cdf("LOGNORMAL", &x, &Param_4, &Param_5);
%mend CDFLOGNMIX;

/***********************************************************************************//**
 *   VCDFLOGNMIX
 *
 *   What it does  :  Evaluate the cumulative distribution function (CDF).  
 *                    Parameters are assumed to exist as dataset variables when called. 
 *                    (in data step or in fcmp code)
 *
 *   Parameters    :
 *        x         = value at which the CDF is evaluated
 *        returnVal = name of variable into which result is stored
 ***********************************************************************************/
%macro VCDFLOGNMIX(x, returnVal);
    &returnval = (1 - Param_3) * cdf("LOGNORMAL", &x, Param_1, Param_2) 
                 + Param_3 * cdf("LOGNORMAL", &x, Param_4, Param_5);
%mend VCDFLOGNMIX;

/***********************************************************************************//**
 *   PDFLOGNMIX
 *
 *   What it does  :  Evaluate the probability density function (PDF).  
 *                    Parameters are assumed to exist as macro variables when called.
 *
 *   Parameters    :
 *        x         = value at which the CDF is evaluated
 *        returnVal = name of variable into which result is stored
 ***********************************************************************************/
%macro PDFLOGNMIX(x, returnval);
    &returnval = (1 - &Param_3) * pdf("LOGNORMAL", &x, &Param_1, &Param_2) 
                 + &Param_3 * pdf("LOGNORMAL", &x, &Param_4, &Param_5);
%mend PDFLOGNMIX;

/***********************************************************************************//**
 *   VCDFLOGNMIX
 *
 *   What it does  :  Evaluate the probability density function (PDF).  
 *                    Pametersb are assumed to exist as variables when called 
 *                    (in data step or in fcmp code).
 *
 *   Parameters    :
 *        x         = value at which the CDF is evaluated
 *        returnVal = name of variable into which result is stored
 ***********************************************************************************/
%macro VPDFLOGNMIX(x, returnVal);
    &returnval = (1 - Param_3) * pdf("LOGNORMAL", &x, Param_1, Param_2) 
                 + Param_3 * pdf("LOGNORMAL", &x, Param_4, Param_5);
%mend VPDFLOGNMIX;

/***********************************************************************************//**
 *   QUANTLOGNMIX
 *
 *   What it does  :  Assumes a dataset named _temp_  in the WORK library exists
 *                    and contains ecdf values. It computes the inverse cdf at each 
 *                    point.  For use by orva_diagplot in producing QQ plots.
 *
 *   Parameters    :
 *        scalefactor = Scale factor used in mixture distributions.  
 *                      The default value is 1.
 ***********************************************************************************/
%macro QuantLOGNMIX(scaleFactor = 1);
   /*   Invert the first LogNormal component */
   %QuantLogn(scaleFactor = &scaleFactor );

   data _temp_;
      set 
         _temp_
         ;
      rename 
         quant = quant1
         ;
      drop 
         cdf 
         truncdf 
         ;
   run;

   /*   Invert the second LogNormal component */
   %let tmp_Param_1 = &Param_1;
   %let tmp_Param_2 = &Param_2;
   %let Param_1     = &Param_4;
   %let Param_2     = &Param_5;

   %QuantLogn(scaleFactor = &scaleFactor );

   %let Param_1     = &tmp_Param_1;
   %let Param_2     = &tmp_Param_2;

   data _temp_;
      set _temp_(drop = truncdf cdf);
   run;
    
   /* Do a bisection search for inverse of mixture */
   proc fcmp 
            data =_temp_
            out  = _temp_
            ;
      retain truncdf 0;
      eps = 0.000001; 

      Param_1 = &Param_1;
      Param_2 = &Param_2;
      Param_3 = &Param_3;
      Param_4 = &Param_4;
      Param_5 = &Param_5;

      if _N_= 1 then do;
         if THRESHOLD = 0 then do;
            truncdf = 0;
         end;
         else do;
            %CDFLognMix(THRESHOLD,truncdf);
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf = ecdf * (1 - truncdf) + truncdf;

      if  cdf/(&scaleFactor) LE 1.0 then do; 
         if quant1 < quant then do;
            LB = quant1;
            UB = quant;
         end;
         else do;
            LB = quant;
            UB = quant1;
         end;

         guess = (UB + LB)/2;
         delta = abs(UB - LB);

         do while(delta > eps);
            %VcdfLognMix(guess,val);

            if val > cdf/&scaleFactor then do;
               UB = guess;
            end;
            else do;
               LB = guess;
            end;

            newGuess =(UB + LB)/2;
            delta = abs(guess - newGuess)/guess;
            guess = newGuess;
         end;

         quant = guess;
      end;
      else do;
         quant = .;
      end;

      keep ecdf cdf amount truncdf quant threshold financialdate;
   run;

%mend QuantLOGNMIX;

/*****************************************************************************************//**
 *   CdfDerLOGNMIX
 *
 *   What it does  :  Calculates derivatives of the CDF function with respect to its parameters.
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerLOGNMIX(scaleFactor = 1, indata =, outdata =);
   data _temp_0;
      set &indata ;

      cdf =  &scaleFactor.*( (1 - &Param_3) * cdf("LOGNORMAL", quant, &Param_1, &Param_2)
            + &Param_3 * cdf("LOGNORMAL", quant, &Param_4, &Param_5) );
      cdf1 = cdf;
      cdf2 = cdf;
      cdf3 = cdf;
      cdf4 = cdf;
   run;

   proc model data=_temp_0 noprint;
      dependent cdf cdf1 cdf2 cdf3 cdf4;
      retain 
               trunquant 0 
               Param_1 &Param_1 
               Param_2 &Param_2 
               Param_3 &Param_3 
               Param_4 &Param_4 
               Param_5 &Param_5
               ;

      cdf =  &scaleFactor.*( (1 - Param_3) * cdf("LOGNORMAL", quant, Param_1, Param_2)
                             + Param_3 * cdf("LOGNORMAL", quant, Param_4, Param_5) );
      cdf1 = pred.cdf;
      cdf2 = pred.cdf;
      cdf3 = pred.cdf;
      cdf4 = pred.cdf;

      DParam_1 = getder(pred.cdf,  Param_1);
      DParam_2 = getder(pred.cdf1, Param_2);
      DParam_3 = getder(pred.cdf2, Param_3);
      DParam_4 = getder(pred.cdf3, Param_4);
      DParam_5 = getder(pred.cdf4, Param_5);

      outvars DParam_1-DParam_5 quant;
      solve 
            Param_1 
            Param_2 
            Param_3 
            Param_4 
            Param_5
            /out = &outdata(keep =  quant 
                                    cdf 
                                    DParam_1-DParam_5
                                    );
   quit;

%mend CdfDerLOGNMIX;

/***********************************************************************************//**
 *   SimLOGNMIX
 *
 *   What it does  :  Simulates a loss 
 *
 *   Parameters    :
 *        loss = name of variable in which to store the simulated loss value.
 ***********************************************************************************/
%macro SimLOGNMIX(loss);
   u = rand('uniform');

   if (u < (1 - Param_3)) then do;
      &loss = exp(rand("NORMAL") * Param_2 + Param_1);
   end;
   else do;
      &loss = exp(rand("NORMAL") * Param_5 + Param_4);
   end;
%mend SimLOGNMIX;

/***********************************************************************************//**
 *   InitFitLOGNMIX
 *
 *   What it does  :  Initializes parameters for fitting
 *
 *   Parameters    :
 *        summaryData = dataset holding the loss data information
 *        KRIFit      = whether or not the fit will be performed with KRIs 
 *                      Possible values: NO or YES.
 ***********************************************************************************/
%macro InitFitLOGNMIX(summaryData = , KRI_FIT = NO);
   %InitFitLOGN(summaryData = &summaryData , KRI_FIT = &KRI_FIT);
   %let tmp_obscount = &obscount;
   %FitLOGN;

   data _NULL_;
      set _parmnlp_;
      where _type_ = "PARMS";
      call symputX("tmp_Param_1", Param_1);
      call symputX("tmp_Param_2", Param_2);
   run;

   proc sort data = _logworkdata_;
      by Amount;
   run;

   /******************************************************************************/
   /* We build the empirical cdf.                                                */
   /* We subtract F out of it                                                    */
   /******************************************************************************/
   data _temp_data;
      retain 
               Fn_0            0 
               Param_1         &tmp_Param_1 
               Param_2         &tmp_Param_2 
               lastAMount      0 
               lastLogAMount   0 
               sample_mean     0 
               sample_log_mean 0
               ;

      set _logworkdata_ end = last;
      Fn_0 = Fn_0 + 1 / &obsCount;
      %VCDFLOGN(Amount, FLognMix);
      diff_Fs = abs(Fn_0 - FLognMix);
   run;

   proc sort 
            data = _temp_data 
            out  = sort_temp_data
            ;
      by descending diff_fs ;
   run;

   data keep_temp_data;
      set sort_temp_data end = last;

      if _N_ > 0.10 * &obscount AND _N_ > 3 then do;
         delete;
      end;

      keep amount ; 
   run;

   proc summary 
               data   = keep_temp_data
               vardef = n 
               noprint
               ;
      var AMOUNT;
      output 
         out = _stat_2(keep =  obsCount 
                              sample_mean 
                              sample_uss 
                              )
         n        =  obsCount 
         mean     =  sample_mean 
         uss      =  sample_uss 
   run;
   
   %InitFitLOGN(summaryData =_STAT_2 , KRI_FIT = NO );
   %let Param_4_init = &Param_1_init;
   %let Param_5_init = &Param_2_init;
   %let Param_1_init = &tmp_Param_1;
   %let Param_2_init = &tmp_Param_2;

   %let Param_3_init    =  0.5;
   %let nparms = 5;    
   %let Obscount = &tmp_Obscount;

   %if &dist NE MIXDIST %then %do;  
      %put ******************************;
      %put Distribution   =  &dist;
      %put Observations   =  &obsCount;
      %put Param_1_init    =  &Param_1_init;
      %put Param_2_init    =  &Param_2_init;
      %put Param_3_init    =  &Param_3_init;
      %put Param_4_init    =  &Param_4_init;
      %put Param_5_init    =  &Param_5_init;
      %put ******************************;
   %end;

   proc datasets library = work memtype = data nolist;
      delete 
            _temp_data 
            sort_temp_data
            keep_temp_data
            _stat_2
            ;
   quit;
%mend InitFitLOGNMIX;


/***********************************************************************************//**
 *   FitLOGNMIX
 *
 *   What it does  :  Fit the model without KRIs
 *
 ***********************************************************************************/
%macro FitLOGNMIX;
   ods listing close;
   ods output 
            IterStop           =  _logl_
            ConvergenceStatus  =  _cs_ 
            covmat             =  _covmat_ 
            ;  
   proc nlp 
            data   = _logworkdata_ 
            outest =  _parmnlp_ 
            &PROC_NLP_OPTIONS. 
            ;
 
      parms 
            Param_1 = &PARAM_1_init, 
            Param_2 = &PARAM_2_init, 
            Param_3 = &PARAM_3_init, 
            Param_4 = &PARAM_4_init,
            Param_5 = &PARAM_5_init
            ;

      bounds Param_1 > 0  %Param_BoundsLOGNMIX(); 
      max logf;
     
      if threshold > 0 then do;
         %VcdfLOGNMIX(THRESHOLD, cdf_temp);
      end;
      else do;
         cdf_temp = 0;
      end;
      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT) *
                  log( cdf_temp + 1E-30);     
      end; 
      else if THRESHOLD = 0 then do;
         lsdf = 0;
      end;
      else do;
         lsdf = log(1 - cdf_temp + 1E-30);
      end;

      %VpdfLOGNMIX(AMOUNT,pdf_temp);
      logf =   log(pdf_temp + 1e-30) - lsdf;
   quit;
   ods listing;
%mend;


/***********************************************************************************//**
 *   FitKriLOGNMIX
 *
 *   What it does  :  Performs the fitting with KRIs
 *   
 *   This has not been worked on, it means nothing as it is.
 *
 ***********************************************************************************/
%macro FitKriLOGNMIX;
   ods listing close;
   ods output 
               IterStop           =  _logl_
               ConvergenceStatus  =  _cs_ 
               covmat             =  _covmat_ 
               ;  

    proc nlp 
            data   =  _LogWorkdata_KRI_ (drop = PARAM_1:)  
            outest =  _parmnlp_ 
            &PROC_NLP_OPTIONS.
            ;

         parms 
               Param_0 = &Param_1_init, 
               Param_2 = &Param_2_init,
               Param_3 = &Param_3_init,
               Param_4 = &Param_4_init,
               Param_5 = &Param_5_init,
               &KRI_INIT_PARM
               ;

      bounds Param_0 > 0 %Param_BoundsLOGNMIX(); 
      max logf;
      LogParam_1 = Safe_ID(log(Param_0) + &KRI_MODEL_YES_INTERCEPT );
      Param_1    = exp(LogParam_1);

      if threshold > 0 then do;
         %VcdfLOGNMIX(THRESHOLD, cdf_temp);
      end;
      else
         cdf_temp = 0;
      
      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
                  log( cdf_temp + 1E-30);     
      end; 
      else if THRESHOLD=0 then do;
         lsdf=0;
      end;
      else do;
         lsdf = log(1 - cdf_temp + 1E-30);
      end;

      %VpdfLOGNMIX(AMOUNT,pdf_temp);
      logf =   log(pdf_temp+ 1e-30) - lsdf;

   quit;
   ods listing;
%mend FitKriLOGNMIX;

/***********************************************************************************//**
 *   MomentsLOGNMIX;
 *
 *   What it does  :  Computes moments
 *   
 *   Needs to be called inside a data step where the parameter variables 
 *   Param_1-Param_5 are defined.
 *
 ***********************************************************************************/
%macro MomentsLOGNMIX;
   %do i=1 %to 4;
      m&i = (1 - Param_3) * exp(&i*param_1 + &i*&i * param_2*param_2 /2) + 
                 Param_3  * exp(&i*param_4 + &i*&i * param_5*param_5 /2);
   %end;

   mean     = m1;
   std      = sqrt(m2 -  mean**2);
   skewness = ( m3 - 3 * mean * m2 + 2 * mean**3 ) / (std**3);
   kurtosis = (m4 - 4 * mean * m3 + 6 * mean**2 * m2 - 3 * mean**4 ) / (std**4); 
%mend MomentsLOGNMIX;

/***********************************************************************************//**
 *   LimMomentsLOGNMIX
 *
 *   What it does  :  Computes the limited moments
 *
 *   Parameters    :
 *        i         = the moment to compute (i=1 is the limited mean, for example)
 *        x         = the limiting value
 *        returnVal = variable that holds the limited moment
 ***********************************************************************************/
%macro LimMomentsLOGNMIX(i, x, returnVal);
   &returnVal = (1 - Param_3) * 
                  (
                     exp(&i * Param_1 + &i * &i * Param_2 * Param_2/2) * 
                     cdf("NORMAL", (log(&x) - Param_1 - &i * Param_2 * Param_2)/Param_2)
                     + 
                     &x**&i * (1 - cdf("NORMAL", (log(&x) - Param_1)/Param_2))
                  ) 
                + 
                param_3 * 
                  (
                     exp(&i * Param_4 + &i * &i * Param_5 * Param_5/2) * 
                     cdf("NORMAL", (log(&x) - Param_4 - &i * Param_5 * Param_5)/Param_5)
                     + 
                     &x**&i * (1 - cdf("NORMAL", (log(&x) - Param_4)/Param_5))
                  ); 
%mend LimMomentsLOGNMIX;

%macro MomFitLOGNMIX;
   /*   Not Used   */
%mend MomFitLOGNMIX;

/***********************************************************************************//**
 *   ParamBoundsLOGNMIX
 *
 *   What it does  :  Defines limits on parameter values in fitting
 *
 ***********************************************************************************/
%macro Param_BoundsLOGNMIX();
   ,Param_2 > 1e-6, Param_3 > 0, Param_3 < 1, Param_5 > 1e-6 
%mend Param_BoundsLOGNMIX;

/***********************************************************************************//**
 *   ValidateParamsLOGNMIX
 *
 *   What it does  :  Dertermines if parameters are valid
 *
 ***********************************************************************************/
%macro ValidateParamsLOGNMIX(Valid, Reason);
   &Valid  = "false";

   if (Param_2 <= 0) then do;
      &Reason = sasmsg("&orvdds", "lognmix_param2_not_pos_error", "quote") ;     
   end;
   else if (Param_5 <= 0) then do;
      &Reason = sasmsg("&orvdds", "lognmix_param5_not_pos_error", "quote") ;     
   end;
   else if (Param_3 < 0) OR (Param_3 > 1) then do;
      &Reason = sasmsg("&orvdds", "lognmix_param3_not_0_1_error", "quote") ;     
   end;
   else do;
      &Valid  = "true";
      &Reason = "OK";
   end;
%mend ValidateParamsLOGNMIX;

%macro ORVD_LOGNMIX(minimalInit = true);
   %let NEEDED_INPUT         = ;
   %let SCALE_TRANSFORM      = LOG;
   %let DIST_NAME            = LOGNMIX;
   %let PARAM_COUNT          = 5;
   %let CDF_TRANSFORM_FACTOR = 1;
   %let ProbMeasure          = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..LOGNMIX;
         function Cdf_LOGNMIX(x,Param_1 &LOGNMIX_Param_List_Minus_1);
            return( (1 - param_3) * 
                     cdf("LOGNORMAL", x, Param_1 , Param_2) + 
                     param_3 * 
                     cdf("LOGNORMAL", x, Param_4 , Param_5)
                     );
         endsub;

         function Pdf_LOGNMIX(x,Param_1 &LOGNMIX_Param_List_Minus_1);
            return( (1 - Param_3) * 
                     pdf("LOGNORMAL", x, Param_1 , Param_2) + 
                     Param_3 * 
                     pdf("LOGNORMAL", x, Param_4 , Param_5)
                     );
         endsub;
      run;
   %end;
%DONE:

   data LOGNMIX_STATUS;
      status          = 0;
   run;  
%mend ORVD_LOGNMIX;
