/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_NEGBIN.sas
 *   
 *   -Provides support for the NegativeBinomial distribution-
 *
 *  This file contains macro fragments to implement calculation specific to the 
 *  NegativeBinomial distribution
 *
 *  The macro %ORVD_NEGBIN is at the end of this file. It is an empty macro that takes
 *  no paramter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_NEGBIN macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ******************************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfNEGBIN(x, returnVal);
      if &FParam_1 < 1E-8 then do;
         &returnVal = 1;
      end;
      else do;
         &returnVal = cdf("NEGBIN", &x, 1/(1 + &FParam_2), &FParam_1);  
      end; 
%mend CdfNEGBIN;

%macro VCdfNEGBIN(x,returnVal);
   if FParam_1 < 1E-8 then do;
      &returnVal = 1;
   end;
   else do;
      &returnVal=cdf("NEGBIN", &x, 1/(1 + FParam_2), FParam_1);   
   end;
%mend VCdfNEGBIN;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfNEGBIN(x,returnVal);
      if &FParam_1 < 1E-8 then do;
         if &x = 0 then do;
            &returnVal = 1;
         end;
         else do;
            &returnVal = 0;
         end;
      end;
      else do;
         &returnVal=pdf("NEGBIN",&x,1/(1+&FParam_2),&FParam_1);   
      end
%mend PdfNEGBIN;

%macro VPdfNEGBIN(x, returnVal);
      if FParam_1 < 1E-8 then do;
         if &x = 0 then do;
            &returnVal = 1;
         end;
         else do;
            &returnVal = 0;
         end;
      end;
      else do;
         &returnVal=pdf("NEGBIN", &x, 1/(1 + FParam_2), FParam_1);   
      end;
%mend VPdfNEGBIN;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantNEGBIN;
   data _temp_;
      cdf   = _N_/(nobs+1);
      if &FParam_1 < 1E-8 then do;
         quant = 0;
      end;
      else do;
         quant = quantile("NEGBINOMIAL",cdf, 1/(1+ &FParam_2), &FParam_1);
      end;
   run;
%mend QuantNEGBIN;

%macro VQuantNEGBIN(x,returnVal);
   if FParam_1 < 1E-8 then do;
      &returnVal = 0;
   end;
   else do;
      &returnVal = quantile("NEGBINOMIAL", &x, 1/(1+ FParam_2), FParam_1);
   end;
%mend VQuantNEGBIN;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimNEGBIN(loss);
   if  fparam_1 GT 1E-8 then do;
      if ( (1/(1 + FParam_2)) GT .9999) then do;
         &loss=rand("POIS",FParam_1 * (1- (1/(1 + FParam_2)))*(1 + FParam_2));
      end;
      else do;
         &loss=rand("NEGBINOMIAL",1/(1+FParam_2),FParam_1);
      end;
   end;
   else do;
      &loss=rand("BINOMIAL",fparam_1*fparam_2/20,20);
   end;
%mend SimNEGBIN;

/*****************************************************************************************
 *   Any intialization or other pre-fit calculations go here.  For example, parameter guesses
 *****************************************************************************************/
%macro InitFitNEGBIN(countdata=);
   /* Not currently used */  
%mend InitFitNEGBIN;
    
/*****************************************************************************************
 *   This macro does the actual fitting.  The count data is passed in, with counts, KRI values
 *   offest (log of net exposure). Selected KRIs are available as macro variables.  
 *****************************************************************************************/
%macro FitNEGBIN(countdata =, outparm =, outcov =);

   %local reason status FP2 int;

   %let methods     = QUANEW NEWRAP TRUREG ; 
   %let try         = 1;
   %let curr_method = %scan(&methods., &try., ' ');

%TRYAGAIN:
   /* perform fitting with proc countreg */
   ods listing close;
   ods output FitSummary = _FitSumamry ;
   ods output ParameterEstimates = _ParameterEstimates_;
   proc countreg 
         data = &CountData.
         outest = count_out_est 
         covout 
         GCONV=1E-14 
         FCONV=1E-20 
         ABSFCONV=0 
         ABSGCONV=1E-10 
         method = &curr_method. 
         type = 
         %if &sysver. < 9.2 %then %do;
            NEGBIN1
         %end;
         %else %do;
            negbin(p=1)
         %end; 
         ;
      model 
         the_count = &SelectedKRIs. 
                     &SelectedDummies.
                     / offset=offset;
   run;
   ods output close;
   ods listing;

   %let status = -1;
   data _null_;
      set count_out_est;
      if _N_ = 1 then do;
         call symput("status", substr(_STATUS_,1,2));
         call symput("reason", substr(_STATUS_,3));
      end;
   run;

   %if &status. NE 0  %then %do;
      %let try = %eval(&try.+1);
      %if &try. LT 4 %then %do; 
         %let prev_method = &curr_method.;
         %let curr_method = %scan(&methods., &try., ' ');
         %put &ORVAR_WARNING: %sysfunc(sasmsg(&orvads, negbin_method_failed, noquote, &prev_method.));
         %put %sysfunc(sasmsg(&orvads, negbin_trying_new_method, noquote, &curr_method));
         %goto TRYAGAIN;
      %end;
      %else %do;
         %let reason = %sysfunc(sasmsg(&orvads, negbin_all_methods_failed, noquote));
         %orva_status_ds( statusds=&outparm., status=1, reason=&reason );
         %return; 
      %end;
   %end;
   %else %do; /* BEGIN: Succesful fit */
      /* WALD test */
      data _NULL_;
         set _ParameterEstimates_;
         if Parameter = "_Alpha" then do;
            if (probt GT 0.10) or (missing(probt)) then do;
               %let status = 2; 
               %let reason = %sysfunc(sasmsg(&orvads, fkrid_wald_test_error, noquote));
            end;
         end;
      run;

      /* Need variable names from the fitsummary dataset to be uniform across loaclizations, 
         so use english names for the variables */
      %orva_unlocalize_countreg_fitsum( summaryData = _FitSumamry );

      /* get the covariance matrix and parameter estimation data */
         data 
            &outcov.(drop = _TYPE_ _STATUS_)
            &outparm.(drop= _NAME_ _TYPE_);
               length _NAME_ $32;
               set count_out_est;
               /* Alpha is Fparam2 */
               label _Alpha="FParam_2"; 
               rename _Alpha=FParam_2;
               if _TYPE_ = "COV" then do;
                  if upcase(substr(_NAME_,1,6)) EQ '_ALPHA' then do;
                     _NAME_ = 'FParam_2';
                  end;
                  output  &outcov.;
               end;
               else if _N_ = 1 then do;
                  output &outparm.;
               end;       
         run;

         data  &outparm.;
            format reason &ORVAR_REASON_FMT;
            status = &status;
            reason = "&reason";
            set _FitSumamry;
            set &outparm.;
         run;

   %end; /* END: Succesful fit */

%mend FitNEGBIN;

/*****************************************************************************************
 *   What it does  :  Adjusts parameters and covariance matrix so that we deal
 *                    with correct parameterization.
 *
 *   Parameters:
 *       parm   - Dataset to modify with the estimated parameters
 *       cov    - Parameter covariance matrix to modify
 *      
 *****************************************************************************************/
%macro PostFitNEGBIN( parm =, cov = );

/************************************************************************
 * Countreg uses a different paramterization than we do.  In countreg, 
 * parameters are mu and alpha, with
 *     mu    = n*beta
 *     alpha = beta
 * countreg regresses the KRIs on the mean, mu.  What we want is the KRIs 
 * regressed on n.  To get this, we will need to adjust the intercept that 
 * countreg has estimated by subtracting log(beta).  n is mu/alpha.  
 * The covariance matrix then needs to be adjusted as well. 
 **************************************************************************/

   %local
      int
      FP2
      varnames_quote
      ;

   /* Adjust the parameter estimation output so that the parameterization is correct. Add information from fitsummary dataset */
   data  &parm.;
      set &parm.;

      /* In countreg, have mu = exp(intercept). Want n = mu/alpha */
      /* For KRI fits, the intercept will also change due to standardization.  This is 
         handled in freq_fit_dist in the unstandardize macro.  Adjustments here are for parmaeterization change only */
      FParam_1 = exp(min(intercept,700))/Fparam_2;
      call symputX("int",put(intercept,BEST20.));
      call symputX("FP2",put(FParam_2,BEST20.));
      /* Going to keep intercept for numerical stability, want intercept
         corresponding to our parmameterization */
      intercept = intercept - log(FParam_2);
   run;

   /* Because the variance of the intercept is multiplied by exp(intercept) twice,
      need to cap this value to avoid overflow */
   %let int = %sysfunc(min(&int., 70));

   /* Now adjust the covariance matrix */
   /* Countreg estimates an intercept, but we use FParam_1, which is exp(intercept)/FParam_2 */
   data &cov.;
      set &cov.;
      label intercept = "FParam_1";
      if trim(lowcase(_NAME_)) = "intercept" then do;
         _NAME_ = "FParam_1";
      end;
      rename intercept = FParam_1;
   run;      

   %let varnames_quote = ;
   data _null_;
      set _ffit_covmat_&ndist.;
      if _N_ = 1;
      length 
         varnames_quote $ 32000
         ;
      array vars _NUMERIC_;
      do i = 1 to dim(vars);
         varnames_quote = strip(varnames_quote) || ' "' || vname(vars[i]) || '"';
      end;
      call symput("varnames_quote",varnames_quote);
   run;

   /* Get Hessian Matrix */
   data _negfit_hessian;
      set &cov.;
      retain  ind_KRI 1 ind_DUMM 1;
      array KRIS KRI:;
      array DUMM Dumm:;
      if trim(lowcase(_NAME_)) = "fparam_1" then do;
         FParam_1 = exp(&int.)/&FP2.;
         FParam_2 =  -exp(&int.)/&FP2./&FP2.;
         do i=1 to dim(KRIS);
            KRIS[i] = 0;
         end;
         do i=1 to dim(DUMM);
            DUMM[i] = 0;
         end;
      end;
      else if (trim(lowcase(_NAME_)) eq "fparam_2") then do;
         FParam_1 = 0;
         FParam_2 = 1;
         do i=1 to dim(KRIS);
            KRIS[i] = 0;
         end;
         do i=1 to dim(DUMM);
            DUMM[i] = 0;
         end;
      end;
      else if lowcase(substr(_NAME_,1,3)) eq "kri" then do;
         FParam_1 = 0;
         FParam_2 = 0;
         do i=1 to dim(KRIS);
            KRIS[i] = 0;
         end;
         do i=1 to dim(DUMM);
            DUMM[i] = 0;
         end;
         KRIS[ind_KRI] = 1;
         ind_KRI = ind_KRI + 1;
      end;
      else if lowcase(substr(_NAME_,1,3)) eq "dum" then do;
         FParam_1 = 0;
         FParam_2 = 0;
         do i=1 to dim(KRIS);
            KRIS[i] = 0;
         end;
         do i=1 to dim(DUMM);
            DUMM[i] = 0;
         end;
         DUMM[ind_DUMM] = 1;
         ind_DUMM = ind_DUMM + 1;
      end;
      drop i ind_DUMM ind_KRI;
   run;

   /* Multiply Hessian * Covariance * hessian transposed */
   proc iml;
      _NAME_={ &varnames_quote.};
      use &cov.;
      read all var _num_ into C; 
      close &cov.;
      use _negfit_hessian;
      read all var _num_ into D;
      close _negfit_hessian;
      X=D*C*(D`);
      create &cov. from X[colname=_NAME_ rowname=_NAME_];
      append from X[rowname=_NAME_];
      close &cov.;
   quit;

   %orvu_delete_data( data = _negfit_hessian );

%mend PostFitNEGBIN;
/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsNEGBIN();
    Mean  = FParam_1 * FParam_2;
    Variance= Mean * (1 + FParam_2);
%mend MomentsNEGBIN;

/*****************************************************************************************
  *   What it does:  Does method of moments estimation and assigns the corresponding 
  *                  values to the distributions parameters
  *
  *   Parameters  : assumes &mean and &variance are set
  *****************************************************************************************/
%macro MomFitNEGBIN;
    if &Mean < &Variance then do;
        FParam_2 = &Variance/&Mean - 1;
        FParam_1 = &Mean / FParam_2;
    end;
    else do;
        FParam_1 = .;
        FParam_2 = .;
        Status   = 1;
        Reason   = sasmsg("&orvdds", "negbin_mean_smaller_var_error", "quote");
    end;
%mend MomFitNEGBIN;

/*****************************************************************************************
  *   What it does:  If you write Fparam_1 as function of mean and other parmeters, this is
  *   the derivative of Fparam_1 with respect to the mean.   Used in parameter variance CI calculcation 
  *****************************************************************************************/
%macro MomDerNEGBIN(deriv=);
   &deriv = 1/FParam_2;
%mend MomDerNEGBIN;
   
/*****************************************************************************************
  *   What it does:  Generates fcmp subroutines to compute neg. binomial pgf and its derivatives
  *****************************************************************************************/
%macro pgf_funcsNEGBIN;
/*---------------------------------------------------------------*/
/*             P(z) = B(z)^(-FP1)                */
/*               Define: B(z) = 1-FP2*(z-1)                      */
/*---------------------------------------------------------------*/
   /*----------------------------------------------------------------------------------------
     -
     -   What it does:  Compute the PGF at a complex point z, given as real part (r) and 
     -                  imaginary part (m). Inputs are overwritten with the output. 
     -
     -   Parameters  : FP1, FP2- negbin parameters
     -                 r- array: real part of points at which to evaluate pgf
     -                 m- array: imaginary part of points at which to evaluate pgf
     -                 n- length of arrays
     -
     ----------------------------------------------------------------------------------------*/
   subroutine pgfNEGBIN( FP1,
                         FP2,
                         r[*],
                         m[*],
                         n );
      outargs r,m;

      do i = 1 to n;
         /* Get polar and cartesian coordinates for B(z) */
         r2   = 1-FP2*(r[i]-1);  m2 = -FP2*m[i];
         mod2 = sqrt( r2**2 + m2**2 );
         if r2 = 0 then do;
            theta2 = sign(m2)*1.5707963267948;
         end;
         else do;
            theta2 = atan(m2/r2);
         end;
         /* Make sure that theta is in the correct quadrant- atan always returns 
            values in quadrants 1 and 4 */
         if r2 < 0 then do;
            theta2 = 3.1415926535897932384 + theta2;
         end;

         /* This is P(z):  */
         PWR = -1*FP1;
         r[i] = mod2**(PWR)*cos(PWR*theta2);
         m[i] = mod2**(PWR)*sin(PWR*theta2);
      end;

   endsub;

    /*----------------------------------------------------------------------------------------
     -
     -   What it does:  Compute the derivative of the PGF at a complex point z  given as 
     -                  real part (r) and imaginary part (m) with respect to the parameters FP1 (FParam_1)
     -                  and FP2 (FParam_1)
     -
     -   Parameters  : dFP1_r- array: real part of derivative wrt FP1
     -                 dFP1_m-  array: imaginary part of derivative wrt FP1
     -                 FP1- FParam_1
     -                 dFP2_r-  array: real part of derivative wrt FP2
     -                 dFP2_m-  array: imaginary part of derivative wrt FP2
     -                 FP2- FParam_2
     -                 r- array: real part of points at which to evaluate pgf derivative
     -                 m- array: real part of points at which to evaluate pgf derivative
     -                 n- length of arrays
     -
     ----------------------------------------------------------------------------------------*/
   subroutine pgfderNEGBIN( dFP1_r[*],
                            dFP1_m[*],
                            FP1,
                            dFP2_r[*],
                            dFP2_m[*],
                            FP2,
                            r[*],
                            m[*],
                            n );
   outargs dFP1_r,dFP1_m,dFP2_r,dFP2_m;

      do i=1 to n;
         /* Derivative wrt FP1 */
         /* dPdFP1 = -P(z)*log(1-FP2*(z-1)) */

         /* Get polar and cartesian coordinates for B(z) */
         r2   = 1-FP2*(r[i]-1);  m2 = -FP2*m[i];
         mod2 = sqrt( r2**2 + m2**2 );
         if r2 = 0 then do;
            theta2 = sign(m2)*1.5707963267948;
         end;
         else do;
            theta2 = atan(m2/r2);
         end;
         /* Make sure that theta is in the correct quadrant- atan always returns 
            values in quadrants 1 and 4 */
         if r2 < 0 then do;
            theta2 = 3.1415926535897932384 + theta2;
         end;

         /* This is P(z) */
         r3 = mod2**(-1*FP1)*cos(-1*FP1*theta2);
         m3 = mod2**(-1*FP1)*sin(-1*FP1*theta2);

         /* This is log(1-FP2*(z-1)) */
         r4 = log(mod2);
         m4 = theta2;

         dFP1_r[i] = -1*r3*r4 + m3*m4;
         dFP1_m[i] = -1*r3*m4 - r4*m3;

         /* Derivative wrt FP2 */
         /* FP1*[1-FP2*(z-1)]**(-FP1-1)*(z-1) */
         r5 = FP1*mod2**(-1*FP1-1)*cos((-1*FP1-1)*theta2);
         m5 = FP1*mod2**(-1*FP1-1)*sin((-1*FP1-1)*theta2);

         dFP2_r[i] = (r5*(r[i]-1) - m5*m[i]);
         dFP2_m[i] = (r5*m[i] + m5*(r[i]-1));
      end;
   endsub;

   /*----------------------------------------------------------------------------------------
     -
     -   What it does:  Compute the derivative of the PGF at a (complex) point z with respect 
     -                  to the complex argument z
     -
     -   Parameters  : FP1- mean parameter    
     -                 FP2- shape parameter
     -                 r- array: real part of points at which to evaluate derivative
     -                 m- array: imaginary part of points at which to evaluate derivative
     -                 n- length of the arrays
     -
     ----------------------------------------------------------------------------------------*/
   subroutine pgfCompDerNEGBIN( FP1,
                                FP2,
                                r[*],
                                m[*],
                                n  );
      outargs r,m;
      do i=1 to n;
         /* Derivative wrt z */
         /* FP1*FP2*[1-FP2*(z-1)]**(-FP1-1) */

         /* Get polar and cartesian coordinates for B(z) */
         r2   = 1-FP2*(r[i]-1);  m2 = -FP2*m[i];
         mod2 = sqrt( r2**2 + m2**2 );
         if r2 = 0 then do;
            theta2 = sign(m2)*1.5707963267948;
         end;
         else do;
            theta2 = atan(m2/r2);
         end;
         /* Make sure that theta is in the correct quadrant- atan always returns 
            values in quadrants 1 and 4 */
         if r2 < 0 then do;
            theta2 = 3.1415926535897932384 + theta2;
         end;

         /* This is P(z) */
         r3 = mod2**(-1*FP1-1)*cos((-1*FP1-1)*theta2);
         m3 = mod2**(-1*FP1-1)*sin((-1*FP1-1)*theta2);
         r[i] = FP1*FP2*r3;
         m[i] = FP1*FP2*m3;         
      end;
   endsub;
%mend;

/*****************************************************************************************
 *   Needed to convert from one frequency interval to the next
 *   Variables fparam_1 -> fparam_1 * ratio
 *****************************************************************************************/
%macro ChangeIntervalNEGBIN( ratio ) ;
    fparam_1 =  &ratio * fparam_1;
%mend ChangeIntervalNEGBIN;

%macro ValidateParamsNEGBIN(Valid, Reason);
   if (FParam_1 > 0) AND (FParam_2 > 0) then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = "%sysfunc(sasmsg(&orvdds, negbin_param_not_pos_error, noquote))";
   end;
%mend ValidateParamsNEGBIN;

/* Dummy implementation of StratumModel */
%macro StratumModelNEGBIN(Mean, Model);
   &Mean = FParam_1*FParam_2;
   &Model = put(&Mean, BEST8.);    
   
%mend;

/* The macro provides the resolved dynamic texts for java client/report to display 
 * context sensitive string. It resolves the display names and/or the values of 
 * the parameters used in the distribution in the model string and the frequency 
 * model item title.   
 */ 
%macro DisplayModelNEGBIN(
                           FparmNameDS, 
                           freqModelDS, 
                           countout_estDS,
                           modelString, 
                           modelItemTitle
                           );

   /* retrieve the summation count */
   %let sumN =;
   %orvu_getObsCount(&freqModelDS,sumN);


   /* Are there KRIs? */
   data _null_;
      set &countout_estDS.;
      where upcase(distribution) eq "NEGBIN";

      call symputX("use_KRIs",Freq_Use_Kris);
   run;

   proc sql noprint;
         %do i = 1 %to 2;
            select 
                     fn.displayName, 
                     put(ce.FParam_&i, 10.3)
            into 
                     :DisplayName&i, 
                     :FParam_&i 
            from 
                     &FparmNameDS fn, 
                     &countout_estDS ce
            where 
                     upcase(fn.distribution)="NEGBIN" 
                     and 
                     upcase(fn.datasetName)="FPARAM_&i"
            ;
         %end;
   quit;

   %let wordMean = %sysfunc(sasmsg(&orvdds, model_item_title_mean, noquote));

   %let DisplayName1_header = %trim(&DisplayName1);
   %if %lowcase(&use_KRIs.) eq true or &sumN. > 1 %then %do;
      %let DisplayName1_header = &DisplayName1_header._0;
   %end;
      
   %if &sumN eq 1 %then %do;
      %let &modelString=%BQUOTE(%trim(&DisplayName1_header), %trim(&DisplayName2)); 
      %let &modelItemTitle=%BQUOTE(%trim(&DisplayName1_header));      
   %end;
   %else %do;
      %let &modelString=%BQUOTE(%trim(&DisplayName1_header) = #SUM(%trim(&DisplayName1)_i,i, 1, &sumN), %trim(&DisplayName2)); 
      %let &modelItemTitle=%BQUOTE(%trim(&DisplayName1)_i);     
   %end;

%mend;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_NEGBIN(minimalInit=true);

    %let DIST_NAME= Negative Binomial;
    %let PARAM_COUNT= 2;
    
    data NEGBIN_STATUS;
        status=0;
    run;    
%mend ORVD_NEGBIN;
