/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_PAR.sas
 *   
 *   -Provides support for the Pareto distribution-
 *
 * This file contains macro fragments to implement calculation specific to the 
 * Pareto distribution
 *
 *  The macro %ORVD_PAR is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_PAR macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ********************************************************************/
   
/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfPAR(x,returnVal);
      &returnVal = 1 - (&Param_1/(&x + &Param_1))**&Param_2;
%mend CdfPAR;

%macro VCdfPAR(x,returnVal);
      &returnVal = 1 - (Param_1/(&x + Param_1))**Param_2;   
%mend VCdfPAR;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfPAR(x,returnVal);
      &returnVal = &Param_2*(&Param_1 / (&x + &Param_1))**(&Param_2) / (&x + &Param_1);   
%mend PdfPAR;

%macro VPdfPAR(x,returnVal);
      &returnVal = Param_2 * ( Param_1 / (&x + Param_1))**Param_2 / (&x + Param_1);   
%mend VPdfPAR;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantPAR(scaleFactor=1);
   data _temp_;
      set _temp_ nobs=nobs end=eof;
      retain truncdf 0;

      if _N_ = 1 then do;
         if THRESHOLD = 0 then do;
            truncdf = 0;
         end;
         else do;
            %CDFPAR(THRESHOLD,truncdf)
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf   = ecdf * (1-truncdf) + truncdf;

      if  cdf/(&scaleFactor) LE 1.0 then do; 
         quant = &param_1 * ( 1/((1-cdf/(&scaleFactor))**( 1/&param_2 )) - 1 );
      end;
      else do;
         quant = .;
      end;
   run;
%mend QuantPAR;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerPAR(scaleFactor=1, indata=, outdata=);
   data &outdata;
      set &indata;
      %CdfPAR(quant,cdf);
      cdf =  &scaleFactor.*cdf;
      DParam_1 = &scaleFactor.*(&Param_2)/quant*(1+&Param_1/quant)**(-1*&Param_2 - 1);
      DParam_2 = &scaleFactor.*(1 + &Param_1/quant)*log(1 + &Param_1/quant);
      keep quant cdf DParam_:;
   run;
%mend CdfDerPAR;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimPAR(loss);
    cdf   =  rand("UNIFORM"); 
    &loss =  (Param_1) * ( cdf**(-1/Param_2) - 1 );
%mend SimPAR;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitPAR(summaryData = , KRI_FIT = NO);
    %let nparms=2;    
   
   /*- initial guess based on moments -*/
    data _null_;
      set &summaryData;

      if sample_variance > sample_mean**2 then do;
         alpha = 2 * sample_variance / (sample_variance - sample_mean**2);
      end;
      else do;
         alpha = 2;
      end;

      theta = sample_mean * (alpha - 1);

      call symputX("Param_2_init", put(alpha, BEST20.));
      call symputX("Param_1_init", put(theta, BEST20.));
      call symputX("obsCount",obsCount);
   run;

   %if &dist NE MIXDIST %then %do;  
      %put **********************************;
      %put Distribution  =  PAR;
      %put Observations  =  &obsCount;
      %put Param_1_init  =  &Param_1_init;
      %put Param_2_init  =  &Param_2_init;
      %put **********************************;
   %end;
%mend InitFitPAR;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitPAR;
   ods listing close;
    ods output 
      IterStop          = _logl_
      ConvergenceStatus = _cs_ 
      covmat            = _covmat_ 
      ;  

    proc nlp 
      data   =  _logworkdata_ 
      outest =  _parmnlp_
      &PROC_NLP_OPTIONS.
      ;

    parms 
      Param_1 = &Param_1_init,
      Param_2 = &Param_2_init
      ;

    bounds Param_1 > 0 %Param_BoundsPAR(); 

        max logf;

         if &THRESHOLDPERCENT >= 0 then do;
            lsdf =   - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT) *
                     log(1 - (Param_1/(threshold + Param_1))**Param_2 + 1E-30);
         end;
         else if THRESHOLD = 0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf =  Param_2 * (log(Param_1) - log(THRESHOLD + Param_1));
         end;

         logf = -lsdf + 
               log(Param_2) + 
               Param_2 * log(Param_1) - 
               (Param_2 + 1) * log(AMOUNT + Param_1);
        quit;

   ods listing;
%mend FitPAR;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriPAR;
   ods listing close;
      ods output 
               IterStop          = _logl_
               ConvergenceStatus = _cs_ 
               covmat            = _covmat_ 
               ;  

       proc nlp 
               data   = _LogWorkdata_KRI_(drop= PARAM_1:) 
               outest = _parmnlp_
               &PROC_NLP_OPTIONS.
               ;

         parms 
               Param_0 = &Param_1_init,
               Param_2 = &param_2_init,
               &KRI_INIT_PARM
               ;

         bounds Param_0 > 0 %Param_BoundsPAR(); 
         max logf;
         LogParam_1 = Safe_ID(log(Param_0) + &KRI_MODEL_YES_INTERCEPT);
         Param_1    = exp(LogParam_1 );

         if &THRESHOLDPERCENT >= 0 then do;
            lsdf = - &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT) *
                  log(1 - (Param_1/(threshold + Param_1))**Param_2 + 1E-30);
         end;
         else if THRESHOLD=0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf = Param_2 * (LogParam_1 - log(THRESHOLD + Param_1));
         end;

         logf  =  -lsdf + 
                  log(Param_2) + 
                  Param_2 * LogParam_1  - 
                  (Param_2 + 1) * log(AMOUNT + Param_1)
                  ;
      quit;
   ods listing;
%mend FitKriPAR;

/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsPAR;
    %do i=1 %to 4;
      if &i < Param_2 then do;
         m&i = 1;

         do _par_j = 1 to &i;
            m&i = m&i * Param_1 * _par_j /(param_2 - _par_j);
         end;
      end;
      else do; 
         m&i = .;
      end;
    %end;

    mean=m1;
    std=sqrt(m2-m1*m1);
    skewness=(m3-3*m1*m2+2*(m1**3))/(std**3);
    kurtosis=(m4-4*m1*m3+6*(m1**2)*m2-3*(m1**4))/(std**4);
%mend MomentsPAR;

%macro LimMomentsPAR(i, x, returnVal);
   if &i = 1 AND Param_2 = 1 then do;
      &returnVal =  - Param_1 * log(Param_1/(&x + Param_1));
   end;
   else if &i ge Param_2 then do;
      /* In this case the incomplete Beta function is not defined in the SAS library, so use numerical 
         integration of the ith power of the inverse cdf (Simpson's Rule)
      */
      %local n;
      %let n = 300;   /* Must be an even integer */
      &returnVal = 0;
      _lm__exp__ = 1 / param_2;

      /* CDF evaluated at x divided by number of steps */   
      %VCdfPAR(&x,_lm__h__); 
      _lm__h__ = _lm__h__/&n.;

      do _lm__j__ = 0 to &n;
         if _lm__j__  = 0 or _lm__j__ = &n then do;
            _lm__coeff__ = 1;
         end;
         else if _lm__j__/2 = int(_lm__j__/2) then do;
            _lm__coeff__ = 2;
         end;
         else do;
            _lm__coeff__ = 4;
         end;

         &returnVal = &returnVal + _lm__coeff__*(( 1 / ( 1 - (_lm__h__*_lm__j__) )**_lm__exp__) - 1)**&i.;
      end;

      &returnVal = &returnVal * Param_1**&i.;
      &returnVal = (&returnVal*_lm__h__/3) + (1-_lm__h__*&n.)*(&x.**&i.);
   end;
   else do;
      &returnVal =  Param_1**&i * gamma(&i+1) * gamma(Param_2 - &i) * cdf('BETA',&x/(&x+Param_1), &i+1, Param_2-&i)
      /
      gamma(Param_2)
      +
      &x**&i * (Param_1/(&x + Param_1))**Param_2;
   end;
%mend LimMomentsPAR;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the location, param_1, param_2, and shape2 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitPAR;
    data _null_;
      d = &variance-&mean**2;
      if d <= 0 then do;
         p1 = .;
         p2 = .;
      call symputX("status",1);
      call symput("reason",sasmsg("&orvdds", "burr_no_mom_estimate_error", "quote"));

      end;
      else do;
      p2 = 2*&variance/d;
      p1 = &mean*(P2 - 1);

      end;
      call symputX("Param_1", put(p1, BEST20.));
      call symputX("Param_2", put(p2, BEST20.));
   run;
%mend MomFitPAR;

%macro Param_BoundsPAR();
   , Param_2>0
%mend;

%macro ValidateParamsPAR(valid, Reason);
   if (Param_1 > 0) AND (Param_2 > 0) then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "par_one_param_not_pos_error", "quote");
   end;
%mend ValidateParamsPAR;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_PAR(minimalInit = true);
   %let NEEDED_INPUT         = ;
   %let SCALE_TRANSFORM      = ID;
   %let DIST_NAME            = Pareto;
   %let PARAM_COUNT          = 2;
   %let CDF_TRANSFORM_FACTOR = 1;
   %let ProbMeasure          = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..PAR;
         function Cdf_Par(x,Param_1 &PAR_Param_List_Minus_1);
            return( 1-((Param_1)/(x+(Param_1)))**(Param_2));   
         endsub;

         function Pdf_Par(x,Param_1 &PAR_Param_List_Minus_1);
            return( Param_2 *( Param_1  /(x + Param_1 ))**Param_2 /(x + Param_1 ));   
         endsub;
      run;
   %end;
%DONE:

   data PAR_STATUS;
      status = 0;
   run;  
%mend ORVD_PAR;
