/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_pois.sas
 *   
 *   -Provides support for the Poisson distribution-
 *
 *  This file contains macro fragments to implement calculation specific to the 
 *  POISonential distribution
 *
 *  The macro %ORVD_POIS is at the end of this file. It is an empty macro that takes
 *  no paramter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_POIS macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 ******************************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfPOIS(x,returnVal);
      if &FParam_1 < 1E-8 then do;
         if &x >= 0 then do;
            &returnVal = 1;
         end;
         else do;
            &returnVal = 0;
         end;
      end;
      else do;
         &returnVal = cdf("POISSON",&x,&FParam_1);
      end;
%mend CdfPOIS;

%macro VCdfPOIS(x,returnVal);
      if FParam_1 < 1E-8 then do;
         if &x >= 0 then do;
            &returnVal = 1;
         end;
         else do;
            &returnVal = 0;
         end;
      end;
      else do;
         &returnVal = cdf("POISSON",&x,FParam_1);
      end; 
%mend VCdfPOIS;

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfPOIS(x, returnVal);
      if &FParam_1 < 1E-8 then do;
         if &x = 0 then do;
            &returnVal = 1;
         end;
         else do;
            &returnVal = 0;
         end;
      end;
      else do;
         &returnVal=pdf("POISSON",&x,&FParam_1);
      end; 
%mend PdfPOIS;

%macro VPdfPOIS(x, returnVal);
      if FParam_1 < 1E-8 then do;
         if &x = 0 then do;
            &returnVal = 1;
         end;
         else do;
            &returnVal = 0;
         end;
      end;
      else do;
         &returnVal=pdf("POISSON", &x, FParam_1);
      end; 
%mend VPdfPOIS;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantPOIS;
   data _temp_;
      cdf=_N_/(nobs+1);
      if &FParam_1 < 1E-8 then do;
         quant = 0;
      end;
      else do;
         quant= quantile("POISSON", cdf, &FParam_1 );
      end;
   run;
%mend QuantPOIS;

%macro VQuantPOIS(x, returnVal);
   if FParam_1 < 1E-8 then do;
      &returnVal = 0;
   end;
   else do;
      &returnVal = quantile("POISSON", &x, FParam_1 );
   end;
%mend VQuantPOIS;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimPOIS(loss);
   if fparam_1 > 1 then 
    &loss = rand("POIS",FParam_1);
   else do;
      &loss=rand("BINOMIAL",FParam_1/20,20);
   end;
%mend SimPOIS;

/*****************************************************************************************
 *   What it does  :  Initialize the poisson paramter for fitting.
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitPOIS(countdata=);
   /* Not Used */
%mend InitFitPOIS;
    
/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC COUNTREG
 *
 *   Parameters:
 *       countdata - The data that is used in fitting
 *       outparm   - Dataset to create with the estimated parameters
 *       outcov    - Parameter covariance matrix
 *      
 *****************************************************************************************/
%macro FitPOIS( countdata =, outparm =, outcov = );
   %local reason status intercept;

   %let methods = QUANEW NEWRAP TRUREG ; 
   %let try = 1;
   %let curr_method = %scan(&methods, &try, ' ');

%TRYAGAIN:
   /* proc countreg */
   ods listing close;
   ods output FitSummary = _FitSumamry ;
   proc countreg 
         data     = &CountData
         outest   = count_out_est 
         covout 
         GCONV    = 1E-14 
         FCONV    = 1E-20 
         ABSFCONV = 0 
         ABSGCONV = 1E-10 
         method   = &curr_method.
         type     = POISSON; 
      model 
         the_count = &SelectedKRIs. 
                     &SelectedDummies.
                     / offset = offset;
   run;
   ods output close;
   ods listing;

   %let status = -1;
   data _null_;
      set count_out_est;
      if _N_ = 1 then do;
         call symput("status", substr(_STATUS_,1,2));
         call symput("reason", substr(_STATUS_,3));
      end;
   run;

   %if &status. NE 0  %then %do;
      %let try = %eval(&try. + 1);
      %if &try LT 4 %then %do; 
         %let prev_method = &curr_method;
         %let curr_method = %scan(&methods., &try, ' ');
         %put &ORVAR_WARNING: %sysfunc(sasmsg(&orvads., pois_method_failed, noquote, &prev_method.));
         %put %sysfunc(sasmsg(&orvads., pois_trying_new_method, noquote, &curr_method.));
         %goto TRYAGAIN;
      %end;
      %else %do;
         %let reason = %sysfunc(sasmsg(&orvads., pois_all_methods_failed, noquote));
         %orva_status_ds( statusds=&outparm., status=1, reason=&reason. );

         %return; 
      %end;
   %end;

   /* Need variable names from the fitsummary dataset to be uniform across loaclizations, 
      so use english names for the variables */
   %orva_unlocalize_countreg_fitsum( summaryData = _FitSumamry );

   /* get the covariance matrix and parameter estimation data */
   data 
      &outcov.(drop = _TYPE_ _STATUS_ )
      &outparm.(drop= _NAME_ _TYPE_);
         length _NAME_ $32;
         set count_out_est;
         if _TYPE_ = "COV" then do;
            output  &outcov.;
         end;
         else if _N_ = 1 then do;
            output &outparm.;
         end;       
   run;

   /* Adjust the parameter estimation output: add information from fitsummary dataset */
   data  &outparm.;
      format reason &ORVAR_REASON_FMT;
      status = &status;
      reason = "&reason";
      set _FitSumamry;
      set &outparm.;
   run;

%mend FitPOIS;

/*****************************************************************************************
 *   What it does  :  Adjusts parameters and covariance matrix so that we deal
 *                    with lambda rather than intercept.
 *
 *   Parameters:
 *       parm   - Dataset to modify with the estimated parameters
 *       cov    - Parameter covariance matrix to modify
 *      
 *****************************************************************************************/
%macro PostFitPOIS( parm =, cov = );

   /* Adjust the parameter estimation output - intercept to lambda */
   data  &parm.;
      set &parm.;

      FParam_1 = exp(min(intercept,700));
      call symputX("intercept",put(intercept,BEST20.));
   run;
 
   /* Because the variance of the intercept is multiplied by exp(intercept) twice,
      need to cap this value to avoid overflow */
   %let intercept = %sysfunc(min(&intercept., 70));

   /* Countreg estimates an intercept, but we use FParam_1, which is exp(intercept) */
   /* Adjustment is describned in 'adjusting covariance matrixes' doc */
   data &cov.;
      set &cov.;
      array nums _NUMERIC_;
      label intercept = "FParam_1";

      /* Multiply the intercept row by exp(&intercept) */
      if trim(lowcase(_NAME_)) = "intercept" then do;
         _NAME_ = "FParam_1";
         do i=1 to dim(nums);
            nums[i] = nums[i]*exp(&intercept.);
         end;
      end;

      /* Multiply the intercept column by exp(&intercept) */
      intercept = intercept*exp(&intercept.);

      rename intercept = FParam_1;
      drop i;
   run;

%mend PostFitPOIS;

/*****************************************************************************************
 *   What it does  :  Calculates the moments from fitted params
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsPOIS();
    Mean  = FParam_1;
    Variance = FParam_1;
%mend MomentsPOIS;

/*****************************************************************************************
 *   What it does:  Does method of moments estimation and assigns the corresponding 
 *                  values to the distributions parameters
 *
 *   Parameters  : assumes &mean and &variance are set
 *****************************************************************************************/
%macro MomFitPOIS;
    FParam_1=&mean;
%mend MomFitPOIS;

/*****************************************************************************************
  *   What it does:  If you write Fparam_1 as function of mean and other parmeters, this is
  *   the derivative of Fparam_1 with respect to the mean.   Used in parameter variance CI calculcation 
  *****************************************************************************************/
%macro MomDerPOIS(deriv=);
   &deriv = 1;
%mend MomDerPOIS;

/*****************************************************************************************
  *   What it does:  Generates fcmp subroutines to compute poisson pgf and its derivatives
  *****************************************************************************************/
%macro pgf_funcsPOIS;
   /*----------------------------------------------------------------------------------------
     -
     -   What it does:  Compute the PGF at a complex point z given as real part (r) and 
     -                  imaginary part (m).  Inputs are overwritten with the output.
     -
     -   Parameters  : lambda- poisson parameter
     -                 r- array: real part points at which to evaluate pgf
     -                 m- array: imaginary part of points at which to evaluate pgf
     -                 n- length of the arrays
     -
     ----------------------------------------------------------------------------------------*/
   subroutine pgfPOIS( lambda,
                       r[*],
                       m[*],
                       n );
      outargs r,m;
      do i=1 to n;
         r2 = exp(lambda*((r[i])-1))*cos(lambda*(m[i]));
         m2 = exp(lambda*((r[i])-1))*sin(lambda*(m[i]));
         r[i] = r2;
         m[i] = m2;
      end;
   endsub;

   /*----------------------------------------------------------------------------------------
     -
     -   What it does:  Compute the derivative of the PGF at a complex point z  given as 
     -                  real part (r) and imaginary part (m) with respect to the parameter lambda
     -
     -   Parameters  : dl_r-  array: real part of derivative wrt lambda
     -                 dl_m-  array: imaginary part of derivative wrt lambda
     -                 lambda- poisson parameter
     -                 r- array: real part of points at which to evaluate pgf derivative
     -                 m- array: real part of points at which to evaluate pgf derivative
     -                 n- length of arrays
     -
     ----------------------------------------------------------------------------------------*/
   subroutine pgfderPOIS( dl_r[*],
                          dl_m[*],
                          lambda,
                          r[*],
                          m[*],
                          n );
      outargs dl_r,dl_m;
      do i=1 to n;
         r2 = exp(lambda*((r[i])-1))*cos(lambda*(m[i]));
         m2 = exp(lambda*((r[i])-1))*sin(lambda*(m[i]));
         r3 = r[i]-1;
         m3 = m[i];
         dl_r[i] = r2*r3 - m2*m3;
         dl_m[i] = r2*m3 + m2*r3;
      end;
   endsub;

   /*----------------------------------------------------------------------------------------
     -   What it does:  Compute the derivative of the PGF at a (complex) point z with respect 
     -                  to the complex argument z
     -   Parameters  : lambda- poisson parameter
     -                 r- array: real part of points at which to evaluate pgf derivative
     -                 m- array: imaginary part of points at which to evaluate pgf derivative
     -                 n- length of arrays
     ----------------------------------------------------------------------------------------*/
   subroutine pgfCompDerPOIS( lambda,
                              r[*],
                              m[*],
                              n );
      outargs r,m;
      do i=1 to n;
         r2 = exp(lambda*((r[i])-1))*cos(lambda*(m[i]));
         m2 = exp(lambda*((r[i])-1))*sin(lambda*(m[i]));
         r[i] = lambda*r2;
         m[i] = lambda*m2;
      end;
   endsub;

%mend;
/*****************************************************************************************
 *   Needed to convert from one frequency interval to the next
 *   Variables fparam_1 -> fparam_1 * ratio
 *****************************************************************************************/
%macro ChangeIntervalPOIS( ratio ) ;
    fparam_1 = &ratio * fparam_1;
%mend ChangeIntervalPOIS;

%macro ValidateParamsPOIS(Valid , Reason);
   if FParam_1 > 0 then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = "%sysfunc(sasmsg(&orvlds, imsp_poisson_param_error, noquote))";
   end;
%mend ValidateParamsPOIS;

/* Dummy implementation of StratumModel */
%macro StratumModelPOIS(Mean, Model);
   &Model = put(FParam_1, BEST8.);    
   &Mean = FParam_1;
%mend;

/* The macro provides the resolved dynamic texts for java client/report to display 
 * context sensitive string. It resolves the display names and/or the values of 
 * the parameters used in the distribution in the model string and the frequency 
 * model item title.   
 */ 
%macro DisplayModelPOIS(
                        FparmNameDS, 
                        freqModelDS, 
                        countout_estDS,
                        modelString, 
                        modelItemTitle
                        );

   /* retrieve the summation count */
   %let sumN =;
   %orvu_getObsCount(&freqModelDS,sumN);


   /* Are there KRIs? */
   data _null_;
      set &countout_estDS.;
      where upcase(distribution) eq "POIS";

      call symputX("use_KRIs",Freq_Use_Kris);
   run;

   /* for each distribution, resolve the name mapping */
   proc sql noprint;
      SELECT
               fn.displayName, 
               put(ce.FParam_1, 10.3)
      INTO 
               :DisplayName1, 
               :FParam_1 
      FROM 
               &FparmNameDS fn, 
               &countout_estDS ce
      WHERE 
               upcase(fn.distribution) = "POIS" 
               and 
               upcase(fn.datasetName)="FPARAM_1";
   quit;
   
   %let wordMean = %sysfunc(sasmsg(&orvdds, model_item_title_mean, noquote));

   %let DisplayName1_header = %trim(&DisplayName1);
   %if %lowcase(&use_KRIs.) eq true or &sumN. > 1 %then %do;
      %let DisplayName1_header = &DisplayName1_header._0;
   %end;

   %if &sumN eq 1 %then %do;
      %let &modelString    = %BQUOTE(%trim(&DisplayName1_header));
      %let &modelItemTitle = %BQUOTE(%trim(&DisplayName1_header));
   %end;
   %else %do;
      %let &modelString=%BQUOTE(%trim(&DisplayName1_header)=#SUM(%trim(&DisplayName1)_i,i, 1, &sumN)); 
      %let &modelItemTitle=%BQUOTE(%trim(&DisplayName1)_i);    
   %end;    
%mend;
/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_POIS(minimalInit=true);
    %let DIST_NAME= Poisson;
    %let PARAM_COUNT= 1;
    data POIS_STATUS;
        status=0;
    run;
%mend ORVD_POIS;
