/********************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_WBL.sas
 *   
 *   -Provides support for the Weibull distribution-
 *
 * This file contains macro fragments to implement calculation specific to the 
 * Weibull distribution
 *
 *  The macro %ORVD_WBL is at the end of this file. It is an empty macro that takes
 *  no parameter and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_WBL macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 ********************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro CdfWBL(x,returnVal);
       &returnVal=1-exp(-(&x/&Param_1)**&Param_2);    
%mend CdfWBL;

%macro VCdfWBL(x,returnVal);
       &returnVal=1-exp(-(&x/Param_1)**Param_2);    
%mend VCdfWBL;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        returnVal = name of variable inton which result is stored
 *****************************************************************************************/
%macro PdfWBL(x,returnVal);
       &returnVal = &Param_2 /&x * (&x/&Param_1)**&Param_2 * exp(-(&x/&Param_1)**&Param_2);    
%mend PdfWBL;

%macro VPdfWBL(x,returnVal );
   &returnVal = Param_2 / &x * (&x  /Param_1)**Param_2* exp(-(&x/Param_1)**Param_2  );   
%mend VPdfWBL;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
%macro QuantWBL(scaleFactor=1);
   data _temp_;
      set _temp_ nobs=nobs end=eof;
      retain truncdf 0;

      if _N_ = 1 then do;
         if THRESHOLD = 0 then do;
            truncdf =0;
         end;
         else do;
            %CDFWBL(THRESHOLD, truncdf)
            truncdf = truncdf * (&scaleFactor);
         end;
      end;

      cdf   = ecdf*(1-truncdf) + truncdf;

      if  cdf/(&scaleFactor) LE 1.0 then do; 
         quant = &Param_1 * (-log(1-cdf/(&scaleFactor)))**( 1 / &Param_2 );
      end;
      else do;
         quant = .;
      end;
   run;
%mend QuantWBL;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerWBL(scaleFactor=1, indata=, outdata=);
   data &outdata;
      set &indata;
      %CdfWBL(quant,cdf);
      cdf =  &scaleFactor.*cdf;
      DParam_1 = (cdf - &scaleFactor)*&Param_2/(&Param_1)*(quant/(&Param_1))**(&Param_2);
      DParam_2 = -1*(cdf - &scaleFactor)*(quant/(&Param_1))**(&Param_2)*log(quant/(&Param_1));
      keep quant cdf DParam_:;
   run;
%mend CdfDerWBL;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimWBL(loss);
    &loss = rand("WEIBULL", Param_2, Param_1);
%mend SimWBL;

/*****************************************************************************************
 *   What it does  :  Get an initial value of the parameters
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitWBL(summaryData =, KRI_FIT = NO );
   %let nparms = 2;    

   /*- initial guess based on percentiles -*/     
   data _null_;
      set &summaryData;
      g     = log(log(4)) / log(log(4 / 3));
      sigma = exp((g * log(sample_q1) - log(sample_q3)) / (g - 1));
      c     = max(log(log(4)) / (log(sample_q3) - log(sigma)), 0.1);

      call symputX("Param_1_init", put(sigma, BEST20.));
      call symputX("Param_2_init", put(c, BEST20.));
      call symputX("obsCount", obsCount);
   run;

   %if &dist NE MIXDIST %then %do;
      %put ******************************;
      %put Distribution   =  WBL;
      %put Observations   =  &obsCount;
      %put Param_1_init   =  &Param_1_init;
      %put Param_2_init   =  &Param_2_init;
      %put ******************************;
   %end;
%mend InitFitWBL;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitWBL;
   ods listing close;
   ods output 
            IterStop            =  _logl_
            ConvergenceStatus   =  _cs_ 
            covmat              =  _covmat_ 
            ;  

   proc nlp 
           data   = _logworkdata_ 
           outest = _parmnlp_
           &PROC_NLP_OPTIONS.
           ;

      parms 
            Param_1 = &Param_1_init, 
            Param_2 = &Param_2_init;

      bounds Param_1 > 0 %Param_BoundsWBL(); 

      max logf;

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
               log(cdf("Weibull", Threshold, Param_2, Param_1) + 1E-30);
      end; 
      else if THRESHOLD=0 then do;
         lsdf = 0;
      end;
      else do;
         lsdf = (THRESHOLD/(Param_1))**Param_2;
      end;

      logf =   lsdf + 
               log(Param_2) +
               (Param_2-1) * log(AMOUNT) - 
               Param_2 * log(Param_1) -
               (AMOUNT/(Param_1))**Param_2
               ;
   quit;
   ods listing;
%mend FitWBL;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriWBL;
   ods listing close;
   ods output 
            IterStop          =  _logl_
            ConvergenceStatus =  _cs_ 
            covmat            =  _covmat_ 
            ;  

   proc nlp 
            data   = _LogWorkdata_KRI_(drop= PARAM_1:) 
            outest = _parmnlp_
            &PROC_NLP_OPTIONS.
            ;

      parms 
            Param_0 = &Param_1_init, 
            Param_2 = &Param_2_init,
            &KRI_INIT_PARM
            ;

      bounds Param_0 > 0 %Param_BoundsWBL(); 
      max logf;
      LogParam_1 = Safe_ID(log(Param_0) + &KRI_MODEL_YES_INTERCEPT );
      Param_1    = exp(LogParam_1);

      if &THRESHOLDPERCENT >= 0 then do;
         lsdf = &THRESHOLDPERCENT/(1-&THRESHOLDPERCENT)*
               log(cdf("Weibull", Threshold, Param_2, Param_1) + 1E-30);
      end; 
      else do;
         if THRESHOLD=0 then do;
            lsdf = 0;
         end;
         else do;
            lsdf = (THRESHOLD / Param_1)**Param_2;
         end;
      end;

      logf = lsdf + 
             log(Param_2) + 
             (Param_2 - 1) * Log(AMOUNT) - 
             Param_2 * LogParam_1 - 
             (AMOUNT / Param_1)**Param_2
             ;
   quit;
   ods listing;
%mend FitKriWBL;

/*****************************************************************************************
 *   What it does  :  Calculates the moments
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomentsWBL;
    %do i=1 %to 4;
        m&i=(param_1**&i)*exp(lgamma(1+&i/param_2));
    %end;
    mean=m1;
    std=sqrt(m2-m1*m1);
    skewness=(m3-3*m1*m2+2*(m1**3))/(std**3);
    kurtosis=(m4-4*m1*m3+6*(m1**2)*m2-3*(m1**4))/(std**4);    
%mend MomentsWBL;

%macro LimMomentsWBL(i, x, returnVal);
    &returnVal = Param_1**&i 
                * gamma(1 + &i/Param_2)
                * cdf("GAMMA",(&x/Param_1)**Param_2,1 + &i/Param_2)
                + &x**&i*exp(-(&x/Param_1)**Param_2);
%mend LimMomentsWBL;


/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the location, param_1, param_2, and shape2 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitWBL;
   data _temp_;
        m1      = &mean;
        m2      = &variance + &mean * &mean;
      Param_1 = m1;
        Param_2 = 1;
       call symputX("converge", put(m1*1e-8, BEST20.));
    run;
   Ods listing close;
   Ods output obsused = _cs_;
   proc model data=_temp_;
      dependent Param_2;
      eq.one = m1*sqrt(gamma(1+2/param_2))-sqrt(m2)*gamma(1+1/param_2);
        solve param_2/maxiter=10000 converge=&converge out=_sol_(keep=param_2);
   run;
   quit;
   Ods listing;
   data _null_;
      set _cs_;
      if label1 = "Failed" then do;
         test = 1;
         if cvalue1 = 1 then do;
               call symputX("status",1);
               call symput("reason",sasmsg("&orvdds", "burr_no_mom_estimate_error", "quote"));
         end;
      end;
   run;
   
   data _null_;
      set _sol_;
      param_1 = &mean/gamma(1+1/param_2);
      call symputX("param_1", put(param_1, BEST20.));
      call symputX("param_2", put(param_2, BEST20.));
   run;     
%mend MomFitWBL;

%macro Param_BoundsWBL();
   , Param_2 > 0
%mend;

%macro ValidateParamsWBL(Valid, Reason);
   if (Param_1 > 0) AND (Param_2 > 0) then do;
      &Valid  = "true";
   end;
   else do;
      &Valid  = "false";
      &Reason = sasmsg("&orvdds", "wbl_one_param_not_pos_error", "quote");
   end;
%mend ValidateParamsWBL;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_WBL(minimalInit = true);
   %let NEEDED_INPUT    = ;
   %let SCALE_TRANSFORM = ID;
   %let DIST_NAME       = Weibull;
   %let PARAM_COUNT     = 2;
   %let ProbMeasure     = CONTINUOUS;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp outlib = &ORVAR_CMPLIB..WBL;
         function Cdf_WBL(x,Param_1 &WBL_Param_List_Minus_1);
            return(1 - exp(-(x / Param_1)**Param_2));    
         endsub;

         function Pdf_WBL(x,Param_1 &WBL_Param_List_Minus_1);
            return(Param_2 /x * (x/Param_1)**Param_2 * exp(-(x/Param_1)**Param_2));    
         endsub;
      run;
   %end; 
%DONE:
   data WBL_Status;
      status = 0;
   run;
%mend ORVD_WBL;
