/* --------------------------------------------------------------
|  Copyright SAS Institute 2012
|  Name: orvdb_export_project
|  Purpose: Exports a project to a zip file
|  Inputs:
|      data_store_root  - The fully qualified directory name for the root of the data store
|      version_name     - The name of the data version we are adding this project to
|      project_name     - The name of the project we are adding to the version - can be different from the name exported
|      model_uid        - database key for the exported project
|      exported_project - The fully qualified path to the zip file that represents the exported project
|
|  Example:
|    %orvdb_export_project( data_store_root=c:\opriskvar\9.3\,  
|                          version_name=inc2, project_name=regulatory,
|                          exported_project= c:\temp\foo.zip);
--------------------------------------------------------------------*/ 
%macro orvdb_export_project( 
                            data_store_root  = ,
                            version_name     = ,
                            project_name     = ,
                            model_uid        = ,
                            exported_project = ,
                            outStatus        = ostatus
                            );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orvdb_export_project(; 
      %put .                      data_store_root  = data_store_root.,;
      %put .                      version_name     = version_name.,;
      %put .                      project_name     = project_name.,;
      %put .                      model_uid        = model_uid.,;
      %put .                      exported_project = exported_project.,;
      %put .                      outStatus        = outStatus.;
      %put .                      );
   %end;

    /*- create the status data set up front -*/
   %let reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );

   /* create the vardata libref -------------------------------*/
   %let rc = %sysfunc(libname( vardata, &data_store_root/main));
   %if &rc > 0 %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, data_store_path_error, noquote));
      %put &reason;
      %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );
      %return;
   %end;

   /* Get the folder name */
   /* Get the version_uid --------*/
   %let version_uid = ;

   proc sql noprint;
      select version_uid into :version_uid from vardata.data_version where version_name = "&version_name" ;
      select dir_suffix into :dir_suffix from vardata.data_version where version_name = "&version_name" ;
   quit;

   %if %str('') = &version_uid %then %do;
       %let reason = %sysfunc(sasmsg(&orvads, invalid_version_name noquote));
       %put &reason;
       %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );  
       %return; 
   %end;

   /* make a directory called 'org' with the data sets we need in order to restore keys */
   %let dir_suffix = %eval(&dir_suffix);
   data _NULL_;
      length project_name $ 20; /* Purposeful truncation */
      project_name = translate("&project_name", '_', ' ');
      call symput( "project_name_short", strip(project_name));
   run;
   
   /* try with full project name first */
   %let path = &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid;
   %if %sysfunc(fileexist(&path)) = 0 %then %do;
          /* failed with project name, so just use model uid */
          %let project_name_short=%str();
          %let path = &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid;

          %if %sysfunc(fileexist(&path)) = 0 %then %do;
	          %let reason = %sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
	          %put &reason;
	          %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );
	          %return;
		  %end;
   %end;

   %let returncode = ;
   %_orv_copy_dir( 
                  source    = &data_store_root/ver_&dir_suffix/loss/org, 
                  dest      = &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid/org, 
                  destlib   = tmp_org, 
                  return    = returncode, 
                  outstatus = &outStatus 
                  );
   %if &returncode = 1 %then %return;  
   
   /* Add in strings so that we can correct the keys for certian tables --------*/
   %let rc = %sysfunc(libname( varprj_m, &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid/));
   %if &rc > 0 %then %do;
      %put %sysfunc(sysmsg()) ;
      %let reason = %sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );
      %return;
   %end;

   data work.__bl_access; set varprj_m.model_bl_access;
      %if %orvu_var_exists( INTERNAL_BL_LEVEL, varprj_m.model_bl_access) = 1 %then 
      drop INTERNAL_BL_LEVEL;;
       %if %orvu_var_exists( INTERNAL_BL_NAME, varprj_m.model_bl_access) = 1 %then 
      drop INTERNAL_BL_NAME;;
   run;
   proc sql noprint nowarn;
      CREATE TABLE 
                  tmp 
      AS SELECT 
                  * 
      FROM 
                  work.__bl_access AS a, 
                  tmp_org.internal_businessline AS b
      WHERE 
                  a.internal_bl_uid = b.internal_bl_uid 
                  ;
   quit;

   data varprj_m.model_bl_access;
      set tmp; 
      keep 
          model_uid 
          internal_bl_uid 
          var_user_uid 
          version_uid 
          internal_bl_name 
          internal_bl_level 
          ; 
   run;

   proc datasets library=vardata memtype=data noprint nowarn;
      copy 
          out        = tmp_org 
          constraint = no
          ;
      select var_sequences;
   run;

   /* Create the losses libref  -------------------------------------*/
   %let rc = %sysfunc(libname( losses, &data_store_root/ver_&dir_suffix/loss/losses));
   %if &rc > 0 %then %do;
      %put %sysfunc(sysmsg()) ;
      %let reason = %sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );
      %return;
   %end;
  
   /* copy portions (filtered) of data sets in the Vardata directory into org -----------------*/
   data tmp_org.data_version; 
      set vardata.data_version;
      where version_uid = &version_uid;
   run;

   data tmp_org.Model_ccv_mapping_refs ; 
      set vardata.Model_ccv_mapping_refs ;
      where model_uid = &model_uid and version_uid=&version_uid;
   run;

   data tmp_org.Model_refs ; 
      set vardata.Model_refs ;
      where model_uid = &model_uid and version_uid=&version_uid;
   run;

   data tmp_org.project_insurance_refs ; 
      set vardata.project_insurance_refs ;
      where model_uid = &model_uid and version_uid=&version_uid;
   run;

   data tmp_org.project_scenario_refs ; 
      set vardata.project_scenario_refs ;
      where model_uid = &model_uid and version_uid=&version_uid;
   run;

   data tmp_org.scenario_ccv_refs ; 
      set vardata.scenario_ccv_refs ;
      where version_uid = &version_uid;
   run;

   data tmp_org.var_folders; 
      set vardata.var_folders ;
      where org_structure_uid = &version_uid or parent_uid = &version_uid;
      if folder_type = "Project" and folder_name_ucase ne upcase("&project_name") then delete;
   run;
   
   /* copy the meta directory ------------*/
   %let returncode = ;
   %_orv_copy_dir( 
                  source    = &data_store_root/ver_&dir_suffix/meta, 
                  dest      = &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid/meta, 
                  destlib   = tmp_meta, 
                  return    = returncode, 
                  outstatus = &outStatus 
                  );
   %if &returncode = 1 %then %return;  

   /* copy the components directory -----------------*/
   %let returncode = ;
   %_orv_copy_dir( 
                  source    = &data_store_root/ver_&dir_suffix/components, 
                  dest      = &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid/components, 
                  destlib   = tmp_comp, 
                  return    = returncode, 
                  outstatus = &outStatus 
                  );
   %if &returncode = 1 %then %return;  

   /* copy the losses directory */
   %let returncode = ;
   %_orv_copy_dir( 
                  source    = &data_store_root/ver_&dir_suffix/loss/losses, 
                  dest      = &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid/losses, 
                  destlib   = tmp_loss, 
                  return    = returncode, 
                  outstatus = &outStatus 
                  );
   %if &returncode = 1 %then %return;  

   /* Copy the covmat directory -------------*/
   %let returncode = ;
   %_orv_copy_dir( 
                  source    = &data_store_root/ver_&dir_suffix/covmat, 
                  dest      = &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid/covmat, 
                  destlib   = tmp_covm, 
                  return    = returncode, 
                  outstatus = &outStatus 
                  );
   %if &returncode = 1 %then %return;  

   /* If the lossi_inc (et al) file is present we need to repair the contraints and index */
   proc datasets lib = varprj_m noprint nowarn;
           %if %sysfunc(exist(varprj_m.lossi_inc)) %then repair lossi_inc ;;  
           %if %sysfunc(exist(varprj_m.empirical_functions)) %then repair empirical_functions ;; 
           %if %sysfunc(exist(varprj_m.extension_input)) %then repair extension_input ;;     
   quit;     
  
   /* Zip up the project */
   %orvu_zip( 
             &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid,
             &exported_project
             ); 

   /* delete unneeded dirs --------------*/
   %orvu_delete_dir( lib = tmp_org);
   %orvu_delete_dir( lib = tmp_meta);
   %orvu_delete_dir( lib = tmp_comp);
   %orvu_delete_dir( lib = tmp_covm);
   %orvu_delete_dir( lib = tmp_loss);

   %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));
   %orva_status_ds( statusds = &outStatus, status = 0, reason = &reason, update = true );

%mend orvdb_export_project;

/* Create directory 'Dest' and copy source data sets to dest */
%macro _orv_copy_dir( source=, dest=, destlib=, return=, outStatus= );

   %let &return = 0;
   %orvu_mkdirs( path = &dest);

   /* check for success --------------------------*/
   %if &destlib = %str() %then %let outlib = tmp;
   %else %let outlib = &destlib;

   %let rc = %sysfunc(libname( &destlib, &dest));
   %if &rc > 0 %then %do;
      %put %sysfunc(sysmsg());
      %let reason = %sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );
      %let &return = 1;
      %return;
   %end;

   /* create the source libref itself ----------------------------*/
   %let rc = %sysfunc(libname( source, &source));
   %if &rc > 0 %then %do;
      %put %sysfunc(sysmsg());
      %let reason = %sysfunc(sasmsg(&orvads, data_store_path_error, noquote));
      %put &reason;
      %orva_status_ds( statusds = &outStatus, status = 1, reason = &reason );
      %let &return = 1;
      %return;
   %end;

   proc datasets 
                library = source 
                memtype = data 
                noprint nowarn
                ;
      copy 
          out        = &outlib 
          constraint = no
          ;
   run;quit;

   /* clear some librefs --------------------*/
   %let rc = %sysfunc(libname( source ));
   %if &destlib = %str() %then %let rc = %sysfunc(libname( tmp ));
%mend _orv_copy_dir;
