/* --------------------------------------------------------------
|  Copyright SAS Institute 2012
|  Name: orvdb_import_project
|  Purpose: Imports a project from a another data store to an existing version
|  Inputs:
|      data_store_root  - The fully qualified directory name for the root of the data store
|      user_id          - A valid user_id already in the data store - sasdemo
|      version_name     - The name of the data version we are adding this project to
|      project_name     - The name of the project we are adding to the version - can be different from the name exported
|      exported_project - The fully qualified path to the zip file that represents the exported project
|      parent_uid       - The Var_folders_uid of the parent folder of the project - 20 is the default which implies
|                         no sub folder for the project
|
|  Example:
|    %orvdb_import_project( data_store_root=c:\opriskvar\9.3\, user_id=sasdemo, version_name=inc2, project_name=copy1,
|                          exported_project= c:\temp\foo.zip );
--------------------------------------------------------------------*/ 
%macro orvdb_import_project( data_store_root=,
                             user_id=,
                             version_name=,
                             project_name=,
                             exported_project=,
                             parent_uid= 20,
                             outStatus      =ostatus);
                             
                             
    /*- create the status data set up front -*/
   %let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );


   /* create the vardata libref -------------------------------*/
   %let rc =%sysfunc(libname( vardata, &data_store_root./main));
   %if &rc > 0 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, data_store_path_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;

   /* Create the folder object */
   /* Get the version_uid --------*/
   /* get the user_id ---------------------------------------------------------------------- */
   %local version_uid user_id;
   %let version_uid=;
   %let user_uid=;
   proc sql noprint;
   select version_uid into :version_uid from vardata.data_version where version_name = "&version_name." ;
   select dir_suffix into :dir_suffix from vardata.data_version where version_name = "&version_name." ;
   select var_user_uid into :user_uid from vardata.var_user where user_id="&user_id.";
   quit;
   
   %if %str() = &version_uid %then %do;
       %let reason=%sysfunc(sasmsg(&orvads, invalid_version_name, noquote));
       %put &reason;
       %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );  
       %return; 
   %end;

   %if %str() = &user_uid %then %do;
       %let reason=%sysfunc(sasmsg(&orvads, invalid_user_name, noquote));
       %put &reason;
       %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );  
       %return; 
   %end;

  /* check for project of the same name already existing ------------------*/
   %local exists;
   %let exists=;
   proc sql noprint;
   select folder_name into :exists from VARDATA.VAR_FOLDERS where folder_name_ucase=upcase("&project_name.") and
           org_structure_uid=&version_uid and folder_type = "Project" and parent_uid = &parent_uid;
   quit;
   %if &exists ne %str() %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, Project_already_exists, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
   
   /* get a unique key to create the var_folders_uid --------------------------------------- */
   %orvdb_seq_incrementer(VarData, Folder_SEQ, var_folders, INC=1);
   proc sql noprint;
      select VALUE Into :var_folders_uid From WORK.VAR_SEQUENCES Where NAME = 'var_folders';
   quit;
   %let model_uid=&var_folders_uid;

   %let dir_suffix=%eval(&dir_suffix.);

   /* sets &project_name_short */
   %let project_name_short = %str();
   %orvu_project_name_get(project_name = &project_name.);

   %let path= &data_store_root./ver_&dir_suffix./components/projects/&project_name_short.&Model_uid.;


   /* unzip the files into that folder */
   %orvu_unzip( &exported_project, &path.);
   %if &syserr ne 0 %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, export_file_error, &path, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;

   /* create the varprj_m libref -------------------------------*/
   %let rc =%sysfunc(libname( varprj_m, &path.));
   %if &rc > 0 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
   
   /* Open the old and new org libref  --------------------------------------*/
   %let rc =%sysfunc(libname( org_old, &path./org));
   %if &rc > 0 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
   %let rc =%sysfunc(libname( org, &data_store_root./ver_&dir_suffix./loss/org));
   %if &rc > 0 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
   %let rc =%sysfunc(libname( losses, &data_store_root./ver_&dir_suffix./loss/losses));
   %if &rc > 0 %then %do;
      %put %sysfunc(sysmsg()) ;
      %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
   %let rc =%sysfunc(libname( comp, &data_store_root./ver_&dir_suffix./components));
   %if &rc > 0 %then %do;
      %put %sysfunc(sysmsg()) ;
      %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
   
   /* update the sequence keys so that we don't get collisions later ------------------------ */
   data vardata.var_sequences; set vardata.var_sequences; set org_old.var_sequences(rename=(seq_count=input_seq_count));
   seq_count = max(seq_count,input_seq_count);
   drop input_seq_count;
   run;
   
   /* Create new folder data for the project -----------*/
   proc sql noprint;
    Insert Into 
				VARDATA.VAR_FOLDERS (
									VAR_FOLDERS_UID, 
									FOLDER_NAME, 
									FOLDER_NAME_UCASE, 
									FOLDER_TYPE, 
									STATUS_FLAG, 
									PARENT_UID, 
									ORG_STRUCTURE_UID, 
									USER_UID
									) 
					Values (
							&var_folders_uid, 
							"&project_name.", 
							"%upcase(&project_name)", 
							'Project', 
							0, 
							&parent_uid, 
							&version_uid, 
							&user_uid
							);
   quit;

   /* Update the model_uid ----------------------------------------*/
   data varprj_m.loss_event_audit_for_model; 
		set  varprj_m.loss_event_audit_for_model;
   		model_uid = &model_uid;
   run;

   /* Only the import user will get business line access -----------------------------*/
   data varprj_m.model_bl_access; set  varprj_m.model_bl_access;
   model_uid = &model_uid;
   var_user_uid = &user_uid;
   run;
   
   
        /* The ccv_uids have to be changed to match the new ccv_uids -------*/
        /* create a map for the old to the new -----------------------------*/
   data var_folders; set org_old.var_folders;
     rename var_folders_uid=old_var_folders_uid;
   run;
   %_orv_build_paths( var_folders, out_var_folders);

   data var_folders; set vardata.var_folders;
     rename var_folders_uid=old_var_folders_uid;
     where org_structure_uid = &version_uid;
   run;
   %_orv_build_paths( var_folders, out_map);
   data _vf; set  out_map;
     rename old_var_folders_uid=var_folders_uid;
   run;
   
   proc sql noprint;
          create table _map as select * from
            out_var_folders a,  _vf b where a.folder_type = b.folder_type and 
            a.folder_type ne "Event" 
			and a.folder_name_ucase = b.folder_name_ucase and a.name=b.name ;
        quit;
   
   
   data map; set _map;
       old = old_var_folders_uid;
	   new = var_folders_uid;
	   keep old new;
   run;
   %let ccv_abort=0;
   data varprj_m.model_ccv_mapping; set  varprj_m.model_ccv_mapping;
       length folder_name $50;
       if _N_ = 1 then do;
               declare hash h (dataset: "work.map");
               rc =h.definekey('old');
               rc = h.definedata('new');
               h.definedone();
               call missing(old, new);

               declare hash missccv (dataset: "work.out_var_folders");
               rc =missccv.definekey('old_var_folders_uid');
               rc = missccv.definedata('Folder_Name');
               missccv.definedone();
               call missing(old_var_folders_uid, folder_name);
       end;
   model_uid = &model_uid;
   old = ccv_uid;
   rc = h.find();
   if new = . then do;
      old_var_folders_uid = ccv_uid;
      rc = missccv.find();
      call symput("missing_ccv", trim(Folder_name));
      call symput("ccv_abort",1);
   end;
   ccv_uid = new;
   old = mapping_uid;
   rc = h.find();
   mapping_uid = new;

   drop rc old new;
   run;

   %if &ccv_abort = 1 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, import_missing_ccv, noquote, &missing_ccv));
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason, update=true);
      %return;
   %end;

   data varprj_m.model; set  varprj_m.model;
      if _N_ = 1 then do;
               declare hash h (dataset: "work.map");
               rc =h.definekey('old');
               rc = h.definedata('new');
               h.definedone();
               call missing(old, new);
       end;
   model_uid = &model_uid;
   model_name = "&project_name";
   version_uid = &version_uid;

   old = CONSORTIUM_CRITERIA_UID;
   rc = h.find();
   CONSORTIUM_CRITERIA_UID = new;

   old = EXTERNAL_CRITERIA_UID ; 
   rc = h.find();
   EXTERNAL_CRITERIA_UID = new;

   old = EXT_SCALE_COMPONENT_UID ; 
   rc = h.find();
   EXT_SCALE_COMPONENT_UID = new;

   old = INTERNAL_CRITERIA_UID ;
    rc = h.find();
   INTERNAL_CRITERIA_UID = new;

   old = INTERNAL_SCALE_ADJUSTMENT_UID ;
   rc = h.find(); 
   INTERNAL_SCALE_ADJUSTMENT_UID = new;

   drop old new rc;
   run;
   
   data varprj_m.model_kri_usage; set  varprj_m.model_kri_usage;
   model_uid = &model_uid;
   run;

   data varprj_m.model_kri_value; set  varprj_m.model_kri_value;
   model_uid = &model_uid;
   run;
   
   data varprj_m.normal_mixture_criteria; set  varprj_m.normal_mixture_criteria;
      model_uid = &model_uid;
   run;

   data varprj_m.project_scenarios; set  varprj_m.project_scenarios;
      if _N_ = 1 then do;
               declare hash h (dataset: "work.map");
               rc =h.definekey('old');
               rc = h.definedata('new');
               h.definedone();
               call missing(old, new);
       end;
   model_uid = &model_uid;
   old = scenario_UID ;
   rc = h.find(); 
   scenario_UID = new;
   drop old new rc;
   run;

   data varprj_m.simmix_priority; set  varprj_m.simmix_priority;
   model_uid = &model_uid;
   run;

   data varprj_m.var_audit; set  varprj_m.var_audit;
   model_uid = &model_uid;
   run;

   data varprj_m.var_criteria; set  varprj_m.var_criteria;
   model_uid = &model_uid;
   run;


   
   /* process all of the criteria data sets ----------------------------*/
   %let consortium_criteria_uid=100;
   %let external_criteria_uid=101;
   %let internal_criteria_uid=102;
   %let internal_scale_adjustment_uid=103;
   %_orv_process_criteria(vardata=VarData, criteria=consortium_criteria, uid=CONSORTIUM_CRITERIA_UID, default=100 );
   %_orv_process_criteria(vardata=VarData, criteria=external_criteria, uid=EXTERNAL_CRITERIA_UID, default=101 );
   %_orv_process_criteria(vardata=VarData, criteria=internal_criteria, uid=INTERNAL_CRITERIA_UID, default=102 );
   %_orv_process_criteria(vardata=VarData, criteria=internal_scale_adjustment, uid=INTERNAL_SCALE_ADJUSTMENT_UID, default=103 );
   
   proc sql noprint;
    Insert into VARDATA.MODEL_REFS (MODEL_UID, CONSORTIUM_CRITERIA_UID, INTERNAL_SCALE_ADJUSTMENT_UID, EXTERNAL_CRITERIA_UID, INTERNAL_CRITERIA_UID, VERSION_UID) 
    values (&model_uid, &consortium_criteria_uid, &internal_scale_adjustment_uid, &external_criteria_uid, &internal_criteria_uid, &version_uid);
   quit;
   
   
   /* update the Audit table ----------------------------*/
   %orvdb_seq_incrementer(VarData, Audit_SEQ, var_audit, INC=1);
   proc sql noprint;
         select VALUE Into :var_audit_uid From WORK.VAR_SEQUENCES Where NAME = 'var_audit';
   quit;
   data _null_; call symput("time",datetime());run;
   proc sql noprint;
         Insert into varprj_m.Var_audit (VAR_AUDIT_UID, MODEL_UID, table_name, field_name, record, new_value, var_user_uid, audit_time, reason_for_change, audit_type, description) 
         values (&var_audit_uid, &model_uid, "", "", "", "", &user_uid, &time, "Project Import",' ', "Project imported"); 
   quit;
   
   
   /*  peform the key mappings -------------*/
   
   %_orv_map_keys( source_ds=insurance, source_lib = org, uid=insurance_uid,keys=insurance_name, 
                   update_ds=rawlossi lossi insurance_prop insure_1 insure_2, uid_name =insurance_1 insurance_2 );
   %_orv_map_keys( source_ds=internal_businessline, source_lib = org, uid=internal_bl_uid,
                   keys=internal_bl_name internal_bl_level, 
                   update_ds=exposurei freqcounts_i lossc lossi missing_range model_bl_access rawlossc rawlossi, uid_name =internal_bl_uid);
   %_orv_map_keys( source_ds=internal_risk_category, source_lib = org, uid=internal_rc_uid,
                   keys=internal_er_name internal_er_level, update_ds=rawlossi lossi  missing_range, uid_name =internal_rc_uid);
   %_orv_map_keys( source_ds=basel_businessline_c, source_lib = org, uid=basel_bl_uid,keys=basel_bl_name basel_bl_level, 
                   update_ds=lossc rawlossc, uid_name =basel_bl_uid);
   %_orv_map_keys( source_ds=basel_risk_category_c, source_lib = org, uid=basel_er_uid,keys=basel_er_name basel_er_level, 
                   update_ds=lossc rawlossc, uid_name =basel_er_uid); 
   %_orv_map_keys( source_ds=basel_businessline_e, source_lib = org, uid=basel_bl_uid,keys=basel_bl_name basel_bl_level, 
                   update_ds=lossp rawlossp, uid_name =basel_bl_uid);
   %_orv_map_keys( source_ds=basel_risk_category_e, source_lib = org, uid=basel_er_uid,keys=basel_er_name basel_er_level, 
                   update_ds=lossp rawlossp, uid_name =basel_er_uid); 
   %_orv_map_keys( source_ds=scenario, source_lib = org, uid=scenario_uid,keys=scenario_name, 
                   update_ds=project_scenarios raresim simmix_priority, uid_name =basel_er_uid); 
   %_orv_map_keys( source_ds=country_i, source_lib = org, uid=country_uid,keys=country_name, 
                   update_ds=rawlossi lossi , 
                   uid_name=country_of_legal_entity_uid country_of_incident_uid);

   /* Fix CCVUID_* values -----------------------------------------------------*/
   libname _comp "&path./components";
   %_orv_fix_ccvuid( );

   /* Delete CELL_WHERE, it will be recreated -----------------------------------*/
   proc datasets lib=varprj_m noprint;
   delete cell_where;
   run;quit;
   
   /* Delete temp directories  */
   %orvu_delete_dir( lib=org_old);
   libname _meta "&path./meta";
   %orvu_delete_dir( lib=_meta);
  
   %orvu_delete_dir( lib=_comp);
   libname _covmat "&path./covmat";
   %orvu_delete_dir( lib=_covmat);
   libname _losses "&path./losses";
   %orvu_delete_dir( lib=_losses);

   %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
   %orva_status_ds( statusds=&outStatus, status=0, reason=&reason, update=true );


%mend orvdb_import_project;


%macro _orv_map_keys( 
						source_ds=, 
						source_lib =, 
						uid=,
						keys=, 
						update_ds=, 
						uid_name = 
						);

   %if not %sysfunc(exist(&source_lib..&source_ds)) %then %return;

   /* Create a dataset with old uid and new uid ------------------*/
   %let uid1=%scan( &uid_name, 1);
   data newmap; set org_old.&source_ds; 
        retain missingkey 0;

        if _N_ eq 1 then do;
          declare hash Hash(dataset: "&source_lib..&source_ds");
          key1=scan("&keys",1);
          rc = Hash.defineKey(key1);
          key2=scan("&keys",2);
          if key2 ne " " then rc = Hash.defineKey(key2);       
       
          rc =Hash.defineData("&uid.");
       
          rc = Hash.defineDone(); 
        
         end;

        _old_uid = &uid;
        rc = hash.find();
        if rc ne 0 then do;
           &uid = _old_uid;
           if  missingkey = 0 then do;
           put "WARNING: Unable to map key &uid to the new data version. Some reports won't run properly. To avoid this problem";
           put "WARNING: make sure all the components needed for this project are present and re-import the project.";
           missingkey = 1;
           end;
        end;

        &uid1 = &uid;
        keep &uid1 _old_uid;
 
   run;

/* now do a loop over all the datasets and replace the key with the correct names -----------------*/
   %let nds = 1;
   %let ds   = %scan(&update_ds, &nds, " ");
   %let uid2=%scan( &uid_name, 2);
   
   %do %while(&ds NE %str());
   
   %if %sysfunc(exist(varprj_m.&ds))  %then %do;  
   %if %orvu_var_exists(&uid1, varprj_m.&ds) %then %do; 
   data varprj_m.&ds; set varprj_m.&ds;
        
        if _N_ eq 1 then do;
          declare hash Hash(dataset: "newmap");
          rc = Hash.defineKey("_old_uid"); 
          call missing( _old_uid );         
          rc = Hash.defineData("&uid1.");
          rc = Hash.defineDone(); 
          _new_uid1 = .;
         end;


       if &uid1 ^= . and &uid1 > 0 then do;
          _old_uid =&uid1;
          rc = hash.find();
          if rc ne 0 then do;
              put "ERROR: Unable to map keys for &uid1 at observation " _n_ " and data set varprj_m.&ds";
             abort;
          end;
          %if &uid2 ne %str() %then %do;
          if &uid2 ^= . and &uid2 ^= 0 then do;
             _new_uid1 = &uid1;
             _old_uid =&uid2;
             rc = hash.find();
             if rc ne 0 then do;
                put "ERROR: Mapping keys that don't exist at observation " _n_ " and data set varprj_m.&ds";
                abort;
             end;
             &uid2 = &uid1;
             &uid1 = _new_uid1;
          end;
          %end;

       end;
       drop _old_uid _new_uid1 rc;
   run;
   %end;
   %end;

   %let nds = %eval(&nds + 1);
   %let ds   = %scan(&update_ds, &nds, " ");

   %end;     

%mend _orv_map_keys;

%macro _orv_process_criteria(
 							vardata=, 
 							criteria=, 
							uid=, 
							default= 
							);

   %if not %sysfunc(exist(varprj_m.&criteria)) %then %return;

   proc sql noprint;
         select &uid Into :&uid From varprj_m.&criteria;
   quit;
   
   %if &&&uid ne &Default %then %do;  /* Add this new entry into components.consortium_criteria */
      %orvdb_seq_incrementer(&VarData, CritDataSet_SEQ, &criteria, INC=1);
      proc sql noprint;
         select VALUE Into :&uid From WORK.VAR_SEQUENCES Where NAME = '&criteria';
      quit;
      data varprj_m.&criteria; set varprj_m.&criteria;
         &uid = &&uid;
      run;
      proc append base=comp.&criteria data = varprj_m.&criteria;run;
   %end;


%mend _orv_process_criteria;

   /* Incorporate the whole path for names that aren't unique otherwise -------------------*/
%macro _orv_build_paths( var_folders, out_var_folders);

 proc sort data=&var_folders out=vf; by old_var_folders_uid;
   run;

   data &out_var_folders; set vf;
   length name $ 128 ;
   if _N_ =1 then do;
     declare hash h ();
     rc =h.definekey('old_var_folders_uid');
     rc = h.definedata('folder_name_ucase');
	 rc = h.definedata('parent_uid');
     h.definedone();
   end;

   rc = h.add();

   if folder_type in ( "Mapping" "Cov Matrices" "Filter Internal" "Filter Consortium" 
                           "Filter External" "Scale Internal" "Scale External")  then do;
      puid = parent_uid;
      vfuid = old_var_folders_uid;
      name = folder_name_ucase;
      do while (parent_uid > 100);
         old_var_folders_uid = parent_uid;
         rc = h.find();
	  name = trim(folder_name_ucase) || "/" || trim(name);
      end;
      old_var_folders_uid = vfuid;
      parent_uid = puid;
      folder_name_ucase = upcase(folder_name);
   end;

   drop puid vfuid rc;
   run;

%mend _orv_build_paths;

   /* fix all of the instances of CCVUID_* ---------------------------------*/
   /*  CCUID -> CCVUID_0 CCVUID_1 CCVUID_2   - Dataset CC                   */
   /*  CCVUID_0 -> CCMAP_0                   - Dataset CCMAPI_0             */
   /*  CCMAP_0 .. CCMAP_2 =  INTERNAL_BL_UID, INTERNAL RC_UID ...           */
   /*  CCMAP = DB_VALUE in dataset MAPPING_ITEM_INT                         */
   /*  CCV_VALUE_UID = CCVUID_* in dataset MAPPING_ITEM_INT                 */
%macro _orv_fix_ccvuid( );

   /* How many ccv are there --------------------------*/
   data _NULL_; set varprj_m.cc(obs=1);
   array cc[*] ccvuid_:;
   if dim(cc) > 2 then call symput("THIRD_DIM","True");
   else call symput("THIRD_DIM","False");
   run;

   /* determine the class variable names ------------------------------------*/
   data _NULL_;
      set varprj_m.ccmapval_0; 
      where CCVUID_0 = -899;
      call symput("CLASS_VAR_0", strip(CCVVALUE_0));
   run;
   data _NULL_;
      set varprj_m.ccmapval_0; 
      where CCVUID_0 = -999;
      call symput("FOLDER_CLASS_VAR_0", strip(CCVVALUE_0));
   run;

   data _NULL_;
      set varprj_m.ccmapval_1; 
      where CCVUID_1 = -899;
      call symput("CLASS_VAR_1", strip(CCVVALUE_1));
   run;
   data _NULL_;
      set varprj_m.ccmapval_1; 
      where CCVUID_1 = -999;
      call symput("FOLDER_CLASS_VAR_1", strip(CCVVALUE_1));
   run;
        
   %if &THIRD_DIM = True %then %do;
      data _NULL_;
         set varprj_m.ccmapval_2; 
         where CCVUID_2 = -899;
         call symput("CLASS_VAR_2", strip(CCVVALUE_2));
      run;
      data _NULL_;
         set varprj_m.ccmapval_2; 
         where CCVUID_2 = -999;
         call symput("FOLDER_CLASS_VAR_2", strip(CCVVALUE_2));
      run;
   %end;

   /* get name of org data set needed to perform the mappings ---------------------*/
   %let DS_0=%_orv_get_ds_name( &CLASS_VAR_0);
   %let DS_1=%_orv_get_ds_name( &CLASS_VAR_1);
   %if &THIRD_DIM = True %then %do;
       %let DS_2 = %_orv_get_ds_name( &CLASS_VAR_2);
       %if "&CLASS_VAR_2" = "BASEL_BL" %then %let &CLASS_VAR_2 = BASEL_BL_UID;
   %end;

   %if "&CLASS_VAR_0" = "BASEL_BL" %then %let CLASS_VAR_0 = BASEL_BL_UID;
   %if "&CLASS_VAR_1" = "BASEL_BL" %then %let CLASS_VAR_1 = BASEL_BL_UID;

   /* Get the name of the character column for doing the matching ------------------*/
   /* repair the places where the name convention isn't followed -------------------*/
   data _NULL_;
      length name $32;
      uid0 = "&CLASS_VAR_0";
      if substr(uid0,1,5) = "CLASS" then do;
         name = trim(uid0) || "_NAME";
         call symput("CLASS_VAR_0",  trim(uid0) || "_UID");
      end;
      else
         name = substr(uid0, 1, length(uid0)-3) || "NAME";
      if uid0 = "INTERNAL_RC_UID" then name = "INTERNAL_ER_NAME";
      
      if uid0 = "BASEL_RC" then do;
          call symput("CLASS_VAR_0", "BASEL_ER_UID");
          name = "BASEL_ER_NAME";
      end;
      call symput("CLASS_VAR_NAME0", trim(name));
    
      uid1 = "&CLASS_VAR_1";
      if substr(uid1,1,5) = "CLASS" then do;
         name = trim(uid1) || "_NAME";
         call symput("CLASS_VAR_1",  trim(uid1) || "_UID");
      end;
      else
         name = substr(uid1, 1, length(uid1)-3) || "NAME";
      if uid1 = "INTERNAL_RC_UID" then name = "INTERNAL_ER_NAME";
      if uid1 = "BASEL_RC" then do;
          call symput("CLASS_VAR_1", "BASEL_ER_UID");
          name = "BASEL_ER_NAME";
      end;
      call symput("CLASS_VAR_NAME1", trim(name));

    %if &THIRD_DIM = True %then %do;
      uid2 = "&CLASS_VAR_2";
      if substr(uid2,1,5) = "CLASS" then do;
         name = trim(uid2) || "_NAME";
         call symput("CLASS_VAR_2",  trim(uid2) || "_UID");
      end;
      else
         name = substr(uid2, 1, length(uid2)-3) || "NAME";
      if uid2 = "INTERNAL_RC_UID" then name = "INTERNAL_ER_NAME";
      if uid2 = "BASEL_RC" then do;
          call symput("CLASS_VAR_2", "BASEL_ER_UID");
          name = "BASEL_ER_NAME";
      end;
      call symput("CLASS_VAR_NAME2", trim(name));  
    %end; 
   run;  

   /* First create maps to correct the class variables used - CCMAP -----------------------*/
   proc sql noprint nowarn;
   create table map0 as select new.&CLASS_VAR_0, old.old from org.&DS_0  New,
         org_old.&DS_0(rename=(&CLASS_VAR_0=old))  old
         where old.&CLASS_VAR_NAME0 = new.&CLASS_VAR_NAME0 ;
   create table map1 as select new.&CLASS_VAR_1, old.old from org.&DS_1  New,
         org_old.&DS_1(rename=(&CLASS_VAR_1=old))  old
         where old.&CLASS_VAR_NAME1 = new.&CLASS_VAR_NAME1 ;
   %if &third_dim = True %then %do;
   create table map2 as select new.&CLASS_VAR_2, old.old from org.&DS_2  New,
         org_old.&DS_2(rename=(&CLASS_VAR_2=old))  old
         where old.&CLASS_VAR_NAME2 = new.&CLASS_VAR_NAME2 ;
   %end;
   quit; 

            /* Fix the CCMAP values ------------------------------------*/
   data map0; set map0; rename &CLASS_VAR_0=CCMAP_0; run;
   data map1; set map1; rename &CLASS_VAR_1=CCMAP_1; run;  
   %_orv_fix_var( CCMAP_0, map0, old);
   %_orv_fix_var( CCMAP_1, map1, old);
   %if &THIRD_DIM = True %then %do;
   data map2; set map2; rename &CLASS_VAR_2=CCMAP_2; run;
   %_orv_fix_var( CCMAP_2, map2, old);
   %end;

        /* Get the ccv_uid numbers -------------------------------------*/
   proc sql noprint nowarn;
   select ccv_uid into :CCV_UID separated by " " from varprj_m.model_ccv_mapping order by order_num;
   quit;


   /* Create a hash map table to fix CCVUID_* for internal data -------------------------*/
   proc sql noprint nowarn;
   create table map_ccvuid0 as select new.ccvuid_0, old.old from 
         comp.ccv_values(rename=(ccv_value_uid=ccvuid_0))  New,
         _comp.ccv_values(rename=(ccv_value_uid=old)) old
         where old.ccv_name = new.ccv_name and new.ccv_uid=%scan(&CCV_UID, 1);
   create table map_ccvuid1 as select new.ccvuid_1, old.old from 
         comp.ccv_values(rename=(ccv_value_uid=ccvuid_1))  New,
         _comp.ccv_values(rename=(ccv_value_uid=old)) old
         where old.ccv_name = new.ccv_name and new.ccv_uid=%scan(&CCV_UID, 2);
   %if &THIRD_DIM = True %then %do;
   create table map_ccvuid2 as select new.ccvuid_2, old.old from 
         comp.ccv_values(rename=(ccv_value_uid=ccvuid_2))  New,
         _comp.ccv_values(rename=(ccv_value_uid=old)) old
         where old.ccv_name = new.ccv_name and new.ccv_uid=%scan(&CCV_UID, 3);
   %end;
   quit;

   /* fix the cc table and others -------------------*/
   %_orv_fix_var( CCVUID_0, map_ccvuid0, old);
   %_orv_fix_var( CCVUID_1, map_ccvuid1, old);
   %if &THIRD_DIM = True %then %do;
   data map_ccvuid2;  set map_ccvuid2;
      ccvuid_2 = input(ccvuid_2_c, 8.0);
      old = input(old_c, 8.0);
      drop old_c ccvuid_2_c;
   run;
   %_orv_fix_var( CCVUID_2, map_ccvuid2, old);
   %end;

   /*Fix capital allocation tab ----------------------------------------------*/
   %let CCV3=0;
   proc sql noprint;
   select var_folders_uid into :CCV1 from vardata.var_folders
      where folder_type = "CCV" and org_structure_uid=&version_uid
            and folder_name="&FOLDER_CLASS_VAR_0";
   select var_folders_uid into :CCV2 from vardata.var_folders
      where folder_type = "CCV" and org_structure_uid=&version_uid
            and folder_name="&FOLDER_CLASS_VAR_1";
 %if &THIRD_DIM = True %then %do;
   select var_folders_uid into :CCV3 from vardata.var_folders
      where folder_type = "CCV" and org_structure_uid=&version_uid
            and folder_name="&FOLDER_CLASS_VAR_2";
 %end;
   quit;

   data varprj_m.var_criteria; set  varprj_m.var_criteria;
      CAPITAL_DIST_CCV1 = &ccv1;
      CAPITAL_DIST_CCV2 = &ccv2;
      CAPITAL_DIST_CCV3 = &ccv3;
   run;
   

%mend;

%macro _orv_get_ds_name( CLASS_VAR);

   %if "&CLASS_VAR" = "INTERNAL_BL_UID" %then internal_businessline;
   %if "&CLASS_VAR" = "INTERNAL_RC_UID" %then internal_risk_category;
   %if "&CLASS_VAR" = "LEGAL_ENTITY_UID" %then legal_entity;
   %if "&CLASS_VAR" = "BASEL_BL"        %then basel_businessline_i;
   %if "&CLASS_VAR" = "BASEL_RC"        %then basel_risk_category_i;
   %if "&CLASS_VAR" = "CLASS_1"         %then class_1;
   %if "&CLASS_VAR" = "CLASS_2"         %then class_2;
   %if "&CLASS_VAR" = "CLASS_3"         %then class_3;
   %if "&CLASS_VAR" = "CLASS_4"         %then class_4;
   %if "&CLASS_VAR" = "CLASS_5"         %then class_5;
   %if "&CLASS_VAR" = "CLASS_6"         %then class_6;
   %if "&CLASS_VAR" = "CLASS_7"         %then class_7;
   %if "&CLASS_VAR" = "CLASS_8"         %then class_8;
   %if "&CLASS_VAR" = "CLASS_9"         %then class_9;
   %if "&CLASS_VAR" = "CLASS_10"        %then class_10;
%mend;


%macro _orv_fix_var( variable, mapds, map_from);

   /* get a list of datsets in varprj_m that contain variable---------------------------*/
   proc contents data=varprj_m._ALL_ out=c noprint;
   run;

   data c; set c;
   where upcase(name)=upcase("&variable");
   keep memname;
   run;

   proc sql noprint;
   select memname into :dataset_list separated by " " from c;
   run;
  

   %let nd = 1;
   %let dataset  = %scan(&dataset_list, &nd, ' ');
   
   %do %while( &dataset ne %str()  );

        data varprj_m.&dataset; set varprj_m.&dataset;
	     if _N_ =1 then do;
		
		declare hash h(dataset: "&mapds");
		rc = h.definekey("&map_from");
		rc = h.definedata("&variable");
		call missing( &map_from ); 
		h.definedone();

	     end;

            &map_from = &variable;
	    if h.find() ne 0  then   /* Not found so Restore value */
		&variable = &map_from;
       
        drop rc &map_from;
        run;

        %let nd = %eval(&nd + 1);
        %let dataset  = %scan(&dataset_list, &nd, ' ');
   %end;
  

%mend;
