/* --------------------------------------------------------------
|  Copyright SAS Institute 2012
|  Name: orvdb_import_version
|  Purpose: Imports a project from a another data store to a new data version
|  Inputs:
|      data_store_root  - The fully qualified directory name for the root of the data store
|      user_id          - A valid user_id already in the data store - sasdemo
|      version_name     - The name of the data version we are adding this project to
|      project_name     - The name of the project we are adding to the version - can be different from the name exported
|      exported_project - The fully qualified path to the zip file that represents the exported project
|
|  Example:
|    %orvdb_import_version( data_store_root=c:\opriskvar\9.3\, user_id=sasdemo, version_name=inc2, project_name=copy1,
|                          exported_project= c:\temp\foo.zip );
--------------------------------------------------------------------*/ 
%macro orvdb_import_version(data_store_root=,
                             user_id=,
                             version_name=,
                             project_name=,
                             exported_project=,
                             dir_suffix=0,
                             outStatus      =ostatus);
                                                    
    /*- create the status data set up front -*/
   %let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );


   /* create the vardata libref -------------------------------*/
   %if %sysfunc(libref(vardata)) ne 0 %then %do;
      %let rc =%sysfunc(libname( vardata, &data_store_root/main));
      %if &rc > 0 %then %do;
         %let reason=%sysfunc(sasmsg(&orvads, data_store_path_error, noquote));
         %put &reason;
         %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
         %return;
      %end;
   %end;

   /* check to see if name is unique -------------------------------------------------------- */
   %local version_uid user_id version_path;
   %let version_uid=;
   proc sql noprint;
    select version_uid into :version_uid from vardata.data_version where version_name = "&version_name" ;
    select var_user_uid into :user_uid from vardata.var_user where user_id="&user_id";
   quit;
   %if %str() ^= &version_uid %then %do;
       %let reason=%sysfunc(sasmsg(&orvads, version_name_not_unique, noquote));
       %put &reason;
       %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );  
       %return; 
   %end;



   /* get the last version number ----------------------------*/
   %if &DIR_SUFFIX = 0 %then %do;
      proc sql noprint;
      select max(dir_suffix) into :dir_suffix from vardata.data_version;
      quit;
      %let dir_suffix=%eval(&dir_suffix + 1);
   %end;

   /* create the new directory for the version ---------------------*/
   %let version_root=&data_store_root/ver_&dir_suffix/;
   %let returncode=;
   %_orv_create_dir( dest= &version_root, destlib=, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return; 
   
   /* create all the sub directories -----------------------------*/
   %_orv_create_dir( dest= &version_root/components, destlib=comp, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return;
   %_orv_create_dir( dest= &version_root/components/projects, destlib=, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return;
   %_orv_create_dir( dest= &version_root/covmat, destlib=covmat, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return;
   %_orv_create_dir( dest= &version_root/loss, destlib=, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return;
   %_orv_create_dir( dest= &version_root/loss/losses, destlib=losses, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return;
   %_orv_create_dir( dest= &version_root/loss/org, destlib=org, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return;
   %_orv_create_dir( dest= &version_root/meta, destlib=meta, return=returncode, outStatus=&outstatus );
   %if &returncode = 1 %then %return; 


   /* unzip the project ---------------------------------------------------------------------*/
   /* get a unique key to create the model_uid --------------------------------------- */
   %orvdb_seq_incrementer(VarData, Folder_SEQ, var_folders, INC=4);
   proc sql noprint;
      select VALUE Into :model_uid From WORK.VAR_SEQUENCES Where NAME = 'var_folders';
   quit;
  
   %let model_uid=%eval(&model_uid);
   
   /* sets &project_name_short */
   %let project_name_short=%str();
   %orvu_project_name_get(project_name = &project_name.);

   %let path= &data_store_root/ver_&dir_suffix/components/projects/&project_name_short&Model_uid;  

   /* unzip the files into that folder */
   %orvu_unzip( &exported_project, &path);
   %if &syserr ne 0 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, export_file_error, &path, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;

   /* create the varprj_m libref -------------------------------*/
   %let rc =%sysfunc(libname( varprj_m, &path));
   %if &rc > 0 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
 

   /* update the sequence keys so that we don't get collisions later ------------------------ */
   /* updates where stored in org_old directory -----------------------------------------*/
   %let rc =%sysfunc(libname( org_old, &path/org));
   %if &rc > 0 %then %do;
      %put %sysfunc(sysmsg()) ;
      %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
      %put &reason;
      %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
      %return;
   %end;
   data vardata.var_sequences; set vardata.var_sequences; set org_old.var_sequences(rename=(seq_count=input_seq_count));
   seq_count = max(seq_count,input_seq_count);
   drop input_seq_count;
   run;

   /* Create the folder object for the new version ----------------*/
   /* get the old version number and add the dir_suffix to it -------------------------------------- */
   %orvdb_seq_incrementer(VarData, Folder_SEQ, var_folders, INC=1);
   proc sql noprint;
      select VALUE Into :version_uid From WORK.VAR_SEQUENCES Where NAME = 'var_folders';
   quit;
   
     /* Update the model_uid ----------------------------------------*/
   data varprj_m.loss_event_audit_for_model; set  varprj_m.loss_event_audit_for_model;
   model_uid = &model_uid;
   run;

   data varprj_m.model; set  varprj_m.model;
   model_uid = &model_uid;
   model_name = "&project_name";
   run;
   
   data varprj_m.model_bl_access; set  varprj_m.model_bl_access;
   model_uid = &model_uid;
   var_user_uid = &user_uid;
   run;
   
   data varprj_m.model_ccv_mapping; set  varprj_m.model_ccv_mapping;
   model_uid = &model_uid;
   run;
   
   data varprj_m.model_kri_usage; set  varprj_m.model_kri_usage;
   model_uid = &model_uid;
   run;

   data varprj_m.model_kri_value; set  varprj_m.model_kri_value;
   model_uid = &model_uid;
   run;
   
   data varprj_m.normal_mixture_criteria; set  varprj_m.normal_mixture_criteria;
      model_uid = &model_uid;
   run;

   data varprj_m.project_scenarios; set  varprj_m.project_scenarios;
   model_uid = &model_uid;
   run;

   data varprj_m.simmix_priority; set  varprj_m.simmix_priority;
   model_uid = &model_uid;
   run;

   data varprj_m.var_audit; set  varprj_m.var_audit;
   model_uid = &model_uid;
   run;

   data varprj_m.var_criteria; set  varprj_m.var_criteria;
   model_uid = &model_uid;
   run;

   /* Copy and modify rows for vardata --------------------------------------------------*/
   
  
   data org_old.data_version;set org_old.data_version; 
        org_structure_uid = &version_uid; version_uid = &version_uid;
		version_name="&version_name";
		dir_suffix=&dir_suffix;
   run;
   proc append base=vardata.data_version data=org_old.data_version;run;

   data _null_; set org_old.data_version;
       call symput("internal", status_internal_data);
       call symput("external", status_external_data); 
       call symput("consortium", status_consortium_data);
   run;
   data load_information;  
        length data_type $ 1;
        if &internal = 1 then do;
           load_information_uid = &version_uid;
           version_uid = &version_uid;
           load_date = datetime();
           var_user_uid = &user_uid;
           data_type = "I";
           output;
        end;
        if &external = 1 then do;
           load_information_uid = &version_uid+1;
           version_uid = &version_uid;
           load_date = datetime();
           var_user_uid = &user_uid;
           data_type = "E";
           output;
        end;
        if &consortium = 1 then do;
           load_information_uid = &version_uid+2;
           version_uid = &version_uid;
           load_date = datetime();
           var_user_uid = &user_uid;
           data_type = "C";
           output;
        end;
   run;
   proc append base=vardata.load_information data=load_information force;run;

   data org_old.model_ccv_mapping_refs; set org_old.model_ccv_mapping_refs;
        model_uid = &model_uid;
   run;
   proc append base=vardata.model_ccv_mapping_refs data=org_old.model_ccv_mapping_refs;run;

   data org_old.model_refs; set org_old.model_refs;
        version_uid = &version_uid; model_uid = &model_uid;
   run;
   proc append base=vardata.model_refs data=org_old.model_refs;run;
 
   data org_old.project_insurance_refs; set org_old.project_insurance_refs;
        version_uid = &version_uid; model_uid = &model_uid;
   run;
   proc append base=vardata.project_insurance_refs data=org_old.project_insurance_refs;run;

   data org_old.project_scenario_refs; set org_old.project_scenario_refs;
        version_uid = &version_uid; model_uid = &model_uid;
   run;
   proc append base=vardata.project_scenario_refs data=org_old.project_scenario_refs;run;

   data org_old.Scenario_ccv_refs ; set org_old.Scenario_ccv_refs ;
        version_uid = &version_uid;
   run;
   proc append base=vardata.Scenario_ccv_refs  data=org_old.Scenario_ccv_refs ;run;

   

   /* Fix the keys in the var_folders data set ---------------------------*/
   /* get the old version uid -------------------------------------------  */
   proc sql noprint;
      select org_structure_uid Into :old_org_structure_uid From org_old.var_folders 
            Where folder_type = 'Version';
   quit;

   %orvdb_seq_incrementer(VarData, Folder_SEQ, var_folders, INC=%orvu_attrn(org_old.var_folders, NOBS)) ;
   proc sql noprint;
      select VALUE Into :new_uid From WORK.VAR_SEQUENCES Where NAME = 'var_folders';
   quit;

      /* make a hashmap so we can easily fix the folder tree -----------------*/
   data map; set org_old.var_folders;
         if _n_ = 1 then do; new =0; old=0; output; end;
         new = &new_uid + _N_ -1;
         old = var_folders_uid;
	 if folder_type = "Version" then new = &version_uid;
         if folder_type = "Project" then new = &model_uid;
         if old < 105 then new = old;
	 output;

         keep new old;
   run;

   %let old_org_structure_uid=%eval(&old_org_structure_uid);
   %let version_uid=%eval(&version_uid);
   
   data var_folders ; set org_old.var_folders ; 
       
        if _N_ = 1 then do;
               declare hash h (dataset: "work.map");
               rc =h.definekey('old');
               rc = h.definedata('new');
               h.definedone();
               call missing(old, new);
        end;
        if parent_uid = &old_org_structure_uid then parent_uid = &version_uid;
		else 
		if parent_uid > 80 then do;  /* new parent look up new name */
                   old = parent_uid;
                   rc = h.find();
                   parent_uid = new;
		end;

        if var_folders_uid < 100 or var_folders_uid > 104 then
	   var_folders_uid = &new_uid + _N_ -1;

	if folder_type = "Version" then do; 
             folder_name = "&version_name";
	     folder_name_ucase = upcase("&version_name");
	     var_folders_uid = &version_uid;
             parent_uid = 45;               /* Incremental versions become full versions */
	end;
	else 
    if folder_type = "Project" then do; 
             folder_name = "&project_name";
			 folder_name_ucase = upcase("&project_name");
			 var_folders_uid = &model_uid;
	end;
	if folder_type = "Folder - Version" then org_structure_uid = 0;
	else org_structure_uid = &version_uid; 

	if folder_name_ucase = "VERSIONNAME.TXT [&old_org_structure_uid]" then do;
	   folder_name_ucase = "VERSIONNAME.TXT [&version_uid]";
	   folder_name = "versionName.txt [&version_uid]";
	end;

        user_uid = &user_uid;
        drop new old rc;
   run;
   proc append base=vardata.var_folders  data=var_folders ;run;

   /* copy and make changes to the version specific folders and files -------------------------------------*/
   %let returncode=;
   %_orv_copy_dir2( source=&path/meta, destlib=meta, sourcelib=_meta, return=returncode, outStatus=&outStatus );
   %if &returncode = 1 %then %return;  

   %let returncode=;
   %_orv_copy_dir2( source=&path/components, destlib=comp, sourcelib=_comp, return=returncode, outStatus=&outStatus );
   %if &returncode = 1 %then %return;  

   %let returncode=;
   %_orv_copy_dir2( source=&path/losses, destlib=losses, sourcelib=_losses, return=returncode, outStatus=&outStatus );
   %if &returncode = 1 %then %return;  
   data losses.internal_loss_event; set losses.internal_loss_event;
       version_uid = &version_uid;
        if _N_ = 1 then do;
               declare hash h (dataset: "work.map");
               rc =h.definekey('old');
               rc = h.definedata('new');
               h.definedone();
               call missing(old, new);
        end;
        old = insurance_1;
        rc = h.find();
        insurance_1 = new;
        old = insurance_2;
        rc = h.find();
        insurance_2 = new;
   run;
   data losses.consortium_loss_event; set losses.consortium_loss_event;
       version_uid = &version_uid;
   run;
   data losses.external_loss_event; set losses.external_loss_event;
       version_uid = &version_uid;
   run;

   %let returncode=;
   %_orv_copy_dir2( source=&path/covmat, destlib=covmat, sourcelib=_covmat, return=returncode, outStatus=&outStatus );
   %if &returncode = 1 %then %return;  

      /* org has other stuff in it that we don't want to copy ---------------------------------------*/
   proc datasets library=org_old memtype=data noprint;
	copy out=org constraint=no;
        exclude var_sequences data_version Model_ccv_mapping_refs Model_refs project_insurance_refs
              project_scenario_refs var_folders;
   run;quit;

     /* fix business line access for sole user_id ---------------------------------*/
   data org.businessline_access; set org.businessline_access;
       var_user_uid = &user_uid;
       data_version_uid = &version_uid;
   run;


   /* delete unneeded dirs --------------*/
   %orvu_delete_dir( lib=org_old);
   %orvu_delete_dir( lib=_meta);
   %orvu_delete_dir( lib=_comp);
   %orvu_delete_dir( lib=_covmat);
   %orvu_delete_dir( lib=_losses);

   /* fix remaining data sets -------------------------------------*/
   %_orv_fix_remaining_data_sets();

   %if %sysfunc(exist(meta.external_metadata)) %then %do;
   data meta.external_metadata; set meta.external_metadata;
        version_uid = &version_uid;
   run;
   %end;
   data meta.internal_metadata; set meta.internal_metadata;
        version_uid = &version_uid;
   run;
   %if %sysfunc(exist(meta.consortium_metadata)) %then %do;
   data meta.consortium_metadata; set meta.consortium_metadata;
        version_uid = &version_uid;
   run;
   %end;

   /* check if the libref was defined already so we know whether we need to clear it */
   %if (%sysfunc(libref(vc&DIR_SUFFIX))) %then %let pre_assigned=False;
   %else %let pre_assigned=True;

   %global ORVA_IS_TABLE_SERVER;
   %let ORVA_IS_TABLE_SERVER=0;
   %orvdb_init_ds_globals_ver(IS_TABLE_SERVER=0,
                              IS_DRIVER=true,
                              DIR_SUFFIX=&DIR_SUFFIX);
  
   /* Setup needed librefs for OpRiskVaR data areas */
   /* only if share isn't being used ---------------*/
   %let ERROR_STATUS=0;
   %if &pre_assigned = False %then %do;
      %orvdb_libname_setup(DATA_ROOT=&DATA_STORE_ROOT, IS_CURRENT_LIBREFS=1,
                           DIR_SUFFIX=&DIR_SUFFIX, ERROR_STATE=ERROR_STATUS);
      %if &ERROR_STATUS ne 0 %then %return;
   %end;

   proc datasets library=work memtype=data noprint;
   delete pktest;
   run;quit;

   %orvdb_primary_key_input();
   
   %orvdb_primary_key_prj_comps();

   %orvdb_foreign_key_input(DO_CONS=true);

   %orvdb_foreign_key_prj_comps(DO_CONS=true);

   /* free the librefs  ----------------------------------*/
   libname comp;
   libname covmat;
   libname losses;
   libname org;
   libname meta;
   libname varprj_m;
   libname org_old;
   %if &pre_assigned = False %then %do; 
      libname &ORVAR_VARCOMP;
      libname &ORVAR_VARORG;
      libname &ORVAR_VARLOSS;
      libname &ORVAR_VARMETA;
      libname &ORVAR_VARCOV;
   %end;


   %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
   %orva_status_ds( statusds=&outStatus, status=0, reason=&reason, update=true );   

  
 %mend orvdb_import_version; 


  /* Create directory 'Dest' and check for success and create a libname to new directory */
%macro _orv_create_dir( dest=, destlib=, return=, outStatus= );

	%let &return=0;
	%orvu_mkdirs( path= &dest);

	/* check for success --------------------------*/
	%if &destlib = %str() %then %let outlib = tmp;
	%else %let outlib=&destlib;

	%let rc =%sysfunc(libname( &destlib, &dest));
	%if &rc > 0  %then %do;
	   %put %sysfunc(sysmsg()) ;
	   %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
	   %put &reason;
	   %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
	   %let &return=1;
	   %return;
	%end;
 
%mend _orv_create_dir;

  /* Create libname source  copy source data sets to dest */
%macro _orv_copy_dir2( source=, destlib=, sourcelib=, return=, outStatus= );

	%let rc =%sysfunc(libname( &sourcelib, &source));
	%if &rc > 0 %then %do;
	   %put %sysfunc(sysmsg());
	   %let reason=%sysfunc(sasmsg(&orvads, creation_of_version_dir_error, noquote));
	   %put &reason;
	   %orva_status_ds( statusds=&outStatus, status=1, reason=&reason );
	   %let &return=1;
	   %return;
	%end;

	proc datasets library=&sourcelib memtype=data noprint;
	copy out=&destlib constraint=no;
	run;quit;

%mend _orv_copy_dir2;

     /* go through the remaining data sets and fix the ids -----------------*/
%macro _orv_fix_remaining_data_sets();
        %local keys cols;
        %let keys=
	scenario_uid
	scenario_event_uid 
	insurance_uid
        insurance_1
        insurance_2
	ccv_uid
	cov_matrix_uid
	ext_scale_component_uid
	int_scale_component_uid
	internal_criteria_uid 
	external_criteria_uid 
	consortium_criteria_uid 
	mapping_uid
	version_uid
	capital_dist_ccv1
	capital_dist_ccv2
	capital_dist_ccv3
        ;

        proc contents data=org._all_ out=orgds(keep=libname memname memtype nobs name) noprint;
	run;
        proc contents data=comp._all_ out=compds(keep=libname memname memtype nobs name) noprint;
	run;
        proc contents data=covmat._all_ out=covmatds(keep=libname memname memtype nobs name) noprint;
	run;
	proc contents data=varprj_m._all_ out=varprjds(keep=libname memname memtype nobs name) noprint;
	run;
        data allds; set orgds compds covmatds varprjds; 
        where memtype='DATA' and nobs > 0;
        run;
       
        data _NULL_;
              rc=filename("tmp",,"temp");
        run;

        data _NULL_; set allds;
             file tmp;
                            /* loop over the keys and update them when found */
              rc = findw("&keys", trim(lowcase(name)), ' ');
              if rc ne 0 then do;
                  put '%_orv_convert_keys(' libname ',' memname ',' name ');';        
              end;
        run; 
        %inc tmp;

		     /* Handle a particular case with scenarios ------------------*/
		%let cols=BASEL_BL_UID BASEL_RC_UID CLASS_1 CLASS_2 CLASS_3 CLASS_4 CLASS_5 CLASS_6 CLASS_7 
                 CLASS_8 CLASS_9 CLASS_10 INTERNAL_BL_UID INTERNAL_RC_UID LEGAL_ENTITY_UID PROCESS_UID;    

		data org.scenario_details; set org.scenario_details;
             if _N_ = 1 then do;
               declare hash h (dataset: "work.map");
               rc =h.definekey('old');
               rc = h.definedata('new');
               h.definedone();
               call missing(old, new);
             end;

			 /* fix the rest of the columns ------------------------------ */			 
			 
			 if type_cd eq 'V' then do;
			 %do i = 1 %to 16;
			 new = .;
             old = %scan(&cols,&i);
             rc = h.find();
             %scan(&cols,&i) = new;
			 %end;
			 end;

			 if type_cd ne 'G' and type_cd ne 'E' then do;
                old = parent_uid;
                rc = h.find();
                parent_uid = new;
			 end;
             drop old new rc;
        run;
%mend;

%macro _orv_convert_keys( libname , memname , name );

        data &libname..&memname; set &libname..&memname;
             if _N_ = 1 then do;
               declare hash h (dataset: "work.map");
               rc =h.definekey('old');
               rc = h.definedata('new');
               h.definedone();
               call missing(old, new);
             end;
             old = &name;
             rc = h.find();
             &name = new;
             drop old new rc;
        run;

%mend;
