/*****************************************************************************************
 * Copyright (C), 2006
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orvdb_migrate_projects.sas                                             
 * TITLE:   Migrates project data sets.
 *
 * SUPPORT: mimuha
 *
 * PARAMETERS:
 *
 *     DATA_ROOT       = fullpath name of the version folder created for OpRiskVaR 3.3 data tables.
 *
 *     OLD_ROOT        = fullpath name of the root folder of the old OpRiskVaR 3.2 data tables.
 * 
 *     VAR_FOLDERS_UID = The uid of the project to be migrated.
 * 
 *     FOLDER_NAME     = The name of the project to be migrated.  The directory name of
 *                       the project is a combination of FOLDER_NAME and VAR_FOLDERS_UID.
 * 
 *     NEW_VERSION_UID = The new version uid.
 *
 * Examples:
 * 
 *   %orvdb_migrate(data_root=C:\VarTables33\ver_1, OLDDATA_ROOT=C:\VarTables, 1202, Project 1);
 *   %orvdb_migrate(data_root=/u/root/VarTables33/ver_2, OLDDATA_ROOT=/u/root/VarTables 1202, Project 1);
 *
 *****************************************************************************************/
%macro orvdb_migrate_projects(DATA_ROOT, OLD_ROOT, VAR_FOLDERS_UID, FOLDER_NAME, NEW_VERSION_UID);

   %if %index(WIN DNTHOST,&sysscp) %then %do;
      %let SEP = "\";
   %end;
   %else %do;
      %let SEP = "/";
   %end;
 
   /*****- create USERPATH Macro variable -*****/
   data _NULL_;
      length path $ 256;
      path= "&DATA_ROOT" || &SEP || "components" || &SEP || "projects";
      call symput("PROJDIRPATHNEW", path);
      path= "&OLD_ROOT" || &SEP || "components" || &SEP || "projects";
      call symput("PROJDIRPATHOLD", path);
   run;
   
   /*****- create old project directory -*****/
   data _null_;
      length prjpath $ 256;
      prjpath = "&FOLDER_NAME";

      /* truncate to 20 characters */
      if lengthc(trim(left(prjpath)))>20 then do;
         short_name = substr(trim(left(prjpath)), 1, 20);
      end;
      else do;
         short_name = trim(left(prjpath));
      end;

      /* append uid for uniqueness */ 
      prjname = translate(trim(left(short_name)),'_',' '); 
      prjname = trim(left(prjname)) || strip(put(&VAR_FOLDERS_UID, best32.));

      /* store path in a variable to be used to create a libref */
      /* build the directory of the old project path */
      prjpath = trim(left("&PROJDIRPATHOLD")) || &SEP || trim(left(prjname));
      /* store path in a variable to be used to create a libref */
      call symput("PROJDIRPATHOLD", prjpath);
   run;
   
   /*construct migration target path*/
   data _null_;
      length prjpath $ 256;
      prjpath = "&FOLDER_NAME";

      /* truncate to 20 characters */
      if lengthc(trim(left(prjpath))) > 20 then do;
          short_name = substr(trim(left(prjpath)), 1, 20);
      end;
      else do;
        short_name = trim(left(prjpath));
      end;

      short_name = translate(trim(left(short_name)),'_',' '); 
      /*based on the logic given in defect S0693911, need to check if all the chars are alphabeta or number or _*/      
      hasBadChar = 0;
      
      do i = 1 to length(short_name);
          ithChar = substr(short_name,i,1);
            if upcase(ithChar) not in ("A", "B", "C", "D", "E", "F", "G", "H", "I", "J",
                                       "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T",  
                                       "U", "V", "W", "X", "Y", "Z", "_",  
                                       "1", "2", "3", "4", "5", "6", "7", "8", "9", "0" 
                                      ) then do;
              hasBadChar = 1;
              leave;
            end;
      end;

      if hasBadChar = 1 then do;
          prjname = strip(put(&VAR_FOLDERS_UID, best32.));
      end;
      else do;
          prjname = trim(left(short_name)) || strip(put(&VAR_FOLDERS_UID, best32.));
      end;

      prjpath = trim(left("&PROJDIRPATHNEW"))|| &SEP. || (trim(left(prjname)));
      call symput("PROJDIRPATHNEW", prjpath);            
   run;   
   
   /* create directory */
   %orvu_mkdirs(path=&PROJDIRPATHNEW.);

   libname VARCPIN "&PROJDIRPATHOLD";
   libname VARCPOUT "&PROJDIRPATHNEW";

   proc migrate in = VARCPIN out = VARCPOUT;
   run;

   /*****- Copy all datasets from the old data store into the new data store -*****/
   /*    proc datasets nolist nodetails;
            copy in=VARCPIN out=VARCPOUT;
         quit;
   */   
   /******************************************************/
   /*****- any changes to project tables goes here  -*****/
   /******************************************************/
   /** change all occurance of exceedance to pobserv, "do not use" to DO_NOT_USE, 
    *  "user specified" to USER_SPECIFIED to data set fitopts
    */

   /* -------------   fitopts   ------------- */ 
   %orvl_ds_exists(VARCPOUT.fitopts, "update fitopts", false);
   %if (&ds_nobs. > 0) %then %do;

      /* Migration to 4.2 */
      %if &OLDDATA_VERSION. < 4.2 %then %do;
         data VARCPOUT.fitopts;
             set VARCPOUT.fitopts;
             name   = transtrn(name, "Exceedance", "PObserv");
             name   = transtrn(name, "Sev_RegressionMethod_Kris", "Sev_SelectionMethod_Kris");
             name   = transtrn(name, "Freq_RegressionMethod_Kris", "Freq_SelectionMethod_Kris");
      
             cvalue = transtrn(cvalue, "Do not use", "DO_NOT_USE");
             cvalue = transtrn(cvalue, "User Specified", "USER_SPECIFIED");
             type   = transtrn(type, "EXCEEDANCE", "POBSERV");

             if name = 'Sev_Distribution' and CValue = 'LOGNGPD1' then do;
                 CValue = '_GPD_LOGN'; 
             end;

             if name = 'Sev_Distribution' and CValue = 'LOGNGPD' then do;
                 CValue = '_GPD_LOGN'; 
             end;
         run;

         data _NULL_;
            set VARCPOUT.fitopts;
            where missing(CCUID) and Name = "Min_Fitting_Count";
            call symput("TheMinFittingCount", NValue);
         run;
      %end;

      /* Migration to 5.1 */
      %if &OLDDATA_VERSION. < 5.1 %then %do;
         /*defect: S0635271: add 3 new rows in the fitopts ds*/
         &ORVAR_PROC_SQL.;
            INSERT INTO VARCPOUT.fitopts 
                     (Name, Nvalue, Cvalue, CCUID, Type, datatype) 
            VALUES 
                     ("Freq_Indicator_Properties", ., "INTERNAL_BL_UID", ., "INDICATOR", "")
            VALUES 
                     ("Freq_Model_Basis_Props", ., "INTERNAL_BL_UID", ., "INDICATOR", "")
            VALUES 
                     ("Freq_Model_Is_Stratified", ., "true", ., "INDICATOR", "")      
            ;
         quit;
      %end;
   %end;
   
   /* -------------   strataprops_X   ------------- */ 
   
   %orvl_ds_exists(VARCPOUT.strataprops_i, "update strataprops", false);
   %if (&ds_nobs. > 0) %then %do;

      /* Migration to 5.1 */
      %if &OLDDATA_VERSION. < 5.1 %then %do;
         /* Switched from 'Thestrataprops' to 'ClassificationVariable' */
         data VARCPOUT.strataprops_i;
            set VARCPOUT.strataprops_i;
            rename TheStrataProps = ClassificationVariable;
         run;
      %end;

   %end;

   %orvl_ds_exists(VARCPOUT.strataprops_c, "update strataprops", false);
   %if (&ds_nobs. > 0) %then %do;

      /* Migration to 5.1 */
      %if &OLDDATA_VERSION. < 5.1 %then %do;
         /* Switched from 'Thestrataprops' to 'ClassificationVariable' */
         data VARCPOUT.strataprops_i;
            set VARCPOUT.strataprops_i;
            rename TheStrataProps = ClassificationVariable;
         run;
      %end;

   %end;

   /* -------------   var_criteria   ------------- */

   /*default COPULA_TARGET to freq */
   %orvl_ds_exists(VARCPOUT.var_criteria, "update var_criteria", false);
   %if (&ds_nobs. > 0) %then %do;

      /* Migration to 4.2 */
      %if &OLDDATA_VERSION. < 4.2 %then %do;

         /* Add COPULA_TARGET column to VAR_CRITERIA */
         &ORVAR_PROC_SQL.;
            ALTER TABLE 
                  VARCPOUT.var_criteria 
            ADD   
                  COPULA_TARGET CHAR(4)
                  ;
            ALTER TABLE 
                  VARCPOUT.var_criteria 
            ADD 
                  COMBINING_METHOD VARCHAR(32)
                  ;
         quit;
 
         data VARCPOUT.var_criteria;
             set VARCPOUT.var_criteria;

             /*if the corr type is user defined aggr "A"*/
             if Correlation_type = "A" then do;
                 COPULA_TARGET = "AGGR";
             end;
             else do;
                 COPULA_TARGET = "FREQ";
             end;
         run;

         /*default COMBINING_METHOD to "MAXIMUM" */
         data VARCPOUT.var_criteria;
             set VARCPOUT.var_criteria;
             COMBINING_METHOD = "MAXIMUM"; 
         run;
   
         &ORVAR_PROC_SQL.;
             ALTER TABLE 
                  VARCPOUT.var_criteria 
             ADD CONSTRAINT 
                  CP_TGT_not_null not null(COPULA_TARGET)
                  ;
         quit;
   
         /* extensions.xml has changed from 4.1 to 4.2. in var_criteria, capital_dist_method 
          *  needs to change as followings. 
          *    BLOCK_MAX -> Block_Maxima 
          *    CVAR95 -> CVaR95
          *    CVAR -> CVaR
          *    VAR95 -> VaR95
          *    VAR -> _VaR_
          *    DUAL_BLOCK_MIN -> Dual_Block_Minima
          */
         data VARCPOUT.var_criteria;
             set VARCPOUT.var_criteria;
             /*if the corr type is user defined aggr "A"*/
             if CAPITAL_DIST_METHOD = "BLOCK_MAX" then do;
                 CAPITAL_DIST_METHOD = "Block_Maxima";
             end;

             if CAPITAL_DIST_METHOD = "CVAR95" then do;
                 CAPITAL_DIST_METHOD = "CVaR95";
             end;

             if CAPITAL_DIST_METHOD = "CVAR" then do;
                 CAPITAL_DIST_METHOD = "CVaR";
             end;

             if CAPITAL_DIST_METHOD = "VAR95" then do;
                 CAPITAL_DIST_METHOD = "VaR95";
             end;

             if CAPITAL_DIST_METHOD = "VAR" then do;
                 CAPITAL_DIST_METHOD = "_VaR_";
             end;

             if CAPITAL_DIST_METHOD = "DUAL_BLOCK_MIN" then do;
                 CAPITAL_DIST_METHOD = "Dual_Block_Minima";
             end;
         run;
      %end; /* END: Migration to 4.2 */

      /* Migration to 5.1 */
      %if &OLDDATA_VERSION. < 5.1 %then %do;

         /* Add the copula_extension_ID to var_criteria for compatibility with Copulas
            ( see US3899 in Rally ) */
         data VARCPOUT.var_criteria;
            set VARCPOUT.var_criteria;
            length 
               copula_extension_ID   8
               copula_ID           $ 32
               ;
         run;

      %end;

   %end;

   /* -------------   model   ------------- */

   /* Migration to 4.2 */
   %if &OLDDATA_VERSION. < 4.2 %then %do;

      %orvl_ds_exists(VARCPOUT.MODEL, "update model", false);

      %if (&ds_nobs. > 0) %then %do;   
         
         /* Migration to 4.2 */
         %if &OLDDATA_VERSION. < 4.2 %then %do;

            /* Add INSURANCE_METHOD column to MODEL */
            &ORVAR_PROC_SQL.;
               ALTER TABLE 
                     VARCPOUT.MODEL 
               ADD 
                     INSURANCE_METHOD VARCHAR(50)
                     ;
            quit;
            
            /*default INSURANCE_METHOD to BASEL */
            data VARCPOUT.MODEL;
                set VARCPOUT.MODEL;
                INSURANCE_METHOD = "BASEL";
            run;
         
            &ORVAR_PROC_SQL.;
               ALTER TABLE 
                        VARCPOUT.MODEL 
               ADD CONSTRAINT 
                        INS_MTD_not_null not null(INSURANCE_METHOD)
                        ;
            quit;
         %end;
      %end;
   %end;

   /* -------------------  Extension_Input  ----------------- */
   proc sql noprint;
      CREATE TABLE 
                  VARCPOUT.EXTENSION_INPUT (
                                            PROCESS           VARCHAR(16),           /* Identify subsystem for the input */
                                            INPUT             VARCHAR(16),         /* Identify function for the input */
                                            ID                num          NOT NULL, 
                                            XML_ID            VARCHAR(16)  NOT NULL, /* <id> in Extensions.xml */
                                            NAME              VARCHAR(16)  NOT NULL, /* variable name */
                                            DISPLAY           VARCHAR(128),          /* user-displayed text for Name */
                                            NVALUE            num,                   /* number value of variable (when applicable) */
                                            CVALUE            VARCHAR(128),          /* character value of variable (when applicable) */
                                            VALUE_TYPE        CHAR(1)                /* One of 'N', 'C', 'D', or 'F' for numeric, character, dataset, or file */
                                            );
	quit;

   /* -------------   fparmX (X = i, c, cov_i, or cov_c)  ------------- */

   /* Migration to 4.2 */
   %if &OLDDATA_VERSION. < 4.2 %then %do;
      %orv_addFreqModelBasisPropVars(dataSet = VARCPOUT.fparmi);
      %orv_addFreqModelBasisPropVars(dataSet = VARCPOUT.fparmc);
      %orv_rename_parmDS(dataSet = VARCPOUT.fparmi);
      %orv_rename_parmDS(dataSet = VARCPOUT.fparmc);
      /*there is no fparmp  since public does not have freq fitting*/
   %end;

   /* Migration to 5.1 */
   %if &OLDDATA_VERSION. < 5.1 %then %do;
      %orv_migrate_Fparm_To_51(ds = VARCPOUT.fparmi, type = i);
      %orv_migrate_Fparm_To_51(ds = VARCPOUT.fparmc, type = c);

      %orv_CorrectFrequencyCov51(ds = VARCPOUT.fparm_covi);
      %orv_CorrectFrequencyCov51(ds = VARCPOUT.fparm_covc);
   %end;

   /* Migration to 6.1 */
   %if &OLDDATA_VERSION. < 5.1 %then %do;
      %orv_migrate_Fparm_To_61(ds = VARCPOUT.fparmi, type = i);
      %orv_migrate_Fparm_To_61(ds = VARCPOUT.fparmc, type = c);
   %end;

   /* -------------   selfparm   ------------- */

   /* Migration to 4.2 */
   %if &OLDDATA_VERSION. < 4.2 %then %do;
      %orv_addFreqModelBasisPropVars(dataSet = VARCPOUT.selfparmi);
      %orv_addFreqModelBasisPropVars(dataSet = VARCPOUT.selfparmc);   
      %orv_rename_parmDS(dataSet = VARCPOUT.selfparmi);
      %orv_rename_parmDS(dataSet = VARCPOUT.selfparmc);
   %end;

   /* Migration to 5.1 */
   %if &OLDDATA_VERSION. < 5.1 %then %do;
      %orv_migrate_Fparm_To_51(ds = VARCPOUT.selfparmi, type = i);
      %orv_migrate_Fparm_To_51(ds = VARCPOUT.selfparmc, type = c);
   %end;

   /* -------------   fparm_mdlX   ------------- */

   /* Migration to 5.1 */
   %if &OLDDATA_VERSION. < 5.1 %then %do;

      %orvl_ds_exists(VARCPOUT.fparm_mdli, "update fparm_mdli", false);
      %if (&ds_nobs. > 0) %then %do;   
         %orv_CorrectFparm_mdlX51(ds = VARCPOUT.fparm_mdli);
      %end;

      %orvl_ds_exists(VARCPOUT.fparm_mdlc, "update fparm_mdlc", false);
      %if (&ds_nobs. > 0) %then %do;   
         %orv_CorrectFparm_mdlX51(ds = VARCPOUT.fparm_mdlc);
      %end;
   %end;

   /* -------------    SparmX and SelsparmX     ------------- */

   /* Migration to 4.2 */
   %if &OLDDATA_VERSION. < 4.2 %then %do;

      /* correct the severity distributions: 
       * 1. change LOGNGPD and LOGNGPD1 to _GPD_LOGN.
       *    for LOGNGPD1 to __GPD_LOGN, the param_4 = 1/param_4.
       * 2. ROBUSTLNG, if COUNT_KRI > 0, drop the fittings,
       *    otherwise, keep it. Since we remove the KRI fitting with ROBUSTLNG.
       * 3. if dist = LOGNGPD_HILL, in the sparm[icp], TailBoundaryChoice column
       *    needs to be updated to GET_HILL_ESTIMATE
       * 4. for covariance matrix for LOGNGPD1, the columns and rows of Param_4 
       *    needs to multiply -(1/Param_4^2), where the Param_4 is the original
       *    value of param_4 of the LOGNGPD1. 
       * Note: The _GPD_LNG may need to be removed as well, but the late change 
       * (4/24/09)on this in 4.1 makes this unnucessary. But I am afraid if the 
       * 3.2 migration to 4.2 may need this _GPD_LNG since the 3.2 to 4.1 
       * migration seems did not change this.
      */
    
      %orv_rename_parmDS(dataSet = VARCPOUT.sparmi);
      %orv_rename_parmDS(dataSet = VARCPOUT.sparmc);
      %orv_rename_parmDS(dataSet = VARCPOUT.sparmp);

      %orv_rename_parmDS(dataSet = VARCPOUT.selsparmi);
      %orv_rename_parmDS(dataSet = VARCPOUT.selsparmc);
      %orv_rename_parmDS(dataSet = VARCPOUT.selsparmp);

      %orv_migrateTailBoundary(dataSet = VARCPOUT.sparmi);
      %orv_migrateTailBoundary(dataSet = VARCPOUT.sparmc);
      %orv_migrateTailBoundary(dataSet = VARCPOUT.sparmp);

      %orv_migrateTailBoundary(dataSet = VARCPOUT.selsparmi);
      %orv_migrateTailBoundary(dataSet = VARCPOUT.selsparmc);
      %orv_migrateTailBoundary(dataSet = VARCPOUT.selsparmp);

      /*take care the tailBoundaryChoice*/
      %orv_migrateLOGNGPD_HILL(libr = VARCPOUT.sparmi);
      %orv_migrateLOGNGPD_HILL(libr = VARCPOUT.sparmc);
      %orv_migrateLOGNGPD_HILL(libr = VARCPOUT.sparmp);

      %orv_migrateLOGNGPD_HILL(libr = VARCPOUT.selsparmi);
      %orv_migrateLOGNGPD_HILL(libr = VARCPOUT.selsparmc);
      %orv_migrateLOGNGPD_HILL(libr = VARCPOUT.selsparmp);

      /*LOGNROBUST*/
      %orv_migrateLNGROBUST(libr = VARCPOUT.selsparmi);
      %orv_migrateLNGROBUST(libr = VARCPOUT.selsparmc);
      %orv_migrateLNGROBUST(libr = VARCPOUT.selsparmp);

      %orv_migrateLNGROBUST(libr = VARCPOUT.sparmi);
      %orv_migrateLNGROBUST(libr = VARCPOUT.sparmc);
      %orv_migrateLNGROBUST(libr = VARCPOUT.sparmp);
 
      /* Add min fitting coiunt column to sev parm datasets */
      %orv_Add_New_Columns(data = VARCPOUT.sparmi);
      %orv_Add_New_Columns(data = VARCPOUT.sparmc);
      %orv_Add_New_Columns(data = VARCPOUT.sparmp);

      %orv_Add_New_Columns(data = VARCPOUT.selsparmi);
      %orv_Add_New_Columns(data = VARCPOUT.selsparmc);
      %orv_Add_New_Columns(data = VARCPOUT.selsparmp); 
   %end;

   /* -------------    Sparm_covX      ------------- */

   /* Migration to 4.2 */
   %if &OLDDATA_VERSION. < 4.2 %then %do;

      /*The migiration of covX has to follow the LNGROBUST migarion since there are
       * iterations are removed from there and has to remove the same iteration 
       * and CCUID in covX
       */
       %orv_adj_sev_dist(i);
       %orv_adj_sev_dist(c);
       %orv_adj_sev_dist(p);
 
   %end;
 
   /* -------------------      Distribution Specific Changes       ----------------- */

   /* Migration to 4.2 */
   %if &OLDDATA_VERSION. < 4.2 %then %do;

      /* change names to _GPD_LOGN, this has to after the migration of sparm_covX
       * since we need to original param4 values in cov migiration */    
      %orv_migrate_GPD_LOGN_Name(libr = VARCPOUT.fitopts);

      %orv_migrate_GPD_LOGN_Name(libr = VARCPOUT.selsparmi);
      %orv_migrate_GPD_LOGN_Name(libr = VARCPOUT.selsparmc);
      %orv_migrate_GPD_LOGN_Name(libr = VARCPOUT.selsparmp);
       
      %orv_migrate_GPD_LOGN_Name(libr = VARCPOUT.sparmi);
      %orv_migrate_GPD_LOGN_Name(libr = VARCPOUT.sparmc);
      %orv_migrate_GPD_LOGN_Name(libr = VARCPOUT.sparmp);

   %end;

   /**************************************************/
   /*********** Create Frequency PANEL_PROPS *********/
   /**************************************************/
   /* Migration to 4.1 */
   %if &OLDDATA_VERSION. < 4.1 %then %do;
      %orvl_ds_exists(VARCPOUT.freqmomts_i, "panel props needed", false);
      %if (&ds_nobs. > 0) %then %do;
         %orv_make_panel_prop_for_migrate(varprj=VARCPOUT, type=i);
         %orv_make_panel_prop_for_migrate(varprj=VARCPOUT, type=c);
         %orv_make_panel_prop_for_migrate(varprj=VARCPOUT, type=p);
      %end;
   %end;

   /**************************************************
    *  Rename all parm datasets as <parm>_inc        *
    *  They will be constituted similarly to an      *
    *  incremental load when user first opens the    *
    *  project with VaR                              *
    **************************************************/
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = fparmi);
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = fparm_covi);
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = fparm_blesti);
   
   %orv_adjust_sparm(sparmds = VARCPOUT.sparmi);
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = sparmi);
  
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = fparmc);
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = fparm_covc);
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = fparm_blestc);

   %orv_adjust_sparm(sparmds=VARCPOUT.sparmc);
   %orv_rename_parm_ds(libname=VARCPOUT, oldname=sparmc);
   
   %orv_adjust_sparm(sparmds = VARCPOUT.sparmp);
   %orv_rename_parm_ds(libname = VARCPOUT, oldname = sparmp);
   
   /************************************************************/
   /*- Drop all KRI, MIXING, VAR, and SCENARIO related datasets 
    *  These will be recreated when the project is opened     -*/
   /************************************************************/
 /*  proc datasets library=VARCPOUT memtype=data nolist nodetails;
      delete kribl:  krifirm: 
             mixsparm mixfparm rrfparm rrsparm 
             agg: fdraws feffects freqcounts_:
             selsparm: selfparm: varsim_: varstat_:
             vlosskri vstatkri_: simseeds
             bucket_:  scenario_:
             freqmomts_scenario: sevmomts_scenario:
      ;
   quit;*/

   /**************************************************/
   /****************- New tables -********************/
   /**************************************************/
   libname VARCPIN clear;
   libname VARCPOUT clear;

%mend orvdb_migrate_projects;

%macro orv_adj_sev_dist(dataType);
   %orvl_ds_exists(VARCPOUT.sparm_cov&dataType., "changing sev dist", false);
   %if (&ds_nobs. > 0) %then %do;
       &ORVAR_PROC_SQL.;
            SELECT DISTINCT 
                  _NAME_ 
            INTO 
                  :paramArray SEPARATED BY " " 
            FROM 
                  VARCPOUT.sparm_cov&dataType.
            ;            
       quit;
       
       /*temp data set to avoid the name collision on param_4*/
       data _tmp_sparm&dataType.;
          set VARCPOUT.sparm&dataType.;
          where 
               distribution = "LOGNGPD1" 
               or 
               distribution = "LOGNGPD_HILL" 
               or 
               distribution = "LNGROBUST"
               ;
       run;

      %orvl_ds_exists(_tmp_sparm&dataType., "temp sparmX", false);
       %if (&ds_nobs. > 0) %then %do;
           data _tmp_sparm&dataType.;
               set _tmp_sparm&dataType. (rename = (param_4 = p4));
           run;

           data VARCPOUT.sparm_cov&dataType.;
               set VARCPOUT.sparm_cov&dataType.;
               array Param[*] &paramArray;

               if _N_ = 1 then do;
                   declare hash sfitParam(dataset: "_tmp_sparm&dataType.");
                   rc = sfitParam.defineKey('iteration');
                   rc = sfitParam.defineData('p4');
                   rc = sfitParam.defineDone(); 
                   call missing(p4);
               end;
   
               if DISTRIBUTION = 'LOGNGPD1' or DISTRIBUTION = 'LOGNGPD_HILL' then do;
                   rc = sfitParam.find();

                   if _NAME_ = 'Param_4' then do;
                      do i = 1 to dim(param);
                            param[i] = - param[i]/(p4*p4);
                      end;
                   end;

                   Param_4 = -Param_4/(p4*p4);
                   DISTRIBUTION = '_GPD_LOGN';  
               end;
               
               if DISTRIBUTION = 'LNGROBUST' then do;
                  rc = sfitParam.find();

                  if rc ne 0 then do;
                     DELETE; /* the iteration has been deleted in LNGROBUST migration with KRI.*/
                  end;
               end;
           run;
       %end; /*end of temp sparmX*/
       
       data VARCPOUT.sparm_cov&dataType.;
           set VARCPOUT.sparm_cov&dataType.;

           if DISTRIBUTION = 'LOGNGPD' then do;
               DISTRIBUTION = '_GPD_LOGN';   
           end;

           drop p4 rc i;
       run;
   %end;      
%mend orv_adj_sev_dist;

%macro orv_adjust_sparm(sparmds=);
   /* there is no need to adjust it from 4.1 to 4.2*/
   /*
   data &sparmds; set &sparmds;
   
      if distribution = "LOGNGPD" then do;
         Param_3 = Param_3/exp(Param_1);
         Param_xLimit = Param_xLimit/exp(Param_1);
      end;  
      else if distribution = "LOGNGPD1" then do;
         Param_3 = Param_3/exp(Param_1);
         Param_xLimit = Param_xLimit/exp(Param_1);
      end;  
      else if distribution = "LOGNGPD_HILL" then do;
         Param_3 = Param_3/exp(Param_1);
         Param_xLimit = Param_xLimit/exp(Param_1);
      end;  
      else if distribution = "IGAUSS" then do;
         _tmp_Param_1 = Param_1;
         Param_1 = Param_2;
         Param_2 = _tmp_Param_1/Param_1;
      end;
      
      drop _: ;
   run;
   */
%mend orv_adjust_sparm;

%macro orv_migrateLNGROBUST(libr=);
   %orvl_ds_exists(&libr, "changing LNGROBUST", false);
   %if (&ds_nobs. > 0) %then %do;
       data &libr. ;
           set &libr. ;

           if DISTRIBUTION = 'LNGROBUST' then do;
               if COUNT_KRI ne 0 then do;
                  DELETE;
               end;
           end;
       run;
   %end;
%mend orv_migrateLNGROBUST;

%macro orv_migrate_GPD_LOGN_Name(libr=);
    %orvl_ds_exists(&libr., "changing sev dist", false);
    %if (&ds_nobs. > 0) %then %do;
        %if &libr eq VARCPOUT.fitopts %then %do;
          data &libr.;
             set &libr.;

             if name = 'Sev_Distribution' and CValue = 'LOGNGPD1' then do;
                 CValue = '_GPD_LOGN'; 
             end;

             if name = 'Sev_Distribution' and CValue = 'LOGNGPD_HILL' then do;
                 CValue = '_GPD_LOGN'; 
             end;

             if name = 'Sev_Distribution' and CValue = 'LOGNGPD' then do;
                 CValue = '_GPD_LOGN'; 
             end;
         run;
        %end;
        %else %do;
          data &libr.;
              set &libr.;
              if DISTRIBUTION = 'LOGNGPD1' then do;
                  DISTRIBUTION = '_GPD_LOGN';   
                  Param_4 = 1/Param_4;
              end;

              if DISTRIBUTION = 'LOGNGPD_HILL' then do;
                  DISTRIBUTION = '_GPD_LOGN';   
                  Param_4 = 1/Param_4;
              end;

              if DISTRIBUTION = 'LOGNGPD' then do;
                  DISTRIBUTION = '_GPD_LOGN';
              end;
          run;
        %end;
    %end;
%mend orv_migrate_GPD_LOGN_Name;

/*this only take care the TailBoundaryChoice*/
%macro orv_migrateLOGNGPD_HILL(libr=);
   %orvl_ds_exists(&libr., "Adding new variables selsparmi", false);
   %if (&ds_nobs. > 0) %then %do;
       data &libr.;
           set &libr.;

            if DISTRIBUTION = "LOGNGPD_HILL" then do;
               TailBoundaryChoice = "GET_HILL_ESTIMATE";
            end;
       run;
   %end;
%mend orv_migrateLOGNGPD_HILL;

%macro orv_rename_parm_ds(libname=, oldname=);
/*   %orvl_ds_exists(&libname..&oldname, "rename parmds", false);
   %if (&ds_nobs. > 0) %then %do;
      proc datasets library=&libname. nolist nodetails;
        change &oldname. = &oldname._inc
        ;
      quit;
   %end;*/
%mend orv_rename_parm_ds;

%macro orv_make_panel_prop_for_migrate(varprj=work, type=i);
   ***** Get CCV and MAPPING UIDs; 
   proc sql noprint;  
      SELECT 
            count(CCV_UID)  
      INTO 
            :numCcvs
      FROM 
            &varprj..model_ccv_mapping
            ;

      SELECT 
            FIT_INTERVAL  
      INTO 
            :PROJ_INTERVAL
      FROM 
            &varprj..model
            ;
   quit;  

   proc sql noprint;  
      %do i=0 %to &numCcvs-1;
         %if (%upcase("&type") eq "I") %then %do;
            SELECT 
                     CCVVALUE_&i 
            INTO
                     :ref_&i
            FROM
                     &varprj..ccmapval_&i 
            WHERE
                     CCVUID_&i = -799
                     ;
         %end;
         %else %if (%upcase("&type") eq "C") %then %do;
            SELECT 
                  CCVVALUE_&i 
            INTO
                  :ref_&i
            FROM 
                  &varprj..ccmapval_&i 
            WHERE
                  CCVUID_&i = -699
                  ;
         %end;
         %else %if (%upcase("&type") eq "P") %then %do;
            SELECT 
                  CCVVALUE_&i 
            INTO
                  :ref_&i
            FROM
                  &varprj..ccmapval_&i 
            WHERE
                  CCVUID_&i = -599
                  ;
         %end;
      %end;
   quit;

   data _null_;
      length dbprop $ 1024;
      dbprop = "";

      %do i=0 %to &numCcvs-1;
         %if &&ref_&i ne . %then %do;
            dbprop = strip(dbprop) || " " || strip("&&ref_&i");
         %end;
      %end;

      %if (%upcase("&type") eq "I") %then %do;
         dbprop = strip(dbprop) || " " || "OWN";
         call symput("freqMomtsType", 'I');
      %end;
      %else %if (%upcase("&type") eq "C") %then %do;
         call symput("freqMomtsType", 'C');
      %end;
      %else %if (%upcase("&type") eq "P") %then %do;
         call symput("freqMomtsType", 'E');
      %end;

      call symput("CCVNAMES", trim(dbprop));
   run;

   %if (&CCVNAMES ne %str()) %then %do;
      %orva_getfreqmomts(dataType      = &freqMomtsType,
                         global_range  = &varprj..global_range_i,
                         interval      = &PROJ_INTERVAL,
                         missing_range = &varprj..missing_range,
                         NUMCCVS       = &numCcvs,
                         CCV_NAMES     = &CCVNAMES,
                         outprops      = &varprj..panelprops_&type,
                         outmissprops  = &varprj..outmissprops_&type,
                         outmissrange  = &varprj..outmissrange_&type
                         );
   %end;

%mend orv_make_panel_prop_for_migrate;

%macro orv_addFreqModelBasisPropVars(dataSet=);
   %orvl_ds_exists(&dataSet., "Adding new variables ", false);
   %if (&ds_nobs. > 0) %then %do;
       &ORVAR_PROC_SQL.;
            ALTER TABLE 
                     &dataSet. 
            ADD 
                     Freq_Model_Basis_Props CHAR(256)
                     ;
            ALTER TABLE 
                     &dataSet. 
            ADD
                     Freq_Model_Is_Stratified CHAR(32)
                        ;
            ALTER TABLE 
                     &dataSet. 
            ADD
                     Freq_Indicator_Properties CHAR(256)
                     ;
       quit;
   %end;
%mend orv_addFreqModelBasisPropVars;

%macro orv_rename_parmDS(dataSet=);  
   /*Rename column name from EXCEEDANCE_XXX to POBSERV_*** for sparmX(i, c, p) */
   %let exeMethodExisti            = 0;
   %let exeStyleExisti             = 0;
   %let exeUIDExisti               = 0;
   %let Freq_Exceedance_Source     = 0;
   %let Freq_RegressionMethod_Kris = 0;

   %orvl_ds_exists(&dataSet., "Adding new variables", false);
   %if (&ds_nobs. > 0) %then %do;   
          data _null_;
              dsid=open("&dataSet.");

              if dsid ne 0 then do;
                  check = varnum(dsid,'EXCEEDANCE_METHOD');
                  call symput("exeMethodExisti", left(check));

                  check = varnum(dsid,'EXCEEDANCE_STYPE');
                  call symput("exeStyleExisti", left(check));

                  check = varnum(dsid,'EXCEEDANCE_SUID');
                  call symput("exeUIDExisti", left(check));

                  check = varnum(dsid,'Freq_Exceedance_Source');
                  call symput("Freq_Exceedance_Source", left(check));

                  check = varnum(dsid,'Freq_RegressionMethod_Kris');
                  call symput("Freq_RegressionMethod_Kris", left(check));
              end;
              else do;
                  call symput("exeMethodExisti", 0);
                  call symput("exeStyleExisti", 0);
                  call symput("exeUIDExisti", 0);
                  call symput("Freq_Exceedance_Source", 0);
                  call symput("Freq_RegressionMethod_Kris", 0);
              end;  
              rc = close(dsid);      
          run;
    %end;
    
    %if &exeMethodExisti ne 0 OR &exeStyleExisti ne 0 OR &exeUIDExisti ne 0 
         OR &Freq_Exceedance_Source ne 0 OR &Freq_RegressionMethod_Kris ne 0 %then %do;
          %orvl_ds_exists(&dataSet., "Adding new variables", false);
          %if (&ds_nobs. > 0) %then %do;      
           data &dataSet.;
              set &dataSet. (rename = (
                                       %if &exeMethodExisti ne 0 %then %do;
                                          EXCEEDANCE_METHOD = POBSERV_METHOD
                                       %end;
                                       %if &exeStyleExisti ne 0 %then %do; 
                                          EXCEEDANCE_STYPE  = POBSERV_STYPE
                                       %end;
                                       %if &exeUIDExisti ne 0 %then %do;
                                          EXCEEDANCE_SUID   = POBSERV_SUID
                                       %end;    
                                       %if &Freq_Exceedance_Source ne 0 %then %do;
                                          Freq_Exceedance_Source = Freq_Pobserv_Source
                                       %end;    
                                       %if &Freq_RegressionMethod_Kris ne 0 %then %do;
                                          Freq_RegressionMethod_Kris = Freq_Selection_Method_Kris
                                       %end;    
                                       )
                               );
           run;
        %end;
    %end;
%mend orv_rename_parmDS;

/*udpate tailBoundaryValue, tailBoundaryPercent*/
%macro orv_migrateTailBoundary(dataSet=);
   %orvl_ds_exists(&dataSet., "Adding new variables ", false);
   %if (&ds_nobs. > 0) %then %do;
       &ORVAR_PROC_SQL.;
            ALTER TABLE 
                     &dataSet. 
            ADD 
                     TailBoundaryChoice CHAR(32)
                     ;
            ALTER TABLE 
                     &dataSet. 
            ADD
                     TailBoundaryValue NUM
                     ;
            ALTER TABLE 
                     &dataSet. 
            ADD
                     TailBoundaryPercent NUM
                     ;
       quit;
       
       /*read the tailBoundary values from fitopts*/
       %orvl_ds_exists(VARCPOUT.fitopts, "update tailBounday ", false);
      %if (&ds_nobs. > 0) %then %do;
              /*first get the project level values, ie when ccuid=.*/
          &ORVAR_PROC_SQL.;
               SELECT  
                     CValue 
               INTO 
                     :TailBoundaryChoice 
               FROM 
                     VARCPOUT.fitopts 
               WHERE 
                     ccuid =. 
                  AND
                     Name = "TailBoundaryChoice"
                     ;            
               SELECT  
                     NValue 
               INTO 
                     :TailBoundaryValue 
               FROM 
                     VARCPOUT.fitopts 
               WHERE 
                     ccuid =. 
                  AND 
                     Name = "TailBoundaryValue"
                     ;            
               SELECT  
                     NValue 
               INTO 
                     :TailBoundaryPercent 
               FROM 
                     VARCPOUT.fitopts 
               WHERE
                     ccuid=. 
                  AND
                     Name = "TailBoundaryPercent"
                     ;           
           quit;          
 
           &ORVAR_PROC_SQL.;
              UPDATE 
                     &dataSet. pm 
               SET
                     TailBoundaryChoice = 
                     (SELECT 
                              CVALUE 
                     FROM
                              VARCPOUT.fitopts ops 
                     WHERE
                              pm.ccuid = ops.ccuid
                        AND
                              ops.name = "TailBoundaryChoice"
                              );

              UPDATE 
                     &dataSet. pm
               SET 
                     TailBoundaryValue = 
                     (SELECT 
                              NVALUE 
                     FROM 
                              VARCPOUT.fitopts ops 
                     WHERE
                              pm.ccuid = ops.ccuid
                        AND
                              ops.name = "TailBoundaryValue"
                              );

              UPDATE 
                     &dataSet. pm 
               SET
                     TailBoundaryPercent = 
                     (SELECT 
                              NVALUE 
                      FROM
                              VARCPOUT.fitopts ops 
                      WHERE
                              pm.ccuid = ops.ccuid
                         AND
                              ops.name = "TailBoundaryPercent"
                        );
          quit;
          
          data &dataSet.;
               set &dataSet.;
               if missing(TailBoundaryChoice)  then do;
                   TailBoundaryChoice = "&TailBoundaryChoice";
               end;

               if missing(TailBoundaryValue) then do;
                   TailBoundaryValue = &TailBoundaryValue;
               end;

               if missing(TailBoundaryPercent) then do;
                   TailBoundaryPercent = &TailBoundaryPercent;
               end;
           run;
       %end;
   %end;
%mend orv_migrateTailBoundary;

%macro orv_Add_New_Columns( data=);
   %orvl_ds_exists(&data, "add new column ", false);
   %if (&ds_nobs. > 0) %then %do;
      data &data;
         set &data;
         Min_Fitting_Count = &TheMinFIttingCount;
      run;
   %end;
%mend orv_Add_New_Columns;

%macro orv_migrate_Fparm_To_51(ds = VARCPOUT.fparmi, type = i);
      
      %orvl_ds_exists(&ds., "update &ds.", false);
      %if (&ds_nobs. > 0) %then %do;   
          data  &ds.;
              set &ds.;
              %if %lowcase(&type.) eq i %then %do;
                 Freq_Model_Basis_Props    = "INTERNAL_BL_UID";
                 Freq_Model_Is_Stratified  = "true";
                 Freq_Indicator_Properties = "INTERNAL_BL_UID";
              %end;
              %else %do;
                 Freq_Model_Basis_Props    = "TIME_ONLY";
                 Freq_Model_Is_Stratified  = "false";
                 Freq_Indicator_Properties = "";
              %end;
          run;

         /* Adjust the frequency fits and covariance matrices to reflect new KRI modeling for negbin.  
            Replace all occurences of intercept with fparam_1 = exp(int) fopr pois, exp(intercept)/fparam_2 for negbin */
        
         /* Go into fparm datasets and get rid of intercept. */
         %orv_CorrectFrequencyParms51(ds = &ds., type = &type.);
      %end;

%mend orv_migrate_Fparm_To_51;

%macro orv_migrate_Fparm_To_61(ds = VARCPOUT.fparmi, type = i);
      
   %orvl_ds_exists(&ds., "update &ds.", false);
   %if (&ds_nobs. > 0) %then %do;   
      data  &ds.;
         set &ds.;
         intercept = safe_log(FParam_1);
      run;
   %end;

%mend orv_migrate_Fparm_To_61;

%macro orv_CorrectFrequencyParms51( ds = , type =  );
   /* Old models had 'intercept'.  Now we use Fparam_1 only.  */
   /* ds is  the name of hte parameter dataset to update */

   /* First check that the paramter data is present. */
   %orvl_ds_exists(&ds, "correct fparm ", false);
   %if (&ds_nobs. = 0) %then %do;
      %return;
   %end;

   /* Get freq_strata_count */
   data __temp__;
      set varcpout.fparm_mdl&type.;
   run;
   proc sort data = __temp__;
      by ccuid;
   run;
   data __temp__;
      set __temp__;
      by ccuid;
      retain freq_strata_count 0;
      if first.ccuid then do;
         freq_strata_count = 0;
      end;
      freq_strata_count = freq_strata_count + 1;
      if last.ccuid then do;
         output;
      end;

      keep
         ccuid
         freq_strata_count
         ;
   run;

   data &ds.;
      merge
         &ds.(drop = KRI_Intercept)
         __temp__
         ;
      by ccuid;

      length Freq_KriModel_Full $ 5000;

      array KRIS KRI:;
      array DUMM Dumm:;

      if strip(lowcase(distribution)) eq "pois" then do;
         Fparam_1 = exp(intercept);
      end;
      else if strip(lowcase(distribution)) eq "negbin" then do;
         FParam_2 = Beta;
         Fparam_1 = exp(intercept)/FParam_2;   
      end;

      /* Add in Freq_KriModel_Full */
      Freq_KriModel_Full = "0";
      do i=1 to dim(KRIS);
         if KRIS[i] ne . then do;
            Freq_KriModel_Full = compress(Freq_KriModel_Full)||"+"||put(KRIS[i],BEST10.)||"*"||trim(vname(KRIS[i]));
         end;
      end;
      do i=1 to dim(DUMM);
         if DUMM[i] ne . then do;
            Freq_KriModel_Full = compress(Freq_KriModel_Full)||"+"||put(DUMM[i],BEST10.)||"*("||Freq_Model_Basis_Props||" = " || substr(trim(vname(DUMM[i])),23)|| ")";
         end;
      end;

      Freq_KriModel_Full = "exp("||compress(Freq_KriModel_Full)||")";

      drop 
         intercept 
         i 
         beta
         ;
   run;

   %orvu_delete_data(data = __temp__ );

%mend orv_CorrectFrequencyParms51;

%macro orv_CorrectFrequencyCov51( ds = );
   /* Old models had 'intercept'.  Now we use Fparam_1 only.  */
   /* ds is  the name of the covariance dataset to update */
   %local CCUIDs curr_ccuid loop_ind set_string keep_list select_list;

   /* First check that the covariance matrix is present. */
   %orvl_ds_exists(&ds, "correct fcov ", false);
   %if (&ds_nobs. = 0) %then %do;
      %return;
   %end;

   &ORVAR_PROC_SQL.;
      
      SELECT DISTINCT 
                     iteration,
                     CCUID

      INTO
                     :Iterations separated by " ",
                     :CCUIDs separated by " "
      FROM
                     &ds.
      ;
   quit; 

   %let loop_ind = 1;
   %let curr_it = %scan(&Iterations.,&loop_ind.);
   %do %while (&curr_it. ne %str());

      %let curr_CCUID = %scan(&CCUIDs.,&loop_ind.);

      /* Need the fit parameters from each fit when adjusting covariance matrices */
      data _NULL_;
         name = "&ds.";
         last = substr(name,length(trim(name)),1);
         call symput("last",trim(last));
         lib = substr(name,1,index(name,".")-1);
         call symput("lib",trim(lib));
      run;

      data _NULL_;
         set &lib..fparm&last.;

         where iteration = &curr_it.;

         /*Is the distribution poisson or negative binomial? */
         if strip(lowcase(distribution)) in ("pois" "negbin") then do;
            edit_matrix = "&orv_true.";

            call symputX("int",log(Fparam_1));
            /* For negative binomial */
            if not missing(Fparam_2) then do;
               call symputX("FP2",FParam_2);
            end;
            /* for poissson */
            else do;
               call symputX("FP2",1);
            end;

         end;
         else do;
            edit_matrix = "&orv_false.";           
         end;

         call symputX("edit_matrix",edit_matrix);
      run;

      %if &edit_matrix. eq &orv_false. %then %do;
         data cov_&curr_it.;
            set &ds.;
         run;
      %end;
      %else %do;
         data cov_&curr_it.;
            set &ds.;
            where iteration = &curr_it.;
    
            if trim(lowcase(_NAME_)) = "intercept" then do;
               _NAME_ = "FParam_1";
            end;

            if trim(lowcase(_NAME_)) = "beta" then do;
               _NAME_ = "FParam_2";
            end;

            label
                   beta = "Fparam_2"
                   intercept = "FParam_1"
            ;
            rename 
                   intercept = FParam_1
                   beta      = FParam_2
            ;
         run;      

         /* Pick out the variables that are in this model */
         &ORVAR_PROC_SQL.;
            
            SELECT
                     '"'||trim(_NAME_)||'"',
                     _NAME_
            INTO
                     :select_list separated by " ",
                     :keep_list separated by " "
            FROM
                     cov_&curr_it.
            ;
         quit; 

         data cov_&curr_it.;
            set cov_&curr_it;
            keep _NAME_ &keep_list.;
         run;

         /* Get Hessian Matrix */
         data _curr_hessian;
            set cov_&curr_it.;

            array KRIS KRI:;
            array DUMM Dumm:;
            /* FPARAM_1 Row: */
            if trim(lowcase(_NAME_)) = "fparam_1" then do;
               FParam_1 = exp(&int.)/&FP2.;
               FParam_2 =  -exp(&int.)/&FP2./&FP2.;
               do i=1 to dim(KRIS);
                  if KRIS[i] ne . then do;
                     KRIS[i] = 0;
                  end;
               end;
               do i=1 to dim(DUMM);
                  if DUMM[i] ne . then do;
                     DUMM[i] = 0;
                  end;
               end;
            end;
            /* FPARAM_2 Row: */
            else if (trim(lowcase(_NAME_)) eq "fparam_2") then do;
               FParam_1 = 0;
               FParam_2 = 1;
               do i=1 to dim(KRIS);
                  if KRIS[i] ne . then do;
                     KRIS[i] = 0;
                  end;
               end;
               do i=1 to dim(DUMM);
                  if DUMM[i] ne . then do;
                     DUMM[i] = 0;
                  end;
               end;
            end;
            /* KRI Rows */
            else if lowcase(substr(_NAME_,1,3)) eq "kri" then do;
               FParam_1 = 0;
               FParam_2 = 0;
               do i=1 to dim(KRIS);
                  if KRIS[i] ne . then do;         
                     if lowcase(trim(vname(KRIS[i]))) eq trim(lowcase(_NAME_)) then do;
                        KRIS[i] = 1;
                     end;
                     else do;
                        KRIS[i] = 0;
                     end;
                  end;
               end;
               do i=1 to dim(DUMM);
                  if DUMM[i] ne . then do;
                     DUMM[i] = 0;
                  end;
               end;
            end;
            /* Dummy rows */
            else if lowcase(substr(_NAME_,1,3)) eq "dum" then do;
               FParam_1 = 0;
               FParam_2 = 0;
               do i=1 to dim(KRIS);
                  if KRIS[i] ne . then do;
                     KRIS[i] = 0;
                  end;
               end;
               do i=1 to dim(DUMM);
                  if DUMM[i] ne . then do;
                     if lowcase(trim(vname(DUMM[i]))) eq trim(lowcase(_NAME_)) then do;
                        DUMM[i] = 1;
                     end;
                     else do;
                        DUMM[i] = 0;
                     end;
                  end;
               end;
            end;
            keep _NAME_ &keep_list.;
         run;

         /* Multiply Hessian * Covariance * hessian transposed */
         proc iml;
            _NAME_ = {&select_list.};

            use cov_&curr_it.;
            read all var _num_ into C; 
            close cov_&curr_it.;

            use _curr_hessian;
            read all var _num_ into D;
            close _curr_hessian;

            X=D*C*(D`);

            create cov_&curr_it. from X[colname=_NAME_ rowname=_NAME_];
            append from X[rowname=_NAME_];
            close cov_&curr_it.;
         quit;

         /* Add the iteration and ccuid information back in */
         data cov_&curr_it.;
            set cov_&curr_it.;
            iteration = &curr_it.;
            CCUID     = &curr_CCUID.;
         run;
      %end;

      %let set_string = &set_string. cov_&curr_it.;

      %let loop_ind = %eval(&loop_ind. + 1);
      %let curr_it = %scan(&Iterations.,&loop_ind.);
   %end;

   data &ds.;
      set &set_string.;
   run;

%mend orv_CorrectFrequencyCov51;

%macro orv_CorrectFparm_mdlX51(ds = );

   /* At 5.1 (d3orvar44), we removed the intercept term from the models and replaced it with Fparm_1 */
   data &ds.;
      set &ds.;

      /* New variables for the new frequency model tab */
      length 
         parameter_model  $ 1000
         model_level      $ 1000
         level_display    $ 1000
         ;


      /* Correct the Parameters */
      if strip(lowcase(distribution)) eq "pois" then do;
         FParam_1 = exp(intercept);
      end;
      else if strip(lowcase(distribution)) eq "negbin" then do;
         FParam_2 = Beta;
         Fparam_1 = exp(intercept)/FParam_2;
      end;

      /* Correct meanTableHeader */
      meanTableHeader = "FParam_1";

      /* Correct the stratum model */
      array KRIS KRI_:;
      array DUMMIES Dummy:;
      found_covariate = 0;

      StratumModel = put(FParam_1,BEST20.);

      do _i_ = 1 to dim(DUMMIES);
         if not missing(DUMMIES[_i_]) then do;
            dummy_IBL = input( substr( vname(DUMMIES[_i_]), 23), BEST20.);
            if dummy_IBL = Internal_BL_UID then do;
               if found_covariate ne 1 then do;
                  StratumModel = strip(StratumModel) || " * exp(0";
                  found_covariate = 1;
               end;
               StratumModel = strip(StratumModel) || " + " || put(DUMMIES[_i_], BEST20.);
            end;
         end; 
      end;
      do _i_ = 1 to dim(KRIS);
         if not missing(KRIS[_i_]) then do;
            if found_covariate ne 1 then do;
               StratumModel = strip(StratumModel) || " * exp(0";
               found_covariate = 1;
            end;
            StratumModel = strip(StratumModel) || " + " || put(KRIS[_i_], BEST20.) || "*" || vname(KRIS[_i_]);
         end;
      end;
      StratumModel = strip(StratumModel) || ")";

      /* Frequency model tab changes */
      parameter_model = StratumModel;
      model_level     = "%sysfunc(sasmsg(&orvads, ffdist_internal_business_line, noquote))";
      level_display = put(internal_bl_uid,10.);

      drop 
         intercept 
         beta
         _i_
         dummy_IBL
         found_covariate
         ;
   run;

%mend orv_CorrectFparm_mdlX51;
