/*****************************************************************************************
 * Copyright (C), 2012
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orvr_backtest_freq_dist                                 
 *       
 * PURPOSE: performs frequency model backtesting
 *         
 * CREATES: report
 *                  
 * SUPPORT: stmajo
 *
 * PARAMETERS:
 *
 *     loss_data - Dataset with the incremental losses
 *     
 *     freqparms - dataset with parameters from frequency models fit to base data
 *     
 *     sevparms  - dataset with parameters from severity models fit to base data
 *     
 *****************************************************************************************/
%macro orvr_backtest_freq_dist(
                               loss_data = _bt_losses_,
                               freqparms = _bt_prevfparm_,
                               sevparms  = _bt_prevfparm_
                               );

   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put #orvr_backtest_freq_dist(; 
      %put .                         loss_data = &loss_data,;
      %put .                         freqparms = &freqparms,;
      %put .                         sevparms  = &sevparms;
      %put .                         );
   %end;

   %local 
         model_uid
         fit_interval
         comp_path
         org_path
         maps
         from_props
         load_freq_dists
         ccuids
         ccuid_count
         current_ccuid
         i
         cell_where
         freq_model_basis_props
         freq_model_is_stratified
         Freq_Use_Kris
         freqlist
         NTests
         dsid
         rc
         ;

   /* -------------------------- */
   /* STEP 1: Get the count data */
   /* -------------------------- */

   /* Will need to get the fit interval and CCVs from the parent project */
   /* Get the model_uid */
   %let comp_path = ;
   %let org_path  = ;
   %let model_uid = ;

   data _NULL_;
      set prevprj.model;
      call symputX("model_uid",model_uid);
      call symputX("fit_interval",fit_interval);
   run;

   data _NULL_;
      length string path $ 200;
      length filesep     $ 1;

      /* path of previous project */
      string = pathname("prevprj");

      /* Find the path of the components folder */
      ind = findw(strip(string),"components","\/");
      path = substr(strip(string),1,ind+9);
      call symputX("comp_path",strip(path));

      /* Path for org folder */
      path = substr(strip(string),1,ind-1);
      filesep = substr(strip(string),ind-1,1);
      path = strip(path) || "loss" || filesep || "org";
      call symputX("org_path",strip(path));

   run;

   libname comp_lib "&comp_path.";
   libname org_lib  "&org_path.";

   /* Get the model_uid */
   data _NULL_;
      set prevprj.model;
      call symputX("model_uid",model_uid);
   run;

   /* For this project, what are the mappings?   */
   %let from_props = ;  /* <= Properties with KRIs that are also in this project */
   %let maps       = ;  /* <= The maps that map FROM_PROPS to TO_CCVS */
   
   proc sql noprint;
      select
               MAPPING_UID
      into
               :maps separated by " "
      from 
               prevprj.model_ccv_mapping
      where
               model_uid = &model_uid.
      ;
   quit;

   /* What CCV properties do they map from? */
   proc sql noprint;
      select
            compress(DB_PROP_INT)
      into
            :from_props separated by " "
      from
            comp_lib.mapping
      where
            MAPPING_UID in (&maps.);
      ;
   quit;

   /* Make a fake global_range table.  The start date in the project's
      global range table is the start date of the base project ... */
   /* Get the global start and global end date, make sure the former comes before the latter */
   %let inc_start_date = ;
   %let inc_end_date = ;

   data _null_;
      set varprj.global_range_i;
      call symputX( "inc_end_date", end_date);
   run;

   data _null_;
      set prevprj.global_range_i;
      call symputX( "inc_start_date", end_date + 1);
   run;

   data _bt_global_range;
      start_date = &inc_start_date.;
      end_date   = &inc_end_date.;
   run;
         
   %let count_ccuids = _all_;
   proc sql noprint;
      select distinct
                        ccuid
      into
                        :count_ccuids separated by " "
      from
                        &freqparms.
      ;
   quit;

   %orva_freq_make_counts(
                            lossdata           = &loss_data.,
                            dataType           = i,
                            ForTheseCCUIDs     = &count_ccuids.,
                            cTool              = false,
                            MapDataPrefix      = VARPRJ.ccmapi_,
                            kribl_ds           = VARPRJ.kribl_filtered,
                            krifirm_ds         = VARPRJ.krifirm_filtered,
                            fit_interval       = &fit_interval.,
                            global_range       = _bt_global_range,
                            missing_range      = varprj.missing_Range,
                            CCVs               = &from_props.,
                            numccvs            = %sysfunc(countw(&from_props.)),
                            allibls            = org_lib.INTERNAL_BUSINESSLINE,
                            mapping_uids       = &maps.,
                            outClassifications = _NULL_,
                            outCounts          = _bt_freq_counts,
                            outStatus          = _bt_freq_make_counts_status
                         );

   /* Check status DS for errors */
   data _NULL_;
      set _bt_freq_make_counts_status;
      if status ne 0 then do;
         call symput("ORV_ABORT", "&ORV_TRUE.");
         call symput("ORV_REASON", reason);
      end;
   run;

   %if &ORV_ABORT. eq &ORV_TRUE. %then %do;
      %goto DONE;
   %end;

   /* ------------------------------------------------------------------- */
   /* STEP 2: Match up each observation with the correct model parameters */
   /* ------------------------------------------------------------------- */

   proc sort data = _bt_freq_counts;
      by CCUID;
   run;

   /* Drop the KRI coefficients from the fparm dataset.  They
      share the same names as the KRIs themseslves and so will
      cause roblems for the merge */
   data _bt_freq_parms;
      set &freqparms.;
      drop KRI_:;
   run;

   data _bt_freq_data_;
      merge 
            _bt_freq_counts (in = in_freq_counts)
            _bt_freq_parms  (in = in_prevfparm)
            ;
      by CCUID;

      if in_freq_counts and in_prevfparm then do;
         output;
      end;
   run;

   /* ------------------------------------------------------------------- */
   /* Step 3: add in net exposure and aggregate the counts to the model   */
   /*         level - same thing we do in freq_fit. Do this cell by cell. */
   /* ------------------------------------------------------------------- */

   %let ccuids = ;
   proc sql noprint;
      select distinct
                  CCUID
      into
                  :CCUIDs separated by " "
      from
                  _bt_freq_data_
      ;
   quit;
   %let ccuid_count = %sysfunc(countw("&CCUIDs."));
   %do i=1 %to &ccuid_count.;
      %let current_ccuid = %scan(&ccuids.,&i.,' ');

      /* Create the cell_where string */
      %let CELL_WHERE = ;
      %orvu_cell_where(
                        ccuid = &current_ccUid., 
                        ccData = varprj.cc
                        )

      data _bt_freq_counts&current_ccuid.;
         set _bt_freq_data_;
         where ccuid eq &current_ccUid.;
         call symput("freq_model_basis_props",freq_model_basis_props);
         call symput("freq_model_is_stratified",freq_model_is_stratified);
         call symput("Freq_Use_Kris",Freq_Use_Kris);
      run;

      /* Now apply the threshold options to get the threshold used to calc. pobs */
      %orva_freq_thresholds(  
                              countdata     = _bt_freq_counts&current_ccuid., 
                              instrataprops = &freq_model_basis_props.,
                              outcountdata  = _bt_freq_counts&current_ccuid.
                              );

      /* For the cell, compute net expore (pobs, exposure, FScale) */
      data _bt_temp_sparms;
         set &sevparms.;
         where ccuid eq &current_ccUid.;
      run;

      %orva_freq_compute_net_exposure( 
                                       in_counts    = _bt_freq_counts&current_ccuid.,
                                       in_losses    = &loss_data.,
                                       cell_where   = &CELL_WHERE.,
                                       dataType     = I,
                                       ccv_property = &freq_model_basis_props.,
                                       stratified   = &freq_model_is_stratified.,
                                       sev_model    = _bt_temp_sparms,
                                       scaleData    = VARPRJ.scalei,
                                       out_counts   = _bt_freq_counts&current_ccuid.,
                                       out_pobs     = _bt_junk 
                                       );

      /* Perform any aggregation of counts. If model is cell level, then 
         counts will be summed over the slicing dimension. Net exposure 
         will be a weighted sum of net exposures */

      /* Will need a list of all the KRIs separated by commas to pass in here */
      %let FREQ_NAMESOFKRIS_comma = ;
      data _null_;

         set &loss_data.;
         if _N_ = 1;
         length FREQ_NAMESOFKRIS_comma $ 1000;
         array KRIs kri:;

         if dim(KRIs) > 0 then do;
            FREQ_NAMESOFKRIS_comma = strip(vname(KRIs[1]));

            do i=2 to dim(KRIs);
               FREQ_NAMESOFKRIS_comma = strip(FREQ_NAMESOFKRIS_comma) || "," || strip(vname(KRIs[i]));
            end;
         end;

         call symput("FREQ_NAMESOFKRIS_comma",FREQ_NAMESOFKRIS_comma);
      run;

      %orva_freq_aggregate_counts( 
                                   in_counts                = _bt_freq_counts&current_ccuid., 
                                   freq_model_is_stratified = &freq_model_is_stratified.,
                                   model_ccv_property       = &freq_model_basis_props.,
                                   keep_KRIs                = &Freq_Use_Kris., 
                                   KRI_names                = %quote(&FREQ_NAMESOFKRIS_comma.), 
                                   out_counts               = _bt_freq_counts&current_ccuid.
                                   );

      /* Abort if any errors were found */
      %if &ORV_ABORT. eq &ORV_TRUE. %then %do;
         %goto DONE;
      %end;

      /* freq_agrgegate_counts only keeps the interval, net_exposure, and count
         info, so add CCUID back in */
      data _bt_freq_counts&current_ccuid.;
         set _bt_freq_counts&current_ccuid.;
         ccuid = &current_ccuid.;
      run;

   %end;

   /* Merge the counts for each cell back together */
   data _bt_freq_counts2;
      set 
      %do i=1 %to &ccuid_count.;
         %let current_ccuid = %scan(&ccuids.,&i.,' ');
         _bt_freq_counts&current_ccuid.
      %end;;
   run;

   data _bt_freq_data_;
      merge 
            _bt_freq_counts2 (in = in_freq_counts)
            _bt_freq_parms  (in = in_prevfparm)
            ;
      by CCUID;

      if in_freq_counts and in_prevfparm then do;
         output;
      end;
   run;

   /* ----------------------------------------------------- */
   /* STEP 4: Get the Param_1 implied by the KRI model and  */
   /*         transform counts with the model cdf.          */
   /* ----------------------------------------------------- */

   data _NULL_;  
      rc = filename("tmp01", , "temp");
   run;

   data _NULL_;
      set _bt_freq_parms end=last;
      where CCUID in (&ccuids.);

      file tmp01;

      length KRI_model_string $ 10000;

      if _N_ = 1 then do;
         put 'data _bt_freq_data_;';
         put '   set _bt_freq_data_;';
      end;
      KRI_model_string = strip(substr(Freq_KRIModel_full,4));

      put '      if missing(intercept) then do;';
      put '          intercept = safe_log(FParam_1);';
      put '      end;';
      put '      if CCUID = ' CCUID 'then do;';
      put '         FParam_1 = net_exposure * exp(intercept + ' KRI_model_string ');';
      put '         %VCDF' DISTRIBUTION '(the_count, p1);';
      if the_count = 0 then do;
         put '      p0 = 0;';
      end;
      else do;
         put '      %VCDF' DISTRIBUTION '(the_count - 1, p0);';
      end;
      put '         u = (p1 - p0) * rand("Uniform") + p0;';
      /* Can't have u = 0 */
      put '         u = max(u, 0.0000001);';
      /* Can't have u = 1 */
      put '         u = min(u, 0.999999);';
      put '      end;';
      if last then do;
         put 'run;';
      end;
   run;

   %inc tmp01;

   proc sort data = _bt_freq_data_;
      by u ;
   run ;

   /* ------------------------------------------------------------ */
   /* STEP 5: Now we have the converted data set - test if this is */
   /*         from a uniform [0,1]                                 */
   /* ------------------------------------------------------------ */

   ods proctitle = off;
   ods select none;   
   %let CAN_REJECT    = %sysfunc(sasmsg(&orvrds, bfd_can_reject_hyp_note, noquote, &orv_backtest_sig_level_freq));
   %let CANNOT_REJECT = %sysfunc(sasmsg(&orvrds, bfd_cant_reject_hyp_note, noquote, &orv_backtest_sig_level_freq));

   /* Generate the dataset with the various tests for uniformity */
/*   ods select none;*/
   ods output GoodnessofFit = _bt_ftest_ ;

   proc univariate data = _bt_freq_data_; 
      var u;
      histogram / BETA(A=1 B=1 THETA=0 SIGMA=1);
   run;

   ods select all;

   /* Determine how many tests were performed.  If we do tests for normality, there
      are 4, for tests against the beta distribution (test for uniformity) there are 2.  
      Because we may change things in the future, better to not hard code the number of 
      tests if we don't have to */
   %let NTests = 0;
   %let dsid = %sysfunc(open(_bt_ftest_, I));
   %if &dsid. ne 0 %then %do;
      %let Ntests = %sysfunc(attrn(&dsid., nlobs));
   %end;
   %let rc = %sysfunc(close(&dsid.));

   /* Empirical cdf values */
   data _bt_freq_data_;
      set _bt_freq_data_ nobs = N;
      emp_u = _N_/(N+1);
   run;

   /* Create text for footnote*/
   %do t = 1 %to &NTests.;
      %let fnstr&t. = .;
   %end;

   data _bt_ftest_;
      set _bt_ftest_;
      length footnotestr $256.;
      count = 'fnstr'||left(trim(_n_));

      if ((psign eq '<') OR (psign eq '-')) and (pvalue <&orv_backtest_sig_level_freq) then do;
         footnotestr = "'"|| TEST ||": &CAN_REJECT'";
      end;
      else do;
         if pvalue > &orv_backtest_sig_level_freq then do;
            footnotestr = "'"||TEST||": &CANNOT_REJECT'";
         end;
         else do;
            footnotestr = "'"||TEST||": &CAN_REJECT'";
         end;
      end;
   run;

   data _null_;
      set _bt_ftest_;
      call symput(count,footnotestr); 
   run;

   /* --------------------------- */
   /* STEP 6: Produce the graph   */
   /* --------------------------- */

   /* Define the template to produce the graph */
   proc template;
      define statgraph orvar.freqBacktest;

      begingraph;

      /* Main Title */
      entrytitle "%sysfunc(sasmsg(&orvrds, bfd_graph_title, noquote))";

      layout overlayequated / 
               equatetype = square

               xaxisopts =(label = "%sysfunc(sasmsg(&orvrds, bfd_uniform_quantiles , noquote))")

               yaxisopts =(label="%sysfunc(sasmsg(&orvrds, bfd_cdf_transform_of_count, noquote))" )

               commonaxisopts = (viewmin=0.0 viewmax=1.0);

         scatterplot
               x = emp_u 
               y = u 
               ;
         lineparm 
               x = 0 
               y = 0 
               slope = 1 / 
               lineattrs = GraphFit
               ;
      endlayout;

    endgraph;
   end;

   /* Create the graph */
   ods graphics on;

   footnote1
      height = 12pt
      color = black
      "%sysfunc(sasmsg(&orvrds, bfd_null_hypothesis_descrip, noquote))"
   ;
   %do g = 1 %to &NTests.;
      %let g2 = %eval(&g. + 1);
      footnote&g2. 
         height = 10pt 
         color  = blue 
         &&fnstr&g.. 
      ; 
   %end;  

   proc sgrender
                 data     = _bt_freq_data_
                 template = orvar.freqBacktest;
   run;


   ods graphics off;
   footnote1;

   %DONE:

   %if &ORV_ABORT. eq &ORV_TRUE. %then %do;
      %put &ORVAR_ERROR. &ORV_REASON.;
      %orv_emptyReport(%quote(&ORV_REASON.));
   %end;

%mend orvr_backtest_freq_dist ;
