/*****************************************************************************
 * Copyright (C), 2012
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orvr_backtest_sev_dist                                      
 *       
 * PURPOSE: Severity distrbution backtesting macro    
 *                 
 * SUPPORT: stmajo, jariou
 *
 * PARAMETERS:
 *     
 *   loss_data - The new losses
 *
 *   sevparms  - The fitted model parameters from the parent project
 *
 * NOTES: 
 *
 *****************************************************************************/
%macro orvr_backtest_sev_dist(
                              loss_data = _bt_losses_,
                              sevparms  = _bt_prevsparm_
                              );

   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put #orvr_backtest_sev_dist(; 
      %put .                        loss_data = &loss_data,;
      %put .                        sevparms  = &sevparms;
      %put .                        );
   %end;

   /* --------------------------------------------------------------------- */
   /* STEP 1: Match each loss observation with the correct model parameters */
   /* --------------------------------------------------------------------- */
   data _bt_sev_data_;
      merge 
            &loss_data.(in = in_sev_losses)
            &sevparms. (in = in_sev_parms)
            ;

      by CCUID;

      if in_sev_losses and in_sev_parms then do;
         output;
      end;

      keep 
         CCUID 
         THRESHOLD 
         AMOUNT 
         KRI_:
         param_:
         distribution
         ;
   run;

   /* Make sure there are a minimum number of losses */
   %let total_sev_obs = %orvu_attrn(_bt_sev_data_, NOBS);
   %if &total_sev_obs. < &orv_backtest_min_obs_for_test. %then %do;
      ods proctitle;
      ods proclabel = "&proclabel.";
      %orv_emptyReport(%sysfunc(sasmsg(&orvrds., bsd_insufficient_losses, noquote)));
      %return;
   %end;

   /* ------------------------------------------------------------------ */
   /* STEP 2: If there are KRIs, adjust the loss amount accordingly.     */
   /*         (This is all done when the model CDF is applied in step 3, */
   /*         really what we are doing is making a script to do this)    */
   /* ------------------------------------------------------------------ */

   /* Get the list of distributions and for each, the scale transform. 
      We will need orvd_emp to have run if we have an empirical distribution, 
      so this is run regardless of whether or not we have KRIs. */
   %let sev_dists = ;
   %let sev_transforms = ;
   proc sql noprint;
      SELECT 
            distribution
      INTO
            :sev_dists separated by " "
      FROM
            _bt_sev_data_
      ;
   quit;
   %do i = 1 %to %sysfunc(countw("&sev_dists."));
      %let dist = %scan(&sev_dists., &i., " ");
      %let scale_transform = ;
      %orva_set_sev_dist( 
                         distIn  = &dist., 
                         distOut = dist 
                         );
      %orvd_&dist.();
      %let sev_transforms = &sev_transforms. &scale_transform.;
   %end;

   /* Now if there are KRIs, write code to adjust the loss amount */
   %if %orvu_var_exists(kri_model, _bt_sev_data_) %then %do;
      %let sev_CCUIDs = ;

      data _bt_sev_data_;
         set _bt_sev_data_;

         length transform $ 10;

         transform = scan("&sev_transforms.", _N_, " ");
      run;

      data _null_;
         rc = filename("krimods", , "temp");
      run;

      data _null_ ;
         file krimods ;
         set _bt_sev_data_ ;
         retain first 1;

         /* Adjust the loss amount and threshold by the KRI model */
         /* Why not just get a correct FParam_1 value? Because then we need to
            worry about the proper scale_transform.  While this can be done 
            easily for 'regular' distributions, there are body/tail 
            distributions where the body has a log transform, so that body and 
            tail scale parameters are adjusted differently.  We can always 
            scale down the CDF by exp(KRI_model) without worrying about any of 
            this. */

         if not missing(kri_model) then do;
            if not first then do;
               put 'else';
            end;
            put 'if ccuid = ' CCUID ' then do;';
            if upcase(strip(transform)) = "ID" then do;
               put '   amount = amount /exp(log(' param_1 ') + ' kri_model ');';
               put '   if not missing(threshold) then do;';
               put '      threshold = threshold /exp(log(' param_1 ') + ' kri_model ');';
               put '   end;' ;
               put '   param_1 = 1;';
            end;
            else do;
               put '   amount = amount /exp(' param_1 ' + ' kri_model ');';
               put '   if not missing(threshold) then do;';
               put '      threshold = threshold /exp(' param_1 ' + ' kri_model ');';
               put '   end;' ;
               put '   param_1 = 0;';
            end;
            put 'end;';
            first = 0;
         end;
      run ;
   %end;

   /* ------------------------------------------------------------ */
   /* STEP 3: Apply the model CDF to the losses (get uniform dist) */
   /* ------------------------------------------------------------ */
   
   data _bt_sev_results_;
      set _bt_sev_data_;

      /* Apply KRI model to amoutns first to */
      %if %orvu_var_exists(kri_model, _bt_sev_data_) %then %do;
         %inc krimods;
      %end;

      /* call cdf function on amount */
      cdf = sevCDF(distribution,amount,&theSevSimCallParams.);

      if THRESHOLD eq . or THRESHOLD eq 0 then do;
         truncdf = 0; 
      end;
      else do;
         /* call cdf function on threshold */
         truncdf = sevCDF(distribution, threshold, &theSevSimCallParams.);
      end;    

      /*Convert into Uniform*/
      if cdf = 1 and truncdf = 1 then do;
         u = 0;
      end;
      else do;
         u =(cdf - truncdf)/(1 - truncdf);
      end;

      if u = 0 then do;
         u = 0.00001;
      end;
      else if u = 1 then do;
         u = 0.99999;
      end;
   run;

   /* --------------------------------- */
   /* STEP 4: Define the graph template */
   /* --------------------------------- */
   proc template;
      define statgraph orvar.sevBacktest;

         begingraph;

            /* Main Title */
            entrytitle "%sysfunc(sasmsg(&orvrds, bsd_graph_title, noquote))";

            layout overlayequated / 
                     equatetype = square

                     xaxisopts =(label = "%sysfunc(sasmsg(&orvrds, bsd_uniform_quantiles , noquote))")

                     yaxisopts =(label="%sysfunc(sasmsg(&orvrds, bsd_cdf_transform_of_losses, noquote))" )

                     commonaxisopts = (viewmin=0.0 viewmax=1.0);

               scatterplot
                     x = u 
                     y = emp_u 
                     ;
               lineparm 
                     x = 0 
                     y = 0 
                     slope = 1 / 
                     lineattrs = GraphFit
                     ;
            endlayout;

         endgraph;
      end;
   run;

   /* ----------------------------------------------------------------------- */
   /* STEP 5: If we are pooling the data, just run the test.  Otherwise, need */
   /*         to detemine which cells can be analyzed separately              */
   /* ----------------------------------------------------------------------- */
   %if %upcase(&orv_backtest_pool_sev_data.) EQ FALSE %then %do;
      /* ----- Figure out which CCUIDs can be tested separately. ----- */
      /* For each CCUID, figure out how many obs there are */
      proc sql noprint;
         CREATE TABLE
                           _bt_nobs_by_ccuid
         AS SELECT DISTINCT 
                           ccuid, 
                           sum(ccuid ne .) as obs_count
         FROM 
                           _bt_sev_results_
         GROUP BY 
                           ccuid
                           ;
      quit;

      /* Divide the CCUIDs into 2 groups: those with enough obs to analyze
         separately and those that need to be pooled */
      data _bt_sufficient_ccuids
           _bt_insufficient_ccuids;

         set _bt_nobs_by_ccuid;

         if obs_count >= &orv_backtest_min_obs_for_test. then do;
            output _bt_sufficient_ccuids;
         end;
         else do;
            output _bt_insufficient_ccuids;
         end;
      run;

      /* If any ccuids have enough obs. to be analyzed individually, get a
         list of them */
      %if %orvu_attrn(_bt_sufficient_ccuids, NOBS) gt 0 %then %do;
         /* Get a list of all the ccuids. */
         %let sufficient_ccuids = ;
         proc sql noprint;
            SELECT DISTINCT
                           ccuid
            INTO
                          :sufficient_ccuids separated by " "
            FROM
                          _bt_sufficient_ccuids
                          ;
         quit;
         
         /* Go through the list of CCUIDs - subset the data and call the 
            report macro */
         %do i = 1 %to %sysfunc(countw("&sufficient_ccuids."));
            %let current_ccuid = %scan(&sufficient_ccuids., &i., ' ');

            data _bt_current_test_data;
               set _bt_sev_results_;
               where ccuid = &current_ccuid.;
            run;

            /* For display, get the name of the CCV's for this CCUID */
            proc sql;
               CREATE TABLE 
                                 _bt_work AS
               SELECT DISTINCT 
                                 * 
               FROM 
                                 varprj.cc cc,
                                 varprj.ccmapval_0 ccv0,
                                 varprj.ccmapval_1 ccv1
                                 %if &THIRD_DIM. %then %do; 
                                    ,varprj.ccmapval_2 ccv2 
                                 %end;
               WHERE 
                                 cc.CCUID = &current_ccuid.
                        AND
                                 cc.CCVUID_0 = ccv0.CCVUID_0 
                        AND 
                                 cc.CCVUID_1 = ccv1.CCVUID_1  
                                 %if &THIRD_DIM. %then %do; 
                                    AND 
                                          cc.CCVUID_2 = ccv2.CCVUID_2 
                                 %end;
                                 ;  
            quit;

            data _null_;
               set _bt_work;
               subtitle = strip(CCVVALUE_0) || " / " || strip(CCVVALUE_1)
               %if &THIRD_DIM %then %do; 
                  || txt_and || strip(CCVVALUE_2) 
               %end;
               ;
               call symput("subtitle",subtitle);
            run;

            %orv_performTestOnSevData(
                                      testData = _bt_current_test_data,
                                      subtitle = %quote(&subtitle.)
                                      )
         %end;
      %end;

      /* For the cells with too few observations, determine if we can pool them 
         to get enough observations for testing */
      %let pool_remaining = &orv_false.;
      %if %orvu_attrn(_bt_insufficient_ccuids, NOBS) gt 0 %then %do;
         data _null_;
            set _bt_insufficient_ccuids end=last;
            retain sum_obs 0;

            sum_obs = sum_obs + obs_count;

            if last and sum_obs >= &orv_backtest_min_obs_for_test. then do;
               call symputX("pool_remaining",&orv_true.);
            end;
         run;

         %if &pool_remaining. eq &orv_true. %then %do;
            %let insufficient_ccuids = ;
            proc sql noprint;
               SELECT DISTINCT
                              ccuid
               INTO
                             :insufficient_ccuids separated by " "
               FROM
                             _bt_insufficient_ccuids
                              ;
            quit;

            data _bt_pooled_test_data;
               set _bt_sev_results_;
               where ccuid in (&insufficient_ccuids.);
            run;

            %let subtitle = %sysfunc(sasmsg(&orvrds., bsd_pooled_losses, noquote, &orv_backtest_min_obs_for_test.));
            %orv_performTestOnSevData(
                                      testData = _bt_pooled_test_data,
                                      subtitle = &subtitle.
                                      )
         %end;
      %end;
   %end; /* %upcase(orv_backtest_pool_sev_data.) EQ FALSE */ 
   %else %do;
      %let subtitle = %sysfunc(sasmsg(&orvrds., bsd_all_losses, noquote));
      /* Call the report macro on the whole dataset */
      %orv_performTestOnSevData(
                                testData = _bt_sev_results_,
                                subtitle = &subtitle.
                                );
   %end;

%mend orvr_backtest_sev_dist ;

%macro orv_performTestOnSevData(
                                 testData = _bt_sev_results_,
                                 subtitle = 
                                 );
   %if %upcase(&ORV_DEBUG) EQ TRUE %then %do;
      %put ##orv_performTestOnSevData(; 
      %put .                          testData = &testData,;
      %put .                          subtitle = &subtitle,;
      %put .                          );
   %end;

   /* This performs the test for uniformity - results of test are stored in 
      _bt_stest_ */
   ods proctitle = off;
   ods select none;
   ods output GoodnessofFit = _bt_stest_ ;

   proc univariate data = &testData.; 
      var u ;
      histogram / BETA(A = 1 B = 1 THETA = 0 SIGMA = 1);
   run;

   ods select all;

   /* Sort so that a QQ plot can be created */
   proc sort 
            data =  &testData. 
            out  = _bt_sev_sorted_values_ 
            ;
      by u;
   run;

   /* Determine how many tests were performed.  If we do tests for normality, there
      are 4, for tests against the beta distribution (test for uniformity) there are 2.  
      Because we may change things in the future, better to not hard code the number of 
      tests if we don't have to */
   %let NTests = %orvu_attrn(_bt_stest_, NOBS);

   /* Create text for footnote*/
   %do t = 1 %to &NTests.;
      %let fnstr&t. = .;
   %end;

   data _null_;
      set _bt_stest_;
      length footnotestr $256.;
      count = 'fnstr'||left(trim(_n_));

      if ((psign eq '<') OR (psign eq '-')) and (pvalue < &orv_backtest_sig_level_sev.) then do;
         footnotestr = "'"|| TEST ||": %sysfunc(sasmsg(&orvrds, bsd_can_reject_hyp_note, noquote, &orv_backtest_sig_level_sev))' ";
      end;
      else do;
         if pvalue > &orv_backtest_sig_level_sev. then do;
            footnotestr = "'"||TEST||": %sysfunc(sasmsg(&orvrds, bsd_cant_reject_hyp_note, noquote, &orv_backtest_sig_level_sev))' ";
         end;
         else do;
            footnotestr = "'"||TEST||": %sysfunc(sasmsg(&orvrds, bsd_can_reject_hyp_note, noquote, &orv_backtest_sig_level_sev))' ";
         end; 
      end;

      call symput(count, footnotestr);

   run;

   /* STEP 5: Create the Report*/
   /* This is the graph at the top of the report:
      QQ plot of the transformed amounts (applied model CDF) and empirical quantiles */

   /* Print which data we are testing */
   title6 &titleSpec4. "&subtitle.";   

   /* Prepare data for plotting */
   proc sort data = &testData.;
      by u;
   run;

   /* Use Kaplan-Meier algorithm to get empirical cdf values */
   /* Out variables are 'amount' and 'cdf'.  Amount is the transformed loss 
      amounts. */
   %orva_cdf_kaplan_meier( 
                          indata            = &testData.,
                          outdata           = _bt_km, 
                          variable          = cdf,
                          threshold         = truncdf,
                          threshold_percent = 0
                          );
   
   data _bt_km;
      set _bt_km;
      rename cdf = emp_u;
   run;

   data _bt_plotdata;
      set _bt_km(keep = emp_u);
      set &testData.(keep = u);
   run;

   ods graphics on;

   footnote1
      height = 12pt
      color = black
      "%sysfunc(sasmsg(&orvrds, bsd_null_hypothesis_descrip, noquote))"
      ;
   %do g = 1 %to &NTests.;
      %let g2 = %eval(&g. + 1);
      footnote&g2. 
         height = 10pt 
         color  = blue 
         &&fnstr&g.. 
         ; 
   %end;  

   proc sgrender
                 data = _bt_plotdata
                 template = orvar.sevBacktest
                 ;
   run;

   ods graphics off;
   footnote1;
%mend orv_performTestOnSevData;
