/*-------------------------------------------------------------------------------
 * Copyright (C), 2008
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orvr_freqfit_stats
 * PURPOSE: create frequency fitting statistical report.
 * SUPPORT: stchri
 *
 * INPUT:
 *    fparm       - frequency fit subset result from fparm_i(c) with iteration=&iteration.
 *    fparmcov    - frequency fit subset cov matrix: output from %orva_subset_fparmcov
 *    metalib     - name of the Meta libname.
 *    interval    - project interval.
 *    fileref     - fileref for HTML output, ignored if blank
 *    dataset     - FPARM dataset.
 *    ouputtype   - any valid ods output type (default=HTML)
 *    style       - any valid ods template style (default=Statdoc)
 *    title       - title of current cell
 *    clearTitles - defaults to TRUE
 *------------------------------------------------------------------------------*/
 %macro ORVR_freqfit_stats(
                           fparm       = ,
                           fparmcov    = ,
                           metalib     = ,
                           interval    = ,
                           fileref     = ,
                           outputType  = HTML,
                           style       = ,
                           title       = "", 
                           clearTitles = true,
                           ccLibref    = ,
                           fparmInfo   = 
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orvr_freqfit_stats(;
      %put .                    fparm       = &fparm.,;
      %put .                    fparmcov    = &fparmcov.,;
      %put .                    metalib     = &metalib.,;
      %put .                    interval    = &interval.,;
      %put .                    fileref     = &fileref.,;
      %put .                    outputType  = &outputType.,;
      %put .                    style       = &style.,;
      %put .                    title       = &title.,;
      %put .                    clearTitles = &clearTitles.,;
      %put .                    ccLibref    = &ccLibref.,;
      %put .                    fparmInfo   = &fparmInfo.;
      %put .                    );
   %end;

   %let COVFMT  = NLNUM32.7;
   %let FREQFMT = NLNUM32.5;

   /*** fix for template issue where can't overwrite template in sasuser area ***/
   ods path(prepend) work.xxx(update);

   /*- get kri dataset names from META library -*/
   %let kribl   = &metalib..kribl_&interval;
   %let krifirm = &metalib..krifirm_&interval;

   /*- initialize titles -*/
   %if &cleartitles eq true %then %do;
      title1; 
      title2; 
      title3; 
      title4; 
      title5; 
      title6; 
   %end;
   title7; 
   title8;

   /*- setup OUTPUT file -*/
   %if %substr(%upcase(&outputtype), 1, 3) = PDF %then %do;
      options 
               orientation  = landscape 
               leftmargin   = 0.01in 
               rightmargin  = 0.01in 
               topmargin    = 0.01in 
               bottommargin = 0.01in
               ;
      title;
   %end;

   %if &fileref ^= "" %then %do;
      ods &outputtype 
                     body  = &fileref 
                     style = &style
                     ;
      ods listing close;
   %end;

   /*- Get the frequency distribution from fparm -*/
   /*- There are two intercepts KRI_intercept=intercept -*/
   /*- Drop KRI_intercept to facilitate KRI counting   -*/
   %let dist = %str();

   data _ffs_fparm;
      set &fparm;
      drop KRI_intercept;
      call symput("dist", distribution);   
   run;

   /*- initialize the appropriate distribution -*/
   %orvd_&dist;

   data _NULL_;
      FreqName = FreqName("&dist");
      call symput("Dist_Name", FreqName);
   run;

   /*- do we have covariance matrix data? -*/
   %let fparmcov_nobs = 0;

   %if &fparmcov ne %str() %then %do;     
      %let dsid = %sysfunc(open(&fparmcov, I));

      %if &dsid ^= 0 %then %do;
         %let fparmcov_nobs = %sysfunc(attrn(&dsid, nlobs));
         %let rc = %sysfunc(close(&dsid));
      %end;
   %end;

   /*** fixed up the covariance matrix ***/
   %let ffs_covmat_status = 0;   
   %let ALL_CNT = 0;

   %if &fparmcov_nobs ^= 0 %then %do;      
      %orva_subset_fparmcov(
                            fcovOut     = _ffs_covmat,
                            fcovDataset = &fparmcov,
                            kribl       = &kribl,
                            krifirm     = &krifirm,
                            statusOut   = _ffs_status,
                            ccLibref    = &ccLibref,
                            metaLibref  = &metalib
                            );

      data _NULL_; 
         set _ffs_status;
         call symput('ffs_covmat_status', status);
      run;

      %if &ffs_covmat_status eq 0 %then %do;
         data _NULL_; 
            set _ffs_covmat end = LAST;
            length allm all $ 64;
            length stdm std $ 64;                             
            all_cnt + 1;
            all = strip(_NAME_); 
            std = ' STD_' || strip(_NAME_); 

            allm = "ALL_NAME_" || strip(put(all_cnt, best.));
            call symput(allm, strip(all));

            stdm = "STD_NAME_" || strip(put(all_cnt, best.));
            call symput(stdm, strip(std));

            if LAST then do;
               call symputX('ALL_CNT', all_cnt);
            end;
         run;          
      %end;

      /*- clean up -*/
      %orvu_delete_data( data = _ffs_status);
   %end;

   /**************************************************/
   /**** Distribution Parameter Estimates Report ****/
   /**************************************************/
   %let fparminfo_nobs = 0;

   %if &fparminfo ne %str() %then %do;     
      %let dsid = %sysfunc(open(&fparminfo, I));

      %if &dsid NE 0 %then %do;
         %let fparminfo_nobs = %sysfunc(attrn(&dsid, nlobs));
         %let rc = %sysfunc(close(&dsid));
      %end;
   %end;

   %if &fparminfo_nobs NE 0 %then %do;
      %let PARM_CNT = 0;

      data _NULL__; 
         set &fparmInfo end = LAST; 
         where upcase(Distribution) eq upcase("&dist");
         retain cnt 0;
   
         length str $ 32;
         cnt + 1;

         str = "FREQ_PARM_" ||  strip(put(cnt,best.));
         call symput(str, strip(DatasetName));

         str = "FREQ_PARM_LABEL_" || strip(put(cnt,best.));
         call symput(str, strip(DisplayName));

         if LAST then do;
            call symputX('PARM_CNT',cnt);
         end;
      run;
   %end;
   %else %do;
      %put Input parameter data set fparmInfo is missing or empty.;
      %goto DONE;
   %end;

   %let _ffs_param_est_title = %sysfunc(sasmsg(&orvrds, ffs_param_est_title, noquote, &dist_name));
   %orvr_cover_page(
                     outputtype   = &outputtype, 
                     report_title = &_ffs_param_est_title,
                     style        = &style
                     );

   %if &cleartitles eq true %then %do;
      title1 height = 12pt "&_ffs_param_est_title";
      title2 &title;
   %end;
   %else %do;
      title6 height = 10pt "&_ffs_param_est_title";
   %end;

   data _ffs_report (keep=Parameter Estimate );
      format Parameter $12.;
      format Estimate &FREQFMT.;
      set &fparm;

      %do i = 1 %to &PARM_CNT;
         Parameter = "&&FREQ_PARM_LABEL_&i";
         Estimate  = &&FREQ_PARM_&i;
         output;
      %end;
   run;

   proc print 
               data = _ffs_report 
               noobs 
               label 
               split = "*"
               ;
      var 
            Parameter 
            Estimate 
            ;
      label 
            Parameter = "%sysfunc(sasmsg(&orvrds, ffs_parameter_label, noquote))"
            Estimate  = "%sysfunc(sasmsg(&orvrds, ffs_estimate_label, noquote))"
            ; 
   run;

   /************************************/
   /**** Display Moments Estimates *****/
   /************************************/
   %let _ffs_mom_est_title = %sysfunc(sasmsg(&orvrds, ffs_mom_est_title, noquote, &dist_name));

   %if &cleartitles eq true %then %do;
      title1 height = 12pt "&_ffs_mom_est_title";
      title2 &title;
   %end;
   %else %do;
      title6 height = 10pt "&_ffs_mom_est_title";
   %end;

   data _ffs_report (keep = Name Value);
      format name $12.;
      format Value &FREQFMT.;
      set _ffs_fparm;

      name  = "%sysfunc(sasmsg(&orvrds, ffs_mean_name, noquote))";
      value = mean;
      output;

      name  = "%sysfunc(sasmsg(&orvrds, ffs_variance_name, noquote))";
      value = variance;
      output;
   run;

   proc print data =_ffs_report noobs label;
      label name  = '00'x;
      label value = "%sysfunc(sasmsg(&orvrds, ffs_statistic_label, noquote))";
   run;

   /********************************************/
   /**** Display Goodness of Fit Statistics ****/
   /********************************************/
   %let _ffs_good_stat_fit_title = %sysfunc(sasmsg(&orvrds, ffs_good_stat_fit_title, noquote, &dist_name));

   %if &cleartitles eq true %then %do;
      title1 height = 12pt "&_ffs_good_stat_fit_title";
      title2 &title;
   %end;
   %else %do;
      title6 height = 10pt "&_ffs_good_stat_fit_title";
   %end;
    
   data _ffs_report (keep=Name Value);
      format name $30.; 
      format Value &FREQFMT.;
      set _ffs_fparm;

      name  = "%sysfunc(sasmsg(&orvrds, ffs_crit_akinfo_label, noquote))";
      value = aic;
      output;

      name  = "%sysfunc(sasmsg(&orvrds, ffs_crit_sch_bay_label, noquote))";
      value = sbc;
      output;

      name   = "%sysfunc(sasmsg(&orvrds, ffs_log_likely_label, noquote))";
      value  = logl;
      output;
   run;

   proc print data = _ffs_report noobs label;
      label name  = '00'x;
      label value = "%sysfunc(sasmsg(&orvrds, ffs_statistic_label, noquote))";
   run;

   /***************************************/
   /*** handle covariance matrix stuff  ***/
   /***************************************/
   %if &ffs_covmat_status eq 0 %then %do;
      /* Get the standard deviation for variables */
      proc iml;
         use _ffs_covmat;
         read all var {
                        %do i = 1 %to &ALL_CNT; 
                           &&ALL_NAME_&i. 
                        %end;
                     } 
                     into cov;
         close _ffs_covmat;

         dim    = nrow(cov);
         StdErr = j(1,dim,.);

         /* Need to take care of negative variances */
         /* because of rounding error */
         do i = 1 to dim;
            if cov[i,i]<0 then do;
               StdErr[i] = .;
            end;
            else do;
               StdErr[i] = sqrt(cov[i,i]);
            end;
         end;

         create _ffs_Stddata from StdErr[colname = {
                                                      %do i = 1 %to &ALL_CNT; 
                                                         &&STD_NAME_&i. 
                                                      %end;
                                                      }
                                                      ];
         append from StdErr;
      quit; 

      /* Merge standard error to fparm data set*/
      data _ffs_fparmstd;
         merge 
               _ffs_fparm 
               _ffs_Stddata
               ;
      run;

      data _ffs_report (keep = Variable Estimate StdErr TValue probt);
         format Variable $32. ;
         set _ffs_fparmstd;
         array VARS %do i=1 %to &ALL_CNT; &&ALL_NAME_&i. %end;;
         array STD  %do i=1 %to &ALL_CNT; &&STD_NAME_&i. %end;;

         do i = 1 to DIM( VARS );
            if ( VARS[i] ^= . ) then do;  
               Variable = vname( VARS[i]);
               Estimate = VARS[i];      
               StdErr = STD[i];

               if StdErr ~= 0 then do;
                  TValue = Estimate/StdErr;
               end;
               else do;
                  TValue = .;
               end;      

               probt = (1 - probt(abs(tvalue), Freq_ObsCount - &ALL_CNT))*2;
               output;
            end;
         end;
      run;

      /************************************************/
      /* Regression Parameter Estimates Report        */
      /***********************************************/
      %let _ffs_regression_title = %sysfunc(sasmsg(&orvrds, ffs_regression_title, noquote));
         %if &cleartitles eq true %then %do;
            title1 height = 12pt "&_ffs_regression_title";
            title2 &title;
         %end; 
         %else %do;
            title6 height = 10pt "&_ffs_regression_title";
         %end;

         /* add labels for DUMMYs */
         %orva_add_label_column(
                                column     = Variable, 
                                dsname     = _ffs_report,
                                prefix     = DUMMY_INTERNAL_BL_UID_,
                                prefixL    = 22,
                                ccLibref   = &ccLibref, 
                                type       = BLUID, 
                                metaLibref = &metalib
                                );
         /* add labels for KRIs */
         %orva_add_label_column(
                                column     = Variable, 
                                dsname     = _ffs_report,
                                prefix     = , 
                                prefixL    = 0, 
                                ccLibref   = &ccLibref, 
                                type       = KRI,
                                kribl      = &kribl, 
                                krifirm    = &krifirm, 
                                metaLibref = &metalib
                                );

         /* add labels for parameters */
         data _ffs_report;
            length _LABEL_ $ 256;
            set _ffs_report;

            %do i = 1 %to &PARM_CNT;
               if upcase(Variable) eq "&&FREQ_PARM_&i" then do;
                  _LABEL_ = "&&FREQ_PARM_LABEL_&i";
               end;
            %end;

            if _LABEL_ eq "" then _LABEL_ = Variable;
         run;
        
         proc template;
            define table orvar.FreqParamEstimates;
               column _LABEL_ Variable Estimate StdErr TValue probt;

               define _LABEL_;   
                  header = "%sysfunc(sasmsg(&orvrds, ffs_header_label, noquote))";  
               end;  

               define Variable;  
                  header = "%sysfunc(sasmsg(&orvrds, ffs_header_variable, noquote))"; 
               end;  

               define Estimate;  
                  header = "%sysfunc(sasmsg(&orvrds, ffs_header_estimate, noquote))";
                  format = &FREQFMT.; 
               end;  

               define StdErr;    
                  header = "%sysfunc(sasmsg(&orvrds, ffs_header_stdErr, noquote))";
                  format = &FREQFMT.; 
               end;  

               define TValue;    
                  header = "%sysfunc(sasmsg(&orvrds, ffs_header_tValue, noquote))";
                  format = &FREQFMT.; 
               end;  

               define probt;     
                  header = "%sysfunc(sasmsg(&orvrds, ffs_header_probt, noquote))";
                  format = &FREQFMT.; 
               end;  
            end;
         run;

         data _NULL_; 
            set _ffs_report;           
            file print ods = (template = 'orvar.FreqParamEstimates');
            put _ods_;
         run;           

         /*************************************/
         /* Covariance Matrix Report          */
         /*************************************/
         %let _ffs_cov_mat_title = %sysfunc(sasmsg(&orvrds, ffs_cov_mat_title, noquote, &dist_name));

         %if &ffs_covmat_status eq 0 %then %do;
            %if &cleartitles eq true %then %do;
               title1 height = 12pt "&_ffs_cov_mat_title";
               title2 &title;
            %end;
            %else %do;
               title6 height = 10pt "&_ffs_cov_mat_title";
            %end;

            proc template;
               define table work.orvar_covmat;
                  column _LABEL_ _NAME_ 
                                       %do i = 1 %to &ALL_CNT; 
                                          &&ALL_NAME_&i. 
                                       %end; 
                                       ;

                  define _LABEL_;   
                     header = "%sysfunc(sasmsg(&orvrds, ffs_header_label, noquote))";  
                  end;  

                  %do i = 1 %to &ALL_CNT;
                     define &&ALL_NAME_&i.;   
                        format = &COVFMT.;    
                     end;  
                  %end;
               end;
            run;

            /* add labels for parameters */
            data _ffs_covmat;
               length _LABEL_ $ 256;
               set _ffs_covmat;

               %do i = 1 %to &PARM_CNT;
                  if upcase(_NAME_) eq "&&PARM_&i" then do;
                     _LABEL_ = "&&LABEL_&i";
                  end;
               %end;

               if _LABEL_ eq "" then _LABEL_ = _NAME_;
            run;

            data _NULL_; 
               set _ffs_covmat;           
               file print ods = (template = 'work.orvar_covmat');
               put _ods_;
            run; 
        %end;
     %end; /* have fparmcov */

      %if &cleartitles eq true %then %do;
         title1; 
         title2;
      %end;
%DONE:
   /* close OUTPUT file */
   %if &fileref ^= "" %then %do;
      ods &outputtype close;
      ods listing;
   %end;

   %orvu_delete_data( data = _ffs_:);
%mend ORVR_freqfit_stats;
