/*-------------------------------------------------------------------
 *  Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME    : orva_var_calc.sas  
 * PURPOSE : Produces report for VaR Criteria, VaR Final Matrix and VaR Results     
 * SUPPORT : stchri 
 *
 * Input:
 *    fileref - fileref for HTML output, ignored if blank
 *    model_uid - UID for model.
 *    options - specifies which parts of the report to produce, 
 *              separated by spaces:
 *       VINS - Final Insurance Values. Both Primary and Secondary.
 *       VCRIT - VaR Calculation Criteria (options)
 *       VMAT - Final Matrix for VaR
 *       VFEFFECTS - Frequency Effects 
 *       VRES_WITHOUT  - VaR Result Matrix without insurance
 *       VRES_INSURANCE  - VaR Result Matrix with insurance
 *       VRESBASEL_WITHOUT  - Basel VaR Result Matrix without insurance
 *       VRESBASEL_INSURANCE - Basel VaR Result Matrix with insurance
 *    outputtype - html, pdf, etc.
 *    style - template style
 *    data_root - The root path to the data store, ending with the file separator character
 *
 * Required tables:
 *   vardata.VAR_USER
 *   varorg.INSURANCE
 *   varprj.MODEL
 *   varprj.VAR_CRITERIA
 *   varprj.VAR_MATRIX
 *   varprj.VAR_RESULT_MATRIX
 *   varprj.ccmapval_0
 *   varprj.ccmapval_1
 *   varprj.cc
 *   varprj.aggstat_naked
 *   varprj.aggstat_insured
 *   varprj.varstat_naked
 *   varprj.varstat_insured
 *   varprj.aggstatbasel_naked
 *   varprj.aggstatbasel_insured
 *
 * Notes: Assumes the caller has set the macro varaible _VARLIB to point
 *        to the client install area's admin folder.  This is where
 *        various sas graphics jars are installed and must be set as the
 *        codebase for graphic reports.
 *-------------------------------------------------------------------*/ 
%macro ORVR_var_calc(
                     fileref      = ,
                     model_uid    = ,
                     options      = ,
                     outputType   = HTML,
                     style        = DEFAULT,
                     data_root    =
                     );

   %orvr_report_setup(
                        fileref    = &fileref,
                        model_uid  = &model_uid,
                        outputType = &outputType,
                        style      = &style,
                        data_root  = &data_root
                        );
                      
   /* Check to see if we need a global title ---*/
   %let nreports = %sysfunc( countw("&options"));

   %if &nreports > 1 %then %do;
      %orvr_cover_page( 
                        outputtype   = &outputtype, 
                        report_title = %sysfunc(sasmsg(&orvrds, vc_global_title, noquote)),
                        style        = &style
                        );  
   %end;

   /******************************************/
   /********** formats needed below **********/
   /******************************************/
   proc format;
      value 
         BooleanYN   
                     1 = 'Yes'
                     0 = 'No'
                     ;
   run;
   
   proc format;
      value VarMixingSelection   
            1 = "%sysfunc(sasmsg(&orvrds, vc_mix_sel_value_internal_only, noquote))"
            2 = "%sysfunc(sasmsg(&orvrds, vc_mix_sel_value_relative_relations, noquote))"
            3 = "%sysfunc(sasmsg(&orvrds, vc_mix_sel_value_consortium_only, noquote))"
            5 = "%sysfunc(sasmsg(&orvrds, vc_mix_sel_value_simulation_mixing, noquote))"
            ;
   run;
      
   proc format;
      value ConfLevel  low-high = [best5.1];
   run;

   proc format;
      value $CorrType 
                     'U' = "%sysfunc(sasmsg(&orvrds, vc_corr_type_value_user_defined, noquote))"
                     'T' = "%sysfunc(sasmsg(&orvrds, vc_corr_type_value_internal, noquote))"
                     'I' = "%sysfunc(sasmsg(&orvrds, vc_corr_type_value_shared_internal, noquote))"
                     'C' = "%sysfunc(sasmsg(&orvrds, vc_corr_type_value_shared_consortium, noquote))"
                     'O' = "%sysfunc(sasmsg(&orvrds, vc_corr_type_value_identity, noquote))"
                     ;
   run;

   proc format;
      value $DataType 
                     'I' = "%sysfunc(sasmsg(&orvrds, vc_data_type_value_internal, noquote))"
                     'M' = "%sysfunc(sasmsg(&orvrds, vc_data_type_value_mixed, noquote))"
                     'S' = "%sysfunc(sasmsg(&orvrds, vc_data_type_value_scenario, noquote))"
                     'P' = "%sysfunc(sasmsg(&orvrds, vc_data_type_value_public, noquote))"
                     'C' = "%sysfunc(sasmsg(&orvrds, vc_data_type_value_consortium, noquote))"
                     ;
   run;

   proc format;
      value $IntervalType 
                        'YEAR'      = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_YEAR, noquote))"
                        'SEMIYEAR'  = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_SEMIYEAR, noquote))"
                        'WEEK'      = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_WEEK, noquote))"
                        'WEEK2'     = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_WEEK2, noquote))"
                        'SEMIMONTH' = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_SEMIMONTH, noquote))"
                        'MONTH'     = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_MONTH, noquote))"
                        'MONTH2'    = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_MONTH2, noquote))"
                        'QTR'       = "%sysfunc(sasmsg(&orvrds, vc_intv_type_value_QTR, noquote))"
                        ;
   run;

   /********************************************************/
   /******************** Criteria Data *********************/
   /********************************************************/
   data Crit;
      set varprj.VAR_CRITERIA(where = (MODEL_UID = &model_uid));
      txt = sasmsg("&orvrds", "vc_percentile", "noquote");
      call symput("CONF1", compress(CONFIDENCE_LEVEL_1) || txt );
      call symput("CONF2", compress(CONFIDENCE_LEVEL_2) || txt );
      call symput("CONF3", compress(CONFIDENCE_LEVEL_3) || txt );
      call symput("SIMULATION_MIXING", MIXING_SELECTION);
      call symputx("DFM_TYPE", DEFAULT_MAXLOSS_CALC_TYPE);
   run;
   
   data _NULL_; 
      set &VARMETA..internal_metadata;
      call symputx("DEFAULT_MAX_LOSS", MAX_LOSS_AMOUNT);
   run;

   data Crit; 
      set Crit; 
      DEFAULT_MAX_LOSS = &DEFAULT_MAX_LOSS;
   run;
      
   /****************************************************/
   /* VaR Insurance    *********************************/
   /****************************************************/
   %if %index(&options, VINS) ^= 0 %then %do;
      /********** Primary Insurance **********/
      %let vc_pri_ins_prob   = %sysfunc(sasmsg(&orvrds, vc_pri_insure_prob_varcalc_title, noquote));
      %let vc_finalmat_prins = %sysfunc(sasmsg(&orvrds, vc_rep_var_mat_pri_insure_title,  noquote));

      %if &nreports < 2 %then %do;
         %orvr_cover_page( 
                           outputtype   = &outputtype, 
                           report_title = &vc_finalmat_prins,
                           style        = &style
                           );  
      %end;

      footnote1; 
      %orv_clearReportTitles; 
      title2 &titleSpec2 "&vc_finalmat_prins";
      title3 &titleSpec4 "&vc_pri_ins_prob";
      title4 &titleSpec3 "(&project)";
      title5 &titleSpec3 "(&author)";   
      ods noproctitle;
      ods proclabel="&vc_finalmat_prins";   

      %orvu_combine_data_with_ccv_names(dataIn = varprj.insure_1, dataOut = _work0,
                  third_dim=&THIRD_DIM);

      *** Get Insurance Policy names;
      proc sql noprint;
         CREATE TABLE 
                        _work AS
         SELECT 
                        * 
         FROM 
                        _work0 AS I 
         LEFT JOIN 
                        &varorg..INSURANCE AS N          
         ON  
                        I.INSURANCE_1 eq N.INSURANCE_UID 
                        ; 
      quit;

      *** dump the Nones!;
      data _work; 
         set _work;
         where INSURANCE_1 ne 0;

         %if &THIRD_DIM %then %do;
            rename CCVVALUE_2 = &MAPNAME2;
         %end;
      run;

      * Sort by descending CCVUID_0 in order to list 'Total' last  LH 2/2/07;
      * Sort by INSURANCE_NAME to list insurance companies alphabetically LH 2/2/07;  
      proc sort data = _work; 
         by  
            %if &THIRD_DIM %then %do;
               CCVUID_2 
            %end;  
            descending CCVUID_0 
            CCVVALUE_1 
            INSURANCE_NAME
            ; 
      run;
      
      *** run the report;
      proc template;
         define table orvar_insurance;
            header h1;
            column 
                  (CCVVALUE_0) 
                  CCVVALUE_1 
                  INSURANCE_NAME 
                  PIns1 
                  PInsCurrent1
                  ;
            define h1;
               text "%sysfunc(sasmsg(&orvrds, class_title, noquote)) "; 
               start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
            end;                   

            define CCVVALUE_0;
               header = "&MAPNAME0" blank_dups = on id;
            end; 

            define CCVVALUE_1;
               header = "&MAPNAME1" id;
            end; 

            define INSURANCE_NAME;
               header = "%sysfunc(sasmsg(&orvrds, vc_ins_policy_header, noquote))";
            end; 

            define PIns1;
               header = "%sysfunc(sasmsg(&orvrds, vc_default_header, noquote))";
               print  = false;
            end;  
   
            define PInsCurrent1;
               header = "%sysfunc(sasmsg(&orvrds, vc_prob_ins_header, noquote))";
               cellstyle PIns1 ^= PInsCurrent1 as data
                                                {
                                                   foreground = red 
                                                   font_style = italic 
                                                   font_weight = bold
                                                   },
                                                   1 as data
                                                   ;
            end;

            define footer table_footer;
               text "%sysfunc(sasmsg(&orvrds, vc_red_note, noquote))";
               style = header{foreground = red font_style = italic font_weight = bold};
               ;
            end;
         end;  
      run;

      %if &THIRD_DIM %then %do;
         %do i = 1 %to &num_ccv2;
            %let name = %scan(&ccv_names2, &i, "#");
            title6 "&MAPNAME2_ORIG = &name";

            data _NULL_;
               set _work;
               where &mapname2 = "&name";
               file print ods = (template = 'orvar_insurance');
               put _ods_;
            run; 
         %end;
      %end;
      %else %do;
         data _NULL_;
            set _work;
            file print ods = (template = 'orvar_insurance');
            put _ods_;
         run;   
      %end;  

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;
 
      /********** Secondary Insurance **********/
      %let vc_sec_ins_prob    = %sysfunc(sasmsg(&orvrds, vc_sec_insure_prob_varcalc_title, noquote));
      %let vc_finalmat_secins = %sysfunc(sasmsg(&orvrds, vc_rep_mat_sec_insure_title,      noquote));
      footnote1; 
      %orv_clearReportTitles; 
      title2 &titleSpec2 "&vc_finalmat_secins";
      title3 &titleSpec4 "&vc_sec_ins_prob";
      title4 &titleSpec3 "(&project)";
      title5 &titleSpec3 "(&author)";
      ods noproctitle;
      ods proclabel = "&vc_finalmat_secins";

      %orvu_combine_data_with_ccv_names( dataIn = varprj.insure_2, dataOut = _work0, third_dim=&THIRD_DIM);
      
      *** Get Insurance Policy Names;
      proc sql noprint;
         CREATE TABLE 
                        _work AS
         SELECT 
                        I.*, 
                        N.INSURANCE_NAME 
         FROM 
                        _work0 AS I 
         LEFT JOIN 
                        &varorg..INSURANCE AS N          
               ON  
                        I.INSURANCE_1 eq N.INSURANCE_UID
                        ; 
      quit;

      data _work0; 
         set _work (rename = (INSURANCE_NAME = INSURANCE_NAME_1)); 
      run;

      proc sql noprint;
         CREATE TABLE 
                        _work AS
         SELECT 
                        I.*, 
                        N.INSURANCE_NAME 
         FROM 
                        _work0 AS I 
         LEFT JOIN 
                        &varorg..INSURANCE AS N          
         ON  
                        I.INSURANCE_2 eq N.INSURANCE_UID
                        ; 
      quit;

      *** Dump the Nones;
      data _work; 
         set _work;
         where INSURANCE_1 ne 0 and INSURANCE_2 ne 0;
      run;

      * Sort by descending CCVUID_0 in order to list 'Total' last  LH 2/2/07;
      * Sort by insuracne names to list insurance companies alphabetically LH 2/2/07; 
      proc sort data = _work; 
         by  
            %if &THIRD_DIM %then %do;
               CCVUID_2 
            %end; 
            descending CCVUID_0 
            CCVVALUE_1 
            INSURANCE_NAME_1 
            INSURANCE_NAME
            ; 
      run; 
   
      *** Run Report;
      proc template;
         define table orvar_insurance_2;
            header h1;
            header h2;
            column  
                     (CCVVALUE_0) 
                     CCVVALUE_1
                     INSURANCE_NAME_1 
                     INSURANCE_NAME
                     PIns2 
                     PInsCurrent2
                     ;
                      
            define h1;
               text "%sysfunc(sasmsg(&orvrds, class_title, noquote))" ; 
               start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
            end;                           

            define CCVVALUE_0;
               header = "&MAPNAME0" blank_dups = on id;
            end; 

            define CCVVALUE_1;
               header = "&MAPNAME1";
            end; 

            define h2;
               text "%sysfunc(sasmsg(&orvrds, vc_ins_policies_header, noquote))"; 
               start = INSURANCE_NAME_1  space = 1 expand = '-' end = INSURANCE_NAME;
            end;     

            define INSURANCE_NAME_1;
               header = "%sysfunc(sasmsg(&orvrds, vc_prim_header, noquote))";
            end; 

            define INSURANCE_NAME;
               header = "%sysfunc(sasmsg(&orvrds, vc_secon_header, noquote))";
            end; 
               
            define PIns2;
               header = "%sysfunc(sasmsg(&orvrds, vc_default_header, noquote)) ";
               print  = false;
            end;        
       
            define PInsCurrent2;
               header="%sysfunc(sasmsg(&orvrds, vc_prob_ins_header, noquote)) ";
               cellstyle 
                        PIns2 ^= PInsCurrent2 as data
                        {foreground = red font_style = italic font_weight = bold},
                        1 as data;
            end;

            define footer table_footer;
               text "%sysfunc(sasmsg(&orvrds, vc_red_note, noquote)) ";
               style = header{foreground=red font_style=italic font_weight=bold};
            ;
            end;
         end;  
      run;

      %if &THIRD_DIM %then %do;
         %do i = 1 %to &num_ccv2;
            %let name = %scan(&ccv_names2, &i, "#");
            title6 "&MAPNAME2_ORIG = &name";

            data _NULL_;
               set _work;
               where CCVVALUE_2 = "&name";
               file print ods = (template = 'orvar_insurance_2');
               put _ods_;
            run; 
         %end;
      %end;
      %else %do;
         data _NULL_; 
            set _work;
            file print ods = (template = 'orvar_insurance_2');
            put _ods_;
         run; 
      %end; 

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;
   %end;
   /****************************************************/
   /* END VaR Insurance Matrix *************************/
   /****************************************************/
   
   /****************************************************/
   /* VaR FINAL Matrix *********************************/
   /****************************************************/
   %if %index(&options, VMAT) ^= 0 %then %do;
      /*** Internal Only ***/                       
      %if &SIMULATION_MIXING eq 1 %then %do;
         %let _final_sparm = varprj.selsparmi;
         %let _final_fparm = varprj.selfparmi; 
         %let MIXED  = false;
         %let SIMMIX = false;
      %end;
      /*** Relative Relations ***/                       
      %else %if &SIMULATION_MIXING eq 2 %then %do;
         %let _final_sparm = varprj.selsparmi varprj.mixsparm;
         %let _final_fparm = varprj.selfparmi varprj.mixfparm; 
         %let MIXED  = true;
         %let SIMMIX = false;
      %end;
      /*** Consortium Only ***/                       
      %else %if &SIMULATION_MIXING = 3 %then %do;
         %let _final_sparm = varprj.selsparmc;
         %let _final_fparm = varprj.selfparmc;  
         %let MIXED  = false;
         %let SIMMIX = false;
      %end;
      /*** Simulation Mixing ***/                       
      %else %if &SIMULATION_MIXING eq 5 %then %do;
         %let _final_sparm = varprj.selsparmi;
         %let _final_fparm = varprj.selfparmi; 
         %let SIMMIX = true;
      %end;

      proc datasets lib=work noprint;
      delete _final_sparm _final_fparm;
      run;
      data _final_sparm; set &_final_sparm;run; 
      data _final_fparm; set &_final_fparm;run; 

      /* Check for nothing to show ------------------------------------------*/
      %let nobs1 = 0;
      %orvu_getObsCount(_final_sparm, nobs1 );
      %let nobs2 = 0;
      %orvu_getObsCount(_final_fparm, nobs2 );
      %if &nobs1 = 0 and &nobs2 = 0 %then %do;                
         %let vc_nodat_found = %sysfunc(sasmsg(&orvrds, no_data_find_error, noquote));  
         %orv_emptyReport(&vc_nodat_found);
         %goto next1;
      %end;
   
      *** Frequency ***;
      %orvu_combine_data_with_ccv_names(dataIn = _final_fparm, dataOut = _final, third_dim=&THIRD_DIM);
      %let vc_exec_sum_graph_title = %sysfunc(sasmsg(&orvrds, vc_exec_sum_graph_title,        noquote));
      %let vc_freqdist_report      = %sysfunc(sasmsg(&orvrds, vc_rep_freq_dist_detail_title,  noquote));
      %let vc_finalmat_freqdist    = %sysfunc(sasmsg(&orvrds, vc_rep_var_mat_freq_dist_title, noquote));
      %let vc_graph_xaxis_label    = %sysfunc(sasmsg(&orvrds, vc_graph_xaxis_label,           noquote));
      
      %if &nreports < 2 %then %do;
         %orvr_cover_page( 
                           outputtype   = &outputtype, 
                           report_title = %sysfunc(sasmsg(&orvrds, vc_final_matrix_title, noquote)),
                           style        = &style
                           );  
      %end;

      footnote1; 
      %orv_clearReportTitles; 
      title2 &titleSpec2 "&vc_finalmat_freqdist";
      title3 &titleSpec4 "&vc_freqdist_report";
      title4 &titleSpec3 "(&project)";
      title5 &titleSpec3 "(&author)";
      title6 &titleSpec5 "&vc_exec_sum_graph_title";
      ods noproctitle;
      ods proclabel = "&vc_finalmat_freqdist";
               
      %orv_varCalcGraphReport(
                                 _final,
                                 (ccvuid_0 ne 0) and (ccvuid_1 ne 0)
                                 %if &THIRD_DIM %then %do; 
                                    and (ccvuid_2 ne 0) 
                                 %end;
                                 ,
                                 CCVVALUE_0,
                                 &MAPNAME0,
                                 MEAN,
                                 &vc_graph_xaxis_label,
                                 HBAR3D
                                 );

      %let vc_nodat_for_graph = %sysfunc(sasmsg(&orvrds, vc_no_data_for_graph_error, noquote));                       

      %if &empty = 0 %then %do;                
         %orv_emptyReport(&vc_nodat_for_graph);
      %end;

      footnote1 &footnoteSpec1;
      title6 &titleSpec5 "%sysfunc(sasmsg(&orvrds, vc_rep_freq_dist_detail_title, noquote))";         
      ods noproctitle;
      ods proclabel = "&vc_finalmat_freqdist";

      proc template;
         define table orvar_final_freq;      
            column 
                  (CCVVALUE_0) 
                  CCVVALUE_1 
                  DISTRIBUTION 
                  MEAN 
                  VARIANCE
                  ;
            header h1;

            define h1;
               text "%sysfunc(sasmsg(&orvrds, vc_fitted_title, noquote))";
               start = MEAN  space = 1 expand = '-' end = VARIANCE;
            end;                         

            define CCVVALUE_0;
               header = "&MAPNAME0" blank_dups = on id;
            end; 

            define CCVVALUE_1;
               header = "&MAPNAME1";
            end;        
    
            define DISTRIBUTION;
               header = "%sysfunc(sasmsg(&orvrds, vc_distribution_title, noquote))";
            end;

            define MEAN;
               header = "%sysfunc(sasmsg(&orvrds, mean_title, noquote))" format=&deffmt;
            end;

            define VARIANCE;
               header = "%sysfunc(sasmsg(&orvrds, var_title, noquote))" format=&deffmt;
            end;
         end;  
      run;
                   
      %if &THIRD_DIM %then %do; 
         %do i = 1 %to &num_ccv2;
            %let name = %scan(&ccv_names2, &i, "#");
            title8 "&MAPNAME2_ORIG = &name";

            data _NULL_;
               set _final;
               where CCVVALUE_2 = "&name";
               file print ods = (template = 'orvar_final_freq');
               put _ods_;
            run; 
         %end;
      %end;
      %else %do;
         data _NULL_; 
            set _final;      
            file print ods = (template = 'orvar_final_freq');
            put _ods_;
         run;
      %end;

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;
         
      *** Severity ***; 
      %orvu_combine_data_with_ccv_names(dataIn = _final_sparm, dataOut = _final, third_dim=&THIRD_DIM);
      %let vc_sevdata_rep       = %sysfunc(sasmsg(&orvrds, vc_sev_data_varcalc_title, noquote));
      %let vc_finalmat_sevdist  = %sysfunc(sasmsg(&orvrds, vc_rep_mat_sev_dist_title, noquote));
      %let vc_graph_xaxis_label = %sysfunc(sasmsg(&orvrds, vc_graph_xaxis_label,      noquote));
      footnote1; 
      %orv_clearReportTitles;
      title2 &titleSpec2 "&vc_finalmat_sevdist";
      title3 &titleSpec4 "&vc_sevdata_rep";
      title4 &titleSpec3 "(&project)";
      title5 &titleSpec3 "(&author)";
      title6 &titleSpec5 "&vc_exec_sum_graph_title";
      ods noproctitle;
      ods proclabel = "&vc_finalmat_sevdist";

      %orv_varCalcGraphReport(
                              _final, 
                              (ccvuid_0 ne 0) and (ccvuid_1 ne 0)
                              %if &THIRD_DIM %then %do; 
                                 and (ccvuid_2 ne 0) 
                              %end;
                              ,
                              CCVVALUE_0,
                              &MAPNAME0,
                              MEAN,
                              &vc_graph_xaxis_label,
                              HBAR3D
                              );

      %if &empty = 0 %then %do; 
         %orv_emptyReport(&vc_nodat_forgraph);
      %end;
      
      %let vc_sevdist_report = %sysfunc(sasmsg(&orvrds, vc_rep_sev_dist_detail_title, noquote));
      footnote1 &footnoteSpec1;
      title6 &titleSpec5 "&vc_sevdist_report";
      ods noproctitle;
      ods proclabel = "&vc_finalmat_sevdist";

      proc template;
         define table orvar_final_sev;
            header h1;       
            column 
                     (CCVVALUE_0) 
                     CCVVALUE_1 
                     DISTRIBUTION 
                     MEAN 
                     STD 
                     KURTOSIS 
                     ;               
            define h1;
               text "%sysfunc(sasmsg(&orvrds, vc_fitted_title, noquote))";
               start = MEAN  space = 1 expand = '-' end = KURTOSIS;
            end;          

            define CCVVALUE_0;
               header = "&MAPNAME0" blank_dups = on id;
            end; 

            define CCVVALUE_1;
               header = "&MAPNAME1";
            end;      

            define DISTRIBUTION;
               header = "%sysfunc(sasmsg(&orvrds, vc_distribution_title, noquote))";
            end;        

            define MEAN;
               header = "%sysfunc(sasmsg(&orvrds, mean_title, noquote))" format = &deffmt;
            end;      

            define STD;
               header = "%sysfunc(sasmsg(&orvrds, mix_std_dev_title, noquote))" format = &deffmt;
            end;  

            *Format changed to best12. since kurtosis can be very large # LH 2/7/07;
            define KURTOSIS;
               header = "%sysfunc(sasmsg(&orvrds, kurt_title, noquote))" format = best12.;
            end;
         end;  
      run;
   
      %if &THIRD_DIM %then %do;
         %do i = 1 %to &num_ccv2;
            %let name = %scan(&ccv_names2, &i, "#");
            title6 "&MAPNAME2_ORIG = &name";

            data _NULL_;
               set _final;
               where CCVVALUE_2 = "&name";
               file print ods=(template='orvar_final_sev');
               put _ods_;
            run; 
         %end;
      %end;
      %else %do;
         data _NULL_;
            set _final;
            file print ods = (template = 'orvar_final_sev');
            put _ods_;
         run;
      %end;

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;

      %next1:

      /*************************  SIMMIX ******************************/
      %if &SIMMIX eq true %then %do;
         data _final_sparm;  
            set varprj.simpriority_sev; 
         run;

         data _final_fparm;  
            set varprj.simpriority_freq; 
         run;

         *** Frequency ***;
         %orvu_combine_data_with_ccv_names(dataIn = _final_fparm, dataOut = _final, third_dim=&THIRD_DIM);
         %let vc_freqsim_mixrep = %sysfunc(sasmsg(&orvrds, vc_freq_sim_mix_varcalc_title,     noquote));
         %let vc_mat_freqsim    = %sysfunc(sasmsg(&orvrds, vc_rep_var_mat_freq_sim_mix_title, noquote));
         footnote1; 
         %orv_clearReportTitles; 
         title2 &titleSpec2 "&vc_mat_freqsim";
         title3 &titleSpec4 "&vc_freqsim_mixrep";
         title4 &titleSpec3 "(&project)";
         title5 &titleSpec3 "(&author)";
         footnote1 &footnoteSpec1;
         %let vc_data_source = %sysfunc(sasmsg(&orvrds, vc_data_source_title, noquote));

         proc template;
            define table orvar_freq_priority;      
               header h1;       
               column (CCVVALUE_0) CCVVALUE_1 SOURCE_NAME;                       

               define CCVVALUE_0;
                  header = "&MAPNAME0" blank_dups = on id;
               end; 

               define CCVVALUE_1;
                  header = "&MAPNAME1";
               end;      

               define SOURCE_NAME;
                  header = "&vc_data_source" ;
               end;
            end;  
         run;
   
         %if &THIRD_DIM %then %do;
            %do i = 1 %to &num_ccv2;
               %let name = %scan(&ccv_names2, &i, "#");
               title6 "&MAPNAME2_ORIG = &name";

               data _NULL_;
                  set _final;
                  where CCVVALUE_2 = "&name";
                  file print ods = (template = 'orvar_freq_priority');
                  put _ods_;
               run; 
            %end;
         %end;
         %else %do;
            data _NULL_; 
               set _final;      
               file print ods = (template = 'orvar_freq_priority');
               put _ods_;
            run;
         %end;

         *** Severity ***; 
         %if &outputtype = PDF %then %do;
            ods PDF startpage = yes;
         %end;

         %orvu_combine_data_with_ccv_names(dataIn = _final_sparm, dataOut = _final, third_dim=&THIRD_DIM);
         %let sevsim_mixchoice = %sysfunc(sasmsg(&orvrds, vc_sev_sim_mix_varcalc_title,     noquote));
         %let finalmat_sevsim  = %sysfunc(sasmsg(&orvrds, vc_rep_var_mat_sev_sim_mix_title, noquote));
         footnote1; 
         %orv_clearReportTitles;
         title2 &titleSpec2 "&finalmat_sevsim";
         title3 &titleSpec4 "&sevsim_mixchoice";
         title4 &titleSpec3 "(&project)";
         title5 &titleSpec3 "(&author)";
         footnote1 &footnoteSpec1;

         proc template;
            define table orvar_sev_priority;
               header h1;       
               column 
                        (CCVVALUE_0) 
                        CCVVALUE_1 
                        SOURCE_NAME 
                        LOWER_BOUND 
                        UPPER_BOUND
                        ;
               define h1;
                  text "%sysfunc(sasmsg(&orvrds, vc_bounds_title, noquote))";
                  start = LOWER_BOUND  space = 1 expand = '-' end = UPPER_BOUND;
               end;          

               define CCVVALUE_0;
                  header = "&MAPNAME0" blank_dups = on id;
               end; 

               define CCVVALUE_1;
                  header = "&MAPNAME1";
               end;      

               define SOURCE_NAME;
                  header = "&vc_data_source" ;
               end;        

               define LOWER_BOUND;
                  header = "%sysfunc(sasmsg(&orvrds, vc_lower_title, noquote))" format = ConfLevel.;
               end;      

               define UPPER_BOUND;
                  header = "%sysfunc(sasmsg(&orvrds, vc_upper_title, noquote))" format = ConfLevel.;
               end;              
            end;  
         run;

         %if &THIRD_DIM %then %do;
            %do i = 1 %to &num_ccv2;
               %let name = %scan(&ccv_names2, &i, "#");
               title6 "&MAPNAME2_ORIG = &name";

               data _NULL_;
                  set _final;
                  where CCVVALUE_2 = "&name";
                  file print ods =(template = 'orvar_sev_priority');
                  put _ods_;
               run; 
            %end;
         %end;
         %else %do;
            data _NULL_;
               set _final;
               file print ods = (template = 'orvar_sev_priority');
               put _ods_;
            run;
         %end;
      %end;  /* Else Do */

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;
   %end;
   /****************************************************/
   /* END VaR Final Matrix *****************************/
   /****************************************************/

   /****************************************************/
   /* VaR Calculation Criteria *************************/
   /****************************************************/
   %let vc_optsim_rep = %sysfunc(sasmsg(&orvrds, vc_opt_sim_title, noquote));

   %if %index(&options, VCRIT) ^= 0 %then %do;
      %if &nreports < 2 %then %do;
         %orvr_cover_page( 
                           outputtype   = &outputtype, 
                           report_title = %sysfunc(sasmsg(&orvrds, vc_rep_var_crit_title, noquote)),
                           style        = &style
                           );  
      %end;

      footnote1; 
      %orv_clearReportTitles;   
      title2 &titleSpec2 "%sysfunc(sasmsg(&orvrds, vc_rep_var_crit_title, noquote))" ;
      title3 &titleSpec4 "&vc_optsim_rep";
      title4 &titleSpec3 "(&project)";
      title5 &titleSpec3 "(&author)";
      footnote1 &footnoteSpec1 ;
      ods noproctitle;
      ods proclabel = "%sysfunc(sasmsg(&orvrds, vc_rep_var_crit_title, noquote))";
         
      proc report 
                  data = crit 
                  style(column) = [font_size = 2] 
                  nowd
                  ;
         column 
               ( 
                  "%sysfunc(sasmsg(&orvrds, vc_conf_level_title, noquote))"  
                  CONFIDENCE_LEVEL_1 
                  CONFIDENCE_LEVEL_2 
                  CONFIDENCE_LEVEL_3
                  )
                  SIM_INTERVAL 
                  ITERATIONS 
                  SEED 
                  MAX_NUM_HOSTS      
                  %if &DFM_TYPE eq 2 %then %do; 
                     DEFAULT_MAXLOSS_RATIO 
                  %end;
                  %else %if &DFM_TYPE eq 1 %then %do; 
                     DEFAULT_MAX_LOSS 
                  %end; 
                  %else %do; 
                     DEFAULT_MAXLOSS_USER_DEF 
                  %end; 
                  MIXING_SELECTION 
                  CORRELATION_TYPE 
                  ;
      
         define CONFIDENCE_LEVEL_1 / "1" format = ConfLevel.;
         define CONFIDENCE_LEVEL_2 / "2" format = ConfLevel.;
         define CONFIDENCE_LEVEL_3 / "3" format = ConfLevel.;
         define ITERATIONS / "%sysfunc(sasmsg(&orvrds, vc_iter_title, noquote))";
         define SEED / "%sysfunc(sasmsg(&orvrds, vc_rand_seed_title, noquote))";
         define MIXING_SELECTION / "%sysfunc(sasmsg(&orvrds, vc_mix_sel_title, noquote))" format = VarMixingSelection.;
         define CORRELATION_TYPE / "%sysfunc(sasmsg(&orvrds, vc_freq_corr_title, noquote))" format = $CorrType.;
         define SIM_INTERVAL / "%sysfunc(sasmsg(&orvrds, vc_interval_title, noquote))" format = $IntervalType.;
         define MAX_NUM_HOSTS / "%sysfunc(sasmsg(&orvrds, vc_parahost_title, noquote))";

         %if &DFM_TYPE eq 2 %then %do; 
            define DEFAULT_MAXLOSS_RATIO / "%sysfunc(sasmsg(&orvrds, vc_worst_case_title, noquote))";
         %end;
         %else %if &DFM_TYPE eq 1 %then %do; 
            define DEFAULT_MAX_LOSS / "%sysfunc(sasmsg(&orvrds, vc_def_max_loss_title, noquote))" format = &curfmt;
         %end;
         %else %do; 
            define DEFAULT_MAXLOSS_USER_DEF / "%sysfunc(sasmsg(&orvrds, vc_def_max_loss_title, noquote))" format = &curfmt;
         %end;
      run;
             
      %if &SIMULATION_MIXING eq 5 %then %do;
         /* get macro var for simmiz ranges */
         data _NULL_;
            set varprj.SIMMIX_RANGES(where = (MODEL_UID = &model_uid));
            call symput("NUM_RANGES", NUM_RANGES);
         run;

         data _NULL_;
            set varprj.SIMMIX_RANGES(where = (MODEL_UID = &model_uid));

            %do i = 1 %to &NUM_RANGES;
               call symput("LOWER_&i", LOWER_&i);
               call symput("UPPER_&i", UPPER_&i);
            %end;
         run;

         /* add better range names to simmix_priority dataset */
         data _priority0; 
            set varprj.SIMMIX_PRIORITY(where = (MODEL_UID = &model_uid));
            length RANGE_NAME $100;

            %do i = 1 %to &NUM_RANGES;
               if RANGE_NUM = &i then do;
                  RANGE_NAME = "&&LOWER_&i. - &&UPPER_&i.%";
               end;
            %end;
         run;

         /* add Scenario names */
         proc sql noprint;
            CREATE TABLE 
                           _priority as
            SELECT 
                           p.*, 
                           n.SCENARIO_NAME 
            FROM 
                           _priority0 AS p 
            LEFT JOIN 
                           &varorg..SCENARIO AS n          
            ON 
                           p.SCENARIO_UID eq n.SCENARIO_UID
                           ; 
         quit;      

         /* combine with ccv names */
         %orvu_combine_data_with_ccv_names(dataIn = _priority, dataOut = _priority0, defaults = true, third_dim=&THIRD_DIM);

         proc sort data = _priority0; 
            by 
               %if &THIRD_DIM %then %do; 
                  CCVUID_2 
               %end; 
               RANGE_NUM PRIORITY; 
         run;

         %if &THIRD_DIM %then %do;
            data _priority0; 
               set _priority0;
               rename CCVVALUE_2 = &mapname0;
            run;
         %end;

         /*** severity simmix priorities ***/
         %if &outputtype = PDF %then %do;
            ods PDF startpage = yes;
         %end;

         %let vc_priority        = %sysfunc(sasmsg(&orvrds, vc_priority_title,                    noquote));
         %let vc_range           = %sysfunc(sasmsg(&orvrds, vc_range_title,                       noquote));
         %let vc_sevsim_priority = %sysfunc(sasmsg(&orvrds, vc_sev_sim_mix_pri_for_varcalc_title, noquote));
         %let vc_simpri_sev      = %sysfunc(sasmsg(&orvrds, vc_sim_mix_pri_sev_title,             noquote));
         footnote1; 
         %orv_clearReportTitles;   
         title2 &titleSpec2 "&vc_simpri_sev" ;
         title3 &titleSpec4 "&vc_sevsim_priority";
         title4 &titleSpec3 "(&project)";
         title5 &titleSpec3 "(&author)";      
         footnote1 &footnoteSpec1 ;

         proc report data= _priority0 style(column)=[font_size=2] nowd;
           %if &THIRD_DIM %then %do;
               by &mapname0 NOTSORTED;
            %end;

            where RANGE_NUM ne 0;
            column 
                  (
                     "%sysfunc(sasmsg(&orvrds, class_title, noquote))" 
                     CCVVALUE_0 
                     CCVVALUE_1
                     )
                     RANGE_NAME 
                     PRIORITY 
                     DATA_TYPE 
                     SCENARIO_NAME
                     ;
            define CCVVALUE_0 / "&MAPNAME0";
            define CCVVALUE_1 / "&MAPNAME1";
            define RANGE_NAME / "&vc_range";
            define PRIORITY / "&vc_priority";
            define DATA_TYPE / "&vc_data_source" format = $DataType.;
            define SCENARIO_NAME / "%sysfunc(sasmsg(&orvrds, name_title, noquote))";
         run;

         /*** frequency simmix priorities ***/
         %if &outputtype = PDF %then %do;
            ods PDF startpage = yes;
         %end;

         %let vc_freqsim_priority = %sysfunc(sasmsg(&orvrds, vc_freq_sim_mix_pri_for_varcalc_title, noquote));
         %let vc_simpri_freq      = %sysfunc(sasmsg(&orvrds, vc_sim_mix_pri_freq_title,             noquote));
         footnote1; 
         %orv_clearReportTitles;   
         title2 &titleSpec2 "&vc_simpri_freq" ;
         title3 &titleSpec4 "&vc_freqsim_priority";
         title4 &titleSpec3 "(&project)";
         title5 &titleSpec3 "(&author)";
         footnote1 &footnoteSpec1 ;

         proc report 
                     data = _priority0 
                     style(column) = [font_size = 2] 
                     nowd
                     ;
            %if &THIRD_DIM %then %do;
               by &mapname0 NOTSORTED;
            %end;

            where RANGE_NUM eq 0;
            column 
                  (
                     "%sysfunc(sasmsg(&orvrds, class_title, noquote))" 
                     CCVVALUE_0 
                     CCVVALUE_1
                     )
                     PRIORITY 
                     DATA_TYPE 
                     SCENARIO_NAME
                     ;
            define CCVVALUE_0 / "&MAPNAME0";
            define CCVVALUE_1 / "&MAPNAME1";
            define PRIORITY / "%sysfunc(sasmsg(&orvrds, vc_priority_title, noquote))";
            define DATA_TYPE / "&vc_data_source" format = $DataType.;
            define SCENARIO_NAME / "%sysfunc(sasmsg(&orvrds, name_title, noquote))";
         run;
      %end;

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;
   %end;

   /****************************************************/
   /* VaR Frequency Effects ****************************/
   /****************************************************/
   %if %index(&options, VFEFFECTS) ^= 0 %then %do;
      %let vc_simfreq_effectrep = %sysfunc(sasmsg(&orvrds, vc_rep_sim_freq_eff_title, noquote));
      %let vc_freqeffect_rep    = %sysfunc(sasmsg(&orvrds, vc_freq_eff_sim_title,     noquote));

      %if &nreports < 2 %then %do;
         %orvr_cover_page( 
                           outputtype = &outputtype, 
                           report_title = &vc_simfreq_effectrep,
                           style = &style
                           );  
      %end;

      footnote1; 
      %orv_clearReportTitles;  
      title2 &titleSpec2 "&vc_simfreq_effectrep" ;
      title3 &titleSpec4 "&vc_freqeffect_rep";
      title4 &titleSpec3 "(&project)";
      title5 &titleSpec3 "(&author)";
      footnote1 &footnoteSpec1;
      ods noproctitle;
      ods proclabel = "&vc_simfreq_effectrep";

      /*** get appropriate CCVs ***/
      %orvu_combine_data_with_ccv_names(dataIn = VARPRJ.feffects, dataOut = _feffects, third_dim=&THIRD_DIM);

      proc sort data = _feffects; 
         by CCUID; 
      run;

      /*** create Freq Effects Template ***/
      proc template;
         define table var_feffects;
            header ccHeader;
            column 
                  (CCVVALUE_0) 
                  CCVVALUE_1 
                  DISTRIBUTION 
                  FPARAM_1 
                  FPARAM_2
                  ;
            define ccHeader;
               text "%sysfunc(sasmsg(&orvrds, class_title, noquote))"; 
               start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
            end;                   

            define CCVVALUE_0;
               header = "&MAPNAME0" blank_dups = on id;
            end; 

            define CCVVALUE_1;
               header = "&MAPNAME1";
            end; 

            %let vc_freqdist_header = %sysfunc(sasmsg(&orvrds, vc_freq_dist_title, noquote));
            define DISTRIBUTION; 
               header = "&vc_freqdist_header"; 
            end;

            define FPARAM_1; 
               header = "%sysfunc(sasmsg(&orvrds, vc_param1_title, noquote))"; 
            end;

            define FPARAM_2; 
               header = "%sysfunc(sasmsg(&orvrds, vc_param2_title, noquote))"; 
            end;
         end;  
      run;
   
      /*** print report ***/
      %if &THIRD_DIM %then %do;
         %do i = 1 %to &num_ccv2;
            %let name = %scan(&ccv_names2, &i, "#");
            title6 "&MAPNAME2_ORIG = &name";

            data _NULL_; 
               set _feffects;
               where ccvvalue_2 = "&name";    
               file print ods = (template = 'var_feffects');
               put _ods_;
            run;    
         %end;
      %end;
      %else %do;
         data _NULL_; 
           set _feffects;    
           file print ods = (template = 'var_feffects');
           put _ods_;
         run;
      %end;

      %orvu_delete_data( data = _feffects);

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;
   %end;

   /****************************************************/
   /* VaR Result Matrices ******************************/
   /*******************************************************/
   /* LH 1/23/07 - Separating tables for 99.9%, 99% &95%, */
   /*  so that one entire table fits on one page          */
   /*******************************************************/
   %if 
      %index(&options, VRES_INSURANCE) or 
      %index(&options, VRESBASEL_INSURANCE) or
      %index(&options, VRESBASEL_WITHOUT) or 
      %index(&options, VRES_WITHOUT) or
      %index(&options, VRESFITTED_WITHOUT) or
      %index(&options, VRESFITTED_INSURANCE) %then %do; 
      /* Check for non-missing parameter risk numbers --------------------*/
      %let parameter_risk = 0;

      data _NULL_;set varprj.varstat_insured; 
         if param_risk_lower_1 ne . then do;
            call symput("parameter_risk", trim(1));
         end;
      run;

      /*** create VaR Report Template ***/
      /* For simulation based confidence intervals ----------------*/
      proc template;
         %do i = 1 %to 3;
            define table var_results&i. ;
               header ccHeader confHeader&i. ;
               column 
                     (CCVVALUE_0) 
                     CCVVALUE_1 
                     LCL&i. 
                     P_&i. 
                     UCL&i. 
                     MEAN
                     ;
               define ccHeader;
                  text "%sysfunc(sasmsg(&orvrds, class_title, noquote))"; 
                  start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
               end;                   

               define CCVVALUE_0;
                  header = "&MAPNAME0" blank_dups=on id;
               end; 

               define CCVVALUE_1;
                  header = "&MAPNAME1";
               end; 

               define confHeader&i.;
                  text "&&CONF&i";
                  start = LCL&i space = 1 expand = '-' end = UCL&i;
               end;    

               define LCL&i;  
                  header= "%sysfunc(sasmsg(&orvrds, vc_lowercl_title, noquote))"       format = &curfmt;     
               end; 

               define P_&i;   
                  header = "%sysfunc(sasmsg(&orvrds, vc_quant_var_title, noquote))"   format = &curfmt;     
               end; 

               define UCL&i;  
                  header = "%sysfunc(sasmsg(&orvrds, vc_uppercl_title, noquote))"       format = &curfmt;     
               end; 

               define MEAN; 
                  header="%sysfunc(sasmsg(&orvrds, mean_title, noquote))"  format = &curfmt;  
               end;              
            end;   
         %end;    
      run;
         
      /*** create VaR Report Template ***/
      /* For Parameter risk based confidence intervals ----------------*/
      proc template;
         %do i = 1 %to 3;
            define table var_results_pr&i. ;
               header ccHeader confHeader&i. ;
               column 
                     (CCVVALUE_0) 
                     CCVVALUE_1 
                     param_risk_lower_&i. 
                     P_&i. param_risk_upper_&i. 
                     MEAN
                     ;

               define ccHeader;
                  text "%sysfunc(sasmsg(&orvrds, class_title, noquote))"; 
                  start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
               end;                   

               define CCVVALUE_0;
                  header = "&MAPNAME0" blank_dups = on id;
               end; 

               define CCVVALUE_1;
                  header = "&MAPNAME1";
               end; 

               define confHeader&i.;
                  text "&&CONF&i";
                  start = param_risk_lower_&i space = 1 expand = '-' end = param_risk_upper_&i;
               end;    

               define param_risk_lower_&i;  
                  header = "%sysfunc(sasmsg(&orvrds, vc_lowerpr_title, noquote))"       format = &curfmt;     
               end; 

               define P_&i;   
                  header = "%sysfunc(sasmsg(&orvrds, vc_quant_var_title, noquote))"   format = &curfmt;     
               end; 

               define param_risk_upper_&i;  
                  header = "%sysfunc(sasmsg(&orvrds, vc_upperpr_title, noquote))"       format = &curfmt;     
               end; 

               define MEAN; 
                  header = "%sysfunc(sasmsg(&orvrds, mean_title, noquote))"  format = &curfmt;  
               end;              
            end;   
         %end;    
      run;
      
      %let vc_sasmeth_agg      = %sysfunc(sasmsg(&orvrds, vc_sas_meth_for_agg_cells_title,  noquote));
      %let vc_results_withins  = %sysfunc(sasmsg(&orvrds, vc_var_result_with_insure_title,  noquote));
      %let vc_varesult_withins = %sysfunc(sasmsg(&orvrds, vc_var_results_with_insure_title, noquote));

      %if %index(&options, VRES_INSURANCE) %then %do;
         %let THE_DATA   = varprj.varstat_insured;
         %let TITLETEXT1 = &vc_varesult_withins;
         %let TITLETEXT3 = &vc_results_withins;
         %let AGGDATA    = varprj.aggstat_insured;
         %let TITLETEXT4 = &vc_sasmeth_agg;

         %if %eval(&parameter_risk = 1) %then %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1    , 
                                    &TITLETEXT3    , 
                                    &TITLETEXT4    , 
                                    var_results_pr , 
                                    &THE_DATA      , 
                                    &AGGDATA
                                    );
         %end;
         %else %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1 , 
                                    &TITLETEXT3 , 
                                    &TITLETEXT4 , 
                                    var_results , 
                                    &THE_DATA   , 
                                    &AGGDATA
                                    );
         %end;
      %end;
   
      %let vc_stdmeth_agg = %sysfunc(sasmsg(&orvrds, vc_std_meth_agg_cells_title, noquote));

      %if %index(&options, VRESBASEL_INSURANCE) %then %do;
         %let THE_DATA   = varprj.varstat_insured;
         %let TITLETEXT1 = &vc_varesult_withins;
         %let TITLETEXT3 = &vc_results_withins;
         %let AGGDATA    = varprj.aggstatbasel_insured;
         %let TITLETEXT4 = &vc_stdmeth_agg;

         %if %eval(&parameter_risk = 1) %then %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1    , 
                                    &TITLETEXT3    , 
                                    &TITLETEXT4    , 
                                    var_results_pr , 
                                    &THE_DATA      , 
                                    &AGGDATA
                                    );
         %end;
         %else %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1 , 
                                    &TITLETEXT3 , 
                                    &TITLETEXT4 , 
                                    var_results , 
                                    &THE_DATA   , 
                                    &AGGDATA
                                    );
         %end;
      %end;
   
      %let vc_results_woins  = %sysfunc(sasmsg(&orvrds, vc_var_result_wo_insure_title,  noquote));
      %let vc_varesult_woins = %sysfunc(sasmsg(&orvrds, vc_var_results_wo_insure_title, noquote));

      %if %index(&options, VRESBASEL_WITHOUT) %then %do;
         %let THE_DATA   = varprj.varstat_naked;
         %let TITLETEXT1 = &vc_varesult_woins;
         %let TITLETEXT3 = &vc_results_woins;
         %let AGGDATA    = varprj.aggstatbasel_naked;
         %let TITLETEXT4 = &vc_stdmeth_agg;

         %if %eval(&parameter_risk = 1) %then %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1    , 
                                    &TITLETEXT3    , 
                                    &TITLETEXT4    , 
                                    var_results_pr , 
                                    &THE_DATA      , 
                                    &AGGDATA
                                    );
         %end;
         %else %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1 , 
                                    &TITLETEXT3 , 
                                    &TITLETEXT4 , 
                                    var_results , 
                                    &THE_DATA   , 
                                    &AGGDATA
                                    );
         %end; 
      %end;
   
      %if %index(&options, VRES_WITHOUT) %then %do;
         %let THE_DATA   = varprj.varstat_naked;
         %let TITLETEXT1 = &vc_varesult_woins;
         %let TITLETEXT3 = &vc_results_woins;
         %let AGGDATA    = varprj.aggstat_naked;
         %let TITLETEXT4 = &vc_sasmeth_agg;

         %if %eval(&parameter_risk = 1) %then %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1    , 
                                    &TITLETEXT3    , 
                                    &TITLETEXT4    , 
                                    var_results_pr , 
                                    &THE_DATA      , 
                                    &AGGDATA
                                    );
         %end;
         %else %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1 ,
                                    &TITLETEXT3 , 
                                    &TITLETEXT4 , 
                                    var_results , 
                                    &THE_DATA   , 
                                    &AGGDATA
                                    );
         %end;
      %end;        
   
      %if %index(&options, VRESFITTED_WITHOUT) %then %do;
         %let vc_fit_mod_agg_cells  = %sysfunc(sasmsg(&orvrds, vc_fit_model_for_agg_cells_title,       noquote));
         %let vc_fitted_woins       = %sysfunc(sasmsg(&orvrds, vc_var_fit_tot_wo_insure_title,         noquote));
         %let vc_fitted_woins_title = %sysfunc(sasmsg(&orvrds, vc_var_fit_tot_results_wo_insure_title, noquote));

         %let THE_DATA   = varprj.varstat_naked;
         %let TITLETEXT1 = &vc_fitted_woins_title;
         %let TITLETEXT3 = &vc_fitted_woins;
         %let AGGDATA    = varprj.aggstat_naked;
         %let TITLETEXT4 = &vc_fit_mod_agg_cells;

         %if %eval(&parameter_risk = 1) %then %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1    , 
                                    &TITLETEXT3    , 
                                    &TITLETEXT4    , 
                                    var_results_pr , 
                                    &THE_DATA      , 
                                    &AGGDATA       , 
                                    fitted         = true
                                    );
         %end;
         %else %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1 , 
                                    &TITLETEXT3 , 
                                    &TITLETEXT4 , 
                                    var_results , 
                                    &THE_DATA   , 
                                    &AGGDATA    , 
                                    fitted      = true
                                    );
         %end;
      %end;        

      %let vc_fitted_withins       = %sysfunc(sasmsg(&orvrds, vc_var_fit_tot_with_insure_title,         noquote));
      %let vc_fitted_withins_title = %sysfunc(sasmsg(&orvrds, vc_var_fit_tot_results_with_insure_title, noquote));

      %if %index(&options, VRESFITTED_INSURANCE) %then %do;
         %let THE_DATA   = varprj.varstat_insured;
         %let TITLETEXT1 = &vc_fitted_withins_title;
         %let TITLETEXT3 = &vc_fitted_withins;
         %let AGGDATA    = varprj.aggstat_insured;
         %let TITLETEXT4 = %sysfunc(sasmsg(&orvrds, vc_fit_model_for_agg_cells_title, noquote));

         %if %eval(&parameter_risk = 1) %then %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1    , 
                                    &TITLETEXT3    , 
                                    &TITLETEXT4    , 
                                    var_results_pr , 
                                    &THE_DATA      , 
                                    &AGGDATA       , 
                                    fitted         = true
                                    );
         %end;
         %else %do;
            %_orv_varResultsReport(
                                    &TITLETEXT1 , 
                                    &TITLETEXT3 , 
                                    &TITLETEXT4 , 
                                    var_results , 
                                    &THE_DATA   , 
                                    &AGGDATA    , 
                                    fitted      = true
                                    );
         %end;
      %end;             
   %end;

   %if %index(&options, VRES_DETAIL) ^= 0 %then %do;
     %ORVR_var_details(
                        fileref = &fileref,
                        model_uid    = &model_uid,
                        options      = &options,
                        outputType   = &outputtype,
                        style        = &style
                        );
   %end;
   /****************************************************/
   /* END of VaR Result Matrices ***********************/
   /****************************************************/

   /*- delete unwanted data sets -*/
   %orvu_delete_data( data =
                              ResultInsurance0 
                              ResultInsurance1 
                              ResultInsurance 
                              Result0 
                              Result1 
                              Result 
                              _work_ 
                              _fake_ 
                              var_user
                              Crit 
                              Final
                              );

   /*- general cleanup: reset title, output etc -*/   
   %orvr_report_cleanup();
%mend ORVR_var_calc;

/***********************************************************************************/
/******************************** Report Macro #1 **********************************/
/***********************************************************************************/
%macro _orv_varResultsReport(
                              TITLETEXT1 , 
                              TITLETEXT3 , 
                              TITLETEXT4 , 
                              TEMPLATE   , 
                              THE_DATA   ,  
                              AGGDATA    ,
                              fitted     = false
                              );

   %if &nreports < 2 %then %do;
      %orvr_cover_page( 
                        outputtype   = &outputtype, 
                        report_title = &titletext1,
                        style        = &style
                        ); 
   %end;

   footnote1; 
   %orv_clearReportTitles;
   title1 &titleSpec1 "&TITLETEXT1";
   title2 &titleSpec2 "&TITLETEXT3";
   title3 &titleSpec4 "&TITLETEXT4";
   title4 &titleSpec3 "(&project)";
   title5 &titleSpec3 "(&author)";

   %let nobs = 0;
   %let dsid = %sysfunc(open(&THE_DATA, I));
   %let vc_exec_sum_graph_title = %sysfunc(sasmsg(&orvrds, vc_exec_sum_graph_title, noquote));
   %let vc_graph_xaxis_label    = %sysfunc(sasmsg(&orvrds, vc_graph_xaxis_label,    noquote));

   %if &dsid ^= 0 %then %do;
      %let nobs = %sysfunc(attrn(&dsid,nlobs));
      %let rc   = %sysfunc(close(&dsid));
   %end;

   %let vc_novar_results = %sysfunc(sasmsg(&orvrds, vc_no_var_results_error, noquote));

   %if &nobs eq 0 %then %do;
      ods noproctitle;
      ods proclabel = "&TITLETEXT1";
      %orv_emptyReport(&vc_novar_results);
   %end;
   %else %do;
      /*** combine cells with aggregates ***/
      /*** get Cross-classification names ***/
      %orvu_combine_data_with_ccv_names(dataIn = &THE_DATA, dataOut = VarData0, third_dim=&THIRD_DIM);
      %orvu_combine_data_with_ccv_names(dataIn = &AGGDATA,  dataOut = aggdata0, third_dim=&THIRD_DIM);

      /* Remove the fitted totals -----------*/
      %if &fitted = false %then %do;
         data VarData0; 
            set VarData0;
            where 
                  STATUS = 0 
               and 
                  ccvuid_0 ne 0 
               and 
                  ccvuid_1 ne 0
                  %if &THIRD_DIM %then %do; 
                     and ccvuid_2 ne 0 
                  %end;
                  ;
         run; 

         /* Combine with aggregated totals -------*/
         data ResultInsurance;
            set 
                  vardata0 
                  aggdata0 
                  ;
         run;
      %end;
      %else %do;   /* Use only fitted totals ------------*/
         data ResultInsurance; 
            set VarData0;
            where STATUS = 0 ;
         run;  
      %end;
       
      proc sort data = ResultInsurance; 
         by CCUID; 
      run;
       
      /*** graphical summary ***/
      title6 &titleSpec5 "&vc_exec_sum_graph_title";
      ods noproctitle;
      ods proclabel = "&vc_exec_sum_graph_title";

      %orv_varCalcGraphReport(
                              ResultInsurance,
                              (ccvuid_0 ne 0) and (ccvuid_1 ne 0)
                              %if &THIRD_DIM %then %do; 
                                 and (ccvuid_2 ne 0) 
                              %end;
                              ,
                              CCVVALUE_0,
                              &MAPNAME0,
                              MEAN,
                              &vc_graph_xaxis_label,
                              HBAR3d
                              );

      %let vc_nodat_forgraph_check = %sysfunc(sasmsg(&orvrds, vc_no_dat_graph_check_error, noquote));

      %if &empty = 0 %then %do; 
         %orv_emptyReport(&vc_nodat_forgraph_check);
      %end;

      %if &outputtype = PDF %then %do;
         ods PDF startpage = yes;
      %end;

      /*** detailed report ***/
      %do i = 1 %to 3;
         title2 &titleSpec2 "&TITLETEXT1" " &&CONF&i";
         title3 &titleSpec4 "&TITLETEXT4";
         title4 &titleSpec3 "(&project)";
         title5 &titleSpec3 "(&author)";
         ods noproctitle;
         ods proclabel = "&TITLETEXT1";
         footnote1 &footnoteSpec1;

         %if &THIRD_DIM %then %do;
            %do j = 1 %to &num_ccv2;
               %let name = %scan(&ccv_names2, &j, "#");
               title6 "&MAPNAME2_ORIG = &name";

               data _NULL_; 
                  set ResultInsurance;
                  where ccvvalue_2 = "&name";    
                  file print ods = (template = "var_results&i.");
                  put _ods_;
               run;    
            %end;
         %end;
         %else %do;
            data _NULL_; 
               set ResultInsurance; 
               file print ods = (template = "var_results&i.");
               put _ods_;
            run;
         %end;
      %end; 
   %end;

   %if &outputtype = PDF %then %do;
      ods PDF startpage = yes;
   %end;
%mend _orv_varResultsReport;

/***********************************************************************************/
/******************************** Graph Macro #1 ***********************************/
/***********************************************************************************/
%macro orv_varCalcGraphReport(
                                 indata,
                                 where,
                                 ccvar,
                                 ccvarlabel,
                                 sumvar,
                                 sumvarlabel,
                                 charttype
                                 );
   /* Check to see if dataset is empty */
   %global empty;
   %let empty = 0;                       

   data _NULL_;
      set &indata (where = (&where));
      call symput ("empty", 1);
   run;

   /* Graph size increased from 4in to 6in, too small before  LH 2/7/07*/ 

   goptions 
            device = java 
            hsize  = 6in 
            vsize  = 6in
            ;
   legend1 mode = share cshadow = gray;
   axis2 label = ("&sumvarlabel");
   axis1 label = ("&ccvarlabel");

   proc GCHART data = &indata;
      label &ccvar="&ccvarlabel";
      &charttype  
                  &ccvar /
                  SUMVAR = &sumvar
                  %if %substr(&charttype, 2, 3) = BAR %then %do;
                     Descending
                     PATTERNID = MIDPOINT
                     shape = cylinder
                     CREF = BLACK
                     AUTOREF
                     raxis = axis2
                  %end;
                  %if %substr(&charttype, 1, 3) = PIE %then %do;
                     value = arrow
                     slice = arrow
                  %end;
                  legend =legend1
                  ;
      where &where;
   run;
   quit;
%mend orv_varCalcGraphReport;
